Package: diffuseR
Title: Functional Interface to Diffusion Models in R
Version: 0.2.2
Authors@R: c(
    person("Troy", "Hernandez", email = "troy@cornball.ai", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0005-4248-604X")),
    person("cornball.ai", role = "cph"),
    person("The HuggingFace Team", role = "cph",
           comment = "portions ported from the diffusers library (Apache-2.0); see inst/COPYRIGHTS"),
    person("Lightricks Ltd.", role = "cph",
           comment = "LTX checkpoint layout and pipeline constants; see inst/COPYRIGHTS"))
Description: A native R implementation of diffusion models providing a
    functional interface to state-of-the-art generative AI. Inspired by the
    'Python' library 'diffusers' from 'Hugging Face'
    <https://huggingface.co/>, 'diffuseR' generates and manipulates images
    from text prompts using models such as 'Stable Diffusion', with no
    'Python' dependency. Supports multiple diffusion schedulers and device
    acceleration.
License: Apache License (>= 2)
URL: https://github.com/cornball-ai/diffuseR
BugReports: https://github.com/cornball-ai/diffuseR/issues
Encoding: UTF-8
Imports: torch, jsonlite, grid, png, jpeg
Suggests: av, hfhub, safetensors, simplermarkdown, tinytest
VignetteBuilder: simplermarkdown
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-08-03 06:24:31 UTC; troy
Author: Troy Hernandez [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-4248-604X>),
  cornball.ai [cph],
  The HuggingFace Team [cph] (portions ported from the diffusers library
    (Apache-2.0); see inst/COPYRIGHTS),
  Lightricks Ltd. [cph] (LTX checkpoint layout and pipeline constants;
    see inst/COPYRIGHTS)
Maintainer: Troy Hernandez <troy@cornball.ai>
Repository: CRAN
Date/Publication: 2026-08-08 12:40:02 UTC

# tinyrox says don't edit this manually, but it can't stop you!

export(auto_devices)
export(bpe_tokenizer)
export(clear_vram)
export(clip_pooled_output)
export(CLIPTokenizer)
export(convert_sd21_pt_to_diffusers)
export(ddim_scheduler_create)
export(ddim_scheduler_step)
export(decode_bpe)
export(download_component)
export(download_flux1)
export(download_flux2_klein)
export(download_ltx2)
export(download_model)
export(download_sd21)
export(download_sdxl)
export(download_zimage_turbo)
export(encode_bpe)
export(encode_qwen)
export(encode_unigram)
export(encode_with_gemma3)
export(encode_with_qwen3)
export(encode_with_t5)
export(filename_from_prompt)
export(flowmatch_calculate_shift)
export(flowmatch_scale_noise)
export(flowmatch_scheduler_create)
export(flowmatch_scheduler_step)
export(flowmatch_set_timesteps)
export(flux_ada_layer_norm_continuous)
export(flux_ada_layer_norm_zero)
export(flux_ada_layer_norm_zero_single)
export(flux_apply_rotary_emb)
export(flux_attention)
export(flux_double_block)
export(flux_is_quant_key)
export(flux_load_pipeline)
export(flux_load_transformer)
export(flux_memory_profile)
export(flux_open_checkpoint)
export(flux_open_quantized)
export(flux_pack_latents)
export(flux_pos_embed)
export(flux_prepare_latent_image_ids)
export(flux_quantize)
export(flux_single_block)
export(flux_transformer)
export(flux_unpack_latents)
export(flux2_bn_normalize)
export(flux2_double_block)
export(flux2_empirical_mu)
export(flux2_feed_forward)
export(flux2_is_quant_key)
export(flux2_load_pipeline)
export(flux2_modulation)
export(flux2_pack_latents)
export(flux2_parallel_self_attention)
export(flux2_patchify_latents)
export(flux2_prepare_latent_ids)
export(flux2_prepare_text_ids)
export(flux2_single_block)
export(flux2_transformer)
export(flux2_unpack_latents_with_ids)
export(flux2_unpatchify_latents)
export(flux2_vae_decoder)
export(gemma3_config_ltx2)
export(gemma3_encode_batch)
export(gemma3_quantize_nf4)
export(gemma3_text_model)
export(gemma3_tokenizer)
export(img2img)
export(is_blackwell_gpu)
export(latents_to_video)
export(load_decoder_safetensors)
export(load_decoder_weights)
export(load_flux2_vae_decoder)
export(load_gemma3_nf4)
export(load_gemma3_text_encoder)
export(load_model_component)
export(load_pipeline)
export(load_qwen3_text_encoder)
export(load_t5_text_encoder)
export(load_text_encoder_safetensors)
export(load_text_encoder_weights)
export(load_text_encoder2_safetensors)
export(load_text_encoder2_weights)
export(load_to_gpu)
export(load_unet_safetensors)
export(load_unet_sdxl_safetensors)
export(load_unet_sdxl_weights)
export(load_unet_weights)
export(ltx23_ada_layer_norm_single)
export(ltx23_adain_filter_latent)
export(ltx23_antialias_act1d)
export(ltx23_apply_interleaved_rotary_emb)
export(ltx23_apply_split_rotary_emb)
export(ltx23_attention)
export(ltx23_audio_causal_conv2d)
export(ltx23_audio_decoder)
export(ltx23_audio_downsample)
export(ltx23_audio_encoder)
export(ltx23_audio_mel_frontend)
export(ltx23_audio_resnet_block)
export(ltx23_audio_upsample)
export(ltx23_audio_vae)
export(ltx23_causal_conv3d)
export(ltx23_census)
export(ltx23_connector_transformer_1d)
export(ltx23_denormalize_latents)
export(ltx23_distilled_sigmas)
export(ltx23_downsample1d)
export(ltx23_encode_audio)
export(ltx23_encode_video_frames)
export(ltx23_feed_forward)
export(ltx23_fp8_linear)
export(ltx23_get_timestep_embedding)
export(ltx23_is_fp8_cast_key)
export(ltx23_kaiser_sinc_filter1d)
export(ltx23_latent_upsampler)
export(ltx23_load_group)
export(ltx23_load_pipeline)
export(ltx23_load_transformer_fp8)
export(ltx23_load_transformer_nf4)
export(ltx23_load_upsampler)
export(ltx23_map_audio_vae_key)
export(ltx23_map_connector_key)
export(ltx23_map_dit_key)
export(ltx23_map_vae_key)
export(ltx23_map_vocoder_key)
export(ltx23_mel_stft)
export(ltx23_memory_profile)
export(ltx23_nf4_dequantize)
export(ltx23_nf4_linear)
export(ltx23_nf4_quantize)
export(ltx23_normalize_latents)
export(ltx23_open_checkpoint)
export(ltx23_open_fp8_checkpoint)
export(ltx23_per_channel_rms_norm)
export(ltx23_per_token_rms_norm)
export(ltx23_prepare_conditioned_latents)
export(ltx23_preprocess_frames)
export(ltx23_quantize_fp8)
export(ltx23_quantize_nf4)
export(ltx23_read_audio)
export(ltx23_read_tail_frames)
export(ltx23_release_dequant_buffers)
export(ltx23_rms_norm)
export(ltx23_rotary_pos_embed)
export(ltx23_rotary_pos_embed_1d)
export(ltx23_set_attn_chunk)
export(ltx23_snake_beta)
export(ltx23_split_keys)
export(ltx23_stage2_distilled_sigmas)
export(ltx23_tail_latents)
export(ltx23_text_connectors)
export(ltx23_tone_map_latents)
export(ltx23_transformer)
export(ltx23_transformer_block)
export(ltx23_tune_gc)
export(ltx23_upsample1d)
export(ltx23_video_decoder3d)
export(ltx23_video_down_block3d)
export(ltx23_video_downsampler3d)
export(ltx23_video_encoder3d)
export(ltx23_video_mid_block3d)
export(ltx23_video_resnet_block3d)
export(ltx23_video_up_block3d)
export(ltx23_video_upsampler3d)
export(ltx23_video_vae)
export(ltx23_vocoder)
export(ltx23_vocoder_resblock)
export(ltx23_vocoder_with_bwe)
export(models2devices)
export(offload_to_cpu)
export(post_quant_conv)
export(preprocess_image)
export(quant_conv)
export(qwen_bpe_tokenizer)
export(qwen3_encoder)
export(recommend)
export(reshard_safetensors)
export(resident_activate)
export(resident_deactivate)
export(resident_generate)
export(resident_load)
export(resident_status)
export(resident_unload)
export(save_frames)
export(save_image)
export(save_video)
export(save_video_ltx23)
export(scheduler_add_noise)
export(sd_pipeline_from_safetensors)
export(sdxl_memory_profile)
export(sdxl_pipeline_from_safetensors)
export(serve)
export(t5_encoder)
export(text_encoder_native)
export(text_encoder_native_from_safetensors)
export(text_encoder2_native)
export(text_encoder2_native_from_safetensors)
export(tokenize_gemma3)
export(txt2img)
export(txt2img_flux)
export(txt2img_flux2)
export(txt2img_sd21)
export(txt2img_sdxl)
export(txt2img_zimage)
export(txt2vid_ltx2)
export(unet_native)
export(unet_native_from_safetensors)
export(unet_native_from_torchscript)
export(unet_sdxl_native)
export(unet_sdxl_native_from_safetensors)
export(unet_sdxl_native_from_torchscript)
export(unigram_tokenizer)
export(vae_decoder_native)
export(vae_decoder_native_from_safetensors)
export(vocab_size)
export(vram_report)
export(write_wav)
export(zimage_block)
export(zimage_cap_pos_ids)
export(zimage_feed_forward)
export(zimage_final_layer)
export(zimage_img_pos_ids)
export(zimage_is_quant_key)
export(zimage_load_pipeline)
export(zimage_patchify)
export(zimage_pos_embed)
export(zimage_t_embedder)
export(zimage_transformer)
export(zimage_unpatchify)

S3method(print,bpe_tokenizer)
S3method(print,diffuseR_resident)
S3method(print,ltx23_checkpoint)
S3method(print,qwen_tokenizer)
S3method(print,unigram_tokenizer)

importFrom(utils,head)

# diffuseR

[![CRAN status](https://www.r-pkg.org/badges/version/diffuseR)](https://CRAN.R-project.org/package=diffuseR)
[![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)

## Overview

`diffuseR` is a functional R implementation of diffusion models, inspired by Hugging Face's Python `diffusers` library. The package provides a simple, idiomatic R interface to state-of-the-art generative AI models for image generation and manipulation using base R and the `torch` package. No Python dependencies. Currently supports Windows and Linux cpu and cuda devices.

## Example output

![](man/figures/sdxl_vs_flux2_radio_studio.jpg)

## Installation

First install torch. As per [this comment](https://github.com/mlverse/torch/issues/1198#issuecomment-2419363312), using the pre-built binaries from [https://torch.mlverse.org/docs/articles/installation#pre-built](https://torch.mlverse.org/docs/articles/installation#pre-built) is heavily recommend. "The pre-built binaries bundle the necessary CUDA and cudnn versions, so you don't need a global compatible system version of CUDA":

```r
options(timeout = 600) # increasing timeout is recommended since we will be downloading a 2GB file.
# For Windows and Linux: "cpu", "cu128" are the only currently supported
# For MacOS the supported are: "cpu-intel" or "cpu-m1"
kind <- "cu128"
version <- available.packages()["torch","Version"]
options(repos = c(
  torch = sprintf("https://torch-cdn.mlverse.org/packages/%s/%s/", kind, version),
  CRAN = "https://cloud.r-project.org" # or any other from which you want to install the other R dependencies.
))
install.packages("torch")
```

You can install the development version of diffuseR from GitHub:

```r
# install.packages("remotes")
remotes::install_github("cornball-ai/diffuseR")
```

## Features

- **Text-to-Image Generation**: Stable Diffusion 2.1, SDXL, FLUX.1-schnell, FLUX.2 Klein, and Z-Image-Turbo (fully native R torch implementations)
- **Text-to-Video Generation**: LTX-2.3 with synchronized audio
- **Image-to-Image Generation**: Modify existing images based on text prompts (SD 2.1 / SDXL)
- **GPU-poor support**: NF4 and fp8 quantization run the 12B FLUX.1 and 22B LTX-2.3 transformers on a 16GB card
- **Scheduler Options**: DDIM and FlowMatch Euler (static and dynamic shifting)
- **Device Support**: CPU and CUDA GPUs (including Blackwell RTX 50xx)
- **R-native Interface**: Functional programming approach that feels natural in R
- **HTTP serving**: `serve()` exposes any model over OpenAI-style endpoints
  (images and video) from one persistent process - see `?serve`

## Hardware requirements

Approximate, per model, from the floor configuration (lowest precision,
CPU-only) to the best-quality tier. `recommend("model")` picks the right
configuration for your machine automatically; the `vignette("performance-levers")`
explains the levers.

| model | minimum (floor quality) | best quality |
|---|---|---|
| SD 2.1 | 4 GB RAM, no GPU | 6 GB VRAM + 4 GB RAM (fp16) |
| SDXL | 10 GB RAM, no GPU | 12 GB VRAM + 8 GB RAM (fp16) |
| FLUX.2 Klein | 12 GB RAM, no GPU | 16 GB VRAM + 20 GB RAM (bf16) |
| Z-Image-Turbo | 14 GB RAM, no GPU | 18 GB VRAM + 24 GB RAM (bf16) |
| FLUX.1-schnell | 28 GB RAM, no GPU | 24 GB VRAM + 45 GB RAM (bf16) |
| LTX-2.3 video | 40 GB RAM, no GPU | 16 GB VRAM + 45 GB RAM (fp8, streamed) |

## Quick Start

### Basic Usage

**Warning**: The first time you run the code below, it will download ~5.3GB of Stable Diffusion 2.1 CPU-only model files [from Hugging Face](https://huggingface.co/datasets/cornball-ai/sd21-R/tree/main) and load them into memory. Ensure you have enough RAM, disk space, and a stable internet connection. Memory management with deep learning models is crucial, so consider using a machine with sufficient resources; ~8GB of free RAM is recommended for running Stable Diffusion 2.1 on CPU only.

```r
options(timeout = 600) # increasing timeout is recommended since we will be downloading a 3.5GB file.
library(diffuseR)
torch::local_no_grad()

# Generate an image from text
cat_img <- txt2img(
  prompt = "a photorealistic cat wearing sunglasses",
  model = "sd21", # Specify the model to use, e.g., "sd21" for Stable Diffusion 2.1
  download_models = TRUE, # Automatically download the model if not already present
  steps = 30,
  seed = 42,
  filename = "cat.png",
)

# Clear out pipeline to free up GPU memory
pipeline <- NULL
torch::cuda_empty_cache()
```


### Advanced Usage with GPU

The unet is the most computationally-intensive part of the model, so it is recommended to run it on a GPU if possible. The decoder and text encoder can be run on CPU if you have limited GPU memory. SDXL's unet requires a minimum of 6GB of GPU memory (VRAM), while Stable Diffusion 2.1 requires a minimum of 2GB.

```r
# Increasing timeout is recommended since we will be downloading 5.1 and 2.8GB model files, among others.
options(timeout = 1200) 

library(diffuseR)
torch::local_no_grad() # Prevents torch from tracking gradients, which is not needed for inference

# Assign the various deep learning models to devices
model_name = "sdxl"
devices = list(unet = "cuda", decoder = "cpu",
               text_encoder = "cpu", encoder = "cpu")

m2d <- models2devices(model_name = model_name, devices = devices,
                      unet_dtype_str = "float16", download_models = TRUE)

pipeline <- load_pipeline(model_name = model_name, m2d = m2d, i2i = TRUE,
                          unet_dtype_str = "float16")

# Generate an image from text
cat_img <- txt2img(
  prompt = "a photorealistic cat wearing sunglasses",
  model_name = model_name,
  devices = devices,
  pipeline = pipeline,
  num_inference_steps = 30,
  guidance_scale = 7.5,
  seed = 42,
  filename = "cat2.png",
)

gambling_cat <- img2img(
  input_image = "cat2.png",
  prompt = "a photorealistic cat throwing dice",
  img_dim = 1024,
  model_name = model_name,
  devices = devices,
  pipeline = pipeline,
  num_inference_steps = 30,
  strength = 0.75,
  guidance_scale = 7.5,
  seed = 42,
  filename = "gambling_cat.png"
)

# Clear out pipeline to free up GPU memory
pipeline <- NULL
torch::cuda_empty_cache()
```
![](cat2.png)
![](gambling_cat.png)


### FLUX and Z-Image

FLUX.1-schnell (12B), FLUX.2 Klein (4B), and Z-Image-Turbo (6B) are
step-distilled models: 4-8 denoising steps, no guidance. All are
quantized locally once at download time and fit comfortably on a 16GB
GPU (measured 1024x1024 on an RTX 5060 Ti: FLUX.1 ~55s at 9.6GB peak;
FLUX.2 Klein ~13s at 12.5GB; Z-Image-Turbo ~24s at 13.1GB).

```r
library(diffuseR)

# FLUX.1-schnell: the HuggingFace repo is gated (Apache-2.0 weights,
# license click-through). Accept the license and set HF_TOKEN first.
# ~34GB download, one-time NF4 quantize to a 6.8GB artifact.
download_flux1()
txt2img_flux("An astronaut riding a horse on Mars, photorealistic",
             seed = 7)

# FLUX.2 Klein 4B: ungated. ~16GB download, one-time fp8 quantize to
# a 3.9GB artifact.
download_flux2_klein()
txt2img_flux2("a red fox sitting in a snowy forest, digital art",
              seed = 42)

# Z-Image-Turbo: ungated, strong at legible text in images (EN + CN).
# ~33GB download, one-time fp8 quantize to a 5.9GB artifact.
download_zimage_turbo()
txt2img_zimage(paste("A storefront with a large wooden sign that reads",
                     "\"DIFFUSER\" in bold carved letters"), seed = 42)

### Text-to-Video: LTX-2.3

LTX-2.3 (22B, distilled) generates video with synchronized audio in 8
steps. NF4-quantized it renders 768x512x49 in ~44s warm on an RTX
5060 Ti; audio-driven talking-head generation and clip continuation
are supported (see ?txt2vid_ltx2).

```r
download_ltx2()       # ~46GB download, one-time fp8 quantize
txt2vid_ltx2("A river winding through a misty forest at dawn",
             pipeline = ltx23_load_pipeline(),
             filename = "river.mp4")
```

### Serving over HTTP

```r
# One persistent process, OpenAI-style endpoints, never downloads:
serve(model = "flux2", port = 7812L, token = "my-secret")
# curl -H "Authorization: Bearer my-secret" \
#   -d '{"prompt":"a lighthouse at dusk","size":"512x512"}' \
#   http://localhost:7812/v1/images/generations
```

# Or through the common dispatcher
txt2img("a lighthouse at dusk", model_name = "flux2")
```

### Precision and safetensors

The quantized transformers ship as **nf4** by default (packed uint8 +
float32 blocks). nf4 loads on **stock CRAN safetensors** because the
artifacts are written in sub-2 GB shards; R's safetensors overflows a
32-bit offset on any single file at or above 2^31 bytes (~2.15 GB), so
shard size, not the dtype, is what gates readability.

Higher-quality tiers behave as follows:

- **bf16** (24 GB+ cards) is also CRAN-readable.
- **fp8** (the 12-16 GB sweet spot) needs the
  [`cornball-ai/safetensors`](https://github.com/cornball-ai/safetensors)
  fork until float8 support lands on CRAN (mlverse/safetensors#13).

`recommend(model)` picks the right tier for your VRAM and the
safetensors you have installed, and asking for a tier your safetensors
cannot read falls back to nf4 with a note rather than erroring:

```r
recommend("flux2")            # e.g. list(precision = "fp8", ...) on a 16 GB card
recommend("flux1")$note       # the fork suggestion, when fp8/bf16 would fit but can't load
```

## Supported Models

Currently supported models:

- Stable Diffusion 2.1
- Stable Diffusion XL (SDXL)
- FLUX.1-schnell (12B, 4-step distilled)
- FLUX.2 Klein 4B (4-step distilled)
- Z-Image-Turbo (6B, 8-step distilled, text rendering)
- LTX-2.3 Video (with audio)

### Choosing an image model: SDXL vs FLUX.2

Same prompt, same seed, 1024x1024, measured on an RTX 5060 Ti 16GB:

> "A retro 1970s radio station studio interior, warm wood panels,
> vinyl records on the walls, soft neon glow, detailed illustration"

| Model | Settings | Load | Warm generation | Peak VRAM |
|---|---|---|---|---|
| SDXL | 50 steps, CFG 7.5 | 45 s | 20 s | **12.7 GB** |
| FLUX.2 Klein 4B | 4 steps, guidance-free | 32 s | **13 s** | 12.5 GB |

FLUX.2 is now faster per image (since the allocator gc-gate fix), and
its prompt adherence and coherence are in a different class — SDXL
melts the speaker stacks and turns the wall records into neon smears,
while FLUX.2 draws the studio you asked for (SDXL left, FLUX.2 right):

![SDXL vs FLUX.2, same prompt and seed](man/figures/sdxl_vs_flux2_radio_studio.jpg)

Rule of thumb: reach for FLUX.2. Note FLUX.2 Klein is guidance-free,
so negative prompts do not apply. When the image needs legible text
(signs, posters, labels), reach for Z-Image-Turbo instead.

### Downloading Models

Model downloads are explicit and consent-gated: call the model's `download_*()` function once (interactive sessions are prompted with the size; non-interactive sessions require `options(diffuseR.consent = TRUE)`). For gated models (like FLUX.1-schnell), you need to:

1. Create a HuggingFace account at https://huggingface.co
2. Accept the model's license agreement (visit the model page and click "Agree")
3. Create an access token at https://huggingface.co/settings/tokens
4. Add to your `~/.Renviron`:
   ```
   HF_TOKEN=hf_your_token_here
   ```

**Manual download with hfhub:**

```r
# Install hfhub with HF_TOKEN fix (until PR merged upstream)
remotes::install_github("cornball-ai/hfhub@fix-gated-repos")

# Download a model
library(hfhub)
path <- hub_download("google/gemma-3-12b-it", "config.json")
# Or download entire model:
# path <- hub_snapshot("stabilityai/stable-diffusion-2-1")
```

## Roadmap

Future plans for diffuseR include:

- [ ] Inpainting support
- [ ] Additional schedulers (PNDM, DPMSolverMultistep, Euler ancestral)
- [ ] FLUX.2 reference-image conditioning and img2img
- [x] text-to-video generation (LTX-2.3)

## How It Works

diffuseR supports two execution modes:

**Native R torch (recommended for SDXL)**: Pure R implementations of VAE decoder, text encoders, and UNet. Required for Blackwell GPUs (RTX 50xx series) and recommended for best compatibility. Enable with `use_native_*` flags or use `txt2img_sdxl()` which defaults to native.

**TorchScript (legacy)**: Pre-exported models from PyTorch. Still available for SD21 and older GPUs. Scripts to build TorchScript files are at [diffuseR-TS](https://github.com/cornball-ai/diffuseR-TS/).

No Python dependencies required for either mode.

## Contributing

Contributions are welcome! Please feel free to submit a Pull Request.

## License

This project is licensed under the Apache 2. License - see the LICENSE file for details.

## Acknowledgments

- Hugging Face for the original diffusers library
- Stability AI for Stable Diffusion
- The R and torch communities for their excellent tooling

#' Qwen2 Byte-Level BPE Tokenizer
#'
#' Pure R implementation of the Qwen2 tokenizer (HuggingFace
#' tokenizer.json, BPE model with ByteLevel pre-tokenization), as used by
#' FLUX.2 klein's Qwen3 text encoder. Text is split with the GPT-4-style
#' regex, each pre-token's UTF-8 bytes are mapped through the GPT-2
#' byte-to-unicode table, and rank-based BPE merges produce the ids.
#' Added tokens (\code{<|im_start|>}, \code{<think>}, ...) are split out
#' literally before byte-level encoding.
#'
#' Limitation: the NFC normalizer is not applied (base R has no NFC);
#' input is assumed to already be NFC, which holds for ordinary text.
#'
#' @name tokenizer_qwen
NULL

# GPT-2 byte-to-unicode table: printable bytes map to themselves,
# everything else to 256+n. Returns a character vector indexed by
# byte value + 1.
.qwen_byte_table <- function() {
    bs <- c(33:126, 161:172, 174:255)
    cs <- bs
    n <- 0L
    for (b in 0:255) {
        if (!(b %in% bs)) {
            bs <- c(bs, b)
            cs <- c(cs, 256L + n)
            n <- n + 1L
        }
    }
    out <- character(256)
    out[bs + 1L] <- vapply(cs, intToUtf8, character(1))
    out
}

#' Load a Qwen2 byte-level BPE tokenizer
#'
#' @param tokenizer_path Path to a tokenizer.json (or a directory
#'   containing one).
#'
#' @return A \code{qwen_tokenizer} object.
#'
#' @export
qwen_bpe_tokenizer <- function(tokenizer_path) {
    path <- path.expand(tokenizer_path)
    if (dir.exists(path)) {
        path <- file.path(path, "tokenizer.json")
    }
    if (!file.exists(path)) {
        stop("tokenizer.json not found: ", path)
    }

    tj <- jsonlite::fromJSON(path, simplifyVector = TRUE)
    model <- tj$model
    if (!identical(model$type, "BPE")) {
        stop("Expected a BPE tokenizer, got: ", model$type %||% "none")
    }

    # Integer-id BPE with no persistent string maps: R gc cost scales
    # with live object count, and a 151k-binding vocab environment made
    # every gc ~17x slower (measured 12 -> 203 ms), which multiplied
    # into minutes across the allocator-triggered gcs of a generation.
    # All lookups compile to three atomic vectors + a 256-int byte table.
    vocab <- unlist(model$vocab) # named int: piece -> id (0-based)
    piece_names <- names(vocab)

    merges <- model$merges
    if (is.matrix(merges)) {
        a <- merges[, 1]
        b <- merges[, 2]
    } else {
        sp <- regexpr(" ", merges, fixed = TRUE)
        a <- substr(merges, 1L, sp - 1L)
        b <- substr(merges, sp + 1L, nchar(merges))
    }
    id_a <- unname(vocab[match(a, piece_names)])
    id_b <- unname(vocab[match(b, piece_names)])
    id_r <- unname(vocab[match(paste0(a, b), piece_names)])
    ok <- !is.na(id_a) & !is.na(id_b) & !is.na(id_r)
    key <- id_a[ok] * 2097152 + id_b[ok] # ids < 2^21: exact in doubles
    rank <- seq_along(id_a)[ok]
    result <- id_r[ok]
    ord <- order(key)

    # Initial ids for the 256 byte-unicode characters
    byte_chars <- .qwen_byte_table()
    byte_ids <- unname(vocab[match(byte_chars, piece_names)])

    # Pre-tokenization split regex (GPT-4 style)
    split_regex <- NULL
    pres <- tj$pre_tokenizer$pretokenizers
    if (!is.null(pres) && "pattern" %in% names(pres)) {
        split_regex <- pres$pattern$Regex[!is.na(pres$pattern$Regex)][1]
    }
    if (is.null(split_regex) || !nzchar(split_regex)) {
        stop("No Split pre-tokenizer regex found in ", path)
    }

    added <- tj$added_tokens
    added_env <- list2env(stats::setNames(as.list(added$id), added$content),
                          parent = emptyenv())

    structure(
              list(
                   merge_key = key[ord],
                   merge_rank = rank[ord],
                   merge_result = result[ord],
                   byte_ids = byte_ids,
                   n_pieces = length(vocab),
                   split_regex = split_regex,
                   added = added_env,
                   added_contents = added$content[order(-nchar(added$content))],
                   pad_id = get0("<|endoftext|>", envir = added_env,
                                 ifnotfound = 151643L),
                   path = path
        ),
              class = "qwen_tokenizer"
    )
}

#' @export
print.qwen_tokenizer <- function(x, ...) {
    cat("<qwen_tokenizer>\n")
    cat("  vocab:  ", x$n_pieces, "+", length(ls(x$added)), "added\n")
    cat("  path:   ", x$path, "\n")
    invisible(x)
}

# Rank-based BPE merge loop over token ids: adjacent-pair keys are
# looked up in the sorted merge table via findInterval (C binary
# search), vectorized across all pairs per iteration
.qwen_bpe_merge_ids <- function(ids, tokenizer) {
    mk <- tokenizer$merge_key
    while (length(ids) > 1L) {
        keys <- ids[-length(ids)] * 2097152 + ids[-1L]
        pos <- findInterval(keys, mk)
        hit <- pos > 0L
        hit[hit] <- mk[pos[hit]] == keys[hit]
        if (!any(hit)) {
            break
        }
        ranks <- rep(Inf, length(keys))
        ranks[hit] <- tokenizer$merge_rank[pos[hit]]
        i <- which.min(ranks)
        ids[i] <- tokenizer$merge_result[pos[i]]
        ids <- ids[-(i + 1L)]
    }
    ids
}

# Encode one plain-text segment (no added tokens inside)
.qwen_encode_segment <- function(text, tokenizer) {
    if (!nzchar(text)) {
        return(integer(0))
    }
    text <- enc2utf8(text)
    m <- gregexpr(tokenizer$split_regex, text, perl = TRUE)[[1]]
    pre_tokens <- regmatches(text, list(m))[[1]]

    out <- lapply(pre_tokens, function(tok) {
        bytes <- as.integer(charToRaw(tok))
        .qwen_bpe_merge_ids(tokenizer$byte_ids[bytes + 1L], tokenizer)
    })
    as.integer(unlist(out))
}

# Split text on added-token literals (longest first), returning a list
# of list(text=, id=) chunks
.qwen_split_added <- function(text, tokenizer) {
    chunks <- list(list(text = text, id = NA_integer_))
    for (content in tokenizer$added_contents) {
        out <- list()
        for (chunk in chunks) {
            if (!is.na(chunk$id) || !grepl(content, chunk$text, fixed = TRUE)) {
                out[[length(out) + 1L]] <- chunk
                next
            }
            parts <- strsplit(chunk$text, content, fixed = TRUE)[[1]]
            # strsplit drops trailing separators; recover the layout
            n_seps <- lengths(regmatches(chunk$text,
                    gregexpr(content, chunk$text, fixed = TRUE)))
            if (length(parts) == 0L) {
                parts <- ""
            }
            for (i in seq_along(parts)) {
                if (nzchar(parts[i])) {
                    out[[length(out) + 1L]] <- list(text = parts[i],
                        id = NA_integer_)
                }
                if (i <= n_seps) {
                    out[[length(out) + 1L]] <- list(
                        text = content,
                        id = get0(content, envir = tokenizer$added))
                }
            }
        }
        chunks <- out
    }
    chunks
}

#' Encode prompts with the Qwen tokenizer
#'
#' With \code{chat_template = TRUE} each prompt is wrapped as a single
#' user turn with the generation prompt, matching
#' \code{apply_chat_template(..., add_generation_prompt = TRUE)}. With
#' \code{enable_thinking = FALSE} (the FLUX.2 klein pipeline behavior)
#' the template closes with an empty thinking block; with
#' \code{enable_thinking = TRUE} (the Z-Image pipeline behavior) it ends
#' at the assistant turn. Right-pads with \code{<|endoftext|>}.
#'
#' @param tokenizer A \code{\link{qwen_bpe_tokenizer}}.
#' @param texts Character vector of prompts.
#' @param max_length Integer. Fixed sequence length (klein: 512). NULL
#'   for no truncation/padding.
#' @param chat_template Logical. Wrap in the Qwen3 chat template.
#' @param enable_thinking Logical. Leave the model's thinking enabled
#'   (no empty think block). Default FALSE.
#'
#' @return List with \code{input_ids} and \code{attention_mask} integer
#'   matrices [length(texts), max_length] (ragged lists when
#'   \code{max_length} is NULL). Ids are 0-based.
#'
#' @export
encode_qwen <- function(tokenizer, texts, max_length = 512L,
                        chat_template = TRUE, enable_thinking = FALSE) {
    stopifnot(inherits(tokenizer, "qwen_tokenizer"))

    encode_one <- function(text) {
        if (chat_template) {
            text <- paste0("<|im_start|>user\n", text, "<|im_end|>\n",
                           "<|im_start|>assistant\n",
                if (!enable_thinking) "<think>\n\n</think>\n\n")
        }
        ids <- integer(0)
        for (chunk in .qwen_split_added(text, tokenizer)) {
            if (!is.na(chunk$id)) {
                ids <- c(ids, as.integer(chunk$id))
            } else {
                ids <- c(ids, .qwen_encode_segment(chunk$text, tokenizer))
            }
        }
        ids
    }

    all_ids <- lapply(as.character(texts), encode_one)

    if (is.null(max_length)) {
        return(list(
                    input_ids = all_ids,
                    attention_mask = lapply(all_ids, function(x) rep(1L, length(x)))
            ))
    }

    all_ids <- lapply(all_ids, function(ids) {
        if (length(ids) > max_length) ids[seq_len(max_length)] else ids
    })
    n <- length(all_ids)
    input_ids <- matrix(as.integer(tokenizer$pad_id), nrow = n,
                        ncol = max_length)
    attention_mask <- matrix(0L, nrow = n, ncol = max_length)
    for (i in seq_len(n)) {
        len <- length(all_ids[[i]])
        if (len > 0L) {
            input_ids[i, seq_len(len)] <- all_ids[[i]]
            attention_mask[i, seq_len(len)] <- 1L
        }
    }
    list(input_ids = input_ids, attention_mask = attention_mask)
}

# Exact token-id parity for the Qwen2 byte-level BPE tokenizer against
# HuggingFace reference cases (generated by
# tools/gen_qwen_tokenizer_cases.py, checked in). The 11 MB
# tokenizer.json is not shipped; the test locates one and skips if
# absent.

library(diffuseR)

find_qwen_tokenizer <- function() {
  p <- Sys.getenv("DIFFUSER_QWEN_TOKENIZER", "")
  if (nzchar(p) && file.exists(p)) {
    return(p)
  }
  if (requireNamespace("hfhub", quietly = TRUE)) {
    p <- tryCatch(
      suppressMessages(hfhub::hub_download(
        "black-forest-labs/FLUX.2-klein-4B",
        "tokenizer/tokenizer.json", local_files_only = TRUE
      )),
      error = function(e) ""
    )
    if (nzchar(p) && file.exists(p)) {
      return(p)
    }
  }
  p <- "../../tools/cache/tokenizer_qwen.json"
  if (file.exists(p)) {
    return(p)
  }
  ""
}

tok_path <- find_qwen_tokenizer()
if (!nzchar(tok_path)) exit_file("no Qwen tokenizer.json available")

cases_path <- system.file("tinytest", "fixtures", "qwen_tokenizer_cases.json",
  package = "diffuseR")
if (cases_path == "") cases_path <- "fixtures/qwen_tokenizer_cases.json"
if (!file.exists(cases_path)) exit_file("qwen tokenizer cases missing")

cases <- jsonlite::fromJSON(cases_path, simplifyVector = FALSE)

tok <- qwen_bpe_tokenizer(tok_path)
expect_equal(tok$pad_id, 151643L)

# --- raw encodings (no template, no padding) --------------------------------------

for (case in cases$cases) {
  expected <- as.integer(unlist(case$ids))
  got <- encode_qwen(tok, case$text, max_length = NULL,
    chat_template = FALSE)$input_ids[[1]]
  expect_equal(got, expected,
    info = sprintf("text: %s", substr(case$text, 1, 60)))
}

# --- chat template + padding/truncation ----------------------------------------------

for (case in cases$templated) {
  got <- encode_qwen(tok, case$text, max_length = case$max_length,
    chat_template = TRUE)
  expect_equal(as.integer(got$input_ids[1, ]), as.integer(unlist(case$ids)),
    info = sprintf("templated: %s", substr(case$text, 1, 60)))
  expect_equal(as.integer(got$attention_mask[1, ]),
    as.integer(unlist(case$mask)))
}

# --- batch shape ------------------------------------------------------------------------

batch <- encode_qwen(tok, c("a photo of a cat", "hello"), max_length = 32L)
expect_equal(dim(batch$input_ids), c(2L, 32L))
expect_equal(dim(batch$attention_mask), c(2L, 32L))

# Z-Image Qwen3 delta: the enable_thinking=TRUE chat template against
# shipped-tokenizer renders (tools/gen_zimage_qwen_template_cases.py) and
# the hidden_states[-2] + mask-slice convention against a tiny reference
# model (tools/gen_fixtures_zimage_qwen.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# --- hidden_states[-2] + mask slicing (tiny reference model) ----------------------

fixture_path <- system.file("tinytest", "fixtures", "zimage_qwen.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/zimage_qwen.safetensors"
if (!file.exists(fixture_path)) exit_file("zimage qwen fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

sd <- fx[grep("^sd\\.", names(fx))]
names(sd) <- paste0("model.", sub("^sd\\.", "", names(sd)))

enc <- qwen3_encoder(
  vocab_size = 128L, hidden_size = 32L, intermediate_size = 64L,
  num_hidden_layers = 4L, num_attention_heads = 4L,
  num_key_value_heads = 2L, head_dim = 8L, rope_theta = 1e6,
  rms_norm_eps = 1e-6
)
dests <- c(enc$named_parameters(), enc$named_buffers())
# The final norm never runs for intermediate hidden states; every
# checkpoint key must still land
expect_true(all(names(sd) %in% names(dests)))
torch::with_no_grad({
  for (name in names(sd)) dests[[name]]$copy_(sd[[name]])
})
enc$eval()

ids <- torch::torch_tensor(
  matrix(as.integer(as.array(fx$input_ids)) + 1L, nrow = 1),
  dtype = torch::torch_long()
)
mask <- fx$attention_mask$to(dtype = torch::torch_long())

# hidden_states[-2] of a 4-layer model = state after layer 3
states <- torch::with_no_grad(enc(ids, attention_mask = mask, out_layers = 3L))
expect_true(max_abs_diff(states[[1]], fx$penult) < 1e-5)

# Mask slice to the variable-length caption (right padding -> first n)
n_real <- as.integer(sum(as.array(fx$attention_mask)))
expect_equal(n_real, 9L)
sliced <- states[[1]][1, 1:n_real, ]
expect_true(max_abs_diff(sliced, fx$penult_sliced) < 1e-5)

# --- chat template with enable_thinking = TRUE -------------------------------------

find_qwen_tokenizer <- function() {
  p <- Sys.getenv("DIFFUSER_QWEN_TOKENIZER", "")
  if (nzchar(p) && file.exists(p)) {
    return(p)
  }
  if (requireNamespace("hfhub", quietly = TRUE)) {
    for (repo in c(
      "Tongyi-MAI/Z-Image-Turbo",
      "black-forest-labs/FLUX.2-klein-4B"
    )) {
      p <- tryCatch(
        suppressMessages(hfhub::hub_download(
          repo, "tokenizer/tokenizer.json", local_files_only = TRUE
        )),
        error = function(e) ""
      )
      if (nzchar(p) && file.exists(p)) {
        return(p)
      }
    }
  }
  p <- "../../tools/cache/tokenizer_qwen.json"
  if (file.exists(p)) {
    return(p)
  }
  ""
}

tok_path <- find_qwen_tokenizer()
if (!nzchar(tok_path)) exit_file("no Qwen tokenizer.json available")

cases_path <- system.file("tinytest", "fixtures", "zimage_template_cases.json",
  package = "diffuseR")
if (cases_path == "") cases_path <- "fixtures/zimage_template_cases.json"
if (!file.exists(cases_path)) exit_file("zimage template cases missing")

cases <- jsonlite::fromJSON(cases_path, simplifyVector = FALSE)

expect_equal(cases$meta$padding_side, "right")
expect_equal(cases$meta$pad_token_id, 151643L)

tok <- qwen_bpe_tokenizer(tok_path)

for (case in cases$templated) {
  got <- encode_qwen(tok, case$text, max_length = case$max_length,
    chat_template = TRUE, enable_thinking = TRUE)
  expect_equal(got$input_ids[1, ], as.integer(unlist(case$ids)),
    info = sprintf("ids for: %s", substr(case$text, 1, 40)))
  expect_equal(got$attention_mask[1, ], as.integer(unlist(case$mask)),
    info = sprintf("mask for: %s", substr(case$text, 1, 40)))
}

#' Create a DDIM Scheduler
#'
#' Creates a Denoising Diffusion Implicit Models (DDIM) scheduler for use with
#' diffusion models. DDIM schedulers provide a deterministic sampling process
#' that offers faster inference compared to DDPM while maintaining high quality
#' outputs.
#'
#' @param num_train_timesteps Integer. The number of diffusion steps used to
#'   train the model. Default: 1000
#' @param num_inference_steps Integer. The number of diffusion steps used for
#'   inference. Fewer steps typically means faster inference at the cost of
#'   sample quality. Default: 50
#' @param eta Numeric. Controls the amount of
#'   stochasticity. When eta=0, the sampling process is deterministic.
#'   When eta=1, the sampling process is equivalent to DDPM. Default: 0
#' @param beta_schedule Character. The beta schedule to use. Options are:
#'   \describe{
#'     \item{"linear"}{Linear beta schedule from beta_start to beta_end}
#'     \item{"scaled_linear"}{Scaled linear schedule, generally gives better results}
#'     \item{"cosine"}{Cosine schedule that approaches zero smoothly}
#'   }
#'   Default: "linear"
#' @param beta_start Numeric. The starting value for the beta schedule. Default: 0.00085
#' @param beta_end Numeric. The final value for the beta schedule. Default: 0.012
#' @param rescale_betas_zero_snr Logical. If TRUE, rescales the beta values
#' @param dtype The data type to use for computations. Default is torch_float32(). Options are torch_float16() and torch_float32().
#' @param device The device to use for computations. Options are torch_device("cpu"), torch_device("cuda").
#'
#' @return A DDIM scheduler object that can be used with diffusion models to
#'   generate samples.
#'
#' @details
#' DDIM (Denoising Diffusion Implicit Models) was introduced by Song et al. (2020)
#' as an extension to DDPM (Denoising Diffusion Probabilistic Models). It offers
#' a deterministic sampling process and allows for controlling the number of
#' inference steps independently from the training process.
#'
#' The scheduler contains the noise schedule and methods for computing
#' alpha, beta, and other parameters used in the diffusion process.
#'
#' @references
#' Song, J., Meng, C., & Ermon, S. (2020).
#' "Denoising Diffusion Implicit Models."
#' \url{https://arxiv.org/abs/2010.02502}
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   scheduler <- ddim_scheduler_create(
#'     num_train_timesteps = 1000,
#'     num_inference_steps = 5,
#'     eta = 0.5,
#'     beta_schedule = "scaled_linear"
#'   )
#'   scheduler$timesteps
#' }
#' @export
ddim_scheduler_create <- function(num_train_timesteps = 1000,
                                  num_inference_steps = 50, eta = 0,
                                  beta_schedule = c("linear", "scaled_linear", "cosine"),
                                  beta_start = 0.00085, beta_end = 0.012,
                                  rescale_betas_zero_snr = FALSE,
                                  dtype = torch::torch_float32(),
                                  device = torch::torch_device("cpu")) {
    beta_schedule <- match.arg(beta_schedule)
    betas <- switch(beta_schedule,
                    "linear" = seq(beta_start, beta_end, length.out = num_train_timesteps),
                    "scaled_linear" = seq(sqrt(beta_start), sqrt(beta_end),
            length.out = num_train_timesteps) ^ 2,
                    "cosine" = NULL)
    if (rescale_betas_zero_snr) {
        betas <- rescale_zero_terminal_snr(betas)
    }
    betas <- torch::torch_tensor(betas, dtype = dtype, device = device)
    alphas <- 1 - betas
    alphas <- torch::torch_tensor(alphas, dtype = dtype, device = device)
    alphas_cumprod <- torch::torch_cumprod(alphas, dim = 1, dtype = dtype)
    alphas_cumprod_prev <- torch::torch_cat(list(torch::torch_ones(1, device = device),
            alphas_cumprod[1:(length(alphas_cumprod) - 1)]))

    step_ratio <- num_train_timesteps %/% num_inference_steps
    # Could add 2 here instead of 1 and remove it from timestep_index and prev_timestep_index below
    timesteps <- as.integer(rev(round((0:(num_inference_steps - 1)) * step_ratio) + 1))
    # timesteps <- as.integer(seq(num_train_timesteps, 0, length.out = num_inference_steps + 1)[-1] + 1)

    list(
         betas = betas,
         alphas = alphas,
         alphas_cumprod = alphas_cumprod,
         alphas_cumprod_prev = alphas_cumprod_prev,
         timesteps = timesteps,
         eta = eta
    )
}

#' Perform a DDIM scheduler step
#'
#' Performs a single denoising step using the DDIM (Denoising Diffusion Implicit Models)
#' algorithm. This function takes the output from a diffusion model at a specific timestep
#' and computes the previous (less noisy) sample in the diffusion process.
#'
#' @param model_output Numeric array. The output from the diffusion model, typically
#'   representing predicted noise or the denoised sample depending on `prediction_type`.
#' @param timestep Integer. The current timestep in the diffusion process.
#' @param sample Numeric array. The current noisy sample at timestep `t`.
#' @param eta Numeric. Controls the stochasticity of the process. When eta=0, DDIM is
#'   deterministic. When eta=1, it's equivalent to DDPM. Default: 0
#' @param use_clipped_model_output Logical. Whether to clip the model output before
#'   computing the sample update. Can improve stability. Default: FALSE
#' @param schedule List. The DDIM scheduler object containing the necessary parameters created from ddim_scheduler_create()
#' @param thresholding Logical. Whether to apply thresholding to the output.
#'   Default: FALSE
#' @param generator An optional random number generator for reproducibility.
#'   Default: NULL
#' @param variance_noise Optional pre-generated noise for the variance when eta > 0.
#'   If NULL and eta > 0, noise will be generated. Default: NULL
#' @param clip_sample Logical. Whether to clip the sample. Default: FALSE
#' @param set_alpha_to_one Logical. Whether to override the final alpha value to 1.
#'   Used for numerical stability in the final step. Default: FALSE
#' @param prediction_type Character. The type of prediction the model outputs.
#'   Options are:
#'   \describe{
#'     \item{"epsilon"}{The model predicts the noise}
#'     \item{"sample"}{The model predicts the denoised sample directly}
#'     \item{"v_prediction"}{The model predicts the velocity vector (v)}
#'   }
#'   Default: "epsilon"
#' @param dtype The data type to use for computations. Default is torch_float32().
#' @param device The device to use for computations. Options are "cpu" and "cuda".
#'
#' @return A list containing:
#'   \describe{
#'     \item{`prev_sample`}{The less noisy sample at timestep t-1}
#'     \item{`pred_original_sample`}{The predicted denoised sample}
#'   }
#'
#' @details
#' The DDIM step function implements the core sampling algorithm of DDIM described in
#' Song et al. 2020. It computes the previous sample x_t-1 given the current
#' sample x_t and the model output.
#'
#' The algorithm differs from DDPM by using a non-Markovian diffusion process that
#' allows for deterministic sampling and fewer inference steps without sacrificing
#' quality.
#'
#' When using `prediction_type="epsilon"` (most common), the model predicts the
#' noise that was added to create the current noisy sample. For `prediction_type="sample"`,
#' the model predicts the clean sample directly. The `v_prediction` option implements
#' the v-parameterization from Salimans & Ho (2022).
#'
#' @references
#' Song, J., Meng, C., & Ermon, S. (2020).
#' "Denoising Diffusion Implicit Models."
#' \url{https://arxiv.org/abs/2010.02502}
#'
#' Salimans, T., & Ho, J. (2022).
#' "Progressive Distillation for Fast Sampling of Diffusion Models."
#' \url{https://arxiv.org/abs/2202.00512}
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   scheduler <- ddim_scheduler_create(num_inference_steps = 5)
#'   sample <- torch::torch_randn(c(1, 4, 8, 8))
#'   model_output <- torch::torch_randn(c(1, 4, 8, 8))
#'   result <- ddim_scheduler_step(
#'     model_output = model_output,
#'     timestep = scheduler$timesteps[1],
#'     sample = sample,
#'     schedule = scheduler,
#'     eta = 0, # Deterministic sampling
#'     prediction_type = "epsilon")
#'   result$shape
#' }
#' @export
ddim_scheduler_step <- function(model_output, timestep, sample, schedule,
                                eta = 0, use_clipped_model_output = FALSE,
                                thresholding = FALSE, generator = NULL,
                                variance_noise = NULL, clip_sample = FALSE,
                                set_alpha_to_one = FALSE,
                                prediction_type = c("epsilon", "sample", "v_prediction"),
                                dtype = torch::torch_float32(),
                                device = "cpu") {
    prediction_type <- match.arg(prediction_type)
    # 1. get previous step value (= timestep + 1); i.e. python-indexing
    timestep_index <- torch::torch_tensor(timestep + 1,
        dtype = torch::torch_long(), device = torch::torch_device(device))
    if (as.numeric(timestep_index) <= 2) { #schedule$timesteps[1]) {
        prev_timestep <- 1 #length(schedule$alphas)
        prev_timestep_index <- torch::torch_tensor(prev_timestep,
            dtype = torch::torch_long(),
            device = torch::torch_device(device))
    } else {
        prev_timestep <- schedule$timesteps[which(as.logical(schedule$timesteps == (timestep))) + 1]
        prev_timestep_index <- torch::torch_tensor(prev_timestep + 1,
            dtype = torch::torch_long(),
            device = torch::torch_device(device))
    }

    # 2. compute alphas, betas
    alpha_prod_t <- schedule$alphas_cumprod[timestep_index]
    alpha_prod_t_prev <- schedule$alphas_cumprod[prev_timestep_index]
    # alpha_prod_t <- alpha_prod_t + 1e-7  # Prevent division by zero
    # alpha_prod_t_prev <- alpha_prod_t_prev + 1e-7
    if (set_alpha_to_one & prev_timestep == 1) {
        alpha_prod_t_prev <- torch::torch_tensor(1.0,
            dtype = dtype,
            device = torch::torch_device(device))
    } else {
        alpha_prod_t_prev <- schedule$alphas_cumprod[prev_timestep_index]
        alpha_prod_t_prev <- alpha_prod_t_prev$to(dtype = dtype, device = torch::torch_device(device))
    }

    beta_prod_t <- 1 - alpha_prod_t
    beta_prod_t <- beta_prod_t$to(dtype = dtype, device = torch::torch_device(device))
    beta_prod_t_prev <- 1 - alpha_prod_t_prev
    beta_prod_t_prev <- beta_prod_t_prev$to(dtype = dtype, device = torch::torch_device(device))

    # 3. Handle different prediction types (epsilon or v-prediction)
    preds <- switch(prediction_type,
                    "epsilon" = list(pred_original_sample = (sample - beta_prod_t ^ (0.5) * model_output) / alpha_prod_t ^ (0.5),
                                     pred_epsilon = model_output),
                    "sample" = list(pred_original_sample = model_output,
                                    pred_epsilon = (sample - alpha_prod_t ^ (0.5) * model_output) / beta_prod_t ^ (0.5)),
                    "v_prediction" = list(pred_original_sample = (alpha_prod_t ^ 0.5) * sample - (beta_prod_t ^ 0.5) * model_output,
            pred_epsilon = (alpha_prod_t ^ 0.5) * model_output + (beta_prod_t ^ 0.5) * sample))
    pred_original_sample <- preds$pred_original_sample$to(dtype = dtype, device = torch::torch_device(device))
    pred_epsilon <- preds$pred_epsilon$to(dtype = dtype, device = torch::torch_device(device))
    # if self.config.prediction_type == "epsilon":
    # pred_original_sample = (sample - beta_prod_t ** (0.5) * model_output) / alpha_prod_t ** (0.5)
    # pred_epsilon = model_output

    # 4. Clip or threshold "predicted x_0"
    # Check thresholding sample code
    if (thresholding) {
        flat <- pred_original_sample$view(c(pred_original_sample$size(1), -1))
        abs_flat <- torch::torch_abs(flat, device = torch::torch_device(device))
        s <- torch::torch_quantile(abs_flat,
                                   torch::torch_tensor(0.995, device = torch::torch_device(device)),
                                   dim = 2,
                                   device = torch::torch_device(device)) # dim=2 because batch is dim=1 in R torch
        s <- torch::torch_max(s, device = torch::torch_device(device))
        s <- s$view(c(pred_original_sample$size(1), 1, 1, 1))
        pred_original_sample <- torch::torch_clamp(pred_original_sample,
            min = -s, max = s,
            device = torch::torch_device(device)) / s
    } else {
        if (clip_sample) {
            # Clip to [-1, 1] range
            pred_original_sample <- torch::torch_clamp(pred_original_sample,
                min = -1, max = 1,
                device = torch::torch_device(device))
        }
    }

    # 5. compute variance
    variance <- (beta_prod_t_prev / beta_prod_t) * (1 - alpha_prod_t / alpha_prod_t_prev)
    std_dev_t <- eta * sqrt(variance)

    # Re-compute model_output (noise prediction) from clipped pred_original_sample if needed
    if (use_clipped_model_output) {
        if (prediction_type == "v_prediction") {
            # Recompute v-prediction from clipped pred_original_sample
            model_output <- sqrt(alpha_prod_t) * pred_epsilon - sqrt(beta_prod_t) * pred_original_sample
        } else {
            # Recompute epsilon from clipped pred_original_sample
            model_output <- (sample - sqrt(alpha_prod_t) * pred_original_sample) / sqrt(beta_prod_t)
        }
    }

    # 6. compute "direction pointing to x_t" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
    pred_sample_direction <- sqrt(1 - alpha_prod_t_prev - std_dev_t ^ 2) * pred_epsilon

    # 7. compute x_t without "random noise" of formula (12) from https://arxiv.org/pdf/2010.02502.pdf
    prev_sample <- sqrt(alpha_prod_t_prev) * pred_original_sample + pred_sample_direction

    if (eta > 0) {
        if (!is.null(variance_noise) && !is.null(generator)) {
            stop("Provide either `generator` or `variance_noise`, not both.")
        }
        if (is.null(variance_noise)) {
            variance_noise <- torch::torch_randn(model_output$shape,
                generator = generator,
                device = torch::torch_device(device),
                dtype = dtype)
        }
        variance <- std_dev_t * eta * variance_noise
        prev_sample <- prev_sample + variance
    }

    prev_sample
}

#' Add noise to latents using DDIM scheduler
#'
#' This function adds noise to the original latents according to the DDIM
#' scheduler's diffusion process. It computes the noisy latents based on the
#' original latents, noise, and the current timestep.
#'
#' @param original_latents A torch tensor representing the original latents.
#'
#' @param noise A torch tensor representing the noise to be added.
#'
#' @param timestep An integer representing the current timestep in the diffusion
#' process.
#'
#' @param scheduler_obj A list containing the DDIM scheduler parameters, including
#' alphas_cumprod and timesteps. The alphas_cumprod represents how much of the
#' original signal remains at each timestep of the diffusion process.
#'
#' @return A torch tensor containing the noised latents, which represents the
#' original latents with the appropriate amount of noise added for the given
#' timestep.
#'
#' @details
#' The noise is added according to the standard diffusion forward process formula:
#' noised_latents = sqrt(alpha_cumprod) * original_latents + sqrt(1-alpha_cumprod) * noise
#'
#' Where alpha_cumprod is the cumulative product of (1-beta) values up to the
#' specified timestep, with beta being the noise schedule.
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   scheduler <- ddim_scheduler_create(num_inference_steps = 5)
#'   latents <- torch::torch_randn(c(1, 4, 8, 8))
#'   noised_latents <- scheduler_add_noise(
#'     original_latents = latents,
#'     noise = torch::torch_randn_like(latents),
#'     timestep = scheduler$timesteps[1],
#'     scheduler_obj = scheduler
#'   )
#'   noised_latents$shape
#' }
#'
#' @export
scheduler_add_noise <- function(original_latents, noise, timestep,
                                scheduler_obj) {
    # Get the alpha_cumprod value for this timestep
    # In DDIM/DDPM schedulers, alphas_cumprod represents
    # how much of the original signal remains at each timestep

    timestep_index <- timestep + 1
    # Get alpha_cumprod for this timestep
    alpha_cumprod <- scheduler_obj$alphas_cumprod[timestep_index]

    # Calculate the noise scaling factors
    sqrt_alpha_prod <- torch::torch_sqrt(torch::torch_tensor(alpha_cumprod,
            device = original_latents$device))
    sqrt_one_minus_alpha_prod <- torch::torch_sqrt(
        torch::torch_tensor(1 - alpha_cumprod, device = original_latents$device)
    )

    # Add noise to the original latents according to the diffusion formula:
    # noisy = sqrt(alpha) * original + sqrt(1-alpha) * noise
    noised_latents <- sqrt_alpha_prod * original_latents +
    sqrt_one_minus_alpha_prod * noise

    return(noised_latents)
}

rescale_zero_terminal_snr <- function(betas) {
    alphas <- 1.0 - betas
    alphas_cumprod <- cumprod(alphas)
    alphas_bar_sqrt <- sqrt(alphas_cumprod)

    # Store old values.
    alphas_bar_sqrt_0 <- alphas_bar_sqrt[1]
    alphas_bar_sqrt_T <- alphas_bar_sqrt[length(alphas_bar_sqrt)]

    # Shift so the last timestep is zero.
    alphas_bar_sqrt <- alphas_bar_sqrt - alphas_bar_sqrt_T

    # Scale so the first timestep is back to the old value.
    alphas_bar_sqrt <- alphas_bar_sqrt * alphas_bar_sqrt_0 / (alphas_bar_sqrt_0 - alphas_bar_sqrt_T)

    # Convert alphas_bar_sqrt to betas
    alphas_bar <- alphas_bar_sqrt ^ 2 # Revert sqrt
    alphas <- alphas_bar[-1] / alphas_bar[-length(alphas_bar)] # Revert cumprod
    alphas <- c(alphas_bar[1], alphas)
    betas <- 1 - alphas

    return(betas)
}

#' Load HF safetensors weights into the native SD/SDXL UNet
#'
#' The native UNet modules mirror the diffusers
#' \code{UNet2DConditionModel} state-dict keys 1:1, with the sole
#' exception that the time- (and, for SDXL, add-) embedding MLPs are
#' flattened from dotted to underscored names
#' (\code{time_embedding.linear_1} -> \code{time_embedding_linear_1}).
#' These loaders read \code{unet/diffusion_pytorch_model.safetensors}
#' (single file or sharded via its \code{.index.json}) and copy each
#' weight into the matching native parameter, verifying that every native
#' parameter is filled and no key or shape is left unmatched.
#'
#' Reads route through the shared sharded opener, so an oversize (>2 GB)
#' single-file checkpoint on stock CRAN safetensors surfaces the
#' actionable "rebuild with smaller shards or install the fork" message
#' rather than a raw 32-bit overflow.
#'
#' @name unet_safetensors
NULL

# time_embedding.linear_{1,2} -> time_embedding_linear_{1,2}
.unet_remap_sd21 <- function(key) {
    key <- sub("^time_embedding\\.linear_1", "time_embedding_linear_1", key)
    sub("^time_embedding\\.linear_2", "time_embedding_linear_2", key)
}

# SD21 rules plus add_embedding.linear_{1,2} -> add_embedding_linear_{1,2}
.unet_remap_sdxl <- function(key) {
    key <- .unet_remap_sd21(key)
    key <- sub("^add_embedding\\.linear_1", "add_embedding_linear_1", key)
    sub("^add_embedding\\.linear_2", "add_embedding_linear_2", key)
}

# Shared loader: open the (single or sharded) diffusers UNet directory,
# map each HF key to a native parameter through `remap`, copy with a
# shape check, and fail loudly on any unmapped key, shape mismatch, or
# unfilled native parameter.
.load_unet_safetensors <- function(native_unet, path, remap, label,
                                   verbose = TRUE) {
    if (!requireNamespace("safetensors", quietly = TRUE)) {
        stop("The safetensors package is required to read UNet weights.")
    }
    path <- path.expand(path)
    if (dir.exists(path)) {
        dir <- path
    } else {
        dir <- dirname(path)
    }
    opened <- .flux_open_sharded_dir(dir, "diffusion_pytorch_model")
    keys <- opened$keys

    dests <- native_unet$named_parameters()
    filled <- character(0)
    unmapped <- character(0)
    mismatch <- character(0)

    torch::with_no_grad({
        for (key in keys) {
            native_name <- remap(key)
            dest <- dests[[native_name]]
            if (is.null(dest)) {
                unmapped <- c(unmapped, key)
                next
            }
            src <- opened$handle$get_tensor(key)
            if (!all(dest$shape == src$shape)) {
                mismatch <- c(mismatch, sprintf("%s (%s vs %s)", native_name,
                        paste(as.integer(src$shape), collapse = "x"),
                        paste(as.integer(dest$shape), collapse = "x")))
                next
            }
            dest$copy_(src)
            filled <- c(filled, native_name)
        }
    })

    if (length(unmapped)) {
        stop(label, " load: ", length(unmapped), " unmapped keys, e.g. ",
             paste(utils::head(unmapped, 3), collapse = ", "))
    }
    if (length(mismatch)) {
        stop(label, " load: ", length(mismatch), " shape mismatches, e.g. ",
             paste(utils::head(mismatch, 3), collapse = ", "))
    }
    unfilled <- setdiff(names(dests), filled)
    if (length(unfilled)) {
        stop(label, " load: ", length(unfilled), " unfilled params, e.g. ",
             paste(utils::head(unfilled, 3), collapse = ", "))
    }
    if (verbose) {
        message("Loaded ", length(filled), " ", label, " parameters")
    }
    invisible(native_unet)
}

#' Load HF safetensors weights into the native SD21 UNet
#'
#' @param native_unet A \code{\link{unet_native}} module.
#' @param path Path to the UNet directory (containing
#'   \code{diffusion_pytorch_model.safetensors} or its shard index) or
#'   directly to the single-file checkpoint.
#' @param verbose Print how many parameters were loaded.
#'
#' @return The native UNet with weights loaded (invisibly).
#' @export
load_unet_safetensors <- function(native_unet, path, verbose = TRUE) {
    .load_unet_safetensors(native_unet, path, .unet_remap_sd21, "SD21 UNet",
                           verbose = verbose)
}

#' Load HF safetensors weights into the native SDXL UNet
#'
#' @param native_unet A \code{\link{unet_sdxl_native}} module.
#' @param path Path to the UNet directory (containing
#'   \code{diffusion_pytorch_model.safetensors} or its shard index) or
#'   directly to the single-file checkpoint.
#' @param verbose Print how many parameters were loaded.
#'
#' @return The native UNet with weights loaded (invisibly).
#' @export
load_unet_sdxl_safetensors <- function(native_unet, path, verbose = TRUE) {
    .load_unet_safetensors(native_unet, path, .unet_remap_sdxl, "SDXL UNet",
                           verbose = verbose)
}

#' Build a native SD21 UNet from a diffusers safetensors directory
#'
#' The safetensors counterpart to
#' \code{\link{unet_native_from_torchscript}}: constructs
#' \code{\link{unet_native}} and loads its weights from
#' \code{unet/diffusion_pytorch_model.safetensors} (no TorchScript, so it
#' works on Blackwell). The default construction matches the canonical
#' Stable Diffusion 2.1 UNet; pass constructor overrides through
#' \code{...} for a variant checkpoint (the loader fails loudly on any
#' shape mismatch, so a wrong architecture surfaces immediately rather
#' than loading silently wrong weights).
#'
#' @param path Path to the UNet directory or its single-file checkpoint.
#' @param verbose Print how many parameters were loaded.
#' @param ... Overrides for \code{\link{unet_native}} constructor args.
#'
#' @return The native SD21 UNet in eval mode.
#' @export
unet_native_from_safetensors <- function(path, verbose = TRUE, ...) {
    model <- unet_native(...)
    load_unet_safetensors(model, path, verbose = verbose)
    model$eval()
    model
}

#' Build a native SDXL UNet from a diffusers safetensors directory
#'
#' The safetensors counterpart to
#' \code{\link{unet_sdxl_native_from_torchscript}}: constructs
#' \code{\link{unet_sdxl_native}} and loads its weights from
#' \code{unet/diffusion_pytorch_model.safetensors}. Validated against the
#' cached \code{stabilityai/stable-diffusion-xl-base-1.0} UNet (all 1680
#' parameters map with matching shapes). Pass constructor overrides
#' through \code{...} for a variant checkpoint.
#'
#' @param path Path to the UNet directory or its single-file checkpoint.
#' @param verbose Print how many parameters were loaded.
#' @param ... Overrides for \code{\link{unet_sdxl_native}} constructor args.
#'
#' @return The native SDXL UNet in eval mode.
#' @export
unet_sdxl_native_from_safetensors <- function(path, verbose = TRUE, ...) {
    model <- unet_sdxl_native(...)
    load_unet_sdxl_safetensors(model, path, verbose = verbose)
    model$eval()
    model
}

# SD/SDXL UNet safetensors loader: HF-key remap rules, the copy path,
# completeness, and error reporting. The remap is validated against real
# cached SDXL weights out of band (all 1680 keys map with matching
# shapes); here we exercise the loader logic portably through a mock
# module so no multi-GB checkpoint is needed.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}
library(diffuseR)

r21 <- diffuseR:::.unet_remap_sd21
rxl <- diffuseR:::.unet_remap_sdxl

# --- remap rules ------------------------------------------------------------------

expect_equal(r21("time_embedding.linear_1.weight"),
  "time_embedding_linear_1.weight")
expect_equal(r21("time_embedding.linear_2.bias"),
  "time_embedding_linear_2.bias")
# non-embedding keys pass through untouched (dotted block paths)
expect_equal(r21("down_blocks.0.resnets.0.norm1.weight"),
  "down_blocks.0.resnets.0.norm1.weight")
expect_equal(r21("conv_out.weight"), "conv_out.weight")
# SDXL adds add_embedding and keeps the time_embedding rule
expect_equal(rxl("add_embedding.linear_1.weight"),
  "add_embedding_linear_1.weight")
expect_equal(rxl("add_embedding.linear_2.bias"), "add_embedding_linear_2.bias")
expect_equal(rxl("time_embedding.linear_1.weight"),
  "time_embedding_linear_1.weight")
# SD21 rule leaves add_embedding alone (SD21 has none)
expect_equal(r21("add_embedding.linear_1.weight"),
  "add_embedding.linear_1.weight")

# --- round-trip through a mock module ---------------------------------------------

native_names <- c("conv_in.weight", "conv_in.bias",
  "time_embedding_linear_1.weight", "time_embedding_linear_1.bias",
  "time_embedding_linear_2.weight",
  "down_blocks.0.resnets.0.norm1.weight", "add_embedding_linear_1.weight")

# inverse remap (native -> HF) to author a synthetic checkpoint
to_hf <- function(n) {
  n <- sub("^time_embedding_linear_1", "time_embedding.linear_1", n)
  n <- sub("^time_embedding_linear_2", "time_embedding.linear_2", n)
  sub("^add_embedding_linear_1", "add_embedding.linear_1", n)
}

make_params <- function(gen) {
  p <- lapply(native_names, function(i) gen())
  names(p) <- native_names
  p
}
truth <- make_params(function() torch::torch_randn(2L, 3L))

dir <- tempfile()
dir.create(dir)
hf <- truth
names(hf) <- vapply(native_names, to_hf, character(1))
safetensors::safe_save_file(hf,
  file.path(dir, "diffusion_pytorch_model.safetensors"))

# fresh zeroed params; the loader must copy truth into them
params <- make_params(function() torch::torch_zeros(2L, 3L))
mock <- list(named_parameters = function() params)
diffuseR:::.load_unet_safetensors(mock, dir, rxl, "mock UNet", verbose = FALSE)
for (nm in native_names) {
  expect_true(as.logical(torch::torch_allclose(params[[nm]], truth[[nm]])))
}

# passing the file directly (not the dir) also works
params_b <- make_params(function() torch::torch_zeros(2L, 3L))
mock_b <- list(named_parameters = function() params_b)
diffuseR:::.load_unet_safetensors(mock_b,
  file.path(dir, "diffusion_pytorch_model.safetensors"), rxl, "mock",
  verbose = FALSE)
expect_true(as.logical(torch::torch_allclose(params_b[["conv_in.weight"]],
  truth[["conv_in.weight"]])))

# --- error paths ------------------------------------------------------------------

# extra HF key with no native destination -> unmapped
dir2 <- tempfile()
dir.create(dir2)
hf_extra <- hf
hf_extra[["nonexistent.layer.weight"]] <- torch::torch_zeros(2L, 3L)
safetensors::safe_save_file(hf_extra,
  file.path(dir2, "diffusion_pytorch_model.safetensors"))
expect_error(
  diffuseR:::.load_unet_safetensors(
    list(named_parameters = function() make_params(function() torch::torch_zeros(2L, 3L))),
    dir2, rxl, "mock", verbose = FALSE),
  pattern = "unmapped")

# a native param with no checkpoint key -> unfilled
params_missing <- c(make_params(function() torch::torch_zeros(2L, 3L)),
  list("extra.param.weight" = torch::torch_zeros(2L, 3L)))
expect_error(
  diffuseR:::.load_unet_safetensors(
    list(named_parameters = function() params_missing), dir, rxl, "mock",
    verbose = FALSE),
  pattern = "unfilled")

# shape mismatch -> error (checked before unfilled)
params_wrong <- make_params(function() torch::torch_zeros(2L, 3L))
params_wrong[["conv_in.weight"]] <- torch::torch_zeros(5L, 5L)
expect_error(
  diffuseR:::.load_unet_safetensors(
    list(named_parameters = function() params_wrong), dir, rxl, "mock",
    verbose = FALSE),
  pattern = "shape mismatch")

unlink(c(dir, dir2), recursive = TRUE)

# --- end-to-end against real SDXL weights (opt-in; heavy ~9.6 GB load) -------------
# Runs only with DIFFUSER_TEST_SDXL_LOAD=1 and the cached diffusers UNet
# present, so a normal suite run never pulls 9.6 GB. Validated manually:
# builds + loads all 1680 params in ~20 s.
sdxl_dir <- Sys.glob(file.path("~/.cache/huggingface/hub",
  "models--stabilityai--stable-diffusion-xl-base-1.0/snapshots/*/unet"))
if (at_home() && nzchar(Sys.getenv("DIFFUSER_TEST_SDXL_LOAD")) &&
    length(sdxl_dir) && dir.exists(sdxl_dir[1])) {
  m <- unet_sdxl_native_from_safetensors(sdxl_dir[1], verbose = FALSE)
  expect_equal(length(m$named_parameters()), 1680L)
  w <- m$named_parameters()[["conv_in.weight"]]
  expect_true(as.numeric(w$abs()$sum()$item()) > 0)
  expect_false(m$training)
  rm(m)
  gc()
}

#' FLUX Transformer Quantization and Loading
#'
#' Quantize the 12B FLUX transformer to NF4 (~7 GB, GPU-resident on
#' 16 GB cards) or fp8 (~12 GB, CPU-resident and streamed per forward),
#' and load any format back into \code{\link{flux_transformer}}. Reuses
#' the LTX-2.3 quantization machinery (\code{ltx23_nf4_quantize},
#' \code{ltx23_nf4_linear}, \code{ltx23_fp8_linear}); only the cast set
#' and the diffusers directory layout are FLUX-specific.
#'
#' @name quantize_flux
NULL

.flux_dtype <- function(dtype) {
    switch(dtype, bfloat16 = torch::torch_bfloat16(),
           float16 = torch::torch_float16(), float32 = torch::torch_float32(),
           stop("Unsupported dtype: ", dtype))
}

# Capability probe: can the installed safetensors round-trip this dtype?
# CRAN safetensors (<= 0.2.1) can read bfloat16 but not write it, and has
# no float8 support; the fixes are upstream PRs. Cached per session;
# options(diffuseR.st_caps = list(bfloat16 = FALSE, ...)) overrides for
# tests.
.st_caps <- new.env(parent = emptyenv())
.st_can_write <- function(dtype = c("bfloat16", "float8_e4m3fn")) {
    dtype <- match.arg(dtype)
    override <- getOption("diffuseR.st_caps")
    if (!is.null(override) && !is.null(override[[dtype]])) {
        return(isTRUE(override[[dtype]]))
    }
    cached <- .st_caps[[dtype]]
    if (!is.null(cached)) {
        return(cached)
    }
    ok <- requireNamespace("safetensors", quietly = TRUE) && tryCatch({
        target <- switch(dtype, bfloat16 = torch::torch_bfloat16(),
                         float8_e4m3fn = torch::torch_float8_e4m3fn())
        x <- torch::torch_zeros(2L)$to(dtype = target)
        tmp <- tempfile(fileext = ".safetensors")
        on.exit(unlink(tmp), add = TRUE)
        safetensors::safe_save_file(list(w = x), tmp)
        y <- safetensors::safe_load_file(tmp, framework = "torch")
        !is.null(y$w)
    }, error = function(e) FALSE)
    .st_caps[[dtype]] <- ok
    ok
}

# Resolve precision = "auto": prefer an existing quantized artifact
# (fp8 first), else pick by float8 write capability. An fp8 artifact is
# only chosen if the installed safetensors can actually read float8 -
# otherwise a fork-built fp8 artifact on a CRAN-safetensors machine
# would be selected and then fail at read time. Write capability is a
# sound proxy for read capability (nothing writes fp8 but cannot read
# it).
.flux_resolve_precision <- function(precision, artifact_prefix = NULL) {
    if (!identical(precision, "auto")) {
        return(precision)
    }
    fp8_ok <- .st_can_write("float8_e4m3fn")
    if (!is.null(artifact_prefix)) {
        for (p in c("fp8", "nf4")) {
            if (p == "fp8" && !fp8_ok) {
                next
            }
            if (file.exists(file.path(paste0(artifact_prefix, p),
                                      "manifest.json"))) {
                return(p)
            }
        }
    }
    if (fp8_ok) {
        "fp8"
    } else {
        "nf4"
    }
}

# Locate a flux-family model's unquantized transformer in the hfhub
# cache. bf16 is not a built artifact: it IS the source the quantizers
# read, so the "bf16 tier" means loading those weights directly
# (flux_load_transformer() handles them as format "full"). Resolves from
# the cache only -- a loader must never trigger a multi-GB download as a
# side effect of a precision choice.
.flux_source_dir <- function(model) {
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to locate bf16 source weights.",
             call. = FALSE)
    }
    spec <- switch(model,
                   flux1 = list(repo = .flux1_repo,
                                files = .flux1_transformer_files,
                                fn = "download_flux1"),
                   flux2 = list(repo = .flux2_repo,
                                files = .flux2_transformer_files,
                                fn = "download_flux2_klein"),
                   zimage = list(repo = .zimage_repo,
                                 files = .zimage_transformer_files,
                                 fn = "download_zimage_turbo"),
                   stop("No bf16 source known for model '", model, "'",
                        call. = FALSE))
    paths <- lapply(spec$files, function(f) {
        tryCatch(hfhub::hub_download(spec$repo, f, local_files_only = TRUE),
                 error = function(e) NULL)
    })
    if (any(vapply(paths, is.null, logical(1)))) {
        stop(sprintf(paste0("bf16 needs the unquantized %s transformer, ",
                            "which is not in the HuggingFace cache. Run ",
                            "%s(quantize = FALSE) to fetch it, or pick a ",
                            "quantized precision."),
                     spec$repo, spec$fn), call. = FALSE)
    }
    # Every file list starts with transformer/config.json, so the parent
    # of the first hit is the transformer directory.
    dirname(paths[[1]])
}

# Where a loader should read weights for a resolved precision: the
# built artifact for nf4/fp8, the cached source for bf16.
.flux_model_dir <- function(model, precision, prefix) {
    if (identical(precision, "bf16")) {
        .flux_source_dir(model)
    } else {
        paste0(prefix, precision)
    }
}

# Transformer constructor arguments from a diffusers config.json
.flux_transformer_args <- function(config) {
    if (is.null(config)) {
        return(list())
    }
    if (isTRUE(config$guidance_embeds)) {
        stop("This checkpoint uses guidance embeddings (FLUX.1-dev); ",
             "only FLUX.1-schnell (guidance_embeds = false) is supported.")
    }
    args <- list(
                 in_channels = config$in_channels,
                 num_layers = config$num_layers,
                 num_single_layers = config$num_single_layers,
                 attention_head_dim = config$attention_head_dim,
                 num_attention_heads = config$num_attention_heads,
                 joint_attention_dim = config$joint_attention_dim,
                 pooled_projection_dim = config$pooled_projection_dim,
                 axes_dims_rope = config$axes_dims_rope,
                 out_channels = config$out_channels
    )
    # JSON roundtrips turn null into empty lists; drop both
    args <- Filter(function(x) !is.null(x) && length(x) > 0L, args)
    lapply(args, function(x) if (is.numeric(x)) as.integer(x) else x)
}

.flux2_transformer_args <- function(config) {
    if (is.null(config)) {
        return(list())
    }
    if (isTRUE(config$guidance_embeds)) {
        stop("This checkpoint uses guidance embeddings (FLUX.2-dev); ",
             "only the klein variants (guidance_embeds = false) are supported.")
    }
    args <- list(
                 in_channels = config$in_channels,
                 num_layers = config$num_layers,
                 num_single_layers = config$num_single_layers,
                 attention_head_dim = config$attention_head_dim,
                 num_attention_heads = config$num_attention_heads,
                 joint_attention_dim = config$joint_attention_dim,
                 timestep_guidance_channels = config$timestep_guidance_channels,
                 axes_dims_rope = config$axes_dims_rope,
                 out_channels = config$out_channels
    )
    args <- Filter(function(x) !is.null(x) && length(x) > 0L, args)
    args <- lapply(args, function(x) if (is.numeric(x)) as.integer(x) else x)
    for (field in c("mlp_ratio", "rope_theta", "eps")) {
        v <- config[[field]]
        if (!is.null(v) && length(v) == 1L) {
            args[[field]] <- as.numeric(v)
        }
    }
    args
}

.zimage_transformer_args <- function(config) {
    if (is.null(config)) {
        return(list())
    }
    args <- list(in_channels = config$in_channels, dim = config$dim,
                 n_layers = config$n_layers,
                 n_refiner_layers = config$n_refiner_layers,
                 n_heads = config$n_heads, cap_feat_dim = config$cap_feat_dim,
                 axes_dims = config$axes_dims,
                 patch_size = config$all_patch_size[[1]],
                 f_patch_size = config$all_f_patch_size[[1]])
    args <- Filter(function(x) !is.null(x) && length(x) > 0L, args)
    args <- lapply(args, function(x) if (is.numeric(x)) as.integer(x) else x)
    for (field in c("norm_eps", "rope_theta", "t_scale")) {
        v <- config[[field]]
        if (!is.null(v) && length(v) == 1L) {
            args[[field]] <- as.numeric(v)
        }
    }
    args
}

# Move plain-field fp8 weights (and their scales) onto a device; used
# for resident fp8 where the whole quantized model fits on the GPU
.flux_fp8_to_device <- function(module, device) {
    for (name in names(module$children)) {
        child <- module$children[[name]]
        if (!is.null(child$weight_fp8)) {
            child$weight_fp8 <- child$weight_fp8$to(device = device)
            child$weight_scale <- child$weight_scale$to(device = device)
        }
        .flux_fp8_to_device(child, device)
    }
    invisible(module)
}

# Collect an fp8 transformer's plain-field weight tensors. weight_fp8
# and weight_scale live outside parameters/buffers, so .pin_component's
# module walk misses them; this mirrors .flux_fp8_to_device's recursion
# and returns the tensors so they can be pinned and staged alongside the
# module. set_data (in .pin_component) mutates each in place, so the
# field reference stays valid and staged onload/offload covers them -
# replacing the per-generation .flux_fp8_to_device reassignment.
.flux_fp8_collect <- function(module, out = list()) {
    for (name in names(module$children)) {
        child <- module$children[[name]]
        if (!is.null(child$weight_fp8)) {
            out[[length(out) + 1L]] <- child$weight_fp8
            out[[length(out) + 1L]] <- child$weight_scale
        }
        out <- .flux_fp8_collect(child, out)
    }
    out
}

# Pin the phase-swapped components of a flux-family pipeline so the
# per-generation CPU<->GPU moves run at DMA rate (see staging.R).
# Returns a named list of staging pairs keyed by component, or NULL when
# staging does not apply (opted out, no phase offload, or CPU device).
# Only the listed components are pinned; text encoders resident on the
# host (text_device == "cpu") are omitted by the caller since they
# compute in place. Resident-fp8 transformers (flux2/zimage) also pin
# their plain weight_fp8/weight_scale fields so staged transfer covers
# them; nf4 packed weights are buffers and ride along automatically.
.flux_build_staging <- function(pipe, pin, phase_offload, device, components,
                                verbose = TRUE) {
    if (!(isTRUE(pin) && isTRUE(phase_offload) &&
            identical(device, "cuda") && torch::cuda_is_available())) {
        return(NULL)
    }
    if (verbose) {
        message("Pinning host staging buffers...")
    }
    staging <- list()
    for (nm in components) {
        module <- pipe[[nm]]
        if (is.null(module)) {
            next
        }
        extra <- if (identical(nm, "transformer") &&
            isTRUE(pipe$fp8_resident)) {
            .flux_fp8_collect(module)
        } else {
            NULL
        }
        st <- .pin_component(module, extra = extra)
        if (!is.null(st)) {
            staging[[nm]] <- st
        }
    }
    if (length(staging)) {
        staging
    } else {
        NULL
    }
}

# Family-specific hooks for quantization and loading
.flux_family_hooks <- function(config) {
    family <- .flux_family(config)
    if (family == "flux2") {
        list(model_fn = flux2_transformer, args_fn = .flux2_transformer_args,
             is_quant_key = flux2_is_quant_key, shard_prefix = "flux2-klein")
    } else if (family == "zimage") {
        list(model_fn = zimage_transformer,
             args_fn = .zimage_transformer_args,
             is_quant_key = zimage_is_quant_key,
             shard_prefix = "zimage-turbo")
    } else {
        list(model_fn = flux_transformer,
             args_fn = .flux_transformer_args,
             is_quant_key = flux_is_quant_key,
             shard_prefix = "flux1")
    }
}

#' Quantize a FLUX transformer to NF4 or fp8 shards
#'
#' Streams the bf16 diffusers checkpoint tensor by tensor. Cast-set
#' weights (see \code{\link{flux_is_quant_key}}) are stored as NF4
#' (packed uint8 + \code{<key>_absmax} float32 blocks) or as
#' float8_e4m3fn with an absmax/448 per-tensor \code{<key>_scale};
#' everything else is copied through unchanged. The manifest embeds the
#' transformer config, so the source checkpoint is not needed again
#' after quantization.
#'
#' @param transformer_dir Source diffusers transformer directory.
#' @param output_dir Output directory for shards + manifest (default:
#'   the per-format location under \code{tools::R_user_dir}).
#' @param format "nf4" or "fp8".
#' @param shard_bytes Numeric. Target shard size in bytes. The default
#'   1.9e9 keeps every shard under the 2^31-byte (~2.15 GB) ceiling that
#'   stock CRAN safetensors can read, so the artifact loads fork-free.
#'   Pass a larger value (e.g. 4e9) only for local builds you will read
#'   back with a fork-patched safetensors.
#' @param force Logical. Re-quantize even if a valid manifest exists.
#' @param verbose Logical.
#'
#' @return Invisibly, the manifest list.
#'
#' @export
flux_quantize <- function(transformer_dir, output_dir = NULL,
                          format = c("nf4", "fp8"), shard_bytes = 1.9e9,
                          force = FALSE, verbose = TRUE) {
    format <- match.arg(format)
    if (format == "fp8" && !.st_can_write("float8_e4m3fn")) {
        stop("The installed safetensors package cannot write float8 ",
             "tensors (needs the float8 support from ",
             "mlverse/safetensors#13, which is merged upstream but not ",
             "yet on CRAN; install the development version of ",
             "safetensors from GitHub, or use format = \"nf4\").",
             call. = FALSE)
    }
    # Residents load into the compute dtype either way; bf16 halves the
    # artifact but the CRAN build of safetensors 0.2.1 cannot write it
    # (the fix is merged upstream; the probe decides, not the version)
    resident_dtype <- if (.st_can_write("bfloat16")) {
        torch::torch_bfloat16()
    } else {
        message("safetensors cannot write bfloat16; storing resident ",
                "tensors as float32 (larger artifact, same results)")
        torch::torch_float32()
    }
    if (is.null(output_dir)) {
        output_dir <- file.path(tools::R_user_dir("diffuseR", "data"),
                                paste0("flux1-schnell-", format))
    }

    manifest_path <- file.path(output_dir, "manifest.json")
    if (!force && file.exists(manifest_path)) {
        manifest <- jsonlite::fromJSON(manifest_path)
        if (identical(manifest$format, format) &&
            all(file.exists(file.path(output_dir, manifest$shards)))) {
            if (verbose) {
                message(toupper(format), " artifact already present: ",
                        output_dir)
            }
            return(invisible(manifest))
        }
    }
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

    ckpt <- flux_open_checkpoint(transformer_dir)
    hooks <- .flux_family_hooks(ckpt$config)
    if (format == "fp8") {
        fp8 <- torch::torch_float8_e4m3fn()
    }

    shard <- list()
    shard_size <- 0
    shard_files <- character(0)
    n_cast <- 0L

    flush_shard <- function() {
        if (!length(shard)) {
            return()
        }
        fname <- sprintf("%s-%s-%05d.safetensors", hooks$shard_prefix,
                         format, length(shard_files) + 1L)
        safetensors::safe_save_file(shard, file.path(output_dir, fname))
        shard_files[[length(shard_files) + 1L]] <<- fname
        if (verbose) {
            message(sprintf("  wrote %s (%.2f GB, %d tensors)", fname,
                            shard_size / 1e9, length(shard)))
        }
        shard <<- list()
        shard_size <<- 0
        gc(verbose = FALSE)
    }

    keys <- ckpt$keys
    for (i in seq_along(keys)) {
        key <- keys[[i]]
        tensor <- ckpt$handle$get_tensor(key)

        if (hooks$is_quant_key(key)) {
            torch::with_no_grad({
                if (format == "nf4") {
                    q <- ltx23_nf4_quantize(tensor)
                    shard[[key]] <- q$packed
                    shard[[paste0(key, "_absmax")]] <- q$absmax
                    shard_size <- shard_size + prod(tensor$shape) * 0.5625
                } else {
                    scale <- tensor$abs()$max()$to(dtype = torch::torch_float32())$
                    clamp(min = 1e-12) / 448
                    shard[[key]] <- (tensor$to(dtype = torch::torch_float32()) / scale)$to(dtype = fp8)
                    shard[[paste0(key, "_scale")]] <- scale
                    shard_size <- shard_size + prod(tensor$shape)
                }
            })
            n_cast <- n_cast + 1L
        } else {
            # Residents load into the compute dtype anyway; storing bf16
            # keeps fp32-shipped checkpoints (Z-Image) from doubling the
            # artifact. Identity for bf16 sources.
            shard[[key]] <- tensor$to(dtype = resident_dtype)
            shard_size <- shard_size + prod(tensor$shape) * 2
        }
        rm(tensor)

        if (shard_size >= shard_bytes) {
            flush_shard()
        }
        if (i %% 200L == 0L) {
            gc(verbose = FALSE)
            if (verbose) {
                message(sprintf("  quantizing %d/%d tensors", i, length(keys)))
            }
        }
    }
    flush_shard()

    manifest <- list(
                     source = basename(transformer_dir),
                     format = format,
                     shards = shard_files,
                     tensors = length(keys),
                     cast = n_cast,
                     config = ckpt$config
    )
    jsonlite::write_json(manifest, manifest_path, auto_unbox = TRUE,
                         pretty = TRUE)
    if (verbose) {
        message(sprintf("Quantized %d/%d tensors to %s across %d shards: %s",
                        n_cast, length(keys), format, length(shard_files),
                        output_dir))
    }
    invisible(manifest)
}

#' Load a FLUX transformer from any checkpoint format
#'
#' Builds \code{\link{flux_transformer}} from the checkpoint's embedded
#' config and loads the weights. Dispatches on the checkpoint format:
#'
#' \itemize{
#'   \item full precision (\code{\link{flux_open_checkpoint}}): weights
#'     stream into the model in \code{dtype} on \code{device}.
#'   \item \code{"nf4"} (\code{\link{flux_open_quantized}}): cast-set
#'     linears become \code{ltx23_nf4_linear}; the whole model (packed
#'     weights included) moves to \code{device} and stays resident.
#'   \item \code{"fp8"}: cast-set linears become
#'     \code{ltx23_fp8_linear}; fp8 weights stay CPU-resident (optionally
#'     pinned) and stream to \code{device} inside each forward.
#' }
#'
#' @param ckpt A checkpoint from \code{\link{flux_open_checkpoint}} or
#'   \code{\link{flux_open_quantized}}.
#' @param device Character. Compute device.
#' @param dtype Character. Model dtype ("bfloat16" or "float32"). For
#'   quantized formats this sets the resident (non-quantized) tensors
#'   and must match the compute dtype: bfloat16 for GPU compute,
#'   float32 for CPU compute.
#' @param pin Logical. Pin fp8 host memory for faster transfers
#'   (streamed fp8 only).
#' @param fp8_resident Logical. Keep the fp8 weights on \code{device}
#'   instead of streaming from the CPU - right for models whose whole
#'   quantized footprint fits in VRAM (FLUX.2 klein-4B: ~4 GB).
#' @param verbose Logical.
#' @param ... Overrides for \code{\link{flux_transformer}} arguments
#'   (tiny test configs).
#'
#' @return The loaded \code{flux_transformer} in eval mode.
#'
#' @export
flux_load_transformer <- function(ckpt, device = "cuda", dtype = "bfloat16",
                                  pin = TRUE, fp8_resident = FALSE,
                                  verbose = TRUE, ...) {
    stopifnot(inherits(ckpt, "ltx23_checkpoint"))
    format <- ckpt$format %||% "full"
    if (identical(format, "fp8") && !.st_can_write("float8_e4m3fn")) {
        stop("This fp8 artifact needs float8 support the installed ",
             "safetensors lacks (mlverse/safetensors#13, merged upstream ",
             "but not yet on CRAN). Install the development version of ",
             "safetensors from GitHub, or rebuild the artifact as nf4.",
             call. = FALSE)
    }
    hooks <- .flux_family_hooks(ckpt$config)

    args <- utils::modifyList(hooks$args_fn(ckpt$config), list(...))
    model <- do.call(hooks$model_fn, args)

    if (format == "full") {
        model$to(dtype = .flux_dtype(dtype))
        res <- ltx23_load_group(ckpt, ckpt$keys, model, verbose = verbose)
        if (length(res$unmapped) || length(res$unfilled)) {
            stop("FLUX transformer load: ", length(res$unmapped),
                 " unmapped keys, ", length(res$unfilled), " unfilled params")
        }
        model$to(device = device)
        model$eval()
        return(model)
    }

    if (!format %in% c("nf4", "fp8")) {
        stop("Unknown checkpoint format: ", format)
    }
    # Residents (embedders, norms, biases) in the compute dtype: the
    # quantized linears dequantize into the input's dtype at forward, so
    # the two must agree (bfloat16 on GPU, float32 for CPU compute)
    model$to(dtype = .flux_dtype(dtype))

    if (format == "nf4") {
        sib_suffix <- "_absmax"
    } else {
        sib_suffix <- "_scale"
    }
    sib_keys <- ckpt$keys[endsWith(ckpt$keys, paste0(".weight", sib_suffix))]
    main_keys <- setdiff(ckpt$keys, sib_keys)

    dests <- c(model$named_parameters(), model$named_buffers())
    filled <- character(0)
    unmapped <- character(0)

    torch::with_no_grad({
        for (i in seq_along(main_keys)) {
            key <- main_keys[[i]]

            if (hooks$is_quant_key(key) &&
                        paste0(key, sib_suffix) %in% sib_keys) {
                segments <- strsplit(key, ".", fixed = TRUE)[[1]]
                parent <- .ltx23_walk_module(model, utils::head(segments, -2L))
                leaf <- segments[length(segments) - 1L]
                old <- .ltx23_walk_module(parent, leaf)
                if (is.null(old)) {
                    unmapped <- c(unmapped, key)
                    next
                }
                w_shape <- old$weight$shape
                has_bias <- !is.null(old$bias)
                quant_mod <- if (format == "nf4") {
                    ltx23_nf4_linear(w_shape[1], w_shape[2], bias = has_bias)
                } else {
                    ltx23_fp8_linear(w_shape[1], w_shape[2], bias = has_bias)
                }
                if (has_bias) {
                    # Adopt the original bias parameter; its checkpoint key
                    # loads through the pre-swap destination map
                    quant_mod$bias <- old$bias
                }
                if (format == "nf4") {
                    quant_mod$set_nf4_weight(
                        ckpt$handle$get_tensor(key),
                        ckpt$handle$get_tensor(paste0(key, sib_suffix))
                    )
                } else {
                    quant_mod$set_fp8_weight(
                        ckpt$handle$get_tensor(key),
                        ckpt$handle$get_tensor(paste0(key, sib_suffix)),
                        pin = pin
                    )
                }
                do.call(`$<-`, list(parent, leaf, quant_mod))
                filled <- c(filled, key)
            } else {
                dest <- dests[[key]]
                if (is.null(dest)) {
                    unmapped <- c(unmapped, key)
                    next
                }
                dest$copy_(ckpt$handle$get_tensor(key))
                filled <- c(filled, key)
            }

            if (i %% 100L == 0L) {
                gc(verbose = FALSE)
                if (verbose && i %% 500L == 0L) {
                    message(sprintf("  loaded %d/%d transformer tensors", i,
                                    length(main_keys)))
                }
            }
        }
    })
    gc(verbose = FALSE)

    if (length(unmapped)) {
        stop("FLUX ", format, " load: ", length(unmapped),
             " unmapped keys, e.g. ",
             paste(utils::head(unmapped, 3), collapse = ", "))
    }
    # Weight params replaced by quantized modules won't be "filled"
    expected_missing <- hooks$is_quant_key(names(dests))
    unfilled <- setdiff(names(dests)[!expected_missing], filled)
    if (length(unfilled)) {
        stop("FLUX ", format, " load: ", length(unfilled),
             " unfilled params, e.g. ",
             paste(utils::head(unfilled, 3), collapse = ", "))
    }

    # NF4: everything (packed buffers included) onto the device.
    # FP8: residents move; the plain-field fp8 weights stay CPU-side
    # unless fp8_resident moves them (small models that fit on the GPU).
    model$to(device = device)
    if (format == "fp8" && fp8_resident) {
        .flux_fp8_to_device(model, device)
    }
    model$eval()
    # Block intermediates are large at image resolutions; per-block gc
    # keeps the quantized-linear temporaries bounded
    options(diffuseR.block_gc = TRUE)
    if (verbose) {
        message("FLUX transformer ready (", format, ") on ", device)
    }
    model
}

# safetensors capability gating: dtype write probes, precision "auto"
# resolution, and the quantizer's resident-dtype fallback for CRAN
# safetensors (no bfloat16 write, no float8).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

can_write <- diffuseR:::.st_can_write
resolve_precision <- diffuseR:::.flux_resolve_precision

# --- capability probe -------------------------------------------------------------

expect_true(is.logical(can_write("bfloat16")))
expect_true(is.logical(can_write("float8_e4m3fn")))

# Option override wins over the probe (and its cache)
options(diffuseR.st_caps = list(bfloat16 = FALSE, float8_e4m3fn = FALSE))
expect_false(can_write("bfloat16"))
expect_false(can_write("float8_e4m3fn"))
options(diffuseR.st_caps = list(float8_e4m3fn = TRUE))
expect_true(can_write("float8_e4m3fn"))
options(diffuseR.st_caps = NULL)

# --- precision "auto" resolution ---------------------------------------------------

# Explicit precision passes through untouched
expect_equal(resolve_precision("nf4"), "nf4")
expect_equal(resolve_precision("fp8"), "fp8")

# An existing artifact wins, fp8 preferred. Pin the fp8 capability so
# the assertion is hermetic: on stock CRAN safetensors the ambient probe
# is FALSE and resolve_precision correctly refuses the fp8 artifact.
options(diffuseR.st_caps = list(float8_e4m3fn = TRUE))
prefix <- file.path(tempdir(), "stcaps-test-")
nf4_dir <- paste0(prefix, "nf4")
fp8_dir <- paste0(prefix, "fp8")
dir.create(nf4_dir, showWarnings = FALSE)
writeLines("{}", file.path(nf4_dir, "manifest.json"))
expect_equal(resolve_precision("auto", prefix), "nf4")
dir.create(fp8_dir, showWarnings = FALSE)
writeLines("{}", file.path(fp8_dir, "manifest.json"))
expect_equal(resolve_precision("auto", prefix), "fp8")
unlink(c(nf4_dir, fp8_dir), recursive = TRUE)
options(diffuseR.st_caps = NULL)

# No artifact: capability decides
options(diffuseR.st_caps = list(float8_e4m3fn = FALSE))
expect_equal(resolve_precision("auto", prefix), "nf4")
options(diffuseR.st_caps = list(float8_e4m3fn = TRUE))
expect_equal(resolve_precision("auto", prefix), "fp8")
options(diffuseR.st_caps = NULL)

# An fp8 artifact present but unreadable (CRAN safetensors) is NOT
# selected - it would fail at read time. Only nf4 (or a build) is safe.
dir.create(fp8_dir, showWarnings = FALSE)
writeLines("{}", file.path(fp8_dir, "manifest.json"))
options(diffuseR.st_caps = list(float8_e4m3fn = FALSE))
expect_equal(resolve_precision("auto", prefix), "nf4")
# With float8 support the same fp8 artifact IS selected
options(diffuseR.st_caps = list(float8_e4m3fn = TRUE))
expect_equal(resolve_precision("auto", prefix), "fp8")
options(diffuseR.st_caps = NULL)
unlink(fp8_dir, recursive = TRUE)

# --- quantizer gates (tiny checkpoint, CRAN-safetensors emulation) ------------------

ckpt_dir <- system.file("tinytest", "fixtures", "zimage_tiny_ckpt",
  package = "diffuseR")
if (ckpt_dir == "") ckpt_dir <- "fixtures/zimage_tiny_ckpt"
if (!dir.exists(ckpt_dir)) exit_file("zimage tiny checkpoint missing")

options(diffuseR.st_caps = list(bfloat16 = FALSE, float8_e4m3fn = FALSE))

# fp8 quantization is refused with an actionable error
expect_error(
  flux_quantize(ckpt_dir, file.path(tempdir(), "stcaps-fp8"),
    format = "fp8", verbose = FALSE),
  pattern = "float8"
)

# NF4 quantization falls back to float32 residents
nf4_out <- file.path(tempdir(), "stcaps-nf4")
unlink(nf4_out, recursive = TRUE)
expect_message(
  manifest <- flux_quantize(ckpt_dir, nf4_out, format = "nf4",
    verbose = FALSE),
  pattern = "float32"
)
ck <- flux_open_quantized(nf4_out)
resident <- ck$handle$get_tensor("cap_embedder.1.weight")
expect_equal(as.character(resident$dtype), "Float")

# And the artifact still loads and runs
model <- flux_load_transformer(ck, device = "cpu", dtype = "float32",
  verbose = FALSE)
out <- torch::with_no_grad(model(
  torch::torch_randn(4L, 1L, 12L, 20L),
  torch::torch_tensor(0.5)$reshape(1L),
  torch::torch_randn(37L, 24L)
))
expect_equal(out$shape, c(4L, 1L, 12L, 20L))
ltx23_release_dequant_buffers()

options(diffuseR.st_caps = NULL)
options(diffuseR.block_gc = NULL)
unlink(c(nf4_out, file.path(tempdir(), "stcaps-fp8")), recursive = TRUE)

# --- load-path guard: fp8 artifact opened without float8 support --------------------
# Build a real fp8 artifact (needs actual float8 support), then force the
# probe to "no float8" and confirm loading errors actionably rather than
# with a raw dtype failure.
if (can_write("float8_e4m3fn")) {
  fp8_out <- file.path(tempdir(), "stcaps-loadguard-fp8")
  unlink(fp8_out, recursive = TRUE)
  flux_quantize(ckpt_dir, fp8_out, format = "fp8", verbose = FALSE)
  options(diffuseR.st_caps = list(float8_e4m3fn = FALSE))
  expect_error(
    flux_load_transformer(flux_open_quantized(fp8_out), device = "cpu",
      verbose = FALSE),
    pattern = "float8"
  )
  options(diffuseR.st_caps = NULL)
  unlink(fp8_out, recursive = TRUE)
}

#' Quantize a Gemma3 text encoder to NF4 shards
#'
#' Streams the HuggingFace Gemma3 checkpoint tensor by tensor. The
#' language model's projection weights (q/k/v/o and gate/up/down, ~11B
#' of the 12B parameters) are stored as NF4 (packed uint8 +
#' \code{<key>_absmax} float32 blocks); embeddings and norms are copied
#' at the resident dtype. Vision-tower and projector weights are
#' dropped - the text encoder never uses them. The result is a ~8 GB
#' artifact that fits a 16 GB card during the encode phase (vs 45 GB
#' of host RAM for the fp32 CPU path).
#'
#' Keys are stored normalized (\code{language_model.} / \code{model.}
#' prefixes stripped), matching the module tree of
#' \code{\link{gemma3_text_model}}.
#'
#' @param model_path HuggingFace snapshot directory (config.json +
#'   model-*.safetensors).
#' @param output_dir Output directory for shards + manifest (default:
#'   \code{gemma3-nf4} under \code{tools::R_user_dir}).
#' @param shard_bytes Numeric. Target shard size in bytes; the 1.9e9
#'   default keeps shards readable by stock CRAN safetensors.
#' @param force Logical. Re-quantize even if a valid manifest exists.
#' @param verbose Logical.
#'
#' @return Invisibly, the manifest list.
#'
#' @export
gemma3_quantize_nf4 <- function(model_path, output_dir = NULL,
                                shard_bytes = 1.9e9, force = FALSE,
                                verbose = TRUE) {
    model_path <- path.expand(model_path)
    if (is.null(output_dir)) {
        output_dir <- file.path(tools::R_user_dir("diffuseR", "data"),
                                "gemma3-nf4")
    }

    manifest_path <- file.path(output_dir, "manifest.json")
    if (!force && file.exists(manifest_path)) {
        manifest <- jsonlite::fromJSON(manifest_path)
        if (identical(manifest$format, "nf4") &&
            all(file.exists(file.path(output_dir, manifest$shards)))) {
            if (verbose) {
                message("NF4 artifact already present: ", output_dir)
            }
            return(invisible(manifest))
        }
    }

    config_path <- file.path(model_path, "config.json")
    if (!file.exists(config_path)) {
        stop("Config file not found: ", config_path)
    }
    config_raw <- jsonlite::fromJSON(config_path)
    if (!is.null(config_raw$text_config)) {
        config_raw <- config_raw$text_config
    }

    resident_dtype <- if (.st_can_write("bfloat16")) {
        torch::torch_bfloat16()
    } else {
        message("safetensors cannot write bfloat16; storing resident ",
                "tensors as float32 (larger artifact, same results)")
        torch::torch_float32()
    }

    opened <- .flux_open_sharded_dir(model_path, "model")
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

    shard <- list()
    shard_size <- 0
    shard_files <- character(0)
    n_cast <- 0L

    flush_shard <- function() {
        if (!length(shard)) {
            return()
        }
        fname <- sprintf("gemma3-nf4-%05d.safetensors",
                         length(shard_files) + 1L)
        safetensors::safe_save_file(shard, file.path(output_dir, fname))
        shard_files[[length(shard_files) + 1L]] <<- fname
        if (verbose) {
            message(sprintf("  wrote %s (%.2f GB, %d tensors)", fname,
                            shard_size / 1e9, length(shard)))
        }
        shard <<- list()
        shard_size <<- 0
        gc(verbose = FALSE)
    }

    for (key in opened$keys) {
        if (grepl("^(vision_tower|multi_modal_projector|lm_head)\\.", key)) {
            next
        }
        norm_key <- sub("^model\\.", "", sub("^language_model\\.", "", key))
        tensor <- opened$handle$get_tensor(key)

        if (.gemma3_is_quant_key(norm_key)) {
            torch::with_no_grad({
                q <- ltx23_nf4_quantize(tensor)
                shard[[norm_key]] <- q$packed
                shard[[paste0(norm_key, "_absmax")]] <- q$absmax
            })
            shard_size <- shard_size + prod(tensor$shape) * 0.5625
            n_cast <- n_cast + 1L
        } else {
            shard[[norm_key]] <- tensor$to(dtype = resident_dtype)
            shard_size <- shard_size + prod(tensor$shape) * 2
        }
        rm(tensor)
        if (shard_size >= shard_bytes) {
            flush_shard()
        }
    }
    flush_shard()

    manifest <- list(model = "gemma3", format = "nf4", shards = shard_files,
                     n_cast = n_cast, config = config_raw)
    jsonlite::write_json(manifest, manifest_path, auto_unbox = TRUE,
                         pretty = TRUE)
    if (verbose) {
        message(sprintf("Quantized %d weights to NF4 across %d shards: %s",
                        n_cast, length(shard_files), output_dir))
    }
    invisible(manifest)
}

# The 7 per-layer projection weights carrying ~11B of the 12B params
.gemma3_is_quant_key <- function(key) {
    grepl(paste0("^layers\\.[0-9]+\\.(self_attn\\.(q|k|v|o)_proj|",
                 "mlp\\.(gate|up|down)_proj)\\.weight$"), key)
}

#' Load a Gemma3 text encoder from an NF4 artifact
#'
#' Builds the model as a skeleton at the compute dtype, swaps the
#' projection linears for NF4 modules filled from the artifact
#' (dequantized per forward through the shared byte-LUT), copies the
#' residents, and hard-errors on any parameter the artifact does not
#' fill.
#'
#' @param artifact_dir Directory produced by
#'   \code{\link{gemma3_quantize_nf4}}.
#' @param device "cuda" or "cpu".
#' @param dtype Compute dtype ("bfloat16" default).
#' @param pin Logical. When loading to the CPU, page-lock the weights
#'   so \code{\link{encode_with_gemma3}} can swap the model to the GPU
#'   at DMA speed per encode and back for free (see
#'   \code{\link{staging}}). Default follows
#'   \code{options(diffuseR.pin_staging)}.
#' @param verbose Logical.
#'
#' @return A \code{gemma3_text_model} ready for
#'   \code{\link{encode_with_gemma3}}.
#'
#' @export
load_gemma3_nf4 <- function(artifact_dir, device = "cuda",
                            dtype = "bfloat16",
                            pin = getOption("diffuseR.pin_staging", TRUE),
                            verbose = TRUE) {
    ckpt <- ltx23_open_fp8_checkpoint(artifact_dir)
    manifest <- jsonlite::fromJSON(file.path(artifact_dir, "manifest.json"))
    if (!identical(manifest$model, "gemma3")) {
        stop("Not a Gemma3 NF4 artifact (manifest model is ",
             manifest$model %||% "absent", "): ", artifact_dir)
    }
    config <- .gemma3_build_config(manifest$config)

    torch_dtype <- switch(dtype,
                          "float32" = torch::torch_float32(),
                          "float16" = torch::torch_float16(),
                          "bfloat16" = torch::torch_bfloat16(),
                          torch::torch_bfloat16())

    if (verbose) {
        message(sprintf("Creating Gemma3 NF4 model: %d layers, hidden_size=%d",
                        config$num_hidden_layers, config$hidden_size))
    }
    model <- .construct_skeleton(gemma3_text_model, config, dtype = torch_dtype)

    cast_keys <- grep("_absmax$", ckpt$keys, value = TRUE)
    cast_keys <- sub("_absmax$", "", cast_keys)
    torch::with_no_grad({
        for (key in cast_keys) {
            segs <- strsplit(sub("\\.weight$", "", key), ".", fixed = TRUE)[[1]]
            # layers.<l>.<group>.<proj>
            layer <- model$layers[[as.integer(segs[2]) + 1L]]
            parent <- layer[[segs[3]]]
            old <- parent[[segs[4]]]
            nf4 <- ltx23_nf4_linear(old$weight$shape[1], old$weight$shape[2],
                                    bias = FALSE)
            nf4$set_nf4_weight(ckpt$handle$get_tensor(key),
                               ckpt$handle$get_tensor(paste0(key, "_absmax")))
            do.call(`$<-`, list(parent, segs[4], nf4))
        }

        dests <- model$parameters
        filled <- character(0)
        for (key in setdiff(ckpt$keys, c(cast_keys,
                    paste0(cast_keys, "_absmax")))) {
            dest <- dests[[key]]
            if (is.null(dest)) {
                next
            }
            dest$copy_(ckpt$handle$get_tensor(key))
            filled <- c(filled, key)
        }
        unfilled <- setdiff(names(dests), filled)
        if (length(unfilled)) {
            stop("Gemma3 NF4 load: ", length(unfilled),
                 " parameters not filled, e.g. ",
                 paste(utils::head(unfilled, 3), collapse = ", "))
        }
    })

    model$to(device = device)
    model$eval()
    if (pin && device == "cpu") {
        st <- .pin_component(model)
        if (!is.null(st)) {
            attr(model, "staging") <- st
            if (verbose) {
                message("Gemma3 weights pinned for staged transfer")
            }
        }
    }
    if (verbose) {
        message("Gemma3 NF4 encoder ready on ", device)
    }
    model
}

# gemma3_quantize_nf4 + load_gemma3_nf4: round-trip through a tiny
# on-disk HF-layout checkpoint, projection swap to NF4 modules, forward
# parity within NF4 quantization tolerance, dispatch through
# load_gemma3_text_encoder, and the unfilled-parameter hard error.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

config <- list(
  vocab_size = 64L, hidden_size = 16L, intermediate_size = 32L,
  num_hidden_layers = 2L, num_attention_heads = 2L,
  num_key_value_heads = 1L, head_dim = 8L, query_pre_attn_scalar = 8L,
  sliding_window = 4L, sliding_window_pattern = 2L
)

ref_config <- gemma3_config_ltx2()
for (nm in names(config)) ref_config[[nm]] <- config[[nm]]
torch::torch_manual_seed(23)
ref <- gemma3_text_model(ref_config)
ref$eval()

# HF-layout snapshot: config.json, one shard, and the weight-map index
# the quantizer's sharded-dir opener requires
src <- file.path(tempdir(), "gemma3-nf4-src")
dir.create(src, showWarnings = FALSE)
jsonlite::write_json(config, file.path(src, "config.json"), auto_unbox = TRUE)
params <- ref$parameters
weights <- stats::setNames(
  lapply(params, function(p) p$detach()$contiguous()),
  paste0("model.", names(params))
)
shard_name <- "model-00001-of-00001.safetensors"
safetensors::safe_save_file(weights, file.path(src, shard_name))
jsonlite::write_json(
  list(weight_map = stats::setNames(
    as.list(rep(shard_name, length(weights))), names(weights))),
  file.path(src, "model.safetensors.index.json"), auto_unbox = TRUE
)

# --- quantize -----------------------------------------------------------------------

art <- file.path(tempdir(), "gemma3-nf4-art")
unlink(art, recursive = TRUE)
# Pin residents to float32 so the exactness assertions below are
# hermetic across CRAN (no bf16 write) and fork safetensors
options(diffuseR.st_caps = list(bfloat16 = FALSE))
manifest <- gemma3_quantize_nf4(src, art, verbose = FALSE)
options(diffuseR.st_caps = NULL)
expect_equal(manifest$model, "gemma3")
# 2 layers x 7 projections
expect_equal(manifest$n_cast, 14L)
expect_true(file.exists(file.path(art, "manifest.json")))

# --- load + forward parity ----------------------------------------------------------

enc <- load_gemma3_nf4(art, device = "cpu", dtype = "float32",
                       verbose = FALSE)
expect_true(inherits(enc$layers[[1]]$self_attn$q_proj, "ltx23_nf4_linear"))
expect_true(inherits(enc$layers[[2]]$mlp$down_proj, "ltx23_nf4_linear"))
# Residents are exact
expect_true(as.numeric((enc$parameters[["embed_tokens.weight"]] -
                        params[["embed_tokens.weight"]])$abs()$max()) == 0)

ids <- torch::torch_randint(1L, 64L, size = c(1L, 5L), dtype = torch::torch_long())
mask <- torch::torch_ones(1L, 5L)
torch::with_no_grad({
  out_ref <- ref(ids, attention_mask = mask, output_hidden_states = TRUE)
  out_nf4 <- enc(ids, attention_mask = mask, output_hidden_states = TRUE)
})
h_ref <- out_ref$last_hidden_state
h_nf4 <- out_nf4$last_hidden_state
rel <- as.numeric((h_nf4 - h_ref)$abs()$max()) /
  as.numeric(h_ref$abs()$max())
# 4-bit weights: outputs agree to quantization noise, not float noise.
# On this tiny 2-layer model the worst element sits ~0.15 relative
# (measured 0.1505); the direction check below is the real quality
# gate, and full-model embedding parity is validated on GPU.
expect_true(rel < 0.25, info = sprintf("relative max diff %.4f", rel))
cs <- as.numeric(torch::nnf_cosine_similarity(
  h_ref$flatten()$unsqueeze(1L), h_nf4$flatten()$unsqueeze(1L)))
expect_true(cs > 0.99, info = sprintf("cosine %.5f", cs))

# --- dispatch through load_gemma3_text_encoder --------------------------------------

enc2 <- load_gemma3_text_encoder(art, device = "cpu", dtype = "float32",
                                 verbose = FALSE)
expect_true(inherits(enc2$layers[[1]]$self_attn$q_proj, "ltx23_nf4_linear"))

# --- an artifact missing a resident is a hard error ---------------------------------

art2 <- file.path(tempdir(), "gemma3-nf4-art2")
unlink(art2, recursive = TRUE)
dir.create(art2)
m2 <- jsonlite::fromJSON(file.path(art, "manifest.json"))
for (s in m2$shards) {
  w <- safetensors::safe_load_file(file.path(art, s), framework = "torch")
  w[["norm.weight"]] <- NULL
  safetensors::safe_save_file(w, file.path(art2, s))
}
jsonlite::write_json(m2, file.path(art2, "manifest.json"), auto_unbox = TRUE)
expect_error(
  load_gemma3_nf4(art2, device = "cpu", verbose = FALSE),
  "not filled"
)

unlink(c(src, art, art2), recursive = TRUE)

# Skeleton construction for checkpoint loading
#
# Building a module allocates and initializes every weight, only for
# the loader to overwrite each one from the checkpoint or swap the
# layer for a quantized module. At LTX-2.3 scale that wrote ~90 GB of
# float32 pages before the first checkpoint byte was read (and the
# follow-up bfloat16 cast held both copies alive), OOM-killing a
# 125 GB machine. A skeleton skips the writes: linear and embedding
# weights are allocated with torch_empty() and never touched (the
# pages stay virtual), and tensor creation defaults to the target
# dtype so no post-construction cast pass is needed.

.noinit_state <- new.env(parent = emptyenv())
.noinit_state$active <- FALSE

# Construct ctor(...) as a load skeleton: linear_noinit/embedding_noinit
# weights stay uninitialized and the torch default dtype is `dtype` for
# the duration. Only for modules whose parameters and buffers are all
# filled from a checkpoint afterwards (the group loaders validate
# exactly that); constructor-computed buffers keep their values but
# follow `dtype`, so precision-sensitive constructors (the vocoder's
# filter banks) must not go through here.
.construct_skeleton <- function(ctor, ..., dtype = torch::torch_bfloat16()) {
    old_dtype <- torch::torch_get_default_dtype()
    torch::torch_set_default_dtype(dtype)
    .noinit_state$active <- TRUE
    on.exit(
            {
        torch::torch_set_default_dtype(old_dtype)
        .noinit_state$active <- FALSE
    },
            add = TRUE
    )
    ctor(...)
}

# Drop-in nn_linear that skips reset_parameters() inside
# .construct_skeleton(); bare construction initializes exactly like
# torch::nn_linear, so random-weight forwards in tests are unaffected.
linear_noinit <- torch::nn_module(
                                  "linear_noinit",
                                  initialize = function(in_features, out_features, bias = TRUE) {
    self$in_features <- in_features
    self$out_features <- out_features
    self$weight <- torch::nn_parameter(torch::torch_empty(out_features,
            in_features))
    if (bias) {
        self$bias <- torch::nn_parameter(torch::torch_empty(out_features))
    } else {
        self$bias <- NULL
    }
    if (!.noinit_state$active) {
        self$reset_parameters()
    }
},
                                  reset_parameters = function() {
    torch::nn_init_kaiming_uniform_(self$weight, a = sqrt(5))
    if (!is.null(self$bias)) {
        bound <- 1 / sqrt(self$in_features)
        torch::nn_init_uniform_(self$bias, -bound, bound)
    }
},
                                  forward = function(input) {
    torch::nnf_linear(input, self$weight, self$bias)
}
)

# Drop-in nn_embedding (no padding/norm options) with the same skip.
embedding_noinit <- torch::nn_module(
                                     "embedding_noinit",
                                     initialize = function(num_embeddings, embedding_dim) {
    self$num_embeddings <- num_embeddings
    self$embedding_dim <- embedding_dim
    self$weight <- torch::nn_parameter(torch::torch_empty(num_embeddings,
            embedding_dim))
    if (!.noinit_state$active) {
        torch::nn_init_normal_(self$weight)
    }
},
                                     forward = function(input) {
    torch::nnf_embedding(input, self$weight)
}
)

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

# Test native CLIP text encoder

# Skip if models not available
skip_if_not <- function(cond, msg) {
  if (!cond) exit_file(msg)
}

library(torch)

# Test module architecture
encoder <- text_encoder_native()
expect_true(inherits(encoder, "nn_module"))

# Check parameters exist
params <- names(encoder$parameters)
expect_true(length(params) > 0)
expect_true("token_embedding.weight" %in% params)
expect_true("position_embedding" %in% params)
expect_true("final_layer_norm.weight" %in% params)

# Test forward pass with random weights
with_no_grad({
  tokens <- torch_tensor(matrix(c(49406, 320, 4380, 49407, rep(49407, 73)), nrow = 1),
                         dtype = torch_long())
  out <- encoder(tokens)
  expect_equal(out$shape[1], 1L)
  expect_equal(out$shape[2], 77L)
  expect_equal(out$shape[3], 768L)
})

# Test load_model_component with use_native
model_dir <- tools::R_user_dir("diffuseR", "data")
encoder_file <- file.path(model_dir, "sdxl", "text_encoder-cpu.pt")
skip_if_not(file.exists(encoder_file), "SDXL text_encoder not available")

native_enc <- load_model_component("text_encoder", "sdxl", "cpu", use_native = TRUE)
expect_true(inherits(native_enc, "TextEncoderNative"))

# Verify weight loading (parameter count)
expect_equal(length(native_enc$parameters), 196L)

# Test output equivalency with TorchScript
ts_enc <- torch::jit_load(encoder_file)
tokens <- torch_tensor(matrix(c(49406, 320, 4380, 49407, rep(49407, 73)), nrow = 1),
                       dtype = torch_long())
with_no_grad({
  ts_out <- ts_enc(tokens)
  native_out <- native_enc$forward(tokens)
})

# Check output shapes match
expect_equal(as.integer(ts_out$shape), as.integer(native_out$shape))

# Check outputs are close (within float32 precision)
diff <- (ts_out - native_out)$abs()
max_diff <- as.numeric(diff$max())
expect_true(max_diff < 5.0, info = paste("max_diff:", max_diff))  # relaxed threshold

# Skeleton construction (noinit_modules.R): linear_noinit /
# embedding_noinit initialize exactly like their torch counterparts on
# bare construction, skip initialization inside .construct_skeleton(),
# and the skeleton scope sets/restores the torch default dtype.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)

skel <- diffuseR:::.construct_skeleton
linear_noinit <- diffuseR:::linear_noinit
embedding_noinit <- diffuseR:::embedding_noinit
noinit_state <- diffuseR:::.noinit_state

# --- bare construction initializes like torch -------------------------------------

torch::torch_manual_seed(7)
l <- linear_noinit(8L, 4L)
expect_true(all(is.finite(as.numeric(l$weight))))
expect_true(all(is.finite(as.numeric(l$bias))))
# kaiming_uniform(a = sqrt(5)) bound and the bias bound are both 1/sqrt(fan_in)
expect_true(as.numeric(l$weight$abs()$max()) <= 1 / sqrt(8) + 1e-6)
expect_true(as.numeric(l$bias$abs()$max()) <= 1 / sqrt(8) + 1e-6)

l_nobias <- linear_noinit(8L, 4L, bias = FALSE)
expect_null(l_nobias$bias)

e <- embedding_noinit(16L, 4L)
expect_true(all(is.finite(as.numeric(e$weight))))

# --- forward parity with torch modules --------------------------------------------

ref <- torch::nn_linear(8L, 4L)
torch::with_no_grad({
  ref$weight$copy_(l$weight)
  ref$bias$copy_(l$bias)
})
x <- torch::torch_randn(2L, 8L)
expect_true(as.numeric((l(x) - ref(x))$abs()$max()) == 0)

eref <- torch::nn_embedding(16L, 4L)
torch::with_no_grad(eref$weight$copy_(e$weight))
idx <- torch::torch_randint(1L, 16L, size = c(2L, 3L), dtype = torch::torch_long())
expect_true(as.numeric((e(idx) - eref(idx))$abs()$max()) == 0)

# --- skeleton scope: target dtype, state restored ----------------------------------

old_dtype <- torch::torch_get_default_dtype()
m <- skel(linear_noinit, 8L, 4L, dtype = torch::torch_float16())
expect_true(m$weight$dtype == torch::torch_float16())
expect_true(m$bias$dtype == torch::torch_float16())
expect_true(torch::torch_get_default_dtype() == old_dtype)
expect_false(noinit_state$active)

me <- skel(embedding_noinit, 16L, 4L, dtype = torch::torch_bfloat16())
expect_true(me$weight$dtype == torch::torch_bfloat16())

# state restored even when the constructor throws
expect_error(skel(function() stop("boom")), "boom")
expect_false(noinit_state$active)
expect_true(torch::torch_get_default_dtype() == old_dtype)

# --- skeleton weights are usable once filled --------------------------------------

m2 <- skel(linear_noinit, 4L, 2L, dtype = torch::torch_float32())
torch::with_no_grad({
  m2$weight$fill_(0.5)
  m2$bias$fill_(0)
})
y <- m2(torch::torch_ones(1L, 4L))
expect_true(as.numeric((y - 2)$abs()$max()) < 1e-6)

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

# Test native UNet

# Skip if models not available
skip_if_not <- function(cond, msg) {
  if (!cond) exit_file(msg)
}

library(diffuseR)
library(torch)

# Test module architecture
unet <- unet_native(
  in_channels = 4L,
  out_channels = 4L,
  block_out_channels = c(320L, 640L, 1280L, 1280L),
  layers_per_block = 2L,
  cross_attention_dim = 1024L,
  attention_head_dim = 64L
)
expect_true(inherits(unet, "nn_module"))

# Check parameters exist
params <- names(unet$parameters)
expect_true(length(params) > 0)
expect_true("conv_in.weight" %in% params)
expect_true("conv_out.weight" %in% params)
expect_true("time_embedding_linear_1.weight" %in% params)

# Test forward pass with random weights
with_no_grad({
  sample <- torch_randn(c(1L, 4L, 32L, 32L))
  timestep <- torch_tensor(c(500L))
  encoder_hidden_states <- torch_randn(c(1L, 77L, 1024L))
  out <- unet(sample, timestep, encoder_hidden_states)
  expect_equal(out$shape[1], 1L)
  expect_equal(out$shape[2], 4L)
  expect_equal(out$shape[3], 32L)
  expect_equal(out$shape[4], 32L)
})

# Test load_model_component with use_native
model_dir <- tools::R_user_dir("diffuseR", "data")
unet_file <- file.path(model_dir, "sd21", "unet-cpu.pt")
skip_if_not(file.exists(unet_file), "SD21 unet not available")

native_unet <- load_model_component("unet", "sd21", "cpu", use_native = TRUE)
expect_true(inherits(native_unet, "UNetNative"))

# Verify weight loading (parameter count)
expect_equal(length(native_unet$parameters), 686L)

# Test output equivalency with TorchScript (deterministic inputs)
ts_unet <- torch::jit_load(unet_file)

# Use deterministic inputs for consistent comparison
sample <- torch_ones(c(1L, 4L, 32L, 32L)) * 0.1
timestep <- torch_tensor(c(500L))
encoder_hidden_states <- torch_ones(c(1L, 77L, 1024L)) * 0.1

with_no_grad({
  ts_out <- ts_unet(sample, timestep, encoder_hidden_states)
  native_out <- native_unet$forward(sample, timestep, encoder_hidden_states)
})

# Check output shapes match
expect_equal(as.integer(ts_out$shape), as.integer(native_out$shape))

# Check outputs are close (within float32 precision for deep attention models)
diff <- (ts_out - native_out)$abs()
max_diff <- as.numeric(diff$max())
mean_diff <- as.numeric(diff$mean())
expect_true(max_diff < 0.1, info = paste("max_diff:", max_diff))  # deterministic inputs
expect_true(mean_diff < 0.02, info = paste("mean_diff:", mean_diff))

# Random (non-constant) inputs at generation resolution. Constant inputs
# leave GroupNorm at its bias and attention uniform, so they cannot catch
# a scrambled timestep embedding; random inputs give cos ~0.82 with that
# bug and ~1.0 when correct.
torch::torch_manual_seed(123)
r_sample <- torch::torch_randn(c(1L, 4L, 64L, 64L))
r_ctx <- torch::torch_randn(c(1L, 77L, 1024L))
with_no_grad({
  ts_r <- ts_unet(r_sample, torch::torch_tensor(c(500L)), r_ctx)
  nt_r <- native_unet$forward(r_sample, torch::torch_tensor(c(500L)), r_ctx)
})
cos_sim <- as.numeric(torch::torch_sum(ts_r * nt_r)$item() /
  (sqrt(torch::torch_sum(ts_r * ts_r)$item()) *
   sqrt(torch::torch_sum(nt_r * nt_r)$item())))
expect_true(cos_sim > 0.999, info = paste("random-input cos:", cos_sim))

#' Generate an image from a text prompt using a diffusion pipeline
#'
#' @param prompt A character string prompt describing the image to generate.
#' @param model_name Name of the model to use (e.g., `"sd21"`).
#' @param ... Additional parameters passed to the diffusion process.
#'
#' @return A tensor or image object, depending on implementation.
#' @export
#'
#' @examples
#' \dontrun{
#' img <- txt2img("a cat wearing sunglasses in space", device = "cuda")
#' }
txt2img <- function(prompt,
                    model_name = c("sd21", "sdxl", "flux1", "flux2", "zimage"),
                    ...) {
    switch(model_name,
           # "sd15" = txt2img_sd15(prompt, ...),
           "sd21" = txt2img_sd21(prompt, ...),
           "sdxl" = txt2img_sdxl(prompt, ...),
           "flux1" = txt2img_flux(prompt, ...),
           "flux2" = txt2img_flux2(prompt, ...),
           "zimage" = txt2img_zimage(prompt, ...),
           # "sd3" = txt2img_sd3(prompt, ...),
           stop("Unsupported model: ", model_name)
    )
}

#' QuickGELU activation
#'
#' GELU approximation used by OpenAI CLIP: x * sigmoid(1.702 * x)
#' @param x Input tensor
#' @keywords internal
quick_gelu <- function(x) {
    x * torch::torch_sigmoid(1.702 * x)
}

#' CLIP MLP Block
#'
#' Feed-forward network with configurable activation
#' @param in_dim Input dimension
#' @param hidden_dim Hidden dimension
#' @param gelu_type GELU variant: "tanh" (tanh approximation), "quick" (QuickGELU), "exact" (standard GELU)
#' @keywords internal
CLIPMLP <- torch::nn_module(
                            "CLIPMLP",

                            initialize = function(
        in_dim,
        hidden_dim,
        gelu_type = "tanh"
    ) {
    self$fc1 <- torch::nn_linear(in_dim, hidden_dim)
    self$fc2 <- torch::nn_linear(hidden_dim, in_dim)
    self$gelu_type <- gelu_type
},

                            forward = function(x) {
    x <- self$fc1(x)
    if (self$gelu_type == "quick") {
        x <- quick_gelu(x)
    } else if (self$gelu_type == "tanh") {
        x <- torch::nnf_gelu(x, approximate = "tanh")
    } else {
        x <- torch::nnf_gelu(x)
    }
    x <- self$fc2(x)
    x
}
)

#' CLIP Attention Block
#'
#' Multi-head self-attention with separate Q/K/V projections (HuggingFace style)
#' @param embed_dim Embedding dimension
#' @param num_heads Number of attention heads
#' @keywords internal
CLIPAttention <- torch::nn_module(
                                  "CLIPAttention",

                                  initialize = function(
        embed_dim,
        num_heads
    ) {
    self$embed_dim <- embed_dim
    self$num_heads <- num_heads
    self$head_dim <- embed_dim %/% num_heads
    self$scale <- self$head_dim ^ (-0.5)

    # Separate Q/K/V projections (HuggingFace CLIPTextModel style)
    self$q_proj <- torch::nn_linear(embed_dim, embed_dim)
    self$k_proj <- torch::nn_linear(embed_dim, embed_dim)
    self$v_proj <- torch::nn_linear(embed_dim, embed_dim)
    self$out_proj <- torch::nn_linear(embed_dim, embed_dim)
},

                                  forward = function(
        x,
        causal_mask = TRUE
    ) {
    batch_size <- x$shape[1]
    seq_len <- x$shape[2]

    # Separate projections
    q <- self$q_proj(x)
    k <- self$k_proj(x)
    v <- self$v_proj(x)

    # Reshape for multi-head attention
    q <- q$view(c(batch_size, seq_len, self$num_heads, self$head_dim))$transpose(2,
        3)
    k <- k$view(c(batch_size, seq_len, self$num_heads, self$head_dim))$transpose(2,
        3)
    v <- v$view(c(batch_size, seq_len, self$num_heads, self$head_dim))$transpose(2,
        3)

    # Scaled dot-product attention
    attn_weights <- torch::torch_matmul(q, k$transpose(3, 4)) * self$scale

    # Apply causal mask
    if (causal_mask) {
        mask <- torch::torch_ones(seq_len, seq_len, device = x$device)$triu(diagonal = 1)$bool()
        attn_weights <- attn_weights$masked_fill(mask, -Inf)
    }

    attn_weights <- torch::nnf_softmax(attn_weights, dim = -1)
    attn_output <- torch::torch_matmul(attn_weights, v)

    # Reshape back
    attn_output <- attn_output$transpose(2, 3)$contiguous()$view(c(batch_size, seq_len, self$embed_dim))
    attn_output <- self$out_proj(attn_output)

    attn_output
}
)

#' CLIP Transformer Block
#'
#' Pre-norm transformer block with attention and MLP (HuggingFace style)
#' @param embed_dim Embedding dimension
#' @param num_heads Number of attention heads
#' @param mlp_dim MLP hidden dimension
#' @param gelu_type GELU variant: "tanh", "quick", or "exact"
#' @keywords internal
CLIPTransformerBlock <- torch::nn_module(
    "CLIPTransformerBlock",

    initialize = function(
                          embed_dim,
                          num_heads,
                          mlp_dim,
                          gelu_type = "tanh"
    ) {
    self$attention <- CLIPAttention(embed_dim, num_heads)
    self$layernorm_1 <- torch::nn_layer_norm(embed_dim)
    self$mlp <- CLIPMLP(embed_dim, mlp_dim, gelu_type)
    self$layernorm_2 <- torch::nn_layer_norm(embed_dim)
},

    forward = function(x) {
    # Pre-norm attention: norm -> attn -> residual
    residual <- x
    x <- self$layernorm_1(x)
    x <- residual + self$attention(x)

    # Pre-norm MLP: norm -> mlp -> residual
    residual <- x
    x <- self$layernorm_2(x)
    x <- residual + self$mlp(x)

    x
}
)

#' Native CLIP Text Encoder
#'
#' Native R torch implementation of CLIP text encoder.
#' Replaces TorchScript for better GPU compatibility.
#'
#' @param vocab_size Vocabulary size (default 49408)
#' @param context_length Maximum sequence length (default 77)
#' @param embed_dim Embedding dimension
#' @param num_layers Number of transformer layers
#' @param num_heads Number of attention heads
#' @param mlp_dim MLP hidden dimension
#' @param apply_final_ln Whether to apply final layer norm (default TRUE).
#'   Set to FALSE to match TorchScript exports that don't include final LN.
#' @param return_penultimate Return the second-to-last transformer block's
#'   output (\code{hidden_states[-2]}, pre-final-LN) instead of the last
#'   layer. This is what SDXL feeds the UNet cross-attention; final LN is
#'   never applied to the penultimate output (default FALSE).
#' @param gelu_type GELU variant: "tanh" (matches the TorchScript exports),
#'   "quick" (HF CLIP ViT-L, used by SDXL text_encoder and FLUX), or "exact"
#'
#' @return An nn_module representing the text encoder
#' @export
text_encoder_native <- torch::nn_module(
                                        "TextEncoderNative",

                                        initialize = function(
        vocab_size = 49408,
        context_length = 77,
        embed_dim = 768,
        num_layers = 12,
        num_heads = 12,
        mlp_dim = 3072,
        apply_final_ln = TRUE,
        return_penultimate = FALSE,
        gelu_type = "tanh"
    ) {
    self$context_length <- context_length
    self$embed_dim <- embed_dim
    self$num_layers <- num_layers
    self$apply_final_ln <- apply_final_ln
    self$return_penultimate <- return_penultimate

    # Embeddings
    self$token_embedding <- torch::nn_embedding(vocab_size, embed_dim)
    self$position_embedding <- torch::nn_parameter(
        torch::torch_zeros(context_length, embed_dim)
    )

    self$transformer_blocks <- torch::nn_module_list()
    for (i in seq_len(num_layers)) {
        self$transformer_blocks$append(
                                       CLIPTransformerBlock(embed_dim, num_heads, mlp_dim,
                gelu_type = gelu_type)
        )
    }

    # Final layer norm
    self$final_layer_norm <- torch::nn_layer_norm(embed_dim)
},

                                        forward = function(input_ids) {
    # Move input to model's device
    input_ids <- input_ids$to(device = self$token_embedding$weight$device)

    batch_size <- input_ids$shape[1]
    seq_length <- input_ids$shape[2]

    # Token + position embeddings
    # Add 1 because R torch is 1-indexed but tokens are 0-indexed (Python convention)
    token_embeds <- self$token_embedding(input_ids + 1L)
    pos_embeds <- self$position_embedding[1:seq_length,]$unsqueeze(1)$expand(c(batch_size, -1, -1))
    hidden_states <- token_embeds + pos_embeds

    # Transformer layers. Capture the input to the last block, which is
    # the output of the second-to-last block = hidden_states[-2] (the
    # penultimate layer SDXL feeds the UNet).
    n <- self$num_layers
    penultimate <- NULL
    for (i in seq_len(n)) {
        if (i == n) {
            penultimate <- hidden_states
        }
        hidden_states <- self$transformer_blocks[[i]](hidden_states)
    }

    if (self$return_penultimate) {
        return(penultimate)
    }

    # Conditionally apply final layer norm to match TorchScript behavior
    if (self$apply_final_ln) {
        hidden_states <- self$final_layer_norm(hidden_states)
    }

    hidden_states
}
)

#' Detect text encoder architecture from TorchScript file
#'
#' @param torchscript_path Path to TorchScript encoder .pt file
#' @return List with vocab_size, context_length, embed_dim, num_layers, num_heads, mlp_dim
#' @keywords internal
detect_text_encoder_architecture <- function(torchscript_path) {
    ts_encoder <- torch::jit_load(torchscript_path)
    ts_params <- ts_encoder$parameters
    param_names <- names(ts_params)

    # Detect prefix style (SD21 vs SDXL)
    if (any(grepl("^text_encoder\\.", param_names))) {
        prefix <- "text_encoder.text_model."
    } else if (any(grepl("^enc\\.", param_names))) {
        prefix <- "enc.text_model."
    } else {
        stop("Unknown TorchScript parameter prefix")
    }

    # Get dimensions from embeddings
    tok_emb <- ts_params[[paste0(prefix, "embeddings.token_embedding.weight")]]
    pos_emb <- ts_params[[paste0(prefix,
                                 "embeddings.position_embedding.weight")]]
    fc1 <- ts_params[[paste0(prefix, "encoder.layers.0.mlp.fc1.weight")]]

    vocab_size <- as.integer(tok_emb$shape[1])
    embed_dim <- as.integer(tok_emb$shape[2])
    context_length <- as.integer(pos_emb$shape[1])
    mlp_dim <- as.integer(fc1$shape[1])

    # Count layers
    layer_params <- grep("encoder\\.layers\\.[0-9]+\\.", param_names, value = TRUE)
    layer_nums <- unique(as.integer(gsub(".*encoder\\.layers\\.([0-9]+)\\..*", "\\1", layer_params)))
    num_layers <- length(layer_nums)

    # Infer heads from embed_dim (typical head_dim is 64)
    num_heads <- embed_dim %/% 64L

    # Detect if final layer norm is applied in TorchScript output
    # by checking output range with test input
    tokens <- torch::torch_tensor(matrix(c(49406, 320, 4380, 49407, rep(49407, 73)), nrow = 1),
                                  dtype = torch::torch_long())
    torch::with_no_grad({
        test_out <- ts_encoder(tokens)
    })
    # Handle both single tensor (text_encoder) and list output (text_encoder2)
    if (is.list(test_out)) {
        hidden_states <- test_out[[1]]
    } else {
        hidden_states <- test_out
    }
    # If output range is large (>100), final LN is NOT applied
    max_abs <- max(abs(as.numeric(hidden_states$min())), abs(as.numeric(hidden_states$max())))
    apply_final_ln <- max_abs < 100

    list(
         vocab_size = vocab_size,
         context_length = context_length,
         embed_dim = embed_dim,
         num_layers = num_layers,
         num_heads = num_heads,
         mlp_dim = mlp_dim,
         prefix = prefix,
         apply_final_ln = apply_final_ln
    )
}

#' Pooled CLIP output at the EOS position
#'
#' The HF CLIPTextModel pooler_output: the final-layer-norm hidden state
#' at the EOS token position, located by argmax over the token ids (EOS
#' is the highest id in the CLIP vocab, and causal attention makes any
#' padding after it irrelevant). No text projection is applied - this is
#' what FLUX uses as pooled_projections.
#'
#' @param hidden_states Final-LN hidden states [B, S, D] from
#'   \code{\link{text_encoder_native}} (with \code{apply_final_ln = TRUE}).
#' @param input_ids Token ids [B, S] (0-based, as fed to the encoder).
#'
#' @return Tensor [B, D].
#'
#' @export
clip_pooled_output <- function(hidden_states, input_ids) {
    input_ids <- input_ids$to(device = hidden_states$device)
    eos_indices <- torch::torch_argmax(input_ids, dim = 2L, keepdim = TRUE)
    hidden_states$gather(
                         dim = 2L,
                         index = eos_indices$unsqueeze(-1L)$expand(c(-1L, -1L,
                hidden_states$shape[3]))
    )$squeeze(2L)
}

# HF CLIPTextModel state-dict key -> native text-encoder parameter name.
# Shared by the encoder-1 and encoder-2 safetensors loaders (the diffusers
# key layout is identical; only encoder 2 adds a top-level text_projection,
# handled by its own remap wrapper).
.clip_st_remap_key <- function(key) {
    key <- sub("^text_model\\.", "", key)
    key <- sub("^embeddings\\.token_embedding\\.", "token_embedding.", key)
    key <- sub("^embeddings\\.position_embedding\\.weight$",
               "position_embedding", key)
    key <- gsub("^encoder\\.layers\\.", "transformer_blocks.", key)
    key <- gsub("\\.self_attn\\.", ".attention.", key)
    key <- gsub("\\.layer_norm1\\.", ".layernorm_1.", key)
    key <- gsub("\\.layer_norm2\\.", ".layernorm_2.", key)
    key
}

# Shared safetensors loader for the native CLIP encoders: open
# model.safetensors, map each HF key through `remap`, copy into the
# matching native parameter, and fail loudly on any unmapped key or
# unfilled native parameter.
.load_clip_st <- function(native_encoder, path, remap, label = "CLIP",
                          verbose = TRUE) {
    path <- path.expand(path)
    if (dir.exists(path)) {
        path <- file.path(path, "model.safetensors")
    }
    handle <- safetensors::safetensors$new(path, framework = "torch")
    keys <- setdiff(handle$keys(), "__metadata__")
    # Non-parameter buffers in some exports
    keys <- keys[!endsWith(keys, "position_ids")]

    dests <- native_encoder$named_parameters()
    filled <- character(0)
    unmapped <- character(0)
    torch::with_no_grad({
        for (key in keys) {
            native_name <- remap(key)
            dest <- dests[[native_name]]
            if (is.null(dest)) {
                unmapped <- c(unmapped, key)
                next
            }
            dest$copy_(handle$get_tensor(key))
            filled <- c(filled, native_name)
        }
    })

    unfilled <- setdiff(names(dests), filled)
    if (length(unmapped)) {
        stop(label, " safetensors load: ", length(unmapped),
             " unmapped keys, e.g. ",
             paste(utils::head(unmapped, 3), collapse = ", "))
    }
    if (length(unfilled)) {
        stop(label, " safetensors load: ", length(unfilled),
             " unfilled params, e.g. ",
             paste(utils::head(unfilled, 3), collapse = ", "))
    }
    if (verbose) {
        message("Loaded ", length(filled), " ", label, " parameters from ",
                path)
    }
    invisible(native_encoder)
}

#' Load HF safetensors weights into the native CLIP text encoder
#'
#' Loads a HuggingFace CLIPTextModel \code{model.safetensors} (e.g.
#' FLUX.1-schnell's \code{text_encoder} or SDXL's \code{text_encoder})
#' into \code{\link{text_encoder_native}}, reusing the TorchScript key
#' remaps minus the export prefixes.
#'
#' @param native_encoder Native text encoder module
#' @param path Path to model.safetensors (or a directory containing it)
#' @param verbose Print loading progress
#'
#' @return The native encoder with loaded weights (invisibly)
#' @export
load_text_encoder_safetensors <- function(native_encoder, path,
    verbose = TRUE) {
    .load_clip_st(native_encoder, path, .clip_st_remap_key, "CLIP",
                  verbose = verbose)
}

# Detect a CLIP text encoder's architecture from a diffusers
# config.json (the config counterpart to
# detect_text_encoder_architecture, which reads a TorchScript file).
# Accepts a directory or a config.json path.
.detect_text_encoder_config <- function(path) {
    if (dir.exists(path)) {
        cfg_path <- file.path(path, "config.json")
    } else {
        cfg_path <- path
    }
    if (!file.exists(cfg_path)) {
        stop("No config.json for the text encoder at ", path)
    }
    cfg <- jsonlite::fromJSON(cfg_path, simplifyVector = TRUE)
    list(vocab_size = as.integer(cfg$vocab_size),
         context_length = as.integer(cfg$max_position_embeddings),
         embed_dim = as.integer(cfg$hidden_size),
         num_layers = as.integer(cfg$num_hidden_layers),
         num_heads = as.integer(cfg$num_attention_heads),
         mlp_dim = as.integer(cfg$intermediate_size))
}

#' Build a native CLIP text encoder from a diffusers safetensors directory
#'
#' Reads the CLIPTextConfig from \code{<dir>/config.json}, constructs
#' \code{\link{text_encoder_native}} to match, and loads
#' \code{model.safetensors} - the safetensors counterpart to the
#' TorchScript text-encoder path (no TorchScript, Blackwell-safe).
#' Handles SD21's OpenCLIP ViT-H and SDXL's CLIP ViT-L (which is the same
#' checkpoint as FLUX's \code{text_encoder}). \code{apply_final_ln}
#' governs only the forward output; the \code{final_layer_norm} weights
#' load either way. Use \code{TRUE} for SD21 and pooled CLIP outputs,
#' \code{FALSE} for the SDXL penultimate-layer prompt embeds.
#'
#' @param path diffusers text_encoder directory (config.json +
#'   model.safetensors) or the config.json path.
#' @param apply_final_ln Apply the final layer norm in forward (default TRUE).
#' @param verbose Print how many parameters were loaded.
#' @param ... Overrides for \code{\link{text_encoder_native}} args (e.g.
#'   \code{gelu_type}).
#'
#' @return The native text encoder in eval mode.
#' @export
text_encoder_native_from_safetensors <- function(path, apply_final_ln = TRUE,
    verbose = TRUE, ...) {
    arch <- .detect_text_encoder_config(path)
    args <- list(vocab_size = arch$vocab_size,
                 context_length = arch$context_length,
                 embed_dim = arch$embed_dim, num_layers = arch$num_layers,
                 num_heads = arch$num_heads, mlp_dim = arch$mlp_dim,
                 apply_final_ln = apply_final_ln)
    args <- utils::modifyList(args, list(...))
    model <- do.call(text_encoder_native, args)
    load_text_encoder_safetensors(model, path, verbose = verbose)
    model$eval()
    model
}

#' Load weights from TorchScript text encoder into native encoder
#'
#' @param native_encoder Native text encoder module
#' @param torchscript_path Path to TorchScript encoder .pt file
#' @param verbose Print loading progress
#'
#' @return The native encoder with loaded weights (invisibly)
#' @export
load_text_encoder_weights <- function(native_encoder, torchscript_path,
                                      verbose = TRUE) {
    ts_encoder <- torch::jit_load(torchscript_path)
    ts_params <- ts_encoder$parameters
    param_names <- names(ts_params)

    # Build mapping from TorchScript names to native names
    # TorchScript format: enc.text_model.encoder.layers.0.self_attn.q_proj.weight
    #                or: text_encoder.text_model.encoder.layers.0.self_attn.q_proj.weight
    # Native format: transformer_blocks.1.attention.q_proj.weight
    remap_key <- function(key) {
        # Strip text_encoder. prefix (SD21 style)
        key <- sub("^text_encoder\\.", "", key)

        # Strip enc. prefix (SDXL style)
        key <- sub("^enc\\.", "", key)

        # Strip text_model. prefix
        key <- sub("^text_model\\.", "", key)

        # Embeddings
        key <- sub("^embeddings\\.token_embedding\\.", "token_embedding.", key)
        # position_embedding.weight -> position_embedding (it's an nn_parameter, not nn_embedding)
        key <- sub("^embeddings\\.position_embedding\\.weight$",
                   "position_embedding", key)

        # Encoder layers -> transformer_blocks (keep 0-indexed)
        key <- gsub("^encoder\\.layers\\.", "transformer_blocks.", key)

        # self_attn -> attention
        key <- gsub("\\.self_attn\\.", ".attention.", key)

        # Layer norms
        key <- gsub("\\.layer_norm1\\.", ".layernorm_1.", key)
        key <- gsub("\\.layer_norm2\\.", ".layernorm_2.", key)

        # Final layer norm
        key <- sub("^final_layer_norm\\.", "final_layer_norm.", key)

        # MLP (HuggingFace uses fc1/fc2 already)
        # No change needed

        key
    }

    loaded <- 0
    skipped <- 0
    unmapped <- character(0)

    torch::with_no_grad({
        for (ts_name in names(ts_params)) {
            native_name <- remap_key(ts_name)

            if (native_name %in% names(native_encoder$parameters)) {
                ts_tensor <- ts_params[[ts_name]]
                native_tensor <- native_encoder$parameters[[native_name]]

                if (all(ts_tensor$shape == native_tensor$shape)) {
                    native_tensor$copy_(ts_tensor)
                    loaded <- loaded + 1
                } else if (verbose) {
                    message("Shape mismatch: ", native_name,
                            " (", paste(as.integer(ts_tensor$shape), collapse = "x"), " vs ",
                            paste(as.integer(native_tensor$shape), collapse = "x"), ")")
                    skipped <- skipped + 1
                }
            } else {
                skipped <- skipped + 1
                unmapped <- c(unmapped, paste0(ts_name, " -> ", native_name))
            }
        }
    })

    if (verbose) {
        if (length(unmapped) > 0 && length(unmapped) <= 10) {
            message("Unmapped parameters:")
            for (u in unmapped) {
                message("  ", u)
            }
        } else if (length(unmapped) > 10) {
            message("Unmapped: ", length(unmapped), " parameters (showing first 5)")
            for (u in head(unmapped, 5)) {
                message("  ", u)
            }
        }
        message("Loaded ", loaded, "/", loaded + skipped, " parameters")
    }

    invisible(native_encoder)
}

#' Native CLIP Text Encoder 2 (OpenCLIP ViT-bigG for SDXL)
#'
#' Native R torch implementation of OpenCLIP text encoder used in SDXL.
#' Returns both hidden states and pooled output.
#'
#' @param vocab_size Vocabulary size (default 49408)
#' @param context_length Maximum sequence length (default 77)
#' @param embed_dim Embedding dimension (default 1280)
#' @param num_layers Number of transformer layers (default 32)
#' @param num_heads Number of attention heads (default 20)
#' @param mlp_dim MLP hidden dimension (default 5120)
#' @param return_penultimate Return the penultimate hidden state
#'   alongside the pooled output (SDXL's \code{hidden_states[-2]}
#'   prompt embeds; the pooled output still comes from the full stack)
#'
#' @return An nn_module representing the text encoder
#' @export
text_encoder2_native <- torch::nn_module(
    "TextEncoder2Native",

    initialize = function(
                          vocab_size = 49408,
                          context_length = 77,
                          embed_dim = 1280,
                          num_layers = 32,
                          num_heads = 20,
                          mlp_dim = 5120,
                          return_penultimate = FALSE
    ) {
    self$context_length <- context_length
    self$embed_dim <- embed_dim
    self$num_layers <- num_layers
    self$return_penultimate <- return_penultimate

    # Embeddings
    self$token_embedding <- torch::nn_embedding(vocab_size, embed_dim)
    self$position_embedding <- torch::nn_parameter(
        torch::torch_zeros(context_length, embed_dim)
    )

    # Transformer blocks with standard GELU (OpenCLIP style)
    self$transformer_blocks <- torch::nn_module_list()
    for (i in seq_len(num_layers)) {
        self$transformer_blocks$append(
                                       CLIPTransformerBlock(embed_dim, num_heads, mlp_dim,
                gelu_type = "exact")
        )
    }

    # Final layer norm
    self$final_layer_norm <- torch::nn_layer_norm(embed_dim)

    # Text projection for pooled output
    self$text_projection <- torch::nn_linear(embed_dim, embed_dim, bias = FALSE)
},

    forward = function(input_ids) {
    # Move input to model's device
    input_ids <- input_ids$to(device = self$token_embedding$weight$device)

    batch_size <- input_ids$shape[1]
    seq_length <- input_ids$shape[2]

    # Token + position embeddings
    # Add 1 because R torch is 1-indexed but tokens are 0-indexed (Python convention)
    token_embeds <- self$token_embedding(input_ids + 1L)
    pos_embeds <- self$position_embedding[1:seq_length,]$unsqueeze(1)$expand(c(batch_size, -1, -1))
    hidden_states <- token_embeds + pos_embeds

    # Transformer layers. Capture the penultimate layer output
    # (hidden_states[-2]) for the SDXL cross-attention embeds, while the
    # pooled output still comes from the full stack.
    n <- self$num_layers
    penultimate <- NULL
    for (i in seq_len(n)) {
        if (i == n) {
            penultimate <- hidden_states
        }
        hidden_states <- self$transformer_blocks[[i]](hidden_states)
    }

    # hidden output: penultimate for SDXL, else the last layer (matches
    # the TorchScript export, which omits final LN on the hidden output)
    hidden_out <- if (self$return_penultimate) penultimate else hidden_states

    # Pooled output: EOS token of the final-LN'd LAST hidden state, then
    # the text projection. EOS is the highest token id (49407), so argmax
    # locates it; causal attention makes trailing padding irrelevant.
    hidden_states_normalized <- self$final_layer_norm(hidden_states)
    eos_indices <- torch::torch_argmax(input_ids, dim = 2L, keepdim = TRUE)
    pooled_output <- hidden_states_normalized$gather(
        dim = 2L,
        index = eos_indices$unsqueeze(-1L)$expand(c(-1L, -1L, self$embed_dim))
    )$squeeze(2L)
    pooled_output <- self$text_projection(pooled_output)

    list(hidden_out, pooled_output)
}
)

#' Load HF safetensors weights into the native SDXL text encoder 2
#'
#' Loads a HuggingFace CLIPTextModelWithProjection \code{model.safetensors}
#' (SDXL's OpenCLIP ViT-bigG \code{text_encoder_2}) into
#' \code{\link{text_encoder2_native}}. Same key layout as encoder 1, plus
#' the top-level \code{text_projection.weight} that produces the pooled
#' \code{text_embeds}.
#'
#' @param native_encoder Native text encoder 2 module
#' @param path Path to model.safetensors (or a directory containing it)
#' @param verbose Print loading progress
#'
#' @return The native encoder with loaded weights (invisibly)
#' @export
load_text_encoder2_safetensors <- function(native_encoder, path,
    verbose = TRUE) {
    remap <- function(key) {
        if (grepl("^text_projection", key)) {
            return("text_projection.weight")
        }
        .clip_st_remap_key(key)
    }
    .load_clip_st(native_encoder, path, remap, "CLIP-bigG", verbose = verbose)
}

#' Build a native SDXL text encoder 2 from a diffusers safetensors directory
#'
#' Reads the CLIPTextConfig from \code{<dir>/config.json}, constructs
#' \code{\link{text_encoder2_native}} to match, and loads
#' \code{model.safetensors} - the safetensors counterpart to the
#' TorchScript \code{text_encoder_2} path (no TorchScript, Blackwell-safe).
#'
#' @param path diffusers text_encoder_2 directory (config.json +
#'   model.safetensors) or the config.json path.
#' @param return_penultimate Return the penultimate hidden state for the
#'   SDXL cross-attention embeds (default TRUE); the pooled output is
#'   always computed from the full stack.
#' @param verbose Print how many parameters were loaded.
#' @param ... Overrides for \code{\link{text_encoder2_native}} args.
#'
#' @return The native text encoder 2 in eval mode.
#' @export
text_encoder2_native_from_safetensors <- function(path,
    return_penultimate = TRUE,
    verbose = TRUE, ...) {
    arch <- .detect_text_encoder_config(path)
    args <- list(vocab_size = arch$vocab_size,
                 context_length = arch$context_length,
                 embed_dim = arch$embed_dim, num_layers = arch$num_layers,
                 num_heads = arch$num_heads, mlp_dim = arch$mlp_dim,
                 return_penultimate = return_penultimate)
    args <- utils::modifyList(args, list(...))
    model <- do.call(text_encoder2_native, args)
    load_text_encoder2_safetensors(model, path, verbose = verbose)
    model$eval()
    model
}

#' Load weights from TorchScript text encoder 2 into native encoder
#'
#' @param native_encoder Native text encoder 2 module
#' @param torchscript_path Path to TorchScript encoder .pt file
#' @param verbose Print loading progress
#'
#' @return The native encoder with loaded weights (invisibly)
#' @export
load_text_encoder2_weights <- function(native_encoder, torchscript_path,
                                       verbose = TRUE) {
    ts_encoder <- torch::jit_load(torchscript_path)
    ts_params <- ts_encoder$parameters

    # Build mapping from TorchScript names to native names
    # TorchScript format: enc.text_model.encoder.layers.0.self_attn.q_proj.weight
    # Native format: transformer_blocks.1.attention.q_proj.weight
    remap_key <- function(key) {
        # Strip enc. prefix
        key <- sub("^enc\\.", "", key)

        # text_projection is at enc level, not text_model level
        if (grepl("^text_projection", key)) {
            return(key)
        }

        # Strip text_model. prefix
        key <- sub("^text_model\\.", "", key)

        # Embeddings
        key <- sub("^embeddings\\.token_embedding\\.", "token_embedding.", key)
        # position_embedding.weight -> position_embedding (it's an nn_parameter, not nn_embedding)
        key <- sub("^embeddings\\.position_embedding\\.weight$",
                   "position_embedding", key)

        # Encoder layers -> transformer_blocks (keep 0-indexed)
        key <- gsub("^encoder\\.layers\\.", "transformer_blocks.", key)

        # self_attn -> attention
        key <- gsub("\\.self_attn\\.", ".attention.", key)

        # Layer norms
        key <- gsub("\\.layer_norm1\\.", ".layernorm_1.", key)
        key <- gsub("\\.layer_norm2\\.", ".layernorm_2.", key)

        # Final layer norm
        key <- sub("^final_layer_norm\\.", "final_layer_norm.", key)

        # MLP (HuggingFace uses fc1/fc2 already)
        # No change needed

        key
    }

    loaded <- 0
    skipped <- 0
    unmapped <- character(0)

    torch::with_no_grad({
        for (ts_name in names(ts_params)) {
            native_name <- remap_key(ts_name)

            if (native_name %in% names(native_encoder$parameters)) {
                ts_tensor <- ts_params[[ts_name]]
                native_tensor <- native_encoder$parameters[[native_name]]

                if (all(ts_tensor$shape == native_tensor$shape)) {
                    native_tensor$copy_(ts_tensor)
                    loaded <- loaded + 1
                } else if (verbose) {
                    message("Shape mismatch: ", native_name,
                            " (", paste(as.integer(ts_tensor$shape), collapse = "x"), " vs ",
                            paste(as.integer(native_tensor$shape), collapse = "x"), ")")
                    skipped <- skipped + 1
                }
            } else {
                skipped <- skipped + 1
                unmapped <- c(unmapped, paste0(ts_name, " -> ", native_name))
            }
        }
    })

    if (verbose) {
        if (length(unmapped) > 0 && length(unmapped) <= 10) {
            message("Unmapped parameters:")
            for (u in unmapped) {
                message("  ", u)
            }
        } else if (length(unmapped) > 10) {
            message("Unmapped: ", length(unmapped), " parameters (showing first 5)")
            for (u in head(unmapped, 5)) {
                message("  ", u)
            }
        }
        message("Loaded ", loaded, "/", loaded + skipped, " parameters")
    }

    invisible(native_encoder)
}

# Parity tests for the FLUX CLIP extensions (quick_gelu, safetensors
# loader, pooled output) and the 16-channel VAE decoder (fixtures from
# tools/gen_fixtures_flux_clip_vae.py, checked in).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture <- function(name) {
  p <- system.file("tinytest", "fixtures", name, package = "diffuseR")
  if (p == "") p <- file.path("fixtures", name)
  p
}
io_path <- fixture("clip_vae_io.safetensors")
clip_path <- fixture("clip_tiny.safetensors")
vae_path <- fixture("vae16_tiny.safetensors")
if (!file.exists(io_path)) exit_file("clip/vae fixtures missing")

fx <- safetensors::safe_load_file(io_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# --- CLIP: quick_gelu forward + argmax pooled output -------------------------------

enc <- text_encoder_native(
  vocab_size = 1000L, context_length = 77L, embed_dim = 16L,
  num_layers = 2L, num_heads = 2L, mlp_dim = 32L,
  apply_final_ln = TRUE, gelu_type = "quick"
)
enc$eval()
load_text_encoder_safetensors(enc, clip_path, verbose = FALSE)

ids <- fx$clip_input_ids$to(dtype = torch::torch_long())
hidden <- torch::with_no_grad(enc(ids))
expect_true(max_abs_diff(hidden, fx$clip_last_hidden) < 1e-5)

pooled <- clip_pooled_output(hidden, ids)
expect_equal(as.integer(pooled$shape), c(2L, 16L))
expect_true(max_abs_diff(pooled, fx$clip_pooled) < 1e-5)

# --- 16-channel VAE decoder ----------------------------------------------------------

dec <- vae_decoder_native(
  latent_channels = 16L,
  block_channels = c(32L, 32L, 16L, 8L),
  norm_groups = 8L
)
dec$eval()
load_decoder_safetensors(dec, vae_path, verbose = FALSE)

img <- torch::with_no_grad(dec(fx$vae_latent))
expect_equal(as.integer(img$shape), as.integer(fx$vae_image$shape))
expect_true(max_abs_diff(img, fx$vae_image) < 1e-5)

# SD-from-safetensors component constructors: config-based CLIP arch
# detection (portable), and the VAE-decoder + CLIP text-encoder
# from-safetensors builders (validated against cached FLUX analogs -
# FLUX's VAE exercises the shared decoder, and FLUX's text_encoder IS
# the SDXL CLIP ViT-L).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}
library(diffuseR)

# --- config-based CLIP arch detection (portable) ----------------------------------

tmp <- tempfile()
dir.create(tmp)
writeLines(jsonlite::toJSON(list(vocab_size = 49408L,
  max_position_embeddings = 77L, hidden_size = 1280L,
  num_hidden_layers = 32L, num_attention_heads = 20L,
  intermediate_size = 5120L), auto_unbox = TRUE),
  file.path(tmp, "config.json"))
arch <- diffuseR:::.detect_text_encoder_config(tmp)
expect_equal(arch$vocab_size, 49408L)
expect_equal(arch$context_length, 77L)
expect_equal(arch$embed_dim, 1280L)      # bigG dims
expect_equal(arch$num_layers, 32L)
expect_equal(arch$num_heads, 20L)
expect_equal(arch$mlp_dim, 5120L)
# accepts a config.json path directly too
expect_equal(diffuseR:::.detect_text_encoder_config(
  file.path(tmp, "config.json"))$embed_dim, 1280L)
expect_error(diffuseR:::.detect_text_encoder_config(tempfile()),
  pattern = "config.json")
unlink(tmp, recursive = TRUE)

# --- against cached FLUX analogs (skipped where the cache is absent) ---------------

flux_vae <- Sys.glob(file.path("~/.cache/huggingface/hub",
  "models--black-forest-labs--FLUX.1-schnell/snapshots/*/vae"))
flux_te <- Sys.glob(file.path("~/.cache/huggingface/hub",
  "models--black-forest-labs--FLUX.1-schnell/snapshots/*/text_encoder"))

if (at_home() && length(flux_vae) && dir.exists(flux_vae[1])) {
  d <- vae_decoder_native_from_safetensors(flux_vae[1], latent_channels = 16L,
    verbose = FALSE)
  ci <- d$named_parameters()[["conv_in.weight"]]
  expect_equal(as.integer(ci$shape[2]), 16L)                 # 16-channel latent
  expect_true(as.numeric(ci$abs()$sum()$item()) > 0)         # actually loaded
  expect_false(d$training)
  rm(d)
  gc()
}

if (at_home() && length(flux_te) && dir.exists(flux_te[1])) {
  a <- diffuseR:::.detect_text_encoder_config(flux_te[1])
  expect_equal(a$embed_dim, 768L)                            # CLIP ViT-L
  expect_equal(a$num_layers, 12L)
  enc <- text_encoder_native_from_safetensors(flux_te[1], verbose = FALSE)
  expect_false(enc$training)
  tok <- torch::torch_tensor(
    matrix(c(49406L, 320L, 49407L, rep(49407L, 74)), nrow = 1),
    dtype = torch::torch_long())
  out <- torch::with_no_grad(enc(tok))
  expect_equal(as.integer(out$shape), c(1L, 77L, 768L))
  rm(enc)
  gc()
}

#' UNet Modules for Diffusion Models
#'
#' Native R torch modules for UNet architecture.
#' @keywords internal
#' @name unet_modules
NULL

#' Group Normalization (32 groups)
#' @keywords internal
group_norm_32 <- function(channels) {
    torch::nn_group_norm(32, channels, eps = 1e-5)
}

#' Sinusoidal Timestep Embedding
#'
#' @param timesteps Tensor of timesteps (batch_size,)
#' @param dim Embedding dimension
#' @param flip_sin_to_cos If TRUE, output [cos, sin] instead of [sin, cos].
#'   SDXL uses TRUE (default), SD21 uses FALSE.
#' @param downscale_freq_shift Frequency shift parameter. SDXL uses 0 (default),
#'   SD21 uses 1. With 0: exponent = log(10000) / half_dim. With 1: exponent = log(10000) / (half_dim - 1).
#' @return Tensor (batch_size, dim)
#' @keywords internal
timestep_embedding <- function(timesteps, dim, flip_sin_to_cos = TRUE,
                               downscale_freq_shift = 0L) {
    half_dim <- dim %/% 2L

    # SDXL uses downscale_freq_shift=0, SD21 uses downscale_freq_shift=1
    emb_scale <- log(10000) / (half_dim - downscale_freq_shift)

    # Create frequency bands: [0, 1, ..., half_dim-1]
    freqs <- torch::torch_exp(torch::torch_arange(0, half_dim - 1L) * -emb_scale)
    freqs <- freqs$to(device = timesteps$device, dtype = torch::torch_float32())

    # Expand and compute embeddings
    # timesteps: [batch] -> [batch, 1]
    # freqs: [half_dim] -> [1, half_dim]
    timesteps_float <- timesteps$to(dtype = torch::torch_float32())$unsqueeze(2L)
    args <- timesteps_float * freqs$unsqueeze(1L)

    # SDXL uses flip_sin_to_cos=True -> [cos, sin]
    # SD21 uses flip_sin_to_cos=False -> [sin, cos]
    if (flip_sin_to_cos) {
        embedding <- torch::torch_cat(list(torch::torch_cos(args),
                torch::torch_sin(args)),
                                      dim = 2L)
    } else {
        embedding <- torch::torch_cat(list(torch::torch_sin(args), torch::torch_cos(args)), dim = 2L)
    }

    # Pad if odd dimension
    if (dim %% 2L == 1L) {
        embedding <- torch::nnf_pad(embedding, c(0L, 1L, 0L, 0L))
    }

    embedding
}

#' ResNet Block for UNet
#'
#' @keywords internal
UNetResBlock <- torch::nn_module(
                                 "UNetResBlock",

                                 initialize = function(
        in_channels,
        out_channels,
        time_embed_dim
    ) {
    self$in_channels <- in_channels
    self$out_channels <- out_channels

    # First conv block
    self$norm1 <- group_norm_32(in_channels)
    self$conv1 <- torch::nn_conv2d(in_channels, out_channels, 3L, padding = 1L)

    # Time embedding projection
    self$time_emb_proj <- torch::nn_linear(time_embed_dim, out_channels)

    # Second conv block
    self$norm2 <- group_norm_32(out_channels)
    self$conv2 <- torch::nn_conv2d(out_channels, out_channels, 3L, padding = 1L)

    # Skip connection (if channels differ)
    if (in_channels != out_channels) {
        self$conv_shortcut <- torch::nn_conv2d(in_channels, out_channels, 1L)
    } else {
        self$conv_shortcut <- NULL
    }
},

                                 forward = function(
        x,
        temb
    ) {
    h <- x

    # First block
    h <- self$norm1(h)
    h <- torch::nnf_silu(h)
    h <- self$conv1(h)

    # Add time embedding
    temb_out <- torch::nnf_silu(temb)
    temb_out <- self$time_emb_proj(temb_out)
    # Expand to spatial dimensions: (B, C) -> (B, C, 1, 1)
    temb_out <- temb_out$unsqueeze(-1L)$unsqueeze(-1L)
    h <- h + temb_out

    # Second block
    h <- self$norm2(h)
    h <- torch::nnf_silu(h)
    h <- self$conv2(h)

    # Skip connection
    if (!is.null(self$conv_shortcut)) {
        x <- self$conv_shortcut(x)
    }

    h + x
}
)

#' Downsample Block
#' @keywords internal
Downsample2D <- torch::nn_module(
                                 "Downsample2D",

                                 initialize = function(channels) {
    self$conv <- torch::nn_conv2d(channels, channels, 3L, stride = 2L,
                                  padding = 1L)
},

                                 forward = function(x) {
    self$conv(x)
}
)

#' Upsample Block
#' @keywords internal
Upsample2D <- torch::nn_module(
                               "Upsample2D",

                               initialize = function(channels) {
    self$conv <- torch::nn_conv2d(channels, channels, 3L, padding = 1L)
},

                               forward = function(x) {
    # Upsample 2x then conv
    x <- torch::nnf_interpolate(x, scale_factor = 2, mode = "nearest")
    self$conv(x)
}
)

#' Cross-Attention for UNet
#' @keywords internal
UNetCrossAttention <- torch::nn_module(
                                       "UNetCrossAttention",

                                       initialize = function(
        query_dim,
        context_dim = NULL,
        heads = 8L,
        dim_head = 64L
    ) {
    self$heads <- heads
    self$dim_head <- dim_head
    inner_dim <- heads * dim_head
    context_dim <- context_dim %||% query_dim

    self$scale <- dim_head ^ (-0.5)

    # Query, Key, Value projections
    self$to_q <- torch::nn_linear(query_dim, inner_dim, bias = FALSE)
    self$to_k <- torch::nn_linear(context_dim, inner_dim, bias = FALSE)
    self$to_v <- torch::nn_linear(context_dim, inner_dim, bias = FALSE)

    # Output projection
    self$to_out <- torch::nn_sequential(torch::nn_linear(inner_dim, query_dim),
                                        torch::nn_dropout(0))
},

                                       forward = function(
        x,
        context = NULL
    ) {
    if (is.null(context)) context <- x

    b <- x$shape[1]
    seq_len <- x$shape[2]
    context_len <- context$shape[2]
    h <- self$heads

    # Project
    q <- self$to_q(x)
    k <- self$to_k(context)
    v <- self$to_v(context)

    # Reshape for multi-head attention: (B, S, H*D) -> (B, H, S, D)
    q <- q$view(c(b, seq_len, h, self$dim_head))$transpose(2L, 3L)
    k <- k$view(c(b, context_len, h, self$dim_head))$transpose(2L, 3L)
    v <- v$view(c(b, context_len, h, self$dim_head))$transpose(2L, 3L)

    # Scaled dot-product attention
    attn <- torch::torch_matmul(q, k$transpose(3L, 4L)) * self$scale
    attn <- torch::nnf_softmax(attn, dim = -1L)

    # Apply attention to values
    out <- torch::torch_matmul(attn, v)

    # Reshape back: (B, H, S, D) -> (B, S, H*D)
    out <- out$transpose(2L, 3L)$contiguous()$view(c(b, seq_len, -1L))

    self$to_out(out)
}
)

#' GEGLU Feedforward
#' @keywords internal
GEGLU <- torch::nn_module(
                          "GEGLU",

                          initialize = function(
        dim_in,
        dim_out
    ) {
    self$proj <- torch::nn_linear(dim_in, dim_out * 2L)
},

                          forward = function(x) {
    x_proj <- self$proj(x)
    chunks <- torch::torch_chunk(x_proj, 2L, dim = -1L)
    chunks[[1]] * torch::nnf_gelu(chunks[[2]])
}
)

#' FeedForward Network
#' @keywords internal
FeedForward <- torch::nn_module(
                                "FeedForward",

                                initialize = function(
        dim,
        mult = 4L
    ) {
    inner_dim <- dim * mult
    self$net <- torch::nn_sequential(GEGLU(dim, inner_dim),
                                     torch::nn_dropout(0),
                                     torch::nn_linear(inner_dim, dim))
},

                                forward = function(x) {
    self$net(x)
}
)

#' Basic Transformer Block
#' @keywords internal
BasicTransformerBlock <- torch::nn_module(
    "BasicTransformerBlock",

    initialize = function(
                          dim,
                          n_heads,
                          d_head,
                          context_dim = NULL
    ) {
    # Self-attention
    self$attn1 <- UNetCrossAttention(dim, heads = n_heads, dim_head = d_head)

    # Cross-attention
    self$attn2 <- UNetCrossAttention(dim, context_dim = context_dim,
                                     heads = n_heads, dim_head = d_head)

    # Feedforward
    self$ff <- FeedForward(dim)

    # Layer norms
    self$norm1 <- torch::nn_layer_norm(dim)
    self$norm2 <- torch::nn_layer_norm(dim)
    self$norm3 <- torch::nn_layer_norm(dim)
},

    forward = function(
                       x,
                       context = NULL
    ) {
    # Self-attention
    x <- x + self$attn1(self$norm1(x))

    # Cross-attention
    x <- x + self$attn2(self$norm2(x), context)

    # Feedforward
    x <- x + self$ff(self$norm3(x))

    x
}
)

#' Spatial Transformer (Attention Block)
#' @keywords internal
SpatialTransformer <- torch::nn_module(
                                       "SpatialTransformer",

                                       initialize = function(
        in_channels,
        n_heads,
        d_head,
        depth = 1L,
        context_dim = NULL
    ) {
    inner_dim <- n_heads * d_head
    self$in_channels <- in_channels

    # Normalization
    self$norm <- group_norm_32(in_channels)

    # Projection in/out
    self$proj_in <- torch::nn_linear(in_channels, inner_dim)
    self$proj_out <- torch::nn_linear(inner_dim, in_channels)

    # Transformer blocks
    self$transformer_blocks <- torch::nn_module_list()
    for (i in seq_len(depth)) {
        self$transformer_blocks$append(
                                       BasicTransformerBlock(inner_dim, n_heads, d_head, context_dim)
        )
    }
},

                                       forward = function(
        x,
        context = NULL
    ) {
    b <- x$shape[1]
    c <- x$shape[2]
    h <- x$shape[3]
    w <- x$shape[4]

    x_in <- x

    # Normalize
    x <- self$norm(x)

    # Reshape to sequence: (B, C, H, W) -> (B, H*W, C)
    x <- x$permute(c(1L, 3L, 4L, 2L))$reshape(c(b, h * w, c))

    # Project in
    x <- self$proj_in(x)

    # Transformer blocks
    for (i in seq_along(self$transformer_blocks)) {
        x <- self$transformer_blocks[[i]](x, context)
    }

    # Project out
    x <- self$proj_out(x)

    # Reshape back: (B, H*W, C) -> (B, C, H, W)
    x <- x$reshape(c(b, h, w, c))$permute(c(1L, 4L, 2L, 3L))

    # Residual
    x + x_in
}
)

#' Z-Image Transformer Block Modules
#'
#' Fresh R port of the Z-Image DiT building blocks from the diffusers
#' reference (Apache-2.0,
#' src/diffusers/models/transformers/transformer_z_image.py). Z-Image is
#' a single-stream DiT: text and image tokens share one sequence and one
#' set of block weights. Each block uses sandwich RMSNorms (a learned
#' norm before AND after both the attention and the feed-forward) and a
#' scale/gate-only modulation — four chunks (scale_msa, gate_msa,
#' scale_mlp, gate_mlp), no shift, gates tanh-squashed, scales 1 + x.
#' The attention is plain joint self-attention, so the FLUX attention
#' module is reused with bias = FALSE and eps = 1e-5.
#'
#' @name dit_zimage_modules
NULL

# Modulation embedding width (ADALN_EMBED_DIM in the reference); the
# effective width is min(dim, 256).
.zimage_adaln_dim <- 256L

#' Z-Image feed-forward (SwiGLU with separate gate weights)
#'
#' w2(silu(w1(x)) * w3(x)) with all three linears bias-free. The hidden
#' width is int(dim / 3 * 8).
#'
#' @param dim Integer. Model width.
#' @param hidden_dim Integer. Hidden width.
#'
#' @return Module whose forward(x) returns \code{w2(silu(w1(x)) * w3(x))},
#'   a tensor of the same shape as \code{x}.
#'
#' @export
zimage_feed_forward <- torch::nn_module(
                                        "zimage_feed_forward",
                                        initialize = function(dim, hidden_dim) {
    self$w1 <- torch::nn_linear(dim, hidden_dim, bias = FALSE)
    self$w2 <- torch::nn_linear(hidden_dim, dim, bias = FALSE)
    self$w3 <- torch::nn_linear(dim, hidden_dim, bias = FALSE)
},
                                        forward = function(x) {
    self$w2(torch::nnf_silu(self$w1(x)) * self$w3(x))
}
)

#' Z-Image transformer block
#'
#' Sandwich-norm residual block shared by the noise refiner, the context
#' refiner and the main trunk. With \code{modulation = TRUE} the block
#' carries an adaLN linear producing (scale_msa, gate_msa, scale_mlp,
#' gate_mlp); the context refiner uses \code{modulation = FALSE} and has
#' no adaLN weights at all.
#'
#' @param dim Integer. Model width.
#' @param n_heads Integer. Attention heads; head dim is dim / n_heads.
#' @param norm_eps Numeric. RMSNorm epsilon. Default 1e-5.
#' @param modulation Logical. Whether the block is timestep-modulated.
#'
#' @return Module whose forward(x, freqs, adaln_input, chunk_size)
#'   returns the residual block output, a tensor of the same shape as
#'   \code{x}. \code{adaln_input} is used only when the block was built
#'   with \code{modulation = TRUE}.
#'
#' @export
zimage_block <- torch::nn_module(
                                 "zimage_block",
                                 initialize = function(dim, n_heads, norm_eps = 1e-5,
        modulation = TRUE) {
    self$attention <- flux_attention(query_dim = dim, heads = n_heads,
                                     dim_head = dim %/% n_heads, eps = 1e-5,
                                     bias = FALSE)
    self$feed_forward <- zimage_feed_forward(dim, as.integer(dim / 3 * 8))

    self$attention_norm1 <- ltx23_rms_norm(dim, eps = norm_eps)
    self$ffn_norm1 <- ltx23_rms_norm(dim, eps = norm_eps)
    self$attention_norm2 <- ltx23_rms_norm(dim, eps = norm_eps)
    self$ffn_norm2 <- ltx23_rms_norm(dim, eps = norm_eps)

    self$modulation <- modulation
    if (modulation) {
        self$adaLN_modulation <- torch::nn_sequential(
            torch::nn_linear(min(dim, .zimage_adaln_dim), 4L * dim, bias = TRUE)
        )
    }
},
                                 forward = function(x, freqs, adaln_input = NULL, chunk_size = NULL) {
    if (self$modulation) {
        mod <- self$adaLN_modulation(adaln_input)$unsqueeze(2L)
        chunks <- mod$chunk(4L, dim = 3L)
        scale_msa <- chunks[[1]]$add(1)
        gate_msa <- chunks[[2]]$tanh()
        scale_mlp <- chunks[[3]]$add(1)
        gate_mlp <- chunks[[4]]$tanh()

        attn_out <- self$attention(self$attention_norm1(x) * scale_msa,
                                   image_rotary_emb = freqs,
                                   chunk_size = chunk_size)
        x <- x + gate_msa * self$attention_norm2(attn_out)
        x + gate_mlp * self$ffn_norm2(
                                      self$feed_forward(self$ffn_norm1(x) * scale_mlp)
        )
    } else {
        attn_out <- self$attention(self$attention_norm1(x),
                                   image_rotary_emb = freqs,
                                   chunk_size = chunk_size)
        x <- x + self$attention_norm2(attn_out)
        x + self$ffn_norm2(self$feed_forward(self$ffn_norm1(x)))
    }
}
)

#' Z-Image final layer
#'
#' Parameterless LayerNorm scaled by 1 + adaLN(c) (scale only, no
#' shift), then the token-to-patch projection.
#'
#' @param hidden_size Integer. Model width.
#' @param out_channels Integer. Patch output dim
#'   (patch^2 * f_patch * latent channels).
#'
#' @return Module whose forward(x, c) returns the token-to-patch
#'   projection [B, S, out_channels], ready for unpatchifying into a
#'   latent.
#'
#' @export
zimage_final_layer <- torch::nn_module(
                                       "zimage_final_layer",
                                       initialize = function(hidden_size, out_channels) {
    self$norm_final <- torch::nn_layer_norm(hidden_size, eps = 1e-6,
        elementwise_affine = FALSE)
    self$linear <- torch::nn_linear(hidden_size, out_channels, bias = TRUE)
    self$adaLN_modulation <- torch::nn_sequential(
        torch::nn_silu(),
        torch::nn_linear(min(hidden_size, .zimage_adaln_dim), hidden_size,
                         bias = TRUE)
    )
},
                                       forward = function(x, c) {
    scale <- self$adaLN_modulation(c)$add(1)$unsqueeze(2L)
    self$linear(self$norm_final(x) * scale)
}
)

#' Z-Image timestep embedder
#'
#' 256-dim cos-first sinusoid (computed in float32) through a
#' Linear-SiLU-Linear MLP. The model feeds t * t_scale with the
#' pipeline's t already in [0, 1].
#'
#' @param out_size Integer. Output width, min(dim, 256).
#' @param mid_size Integer. Hidden width. The full model uses 1024.
#' @param freq_size Integer. Sinusoid width. Default 256.
#'
#' @return Module whose forward(t) returns the timestep embedding
#'   [B, out_size].
#'
#' @export
zimage_t_embedder <- torch::nn_module(
                                      "zimage_t_embedder",
                                      initialize = function(out_size, mid_size = 1024L, freq_size = 256L) {
    self$freq_size <- freq_size
    self$mlp <- torch::nn_sequential(
                                     torch::nn_linear(freq_size, mid_size, bias = TRUE),
                                     torch::nn_silu(),
                                     torch::nn_linear(mid_size, out_size, bias = TRUE)
    )
},
                                      forward = function(t) {
    t_freq <- ltx23_get_timestep_embedding(t, self$freq_size,
        flip_sin_to_cos = TRUE, downscale_freq_shift = 0)
    self$mlp(t_freq$to(dtype = self$mlp[[1]]$weight$dtype))
}
)

#!/usr/bin/env r
#
# Full LTX-2 Pipeline Test
#
# Tests: Text encoding -> DiT denoising -> (VAE decode)
#

library(torch)
library(diffuseR)

cat("=== LTX-2 Full Pipeline Test ===\n\n")

# Helper to get GPU memory via nvidia-smi
get_gpu_memory <- function() {
  out <- system("nvidia-smi --query-gpu=memory.used --format=csv,noheader,nounits", intern = TRUE)
  as.numeric(out[1]) / 1024  # Convert MB to GB
}

# Clear GPU
gc(full = TRUE)
torch::cuda_empty_cache()

mem_start <- get_gpu_memory()
cat(sprintf("Initial GPU memory: %.2f GB\n\n", mem_start))

# Test parameters
test_prompt <- "A cat sitting on a table, looking at the camera"
width <- 512L  # Small resolution for testing
height <- 320L
num_frames <- 17L  # Small number of frames
num_steps <- 4L    # Few steps for quick test

cat(sprintf("Test settings:\n"))
cat(sprintf("  Prompt: %s\n", test_prompt))
cat(sprintf("  Resolution: %dx%d, %d frames\n", width, height, num_frames))
cat(sprintf("  Steps: %d\n\n", num_steps))

# Run pipeline (without VAE for now)
cat("Running pipeline...\n")
t0 <- Sys.time()

tryCatch({
  result <- txt2vid_ltx2(
    prompt = test_prompt,
    width = width,
    height = height,
    num_frames = num_frames,
    num_inference_steps = num_steps,
    guidance_scale = 4.0,
    text_backend = "random",  # Skip real text encoder for speed
    verbose = TRUE
  )

  elapsed <- as.numeric(Sys.time() - t0)
  mem_peak <- get_gpu_memory()

  cat(sprintf("\n=== Results ===\n"))
  cat(sprintf("Elapsed time: %.1f seconds\n", elapsed))
  cat(sprintf("Peak GPU memory: %.2f GB\n", mem_peak))
  cat(sprintf("Output shape: %s\n", paste(dim(result$video), collapse = "x")))
  cat(sprintf("Latent range: [%.3f, %.3f]\n",
              min(result$video), max(result$video)))
  cat("Pipeline: PASSED\n")

}, error = function(e) {
  cat(sprintf("\nPipeline: FAILED\n"))
  cat(sprintf("Error: %s\n", e$message))
  print(traceback())
})

# Cleanup
gc(full = TRUE)
torch::cuda_empty_cache()

mem_end <- get_gpu_memory()
cat(sprintf("\nFinal GPU memory: %.2f GB\n", mem_end))
cat("\n=== Test Complete ===\n")

#' VAE ResNet Block
#'
#' @param in_channels Input channels
#' @param out_channels Output channels
#' @keywords internal
VAEResnetBlock <- torch::nn_module(
                                   "VAEResnetBlock",

                                   initialize = function(
        in_channels,
        out_channels,
        norm_groups = 32
    ) {
    self$norm1 <- torch::nn_group_norm(norm_groups, in_channels, eps = 1e-6)
    self$conv1 <- torch::nn_conv2d(in_channels, out_channels,
                                   kernel_size = 3, padding = 1)
    self$norm2 <- torch::nn_group_norm(norm_groups, out_channels, eps = 1e-6)
    self$conv2 <- torch::nn_conv2d(out_channels, out_channels, kernel_size = 3, padding = 1)

    # Shortcut if dimensions change
    if (in_channels != out_channels) {
        self$conv_shortcut <- torch::nn_conv2d(in_channels, out_channels, kernel_size = 1)
    }
},

                                   forward = function(x) {
    h <- x
    h <- self$norm1(h)
    h <- torch::nnf_silu(h)
    h <- self$conv1(h)

    h <- self$norm2(h)
    h <- torch::nnf_silu(h)
    h <- self$conv2(h)

    # Apply shortcut if it exists
    if (!is.null(self$conv_shortcut)) {
        x <- self$conv_shortcut(x)
    }

    h + x
}
)

#' VAE Attention Block
#'
#' Self-attention for VAE mid block
#' @param channels Number of channels
#' @keywords internal
VAEAttentionBlock <- torch::nn_module(
                                      "VAEAttentionBlock",

                                      initialize = function(channels, norm_groups = 32) {
    self$group_norm <- torch::nn_group_norm(norm_groups, channels, eps = 1e-6)
    self$to_q <- torch::nn_linear(channels, channels)
    self$to_k <- torch::nn_linear(channels, channels)
    self$to_v <- torch::nn_linear(channels, channels)
    self$to_out <- torch::nn_module_list(list(
            torch::nn_linear(channels, channels)
        ))
    self$channels <- channels
},

                                      forward = function(x) {
    residual <- x
    batch <- x$shape[1]
    channels <- x$shape[2]
    height <- x$shape[3]
    width <- x$shape[4]

    # Normalize
    x <- self$group_norm(x)

    # Reshape to (batch, h*w, channels)
    x <- x$permute(c(1, 3, 4, 2))$reshape(c(batch, height * width, channels))

    # QKV projections
    q <- self$to_q(x)
    k <- self$to_k(x)
    v <- self$to_v(x)

    # Scaled dot-product attention
    scale <- 1.0 / sqrt(channels)
    attn <- torch::torch_bmm(q, k$transpose(2, 3)) * scale
    attn <- torch::nnf_softmax(attn, dim = -1)
    out <- torch::torch_bmm(attn, v)

    # Project out
    out <- self$to_out[[1]](out)

    # Reshape back to (batch, channels, h, w)
    out <- out$reshape(c(batch, height, width, channels))$permute(c(1, 4, 2, 3))

    out + residual
}
)

#' VAE Up Block
#'
#' @param in_channels Input channels
#' @param out_channels Output channels
#' @param num_resnets Number of resnet blocks (default 3)
#' @param add_upsample Whether to add upsampler
#' @keywords internal
VAEUpBlock <- torch::nn_module(
                               "VAEUpBlock",

                               initialize = function(
        in_channels,
        out_channels,
        num_resnets = 3,
        add_upsample = TRUE,
        norm_groups = 32
    ) {
    self$resnets <- torch::nn_module_list()

    for (i in seq_len(num_resnets)) {
        if (i == 1) {
            res_in <- in_channels
        } else {
            res_in <- out_channels
        }
        self$resnets$append(VAEResnetBlock(res_in, out_channels, norm_groups))
    }

    if (add_upsample) {
        self$upsamplers <- torch::nn_module_list(list(
                torch::nn_module(
                                 "Upsampler",
                                 initialize = function(channels) {
            self$conv <- torch::nn_conv2d(channels, channels,
                kernel_size = 3, padding = 1)
        },
                                 forward = function(x) {
            x <- torch::nnf_interpolate(x, scale_factor = 2, mode = "nearest")
            self$conv(x)
        }
                )(out_channels)
            ))
    }
},

                               forward = function(x) {
    for (i in seq_along(self$resnets)) {
        x <- self$resnets[[i]](x)
    }
    if (!is.null(self$upsamplers)) {
        x <- self$upsamplers[[1]](x)
    }
    x
}
)

#' VAE Mid Block
#'
#' @param channels Number of channels
#' @keywords internal
VAEMidBlock <- torch::nn_module(
                                "VAEMidBlock",

                                initialize = function(channels, norm_groups = 32) {
    self$resnets <- torch::nn_module_list(list(
            VAEResnetBlock(channels, channels, norm_groups),
            VAEResnetBlock(channels, channels, norm_groups)
        ))
    self$attentions <- torch::nn_module_list(list(VAEAttentionBlock(channels,
                norm_groups)))
},

                                forward = function(x) {
    x <- self$resnets[[1]](x)
    x <- self$attentions[[1]](x)
    x <- self$resnets[[2]](x)
    x
}
)

#' Load HF safetensors VAE weights into the native decoder
#'
#' Loads the decoder half of a diffusers AutoencoderKL safetensors file
#' (e.g. FLUX.1-schnell's \code{vae/diffusion_pytorch_model.safetensors}).
#' Keys under \code{decoder.} map to the native module 1:1; encoder and
#' quant-conv keys are skipped (the FLUX VAE has no quant convs, and
#' txt2img needs no encoder).
#'
#' @param native_decoder Native VAE decoder module
#' @param path Path to the VAE .safetensors file (or a directory
#'   containing diffusion_pytorch_model.safetensors)
#' @param verbose Print loading progress
#'
#' @return The native decoder with loaded weights (invisibly)
#' @export
load_decoder_safetensors <- function(native_decoder, path, verbose = TRUE) {
    path <- path.expand(path)
    if (dir.exists(path)) {
        path <- file.path(path, "diffusion_pytorch_model.safetensors")
    }
    handle <- safetensors::safetensors$new(path, framework = "torch")
    keys <- setdiff(handle$keys(), "__metadata__")
    dec_keys <- keys[startsWith(keys, "decoder.")]

    dests <- native_decoder$named_parameters()
    filled <- character(0)
    unmapped <- character(0)
    torch::with_no_grad({
        for (key in dec_keys) {
            native_name <- sub("^decoder\\.", "", key)
            dest <- dests[[native_name]]
            if (is.null(dest)) {
                unmapped <- c(unmapped, key)
                next
            }
            dest$copy_(handle$get_tensor(key))
            filled <- c(filled, native_name)
        }
    })

    unfilled <- setdiff(names(dests), filled)
    if (length(unmapped)) {
        stop("VAE decoder load: ", length(unmapped), " unmapped keys, e.g. ",
             paste(utils::head(unmapped, 3), collapse = ", "))
    }
    if (length(unfilled)) {
        stop("VAE decoder load: ", length(unfilled),
             " unfilled params, e.g. ",
             paste(utils::head(unfilled, 3), collapse = ", "))
    }
    if (verbose) {
        message("Loaded ", length(filled), " decoder parameters from ", path)
    }
    invisible(native_decoder)
}

#' Build a native VAE decoder from a diffusers safetensors directory
#'
#' The safetensors counterpart to the TorchScript decoder path:
#' constructs \code{\link{vae_decoder_native}} and loads the decoder half
#' of a diffusers AutoencoderKL checkpoint (no TorchScript, so it works
#' on Blackwell). \code{latent_channels} defaults to 4 (SD/SDXL); pass 16
#' for the FLUX/SD3 VAE. The SD/SDXL and FLUX VAEs share the decoder
#' shape and differ only in that channel count.
#'
#' @param path Path to the VAE directory (containing
#'   \code{diffusion_pytorch_model.safetensors}) or the file itself.
#' @param latent_channels Latent channel count (4 for SD/SDXL, 16 for FLUX).
#' @param verbose Print how many parameters were loaded.
#' @param ... Overrides for \code{\link{vae_decoder_native}} constructor args.
#'
#' @return The native VAE decoder in eval mode.
#' @export
vae_decoder_native_from_safetensors <- function(path, latent_channels = 4L,
    verbose = TRUE, ...) {
    model <- vae_decoder_native(latent_channels = latent_channels, ...)
    load_decoder_safetensors(model, path, verbose = verbose)
    model$eval()
    model
}

#' Load weights from TorchScript decoder into native decoder
#'
#' @param native_decoder Native VAE decoder module
#' @param torchscript_path Path to TorchScript decoder .pt file
#' @param verbose Print loading progress
#'
#' @return The native decoder with loaded weights (invisibly)
#' @export
load_decoder_weights <- function(native_decoder, torchscript_path,
                                 verbose = TRUE) {
    ts_decoder <- torch::jit_load(torchscript_path)
    ts_params <- ts_decoder$parameters

    loaded <- 0
    torch::with_no_grad({
        for (ts_name in names(ts_params)) {
            # Strip dec. prefix
            native_name <- sub("^dec\\.", "", ts_name)

            if (native_name %in% names(native_decoder$parameters)) {
                ts_tensor <- ts_params[[ts_name]]
                native_tensor <- native_decoder$parameters[[native_name]]

                if (all(ts_tensor$shape == native_tensor$shape)) {
                    native_tensor$copy_(ts_tensor)
                    loaded <- loaded + 1
                } else if (verbose) {
                    cat("Shape mismatch:", native_name, "\n")
                }
            } else if (verbose) {
                cat("Missing param:", native_name, "\n")
            }
        }
    })

    if (verbose) {
        cat("Loaded", loaded, "/", length(names(ts_params)), "parameters\n")
    }

    invisible(native_decoder)
}

#' Native VAE Decoder
#'
#' Native R torch implementation of the SDXL VAE decoder.
#' Replaces TorchScript decoder for better GPU compatibility.
#'
#' @param latent_channels Number of latent channels (4 for SD/SDXL,
#'   16 for FLUX/SD3)
#' @param out_channels Number of output channels (default 3 for RGB)
#' @param block_channels Decoder block channels (reversed encoder
#'   block_out_channels; default matches SD/SDXL and FLUX)
#' @param norm_groups Group norm groups (default 32; must divide every
#'   entry of \code{block_channels})
#'
#' @return An nn_module representing the VAE decoder
#' @export
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   # A small decoder; the SD/SDXL defaults are far too large to build
#'   # inside an example.
#'   decoder <- vae_decoder_native(latent_channels = 4,
#'                                 block_channels = 32,
#'                                 norm_groups = 32)
#'   latents <- torch::torch_randn(c(1, 4, 8, 8))
#'   image <- torch::with_no_grad(decoder(latents))
#'   image$shape
#' }
#'
#' # Real weights come from a downloaded checkpoint.
#' \dontrun{
#' decoder <- vae_decoder_native()
#' load_decoder_weights(decoder, "path/to/decoder.pt")
#' }
vae_decoder_native <- torch::nn_module(
                                       "VAEDecoderNative",

                                       initialize = function(
        latent_channels = 4,
        out_channels = 3,
        block_channels = c(512, 512, 256, 128),
        norm_groups = 32
    ) {
    # Diffusers AutoencoderKL decoder: block channels reversed from the
    # encoder's block_out_channels; upsamplers on all but the last block.
    # The SD/SDXL and FLUX/SD3 VAEs share this exact shape (FLUX differs
    # only in latent_channels = 16).
    n_blocks <- length(block_channels)

    self$conv_in <- torch::nn_conv2d(latent_channels, block_channels[1],
                                     kernel_size = 3, padding = 1)

    self$mid_block <- VAEMidBlock(block_channels[1], norm_groups)

    self$up_blocks <- torch::nn_module_list()
    for (i in seq_len(n_blocks)) {
        in_ch <- block_channels[max(i - 1, 1)]
        self$up_blocks$append(VAEUpBlock(in_ch, block_channels[i],
                num_resnets = 3,
                add_upsample = i < n_blocks,
                norm_groups = norm_groups))
    }

    last <- block_channels[n_blocks]
    self$conv_norm_out <- torch::nn_group_norm(norm_groups, last, eps = 1e-6)
    self$conv_out <- torch::nn_conv2d(last, out_channels, kernel_size = 3, padding = 1)
},

                                       forward = function(x) {
    # Input conv
    x <- self$conv_in(x)

    # Mid block
    x <- self$mid_block(x)

    # Up blocks
    for (i in seq_along(self$up_blocks)) {
        x <- self$up_blocks[[i]](x)
    }

    # Output
    x <- self$conv_norm_out(x)
    x <- torch::nnf_silu(x)
    x <- self$conv_out(x)

    x
}
)

#' LTX-2.3 Single-File Checkpoint Reader
#'
#' LTX 2.3 checkpoints ship as one safetensors file containing every
#' component (transformer, connectors, video VAE, audio VAE, vocoder),
#' with the model version and full component configuration embedded in
#' the safetensors metadata. These helpers open the file, validate the
#' version, split the key space by component, and stream tensors into
#' R torch modules one at a time so the 46 GB file is never fully
#' materialized in memory.
#'
#' @name checkpoint_ltx23
NULL

#' Open an LTX-2.3 checkpoint
#'
#' Opens a single-file LTX checkpoint lazily (header only), validates the
#' \code{model_version} metadata, and parses the embedded component
#' configuration.
#'
#' @param path Path to the checkpoint .safetensors file.
#' @param require_version Character. Required \code{model_version} prefix
#'   (default "2.3"). Set to NULL to skip the check.
#'
#' @return An object of class \code{ltx23_checkpoint}: a list with
#'   \code{handle} (safetensors reader), \code{keys}, \code{version},
#'   \code{config} (parsed component configs, or NULL), and \code{path}.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' ckpt <- ltx23_open_checkpoint("ltx-2.3-22b-distilled-1.1.safetensors")
#' str(ltx23_split_keys(ckpt$keys), max.level = 1)
#' }
ltx23_open_checkpoint <- function(path, require_version = "2.3") {
    if (!requireNamespace("safetensors", quietly = TRUE)) {
        stop("The safetensors package is required to read LTX checkpoints.")
    }
    path <- path.expand(path)
    if (!file.exists(path)) {
        stop("Checkpoint not found: ", path)
    }

    handle <- safetensors::safetensors$new(path, framework = "torch")

    meta <- handle$metadata[["__metadata__"]]
    if (is.list(meta) || is.character(meta)) {
        version <- meta[["model_version"]]
    } else {
        version <- NULL
    }

    if (!is.null(require_version)) {
        if (is.null(version)) {
            stop("Checkpoint has no model_version metadata; expected an LTX ",
                 require_version, " single-file checkpoint: ", path)
        }
        if (!startsWith(version, require_version)) {
            stop(
                 "Checkpoint model_version is '", version, "' but '",
                 require_version, "*' is required: ", path
            )
        }
    }

    config <- NULL
    if (!is.null(meta[["config"]])) {
        config <- tryCatch(
                           jsonlite::fromJSON(meta[["config"]], simplifyVector = TRUE),
                           error = function(e) NULL
        )
    }

    keys <- setdiff(handle$keys(), "__metadata__")

    structure(
              list(
                   handle = handle,
                   keys = keys,
                   version = version,
                   config = config,
                   path = path
        ),
              class = "ltx23_checkpoint"
    )
}

#' @export
print.ltx23_checkpoint <- function(x, ...) {
    cat("<ltx23_checkpoint>\n")
    cat("  path:    ", x$path, "\n")
    cat("  version: ", x$version %||% "(none)", "\n")
    cat("  tensors: ", length(x$keys), "\n")
    groups <- ltx23_split_keys(x$keys)
    for (g in names(groups)) {
        cat(sprintf("  %-10s %d keys\n", g, length(groups[[g]])))
    }
    invisible(x)
}

#' Split checkpoint keys by component
#'
#' Splits the flat key space of an LTX single-file checkpoint into its
#' component groups. Connector tensors live under the
#' \code{model.diffusion_model.} prefix alongside the transformer, plus a
#' top-level \code{text_embedding_projection.} group; both are routed to
#' the \code{connectors} component.
#'
#' @param keys Character vector of checkpoint tensor names.
#'
#' @return Named list of character vectors: \code{dit},
#'   \code{connectors}, \code{vae}, \code{audio_vae}, \code{vocoder},
#'   and \code{other} (anything unrecognized; should be empty).
#'
#' @export
ltx23_split_keys <- function(keys) {
    dm_prefix <- "model.diffusion_model."
    connector_res <- c(
                       "^model\\.diffusion_model\\.video_embeddings_connector\\.",
                       "^model\\.diffusion_model\\.audio_embeddings_connector\\.",
                       "^text_embedding_projection\\."
    )

    is_connector <- Reduce(`|`, lapply(connector_res, grepl, x = keys))
    is_dit <- startsWith(keys, dm_prefix) & !is_connector
    is_vae <- startsWith(keys, "vae.")
    is_audio_vae <- startsWith(keys, "audio_vae.")
    is_vocoder <- startsWith(keys, "vocoder.")

    claimed <- is_connector | is_dit | is_vae | is_audio_vae | is_vocoder

    list(dit = keys[is_dit], connectors = keys[is_connector],
         vae = keys[is_vae], audio_vae = keys[is_audio_vae],
         vocoder = keys[is_vocoder], other = keys[!claimed])
}

#' Stream a checkpoint key group into a module
#'
#' Reads tensors one at a time from an open checkpoint and copies them
#' into the matching parameters/buffers of \code{module}. Destination
#' names are derived by \code{map_key}; \code{$copy_()} handles any
#' dtype/device conversion, so the module may already live on its target
#' device in its target dtype.
#'
#' @param ckpt An \code{ltx23_checkpoint}.
#' @param keys Character vector of checkpoint keys to load (one group
#'   from \code{\link{ltx23_split_keys}}).
#' @param module A torch nn_module to populate.
#' @param map_key Function mapping a checkpoint key to the module's
#'   parameter/buffer name, or NA to skip the key deliberately.
#' @param verbose Logical. Report progress and coverage.
#' @param gc_every Integer. Run \code{gc()} after this many tensors.
#'
#' @return Invisibly, a list with \code{unmapped} (checkpoint keys that
#'   found no destination), \code{skipped} (keys the mapper declined),
#'   and \code{unfilled} (module parameters/buffers never written).
#'   A perfectly loaded group has zero \code{unmapped} and zero
#'   \code{unfilled}.
#'
#' @export
ltx23_load_group <- function(ckpt, keys, module, map_key = identity,
                             verbose = TRUE, gc_every = 50L) {
    stopifnot(inherits(ckpt, "ltx23_checkpoint"))

    params <- module$named_parameters()
    buffers <- module$named_buffers()
    dests <- c(params, buffers)

    unmapped <- character(0)
    skipped <- character(0)
    filled <- character(0)

    n <- length(keys)
    torch::with_no_grad({
        for (i in seq_along(keys)) {
            key <- keys[[i]]
            dest_name <- map_key(key)

            if (length(dest_name) != 1L || is.na(dest_name)) {
                skipped <- c(skipped, key)
                next
            }
            dest <- dests[[dest_name]]
            if (is.null(dest)) {
                unmapped <- c(unmapped, key)
                next
            }

            value <- ckpt$handle$get_tensor(key)
            if (!identical(as.integer(dest$shape), as.integer(value$shape))) {
                stop(sprintf(
                             "Shape mismatch for '%s' -> '%s': checkpoint [%s], module [%s]",
                             key, dest_name,
                             paste(value$shape, collapse = ","),
                             paste(dest$shape, collapse = ",")
                    ))
            }
            dest$copy_(value)
            filled <- c(filled, dest_name)
            rm(value)

            if (i %% gc_every == 0L) {
                gc(verbose = FALSE)
                if (verbose && n > 500L) {
                    message(sprintf("  loaded %d/%d tensors", i, n))
                }
            }
        }
    })
    gc(verbose = FALSE)

    unfilled <- setdiff(names(dests), filled)

    if (verbose) {
        message(sprintf(
                        "Loaded %d/%d tensors (%d skipped); unmapped: %d, unfilled: %d",
                        length(filled), n, length(skipped),
                        length(unmapped), length(unfilled)
            ))
        if (length(unmapped)) {
            message("  unmapped e.g.: ",
                    paste(utils::head(unmapped, 3), collapse = ", "))
        }
        if (length(unfilled)) {
            message("  unfilled e.g.: ", paste(utils::head(unfilled, 3), collapse = ", "))
        }
    }

    invisible(list(unmapped = unmapped, skipped = skipped, unfilled = unfilled))
}

#' Summarize checkpoint key coverage
#'
#' @param ckpt An \code{ltx23_checkpoint}.
#'
#' @return A data.frame with one row per component group and its key count.
#'
#' @export
ltx23_census <- function(ckpt) {
    stopifnot(inherits(ckpt, "ltx23_checkpoint"))
    groups <- ltx23_split_keys(ckpt$keys)
    data.frame(group = names(groups),
               keys = vapply(groups, length, integer(1)), row.names = NULL)
}

# Tests for the LTX-2.3 single-file checkpoint reader.
# Uses a tiny fake checkpoint written with official-style key names and
# model_version metadata; no model downloads required.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

make_fake_checkpoint <- function(path, version = "2.3.0") {
  tensors <- list(
    # DiT group
    "model.diffusion_model.patchify_proj.weight" =
      torch::torch_randn(8, 4),
    "model.diffusion_model.patchify_proj.bias" =
      torch::torch_randn(8),
    "model.diffusion_model.scale_shift_table" =
      torch::torch_randn(6, 8),
    "model.diffusion_model.transformer_blocks.0.attn1.to_q.weight" =
      torch::torch_randn(8, 8),
    # Connector group (inside and outside the diffusion_model prefix)
    "model.diffusion_model.video_embeddings_connector.transformer_1d_blocks.0.attn1.to_q.weight" =
      torch::torch_randn(8, 8),
    "model.diffusion_model.audio_embeddings_connector.transformer_1d_blocks.0.attn1.to_q.weight" =
      torch::torch_randn(8, 8),
    "text_embedding_projection.video_aggregate_embed.weight" =
      torch::torch_randn(8, 12),
    "text_embedding_projection.audio_aggregate_embed.weight" =
      torch::torch_randn(4, 12),
    # VAE / audio VAE / vocoder groups
    "vae.decoder.conv_in.conv.weight" = torch::torch_randn(4, 4, 3, 3, 3),
    "audio_vae.decoder.conv_in.weight" = torch::torch_randn(4, 2, 3, 3),
    "vocoder.vocoder.conv_in.weight" = torch::torch_randn(4, 4, 7),
    "vocoder.bwe_generator.conv_in.weight" = torch::torch_randn(4, 4, 7),
    "vocoder.mel_stft.mel_basis" = torch::torch_randn(64, 257)
  )
  metadata <- list(model_version = version)
  safetensors::safe_save_file(tensors, path, metadata = metadata)
  invisible(tensors)
}

tmp <- tempfile(fileext = ".safetensors")
on.exit(unlink(tmp), add = TRUE)
tensors <- make_fake_checkpoint(tmp)

# --- Open + version gate -----------------------------------------------------

ckpt <- ltx23_open_checkpoint(tmp)
expect_inherits(ckpt, "ltx23_checkpoint")
expect_equal(ckpt$version, "2.3.0")
expect_equal(sort(ckpt$keys), sort(names(tensors)))

# Wrong version is rejected
tmp_old <- tempfile(fileext = ".safetensors")
on.exit(unlink(tmp_old), add = TRUE)
make_fake_checkpoint(tmp_old, version = "2.0.1")
expect_error(ltx23_open_checkpoint(tmp_old), pattern = "2\\.3")

# Version check can be disabled
ckpt_old <- ltx23_open_checkpoint(tmp_old, require_version = NULL)
expect_equal(ckpt_old$version, "2.0.1")

# Missing file
expect_error(ltx23_open_checkpoint(tempfile()), pattern = "not found")

# --- Key group split ---------------------------------------------------------

groups <- ltx23_split_keys(ckpt$keys)
expect_equal(length(groups$dit), 4L)
expect_equal(length(groups$connectors), 4L)
expect_equal(length(groups$vae), 1L)
expect_equal(length(groups$audio_vae), 1L)
expect_equal(length(groups$vocoder), 3L)
expect_equal(length(groups$other), 0L)

# Every key lands in exactly one group
expect_equal(sort(unlist(groups, use.names = FALSE)), sort(ckpt$keys))

# Census matches
cen <- ltx23_census(ckpt)
expect_equal(cen$keys[cen$group == "vocoder"], 3L)

# --- Streaming group load ----------------------------------------------------

toy <- torch::nn_module(
  "toy",
  initialize = function() {
    self$patchify_proj <- torch::nn_linear(4, 8)
    self$scale_shift_table <- torch::nn_parameter(torch::torch_zeros(6, 8))
    self$blocks <- torch::nn_module_list(list(
      torch::nn_module(
        "toy_block",
        initialize = function() {
          self$to_q <- torch::nn_linear(8, 8, bias = FALSE)
        }
      )()
    ))
  }
)()

map_dit_toy <- function(key) {
  key <- sub("^model\\.diffusion_model\\.", "", key)
  key <- sub("^transformer_blocks\\.0\\.attn1\\.", "blocks.0.", key)
  key
}

res <- ltx23_load_group(ckpt, groups$dit, toy, map_key = map_dit_toy, verbose = FALSE)
expect_equal(length(res$unmapped), 0L)
expect_equal(length(res$unfilled), 0L)
expect_equal(length(res$skipped), 0L)

# Values actually copied
expect_equal(
  as.numeric(torch::torch_sum(toy$patchify_proj$weight)),
  as.numeric(torch::torch_sum(tensors[["model.diffusion_model.patchify_proj.weight"]])),
  tolerance = 1e-5
)

# copy_ converts dtype: load into a float64 module
toy64 <- torch::nn_module(
  "toy64",
  initialize = function() {
    self$w <- torch::nn_parameter(torch::torch_zeros(8, 4, dtype = torch::torch_float64()))
  }
)()
res64 <- ltx23_load_group(
  ckpt, "model.diffusion_model.patchify_proj.weight", toy64,
  map_key = function(key) "w", verbose = FALSE
)
expect_equal(length(res64$unmapped), 0L)
expect_equal(toy64$w$dtype$.type(), "Double")

# Unmapped keys are reported, not fatal
res_bad <- ltx23_load_group(
  ckpt, groups$vae, toy,
  map_key = identity, verbose = FALSE
)
expect_equal(length(res_bad$unmapped), 1L)

# Mapper can deliberately skip keys with NA
res_skip <- ltx23_load_group(
  ckpt, groups$vae, toy,
  map_key = function(key) NA_character_, verbose = FALSE
)
expect_equal(length(res_skip$skipped), 1L)
expect_equal(length(res_skip$unmapped), 0L)

# Shape mismatch is a hard error
expect_error(
  ltx23_load_group(
    ckpt, "model.diffusion_model.scale_shift_table", toy,
    map_key = function(key) "patchify_proj.weight", verbose = FALSE
  ),
  pattern = "Shape mismatch"
)

# --- Real checkpoint census (local only, needs the 46GB file) ----------------

if (at_home()) {
  real <- tryCatch(
    hfhub::hub_download(
      "Lightricks/LTX-2.3", "ltx-2.3-22b-distilled-1.1.safetensors",
      local_files_only = TRUE
    ),
    error = function(e) NULL
  )
  if (!is.null(real)) {
    rc <- ltx23_open_checkpoint(real)
    expect_equal(rc$version, "2.3.0")
    expect_equal(length(rc$keys), 5947L)
    rg <- ltx23_split_keys(rc$keys)
    expect_equal(length(rg$other), 0L)
    expect_equal(length(rg$vocoder), 1227L)
    expect_true(!is.null(rc$config$transformer$num_layers))
  }
}

#' Convert cornball SD 2.1 TorchScript weights to a diffusers artifact
#'
#' Rebuilds a diffusers-layout directory (\code{unet/}, \code{vae/},
#' \code{text_encoder/}) from the cornball-ai/sd21-R TorchScript
#' component \code{.pt} files, so the native safetensors pipeline
#' (\code{\link{download_sd21}} / \code{\link{sd_pipeline_from_safetensors}})
#' can load SD 2.1 with no TorchScript.
#'
#' A TorchScript trace preserves the exact parameter tensors, so the
#' result is bit-identical to the source at the chosen dtype. This is the
#' provenance-clean way to build the hosted artifact: the upstream
#' \code{stabilityai/stable-diffusion-2-1} repo was deprecated, SD 2.1 is
#' CreativeML OpenRAIL++-M (redistributable), and cornball already hosts
#' these weights as \code{.pt}. At \code{float16} the components are all
#' sub-2 GB single files (unet ~1.7 GB, text_encoder ~0.65 GB, vae
#' ~0.16 GB), so they load on stock CRAN safetensors.
#'
#' @param pt_dir Directory holding \code{unet-cpu.pt},
#'   \code{decoder-cpu.pt}, \code{text_encoder-cpu.pt} (default: the
#'   diffuseR \code{sd21} data location).
#' @param output_dir Output diffusers directory (default: the
#'   \code{sd_pipeline_from_safetensors} / \code{download_sd21} location).
#' @param dtype \code{"float16"} (default, the hosted tier) or
#'   \code{"float32"}.
#' @param verbose Logical.
#'
#' @return Invisibly, \code{output_dir}.
#'
#' @export
convert_sd21_pt_to_diffusers <- function(pt_dir = NULL, output_dir = NULL,
    dtype = c("float16", "float32"),
    verbose = TRUE) {
    dtype <- match.arg(dtype)
    if (!requireNamespace("safetensors", quietly = TRUE)) {
        stop("The safetensors package is required to write the artifact.")
    }
    if (is.null(pt_dir)) {
        pt_dir <- file.path(tools::R_user_dir("diffuseR", "data"), "sd21")
    }
    if (is.null(output_dir)) {
        output_dir <- file.path(tools::R_user_dir("diffuseR", "data"),
                                "sd21-diffusers")
    }
    pt_dir <- path.expand(pt_dir)
    td <- switch(dtype, float16 = torch::torch_float16(),
                 float32 = torch::torch_float32())

    comps <- list(
                  list(pt = "unet-cpu.pt", strip = "^unet\\.", subdir = "unet",
                       file = "diffusion_pytorch_model.safetensors"),
                  list(pt = "decoder-cpu.pt", strip = "^vae\\.", subdir = "vae",
                       file = "diffusion_pytorch_model.safetensors"),
                  list(pt = "text_encoder-cpu.pt", strip = "^text_encoder\\.",
                       subdir = "text_encoder", file = "model.safetensors")
    )
    absent <- Filter(function(c) !file.exists(file.path(pt_dir, c$pt)), comps)
    if (length(absent)) {
        stop("Missing TorchScript component(s) in ", pt_dir, ": ",
             paste(vapply(absent, `[[`, character(1), "pt"), collapse = ", "),
             ". Fetch them first (e.g. run a native sd21 pipeline once, or ",
             "download_model(\"sd21\")).")
    }
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

    for (comp in comps) {
        m <- torch::jit_load(file.path(pt_dir, comp$pt))
        params <- m$parameters
        keys <- sub(comp$strip, "", names(params))
        sd <- stats::setNames(
                              lapply(params, function(p) p$detach()$to(dtype = td)$contiguous()),
                              keys)
        d <- file.path(output_dir, comp$subdir)
        dir.create(d, showWarnings = FALSE)
        safetensors::safe_save_file(sd, file.path(d, comp$file))
        if (identical(comp$subdir, "text_encoder")) {
            .sd21_write_clip_config(params, file.path(d, "config.json"))
        }
        if (verbose) {
            message(sprintf("  %-13s %d tensors -> %s/%s", comp$subdir,
                            length(sd), comp$subdir, comp$file))
        }
        rm(m, params, sd)
        gc(verbose = FALSE)
    }

    if (verbose) {
        message("SD 2.1 diffusers artifact (", dtype, "): ", output_dir)
    }
    invisible(output_dir)
}

# Derive a minimal CLIPTextConfig (what text_encoder_native_from_safetensors
# reads) from the text_encoder .pt parameters, keyed with the export prefix.
.sd21_write_clip_config <- function(params, path) {
    g <- function(k) params[[paste0("text_encoder.", k)]]
    tok <- g("text_model.embeddings.token_embedding.weight")
    pos <- g("text_model.embeddings.position_embedding.weight")
    fc1 <- g("text_model.encoder.layers.0.mlp.fc1.weight")
    layers <- grep("encoder\\.layers\\.", names(params), value = TRUE)
    n_layers <- length(unique(sub(".*layers\\.([0-9]+)\\..*", "\\1", layers)))
    hidden <- as.integer(tok$shape[2])
    cfg <- list(vocab_size = as.integer(tok$shape[1]), hidden_size = hidden,
                num_hidden_layers = as.integer(n_layers),
                num_attention_heads = as.integer(hidden %/% 64L),
                intermediate_size = as.integer(fc1$shape[1]),
                max_position_embeddings = as.integer(pos$shape[1]))
    jsonlite::write_json(cfg, path, auto_unbox = TRUE, pretty = TRUE)
    invisible(cfg)
}

# SD 2.1 .pt -> diffusers converter: CLIPTextConfig derivation from the
# text_encoder parameters, and the missing-component error. The full
# conversion (bit-identical to the source) is validated out of band
# against the cached .pt weights.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
library(diffuseR)

# --- config derivation from mock text_encoder params (small dims) ------------------

z <- function(...) torch::torch_zeros(c(...))
p <- list()
p[["text_encoder.text_model.embeddings.token_embedding.weight"]] <- z(100L, 128L)
p[["text_encoder.text_model.embeddings.position_embedding.weight"]] <- z(77L, 128L)
for (i in 0:3) {
  p[[sprintf("text_encoder.text_model.encoder.layers.%d.mlp.fc1.weight", i)]] <-
    z(512L, 128L)
}
cfg <- diffuseR:::.sd21_write_clip_config(p, tempfile(fileext = ".json"))
expect_equal(cfg$vocab_size, 100L)
expect_equal(cfg$hidden_size, 128L)
expect_equal(cfg$max_position_embeddings, 77L)
expect_equal(cfg$num_hidden_layers, 4L)          # distinct layer indices
expect_equal(cfg$num_attention_heads, 2L)        # hidden / 64
expect_equal(cfg$intermediate_size, 512L)

# --- missing TorchScript components -> actionable error ---------------------------

expect_error(convert_sd21_pt_to_diffusers(pt_dir = tempfile()),
  pattern = "Missing")

#' Download and Prepare LTX-2.3 Model Weights
#'
#' Downloads the LTX-2.3 distilled checkpoint (46 GB, LTX-2 Community
#' License) and the Gemma3 text encoder from HuggingFace with an explicit
#' consent prompt, then quantizes the transformer to the local fp8
#' artifact (~26 GB) used by the GPU-poor pipeline.
#'
#' @name download_ltx23
NULL

.ltx23_checkpoint_repo <- "Lightricks/LTX-2.3"
.ltx23_checkpoint_file <- "ltx-2.3-22b-distilled-1.1.safetensors"
.ltx23_text_encoder_repo <- "Lightricks/LTX-2"

.ltx23_disk_free_gb <- function(path) {
    out <- tryCatch(
                    system2("df", c("-Pk", shQuote(path)), stdout = TRUE, stderr = FALSE),
                    error = function(e) NULL
    )
    if (is.null(out) || length(out) < 2L) {
        return(NA_real_)
    }
    fields <- strsplit(trimws(out[[2]]), "\\s+")[[1]]
    as.numeric(fields[4]) / 1024 ^ 2
}

.ltx23_consent <- function(what) {
    if (isTRUE(getOption("diffuseR.consent"))) {
        return(TRUE)
    }
    if (!interactive()) {
        stop(
             "Cannot download models in non-interactive mode without consent. ",
             "Set options(diffuseR.consent = TRUE) to allow downloads.",
             call. = FALSE
        )
    }
    isTRUE(utils::askYesNo(paste0("Download ", what, "?")))
}

#' Download the LTX-2.3 checkpoint and build a quantized artifact
#'
#' Skips work that is already done: a valid manifest short-circuits
#' everything; a cached 46 GB source skips the download. The source file
#' may be deleted after quantization (it is never removed automatically).
#'
#' Both quantized tiers are buildable here. \code{\link{recommend}}
#' returns nf4 for LTX on any card with 14 GB or more (it prefers nf4 at
#' 1280 px over fp8 at 1024 px, since video trades weight precision for
#' resolution), so nf4 is the tier most users want. fp8 additionally
#' needs a safetensors that can \emph{write} float8; asking for it
#' without one warns and builds nf4 instead rather than failing inside
#' the quantizer.
#'
#' @param quantize Logical. Build the quantized artifact after downloading.
#' @param precision "nf4" (~19 GB, readable by every safetensors) or
#'   "fp8" (~26 GB, needs float8 write support).
#' @param output_dir Directory for the artifact. NULL derives it from
#'   \code{precision}.
#' @param text_encoder Logical. Also fetch the Gemma3 text encoder and
#'   tokenizer (~25 GB, shared with LTX-2.0; from the Lightricks/LTX-2
#'   repo).
#' @param verbose Logical.
#'
#' @return Invisibly, a list with \code{checkpoint} (source path or NULL),
#'   \code{artifact_dir}, \code{precision}, \code{text_encoder_dir}, and
#'   \code{fp8_dir} for back-compatibility -- the artifact directory when
#'   \code{precision} is "fp8", NULL otherwise, since a field named
#'   \code{fp8_dir} pointing at an nf4 artifact would be a trap.
#'
#' @export
download_ltx2 <- function(quantize = TRUE, precision = c("nf4", "fp8"),
                          output_dir = NULL,
                          text_encoder = TRUE, verbose = TRUE) {
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to download model weights.")
    }
    precision <- match.arg(precision)
    # Explicit fp8 without float8 write support: warn and build nf4
    # instead of failing deep inside the quantizer.
    precision <- .st_graceful_precision(precision, mode = "write")
    if (is.null(output_dir)) {
        output_dir <- file.path(tools::R_user_dir("diffuseR", "data"),
                                paste0("ltx2.3-", precision))
    }
    art_gb <- if (identical(precision, "fp8")) 26 else 19
    result <- list(checkpoint = NULL, artifact_dir = output_dir,
                   precision = precision,
                   fp8_dir = if (identical(precision, "fp8")) output_dir,
                   text_encoder_dir = NULL)

    manifest_path <- file.path(output_dir, "manifest.json")
    have_artifact <- file.exists(manifest_path) && {
        m <- jsonlite::fromJSON(manifest_path)
        all(file.exists(file.path(output_dir, m$shards)))
    }

    if (!have_artifact || !quantize) {
        cached <- tryCatch(
                           hfhub::hub_download(.ltx23_checkpoint_repo, .ltx23_checkpoint_file,
                local_files_only = TRUE),
                           error = function(e) NULL
        )
        if (is.null(cached)) {
            free <- .ltx23_disk_free_gb(path.expand("~"))
            if (!is.na(free) && free < 46 + art_gb) {
                warning(sprintf(paste0("Only %.0f GB free; the download + ",
                                       "%s artifact need ~%d GB."),
                                free, precision, 46 + art_gb))
            }
            ok <- .ltx23_consent(sprintf(paste0(
                                        "the LTX-2.3 distilled checkpoint (46 GB) plus a ~%d GB local %s ",
                                        "artifact from HuggingFace (weights under the LTX-2 Community License)"),
                                        art_gb, precision))
            if (!ok) {
                stop("Download cancelled.", call. = FALSE)
            }
            if (verbose) {
                message("Downloading ", .ltx23_checkpoint_file, " (46 GB)...")
            }
            cached <- hfhub::hub_download(.ltx23_checkpoint_repo, .ltx23_checkpoint_file)
        }
        result$checkpoint <- cached

        if (quantize && !have_artifact) {
            if (verbose) {
                message("Quantizing transformer linears to ", precision,
                        " (one-time)...")
            }
            if (identical(precision, "fp8")) {
                ltx23_quantize_fp8(cached, output_dir, verbose = verbose)
            } else {
                ltx23_quantize_nf4(cached, output_dir, verbose = verbose)
            }
            if (verbose) {
                message(
                        toupper(precision), " artifact ready: ", output_dir, "\n",
                        "The 46 GB source in the HuggingFace cache may be deleted if ",
                        "you do not need bf16 weights."
                )
            }
        }
    } else if (verbose) {
        message(toupper(precision), " artifact already present: ", output_dir)
    }

    if (text_encoder) {
        te_files <- c(
                      "text_encoder/config.json",
                      sprintf("text_encoder/model-%05d-of-00011.safetensors", 1:11),
                      "text_encoder/model.safetensors.index.json",
                      "tokenizer/tokenizer.json", "tokenizer/tokenizer_config.json",
                      "tokenizer/special_tokens_map.json"
        )
        have_te <- .hub_all_cached(.ltx23_text_encoder_repo, te_files)
        if (!have_te) {
            ok <- .ltx23_consent(
                                 "the Gemma3 text encoder and tokenizer (~25 GB, Lightricks/LTX-2)"
            )
            if (!ok) {
                stop("Download cancelled.", call. = FALSE)
            }
        }
        if (verbose && !have_te) {
            message("Downloading Gemma3 text encoder...")
        }
        paths <- vapply(te_files, function(f) {
            tryCatch(hfhub::hub_download(.ltx23_text_encoder_repo, f),
                     error = function(e) NA_character_)
        }, character(1))
        result$text_encoder_dir <- dirname(paths[[2]])
    }

    invisible(result)
}

# TRUE when every file is already in the local hub cache, i.e. no
# network fetch will happen. Consent gates key on this - never on a
# single sentinel file, and never on artifact presence (a completed
# artifact must not license re-downloading sources).
.hub_all_cached <- function(repo, files, repo_type = NULL) {
    all(vapply(files, function(f) {
        args <- list(repo, f, local_files_only = TRUE)
        if (!is.null(repo_type)) {
            args$repo_type <- repo_type
        }
        !is.null(tryCatch(do.call(hfhub::hub_download, args),
                          error = function(e) NULL))
    }, logical(1)))
}

#' @importFrom utils head
NULL

# torch indexing uses `..` as a Python-style ellipsis
utils::globalVariables("..")

.onLoad <- function(libname, pkgname) {
    # start_torch() reads torch.threshold_call_gc exactly once at torch
    # init and there is no live setter (unlike the CUDA gates, which
    # ltx23_tune_gc pushes via cpp), so this is the only place the
    # option can land in time. The 4000 MB default fires an R gc for
    # every few GB of host allocation; raising it measurably cuts gc
    # counts, though an LTX load-time A/B showed no wall-clock change
    # (those gcs were cheap). A user-set option wins.
    if (is.null(getOption("torch.threshold_call_gc"))) {
        options(torch.threshold_call_gc = 16000)
    }
}

#' Save Video to File
#'
#' Saves a video array to a file in various formats.
#'
#' @param video Array of video frames with shape [T, H, W, C] where C is 3 (RGB).
#'   Values should be in [0, 1] range.
#' @param file Character. Output file path. Extension determines format.
#' @param fps Numeric. Frames per second (default 24).
#' @param format Character. Output format: "mp4", "gif", "webm", or "frames".
#'   If NULL, inferred from file extension.
#' @param backend Character. Backend to use: "ffmpeg", "av", or "auto".
#' @param quality Integer. Quality level 1-100 (for lossy formats).
#' @param verbose Logical. Print progress messages.
#'
#' @return Invisibly returns the output file path.
#'
#' @export
#'
#' @examples
#' video <- array(runif(4 * 16 * 16 * 3), dim = c(4, 16, 16, 3))
#'
#' # Individual PNG frames need no external encoder.
#' frame_dir <- file.path(tempdir(), "frames")
#' save_video(video, frame_dir, format = "frames", verbose = FALSE)
#' length(list.files(frame_dir, pattern = "[.]png$"))
#' unlink(frame_dir, recursive = TRUE)
#'
#' # MP4 and GIF need an ffmpeg binary or the 'av' package. Not shown as
#' # a live example: both backends hand off to an ffmpeg process that
#' # inherits this session's stdin, and R CMD check feeds the example
#' # script to R on stdin, so the encoder eats a byte of the script and
#' # every later example parses one character short.
#' # save_video(video, "output.mp4", fps = 24)
#' # save_video(video, "output.gif", fps = 10)
save_video <- function(video, file, fps = 24, format = NULL,
                       backend = "auto", quality = 85, verbose = TRUE) {
    # Validate input
    if (!is.array(video)) {
        stop("video must be an array")
    }

    dims <- dim(video)
    if (length(dims) != 4 || dims[4] != 3) {
        stop("video must have shape [T, H, W, 3] (RGB)")
    }

    num_frames <- dims[1]
    height <- dims[2]
    width <- dims[3]

    if (verbose) {
        message(sprintf("Video: %d frames, %dx%d", num_frames, width, height))
    }

    # Infer format from extension if not specified
    if (is.null(format)) {
        ext <- tolower(tools::file_ext(file))
        format <- switch(ext,
                         "mp4" = "mp4",
                         "gif" = "gif",
                         "webm" = "webm",
                         "png" = "frames",
                         "frames" # default to frames for directories
        )
        if (ext == "" && dir.exists(dirname(file))) {
            format <- "frames"
        }
    }

    # Clamp values to [0, 1]
    video <- pmax(pmin(video, 1), 0)

    # Dispatch to backend
    if (format == "frames") {
        save_frames(video, file, verbose = verbose)
    } else if (backend == "auto") {
        # Try ffmpeg first, fall back to av
        if (.ffmpeg_available()) {
            save_video_ffmpeg(video, file, fps = fps, format = format,
                              quality = quality, verbose = verbose)
        } else if (requireNamespace("av", quietly = TRUE)) {
            save_video_av(video, file, fps = fps, verbose = verbose)
        } else {
            stop("No video backend available. Install ffmpeg or the 'av' R package.")
        }
    } else if (backend == "ffmpeg") {
        save_video_ffmpeg(video, file, fps = fps, format = format,
                          quality = quality, verbose = verbose)
    } else if (backend == "av") {
        save_video_av(video, file, fps = fps, verbose = verbose)
    } else {
        stop("Unknown backend: ", backend)
    }

    invisible(file)
}

#' Save Video Frames as Individual Images
#'
#' @param video Array of video frames [T, H, W, C].
#' @param dir Directory to save frames in.
#' @param prefix Character. Filename prefix (default "frame_").
#' @param format Character. Image format: "png" or "jpg".
#' @param verbose Logical.
#'
#' @return Invisibly returns vector of saved file paths.
#'
#' @export
save_frames <- function(video, dir, prefix = "frame_", format = "png",
                        verbose = TRUE) {
    # Create directory if needed
    if (!dir.exists(dir)) {
        dir.create(dir, recursive = TRUE)
    }

    num_frames <- dim(video)[1]
    files <- character(num_frames)

    for (i in seq_len(num_frames)) {
        frame <- video[i,,,]
        filename <- file.path(dir, sprintf("%s%04d.%s", prefix, i, format))

        # Convert to 8-bit integer array
        frame_int <- as.integer(frame * 255)
        dim(frame_int) <- dim(frame)

        # Save using png or jpeg
        if (format == "png") {
            png::writePNG(frame_int / 255, filename)
        } else {
            jpeg::writeJPEG(frame_int / 255, filename)
        }

        files[i] <- filename
    }

    if (verbose) {
        message(sprintf("Saved %d frames to %s", num_frames, dir))
    }

    invisible(files)
}

#' Save Video using FFmpeg
#'
#' @param video Array of video frames [T, H, W, C].
#' @param file Output file path.
#' @param fps Frames per second.
#' @param format Output format.
#' @param quality Quality level 1-100.
#' @param verbose Logical.
#'
#' @keywords internal
save_video_ffmpeg <- function(video, file, fps = 24, format = "mp4",
                              quality = 85, verbose = TRUE) {
    if (!.ffmpeg_available()) {
        stop("ffmpeg not found in PATH")
    }

    # Create temp directory for frames
    temp_dir <- tempfile("frames_")
    dir.create(temp_dir)
    on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

    # Save frames as PNG
    if (verbose) {
        message("Writing frames...")
    }
    num_frames <- dim(video)[1]

    for (i in seq_len(num_frames)) {
        frame <- video[i,,,]
        filename <- file.path(temp_dir, sprintf("frame_%04d.png", i))
        png::writePNG(frame, filename)
    }

    # Build ffmpeg command
    input_pattern <- file.path(temp_dir, "frame_%04d.png")

    # Quality mapping (CRF for H.264: 0-51, lower is better)
    crf <- as.integer(51 - (quality / 100) * 51)

    if (format == "mp4") {
        cmd <- sprintf(
                       'ffmpeg -y -framerate %d -i "%s" -c:v libx264 -crf %d -pix_fmt yuv420p "%s"',
                       fps, input_pattern, crf, file
        )
    } else if (format == "webm") {
        cmd <- sprintf(
                       'ffmpeg -y -framerate %d -i "%s" -c:v libvpx-vp9 -crf %d -b:v 0 "%s"',
                       fps, input_pattern, crf, file
        )
    } else if (format == "gif") {
        # GIF requires palette generation for quality
        palette <- file.path(temp_dir, "palette.png")
        cmd1 <- sprintf(
                        'ffmpeg -y -framerate %d -i "%s" -vf "palettegen=stats_mode=diff" "%s"',
                        fps, input_pattern, palette
        )
        cmd2 <- sprintf(
                        'ffmpeg -y -framerate %d -i "%s" -i "%s" -lavfi "paletteuse=dither=bayer" "%s"',
                        fps, input_pattern, palette, file
        )
        if (verbose) {
            message("Generating palette...")
        }
        system(cmd1, ignore.stdout = !verbose, ignore.stderr = !verbose)
        cmd <- cmd2
    } else {
        stop("Unsupported format for ffmpeg: ", format)
    }

    if (verbose) {
        message(sprintf("Encoding %s...", format))
    }
    result <- system(cmd, ignore.stdout = !verbose, ignore.stderr = !verbose)

    if (result != 0) {
        stop("ffmpeg encoding failed")
    }

    if (verbose) {
        message(sprintf("Saved: %s", file))
    }
}

#' Save Video using av Package
#'
#' @param video Array of video frames [T, H, W, C].
#' @param file Output file path.
#' @param fps Frames per second.
#' @param verbose Logical.
#'
#' @keywords internal
save_video_av <- function(video, file, fps = 24, verbose = TRUE) {
    if (!requireNamespace("av", quietly = TRUE)) {
        stop("Package 'av' is required for this backend")
    }

    num_frames <- dim(video)[1]
    height <- dim(video)[2]
    width <- dim(video)[3]

    if (verbose) {
        message("Encoding video with av...")
    }

    # av::av_encode_video expects a function that returns frames
    # or a matrix/array. We need to convert our [T,H,W,C] to what av expects.

    # Create temp directory for frames
    temp_dir <- tempfile("frames_")
    dir.create(temp_dir)
    on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

    frame_files <- character(num_frames)
    for (i in seq_len(num_frames)) {
        frame <- video[i,,,]
        filename <- file.path(temp_dir, sprintf("frame_%04d.png", i))
        png::writePNG(frame, filename)
        frame_files[i] <- filename
    }

    # Use av to encode from image files
    av::av_encode_video(input = frame_files, output = file, framerate = fps,
                        verbose = verbose)

    if (verbose) {
        message(sprintf("Saved: %s", file))
    }
}

#' Check if FFmpeg is Available
#'
#' @return Logical. TRUE if ffmpeg is in PATH.
#' @keywords internal
.ffmpeg_available <- function() {
    result <- tryCatch(
                       system2("ffmpeg", "-version", stdout = FALSE, stderr = FALSE),
                       error = function(e) 1
    )
    result == 0
}

#' Create Video from Latents (Helper)
#'
#' Convenience function to decode latents and save video in one step.
#'
#' @param latents Tensor of latents from generation.
#' @param vae VAE decoder module.
#' @param file Output file path.
#' @param fps Frames per second.
#' @param ... Additional arguments to save_video.
#'
#' @return Invisibly returns the output file path.
#'
#' @export
latents_to_video <- function(latents, vae, file, fps = 24, ...) {
    torch::with_no_grad({
        # Decode latents
        video_tensor <- vae$decode(latents)

        # Convert to array [T, H, W, C]
        video_array <- video_tensor$squeeze(1L)$permute(c(2, 3, 4, 1))$cpu()$numpy()

        # Clamp to [0, 1]
        video_array <- pmax(pmin(video_array, 1), 0)
    })

    # Save video
    save_video(video_array, file, fps = fps, ...)
}

# Tests for Video Output Utilities

# Test 1: save_video function exists
cat("Test 1: save_video function signature\n")
expect_true(is.function(save_video), info = "save_video should be a function")
expect_true(is.function(save_frames), info = "save_frames should be a function")
expect_true(is.function(latents_to_video), info = "latents_to_video should be a function")

# Test 2: save_video parameters
cat("Test 2: save_video parameters\n")
params <- names(formals(save_video))
expect_true("video" %in% params, info = "Should have video param")
expect_true("file" %in% params, info = "Should have file param")
expect_true("fps" %in% params, info = "Should have fps param")
expect_true("format" %in% params, info = "Should have format param")
expect_true("backend" %in% params, info = "Should have backend param")

# Test 3: Create test video array
cat("Test 3: Create test video array\n")
# Create a simple 10-frame gradient video [T, H, W, C]
num_frames <- 10L
height <- 64L
width <- 64L
video <- array(0, dim = c(num_frames, height, width, 3L))

for (t in seq_len(num_frames)) {
  # Create a gradient that changes over time
  for (h in seq_len(height)) {
    for (w in seq_len(width)) {
      video[t, h, w, 1] <- (h - 1) / (height - 1)  # Red gradient vertical
      video[t, h, w, 2] <- (w - 1) / (width - 1)  # Green gradient horizontal
      video[t, h, w, 3] <- (t - 1) / (num_frames - 1)  # Blue changes over time
    }
  }
}

expect_equal(dim(video), c(num_frames, height, width, 3L), info = "Video array shape correct")
expect_true(min(video) >= 0, info = "Video values >= 0")
expect_true(max(video) <= 1, info = "Video values <= 1")

# Test 4: save_frames
cat("Test 4: save_frames to temp directory\n")
temp_dir <- tempfile("frames_test_")
saved_files <- save_frames(video, temp_dir, verbose = FALSE)
expect_equal(length(saved_files), num_frames, info = "Should save all frames")
expect_true(dir.exists(temp_dir), info = "Directory should exist")
expect_true(file.exists(saved_files[1]), info = "First frame should exist")
expect_true(file.exists(saved_files[num_frames]), info = "Last frame should exist")

# Check file naming
expect_true(grepl("frame_0001\\.png$", saved_files[1]), info = "Frame naming correct")

# Cleanup
unlink(temp_dir, recursive = TRUE)

# Test 5: ffmpeg availability check
cat("Test 5: FFmpeg availability check\n")
has_ffmpeg <- diffuseR:::.ffmpeg_available()
expect_true(is.logical(has_ffmpeg), info = ".ffmpeg_available returns logical")
cat(sprintf("  FFmpeg available: %s\n", has_ffmpeg))

# Test 6: save_video with frames format
cat("Test 6: save_video with frames format\n")
temp_dir2 <- tempfile("video_frames_")
save_video(video, temp_dir2, format = "frames", verbose = FALSE)
expect_true(dir.exists(temp_dir2), info = "Output directory created")
frame_files <- list.files(temp_dir2, pattern = "\\.png$")
expect_equal(length(frame_files), num_frames, info = "All frames saved")
unlink(temp_dir2, recursive = TRUE)

# Test 7: Format inference from extension
cat("Test 7: Format inference\n")
# Create temp files to test format detection
temp_mp4 <- tempfile(fileext = ".mp4")
temp_gif <- tempfile(fileext = ".gif")
temp_webm <- tempfile(fileext = ".webm")

expect_equal(tools::file_ext(temp_mp4), "mp4", info = "MP4 extension detected")
expect_equal(tools::file_ext(temp_gif), "gif", info = "GIF extension detected")
expect_equal(tools::file_ext(temp_webm), "webm", info = "WebM extension detected")

# Test 8: save_video to MP4 (if ffmpeg available)
cat("Test 8: save_video to MP4\n")
if (has_ffmpeg) {
  temp_mp4 <- tempfile(fileext = ".mp4")
  tryCatch({
    save_video(video, temp_mp4, fps = 10, verbose = FALSE)
    expect_true(file.exists(temp_mp4), info = "MP4 file created")
    expect_true(file.size(temp_mp4) > 0, info = "MP4 file has content")
    cat("  MP4 saved successfully\n")
  }, error = function(e) {
    cat(sprintf("  MP4 save failed: %s\n", e$message))
  })
  unlink(temp_mp4)
} else {
  cat("  Skipped (ffmpeg not available)\n")
}

# Test 9: save_video to GIF (if ffmpeg available)
cat("Test 9: save_video to GIF\n")
if (has_ffmpeg) {
  temp_gif <- tempfile(fileext = ".gif")
  tryCatch({
    save_video(video, temp_gif, fps = 5, verbose = FALSE)
    expect_true(file.exists(temp_gif), info = "GIF file created")
    expect_true(file.size(temp_gif) > 0, info = "GIF file has content")
    cat("  GIF saved successfully\n")
  }, error = function(e) {
    cat(sprintf("  GIF save failed: %s\n", e$message))
  })
  unlink(temp_gif)
} else {
  cat("  Skipped (ffmpeg not available)\n")
}

cat("\nVideo output tests completed\n")

#' LTX-2.3 Transformer Building Blocks
#'
#' Fresh R port of the LTX-2 transformer components from the diffusers
#' reference (Apache-2.0, src/diffusers/models/transformers/
#' transformer_ltx2.py and shared modules). Field names mirror the
#' diffusers module tree so checkpoint keys map 1:1.
#'
#' @name dit_ltx23_modules
NULL

#' RMS normalization
#'
#' Variance is computed in float32; the result is cast back to the input
#' dtype (or the weight dtype when elementwise affine).
#'
#' @param dim Integer. Normalized dimension size.
#' @param eps Numeric. Stability epsilon.
#' @param elementwise_affine Logical. Learn a scale weight.
#'
#' @return Module whose forward(x) returns \code{x} RMS-normalized over
#'   the last axis and cast back to the input dtype, a tensor of the
#'   same shape.
#'
#' @export
ltx23_rms_norm <- torch::nn_module(
                                   "ltx23_rms_norm",
                                   initialize = function(dim, eps = 1e-6, elementwise_affine = TRUE) {
    self$eps <- eps
    self$elementwise_affine <- elementwise_affine
    if (elementwise_affine) {
        self$weight <- torch::nn_parameter(torch::torch_ones(dim))
    }
},
                                   forward = function(x) {
    input_dtype <- x$dtype
    variance <- x$to(dtype = torch::torch_float32())$pow(2)$mean(dim = -1L,
        keepdim = TRUE)
    x <- x * torch::torch_rsqrt(variance + self$eps)
    if (self$elementwise_affine) {
        x <- x$to(dtype = self$weight$dtype) * self$weight
    } else {
        x <- x$to(dtype = input_dtype)
    }
    x
}
)

#' Sinusoidal timestep embedding
#'
#' DDPM-style sinusoidal embedding. LTX uses \code{flip_sin_to_cos=TRUE}
#' (cos first) and \code{downscale_freq_shift=0}.
#'
#' @param timesteps 1D tensor of timestep values.
#' @param embedding_dim Integer. Output embedding size.
#' @param flip_sin_to_cos Logical. Put cos before sin.
#' @param downscale_freq_shift Numeric. Frequency delta control.
#' @param max_period Numeric. Maximum embedding frequency period.
#'
#' @return Tensor [N, embedding_dim].
#'
#' @export
ltx23_get_timestep_embedding <- function(timesteps, embedding_dim,
    flip_sin_to_cos = TRUE,
    downscale_freq_shift = 0,
    max_period = 10000) {
    stopifnot(timesteps$ndim == 1L)
    half_dim <- embedding_dim %/% 2L
    exponent <- -log(max_period) * torch::torch_arange(start = 0,
        end = half_dim - 1, dtype = torch::torch_float32(),
        device = timesteps$device)
    exponent <- exponent / (half_dim - downscale_freq_shift)

    emb <- torch::torch_exp(exponent)
    emb <- timesteps$unsqueeze(2L)$to(dtype = torch::torch_float32()) * emb$unsqueeze(1L)
    emb <- torch::torch_cat(list(torch::torch_sin(emb), torch::torch_cos(emb)), dim = -1L)

    if (flip_sin_to_cos) {
        emb <- torch::torch_cat(
                                list(emb[, (half_dim + 1):(2 * half_dim)], emb[, 1:half_dim]),
                                dim = -1L
        )
    }
    if (embedding_dim %% 2L == 1L) {
        emb <- torch::nnf_pad(emb, c(0L, 1L, 0L, 0L))
    }
    emb
}

# Two-layer timestep MLP (linear_1 -> silu -> linear_2), matching
# diffusers TimestepEmbedding state-dict names.
ltx23_timestep_embedding <- torch::nn_module(
    "ltx23_timestep_embedding",
    initialize = function(in_channels, time_embed_dim, bias = TRUE) {
    self$linear_1 <- linear_noinit(in_channels, time_embed_dim, bias = bias)
    self$linear_2 <- linear_noinit(time_embed_dim, time_embed_dim, bias = bias)
},
    forward = function(sample) {
    self$linear_2(torch::nnf_silu(self$linear_1(sample)))
}
)

# Sinusoidal projection (256 channels) + MLP, matching diffusers
# PixArtAlphaCombinedTimestepSizeEmbeddings without additional conditions.
ltx23_combined_timestep_embed <- torch::nn_module(
    "ltx23_combined_timestep_embed",
    initialize = function(embedding_dim) {
    self$timestep_embedder <- ltx23_timestep_embedding(256L, embedding_dim)
},
    forward = function(timestep, hidden_dtype) {
    proj <- ltx23_get_timestep_embedding(timestep, 256L,
        flip_sin_to_cos = TRUE, downscale_freq_shift = 0)
    self$timestep_embedder(proj$to(dtype = hidden_dtype))
}
)

#' Adaptive layer norm single (adaLN-single)
#'
#' Embeds a timestep/sigma and projects it to a configurable number of
#' modulation parameter vectors.
#'
#' @param embedding_dim Integer. Model dimension.
#' @param num_mod_params Integer. Number of modulation parameter vectors.
#'
#' @return Module whose forward returns
#'   \code{list(mod_params [N, num_mod_params * dim], embedded_timestep [N, dim])}.
#'
#' @export
ltx23_ada_layer_norm_single <- torch::nn_module(
    "ltx23_ada_layer_norm_single",
    initialize = function(embedding_dim, num_mod_params = 6L) {
    self$num_mod_params <- num_mod_params
    self$emb <- ltx23_combined_timestep_embed(embedding_dim)
    self$linear <- linear_noinit(embedding_dim,
                                 num_mod_params * embedding_dim, bias = TRUE)
},
    forward = function(timestep, hidden_dtype) {
    embedded_timestep <- self$emb(timestep, hidden_dtype = hidden_dtype)
    list(self$linear(torch::nnf_silu(embedded_timestep)), embedded_timestep)
}
)

# torch_matmul_out / torch_softmax_out are unexported torch internals;
# resolve them once and fall back to allocating ops if they disappear
.ltx23_attn_out_fns <- local({
    fns <- NULL
    function() {
        if (is.null(fns)) {
            fns <<- tryCatch(
                             list(
                                  matmul = get("torch_matmul_out",
                        envir = asNamespace("torch")),
                                  softmax = get("torch_softmax_out", envir = asNamespace("torch"))
                ),
                             error = function(e) FALSE
            )
        }
        fns
    }
})

# Persistent attention scratch (score matrix + context output), keyed by
# role/shape/dtype/device. Role is required because score and output shapes can
# coincide; out= matmul must never alias one of its inputs.
.ltx23_attn_scratch <- new.env(parent = emptyenv())

.ltx23_get_attn_buffer <- function(role, shape, dtype, device) {
    key <- paste(role, paste(shape, collapse = "x"), dtype$.type(),
                 paste(device$type, device$index %||% 0L), sep = "|")
    buf <- .ltx23_attn_scratch[[key]]
    if (is.null(buf)) {
        buf <- torch::torch_empty(shape, dtype = dtype, device = device)
        .ltx23_attn_scratch[[key]] <- buf
    }
    buf
}

# Manual scaled dot-product attention on [B, H, S, D] tensors. The
# [B, H, Sq, Sk] score matrix materializes, so queries are chunked
# adaptively against a memory budget and all large temporaries live in
# reusable scratch buffers.
.ltx23_sdpa_manual <- function(query, key, value, attention_mask = NULL,
                               chunk_size = NULL) {
    head_dim <- query$shape[length(query$shape)]
    scale <- 1.0 / sqrt(head_dim)
    key_t <- key$transpose(-2L, -1L)
    fns <- .ltx23_attn_out_fns()

    b <- query$shape[1]
    heads <- query$shape[2]
    n_q <- query$shape[3]
    n_k <- key$shape[3]
    d_v <- value$shape[4]

    attend <- function(q, mask) {
        attn <- torch::torch_matmul(q$mul(scale), key_t)
        if (!is.null(mask)) {
            attn <- attn + mask
        }
        attn <- torch::nnf_softmax(attn, dim = -1L)
        torch::torch_matmul(attn, value)
    }

    # Buffered variant: score matrix and context reuse persistent scratch
    attend_into <- function(q, mask, attn_buf, out_buf) {
        fns$matmul(attn_buf, q$mul(scale), key_t)
        if (!is.null(mask)) {
            attn_buf$add_(mask)
        }
        fns$softmax(attn_buf, attn_buf, dim = -1L, dtype = attn_buf$dtype)
        fns$matmul(out_buf, attn_buf, value)
        out_buf
    }

    # Adaptive query chunking: bound the [B, H, chunk, S_k] score matrix
    # to a memory budget regardless of sequence length. The default is
    # sized so the self+cross scratch pair stays under ~400MB at
    # 1280x704x121 (14080 video tokens), the validated ceiling of the
    # 16GB NF4-resident profile; below ~2000 tokens it has no effect.
    budget <- getOption("diffuseR.attn_budget", 1.5e8)
    auto_chunk <- max(256L, as.integer(budget / (b * heads * n_k * 2)))
    if (is.null(chunk_size)) {
        chunk_size <- auto_chunk
    } else {
        chunk_size <- min(chunk_size, auto_chunk)
    }

    if (isFALSE(fns)) {
        # Fallback: allocating path
        if (n_q <= chunk_size) {
            return(attend(query, attention_mask))
        }
        outs <- list()
        start <- 1L
        while (start <= n_q) {
            len <- min(chunk_size, n_q - start + 1L)
            q_chunk <- query$narrow(3L, start, len)
            mask_chunk <- attention_mask
            if (!is.null(mask_chunk) && mask_chunk$shape[3] > 1L) {
                mask_chunk <- mask_chunk$narrow(3L, start, len)
            }
            outs[[length(outs) + 1L]] <- attend(q_chunk, mask_chunk)
            start <- start + len
        }
        return(torch::torch_cat(outs, dim = 3L))
    }

    rows <- min(n_q, chunk_size)
    attn_buf <- .ltx23_get_attn_buffer("scores", c(b, heads, rows, n_k),
                                       query$dtype, query$device)
    out_buf <- .ltx23_get_attn_buffer(
                                      "output", c(b, heads, n_q, d_v), query$dtype, query$device
    )

    if (n_q <= chunk_size) {
        return(attend_into(query, attention_mask, attn_buf, out_buf))
    }

    start <- 1L
    while (start <= n_q) {
        len <- min(chunk_size, n_q - start + 1L)
        q_chunk <- query$narrow(3L, start, len)
        mask_chunk <- attention_mask
        if (!is.null(mask_chunk) && mask_chunk$shape[3] > 1L) {
            mask_chunk <- mask_chunk$narrow(3L, start, len)
        }
        attend_into(
                    q_chunk, mask_chunk,
                    attn_buf$narrow(3L, 1L, len),
                    out_buf$narrow(3L, start, len)
        )
        start <- start + len
    }
    out_buf
}

# Resolve the exported fused SDPA binding once. torch 0.17.0 ships this,
# but capability-based lookup preserves compatibility with older builds.
.ltx23_fused_sdpa_fn <- local({
    fn <- NULL
    resolved <- FALSE
    function() {
        if (!resolved) {
            if ("torch_scaled_dot_product_attention" %in%
                              getNamespaceExports("torch")) {
                fn <<- getExportedValue("torch",
                                        "torch_scaled_dot_product_attention")
            }
            resolved <<- TRUE
        }
        fn
    }
})

.ltx23_fused_sdpa <- function(query, key, value, attention_mask = NULL) {
    fn <- .ltx23_fused_sdpa_fn()
    if (is.null(fn)) {
        stop("Fused scaled-dot-product attention is unavailable.",
             call. = FALSE)
    }
    if (is.null(attention_mask)) {
        fn(query, key, value, dropout_p = 0, is_causal = FALSE)
    } else {
        fn(query, key, value, attn_mask = attention_mask,
            dropout_p = 0, is_causal = FALSE)
    }
}

# Prefer native fused SDPA when no explicit query chunk was requested. Set
# options(diffuseR.ltx23_fused_sdpa = FALSE) to force the manual fallback.
.ltx23_sdpa <- function(query, key, value, attention_mask = NULL,
                        chunk_size = NULL) {
    use_fused <- is.null(chunk_size) &&
    isTRUE(getOption("diffuseR.ltx23_fused_sdpa", TRUE)) &&
    !is.null(.ltx23_fused_sdpa_fn())
    if (use_fused) {
        return(.ltx23_fused_sdpa(query, key, value, attention_mask))
    }
    .ltx23_sdpa_manual(query, key, value, attention_mask, chunk_size)
}

#' Release the attention scratch buffers
#'
#' @return Invisibly, NULL.
#'
#' @keywords internal
.ltx23_release_attn_buffers <- function() {
    rm(list = ls(.ltx23_attn_scratch), envir = .ltx23_attn_scratch)
    invisible(NULL)
}

#' LTX-2 attention layer
#'
#' Attention with RMS q/k norms across heads, optional per-head output
#' gating (LTX-2.3), separate query/key RoPE (for a2v/v2a cross
#' attention), and optional STG perturbation (skip attention, use the
#' value projection).
#'
#' @param query_dim Integer. Query feature dimension.
#' @param heads,kv_heads Integers. Attention head counts.
#' @param dim_head Integer. Per-head dimension.
#' @param bias,out_bias Logicals. Projection biases.
#' @param cross_attention_dim Integer or NULL. Key/value input dimension
#'   (NULL for self-attention).
#' @param norm_eps Numeric. RMS norm epsilon.
#' @param norm_elementwise_affine Logical. RMS norms carry weights.
#' @param rope_type "split" or "interleaved".
#' @param apply_gated_attention Logical. Add per-head sigmoid output gates.
#'
#' @return Module whose forward(hidden_states, ...) returns the attended
#'   states [B, S, query_dim] after the output projection.
#'
#' @export
ltx23_attention <- torch::nn_module(
                                    "ltx23_attention",
                                    initialize = function(
        query_dim,
        heads = 8L,
        kv_heads = NULL,
        dim_head = 64L,
        bias = TRUE,
        cross_attention_dim = NULL,
        out_bias = TRUE,
        norm_eps = 1e-6,
        norm_elementwise_affine = TRUE,
        rope_type = "split",
        apply_gated_attention = FALSE
    ) {
    kv_heads <- kv_heads %||% heads
    self$heads <- as.integer(heads)
    self$head_dim <- as.integer(dim_head)
    inner_dim <- dim_head * heads
    inner_kv_dim <- dim_head * kv_heads
    cross_dim <- cross_attention_dim %||% query_dim
    self$rope_type <- rope_type
    self$attn_chunk <- NULL # set by memory profiles; NULL = unchunked

    self$norm_q <- ltx23_rms_norm(inner_dim, eps = norm_eps,
                                  elementwise_affine = norm_elementwise_affine)
    self$norm_k <- ltx23_rms_norm(inner_kv_dim, eps = norm_eps,
                                  elementwise_affine = norm_elementwise_affine)
    self$to_q <- linear_noinit(query_dim, inner_dim, bias = bias)
    self$to_k <- linear_noinit(cross_dim, inner_kv_dim, bias = bias)
    self$to_v <- linear_noinit(cross_dim, inner_kv_dim, bias = bias)
    self$to_out <- torch::nn_module_list(list(
            linear_noinit(inner_dim, query_dim, bias = out_bias)
        ))

    self$apply_gated_attention <- apply_gated_attention
    if (apply_gated_attention) {
        # Per-head gate logits, computed on the raw block input (pre-Q)
        self$to_gate_logits <- linear_noinit(query_dim, heads, bias = TRUE)
    }
},
                                    forward = function(
        hidden_states,
        encoder_hidden_states = NULL,
        attention_mask = NULL,
        query_rotary_emb = NULL,
        key_rotary_emb = NULL,
        perturbation_mask = NULL,
        all_perturbed = FALSE
    ) {
    if (is.null(encoder_hidden_states)) {
        encoder_hidden_states <- hidden_states
    }
    if (!is.null(attention_mask)) {
        # [B, 1, S] or [B, Tq, S] additive bias -> [B, 1, *, S]
        if (attention_mask$ndim == 3L) {
            attention_mask <- attention_mask$unsqueeze(2L)
        }
    }

    gate_logits <- NULL
    if (self$apply_gated_attention) {
        gate_logits <- self$to_gate_logits(hidden_states) # [B, T, H]
    }

    value_flat <- self$to_v(encoder_hidden_states)

    if (isTRUE(all_perturbed)) {
        # STG: skip attention entirely, use the value projection
        out <- value_flat
    } else {
        query <- self$norm_q(self$to_q(hidden_states))
        key <- self$norm_k(self$to_k(encoder_hidden_states))

        key_rope <- key_rotary_emb %||% query_rotary_emb
        if (!is.null(query_rotary_emb)) {
            if (self$rope_type == "interleaved") {
                query <- ltx23_apply_interleaved_rotary_emb(query, query_rotary_emb)
                key <- ltx23_apply_interleaved_rotary_emb(key, key_rope)
            } else {
                query <- ltx23_apply_split_rotary_emb(query, query_rotary_emb)
                key <- ltx23_apply_split_rotary_emb(key, key_rope)
            }
        }

        # [B, S, H*D] -> [B, H, S, D]
        query <- query$unflatten(3L, c(self$heads, -1L))$transpose(2L, 3L)
        key <- key$unflatten(3L, c(self$heads, -1L))$transpose(2L, 3L)
        value <- value_flat$unflatten(3L, c(self$heads, -1L))$transpose(2L, 3L)

        out <- .ltx23_sdpa(query, key, value, attention_mask, self$attn_chunk)
        out <- out$transpose(2L, 3L)$flatten(start_dim = 3L)
        out <- out$to(dtype = hidden_states$dtype)

        if (!is.null(perturbation_mask)) {
            # Interpolate between the perturbed (value) and full attention paths
            out <- torch::torch_lerp(value_flat, out, perturbation_mask)
        }
    }

    if (self$apply_gated_attention) {
        out <- out$unflatten(3L, c(self$heads, -1L)) # [B, T, H, D]
        # Factor 2 so zero-initialized gate logits give unit gates
        gates <- gate_logits$sigmoid()$mul(2.0) # [B, T, H]
        out <- out * gates$unsqueeze(-1L)
        out <- out$flatten(start_dim = 3L)
    }

    self$to_out[[1]](out)
}
)

# In-place tanh GELU for large activations: nnf_gelu allocates a second
# copy of its input, which at high resolution doubles the feed-forward
# intermediate (the single largest transient in a block). Uses the
# internal in-place kernel when available, chunked so any internal
# temporaries stay small; falls back to the allocating path.
.ltx23_gelu_tanh_inplace <- function(x, chunk_elements = 2 ^ 24) {
    fn <- get0("torch_gelu_", envir = asNamespace("torch"))
    if (is.null(fn)) {
        return(torch::nnf_gelu(x, approximate = "tanh"))
    }
    flat <- x$view(-1L)
    n <- flat$numel()
    start <- 1
    while (start <= n) {
        len <- min(chunk_elements, n - start + 1)
        fn(flat$narrow(1L, start, len), approximate = "tanh")
        start <- start + len
    }
    x
}

#' LTX feed-forward layer
#'
#' Linear -> GELU (tanh approximation) -> Linear with 4x hidden dim,
#' matching diffusers \code{FeedForward(activation_fn="gelu-approximate")}
#' state-dict names (net.0.proj, net.2).
#'
#' @param dim Integer. Input/output dimension.
#' @param mult Integer. Hidden dimension multiplier.
#'
#' @return Module whose forward(x) returns the projected states passed
#'   through a tanh-approximated GELU, a tensor of the same shape as
#'   \code{x}.
#'
#' @export
ltx23_feed_forward <- torch::nn_module(
                                       "ltx23_feed_forward",
                                       initialize = function(dim, mult = 4L) {
    inner_dim <- as.integer(dim * mult)
    gelu_proj <- torch::nn_module(
                                  "ltx23_gelu",
                                  initialize = function(dim_in, dim_out) {
        self$proj <- linear_noinit(dim_in, dim_out)
    },
                                  forward = function(x) {
        h <- self$proj(x)
        if (h$numel() > getOption("diffuseR.gelu_inplace_min", 1e8)) {
            return(.ltx23_gelu_tanh_inplace(h))
        }
        torch::nnf_gelu(h, approximate = "tanh")
    }
    )
    self$net <- torch::nn_module_list(list(
            gelu_proj(dim, inner_dim),
            torch::nn_identity(), # dropout slot in the reference; inference no-op
            linear_noinit(inner_dim, dim)
        ))
},
                                       forward = function(x) {
    self$net[[3]](self$net[[1]](x))
}
)

# Combine a per-block scale/shift table with global modulation params:
# ada_values = table[None, None] + temb.reshape(B, temb_tokens, num, -1),
# returned as a list split along the parameter axis.
ltx23_get_mod_params <- function(scale_shift_table, temb, batch_size) {
    num_params <- scale_shift_table$shape[1]
    ada_values <- scale_shift_table$unsqueeze(1L)$unsqueeze(1L)$to(device = temb$device) +
    temb$reshape(c(batch_size, temb$shape[2], num_params, -1L))
    torch::torch_unbind(ada_values, dim = 3L)
}

#' LTX-2 transformer block
#'
#' Dual-stream (video + audio) block: modulated self-attention per
#' modality, text cross-attention per modality (with LTX-2.3 query and
#' key/value modulation), bidirectional audio-video cross-attention with
#' global+per-block modulation, and modulated feed-forward.
#'
#' @param dim,audio_dim Integers. Video/audio stream dimensions.
#' @param num_attention_heads,attention_head_dim Video attention shape.
#' @param cross_attention_dim Integer. Text embedding dim for video.
#' @param audio_num_attention_heads,audio_attention_head_dim Audio attention shape.
#' @param audio_cross_attention_dim Integer. Text embedding dim for audio.
#' @param video_gated_attn,audio_gated_attn Logicals. Per-head output gates.
#' @param video_cross_attn_adaln,audio_cross_attn_adaln Logicals. LTX-2.3
#'   text cross-attention modulation (9 mod params instead of 6).
#' @param eps Numeric. Norm epsilon.
#' @param elementwise_affine Logical. Block norms carry weights (FALSE for LTX).
#' @param rope_type "split" or "interleaved".
#' @param perturbed_attn Logical. Enable the STG perturbation arguments.
#'
#' @return Module whose forward(hidden_states, ...) returns
#'   \code{list(hidden_states, audio_hidden_states)}, the video and audio
#'   streams after self-attention, cross-attention and the feed-forward,
#'   each the same shape as its input.
#'
#' @export
ltx23_transformer_block <- torch::nn_module(
    "ltx23_transformer_block",
    initialize = function(
                          dim,
                          num_attention_heads,
                          attention_head_dim,
                          cross_attention_dim,
                          audio_dim,
                          audio_num_attention_heads,
                          audio_attention_head_dim,
                          audio_cross_attention_dim,
                          video_gated_attn = TRUE,
                          video_cross_attn_adaln = TRUE,
                          audio_gated_attn = TRUE,
                          audio_cross_attn_adaln = TRUE,
                          eps = 1e-6,
                          elementwise_affine = FALSE,
                          rope_type = "split",
                          perturbed_attn = TRUE
    ) {
    self$perturbed_attn <- perturbed_attn

    # 1. Self-attention (video and audio)
    self$norm1 <- ltx23_rms_norm(dim, eps = eps,
                                 elementwise_affine = elementwise_affine)
    self$attn1 <- ltx23_attention(
                                  query_dim = dim, heads = num_attention_heads, dim_head = attention_head_dim,
                                  rope_type = rope_type, apply_gated_attention = video_gated_attn
    )
    self$audio_norm1 <- ltx23_rms_norm(audio_dim, eps = eps, elementwise_affine = elementwise_affine)
    self$audio_attn1 <- ltx23_attention(
                                        query_dim = audio_dim, heads = audio_num_attention_heads,
                                        dim_head = audio_attention_head_dim,
                                        rope_type = rope_type, apply_gated_attention = audio_gated_attn
    )

    # 2. Prompt cross-attention
    self$norm2 <- ltx23_rms_norm(dim, eps = eps, elementwise_affine = elementwise_affine)
    self$attn2 <- ltx23_attention(
                                  query_dim = dim, cross_attention_dim = cross_attention_dim,
                                  heads = num_attention_heads, dim_head = attention_head_dim,
                                  rope_type = rope_type, apply_gated_attention = video_gated_attn
    )
    self$audio_norm2 <- ltx23_rms_norm(audio_dim, eps = eps, elementwise_affine = elementwise_affine)
    self$audio_attn2 <- ltx23_attention(
                                        query_dim = audio_dim, cross_attention_dim = audio_cross_attention_dim,
                                        heads = audio_num_attention_heads, dim_head = audio_attention_head_dim,
                                        rope_type = rope_type, apply_gated_attention = audio_gated_attn
    )

    # 3. Audio-to-video (Q: video) and video-to-audio (Q: audio) cross-attention
    self$audio_to_video_norm <- ltx23_rms_norm(dim, eps = eps, elementwise_affine = elementwise_affine)
    self$audio_to_video_attn <- ltx23_attention(
        query_dim = dim, cross_attention_dim = audio_dim,
        heads = audio_num_attention_heads, dim_head = audio_attention_head_dim,
        rope_type = rope_type, apply_gated_attention = video_gated_attn
    )
    self$video_to_audio_norm <- ltx23_rms_norm(audio_dim, eps = eps, elementwise_affine = elementwise_affine)
    self$video_to_audio_attn <- ltx23_attention(
        query_dim = audio_dim, cross_attention_dim = dim,
        heads = audio_num_attention_heads, dim_head = audio_attention_head_dim,
        rope_type = rope_type, apply_gated_attention = audio_gated_attn
    )

    # 4. Feed-forward
    self$norm3 <- ltx23_rms_norm(dim, eps = eps, elementwise_affine = elementwise_affine)
    self$ff <- ltx23_feed_forward(dim)
    self$audio_norm3 <- ltx23_rms_norm(audio_dim, eps = eps, elementwise_affine = elementwise_affine)
    self$audio_ff <- ltx23_feed_forward(audio_dim)

    # 5. Per-block modulation tables
    self$video_cross_attn_adaln <- video_cross_attn_adaln
    self$audio_cross_attn_adaln <- audio_cross_attn_adaln
    self$cross_attn_adaln <- video_cross_attn_adaln || audio_cross_attn_adaln
    video_mod_params <- if (video_cross_attn_adaln) 9L else 6L
    audio_mod_params <- if (audio_cross_attn_adaln) 9L else 6L
    self$scale_shift_table <- torch::nn_parameter(
        torch::torch_randn(video_mod_params, dim) / sqrt(dim)
    )
    self$audio_scale_shift_table <- torch::nn_parameter(
        torch::torch_randn(audio_mod_params, audio_dim) / sqrt(audio_dim)
    )
    if (self$cross_attn_adaln) {
        self$prompt_scale_shift_table <- torch::nn_parameter(torch::torch_randn(2L, dim))
        self$audio_prompt_scale_shift_table <- torch::nn_parameter(torch::torch_randn(2L, audio_dim))
    }
    self$video_a2v_cross_attn_scale_shift_table <- torch::nn_parameter(torch::torch_randn(5L, dim))
    self$audio_a2v_cross_attn_scale_shift_table <- torch::nn_parameter(torch::torch_randn(5L, audio_dim))
},
    forward = function(
                       hidden_states,
                       audio_hidden_states,
                       encoder_hidden_states,
                       audio_encoder_hidden_states,
                       temb,
                       temb_audio,
                       temb_ca_scale_shift,
                       temb_ca_audio_scale_shift,
                       temb_ca_gate,
                       temb_ca_audio_gate,
                       temb_prompt = NULL,
                       temb_prompt_audio = NULL,
                       video_rotary_emb = NULL,
                       audio_rotary_emb = NULL,
                       ca_video_rotary_emb = NULL,
                       ca_audio_rotary_emb = NULL,
                       encoder_attention_mask = NULL,
                       audio_encoder_attention_mask = NULL,
                       self_attention_mask = NULL,
                       use_a2v_cross_attention = TRUE,
                       use_v2a_cross_attention = TRUE,
                       perturbation_mask = NULL,
                       all_perturbed = FALSE
    ) {
    batch_size <- hidden_states$shape[1]

    # 1.1 Video self-attention
    video_ada <- ltx23_get_mod_params(self$scale_shift_table, temb, batch_size)
    shift_msa <- video_ada[[1]]; scale_msa <- video_ada[[2]]; gate_msa <- video_ada[[3]]
    shift_mlp <- video_ada[[4]]; scale_mlp <- video_ada[[5]]; gate_mlp <- video_ada[[6]]

    norm_hidden_states <- self$norm1(hidden_states)
    norm_hidden_states <- norm_hidden_states * scale_msa$add(1) + shift_msa

    attn_hidden_states <- self$attn1(
                                     norm_hidden_states,
                                     query_rotary_emb = video_rotary_emb,
                                     attention_mask = self_attention_mask,
                                     perturbation_mask = if (self$perturbed_attn) {
            perturbation_mask
        } else {
            NULL
        },
                                     all_perturbed = if (self$perturbed_attn) all_perturbed else FALSE
    )
    hidden_states <- hidden_states + attn_hidden_states * gate_msa

    # 1.2 Audio self-attention
    audio_ada <- ltx23_get_mod_params(self$audio_scale_shift_table, temb_audio, batch_size)
    audio_shift_msa <- audio_ada[[1]]; audio_scale_msa <- audio_ada[[2]]
    audio_gate_msa <- audio_ada[[3]]; audio_shift_mlp <- audio_ada[[4]]
    audio_scale_mlp <- audio_ada[[5]]; audio_gate_mlp <- audio_ada[[6]]

    norm_audio_hidden_states <- self$audio_norm1(audio_hidden_states)
    norm_audio_hidden_states <- norm_audio_hidden_states * audio_scale_msa$add(1) + audio_shift_msa

    attn_audio_hidden_states <- self$audio_attn1(
        norm_audio_hidden_states,
        query_rotary_emb = audio_rotary_emb,
        perturbation_mask = if (self$perturbed_attn) {
            perturbation_mask
        } else {
            NULL
        },
        all_perturbed = if (self$perturbed_attn) all_perturbed else FALSE
    )
    audio_hidden_states <- audio_hidden_states + attn_audio_hidden_states * audio_gate_msa

    # 2. Text cross-attention modulation params (LTX-2.3)
    if (self$cross_attn_adaln) {
        prompt_ada <- ltx23_get_mod_params(self$prompt_scale_shift_table, temb_prompt, batch_size)
        shift_text_kv <- prompt_ada[[1]]; scale_text_kv <- prompt_ada[[2]]
        audio_prompt_ada <- ltx23_get_mod_params(
            self$audio_prompt_scale_shift_table, temb_prompt_audio, batch_size
        )
        audio_shift_text_kv <- audio_prompt_ada[[1]]; audio_scale_text_kv <- audio_prompt_ada[[2]]
    }

    # 2.1 Video-text cross-attention
    norm_hidden_states <- self$norm2(hidden_states)
    if (self$video_cross_attn_adaln) {
        shift_text_q <- video_ada[[7]]; scale_text_q <- video_ada[[8]]; gate_text_q <- video_ada[[9]]
        norm_hidden_states <- norm_hidden_states * scale_text_q$add(1) + shift_text_q
    }
    enc_states <- encoder_hidden_states
    if (self$cross_attn_adaln) {
        enc_states <- enc_states * scale_text_kv$add(1) + shift_text_kv
    }
    attn_hidden_states <- self$attn2(
                                     norm_hidden_states,
                                     encoder_hidden_states = enc_states,
                                     attention_mask = encoder_attention_mask
    )
    if (self$video_cross_attn_adaln) {
        attn_hidden_states <- attn_hidden_states * gate_text_q
    }
    hidden_states <- hidden_states + attn_hidden_states

    # 2.2 Audio-text cross-attention
    norm_audio_hidden_states <- self$audio_norm2(audio_hidden_states)
    if (self$audio_cross_attn_adaln) {
        audio_shift_text_q <- audio_ada[[7]]; audio_scale_text_q <- audio_ada[[8]]
        audio_gate_text_q <- audio_ada[[9]]
        norm_audio_hidden_states <- norm_audio_hidden_states * audio_scale_text_q$add(1) +
        audio_shift_text_q
    }
    audio_enc_states <- audio_encoder_hidden_states
    if (self$cross_attn_adaln) {
        audio_enc_states <- audio_enc_states * audio_scale_text_kv$add(1) + audio_shift_text_kv
    }
    attn_audio_hidden_states <- self$audio_attn2(
        norm_audio_hidden_states,
        encoder_hidden_states = audio_enc_states,
        attention_mask = audio_encoder_attention_mask
    )
    if (self$audio_cross_attn_adaln) {
        attn_audio_hidden_states <- attn_audio_hidden_states * audio_gate_text_q
    }
    audio_hidden_states <- audio_hidden_states + attn_audio_hidden_states

    # 3. Audio-to-video and video-to-audio cross-attention
    if (use_a2v_cross_attention || use_v2a_cross_attention) {
        norm_hidden_states <- self$audio_to_video_norm(hidden_states)
        norm_audio_hidden_states <- self$video_to_audio_norm(audio_hidden_states)

        video_ca_ada <- ltx23_get_mod_params(
            self$video_a2v_cross_attn_scale_shift_table$narrow(1L, 1L, 4L),
            temb_ca_scale_shift, batch_size
        )
        video_ca_gate <- ltx23_get_mod_params(
            self$video_a2v_cross_attn_scale_shift_table$narrow(1L, 5L, 1L),
            temb_ca_gate, batch_size
        )
        a2v_gate <- video_ca_gate[[1]]$squeeze(3L)

        audio_ca_ada <- ltx23_get_mod_params(
            self$audio_a2v_cross_attn_scale_shift_table$narrow(1L, 1L, 4L),
            temb_ca_audio_scale_shift, batch_size
        )
        audio_ca_gate <- ltx23_get_mod_params(
            self$audio_a2v_cross_attn_scale_shift_table$narrow(1L, 5L, 1L),
            temb_ca_audio_gate, batch_size
        )
        v2a_gate <- audio_ca_gate[[1]]$squeeze(3L)

        if (use_a2v_cross_attention) {
            mod_norm_hidden <- norm_hidden_states *
            video_ca_ada[[1]]$squeeze(3L)$add(1) + video_ca_ada[[2]]$squeeze(3L)
            mod_norm_audio <- norm_audio_hidden_states *
            audio_ca_ada[[1]]$squeeze(3L)$add(1) + audio_ca_ada[[2]]$squeeze(3L)

            a2v_attn <- self$audio_to_video_attn(
                mod_norm_hidden,
                encoder_hidden_states = mod_norm_audio,
                query_rotary_emb = ca_video_rotary_emb,
                key_rotary_emb = ca_audio_rotary_emb
            )
            hidden_states <- hidden_states + a2v_gate * a2v_attn
        }

        if (use_v2a_cross_attention) {
            mod_norm_hidden <- norm_hidden_states *
            video_ca_ada[[3]]$squeeze(3L)$add(1) + video_ca_ada[[4]]$squeeze(3L)
            mod_norm_audio <- norm_audio_hidden_states *
            audio_ca_ada[[3]]$squeeze(3L)$add(1) + audio_ca_ada[[4]]$squeeze(3L)

            v2a_attn <- self$video_to_audio_attn(
                mod_norm_audio,
                encoder_hidden_states = mod_norm_hidden,
                query_rotary_emb = ca_audio_rotary_emb,
                key_rotary_emb = ca_video_rotary_emb
            )
            audio_hidden_states <- audio_hidden_states + v2a_gate * v2a_attn
        }
    }

    # 4. Feed-forward
    norm_hidden_states <- self$norm3(hidden_states) * scale_mlp$add(1) + shift_mlp
    hidden_states <- hidden_states + self$ff(norm_hidden_states) * gate_mlp

    norm_audio_hidden_states <- self$audio_norm3(audio_hidden_states) *
    audio_scale_mlp$add(1) + audio_shift_mlp
    audio_hidden_states <- audio_hidden_states + self$audio_ff(norm_audio_hidden_states) *
    audio_gate_mlp

    list(hidden_states, audio_hidden_states)
}
)

# Parity tests for the LTX-2.3 DiT port against diffusers reference
# fixtures (generated by tools/gen_fixtures_dit.py, checked in).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "dit_ltx23.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/dit_ltx23.safetensors"
if (!file.exists(fixture_path)) exit_file("dit fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# Copy fixture weights (diffusers state-dict names) into an R module.
# Errors if any name has no destination or any destination stays unfilled.
load_named_weights <- function(module, weights) {
  dests <- c(module$named_parameters(), module$named_buffers())
  missing_dest <- setdiff(names(weights), names(dests))
  if (length(missing_dest)) {
    stop("No destination for: ", paste(utils::head(missing_dest, 5), collapse = ", "))
  }
  unfilled <- setdiff(names(dests), names(weights))
  if (length(unfilled)) {
    stop("Unfilled params: ", paste(utils::head(unfilled, 5), collapse = ", "))
  }
  torch::with_no_grad({
    for (name in names(weights)) dests[[name]]$copy_(weights[[name]])
  })
  invisible(module)
}

fixture_group <- function(prefix) {
  keys <- grep(paste0("^", prefix, "\\."), names(fx), value = TRUE)
  w <- fx[keys]
  names(w) <- sub(paste0("^", prefix, "\\."), "", keys)
  w
}

# --- Scratch buffers keep score and output roles distinct -------------------------
release_attn_buffers <- diffuseR:::.ltx23_release_attn_buffers
get_attn_buffer <- diffuseR:::.ltx23_get_attn_buffer
release_attn_buffers()
collision_shape <- c(1L, 2L, 8L, 8L)
score_buf <- get_attn_buffer(
  "scores", collision_shape, torch::torch_float32(), torch::torch_device("cpu")
)
output_buf <- get_attn_buffer(
  "output", collision_shape, torch::torch_float32(), torch::torch_device("cpu")
)
score_buf$fill_(1)
output_buf$fill_(2)
score_buf$add_(1)
expect_equal(as.numeric(output_buf$mean()), 2)
release_attn_buffers()
# --- Fused SDPA matches the chunked manual fallback ---------------------------
sdpa <- diffuseR:::.ltx23_sdpa
sdpa_manual <- diffuseR:::.ltx23_sdpa_manual
fused_sdpa <- diffuseR:::.ltx23_fused_sdpa
fused_sdpa_fn <- diffuseR:::.ltx23_fused_sdpa_fn
attn_scratch <- get(".ltx23_attn_scratch", envir = asNamespace("diffuseR"))

torch::torch_manual_seed(23)
q <- torch::torch_randn(c(1L, 2L, 5L, 8L))
k <- torch::torch_randn(c(1L, 2L, 7L, 8L))
v <- torch::torch_randn(c(1L, 2L, 7L, 6L))
mask <- torch::torch_zeros(c(1L, 1L, 5L, 7L))
mask[, , , 7] <- -10000

manual_mask <- sdpa_manual(q, k, v, mask, chunk_size = 2L)$clone()
manual_nomask <- sdpa_manual(q, k, v, chunk_size = 2L)$clone()
if (!is.null(fused_sdpa_fn())) {
  release_attn_buffers()
  fused_mask <- fused_sdpa(q, k, v, mask)
  fused_nomask <- fused_sdpa(q, k, v)
  expect_true(max_abs_diff(fused_mask, manual_mask) < 1e-5)
  expect_true(max_abs_diff(fused_nomask, manual_nomask) < 1e-5)

  old <- options(diffuseR.ltx23_fused_sdpa = TRUE)
  auto_mask <- sdpa(q, k, v, mask)
  options(old)
  expect_true(max_abs_diff(auto_mask, fused_mask) < 1e-6)
  expect_equal(length(ls(attn_scratch)), 0L)
}

release_attn_buffers()
old <- options(diffuseR.ltx23_fused_sdpa = FALSE)
fallback_mask <- sdpa(q, k, v, mask)$clone()
options(old)
expect_true(max_abs_diff(fallback_mask, manual_mask) < 1e-5)
expect_true(length(ls(attn_scratch)) > 0L)
release_attn_buffers()

# --- Gated attention with split RoPE ------------------------------------------

attn <- ltx23_attention(
  query_dim = 16L, heads = 2L, dim_head = 8L,
  rope_type = "split", apply_gated_attention = TRUE
)
attn$eval()
load_named_weights(attn, fixture_group("attn"))

rope <- list(fx$attn_cos, fx$attn_sin)
torch::with_no_grad({
  out <- attn(fx$attn_x, attention_mask = fx$attn_mask, query_rotary_emb = rope)
  out_nomask <- attn(fx$attn_x, query_rotary_emb = rope)
})
expect_true(max_abs_diff(out, fx$attn_out) < 1e-4)
expect_true(max_abs_diff(out_nomask, fx$attn_out_nomask) < 1e-4)

# Query chunking is numerically identical
attn$attn_chunk <- 2L
torch::with_no_grad({
  out_chunked <- attn(fx$attn_x, attention_mask = fx$attn_mask, query_rotary_emb = rope)
})
expect_true(max_abs_diff(out_chunked, fx$attn_out) < 1e-5)
attn$attn_chunk <- NULL

# --- AdaLN single (9 mod params) -----------------------------------------------

ada <- ltx23_ada_layer_norm_single(16L, num_mod_params = 9L)
ada$eval()
load_named_weights(ada, fixture_group("ada"))
torch::with_no_grad({
  ada_res <- ada(fx$ada_t, hidden_dtype = torch::torch_float32())
})
expect_true(max_abs_diff(ada_res[[1]], fx$ada_out) < 1e-4)
expect_true(max_abs_diff(ada_res[[2]], fx$ada_emb) < 1e-4)

# --- Tiny full model -------------------------------------------------------------

model <- ltx23_transformer(
  in_channels = 4L,
  out_channels = 4L,
  num_attention_heads = 2L,
  attention_head_dim = 8L,
  cross_attention_dim = 16L,
  audio_in_channels = 4L,
  audio_out_channels = 4L,
  audio_num_attention_heads = 2L,
  audio_attention_head_dim = 4L,
  audio_cross_attention_dim = 8L,
  num_layers = 2L,
  rope_type = "split"
)
model$eval()
load_named_weights(model, fixture_group("model"))

run_model <- function(...) {
  torch::with_no_grad({
    model(
      hidden_states = fx$m_hidden,
      audio_hidden_states = fx$m_audio_hidden,
      encoder_hidden_states = fx$m_enc,
      audio_encoder_hidden_states = fx$m_audio_enc,
      timestep = fx$m_timestep,
      sigma = fx$m_sigma,
      num_frames = 2L, height = 3L, width = 4L, fps = 24.0,
      audio_num_frames = 5L,
      use_cross_timestep = TRUE,
      ...
    )
  })
}

res <- run_model(
  encoder_attention_mask = fx$m_enc_mask,
  audio_encoder_attention_mask = fx$m_enc_mask
)
expect_equal(as.integer(res$sample$shape), as.integer(fx$m_out_video$shape))
expect_true(max_abs_diff(res$sample, fx$m_out_video) < 1e-4)
expect_true(max_abs_diff(res$audio_sample, fx$m_out_audio) < 1e-4)

# Isolated modalities (a2v/v2a off)
res_iso <- run_model(isolate_modalities = TRUE)
expect_true(max_abs_diff(res_iso$sample, fx$m_out_video_iso) < 1e-4)
expect_true(max_abs_diff(res_iso$audio_sample, fx$m_out_audio_iso) < 1e-4)

# STG perturbation on block index 1 (0-based, i.e. the second block)
res_stg <- run_model(spatio_temporal_guidance_blocks = c(1L))
expect_true(max_abs_diff(res_stg$sample, fx$m_out_video_stg) < 1e-4)
expect_true(max_abs_diff(res_stg$audio_sample, fx$m_out_audio_stg) < 1e-4)

# --- Key mapper ------------------------------------------------------------------

expect_equal(
  ltx23_map_dit_key("model.diffusion_model.patchify_proj.weight"),
  "proj_in.weight"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.audio_patchify_proj.bias"),
  "audio_proj_in.bias"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.adaln_single.emb.timestep_embedder.linear_1.weight"),
  "time_embed.emb.timestep_embedder.linear_1.weight"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.audio_adaln_single.linear.bias"),
  "audio_time_embed.linear.bias"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.prompt_adaln_single.linear.weight"),
  "prompt_adaln.linear.weight"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.audio_prompt_adaln_single.linear.weight"),
  "audio_prompt_adaln.linear.weight"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.av_ca_a2v_gate_adaln_single.linear.weight"),
  "av_cross_attn_video_a2v_gate.linear.weight"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.transformer_blocks.0.attn1.q_norm.weight"),
  "transformer_blocks.0.attn1.norm_q.weight"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.transformer_blocks.5.scale_shift_table_a2v_ca_video"),
  "transformer_blocks.5.video_a2v_cross_attn_scale_shift_table"
)
expect_equal(
  ltx23_map_dit_key("model.diffusion_model.transformer_blocks.0.ff.net.0.proj.weight"),
  "transformer_blocks.0.ff.net.0.proj.weight"
)
expect_equal(ltx23_map_dit_key("model.diffusion_model.scale_shift_table"), "scale_shift_table")

# --- In-place tanh GELU (large-activation feed-forward path) ------------------------

ff <- ltx23_feed_forward(8L)
x_ff <- torch::torch_randn(1L, 37L, 8L)
torch::with_no_grad({
  ref_ff <- ff(x_ff)
  # Force the in-place chunked path (threshold below this tensor's numel,
  # chunk smaller than the intermediate so several chunks run)
  op <- options(diffuseR.gelu_inplace_min = 1)
  inp_ff <- ff(x_ff)
  options(op)
})
expect_true(as.numeric((ref_ff - inp_ff)$abs()$max()) < 1e-6)

h_g <- torch::torch_randn(3L, 41L)
g_ref <- torch::nnf_gelu(h_g, approximate = "tanh")
torch::with_no_grad(
  g_inp <- diffuseR:::.ltx23_gelu_tanh_inplace(h_g$clone(), chunk_elements = 16)
)
expect_true(as.numeric((g_ref - g_inp)$abs()$max()) < 1e-6)

#' VRAM Detection and Management Utilities
#'
#' Device detection, VRAM reporting, and module offloading helpers shared
#' by the image and video pipelines.
#'
#' @name vram
NULL

#' Check if GPU is Blackwell Architecture
#'
#' Blackwell GPUs (RTX 50xx) may need special handling.
#'
#' @return Logical. TRUE if Blackwell GPU detected.
#'
#' @export
#'
#' @examples
#' # Soft probe: FALSE on any machine without a Blackwell card, and on
#' # machines where torch has no lantern binaries.
#' is_blackwell_gpu()
is_blackwell_gpu <- function() {
    # Check compute capability via torch. cuda_is_available() ERRORS
    # (not FALSE) when the torch package is installed without its
    # lantern binaries - fresh installs, win-builder, CRAN checks - so
    # probe soft: no lantern means no CUDA means not Blackwell.
    avail <- tryCatch(torch::cuda_is_available(), error = function(e) FALSE)
    if (avail) {
        cap <- tryCatch(torch::cuda_get_device_capability(0L),
                        error = function(e) NULL)
        if (!is.null(cap)) {
            # Blackwell is compute 12.x
            return(as.integer(cap[[1]]) >= 12L)
        }
    }

    FALSE
}

#' Detect Available VRAM
#'
#' Asks nvidia-smi.
#'
#' @param use_free Logical. If TRUE, return free VRAM. If FALSE, return total.
#'
#' @return Numeric. VRAM in GB, or 0 if no GPU detected.
#' @keywords internal
.detect_vram <- function(use_free = FALSE) {
    smi <- suppressWarnings(tryCatch(
                                     system2("nvidia-smi",
                c(paste0("--query-gpu=memory.",
                        if (use_free) "free" else "total"),
                    "--format=csv,noheader,nounits"),
                stdout = TRUE, stderr = FALSE),
                                     error = function(e) character(0)
        ))
    mb <- suppressWarnings(as.numeric(smi[1]))
    if (isTRUE(mb > 0)) {
        return(mb / 1024)
    }

    # Fallback: check if CUDA available but can't determine VRAM.
    # cuda_is_available() ERRORS (not FALSE) when torch is installed
    # without its lantern binaries - fresh installs, win-builder, CRAN -
    # and this branch is exactly where those machines land, since they
    # have no nvidia-smi either. Probe soft, same as is_blackwell_gpu().
    if (isTRUE(tryCatch(torch::cuda_is_available(),
                        error = function(e) FALSE))) {
        # Conservative estimate - assume 8GB if we can't detect
        message("Could not detect VRAM via nvidia-smi; assuming 8 GB.")
        return(8)
    }

    # No GPU detected
    0
}

#' Offload Module to CPU
#'
#' Moves a torch module and all its parameters to CPU.
#'
#' @param module A torch nn_module.
#' @param gc Logical. Run garbage collection after offload.
#'
#' @return The module (modified in place).
#'
#' @export
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   model <- torch::nn_linear(4, 2)
#'   offload_to_cpu(model)
#' }
offload_to_cpu <- function(module, gc = TRUE) {
    module$to(device = "cpu")
    if (gc && torch::cuda_is_available()) {
        gc()
        torch::cuda_empty_cache()
    }
    invisible(module)
}

#' Load Module to GPU
#'
#' Moves a torch module and all its parameters to CUDA.
#'
#' @param module A torch nn_module.
#' @param device Character. Target device (default "cuda").
#'
#' @return The module (modified in place).
#'
#' @export
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   model <- torch::nn_linear(4, 2)
#'   # "cuda" needs a GPU; "cpu" is the portable round trip.
#'   load_to_gpu(model, device = "cpu")
#' }
load_to_gpu <- function(module, device = "cuda") {
    module$to(device = device)
    invisible(module)
}

#' Report VRAM Usage
#'
#' Prints current VRAM usage from nvidia-smi.
#'
#' @param label Character. Label for the report.
#'
#' @return Invisibly returns a list with used and free VRAM in GB.
#'
#' @export
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   vram_report("After model load")
#' }
vram_report <- function(label = "") {
    if (!torch::cuda_is_available()) {
        message("[", label, "] No CUDA available")
        return(invisible(list(used = 0, free = 0)))
    }

    smi <- suppressWarnings(tryCatch(
                                     system2("nvidia-smi",
                c("--query-gpu=memory.used,memory.free",
                    "--format=csv,noheader,nounits"),
                stdout = TRUE, stderr = FALSE),
                                     error = function(e) character(0)
        ))
    vals <- suppressWarnings(as.numeric(strsplit(smi[1], ",")[[1]]))
    if (length(vals) == 2L && all(is.finite(vals))) {
        used <- vals[1] / 1024
        free <- vals[2] / 1024
        message(sprintf("[%s] VRAM: %.2f GB used, %.2f GB free", label,
                        used, free))
        return(invisible(list(used = used, free = free)))
    }

    message("[", label, "] VRAM: (nvidia-smi unavailable)")
    invisible(list(used = NA, free = NA))
}

#' Clear VRAM Cache
#'
#' Forces garbage collection and clears CUDA memory cache.
#'
#' @param verbose Logical. Print memory status before/after.
#'
#' @return Invisibly returns NULL.
#'
#' @export
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   clear_vram()
#' }
clear_vram <- function(verbose = FALSE) {
    if (!torch::cuda_is_available()) {
        return(invisible(NULL))
    }

    if (verbose) {
        vram_report("Before clear")
    }

    gc()
    tryCatch(torch::cuda_empty_cache(), error = function(e) NULL)

    if (verbose) {
        vram_report("After clear")
    }

    invisible(NULL)
}

# Helper: get all adjacent symbol pairs
token_get_pairs <- function(symbols) {
    pairs <- character(0)
    if (length(symbols) > 1) {
        for (i in seq_len(length(symbols) - 1)) {
            pairs <- c(pairs, paste(symbols[i], symbols[i + 1], sep = " "))
        }
    }
    unique(pairs)
}

# Helper: merge the first occurrence of a given bigram
token_merge_pair_once <- function(symbols, bigram) {
    parts <- strsplit(bigram, " ")[[1]]
    out <- c()
    i <- 1
    while (i <= length(symbols)) {
        if (i < length(symbols) && symbols[i] == parts[1] &&
            symbols[i + 1] == parts[2]) {
            out <- c(out, paste0(parts[1], parts[2]))
            i <- i + 2
        } else {
            out <- c(out, symbols[i])
            i <- i + 1
        }
    }
    out
}

#' Tokenize a prompt
#'
#' @param prompt A character string prompt describing the image to generate.
#' @param merges Path to the merges file (BPE merges).
#' @param vocab_file Path to the vocabulary file (token->id mapping).
#' @param pad_token The token ID used for padding (default is 0).
#'
#' @return A 2D torch tensor of shape c(1, 77) containing the token IDs.
#' @export
CLIPTokenizer <- function(prompt,
                          merges = system.file("tokenizer/merges.txt", package = "diffuseR"),
                          vocab_file = system.file("tokenizer/vocab.json", package = "diffuseR"),
                          pad_token = 0L) {
    # 1. Load merges and build BPE rank map
    merge_lines <- readLines(merges, encoding = "UTF-8")
    merge_lines <- merge_lines[-1] # drop header
    merge_lines <- merge_lines[nzchar(merge_lines)] # drop blanks
    merges_list <- strsplit(merge_lines, " ")
    merge_keys <- sapply(merges_list, paste, collapse = " ")
    bpe_ranks <- stats::setNames(seq_along(merge_keys), merge_keys)

    # 2. Load vocabulary JSON (token->id)
    vocab <- jsonlite::fromJSON(vocab_file)

    # 3. Prepare text: lowercase and split into words
    text <- tolower(prompt)
    words <- strsplit(text, "\\s+")[[1]]

    # 4. BPE tokenization per word
    tokens <- character(0)
    for (word in words) {
        chars <- strsplit(word, "")[[1]]
        symbols <- if (length(chars) > 0) {
            c(if (length(chars) > 1) chars[-length(chars)] else character(0),
                paste0(chars[length(chars)], "</w>"))
        } else {
            character(0)
        }
        repeat {
            pairs <- token_get_pairs(symbols)
            valid <- intersect(pairs, names(bpe_ranks))
            if (length(valid) == 0) {
                break
            }
            best <- valid[which.min(bpe_ranks[valid])]
            symbols <- token_merge_pair_once(symbols, best)
        }
        tokens <- c(tokens, symbols)
    }

    # Truncate tokens to max length
    max_tokens <- 77L - 2L # 2 for start and end tokens
    if (length(tokens) > max_tokens) {
        truncated_tokens <- tokens[(max_tokens + 1):length(tokens)]
        tokens <- tokens[1:max_tokens]
        warning("Prompt was truncated. Consider shortening it.")
        warning("Dropped prompt: ", paste(truncated_tokens, collapse = " "))
    }

    # 5. Map token strings to vocabulary IDs
    ids <- vapply(tokens, function(tok) {
        if (!tok %in% names(vocab)) stop(paste("Unknown token:", tok))
        vocab[[tok]]
    }, integer(1))

    # 6. Add special tokens and pad/truncate
    sot_id <- 49406L
    eot_id <- 49407L
    seq_ids <- c(sot_id, ids, eot_id)
    max_len <- 77L
    if (length(seq_ids) < max_len) {
        seq_ids <- c(seq_ids, rep(pad_token, max_len - length(seq_ids)))
    } else if (length(seq_ids) > max_len) {
        seq_ids <- seq_ids[1:max_len]
    }

    # 7. Return as 2D torch tensor [1, 77]
    t <- torch::torch_tensor(seq_ids, dtype = torch::torch_long())
    t <- t$unsqueeze(1) # prepend batch dimension
    t
}

#' Get SDXL Memory Profile
#'
#' Determines optimal memory configuration for SDXL image generation
#' based on available VRAM.
#'
#' @param vram_gb Numeric. Available VRAM in GB, or NULL for auto-detection.
#'
#' @return A list with memory profile settings.
#'
#' @details
#' Memory profiles for SDXL:
#' \describe{
#'   \item{full_gpu}{16GB+ - All components on CUDA}
#'   \item{balanced}{10-12GB - UNet + decoder on CUDA, text encoders on CPU}
#'   \item{unet_gpu}{6-10GB - Only UNet on CUDA, everything else CPU}
#'   \item{cpu_only}{<6GB - All on CPU}
#' }
#'
#' Each profile also specifies:
#' - cfg_mode: "batched" or "sequential" (sequential halves peak memory)
#' - cleanup: "none", "phase", or "step" (when to clear VRAM)
#' - dtype: "float16" or "float32"
#' - max_resolution: maximum image dimension
#'
#' @export
#'
#' @examples
#' # A stated VRAM budget is deterministic and needs no GPU.
#' str(sdxl_memory_profile(vram_gb = 8))
#'
#' str(sdxl_memory_profile(vram_gb = 24))
#'
#' # Auto-detect free VRAM on this machine.
#' str(sdxl_memory_profile())
sdxl_memory_profile <- function(vram_gb = NULL) {
    # Auto-detect free VRAM if not provided
    if (is.null(vram_gb)) {
        vram_gb <- .detect_vram(use_free = TRUE)
        message(sprintf("Detected %.1f GB free VRAM", vram_gb))
    }

    # Determine profile level
    if (vram_gb >= 16) {
        profile <- "full_gpu"
    } else if (vram_gb >= 10) {
        profile <- "balanced"
    } else if (vram_gb >= 6) {
        profile <- "unet_gpu"
    } else {
        profile <- "cpu_only"
    }

    # Build profile config
    profiles <- list(
                     full_gpu = list(
                                     name = "full_gpu",
                                     devices = list(unet = "cuda", decoder = "cuda",
                text_encoder = "cuda", text_encoder2 = "cuda",
                encoder = "cuda"),
                                     dtype = "float16",
                                     cfg_mode = "batched",
                                     cleanup = "none",
                                     max_resolution = 1536L,
                                     step_cleanup_interval = 0L # No step cleanup
        ),
                     balanced = list(
                                     name = "balanced",
                                     devices = list(
                unet = "cuda",
                decoder = "cuda",
                text_encoder = "cpu",
                text_encoder2 = "cpu",
                encoder = "cpu"
            ),
                                     dtype = "float16",
                                     cfg_mode = "batched",
                                     cleanup = "phase", # Cleanup between text encoding and denoising
                                     max_resolution = 1024L,
                                     step_cleanup_interval = 0L
        ),
                     unet_gpu = list(
                                     name = "unet_gpu",
                                     devices = list(
                unet = "cuda",
                decoder = "cpu",
                text_encoder = "cpu",
                text_encoder2 = "cpu",
                encoder = "cpu"
            ),
                                     dtype = "float16",
                                     cfg_mode = "sequential", # Sequential CFG halves peak memory
                                     cleanup = "phase",
                                     max_resolution = 1024L,
                                     step_cleanup_interval = 10L # Cleanup every 10 steps
        ),
                     cpu_only = list(
                                     name = "cpu_only",
                                     devices = list(
                unet = "cpu",
                decoder = "cpu",
                text_encoder = "cpu",
                text_encoder2 = "cpu",
                encoder = "cpu"
            ),
                                     dtype = "float32", # CPU often faster with float32
                                     cfg_mode = "sequential",
                                     cleanup = "none", # No GPU to clean
                                     max_resolution = 768L,
                                     step_cleanup_interval = 0L
        )
    )

    profiles[[profile]]
}

# Tests for SDXL GPU-Poor Memory Profiles

# Test 1: sdxl_memory_profile exists
cat("Test 1: sdxl_memory_profile function\n")
expect_true(is.function(sdxl_memory_profile), info = "sdxl_memory_profile should be a function")

# Test 2: Profile detection by VRAM
cat("Test 2: Profile detection by VRAM\n")
profile_full <- sdxl_memory_profile(vram_gb = 20)
expect_equal(profile_full$name, "full_gpu", info = "20GB should be full_gpu")

profile_balanced <- sdxl_memory_profile(vram_gb = 12)
expect_equal(profile_balanced$name, "balanced", info = "12GB should be balanced")

profile_unet <- sdxl_memory_profile(vram_gb = 8)
expect_equal(profile_unet$name, "unet_gpu", info = "8GB should be unet_gpu")

profile_cpu <- sdxl_memory_profile(vram_gb = 4)
expect_equal(profile_cpu$name, "cpu_only", info = "4GB should be cpu_only")

# Test 3: Profile structure
cat("Test 3: Profile structure\n")
profile <- sdxl_memory_profile(vram_gb = 10)
expect_true("devices" %in% names(profile), info = "Should have devices")
expect_true("dtype" %in% names(profile), info = "Should have dtype")
expect_true("cfg_mode" %in% names(profile), info = "Should have cfg_mode")
expect_true("cleanup" %in% names(profile), info = "Should have cleanup")
expect_true("max_resolution" %in% names(profile), info = "Should have max_resolution")
expect_true("step_cleanup_interval" %in% names(profile), info = "Should have step_cleanup_interval")

# Test 4: Device configuration
cat("Test 4: Device configurations\n")
full <- sdxl_memory_profile(vram_gb = 20)
expect_equal(full$devices$unet, "cuda", info = "full_gpu: unet on cuda")
expect_equal(full$devices$decoder, "cuda", info = "full_gpu: decoder on cuda")
expect_equal(full$devices$text_encoder, "cuda", info = "full_gpu: text_encoder on cuda")

balanced <- sdxl_memory_profile(vram_gb = 12)
expect_equal(balanced$devices$unet, "cuda", info = "balanced: unet on cuda")
expect_equal(balanced$devices$decoder, "cuda", info = "balanced: decoder on cuda")
expect_equal(balanced$devices$text_encoder, "cpu", info = "balanced: text_encoder on cpu")

unet <- sdxl_memory_profile(vram_gb = 8)
expect_equal(unet$devices$unet, "cuda", info = "unet_gpu: unet on cuda")
expect_equal(unet$devices$decoder, "cpu", info = "unet_gpu: decoder on cpu")
expect_equal(unet$devices$text_encoder, "cpu", info = "unet_gpu: text_encoder on cpu")

cpu <- sdxl_memory_profile(vram_gb = 4)
expect_equal(cpu$devices$unet, "cpu", info = "cpu_only: unet on cpu")
expect_equal(cpu$devices$decoder, "cpu", info = "cpu_only: decoder on cpu")

# Test 5: CFG mode by profile
cat("Test 5: CFG mode by profile\n")
expect_equal(sdxl_memory_profile(vram_gb = 20)$cfg_mode, "batched", info = "full_gpu uses batched CFG")
expect_equal(sdxl_memory_profile(vram_gb = 8)$cfg_mode, "sequential", info = "unet_gpu uses sequential CFG")

# Test 6: Cleanup settings
cat("Test 6: Cleanup settings\n")
expect_equal(sdxl_memory_profile(vram_gb = 20)$cleanup, "none", info = "full_gpu: no cleanup")
expect_equal(sdxl_memory_profile(vram_gb = 12)$cleanup, "phase", info = "balanced: phase cleanup")
expect_equal(sdxl_memory_profile(vram_gb = 8)$cleanup, "phase", info = "unet_gpu: phase cleanup")

# Test 7: Step cleanup interval
cat("Test 7: Step cleanup interval\n")
expect_equal(sdxl_memory_profile(vram_gb = 20)$step_cleanup_interval, 0L, info = "full_gpu: no step cleanup")
expect_equal(sdxl_memory_profile(vram_gb = 8)$step_cleanup_interval, 10L, info = "unet_gpu: cleanup every 10 steps")

# Test 8: txt2img_sdxl accepts memory_profile parameter
cat("Test 8: txt2img_sdxl memory_profile parameter\n")
params <- names(formals(txt2img_sdxl))
expect_true("memory_profile" %in% params, info = "txt2img_sdxl should have memory_profile param")
expect_true("verbose" %in% params, info = "txt2img_sdxl should have verbose param")

cat("\nSDXL GPU-poor tests completed\n")

#' Download and Prepare FLUX.1-schnell Weights
#'
#' Downloads FLUX.1-schnell from HuggingFace (weights Apache-2.0, but
#' the repo is gated behind a license click-through) and quantizes the
#' 12B transformer to a local NF4 (~7 GB) or fp8 (~12 GB) artifact.
#'
#' @name download_flux
NULL

.flux1_repo <- "black-forest-labs/FLUX.1-schnell"

.flux1_transformer_files <- c(
                              "transformer/config.json",
                              "transformer/diffusion_pytorch_model.safetensors.index.json",
                              sprintf("transformer/diffusion_pytorch_model-%05d-of-00003.safetensors",
                                      1:3)
)

.flux1_support_files <- c("vae/config.json",
                          "vae/diffusion_pytorch_model.safetensors",
                          "text_encoder/config.json",
                          "text_encoder/model.safetensors",
                          "scheduler/scheduler_config.json")

.flux1_t5_files <- c(
                     "text_encoder_2/config.json",
                     "text_encoder_2/model.safetensors.index.json",
                     sprintf("text_encoder_2/model-%05d-of-00002.safetensors", 1:2),
                     "tokenizer_2/tokenizer.json",
                     "tokenizer_2/tokenizer_config.json",
                     "tokenizer_2/special_tokens_map.json"
)

# hub_download with the gated-repo 401/403 turned into an actionable error
.flux1_download <- function(file, ...) {
    tryCatch(
             hfhub::hub_download(.flux1_repo, file, ...),
             error = function(e) {
        msg <- conditionMessage(e)
        if (grepl("401|403|[Uu]nauthorized|[Ff]orbidden", msg)) {
            stop(
                 "FLUX.1-schnell is a gated HuggingFace repo (the weights are ",
                 "Apache-2.0; the gate is a license click-through). To download:\n",
                 "  1. Log in at https://huggingface.co/black-forest-labs/FLUX.1-schnell ",
                 "and accept the license.\n",
                 "  2. Create a read token at https://huggingface.co/settings/tokens\n",
                 "  3. Sys.setenv(HF_TOKEN = \"hf_...\") and retry.",
                 call. = FALSE
            )
        }
        stop(e)
    }
    )
}

#' Download FLUX.1-schnell and build the quantized artifact
#'
#' Skips work already done: a valid quantized manifest short-circuits
#' the transformer download; cached files are not re-fetched. Needs
#' \code{HF_TOKEN} set for the gated repo (see the error message it
#' raises without one). The bf16 transformer source (~24 GB in the
#' HuggingFace cache) may be deleted after quantization.
#'
#' @param quantize Logical. Build the quantized artifact after
#'   downloading.
#' @param precision "nf4" (~7 GB, GPU-resident on 16 GB cards) or
#'   "fp8" (~12 GB, CPU-resident, streamed; near-bf16 quality).
#' @param output_dir Directory for the quantized artifact.
#' @param text_encoders Logical. Also fetch the CLIP + T5 text encoders,
#'   tokenizer, VAE, and scheduler config (~10 GB).
#' @param verbose Logical.
#'
#' @return Invisibly, a list with \code{transformer_dir},
#'   \code{artifact_dir}, and \code{support} (named file paths).
#'
#' @export
download_flux1 <- function(quantize = TRUE, precision = c("nf4", "fp8"),
                           output_dir = NULL, text_encoders = TRUE,
                           verbose = TRUE) {
    precision <- match.arg(precision)
    # Explicit fp8 without float8 support: warn + build nf4 instead of
    # failing in flux_quantize.
    precision <- .st_graceful_precision(precision, mode = "write")
    if (is.null(output_dir)) {
        output_dir <- file.path(tools::R_user_dir("diffuseR", "data"),
                                paste0("flux1-schnell-", precision))
    }
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to download model weights.")
    }
    result <- list(transformer_dir = NULL, artifact_dir = output_dir,
                   support = character(0))

    manifest_path <- file.path(output_dir, "manifest.json")
    have_artifact <- file.exists(manifest_path) && {
        m <- jsonlite::fromJSON(manifest_path)
        all(file.exists(file.path(output_dir, m$shards)))
    }

    if (!have_artifact || !quantize) {
        cached <- tryCatch(
                           hfhub::hub_download(.flux1_repo, .flux1_transformer_files[[3]],
                local_files_only = TRUE),
                           error = function(e) NULL
        )
        if (is.null(cached) ||
            !.hub_all_cached(.flux1_repo, .flux1_transformer_files)) {
            free <- .ltx23_disk_free_gb(path.expand("~"))
            if (!is.na(free) && free < 45) {
                warning(sprintf(
                                "Only %.0f GB free; the download + %s artifact need ~45 GB.",
                                free, precision
                    ))
            }
            ok <- .ltx23_consent(paste0(
                                        "FLUX.1-schnell: the 24 GB bf16 transformer plus a ~",
                    if (precision == "nf4") "7" else "12",
                                        " GB local ", precision,
                                        " artifact (weights Apache-2.0, gated HuggingFace repo)"
                ))
            if (!ok) {
                stop("Download cancelled.", call. = FALSE)
            }
            if (verbose) {
                message("Downloading the FLUX.1-schnell transformer (24 GB)...")
            }
        }
        paths <- vapply(.flux1_transformer_files, .flux1_download, character(1))
        result$transformer_dir <- dirname(paths[[1]])

        if (quantize && !have_artifact) {
            if (verbose) {
                message("Quantizing transformer linears to ", precision,
                        " (one-time)...")
            }
            flux_quantize(result$transformer_dir, output_dir,
                          format = precision, verbose = verbose)
            if (verbose) {
                message(
                        toupper(precision), " artifact ready: ", output_dir, "\n",
                        "The 24 GB source in the HuggingFace cache may be deleted ",
                        "if you do not need bf16 weights."
                )
            }
        }
    } else if (verbose) {
        message(toupper(precision), " artifact already present: ", output_dir)
    }

    if (text_encoders) {
        files <- c(.flux1_support_files, .flux1_t5_files)
        have_t5 <- .hub_all_cached(.flux1_repo, files)
        if (!have_t5) {
            ok <- .ltx23_consent(
                                 "the FLUX text encoders, VAE, and tokenizer (~10 GB)"
            )
            if (!ok) {
                stop("Download cancelled.", call. = FALSE)
            }
            if (verbose) {
                message("Downloading text encoders + VAE...")
            }
        }
        result$support <- vapply(files, .flux1_download, character(1))
    }

    invisible(result)
}

#' Z-Image-Turbo Text-to-Image Pipeline
#'
#' Z-Image-Turbo text-to-image, ported from the diffusers reference
#' (Apache-2.0, src/diffusers/pipelines/z_image/pipeline_z_image.py).
#' Turbo is guidance-distilled: 8 steps, no classifier-free guidance.
#' The FlowMatch schedule uses the checkpoint's static shift (3.0) on
#' sigmas linspace(1, 1/N, N); the model consumes the REVERSED
#' normalized timestep (1000 - t)/1000 and its output is negated before
#' the Euler step. The VAE is the FLUX.1 16-channel autoencoder
#' verbatim.
#'
#' @name txt2img_zimage
NULL

# Resolve a Z-Image support file from the HuggingFace cache
.zimage_cached <- function(file) {
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to locate model files.")
    }
    tryCatch(
             hfhub::hub_download(.zimage_repo, file, local_files_only = TRUE),
             error = function(e) {
        stop("Missing ", file, " in the HuggingFace cache; ",
             "run download_zimage_turbo() first.", call. = FALSE)
    }
    )
}

#' Load the Z-Image-Turbo pipeline
#'
#' Loads the quantized transformer artifact plus the 16-channel VAE
#' decoder, Qwen3-4B text encoder, and tokenizer from the HuggingFace
#' cache populated by \code{\link{download_zimage_turbo}}. With fp8
#' precision the ~6.3 GB transformer rides to the GPU per phase.
#'
#' @param model_dir Quantized artifact directory (default: the
#'   \code{download_zimage_turbo} location for \code{precision}), or a
#'   raw diffusers transformer directory.
#' @param device Character. Compute device.
#' @param precision "auto" (default: reuse an existing artifact, else
#'   fp8 when safetensors supports float8, else nf4), "fp8", or "nf4".
#' @param text_device Device for the Qwen3 encoder (default:
#'   \code{device}; it encodes in its own phase and offloads).
#' @param attn_chunk Integer or NULL. Attention query-chunk override.
#' @param phase_offload Logical. One GPU tenant per phase.
#' @param pin Logical or NULL. Page-lock the phase-swapped weights for
#'   DMA-rate transfer (see \code{\link{staging}}). NULL (default)
#'   resolves via \code{options(diffuseR.pin_staging)} then the
#'   host-RAM-aware \code{\link{recommend}} decision.
#' @param verbose Logical.
#'
#' @return A \code{zimage_pipeline} list.
#'
#' @export
zimage_load_pipeline <- function(model_dir = NULL, device = "cuda",
                                 precision = c("auto", "fp8", "nf4", "bf16"),
                                 text_device = NULL, attn_chunk = NULL,
                                 phase_offload = TRUE, pin = NULL,
                                 verbose = TRUE) {
    precision <- .flux_resolve_precision(match.arg(precision),
        file.path(tools::R_user_dir("diffuseR", "data"), "zimage-turbo-"))
    if (is.null(text_device)) {
        if (device == "cuda") {
            text_device <- "cuda"
        } else {
            text_device <- "cpu"
        }
    }
    if (is.null(model_dir)) {
        model_dir <- .flux_model_dir("zimage", precision,
            file.path(tools::R_user_dir("diffuseR", "data"),
                      "zimage-turbo-"))
    }

    ckpt <- if (file.exists(file.path(model_dir, "manifest.json"))) {
        flux_open_quantized(model_dir)
    } else {
        flux_open_checkpoint(model_dir)
    }
    pin <- .resolve_pin(pin, "zimage", ckpt$format %||% "full")

    if (!nzchar(Sys.getenv("PYTORCH_CUDA_ALLOC_CONF"))) {
        # Stable resident footprint: the native backend avoids
        # expandable_segments' page-unmap cost on per-step churn
        Sys.setenv(PYTORCH_CUDA_ALLOC_CONF = "backend:native")
    }
    if (device == "cuda") {
        # Footprint sized to the largest phase (the 8 GB Qwen3 encode),
        # same reasoning as the FLUX.2 klein pipeline
        .flux_gc_gates(footprint_gb = 12)
    }

    if (phase_offload) {
        component_device <- "cpu"
    } else {
        component_device <- device
    }

    pipe <- list(
                 format = ckpt$format %||% "full",
                 device = device,
                 text_device = text_device,
                 phase_offload = phase_offload,
                 attn_chunk = if (is.null(attn_chunk)) NULL else as.integer(attn_chunk),
                 config = ckpt$config
    )

    if (verbose) {
        message("Loading transformer (", pipe$format, ")...")
    }
    pipe$transformer <- flux_load_transformer(
        ckpt, device = component_device,
        dtype = if (device == "cpu") "float32" else "bfloat16",
        pin = FALSE,
        fp8_resident = FALSE,
        verbose = verbose
    )
    # Resident fp8 happens at onload time (the weights ride to the GPU
    # with the phase and back off after)
    pipe$fp8_resident <- identical(pipe$format, "fp8") && device == "cuda"

    if (verbose) {
        message("Loading VAE decoder...")
    }
    vae_config <- jsonlite::fromJSON(.zimage_cached("vae/config.json"))
    pipe$vae_scaling_factor <- vae_config$scaling_factor %||% 0.3611
    pipe$vae_shift_factor <- vae_config$shift_factor %||% 0.1159
    dec <- vae_decoder_native(
                              latent_channels = as.integer(vae_config$latent_channels %||% 16L)
    )
    load_decoder_safetensors(
                             dec, .zimage_cached("vae/diffusion_pytorch_model.safetensors"),
                             verbose = verbose
    )
    dec$to(device = component_device)
    if (device == "cuda" && phase_offload) {
        # Cast to compute dtype once so the per-generation onload is
        # device-only and the pinned copy is bf16 (see flux2 loader).
        dec$to(dtype = torch::torch_bfloat16())
    }
    dec$eval()
    pipe$decoder <- dec

    sched_config <- jsonlite::fromJSON(.zimage_cached("scheduler/scheduler_config.json"))
    pipe$sched_shift <- sched_config$shift %||% 3.0

    if (verbose) {
        message("Loading Qwen3 text encoder...")
    }
    te_dir <- dirname(.zimage_cached("text_encoder/config.json"))
    te_config <- jsonlite::fromJSON(file.path(te_dir, "config.json"))
    # hidden_states[-2]: the state after num_hidden_layers - 1 layers
    pipe$te_penult_layer <- as.integer(te_config$num_hidden_layers %||% 36L) - 1L
    pipe$text_encoder <- load_qwen3_text_encoder(
        te_dir, device = if (phase_offload) "cpu" else text_device,
        dtype = if (text_device == "cpu") "float32" else "bfloat16",
        verbose = verbose
    )
    pipe$tokenizer <- qwen_bpe_tokenizer(.zimage_cached("tokenizer/tokenizer.json"))

    components <- c("transformer", "decoder")
    if (!identical(text_device, "cpu")) {
        components <- c(components, "text_encoder")
    }
    pipe$staging <- .flux_build_staging(pipe, pin, phase_offload, device,
                                        components, verbose = verbose)

    structure(pipe, class = "zimage_pipeline")
}

# Qwen3 prompt encoding, Z-Image style: thinking-enabled chat template,
# penultimate hidden state, valid tokens only (right padding -> first n)
.zimage_encode_prompt <- function(prompt, model, tokenizer, penult_layer,
                                  max_sequence_length = 512L, device = NULL) {
    enc <- encode_qwen(tokenizer, prompt, max_length = max_sequence_length,
                       chat_template = TRUE, enable_thinking = TRUE)
    device <- device %||% model$model$embed_tokens$weight$device
    long <- torch::torch_long()
    ids <- torch::torch_tensor(enc$input_ids + 1L, dtype = long,
                               device = device)
    mask <- torch::torch_tensor(enc$attention_mask, dtype = long,
                                device = device)

    states <- torch::with_no_grad(model(ids, attention_mask = mask,
                                        out_layers = penult_layer))
    n_real <- sum(enc$attention_mask[1,])
    states[[1]][1, 1:n_real,] # [L, hidden]
}

# Flow-matching Euler loop; Turbo is CFG-free (one forward per step).
# The model sees the reversed normalized timestep and predicts the
# negated velocity.
.zimage_denoise <- function(transformer, latents, schedule, cap_feats,
                            compute_dtype, chunk_size = NULL, verbose = TRUE) {
    timesteps <- as.numeric(schedule$timesteps$cpu())
    n <- length(timesteps)
    pb <- .denoise_progress(n, NULL, verbose)
    f32 <- torch::torch_float32()

    torch::with_no_grad({
        for (i in seq_len(n)) {
            t <- timesteps[[i]]
            t_model <- torch::torch_tensor((1000 - t) / 1000, dtype = f32,
                device = latents$device)$reshape(1L)

            # [1, 16, H8, W8] -> [16, 1, H8, W8] (frame axis)
            x_in <- latents$squeeze(1L)$unsqueeze(2L)$to(dtype = compute_dtype)
            out <- transformer(x_in, t_model, cap_feats,
                               chunk_size = chunk_size)
            noise_pred <- out$squeeze(2L)$unsqueeze(1L)$to(dtype = f32)$neg()

            step <- flowmatch_scheduler_step(noise_pred, t, latents, schedule)
            latents <- step$prev_sample
            schedule <- step$schedule
            rm(out, noise_pred, step)
            pb$tick(i)
        }
    })
    pb$done()
    latents
}

#' Generate an image with Z-Image-Turbo
#'
#' Guidance-distilled text-to-image (8 steps, no CFG): Qwen3-4B prompt
#' encoding (thinking-enabled chat template, penultimate hidden state),
#' FlowMatch denoising with the reversed-timestep convention, and
#' 16-channel VAE decode. Strong at legible text rendering, English and
#' Chinese both.
#'
#' @param prompt Character. The prompt.
#' @param pipeline A \code{zimage_pipeline} from
#'   \code{\link{zimage_load_pipeline}}; NULL loads one (passing
#'   \code{...} through).
#' @param width,height Integers, divisible by 16.
#' @param num_inference_steps Integer. Denoising steps (Turbo: 8).
#' @param max_sequence_length Integer. Qwen3 token length (512).
#' @param seed Integer or NULL. Latents are drawn on the CPU, so a seed
#'   matches a Python diffusers run with a CPU generator.
#' @param prompt_embeds Optional precomputed [L, 2560] caption
#'   embeddings (valid tokens only).
#' @param save_file Logical. Write a PNG.
#' @param filename Output path (default derived from the prompt).
#' @param verbose Logical, or one of "silent", "progress", "steps".
#'   TRUE = "steps" (full per-phase chatter), FALSE = "silent".
#'   "progress" prints a one-line generation summary plus a denoise
#'   progress bar (interactive) or periodic step ticks (captured logs).
#' @param ... Passed to \code{\link{zimage_load_pipeline}} when
#'   \code{pipeline} is NULL.
#'
#' @return Invisibly, \code{list(image, metadata)} where \code{image} is
#'   an [H, W, 3] array in [0, 1].
#'
#' @export
txt2img_zimage <- function(prompt, pipeline = NULL, width = 1024L,
                           height = 1024L, num_inference_steps = 8L,
                           max_sequence_length = 512L, seed = NULL,
                           prompt_embeds = NULL, save_file = TRUE,
                           filename = NULL, verbose = TRUE, ...) {
    level <- .verbosity(verbose)
    verbose <- level == "steps"
    if (level != "silent") {
        message(sprintf("Z-Image Turbo: %dx%d, %d steps", as.integer(width),
                        as.integer(height), as.integer(num_inference_steps)))
    }
    if (is.null(pipeline)) {
        pipeline <- zimage_load_pipeline(..., verbose = verbose)
    }
    device <- pipeline$device
    width <- as.integer(width)
    height <- as.integer(height)
    if (width %% 16L != 0L || height %% 16L != 0L) {
        stop("width and height must be divisible by 16")
    }

    f32 <- torch::torch_float32()
    compute_dtype <- if (device == "cpu") {
        f32
    } else {
        torch::torch_bfloat16()
    }

    phase_offload <- isTRUE(pipeline$phase_offload) && device != "cpu"
    staging <- pipeline$staging %||% list()
    # Components move by name so pinned staging (see staging.R) can carry
    # the CPU<->GPU transfer when the loader prepared it; otherwise the
    # pageable module$to() path runs.
    onload <- function(what) {
        module <- pipeline[[what]]
        if (phase_offload) {
            st <- staging[[what]]
            if (is.null(st)) {
                module$to(device = device)
            } else {
                .staged_onload(st, device)
            }
        }
        module
    }
    offload <- function(what) {
        module <- pipeline[[what]]
        if (phase_offload) {
            st <- staging[[what]]
            if (is.null(st)) {
                module$to(device = "cpu")
            } else {
                .staged_offload(st)
            }
            clear_vram()
        }
        invisible(module)
    }

    t0 <- Sys.time()

    # --- Phase 1: text encoding --------------------------------------------------
    if (is.null(prompt_embeds)) {
        if (verbose) {
            message("Encoding prompt (Qwen3)...")
        }
        # An explicit text_device = "cpu" keeps Qwen3 fp32 on the host
        # (it computes in place); only onload it when it phase-swaps.
        gpu_encode <- !identical(pipeline$text_device, "cpu")
        if (gpu_encode) {
            onload("text_encoder")
        }
        te_device <- pipeline$text_encoder$model$embed_tokens$weight$device
        prompt_embeds <- .zimage_encode_prompt(prompt, pipeline$text_encoder,
            pipeline$tokenizer,
            penult_layer = pipeline$te_penult_layer %||% 35L,
            max_sequence_length = max_sequence_length, device = te_device)
        if (gpu_encode) {
            offload("text_encoder")
        }
    }
    prompt_embeds <- prompt_embeds$to(device = device, dtype = compute_dtype)

    # --- Phase 2: latents and schedule ----------------------------------------------
    h8 <- height %/% 8L
    w8 <- width %/% 8L
    if (!is.null(seed)) {
        torch::torch_manual_seed(seed)
    }
    latents <- torch::torch_randn(c(1L, 16L, h8, w8), dtype = f32)$
    to(device = device)

    n_steps <- as.integer(num_inference_steps)
    sched <- flowmatch_scheduler_create(
                                        shift = pipeline$sched_shift %||% 3.0,
                                        use_dynamic_shifting = FALSE
    )
    sched <- flowmatch_set_timesteps(
                                     sched, n_steps,
                                     sigmas = seq(1, 1 / n_steps, length.out = n_steps)
    )

    # --- Phase 3: denoise ------------------------------------------------------------
    transformer <- onload("transformer")
    # Resident fp8's plain weight fields ride to the GPU with the phase.
    # When staging covers the transformer it already moved them (and the
    # field reassignment here would orphan the staged pairs), so skip it.
    fp8_manual <- isTRUE(pipeline$fp8_resident) &&
    is.null(staging[["transformer"]])
    if (fp8_manual) {
        .flux_fp8_to_device(transformer, device)
    }
    if (verbose) {
        message(sprintf("Denoising: %d steps at %dx%d...", n_steps, width,
                        height))
    }
    latents <- .zimage_denoise(
                               transformer, latents, sched, prompt_embeds,
                               compute_dtype, chunk_size = pipeline$attn_chunk,
                               verbose = level
    )
    if (fp8_manual && phase_offload) {
        .flux_fp8_to_device(pipeline$transformer, "cpu")
    }
    offload("transformer")
    ltx23_release_dequant_buffers()

    # --- Phase 4: decode ---------------------------------------------------------------
    if (verbose) {
        message("Decoding...")
    }
    latents <- latents$div(pipeline$vae_scaling_factor %||% 0.3611)$
    add(pipeline$vae_shift_factor %||% 0.1159)

    decoder <- onload("decoder")
    torch::with_no_grad({
        dec_param <- decoder$conv_in$weight
        img <- decoder(latents$to(device = dec_param$device,
                                  dtype = dec_param$dtype))
        img <- img$to(dtype = f32)$cpu()
    })
    offload("decoder")

    img <- img$squeeze(1)$permute(c(2L, 3L, 1L))
    img <- img$add(1)$div(2)$clamp(0, 1)
    img_array <- as.array(img)

    gen_seconds <- as.numeric(difftime(Sys.time(), t0, units = "secs"))
    if (verbose) {
        message(sprintf("Generated in %.1f s", gen_seconds))
    }

    if (save_file) {
        if (is.null(filename)) {
            filename <- filename_from_prompt(prompt)
        }
        save_image(img_array, filename)
        if (verbose) {
            message("Saved to ", filename)
        }
    }

    metadata <- list(
                     prompt = prompt, width = width, height = height,
                     steps = n_steps, seed = seed, model = "zimage-turbo",
                     precision = pipeline$format, seconds = gen_seconds
    )
    invisible(list(image = img_array, metadata = metadata))
}

# End-to-end smoke test for the Z-Image Turbo pipeline wiring on the CPU
# with tiny random-init components: latents -> denoise (2 steps, static
# shift, reversed timestep + negated output) -> scale/shift -> decode.
# Numeric quality comes from the per-component parity tests; this checks
# the phase plumbing (onload/offload routing, decode).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)

torch::torch_manual_seed(9)

transformer <- zimage_transformer(
  in_channels = 16L, dim = 48L, n_layers = 1L, n_refiner_layers = 1L,
  n_heads = 2L, cap_feat_dim = 24L, axes_dims = c(8L, 8L, 8L)
)
transformer$eval()

decoder <- vae_decoder_native(
  latent_channels = 16L,
  block_channels = c(32L, 32L, 16L, 8L),
  norm_groups = 8L
)
decoder$eval()

pipeline <- structure(
  list(
    transformer = transformer,
    decoder = decoder,
    device = "cpu",
    text_device = "cpu",
    phase_offload = FALSE,
    fp8_resident = FALSE,
    format = "full",
    attn_chunk = NULL,
    config = list(in_channels = 16L),
    sched_shift = 3.0,
    te_penult_layer = 35L,
    vae_scaling_factor = 0.3611,
    vae_shift_factor = 0.1159
  ),
  class = "zimage_pipeline"
)

res <- txt2img_zimage(
  "tiny smoke test",
  pipeline = pipeline,
  width = 64L, height = 64L,
  num_inference_steps = 2L,
  seed = 42L,
  prompt_embeds = torch::torch_randn(7L, 24L),
  save_file = FALSE,
  verbose = FALSE
)

expect_equal(dim(res$image), c(64L, 64L, 3L))
expect_true(all(is.finite(res$image)))
expect_true(all(res$image >= 0) && all(res$image <= 1))
expect_equal(res$metadata$steps, 2L)
expect_equal(res$metadata$model, "zimage-turbo")

#' T5 Text Encoder (T5-v1.1)
#'
#' Fresh R port of the T5 encoder stack from HuggingFace transformers
#' (Apache-2.0, src/transformers/models/t5/modeling_t5.py), as used by
#' FLUX's second text encoder (T5-v1.1-XXL: 24 layers, d_model 4096,
#' 64 heads x d_kv 64, gated-GELU FFN). Distinctives faithfully carried
#' over: RMS layer norms (no mean subtraction), no biases anywhere, no
#' 1/sqrt(d) attention scaling (folded into the weights), and a shared
#' relative position bias computed once from block 1's embedding and
#' added to every layer's attention logits. Module field names mirror
#' the checkpoint keys (minus the \code{encoder.} prefix).
#'
#' FLUX passes no attention mask - padding tokens attend and are
#' attended to - so none is implemented.
#'
#' @name t5_text_encoder
NULL

# Bucketed relative positions (bidirectional): half the buckets split by
# sign, half of those exact small offsets, the rest log-spaced up to
# max_distance. Reference: T5Attention._relative_position_bucket.
.t5_relative_position_bucket <- function(relative_position,
    num_buckets = 32L,
    max_distance = 128L) {
    num_buckets <- num_buckets %/% 2L
    long <- torch::torch_long()
    relative_buckets <- (relative_position > 0)$to(dtype = long)$mul(num_buckets)
    relative_position <- torch::torch_abs(relative_position)

    max_exact <- num_buckets %/% 2L
    is_small <- relative_position < max_exact

    rp_large <- relative_position$to(dtype = torch::torch_float32())$
    div(max_exact)$log()$
    div(log(max_distance / max_exact))$
    mul(num_buckets - max_exact)$
    to(dtype = long)$add(max_exact)
    rp_large <- torch::torch_minimum(
                                     rp_large,
                                     torch::torch_full_like(rp_large, num_buckets - 1L)
    )

    relative_buckets + torch::torch_where(is_small, relative_position, rp_large)
}

# T5 self-attention: no scaling, no biases; the relative position bias
# is added to the logits pre-softmax (softmax in float32)
.t5_attention <- torch::nn_module(
                                  "t5_attention",
                                  initialize = function(d_model, d_kv, num_heads, has_relative_bias = FALSE,
        num_buckets = 32L, max_distance = 128L) {
    inner_dim <- num_heads * d_kv
    self$num_heads <- num_heads
    self$d_kv <- d_kv
    self$num_buckets <- as.integer(num_buckets)
    self$max_distance <- as.integer(max_distance)
    self$q <- torch::nn_linear(d_model, inner_dim, bias = FALSE)
    self$k <- torch::nn_linear(d_model, inner_dim, bias = FALSE)
    self$v <- torch::nn_linear(d_model, inner_dim, bias = FALSE)
    self$o <- torch::nn_linear(inner_dim, d_model, bias = FALSE)
    if (has_relative_bias) {
        self$relative_attention_bias <- torch::nn_embedding(num_buckets,
            num_heads)
    }
},
                                  compute_bias = function(seq_len, device) {
    pos <- torch::torch_arange(start = 0, end = seq_len - 1,
                               dtype = torch::torch_long(), device = device)
    # relative_position[i, j] = j - i
    relative_position <- pos$unsqueeze(1L) - pos$unsqueeze(2L)
    buckets <- .t5_relative_position_bucket(relative_position,
        num_buckets = self$num_buckets,
        max_distance = self$max_distance)
    values <- self$relative_attention_bias(buckets + 1L) # [S, S, H]
    values$permute(c(3L, 1L, 2L))$unsqueeze(1L) # [1, H, S, S]
},
                                  forward = function(x, position_bias) {
    shape <- x$shape
    b <- shape[1]
    s <- shape[2]
    per_head <- c(b, s, self$num_heads, self$d_kv)
    q <- self$q(x)$view(per_head)$transpose(2L, 3L) # [B, H, S, dk]
    k <- self$k(x)$view(per_head)$transpose(2L, 3L)
    v <- self$v(x)$view(per_head)$transpose(2L, 3L)

    scores <- torch::torch_matmul(q, k$transpose(-2L, -1L)) # no 1/sqrt(d)
    scores <- scores + position_bias
    attn <- torch::nnf_softmax(scores$to(dtype = torch::torch_float32()),
                               dim = -1L)$to(dtype = scores$dtype)
    out <- torch::torch_matmul(attn, v)
    out <- out$transpose(2L, 3L)$reshape(c(b, s, -1L))
    self$o(out)
}
)

# layer.0: pre-norm self-attention with residual
.t5_self_attn_layer <- torch::nn_module(
                                        "t5_self_attn_layer",
                                        initialize = function(d_model, d_kv, num_heads, eps,
        has_relative_bias = FALSE, num_buckets = 32L,
        max_distance = 128L) {
    self$SelfAttention <- .t5_attention(d_model, d_kv, num_heads,
                                        has_relative_bias = has_relative_bias,
                                        num_buckets = num_buckets,
                                        max_distance = max_distance)
    self$layer_norm <- ltx23_rms_norm(d_model, eps = eps)
},
                                        forward = function(x, position_bias) {
    x + self$SelfAttention(self$layer_norm(x), position_bias)
}
)

# layer.1: pre-norm gated-GELU feed-forward with residual
.t5_ff_layer <- torch::nn_module(
                                 "t5_ff_layer",
                                 initialize = function(d_model, d_ff, eps) {
    dense <- torch::nn_module(
                              "t5_dense_gated_act_dense",
                              initialize = function(d_model, d_ff) {
        self$wi_0 <- torch::nn_linear(d_model, d_ff, bias = FALSE)
        self$wi_1 <- torch::nn_linear(d_model, d_ff, bias = FALSE)
        self$wo <- torch::nn_linear(d_ff, d_model, bias = FALSE)
    },
                              forward = function(x) {
        h <- torch::nnf_gelu(self$wi_0(x), approximate = "tanh") * self$wi_1(x)
        self$wo(h)
    }
    )
    self$DenseReluDense <- dense(d_model, d_ff)
    self$layer_norm <- ltx23_rms_norm(d_model, eps = eps)
},
                                 forward = function(x) {
    x + self$DenseReluDense(self$layer_norm(x))
}
)

.t5_block <- torch::nn_module(
                              "t5_block",
                              initialize = function(d_model, d_kv, num_heads, d_ff, eps,
        has_relative_bias = FALSE, num_buckets = 32L,
        max_distance = 128L) {
    self$layer <- torch::nn_module_list(list(
            .t5_self_attn_layer(d_model, d_kv, num_heads, eps,
                                has_relative_bias = has_relative_bias,
                                num_buckets = num_buckets,
                                max_distance = max_distance),
            .t5_ff_layer(d_model, d_ff, eps)
        ))
},
                              forward = function(x, position_bias) {
    x <- self$layer[[1]](x, position_bias)
    self$layer[[2]](x)
}
)

#' T5 encoder stack
#'
#' Defaults are the T5-v1.1-XXL configuration used by FLUX.
#'
#' @param vocab_size,d_model,d_kv,num_heads,d_ff,num_layers Integers.
#' @param relative_attention_num_buckets,relative_attention_max_distance
#'   Integers. Relative position bias shape.
#' @param layer_norm_epsilon Numeric.
#'
#' @return Module whose forward(input_ids) (1-based ids [B, S]) returns
#'   the last hidden state [B, S, d_model].
#'
#' @export
t5_encoder <- torch::nn_module(
                               "t5_encoder",
                               initialize = function(vocab_size = 32128L, d_model = 4096L, d_kv = 64L,
        num_heads = 64L, d_ff = 10240L, num_layers = 24L,
        relative_attention_num_buckets = 32L,
        relative_attention_max_distance = 128L,
        layer_norm_epsilon = 1e-6) {
    self$shared <- torch::nn_embedding(vocab_size, d_model)
    self$block <- torch::nn_module_list(
                                        lapply(seq_len(num_layers), function(i) {
        .t5_block(d_model, d_kv, num_heads, d_ff, layer_norm_epsilon,
                  has_relative_bias = (i == 1L),
                  num_buckets = relative_attention_num_buckets,
                  max_distance = relative_attention_max_distance)
    })
    )
    self$final_layer_norm <- ltx23_rms_norm(d_model, eps = layer_norm_epsilon)
},
                               forward = function(input_ids) {
    x <- self$shared(input_ids)
    # Bias comes from block 1 and is shared by every layer
    position_bias <- self$block[[1]]$layer[[1]]$SelfAttention$compute_bias(
        input_ids$shape[2], input_ids$device
    )$to(dtype = x$dtype)
    for (i in seq_along(self$block)) {
        x <- self$block[[i]](x, position_bias)
    }
    self$final_layer_norm(x)
}
)

# Encoder constructor arguments from a transformers config.json
.t5_encoder_args <- function(config) {
    if (is.null(config)) {
        return(list())
    }
    ffp <- config$feed_forward_proj %||% "gated-gelu"
    if (!startsWith(ffp, "gated")) {
        stop("Only gated feed-forward T5 (v1.1) is supported, got: ", ffp)
    }
    args <- list(
                 vocab_size = config$vocab_size,
                 d_model = config$d_model,
                 d_kv = config$d_kv,
                 num_heads = config$num_heads,
                 d_ff = config$d_ff,
                 num_layers = config$num_layers,
                 relative_attention_num_buckets = config$relative_attention_num_buckets,
                 relative_attention_max_distance = config$relative_attention_max_distance
    )
    args <- Filter(function(x) !is.null(x) && length(x) > 0L, args)
    args <- lapply(args, as.integer)
    eps <- config$layer_norm_epsilon
    if (!is.null(eps) && length(eps) == 1L) {
        args$layer_norm_epsilon <- as.numeric(eps)
    }
    args
}

#' Load a T5 encoder from a transformers directory
#'
#' Streams the (possibly sharded) safetensors weights into
#' \code{\link{t5_encoder}}, stripping the \code{encoder.} key prefix
#' and aliasing \code{embed_tokens} to the shared embedding.
#'
#' @param model_path Directory with \code{config.json} and
#'   \code{model*.safetensors} (FLUX.1-schnell's \code{text_encoder_2}).
#' @param device Character. Target device.
#' @param dtype Character. "float32" (CPU default; T5 overflows in
#'   float16) or "bfloat16".
#' @param verbose Logical.
#' @param ... Overrides for \code{\link{t5_encoder}} arguments.
#'
#' @return The loaded \code{t5_encoder} in eval mode.
#'
#' @export
load_t5_text_encoder <- function(model_path, device = "cpu",
                                 dtype = "float32", verbose = TRUE, ...) {
    model_path <- path.expand(model_path)
    config <- NULL
    config_path <- file.path(model_path, "config.json")
    if (file.exists(config_path)) {
        config <- jsonlite::fromJSON(config_path, simplifyVector = TRUE)
    }

    args <- utils::modifyList(.t5_encoder_args(config), list(...))
    model <- do.call(t5_encoder, args)
    model$to(dtype = .flux_dtype(dtype))

    opened <- .flux_open_sharded_dir(model_path, "model")
    ckpt <- structure(
                      list(handle = opened$handle, keys = opened$keys, version = NULL,
                           config = config, path = model_path),
                      class = "ltx23_checkpoint"
    )

    map_key <- function(key) {
        key <- sub("^encoder\\.", "", key)
        if (key == "embed_tokens.weight") {
            key <- "shared.weight"
        }
        key
    }
    res <- ltx23_load_group(ckpt, ckpt$keys, model, map_key = map_key,
                            verbose = verbose)
    if (length(res$unmapped) || length(res$unfilled)) {
        stop("T5 encoder load: ", length(res$unmapped), " unmapped keys, ",
             length(res$unfilled), " unfilled params")
    }

    model$to(device = device)
    model$eval()
    model
}

#' Encode prompts with the T5 encoder
#'
#' Tokenizes with \code{\link{encode_unigram}} (right padding to
#' \code{max_sequence_length}) and runs the encoder. Matching the FLUX
#' reference pipeline, no attention mask is used.
#'
#' @param prompts Character vector.
#' @param model A \code{\link{t5_encoder}}.
#' @param tokenizer A \code{\link{unigram_tokenizer}}.
#' @param max_sequence_length Integer. Fixed token length (schnell: 256).
#' @param device Device for the input ids (defaults to the model's).
#'
#' @return Tensor [length(prompts), max_sequence_length, d_model].
#'
#' @export
encode_with_t5 <- function(prompts, model, tokenizer,
                           max_sequence_length = 256L, device = NULL) {
    enc <- encode_unigram(tokenizer, prompts, max_length = max_sequence_length)
    device <- device %||% model$shared$weight$device
    ids <- torch::torch_tensor(enc$input_ids + 1L,
                               dtype = torch::torch_long(), device = device)
    torch::with_no_grad(model(ids))
}

# Parity tests for the T5 encoder port against HF transformers reference
# fixtures (generated by tools/gen_fixtures_t5.py, checked in).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "t5_flux.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/t5_flux.safetensors"
if (!file.exists(fixture_path)) exit_file("t5 fixtures missing")

ckpt_dir <- system.file("tinytest", "fixtures", "t5_tiny_ckpt",
  package = "diffuseR")
if (ckpt_dir == "") ckpt_dir <- "fixtures/t5_tiny_ckpt"
if (!dir.exists(ckpt_dir)) exit_file("t5 tiny checkpoint missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# --- relative position bucketing: exact integer parity ---------------------------

buckets_r <- diffuseR:::.t5_relative_position_bucket(
  fx$rel_positions$to(dtype = torch::torch_long())
)
expect_equal(
  as.integer(buckets_r$flatten()),
  as.integer(fx$rel_buckets$flatten())
)

# --- tiny encoder parity (padded batch, no mask) -----------------------------------

model <- load_t5_text_encoder(ckpt_dir, device = "cpu", dtype = "float32",
  verbose = FALSE)
out <- torch::with_no_grad(
  model(fx$input_ids$to(dtype = torch::torch_long()) + 1L)
)
expect_equal(as.integer(out$shape), as.integer(fx$out$shape))
expect_true(max_abs_diff(out, fx$out) < 1e-4)

# (encode_with_t5 is exercised end-to-end in the pipeline tests; real
# token ids exceed this fixture's 100-entry vocab.)

#' LTX-2.3 Memory Profiles and CUDA GC Tuning
#'
#' Memory management for running the 22B LTX-2.3 transformer on limited
#' VRAM, built on the patterns proven in the whisper and chatterbox
#' packages: torch allocator GC tuning before the first CUDA op, fp8
#' CPU-resident streaming weights, query-chunked attention, and
#' phase-sequential component placement.
#'
#' @name memory_ltx23
NULL

#' Get an LTX-2.3 memory profile
#'
#' Selects transformer precision, component placement, and attention
#' chunking for the available VRAM. Measured on an RTX 5060 Ti (16 GB):
#' fp8 streaming peaks ~11.6 GB (without phase offloading) at
#' 512x320x49; NF4 keeps the whole 22B transformer resident (~12.5 GB)
#' and removes the ~21 GB/step PCIe weight streaming. The NF4 profile
#' renders 1280x704x121 with audio in ~23 min at a 15.7 GB peak
#' (tiled VAE decode, in-place feed-forward GELU, and the default
#' \code{diffuseR.attn_budget} of 1.5e8 all required at that size).
#'
#' \describe{
#'   \item{precision "nf4"}{Weights resident on the GPU; fastest steps;
#'     about 9 percent weight round-trip error.}
#'   \item{precision "fp8"}{Weights CPU-resident, streamed per linear;
#'     near-bf16 quality; each step pays the PCIe transfer.}
#' }
#'
#' @param vram_gb Numeric or NULL (auto-detect free VRAM).
#'
#' @return Named list with device/dtype placement, \code{attn_chunk},
#'   \code{pin_weights}, and resolution caps.
#'
#' @export
ltx23_memory_profile <- function(vram_gb = NULL) {
    if (is.null(vram_gb)) {
        vram_gb <- .detect_vram(use_free = TRUE)
        message(sprintf("Detected %.1f GB free VRAM", vram_gb))
    }

    profile <- if (vram_gb >= 14) {
        "high"
    } else if (vram_gb >= 10) {
        "medium"
    } else if (vram_gb >= 7) {
        "low"
    } else {
        "cpu_only"
    }

    profiles <- list(
                     high = list(name = "high", device = "cuda", dtype = "bfloat16",
                                 precision = "nf4", phase_offload = TRUE,
                                 pin_weights = FALSE, attn_chunk = NULL,
                                 text_device = "cpu", max_resolution = c(704L, 1280L),
                                 max_frames = 121L),
                     medium = list(
                                   name = "medium",
                                   device = "cuda",
                                   dtype = "bfloat16",
                                   precision = "fp8",
                                   phase_offload = TRUE,
                                   pin_weights = TRUE,
                                   attn_chunk = 4096L,
                                   text_device = "cpu",
                                   max_resolution = c(576L, 1024L),
                                   max_frames = 121L
        ),
                     low = list(
                                name = "low",
                                device = "cuda",
                                dtype = "bfloat16",
                                precision = "fp8",
                                phase_offload = TRUE,
                                pin_weights = TRUE,
                                attn_chunk = 2048L,
                                text_device = "cpu",
                                max_resolution = c(512L, 768L),
                                max_frames = 65L
        ),
                     cpu_only = list(
                                     name = "cpu_only",
                                     device = "cpu",
                                     dtype = "float32",
                                     precision = "fp8",
                                     phase_offload = FALSE,
                                     pin_weights = FALSE,
                                     attn_chunk = NULL,
                                     text_device = "cpu",
                                     max_resolution = c(384L, 640L),
                                     max_frames = 33L
        )
    )

    profiles[[profile]]
}

#' Tune the torch CUDA allocator for large-resident inference
#'
#' Stops the allocator GC storm (cf. ~/skills/torch
#' torch-jit-gc-performance.md): lantern proactively calls R's gc()
#' whenever reserved memory exceeds \code{torch.cuda_allocator_reserved_rate}
#' (default 0.20) of the card. With ~75\% of VRAM occupied by resident
#' weights that fires on nearly every allocation. Raising the rate to the
#' actual footprint is safe here because the LTX hot loops compute into
#' persistent scratch buffers (near-zero per-step garbage). Also raises
#' the host-allocation GC threshold and defaults
#' \code{PYTORCH_CUDA_ALLOC_CONF} to expandable segments. User-set
#' options win.
#'
#' \code{start_torch()} reads the gate options exactly once, so setting
#' them after torch has started is inert on its own. The three CUDA
#' gates are therefore also pushed into the live allocator here (the
#' \code{.flux_gc_gates} pattern), which makes this function effective
#' whenever it runs. The host-side \code{torch.threshold_call_gc} has
#' no live setter; the package defaults it in \code{.onLoad} so torch
#' reads it at init in any session that loads diffuseR before running
#' torch ops.
#'
#' @param footprint_gb Numeric. Expected resident GPU footprint in GB
#'   (NF4 transformer: ~12).
#' @param total_gb Numeric or NULL (auto-detect total VRAM).
#'
#' @return Invisibly, the applied reserved rate (NULL if skipped).
#'
#' @export
ltx23_tune_gc <- function(footprint_gb = 12, total_gb = NULL) {
    if (!nzchar(Sys.getenv("PYTORCH_CUDA_ALLOC_CONF"))) {
        Sys.setenv(PYTORCH_CUDA_ALLOC_CONF = "expandable_segments:True")
    }
    # Options are set before the first torch:: call in this function:
    # torch::cuda_is_available() below starts torch, and start_torch
    # reads these once. (The old order set them one call after init
    # read them - inert by construction.)
    if (is.null(getOption("torch.threshold_call_gc"))) {
        options(torch.threshold_call_gc = 16000)
    }
    # The other two allocator-callback gates (defaults 0.8): measured
    # 32.7s -> 21.5s on the tiled decode under expandable segments, and
    # R-gc share 50% -> 32% on the native backend, with no wall-time
    # downside in any condition
    if (is.null(getOption("torch.cuda_allocator_allocated_rate"))) {
        options(torch.cuda_allocator_allocated_rate = 0.95)
    }
    if (is.null(getOption("torch.cuda_allocator_allocated_reserved_rate"))) {
        options(torch.cuda_allocator_allocated_reserved_rate = 0.95)
    }
    rate <- NULL
    if (is.null(getOption("torch.cuda_allocator_reserved_rate"))) {
        if (is.null(total_gb)) {
            total_gb <- .detect_vram(use_free = FALSE)
        }
        if (isTRUE(total_gb > 0)) {
            rate <- min(0.92, max(0.20, footprint_gb / total_gb))
            options(torch.cuda_allocator_reserved_rate = rate)
        }
    }
    if (!torch::cuda_is_available()) {
        return(invisible(rate))
    }
    # Torch is usually long started by the time a loader calls this:
    # push the CUDA gates into the live allocator directly.
    push <- get0("cpp_set_cuda_allocator_allocator_thresholds",
                 envir = asNamespace("torch"))
    if (is.function(push)) {
        try(push(
                 getOption("torch.cuda_allocator_reserved_rate", 0.2),
                 getOption("torch.cuda_allocator_allocated_rate", 0.8),
                 getOption("torch.cuda_allocator_allocated_reserved_rate", 0.8)
            ), silent = TRUE)
    }
    invisible(rate)
}

# ltx23_tune_gc option semantics: defaults land when unset, user-set
# values win, and the reserved rate clamps to [0.20, 0.92]. The cpp
# gate push itself is CUDA-only and exercised implicitly wherever a
# GPU is present.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)

gate_opts <- c("torch.threshold_call_gc",
               "torch.cuda_allocator_allocated_rate",
               "torch.cuda_allocator_allocated_reserved_rate",
               "torch.cuda_allocator_reserved_rate")
old <- options()[gate_opts]
names(old) <- gate_opts
clear <- function() {
  for (o in gate_opts) {
    opt <- list(NULL)
    names(opt) <- o
    options(opt)
  }
}

clear()
rate <- ltx23_tune_gc(footprint_gb = 12, total_gb = 16)
expect_equal(rate, 0.75)
expect_equal(getOption("torch.cuda_allocator_reserved_rate"), 0.75)
expect_equal(getOption("torch.threshold_call_gc"), 16000)
expect_equal(getOption("torch.cuda_allocator_allocated_rate"), 0.95)
expect_equal(getOption("torch.cuda_allocator_allocated_reserved_rate"), 0.95)

# User-set values win; the reserved rate is then not recomputed
clear()
options(torch.cuda_allocator_reserved_rate = 0.5,
        torch.threshold_call_gc = 4000)
rate <- ltx23_tune_gc(footprint_gb = 12, total_gb = 16)
expect_null(rate)
expect_equal(getOption("torch.cuda_allocator_reserved_rate"), 0.5)
expect_equal(getOption("torch.threshold_call_gc"), 4000)

# Clamps: floor 0.20 for small footprints, ceiling 0.92 for tight fits
clear()
expect_equal(ltx23_tune_gc(footprint_gb = 12, total_gb = 100), 0.20)
clear()
expect_equal(ltx23_tune_gc(footprint_gb = 12, total_gb = 12), 0.92)

options(old)

# post_quant_conv.R

#' Post Quant Conv
#'
#' This function applies a quantized convolution operation to an input tensor.
#' It is typically used in the context of image post  processing, particularly in
#' generative models like Stable Diffusion XL.
#'
#' @param x Input tensor to be processed.
#' @param dtype Data type for the tensor (e.g., "torch_float16" or "torch_float32").
#' @param device Device on which the tensor is located (e.g., "cpu" or "cuda").
#' @return Processed tensor after applying the quantized convolution.
#' @export
post_quant_conv <- function(x, dtype, device) {
    params_path <- system.file("post_quant_conv/", package = "diffuseR")
    qc_weights <- as.matrix(utils::read.csv(paste0(params_path,
                "/post_quant_conv_weights.csv"),
            header = FALSE))
    qc_bias <- as.numeric(utils::read.csv(paste0(params_path,
                "/post_quant_conv_bias.csv"),
            header = FALSE)[[1]])

    # Convert to torch tensors and reshape weights for conv2d
    # Move weights to same device/dtype as input before convolution
    qc_weights_tensor <- torch::torch_tensor(qc_weights)$view(c(4, 4, 1, 1))
    qc_weights_tensor <- qc_weights_tensor$to(dtype = dtype, device = torch::torch_device(device))
    qc_bias_tensor <- torch::torch_tensor(qc_bias)
    qc_bias_tensor <- qc_bias_tensor$to(dtype = dtype, device = torch::torch_device(device))

    torch::nnf_conv2d(x, weight = qc_weights_tensor, bias = qc_bias_tensor)
}

#' FP8 Weight Storage for the LTX-2.3 Transformer
#'
#' GPU-poor weight handling: the large attention/FFN linears of the DiT
#' are stored as float8_e4m3fn with per-tensor scales (the official LTX
#' quantization policy), kept CPU-resident (optionally pinned), and
#' dequantized on the compute device inside each forward. Everything
#' else (norms, embeddings, modulation tables, biases) stays bfloat16.
#' Requires a safetensors build with F8 dtype support.
#'
#' @name fp8_ltx23
NULL

# Official LTX fp8 cast policy: attention and FFN linear weights inside
# the transformer blocks (the dotless ff suffixes also catch audio_ff)
.ltx23_fp8_cast_pattern <- paste0(
                                  "^transformer_blocks\\.[0-9]+\\..*",
                                  "(to_q|to_k|to_v|to_out\\.0|ff\\.net\\.0\\.proj|ff\\.net\\.2)\\.weight$"
)

#' Test whether a mapped DiT key is in the official fp8 cast set
#'
#' @param mapped_key Character vector of mapped (diffusers-style)
#'   parameter names.
#'
#' @return Logical vector.
#'
#' @export
ltx23_is_fp8_cast_key <- function(mapped_key) {
    grepl(.ltx23_fp8_cast_pattern, mapped_key)
}

#' FP8 linear layer
#'
#' Weight lives as float8_e4m3fn plus a float32 scale in plain module
#' fields (so \code{$to(device)} moves only the bias); the forward pass
#' ships 1 byte/param to the input's device, upcasts, rescales, and runs
#' \code{nnf_linear}.
#'
#' @param out_features,in_features Integers.
#' @param bias Logical.
#'
#' @return Module whose forward(x) returns the linear projection of
#'   \code{x}, with the fp8 weight bytes transferred and cast up to the
#'   compute dtype for the matmul. Same result as an
#'   \code{nn_linear} of the same shape, at a quarter of the resident
#'   weight bytes.
#'
#' @export
ltx23_fp8_linear <- torch::nn_module(
                                     "ltx23_fp8_linear",
                                     initialize = function(out_features, in_features, bias = TRUE) {
    self$out_features <- as.integer(out_features)
    self$in_features <- as.integer(in_features)
    self$weight_fp8 <- NULL
    self$weight_scale <- NULL
    if (bias) {
        self$bias <- torch::nn_parameter(torch::torch_zeros(out_features))
    }
},
                                     set_fp8_weight = function(weight, scale, pin = FALSE) {
    weight <- weight$to(device = "cpu")
    if (pin && torch::cuda_is_available()) {
        weight <- .pin_host(weight)
    }
    self$weight_fp8 <- weight
    self$weight_scale <- scale$to(device = "cpu",
                                  dtype = torch::torch_float32())
    invisible(self)
},
                                     forward = function(x) {
    # Transfer fp8 bytes first (no-op when resident), then cast into a
    # persistent per-shape buffer and scale in place: zero fresh
    # allocations per call, same math as the allocating upcast
    w8 <- self$weight_fp8$to(device = x$device, non_blocking = TRUE)
    w <- .ltx23_get_dequant_buffer(
                                   c(self$out_features, self$in_features), x$dtype, x$device
    )
    torch::with_no_grad({
        w$copy_(w8)
        w$mul_(self$weight_scale$to(device = x$device, dtype = x$dtype))
    })
    torch::nnf_linear(x, w, self$bias)
}
)

# Navigate a dotted parameter path to a submodule. Numeric segments
# index module lists (0-based names -> 1-based R); named segments go
# through the nn_module `$` accessor.
.ltx23_walk_module <- function(module, segments) {
    cur <- module
    for (seg in segments) {
        cur <- if (grepl("^[0-9]+$", seg)) {
            cur[[as.integer(seg) + 1L]]
        } else {
            do.call(`$`, list(cur, seg))
        }
        if (is.null(cur)) {
            return(NULL)
        }
    }
    cur
}

#' Quantize an LTX-2.3 checkpoint to FP8 shards
#'
#' Streams the single-file bf16 checkpoint tensor by tensor. DiT
#' attention/FFN linear weights are stored as float8_e4m3fn with a
#' float32 absmax/448 per-tensor scale (\code{<key>_scale} sibling);
#' everything else is copied through unchanged. Output shards carry the
#' original key names plus a manifest for skip-if-exists.
#'
#' @param checkpoint_path Source .safetensors (46 GB bf16 single file).
#' @param output_dir Output directory for shards + manifest; NULL (the
#'   default) resolves under tools::R_user_dir("diffuseR", "data").
#' @param shard_bytes Numeric. Target shard size in bytes. The default
#'   1.9e9 keeps every shard under the 2^31-byte (~2.15 GB) ceiling that
#'   stock CRAN safetensors can read. Pass a larger value (e.g. 4e9) only
#'   for local builds you will read back with a fork-patched safetensors.
#' @param force Logical. Re-quantize even if a valid manifest exists.
#' @param verbose Logical.
#'
#' @return Invisibly, the manifest list.
#'
#' @export
ltx23_quantize_fp8 <- function(checkpoint_path, output_dir = NULL,
                               shard_bytes = 1.9e9, force = FALSE,
                               verbose = TRUE) {
    if (is.null(output_dir)) {
        output_dir <- file.path(tools::R_user_dir("diffuseR", "data"),
                                "ltx2.3-fp8")
    }
    manifest_path <- file.path(output_dir, "manifest.json")
    if (!force && file.exists(manifest_path)) {
        manifest <- jsonlite::fromJSON(manifest_path)
        if (all(file.exists(file.path(output_dir, manifest$shards)))) {
            if (verbose) {
                message("FP8 artifact already present: ", output_dir)
            }
            return(invisible(manifest))
        }
    }
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

    ckpt <- ltx23_open_checkpoint(checkpoint_path)
    fp8 <- torch::torch_float8_e4m3fn()

    shard <- list()
    shard_size <- 0
    shard_files <- character(0)
    n_cast <- 0L

    flush_shard <- function() {
        if (!length(shard)) {
            return()
        }
        fname <- sprintf("ltx2.3-fp8-%05d.safetensors",
                         length(shard_files) + 1L)
        safetensors::safe_save_file(shard, file.path(output_dir, fname))
        shard_files[[length(shard_files) + 1L]] <<- fname
        if (verbose) {
            message(sprintf("  wrote %s (%.2f GB, %d tensors)",
                            fname, shard_size / 1e9, length(shard)))
        }
        shard <<- list()
        shard_size <<- 0
        gc(verbose = FALSE)
    }

    keys <- ckpt$keys
    for (i in seq_along(keys)) {
        key <- keys[[i]]
        tensor <- ckpt$handle$get_tensor(key)

        mapped <- ltx23_map_dit_key(key)
        if (startsWith(key, "model.diffusion_model.") &&
            ltx23_is_fp8_cast_key(mapped)) {
            torch::with_no_grad({
                scale <- tensor$abs()$max()$to(dtype = torch::torch_float32())$
                clamp(min = 1e-12) / 448
                shard[[key]] <- (tensor$to(dtype = torch::torch_float32()) / scale)$to(dtype = fp8)
                shard[[paste0(key, "_scale")]] <- scale
            })
            shard_size <- shard_size + prod(tensor$shape)
            n_cast <- n_cast + 1L
        } else {
            shard[[key]] <- tensor
            shard_size <- shard_size + prod(tensor$shape) * 2
        }
        rm(tensor)

        if (shard_size >= shard_bytes) {
            flush_shard()
        }
        if (i %% 200L == 0L) {
            gc(verbose = FALSE)
            if (verbose) {
                message(sprintf("  quantizing %d/%d tensors", i, length(keys)))
            }
        }
    }
    flush_shard()

    manifest <- list(
                     source = basename(checkpoint_path),
                     model_version = ckpt$version,
                     shards = shard_files,
                     tensors = length(keys),
                     fp8_cast = n_cast,
                     config = ckpt$config
    )
    jsonlite::write_json(manifest, manifest_path, auto_unbox = TRUE, pretty = TRUE)
    if (verbose) {
        message(sprintf("Quantized %d/%d tensors to fp8 across %d shards: %s",
                        n_cast, length(keys), length(shard_files), output_dir))
    }
    invisible(manifest)
}

#' Open an FP8 shard directory as a checkpoint
#'
#' Presents the sharded fp8 artifact through the same interface as
#' \code{\link{ltx23_open_checkpoint}} so the group loaders work
#' unchanged.
#'
#' @param dir The fp8 artifact directory (with manifest.json).
#'
#' @return An \code{ltx23_checkpoint}.
#'
#' @export
ltx23_open_fp8_checkpoint <- function(dir) {
    manifest_path <- file.path(dir, "manifest.json")
    if (!file.exists(manifest_path)) {
        stop("No manifest.json in ", dir, "; run ltx23_quantize_fp8() first.")
    }
    manifest <- jsonlite::fromJSON(manifest_path, simplifyVector = TRUE)

    shard_paths <- file.path(dir, manifest$shards)
    handles <- lapply(shard_paths, function(p) {
        safetensors::safetensors$new(p, framework = "torch")
    })
    key_to_handle <- list()
    key_to_path <- list()
    for (i in seq_along(handles)) {
        h <- handles[[i]]
        for (k in setdiff(h$keys(), "__metadata__")) {
            key_to_handle[[k]] <- h
            key_to_path[[k]] <- shard_paths[[i]]
        }
    }

    handle <- list(
                   get_tensor = function(key) {
        h <- key_to_handle[[key]]
        if (is.null(h)) stop("Key not found in fp8 shards: ", key)
        .st_read_or_breadcrumb(function() h$get_tensor(key), key_to_path[[key]])
    }
    )

    structure(
              list(handle = handle, keys = names(key_to_handle),
                   version = manifest$model_version, config = manifest$config,
                   format = manifest$format %||% "fp8", path = dir),
              class = "ltx23_checkpoint"
    )
}

#' Load the LTX-2.3 transformer with FP8 weights
#'
#' Builds the transformer, swaps the official cast-set linears for
#' \code{\link{ltx23_fp8_linear}}, loads fp8 weights CPU-side (optionally
#' pinned) and everything else as bfloat16 on \code{device}. Sets
#' \code{options(diffuseR.block_gc = TRUE)} so the transformer runs
#' per-block garbage collection over the dequantized temporaries.
#'
#' @param ckpt An fp8 \code{ltx23_checkpoint}
#'   (\code{\link{ltx23_open_fp8_checkpoint}}).
#' @param device Character. Device for the resident (non-fp8) weights.
#' @param pin Logical. Pin the fp8 host memory for faster transfers.
#' @param verbose Logical.
#' @param ... Passed to \code{\link{ltx23_transformer}} (tiny test configs).
#'
#' @return The loaded \code{ltx23_transformer}.
#'
#' @export
ltx23_load_transformer_fp8 <- function(ckpt, device = "cuda", pin = TRUE,
                                       verbose = TRUE, ...) {
    stopifnot(inherits(ckpt, "ltx23_checkpoint"))
    model <- .construct_skeleton(ltx23_transformer, ...)

    groups <- ltx23_split_keys(ckpt$keys)
    dit_keys <- groups$dit
    scale_keys <- dit_keys[endsWith(dit_keys, ".weight_scale")]
    main_keys <- setdiff(dit_keys, scale_keys)

    dests <- c(model$named_parameters(), model$named_buffers())
    filled <- character(0)
    unmapped <- character(0)

    torch::with_no_grad({
        for (i in seq_along(main_keys)) {
            key <- main_keys[[i]]
            mapped <- ltx23_map_dit_key(key)

            if (ltx23_is_fp8_cast_key(mapped) &&
                        paste0(key, "_scale") %in% scale_keys) {
                segments <- strsplit(mapped, ".", fixed = TRUE)[[1]]
                parent <- .ltx23_walk_module(model, utils::head(segments, -2L))
                leaf <- segments[length(segments) - 1L]
                old <- .ltx23_walk_module(parent, leaf)
                if (is.null(old)) {
                    unmapped <- c(unmapped, key)
                    next
                }
                weight <- ckpt$handle$get_tensor(key)
                scale <- ckpt$handle$get_tensor(paste0(key, "_scale"))
                fp8_mod <- ltx23_fp8_linear(weight$shape[1], weight$shape[2],
                    bias = !is.null(old$bias))
                if (!is.null(old$bias)) {
                    # Adopt the original bias parameter so the separate bias key,
                    # which copies through the pre-swap destination map, lands here
                    fp8_mod$bias <- old$bias
                }
                fp8_mod$set_fp8_weight(weight, scale, pin = pin)
                do.call(`$<-`, list(parent, leaf, fp8_mod))
                filled <- c(filled, mapped)
                rm(weight, scale)
            } else {
                dest <- dests[[mapped]]
                if (is.null(dest)) {
                    unmapped <- c(unmapped, key)
                    next
                }
                dest$copy_(ckpt$handle$get_tensor(key))
                filled <- c(filled, mapped)
            }

            if (i %% 100L == 0L) {
                gc(verbose = FALSE)
                if (verbose && i %% 500L == 0L) {
                    message(sprintf("  loaded %d/%d transformer tensors", i, length(main_keys)))
                }
            }
        }
    })
    gc(verbose = FALSE)

    if (length(unmapped)) {
        stop("FP8 transformer load: ", length(unmapped), " unmapped keys, e.g. ",
             paste(utils::head(unmapped, 3), collapse = ", "))
    }

    # Weight params replaced by fp8 modules won't be "filled"; account for them
    expected_missing <- grepl(.ltx23_fp8_cast_pattern, names(dests))
    unfilled <- setdiff(names(dests)[!expected_missing], filled)
    if (length(unfilled)) {
        stop("FP8 transformer load: ", length(unfilled), " unfilled params, e.g. ",
             paste(utils::head(unfilled, 3), collapse = ", "))
    }

    # Residents (norms, embeddings, tables, biases) to the compute device;
    # fp8 fields stay on the CPU because they are plain fields
    model$to(device = device)
    model$eval()
    # fp8 dequant allocates a full weight per linear; per-block gc
    # keeps those temporaries from accumulating
    options(diffuseR.block_gc = TRUE)
    if (verbose) {
        message("Transformer ready: fp8 weights CPU-resident, rest on ", device)
    }
    model
}

#' Set the attention query-chunk size across a transformer
#'
#' R torch has no fused attention, so the [B, H, S, S] matrix
#' materializes; chunking queries bounds the peak. NULL disables
#' chunking.
#'
#' @param transformer An \code{ltx23_transformer}.
#' @param chunk Integer or NULL.
#'
#' @return Invisibly, the transformer.
#'
#' @export
ltx23_set_attn_chunk <- function(transformer, chunk) {
    for (i in seq_along(transformer$transformer_blocks)) {
        block <- transformer$transformer_blocks[[i]]
        for (name in c("attn1", "audio_attn1", "attn2", "audio_attn2",
                       "audio_to_video_attn", "video_to_audio_attn")) {
            block[[name]]$attn_chunk <- chunk
        }
    }
    invisible(transformer)
}

# FP8 quantization round trip: quantize a tiny official-named checkpoint,
# reload with fp8 linears, and compare against the bf16 original.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}
# fp8 needs the F8-capable safetensors build
# Same gate the quantizer uses, so the diffuseR.st_caps override
# reaches this guard too; a private probe would disagree with it.
f8_ok <- diffuseR:::.st_can_write("float8_e4m3fn")
if (!f8_ok) exit_file("safetensors build lacks F8 support")

library(diffuseR)

torch::torch_manual_seed(123)

tiny_cfg <- list(
  in_channels = 4L, out_channels = 4L,
  num_attention_heads = 2L, attention_head_dim = 8L,
  cross_attention_dim = 16L,
  audio_in_channels = 4L, audio_out_channels = 4L,
  audio_num_attention_heads = 2L, audio_attention_head_dim = 4L,
  audio_cross_attention_dim = 8L,
  num_layers = 2L
)
ref_model <- do.call(ltx23_transformer, tiny_cfg)
ref_model$eval()

# Reverse the diffusers renames to synthesize an official-named checkpoint
to_official <- function(name) {
  name <- gsub("av_cross_attn_video_scale_shift", "av_ca_video_scale_shift_adaln_single", name, fixed = TRUE)
  name <- gsub("av_cross_attn_video_a2v_gate", "av_ca_a2v_gate_adaln_single", name, fixed = TRUE)
  name <- gsub("av_cross_attn_audio_scale_shift", "av_ca_audio_scale_shift_adaln_single", name, fixed = TRUE)
  name <- gsub("av_cross_attn_audio_v2a_gate", "av_ca_v2a_gate_adaln_single", name, fixed = TRUE)
  name <- gsub("video_a2v_cross_attn_scale_shift_table", "scale_shift_table_a2v_ca_video", name, fixed = TRUE)
  name <- gsub("audio_a2v_cross_attn_scale_shift_table", "scale_shift_table_a2v_ca_audio", name, fixed = TRUE)
  name <- gsub("prompt_adaln", "prompt_adaln_single", name, fixed = TRUE)
  name <- sub("^audio_time_embed\\.", "audio_adaln_single.", name)
  name <- sub("^time_embed\\.", "adaln_single.", name)
  name <- gsub("proj_in", "patchify_proj", name, fixed = TRUE)
  name <- gsub("norm_q", "q_norm", name, fixed = TRUE)
  name <- gsub("norm_k", "k_norm", name, fixed = TRUE)
  paste0("model.diffusion_model.", name)
}

params <- ref_model$named_parameters()
tensors <- list()
for (name in names(params)) {
  tensors[[to_official(name)]] <- params[[name]]$detach()
}
# Round trip through the mapper must be lossless
expect_equal(
  sort(vapply(names(tensors), ltx23_map_dit_key, character(1), USE.NAMES = FALSE)),
  sort(names(params))
)

src <- tempfile(fileext = ".safetensors")
on.exit(unlink(src), add = TRUE)
safetensors::safe_save_file(tensors, src, metadata = list(model_version = "2.3.0"))

# --- Quantize -----------------------------------------------------------------

fp8_dir <- tempfile("fp8_")
on.exit(unlink(fp8_dir, recursive = TRUE), add = TRUE)
manifest <- ltx23_quantize_fp8(src, fp8_dir, verbose = FALSE)
expect_true(file.exists(file.path(fp8_dir, "manifest.json")))
expect_true(manifest$fp8_cast > 0)
# 2 layers x (3 video self + 3 audio self + 3 x-attn x 2 + 4 a2v/v2a) qkv/out
# + 4 ff projections: just assert the exact official policy count
n_expected <- sum(ltx23_is_fp8_cast_key(names(params)))
expect_equal(manifest$fp8_cast, n_expected)

# Re-running is a no-op (skip-if-exists)
m2 <- ltx23_quantize_fp8(src, fp8_dir, verbose = FALSE)
expect_equal(m2$shards, manifest$shards)

# --- Load fp8 + forward parity ---------------------------------------------------

ckpt <- ltx23_open_fp8_checkpoint(fp8_dir)
expect_equal(ckpt$version, "2.3.0")

old_opt <- getOption("diffuseR.use_fp8")
fp8_model <- do.call(ltx23_load_transformer_fp8, c(
  list(ckpt = ckpt, device = "cpu", pin = FALSE, verbose = FALSE),
  tiny_cfg
))
options(diffuseR.use_fp8 = old_opt)

# Cast linears were swapped
expect_inherits(fp8_model$transformer_blocks[[1]]$attn1$to_q, "ltx23_fp8_linear")
expect_equal(
  fp8_model$transformer_blocks[[1]]$attn1$to_q$weight_fp8$dtype$.type(),
  "Float8_e4m3fn"
)
# Non-cast params match exactly (bf16 storage of an fp32 original)
expect_true(as.numeric(torch::torch_max(torch::torch_abs(
  fp8_model$proj_in$weight$to(dtype = torch::torch_float32()) -
    ref_model$proj_in$weight
))) < 0.01)

# Dequantized weights reconstruct the originals within fp8 e4m3 tolerance
# (a random tiny network amplifies chaotically, so end-to-end output error
# is not a meaningful metric; real-weight quality shows in the E2E render)
fp8_q <- fp8_model$transformer_blocks[[1]]$attn1$to_q
deq <- fp8_q$weight_fp8$to(dtype = torch::torch_float32()) *
  fp8_q$weight_scale
orig <- ref_model$transformer_blocks[[1]]$attn1$to_q$weight
rel_w <- as.numeric((deq - orig)$abs()$mean() / orig$abs()$mean())
expect_true(rel_w < 0.05)

# The fp8_linear layer itself matches a plain linear closely
x_probe <- torch::torch_randn(3L, 16L)
torch::with_no_grad({
  y_fp8 <- fp8_q(x_probe$to(dtype = torch::torch_bfloat16()))
  y_ref <- torch::nnf_linear(x_probe, orig,
    ref_model$transformer_blocks[[1]]$attn1$to_q$bias)
})
rel_y <- as.numeric(
  (y_fp8$to(dtype = torch::torch_float32()) - y_ref)$abs()$mean() /
    y_ref$abs()$mean()
)
expect_true(rel_y < 0.1)

# Full fp8 model forward runs in bf16 and stays finite
run <- function(model, dtype) {
  torch::torch_manual_seed(5)
  hidden <- torch::torch_randn(1L, 24L, 4L, dtype = dtype)
  audio <- torch::torch_randn(1L, 5L, 4L, dtype = dtype)
  enc <- torch::torch_randn(1L, 7L, 16L, dtype = dtype)
  aenc <- torch::torch_randn(1L, 7L, 8L, dtype = dtype)
  t <- torch::torch_tensor(700, dtype = torch::torch_float32())
  torch::with_no_grad({
    model(
      hidden_states = hidden, audio_hidden_states = audio,
      encoder_hidden_states = enc, audio_encoder_hidden_states = aenc,
      timestep = t, sigma = t,
      num_frames = 2L, height = 3L, width = 4L, audio_num_frames = 5L,
      use_cross_timestep = TRUE
    )
  })
}
fp8_out <- run(fp8_model, torch::torch_bfloat16())
expect_equal(as.integer(fp8_out$sample$shape), c(1L, 24L, 4L))
expect_true(as.logical(fp8_out$sample$isfinite()$all()$item()))
expect_true(as.logical(fp8_out$audio_sample$isfinite()$all()$item()))

# --- attn chunking + memory profile -------------------------------------------------

ltx23_set_attn_chunk(fp8_model, 8L)
expect_equal(fp8_model$transformer_blocks[[1]]$attn2$attn_chunk, 8L)
ltx23_set_attn_chunk(fp8_model, NULL)
expect_null(fp8_model$transformer_blocks[[1]]$attn2$attn_chunk)

prof <- ltx23_memory_profile(vram_gb = 16)
expect_equal(prof$name, "high")
expect_true(is.integer(prof$attn_chunk) || is.null(prof$attn_chunk))
expect_equal(prof$precision, "nf4")
expect_equal(ltx23_memory_profile(vram_gb = 4)$name, "cpu_only")

#' LTX-2.3 JIT Block Stack
#'
#' TorchScript compilation of the 48-block NF4 transformer step (cf.
#' the torch skill's JIT-decode pattern proven in whisper and
#' chatterbox). Eager execution crosses R -> lantern per op (~190 us
#' each) and leaves every intermediate as an R tensor handle that only
#' dies at gc(); at high resolution that forces a per-block gc()
#' costing the vast majority of step time. Compiled, the whole block
#' stack is one crossing: intermediates are freed eagerly by libtorch,
#' no R garbage accumulates, no per-block gc is needed, and attention
#' runs through the fused \code{scaled_dot_product_attention} kernel
#' instead of a materialized score matrix.
#'
#' Weights are passed per call as a flat \code{List[Tensor]} (borrowed
#' by reference, no copies) with a fixed per-block layout; the packer
#' and the TorchScript indices must stay in lockstep (parity-tested).
#'
#' @name jit_ltx23
NULL

# Per-attention tensor layout (16 slots):
#   0 gate_w, 1 gate_b,
#   2 q_packed, 3 q_absmax, 4 q_bias,
#   5 k_packed, 6 k_absmax, 7 k_bias,
#   8 v_packed, 9 v_absmax, 10 v_bias,
#   11 out_packed, 12 out_absmax, 13 out_bias,
#   14 norm_q_w, 15 norm_k_w
# Per-block layout (114 slots): attn1 @0, audio_attn1 @16, attn2 @32,
# audio_attn2 @48, audio_to_video_attn @64, video_to_audio_attn @80,
# ff @96 (proj_packed, proj_absmax, proj_bias, net2_packed,
# net2_absmax, net2_bias), audio_ff @102, then the modulation tables:
# 108 scale_shift_table, 109 audio_scale_shift_table,
# 110 prompt_scale_shift_table, 111 audio_prompt_scale_shift_table,
# 112 video_a2v_cross_attn_scale_shift_table,
# 113 audio_a2v_cross_attn_scale_shift_table.
.ltx23_jit_slots_per_block <- 114L

.ltx23_jit_source <- function() {
    "
def nf4_lin(x: Tensor, packed: Tensor, absmax: Tensor, bias: Tensor, table2: Tensor) -> Tensor:
    # table2 is the [256, 2] byte LUT (hi nibble, lo nibble) in the
    # compute dtype: one embedding gather replaces the shift/and/stack
    # int64-index chain, whose intermediates cost ~70 bytes of traffic
    # per half-byte weight and made dequant the per-step wall. absmax
    # is cast down before the scale so the product never promotes to
    # float32 (which would materialize a full-precision weight copy).
    rows = bias.size(0)
    vals = torch.embedding(table2, packed.int())
    w = vals.reshape([-1, 64]) * absmax.type_as(table2).unsqueeze(1)
    return torch.linear(x, w.reshape([rows, -1]).type_as(x), bias)

def rmsn(x: Tensor) -> Tensor:
    v = x.float().pow(2).mean(-1, keepdim=True)
    return (x * torch.rsqrt(v + 1e-6)).type_as(x)

def rmsn_w(x: Tensor, w: Tensor) -> Tensor:
    v = x.float().pow(2).mean(-1, keepdim=True)
    return (x * torch.rsqrt(v + 1e-6)).type_as(w) * w

def rope(x: Tensor, cs: Tensor, sn: Tensor) -> Tensor:
    if cs.dim() == 4:
        # Per-head split layout [B, H, T, r]: x [B, T, H*D] -> [B, H, T, D]
        b = cs.size(0)
        hh = cs.size(1)
        t = cs.size(2)
        xh = x.reshape([b, t, hh, -1]).transpose(1, 2)
        r = xh.size(-1) // 2
        xf = xh.narrow(-1, 0, r).float()
        xs = xh.narrow(-1, r, r).float()
        o = torch.cat([xf * cs - xs * sn, xs * cs + xf * sn], -1)
        return o.transpose(1, 2).reshape([b, t, -1]).type_as(x)
    r2 = x.size(-1) // 2
    xf2 = x.narrow(-1, 0, r2).float()
    xs2 = x.narrow(-1, r2, r2).float()
    return torch.cat([xf2 * cs - xs2 * sn, xs2 * cs + xf2 * sn], -1).type_as(x)

def mods(tbl: Tensor, temb: Tensor, num: int) -> Tensor:
    b = temb.size(0)
    t = temb.size(1)
    return tbl.unsqueeze(0).unsqueeze(0) + temb.reshape([b, t, num, -1])

def modtok(vada: Tensor, j: int, idx: Optional[Tensor]) -> Tensor:
    # One modulation vector [B, T, D]. Without idx, T is 1 (global) or
    # S (per-token) and broadcasts/aligns directly. With idx (prefix
    # conditioning), T holds the distinct timestep variants and idx
    # [S] selects per token -- materializing one [B, S, D] slice on
    # demand instead of a [B, S, num, D] block.
    v = vada.select(2, j)
    if idx is None:
        return v
    return v.index_select(1, idx)

def attn_nf4(x: Tensor, ctx: Tensor, ws: List[Tensor], base: int, table2: Tensor, heads: int,
             q_cos: Optional[Tensor], q_sin: Optional[Tensor],
             k_cos: Optional[Tensor], k_sin: Optional[Tensor],
             mask: Optional[Tensor]) -> Tensor:
    gl = torch.linear(x, ws[base], ws[base + 1])
    q = rmsn_w(nf4_lin(x, ws[base + 2], ws[base + 3], ws[base + 4], table2), ws[base + 14])
    k = rmsn_w(nf4_lin(ctx, ws[base + 5], ws[base + 6], ws[base + 7], table2), ws[base + 15])
    v = nf4_lin(ctx, ws[base + 8], ws[base + 9], ws[base + 10], table2)
    if q_cos is not None and q_sin is not None:
        q = rope(q, q_cos, q_sin)
    if k_cos is not None and k_sin is not None:
        k = rope(k, k_cos, k_sin)
    qh = q.unflatten(-1, [heads, -1]).transpose(1, 2)
    kh = k.unflatten(-1, [heads, -1]).transpose(1, 2)
    vh = v.unflatten(-1, [heads, -1]).transpose(1, 2)
    o = torch.scaled_dot_product_attention(qh, kh, vh, attn_mask=mask)
    o = o.transpose(1, 2).flatten(2).type_as(x)
    gates = torch.sigmoid(gl) * 2.0
    o = (o.unflatten(-1, [heads, -1]) * gates.unsqueeze(-1)).flatten(2)
    return nf4_lin(o, ws[base + 11], ws[base + 12], ws[base + 13], table2)

def ff_nf4(x: Tensor, ws: List[Tensor], base: int, table2: Tensor) -> Tensor:
    h = torch.gelu(nf4_lin(x, ws[base], ws[base + 1], ws[base + 2], table2), approximate=\"tanh\")
    return nf4_lin(h, ws[base + 3], ws[base + 4], ws[base + 5], table2)

def block_nf4(h: Tensor, ah: Tensor, enc: Tensor, aenc: Tensor,
              temb: Tensor, temb_a: Tensor,
              tcss: Tensor, tcass: Tensor, tcg: Tensor, tcag: Tensor,
              tp: Tensor, tpa: Tensor,
              v_cos: Tensor, v_sin: Tensor, a_cos: Tensor, a_sin: Tensor,
              cav_cos: Tensor, cav_sin: Tensor, caa_cos: Tensor, caa_sin: Tensor,
              enc_mask: Optional[Tensor], aenc_mask: Optional[Tensor],
              cond_idx: Optional[Tensor],
              ws: List[Tensor], base: int, table2: Tensor,
              heads: int, aheads: int) -> Tuple[Tensor, Tensor]:
    vada = mods(ws[base + 108], temb, 9)
    aada = mods(ws[base + 109], temb_a, 9)

    nh = rmsn(h) * (modtok(vada, 1, cond_idx) + 1.0) + modtok(vada, 0, cond_idx)
    ax = attn_nf4(nh, nh, ws, base, table2, heads, v_cos, v_sin, v_cos, v_sin, None)
    h = h + ax * modtok(vada, 2, cond_idx)

    nah = rmsn(ah) * (aada.select(2, 1) + 1.0) + aada.select(2, 0)
    aax = attn_nf4(nah, nah, ws, base + 16, table2, aheads, a_cos, a_sin, a_cos, a_sin, None)
    ah = ah + aax * aada.select(2, 2)

    pada = mods(ws[base + 110], tp, 2)
    apada = mods(ws[base + 111], tpa, 2)

    nh = rmsn(h) * (modtok(vada, 7, cond_idx) + 1.0) + modtok(vada, 6, cond_idx)
    encm = enc * (pada.select(2, 1) + 1.0) + pada.select(2, 0)
    ax = attn_nf4(nh, encm, ws, base + 32, table2, heads, None, None, None, None, enc_mask)
    h = h + ax * modtok(vada, 8, cond_idx)

    nah = rmsn(ah) * (aada.select(2, 7) + 1.0) + aada.select(2, 6)
    aencm = aenc * (apada.select(2, 1) + 1.0) + apada.select(2, 0)
    aax = attn_nf4(nah, aencm, ws, base + 48, table2, aheads, None, None, None, None, aenc_mask)
    ah = ah + aax * aada.select(2, 8)

    nh = rmsn(h)
    nah = rmsn(ah)
    vca = mods(ws[base + 112].narrow(0, 0, 4), tcss, 4)
    vcg = mods(ws[base + 112].narrow(0, 4, 1), tcg, 1)
    aca = mods(ws[base + 113].narrow(0, 0, 4), tcass, 4)
    acg = mods(ws[base + 113].narrow(0, 4, 1), tcag, 1)

    mnh = nh * (vca.select(2, 0) + 1.0) + vca.select(2, 1)
    mna = nah * (aca.select(2, 0) + 1.0) + aca.select(2, 1)
    a2v = attn_nf4(mnh, mna, ws, base + 64, table2, aheads, cav_cos, cav_sin, caa_cos, caa_sin, None)
    h = h + vcg.select(2, 0) * a2v

    mnh = nh * (vca.select(2, 2) + 1.0) + vca.select(2, 3)
    mna = nah * (aca.select(2, 2) + 1.0) + aca.select(2, 3)
    v2a = attn_nf4(mna, mnh, ws, base + 80, table2, aheads, caa_cos, caa_sin, cav_cos, cav_sin, None)
    ah = ah + acg.select(2, 0) * v2a

    nh = rmsn(h) * (modtok(vada, 4, cond_idx) + 1.0) + modtok(vada, 3, cond_idx)
    h = h + ff_nf4(nh, ws, base + 96, table2) * modtok(vada, 5, cond_idx)

    nah = rmsn(ah) * (aada.select(2, 4) + 1.0) + aada.select(2, 3)
    ah = ah + ff_nf4(nah, ws, base + 102, table2) * aada.select(2, 5)

    return (h, ah)

def stack_nf4(h: Tensor, ah: Tensor, enc: Tensor, aenc: Tensor,
              temb: Tensor, temb_a: Tensor,
              tcss: Tensor, tcass: Tensor, tcg: Tensor, tcag: Tensor,
              tp: Tensor, tpa: Tensor,
              v_cos: Tensor, v_sin: Tensor, a_cos: Tensor, a_sin: Tensor,
              cav_cos: Tensor, cav_sin: Tensor, caa_cos: Tensor, caa_sin: Tensor,
              enc_mask: Optional[Tensor], aenc_mask: Optional[Tensor],
              cond_idx: Optional[Tensor],
              ws: List[Tensor], table2: Tensor,
              n_blocks: int, heads: int, aheads: int) -> Tuple[Tensor, Tensor]:
    i = 0
    while i < n_blocks:
        h, ah = block_nf4(h, ah, enc, aenc, temb, temb_a, tcss, tcass, tcg, tcag,
                          tp, tpa, v_cos, v_sin, a_cos, a_sin,
                          cav_cos, cav_sin, caa_cos, caa_sin,
                          enc_mask, aenc_mask, cond_idx, ws, i * 114, table2, heads, aheads)
        i += 1
    return (h, ah)
"
}

# Compile once per session
.ltx23_jit_env <- new.env(parent = emptyenv())

.ltx23_jit_unit <- function() {
    unit <- .ltx23_jit_env$unit
    if (is.null(unit)) {
        unit <- torch::jit_compile(.ltx23_jit_source())
        .ltx23_jit_env$unit <- unit
    }
    unit
}

# Flat List[Tensor] for one attention module (16 slots)
.ltx23_jit_pack_attn <- function(attn) {
    list(attn$to_gate_logits$weight, attn$to_gate_logits$bias,
         attn$to_q$weight_nf4, attn$to_q$weight_absmax, attn$to_q$bias,
         attn$to_k$weight_nf4, attn$to_k$weight_absmax, attn$to_k$bias,
         attn$to_v$weight_nf4, attn$to_v$weight_absmax, attn$to_v$bias,
         attn$to_out[[1]]$weight_nf4, attn$to_out[[1]]$weight_absmax,
         attn$to_out[[1]]$bias, attn$norm_q$weight, attn$norm_k$weight)
}

.ltx23_jit_pack_ff <- function(ff) {
    list(ff$net[[1]]$proj$weight_nf4, ff$net[[1]]$proj$weight_absmax,
         ff$net[[1]]$proj$bias, ff$net[[3]]$weight_nf4,
         ff$net[[3]]$weight_absmax, ff$net[[3]]$bias)
}

#' Pack a transformer block's weights for the JIT stack
#'
#' Returns the block's tensors in the fixed 114-slot layout consumed by
#' the compiled \code{stack_nf4}/\code{block_nf4} TorchScript functions.
#' Tensor handles are borrowed, not copied.
#'
#' @param block An NF4-quantized \code{ltx23_transformer_block}.
#'
#' @return List of 114 tensors.
#'
#' @keywords internal
.ltx23_jit_pack_block <- function(block) {
    c(
        .ltx23_jit_pack_attn(block$attn1),
        .ltx23_jit_pack_attn(block$audio_attn1),
        .ltx23_jit_pack_attn(block$attn2),
        .ltx23_jit_pack_attn(block$audio_attn2),
        .ltx23_jit_pack_attn(block$audio_to_video_attn),
        .ltx23_jit_pack_attn(block$video_to_audio_attn),
        .ltx23_jit_pack_ff(block$ff),
        .ltx23_jit_pack_ff(block$audio_ff),
        list(block$scale_shift_table, block$audio_scale_shift_table,
             block$prompt_scale_shift_table,
             block$audio_prompt_scale_shift_table,
             block$video_a2v_cross_attn_scale_shift_table,
             block$audio_a2v_cross_attn_scale_shift_table)
    )
}

# A block is JIT-eligible when every cast linear is NF4 (loader swap
# applied) and the gated/adaln 2.3 features the script bakes in are on
.ltx23_jit_block_ok <- function(block) {
    inherits(block$attn1$to_q, "ltx23_nf4_linear") &&
    inherits(block$ff$net[[3]], "ltx23_nf4_linear") &&
    isTRUE(block$attn1$apply_gated_attention) &&
    isTRUE(block$video_cross_attn_adaln) &&
    isTRUE(block$audio_cross_attn_adaln)
}

# Byte-level NF4 dequant LUT [256, 2] (hi nibble, lo nibble), cached
# per device+dtype. Built in the compute dtype so the in-graph scale
# multiply never promotes: float32 (CPU, tests) reproduces the old
# fp32 dequant exactly; bf16 (GPU) accepts ~0.4% weight rounding (far
# below NF4 quantization noise) for half the dequant traffic.
.ltx23_jit_table <- function(device, dtype) {
    key <- paste(device$type, device$index %||% 0L, dtype$.type(), sep = "|")
    tbl <- .ltx23_jit_env[[key]]
    if (is.null(tbl)) {
        byte <- 0:255
        pairs <- cbind(.ltx23_nf4_table[byte %/% 16L + 1L],
                       .ltx23_nf4_table[byte %% 16L + 1L])
        tbl <- torch::torch_tensor(pairs, dtype = dtype, device = device)
        .ltx23_jit_env[[key]] <- tbl
    }
    tbl
}

#' Run the block stack through the compiled TorchScript path
#'
#' One R-to-libtorch crossing for all blocks: no per-op dispatch, no R
#' tensor garbage, fused SDPA. Masks must already be additive
#' \code{[B, 1, 1, S]} (or NULL); rope tensors are the \code{[.., r]}
#' cos/sin pairs used by the eager path.
#'
#' @return list(hidden_states, audio_hidden_states)
#'
#' @keywords internal
.ltx23_jit_run_stack <- function(blocks, hidden_states, audio_hidden_states,
                                 encoder_hidden_states,
                                 audio_encoder_hidden_states, temb,
                                 temb_audio, temb_ca_scale_shift,
                                 temb_ca_audio_scale_shift, temb_ca_gate,
                                 temb_ca_audio_gate, temb_prompt,
                                 temb_prompt_audio, video_rotary_emb,
                                 audio_rotary_emb, ca_video_rotary_emb,
                                 ca_audio_rotary_emb,
                                 encoder_attention_mask = NULL,
                                 audio_encoder_attention_mask = NULL,
                                 cond_token_index = NULL) {
    unit <- .ltx23_jit_unit()
    # unname: an nn_module_list yields named children, and a named R
    # list marshals to TorchScript as Dict[str, Tensor], not List[Tensor]
    ws <- unname(do.call(c, lapply(blocks, .ltx23_jit_pack_block)))
    table <- .ltx23_jit_table(hidden_states$device, hidden_states$dtype)
    heads <- blocks[[1]]$attn1$heads
    aheads <- blocks[[1]]$audio_attn1$heads

    # Eager attention takes [B, 1, S] additive masks and unsqueezes to
    # [B, 1, 1, S]; SDPA wants them pre-broadcast
    fix_mask <- function(m) {
        if (!is.null(m) && m$ndim == 3L) {
            m$unsqueeze(2L)
        } else {
            m
        }
    }

    res <- unit$stack_nf4(hidden_states, audio_hidden_states,
                          encoder_hidden_states, audio_encoder_hidden_states,
                          temb, temb_audio, temb_ca_scale_shift,
                          temb_ca_audio_scale_shift, temb_ca_gate,
                          temb_ca_audio_gate, temb_prompt, temb_prompt_audio,
                          video_rotary_emb[[1]], video_rotary_emb[[2]],
                          audio_rotary_emb[[1]], audio_rotary_emb[[2]],
                          ca_video_rotary_emb[[1]], ca_video_rotary_emb[[2]],
                          ca_audio_rotary_emb[[1]], ca_audio_rotary_emb[[2]],
                          fix_mask(encoder_attention_mask),
                          fix_mask(audio_encoder_attention_mask),
                          cond_token_index, ws, table,
                          torch::jit_scalar(length(blocks)),
                          torch::jit_scalar(as.integer(heads)),
                          torch::jit_scalar(as.integer(aheads)))
    list(res[[1]], res[[2]])
}

# Parity of the TorchScript block stack (R/jit_ltx23.R) against the
# eager NF4 transformer block. The compiled path must reproduce the
# eager path bit-for-bit up to SDPA-vs-materialized-attention rounding.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)
torch::torch_manual_seed(7)

# The prompt KV modulation tables are dim-sized and apply directly to
# the text streams, so cross_attention_dim == dim (as in the real
# model: 4096/4096 video, 2048/2048 audio)
dim <- 32L; heads <- 2L; head_dim <- 16L; cross_dim <- 32L
adim <- 16L; aheads <- 2L; ahead_dim <- 8L; across_dim <- 16L
B <- 2L; Sv <- 6L; Sa <- 5L; St <- 4L

make_block <- function() {
  blk <- ltx23_transformer_block(
    dim = dim, num_attention_heads = heads, attention_head_dim = head_dim,
    cross_attention_dim = cross_dim,
    audio_dim = adim, audio_num_attention_heads = aheads,
    audio_attention_head_dim = ahead_dim,
    audio_cross_attention_dim = across_dim
  )
  blk$eval()

  # Swap every cast linear for an NF4 module exactly like the loader does
  swap_lin <- function(parent, leaf) {
    old <- diffuseR:::.ltx23_walk_module(parent, leaf)
    q <- ltx23_nf4_quantize(old$weight)
    m <- ltx23_nf4_linear(old$weight$shape[1], old$weight$shape[2],
      bias = !is.null(old$bias))
    if (!is.null(old$bias)) m$bias <- old$bias
    m$set_nf4_weight(q$packed, q$absmax)
    do.call(`$<-`, list(parent, leaf, m))
  }
  for (attn in list(blk$attn1, blk$audio_attn1, blk$attn2, blk$audio_attn2,
                    blk$audio_to_video_attn, blk$video_to_audio_attn)) {
    swap_lin(attn, "to_q")
    swap_lin(attn, "to_k")
    swap_lin(attn, "to_v")
    swap_lin(attn$to_out, "0")
  }
  for (ff in list(blk$ff, blk$audio_ff)) {
    swap_lin(ff$net[[1]], "proj")
    swap_lin(ff$net, "2")
  }
  blk
}

blk1 <- make_block()
blk2 <- make_block()
expect_true(diffuseR:::.ltx23_jit_block_ok(blk1))

# The model's split-rope embedder emits per-head 4D freqs [B, H, T, r]
rope_pair <- function(s, n_heads, r) {
  ang <- torch::torch_rand(B, n_heads, s, r) * 6.28
  list(torch::torch_cos(ang), torch::torch_sin(ang))
}

h <- torch::torch_randn(B, Sv, dim)
ah <- torch::torch_randn(B, Sa, adim)
enc <- torch::torch_randn(B, St, cross_dim)
aenc <- torch::torch_randn(B, St, across_dim)
temb <- torch::torch_randn(B, 1L, 9L * dim)
temb_a <- torch::torch_randn(B, 1L, 9L * adim)
tcss <- torch::torch_randn(B, 1L, 4L * dim)
tcass <- torch::torch_randn(B, 1L, 4L * adim)
tcg <- torch::torch_randn(B, 1L, dim)
tcag <- torch::torch_randn(B, 1L, adim)
tp <- torch::torch_randn(B, 1L, 2L * dim)
tpa <- torch::torch_randn(B, 1L, 2L * adim)
v_rope <- rope_pair(Sv, heads, head_dim %/% 2L)
a_rope <- rope_pair(Sa, aheads, ahead_dim %/% 2L)
cav_rope <- rope_pair(Sv, aheads, ahead_dim %/% 2L)
caa_rope <- rope_pair(Sa, aheads, ahead_dim %/% 2L)
# Additive [B, 1, S] mask with one text token masked out
enc_mask <- torch::torch_zeros(B, 1L, St)
enc_mask[, , St] <- -10000
aenc_mask <- torch::torch_zeros(B, 1L, St)

eager <- function(blk, h, ah) {
  blk(
    hidden_states = h, audio_hidden_states = ah,
    encoder_hidden_states = enc, audio_encoder_hidden_states = aenc,
    temb = temb, temb_audio = temb_a,
    temb_ca_scale_shift = tcss, temb_ca_audio_scale_shift = tcass,
    temb_ca_gate = tcg, temb_ca_audio_gate = tcag,
    temb_prompt = tp, temb_prompt_audio = tpa,
    video_rotary_emb = v_rope, audio_rotary_emb = a_rope,
    ca_video_rotary_emb = cav_rope, ca_audio_rotary_emb = caa_rope,
    encoder_attention_mask = enc_mask,
    audio_encoder_attention_mask = aenc_mask
  )
}

jit <- function(blocks, h, ah) {
  diffuseR:::.ltx23_jit_run_stack(
    blocks, h, ah, enc, aenc,
    temb, temb_a, tcss, tcass, tcg, tcag, tp, tpa,
    v_rope, a_rope, cav_rope, caa_rope,
    encoder_attention_mask = enc_mask,
    audio_encoder_attention_mask = aenc_mask
  )
}

max_abs_diff <- function(a, b) as.numeric((a - b)$abs()$max())

torch::with_no_grad({
  ref1 <- eager(blk1, h, ah)
  out1 <- jit(list(blk1), h, ah)
})
expect_equal(as.integer(out1[[1]]$shape), as.integer(ref1[[1]]$shape))
expect_equal(as.integer(out1[[2]]$shape), as.integer(ref1[[2]]$shape))
expect_true(max_abs_diff(out1[[1]], ref1[[1]]) < 1e-4)
expect_true(max_abs_diff(out1[[2]], ref1[[2]]) < 1e-4)

# Two-block stack must equal two sequential eager blocks (checks the
# per-block base-offset arithmetic in the weight list)
torch::with_no_grad({
  mid <- eager(blk1, h, ah)
  ref2 <- eager(blk2, mid[[1]], mid[[2]])
  out2 <- jit(list(blk1, blk2), h, ah)
})
expect_true(max_abs_diff(out2[[1]], ref2[[1]]) < 1e-4)
expect_true(max_abs_diff(out2[[2]], ref2[[2]]) < 1e-4)

# The mask must actually mask: moving the -10000 changes the output
enc_mask2 <- torch::torch_zeros(B, 1L, St)
enc_mask2[, , 1L] <- -10000
torch::with_no_grad({
  out_m <- diffuseR:::.ltx23_jit_run_stack(
    list(blk1), h, ah, enc, aenc,
    temb, temb_a, tcss, tcass, tcg, tcag, tp, tpa,
    v_rope, a_rope, cav_rope, caa_rope,
    encoder_attention_mask = enc_mask2,
    audio_encoder_attention_mask = aenc_mask
  )
})
expect_true(max_abs_diff(out_m[[1]], out1[[1]]) > 1e-6)

# NULL masks accepted
torch::with_no_grad({
  ref_nm <- blk1(
    hidden_states = h, audio_hidden_states = ah,
    encoder_hidden_states = enc, audio_encoder_hidden_states = aenc,
    temb = temb, temb_audio = temb_a,
    temb_ca_scale_shift = tcss, temb_ca_audio_scale_shift = tcass,
    temb_ca_gate = tcg, temb_ca_audio_gate = tcag,
    temb_prompt = tp, temb_prompt_audio = tpa,
    video_rotary_emb = v_rope, audio_rotary_emb = a_rope,
    ca_video_rotary_emb = cav_rope, ca_audio_rotary_emb = caa_rope
  )
  out_nm <- diffuseR:::.ltx23_jit_run_stack(
    list(blk1), h, ah, enc, aenc,
    temb, temb_a, tcss, tcass, tcg, tcag, tp, tpa,
    v_rope, a_rope, cav_rope, caa_rope
  )
})
expect_true(max_abs_diff(out_nm[[1]], ref_nm[[1]]) < 1e-4)
expect_true(max_abs_diff(out_nm[[2]], ref_nm[[2]]) < 1e-4)

# 3D whole-vector rope layout (the apply fn's other branch) also matches
rope3 <- function(s, r) {
  ang <- torch::torch_rand(1L, s, r) * 6.28
  list(torch::torch_cos(ang), torch::torch_sin(ang))
}
v3 <- rope3(Sv, (heads * head_dim) %/% 2L)
a3 <- rope3(Sa, (aheads * ahead_dim) %/% 2L)
cav3 <- rope3(Sv, (aheads * ahead_dim) %/% 2L)
caa3 <- rope3(Sa, (aheads * ahead_dim) %/% 2L)
torch::with_no_grad({
  ref3 <- blk1(
    hidden_states = h, audio_hidden_states = ah,
    encoder_hidden_states = enc, audio_encoder_hidden_states = aenc,
    temb = temb, temb_audio = temb_a,
    temb_ca_scale_shift = tcss, temb_ca_audio_scale_shift = tcass,
    temb_ca_gate = tcg, temb_ca_audio_gate = tcag,
    temb_prompt = tp, temb_prompt_audio = tpa,
    video_rotary_emb = v3, audio_rotary_emb = a3,
    ca_video_rotary_emb = cav3, ca_audio_rotary_emb = caa3
  )
  out3 <- diffuseR:::.ltx23_jit_run_stack(
    list(blk1), h, ah, enc, aenc,
    temb, temb_a, tcss, tcass, tcg, tcag, tp, tpa,
    v3, a3, cav3, caa3
  )
})
expect_true(max_abs_diff(out3[[1]], ref3[[1]]) < 1e-4)
expect_true(max_abs_diff(out3[[2]], ref3[[2]]) < 1e-4)

# Per-token temb (i2v conditioning: per-token video timestep) matches
temb_tok <- torch::torch_randn(B, Sv, 9L * dim)
torch::with_no_grad({
  ref_pt <- blk1(
    hidden_states = h, audio_hidden_states = ah,
    encoder_hidden_states = enc, audio_encoder_hidden_states = aenc,
    temb = temb_tok, temb_audio = temb_a,
    temb_ca_scale_shift = tcss, temb_ca_audio_scale_shift = tcass,
    temb_ca_gate = tcg, temb_ca_audio_gate = tcag,
    temb_prompt = tp, temb_prompt_audio = tpa,
    video_rotary_emb = v_rope, audio_rotary_emb = a_rope,
    ca_video_rotary_emb = cav_rope, ca_audio_rotary_emb = caa_rope
  )
  out_pt <- diffuseR:::.ltx23_jit_run_stack(
    list(blk1), h, ah, enc, aenc,
    temb_tok, temb_a, tcss, tcass, tcg, tcag, tp, tpa,
    v_rope, a_rope, cav_rope, caa_rope
  )
})
expect_true(max_abs_diff(out_pt[[1]], ref_pt[[1]]) < 1e-4)
expect_true(max_abs_diff(out_pt[[2]], ref_pt[[2]]) < 1e-4)

# Compact conditioned temb (2 variants + per-token index) must equal
# the eager block run with the explicitly expanded per-token temb
temb2 <- torch::torch_randn(B, 2L, 9L * dim)
cond_idx <- torch::torch_tensor(c(1L, 1L, 0L, 0L, 0L, 0L),
  dtype = torch::torch_long()) # first two tokens conditioned
temb_full <- temb2$index_select(2L, cond_idx$add(1L))
torch::with_no_grad({
  ref_ci <- blk1(
    hidden_states = h, audio_hidden_states = ah,
    encoder_hidden_states = enc, audio_encoder_hidden_states = aenc,
    temb = temb_full, temb_audio = temb_a,
    temb_ca_scale_shift = tcss, temb_ca_audio_scale_shift = tcass,
    temb_ca_gate = tcg, temb_ca_audio_gate = tcag,
    temb_prompt = tp, temb_prompt_audio = tpa,
    video_rotary_emb = v_rope, audio_rotary_emb = a_rope,
    ca_video_rotary_emb = cav_rope, ca_audio_rotary_emb = caa_rope
  )
  out_ci <- diffuseR:::.ltx23_jit_run_stack(
    list(blk1), h, ah, enc, aenc,
    temb2, temb_a, tcss, tcass, tcg, tcag, tp, tpa,
    v_rope, a_rope, cav_rope, caa_rope,
    cond_token_index = cond_idx
  )
})
expect_true(max_abs_diff(out_ci[[1]], ref_ci[[1]]) < 1e-4)
expect_true(max_abs_diff(out_ci[[2]], ref_ci[[2]]) < 1e-4)

# The byte LUT decodes every byte to its (hi nibble, lo nibble)
# codebook pair in order
lut <- diffuseR:::.ltx23_jit_table(torch::torch_device("cpu"),
  torch::torch_float32())
expect_equal(as.integer(lut$shape), c(256L, 2L))
nf4_levels <- diffuseR:::.ltx23_nf4_table
for (b in c(0L, 1L, 15L, 16L, 137L, 240L, 255L)) {
  expect_equal(as.numeric(lut[b + 1L, ]),
    c(nf4_levels[b %/% 16L + 1L], nf4_levels[b %% 16L + 1L]))
}

#' FLUX Transformer Building Blocks
#'
#' Fresh R port of the FLUX MMDiT blocks from the diffusers reference
#' implementation (Apache-2.0,
#' src/diffusers/models/transformers/transformer_flux.py and
#' src/diffusers/models/normalization.py). Module field names mirror the
#' diffusers state-dict keys 1:1 so checkpoints load without remapping.
#' Reuses the LTX primitives \code{ltx23_rms_norm}, \code{.ltx23_sdpa}
#' and \code{ltx23_feed_forward}.
#'
#' @name dit_flux_modules
NULL

#' FLUX adaLN-Zero modulation (double-stream)
#'
#' Projects the conditioning embedding to six modulation vectors and
#' returns the msa-modulated input plus the remaining parameters.
#' Reference: diffusers AdaLayerNormZero.
#'
#' @param dim Integer. Model dimension.
#'
#' @return Module whose forward(x, emb) returns
#'   \code{list(x_norm, gate_msa, shift_mlp, scale_mlp, gate_mlp)}.
#'
#' @export
flux_ada_layer_norm_zero <- torch::nn_module(
    "flux_ada_layer_norm_zero",
    initialize = function(dim) {
    self$linear <- torch::nn_linear(dim, 6L * dim, bias = TRUE)
    self$norm <- torch::nn_layer_norm(dim, eps = 1e-6,
                                      elementwise_affine = FALSE)
},
    forward = function(x, emb) {
    emb <- self$linear(torch::nnf_silu(emb))
    p <- emb$chunk(6L, dim = 2L)
    # shift_msa, scale_msa, gate_msa, shift_mlp, scale_mlp, gate_mlp
    x <- self$norm(x) * p[[2]]$unsqueeze(2L)$add(1) + p[[1]]$unsqueeze(2L)
    list(x, p[[3]], p[[4]], p[[5]], p[[6]])
}
)

#' FLUX adaLN-Zero modulation (single-stream)
#'
#' Three modulation vectors: shift, scale, gate. Reference: diffusers
#' AdaLayerNormZeroSingle.
#'
#' @param dim Integer. Model dimension.
#'
#' @return Module whose forward(x, emb) returns \code{list(x_norm, gate)}.
#'
#' @export
flux_ada_layer_norm_zero_single <- torch::nn_module(
    "flux_ada_layer_norm_zero_single",
    initialize = function(dim) {
    self$linear <- torch::nn_linear(dim, 3L * dim, bias = TRUE)
    self$norm <- torch::nn_layer_norm(dim, eps = 1e-6,
                                      elementwise_affine = FALSE)
},
    forward = function(x, emb) {
    emb <- self$linear(torch::nnf_silu(emb))
    p <- emb$chunk(3L, dim = 2L)
    # shift_msa, scale_msa, gate_msa
    x <- self$norm(x) * p[[2]]$unsqueeze(2L)$add(1) + p[[1]]$unsqueeze(2L)
    list(x, p[[3]])
}
)

#' FLUX continuous adaLN (final norm)
#'
#' Scale/shift conditioning of the final norm. Note the chunk order:
#' scale first, then shift (the reverse of adaLN-Zero). Reference:
#' diffusers AdaLayerNormContinuous as used by FLUX norm_out
#' (elementwise_affine = FALSE, eps = 1e-6).
#'
#' @param dim Integer. Model dimension.
#' @param cond_dim Integer. Conditioning embedding dimension.
#' @param bias Logical. Bias on the projection (TRUE for FLUX.1, FALSE
#'   for FLUX.2).
#'
#' @return Module whose forward(x, cond) returns \code{x} normalized and
#'   then scaled and shifted by the conditioning embedding, a tensor of
#'   the same shape as \code{x}.
#'
#' @export
flux_ada_layer_norm_continuous <- torch::nn_module(
    "flux_ada_layer_norm_continuous",
    initialize = function(dim, cond_dim = dim, bias = TRUE) {
    self$linear <- torch::nn_linear(cond_dim, 2L * dim, bias = bias)
    self$norm <- torch::nn_layer_norm(dim, eps = 1e-6,
                                      elementwise_affine = FALSE)
},
    forward = function(x, cond) {
    emb <- self$linear(torch::nnf_silu(cond)$to(dtype = x$dtype))
    p <- emb$chunk(2L, dim = 2L)
    # scale, shift
    self$norm(x) * p[[1]]$unsqueeze(2L)$add(1) + p[[2]]$unsqueeze(2L)
}
)

#' FLUX joint attention
#'
#' Multi-head attention with per-head RMS q/k norms and rotary position
#' embeddings. With \code{added_kv = TRUE} (double-stream blocks) the
#' text stream gets its own q/k/v projections and both streams attend
#' jointly (text tokens first); the outputs are split back and projected
#' per stream. With \code{pre_only = TRUE} (single-stream blocks) there
#' is no output projection. Reference: diffusers FluxAttention +
#' FluxAttnProcessor.
#'
#' @param query_dim Integer. Model dimension.
#' @param heads Integer. Number of attention heads.
#' @param dim_head Integer. Per-head dimension.
#' @param added_kv Logical. Add text-stream projections (double blocks).
#' @param pre_only Logical. Skip the output projection (single blocks).
#' @param eps Numeric. RMS norm epsilon.
#' @param bias Logical. Bias on the linear projections (TRUE for FLUX.1,
#'   FALSE for FLUX.2).
#'
#' @return Module whose forward(hidden_states, encoder_hidden_states,
#'   image_rotary_emb, chunk_size) returns the attended image stream
#'   [B, S, query_dim]. When \code{encoder_hidden_states} is supplied
#'   (double-stream blocks) it returns \code{list(image, text)} instead,
#'   each projected by its own output layer.
#'
#' @export
flux_attention <- torch::nn_module(
                                   "flux_attention",
                                   initialize = function(query_dim, heads, dim_head, added_kv = FALSE,
        pre_only = FALSE, eps = 1e-6, bias = TRUE) {
    inner_dim <- heads * dim_head
    self$heads <- heads
    self$dim_head <- dim_head
    self$added_kv <- added_kv
    self$pre_only <- pre_only

    self$norm_q <- ltx23_rms_norm(dim_head, eps = eps)
    self$norm_k <- ltx23_rms_norm(dim_head, eps = eps)
    self$to_q <- torch::nn_linear(query_dim, inner_dim, bias = bias)
    self$to_k <- torch::nn_linear(query_dim, inner_dim, bias = bias)
    self$to_v <- torch::nn_linear(query_dim, inner_dim, bias = bias)

    if (!pre_only) {
        self$to_out <- torch::nn_module_list(list(
                torch::nn_linear(inner_dim, query_dim, bias = bias)
            ))
    }
    if (added_kv) {
        self$norm_added_q <- ltx23_rms_norm(dim_head, eps = eps)
        self$norm_added_k <- ltx23_rms_norm(dim_head, eps = eps)
        self$add_q_proj <- torch::nn_linear(query_dim, inner_dim, bias = bias)
        self$add_k_proj <- torch::nn_linear(query_dim, inner_dim, bias = bias)
        self$add_v_proj <- torch::nn_linear(query_dim, inner_dim, bias = bias)
        self$to_add_out <- torch::nn_linear(inner_dim, query_dim, bias = bias)
    }
},
                                   forward = function(hidden_states, encoder_hidden_states = NULL,
        image_rotary_emb = NULL, chunk_size = NULL) {
    # Per-head layout [B, S, H, D]
    query <- self$to_q(hidden_states)$unflatten(3L, c(self$heads, -1L))
    key <- self$to_k(hidden_states)$unflatten(3L, c(self$heads, -1L))
    value <- self$to_v(hidden_states)$unflatten(3L, c(self$heads, -1L))

    query <- self$norm_q(query)
    key <- self$norm_k(key)

    if (!is.null(encoder_hidden_states)) {
        txt_len <- encoder_hidden_states$shape[2]
        eq <- self$add_q_proj(encoder_hidden_states)$unflatten(3L,
            c(self$heads, -1L))
        ek <- self$add_k_proj(encoder_hidden_states)$unflatten(3L, c(self$heads, -1L))
        ev <- self$add_v_proj(encoder_hidden_states)$unflatten(3L, c(self$heads, -1L))
        eq <- self$norm_added_q(eq)
        ek <- self$norm_added_k(ek)
        # Text tokens first, matching the rotary frequency layout
        query <- torch::torch_cat(list(eq, query), dim = 2L)
        key <- torch::torch_cat(list(ek, key), dim = 2L)
        value <- torch::torch_cat(list(ev, value), dim = 2L)
    }

    # [B, S, H, D] -> [B, H, S, D] for RoPE and attention
    query <- query$transpose(2L, 3L)
    key <- key$transpose(2L, 3L)
    value <- value$transpose(2L, 3L)

    if (!is.null(image_rotary_emb)) {
        query <- flux_apply_rotary_emb(query, image_rotary_emb)
        key <- flux_apply_rotary_emb(key, image_rotary_emb)
    }

    out <- .ltx23_sdpa(query, key, value, chunk_size = chunk_size)
    # [B, H, S, D] -> [B, S, H*D]
    out <- out$transpose(2L, 3L)$flatten(start_dim = 3L)
    out <- out$to(dtype = hidden_states$dtype)

    if (!is.null(encoder_hidden_states)) {
        seq_len <- out$shape[2]
        ctx <- out$narrow(2L, 1L, txt_len)
        img <- out$narrow(2L, txt_len + 1L, seq_len - txt_len)
        return(list(
                    self$to_out[[1]](img$contiguous()),
                    self$to_add_out(ctx$contiguous())
            ))
    }
    if (self$pre_only) {
        return(out)
    }
    self$to_out[[1]](out)
}
)

#' FLUX double-stream (MMDiT) transformer block
#'
#' Image and text streams each get adaLN-Zero modulation and a
#' feed-forward; attention is joint across both streams. Reference:
#' diffusers FluxTransformerBlock.
#'
#' @param dim Integer. Model dimension.
#' @param num_attention_heads Integer. Attention heads.
#' @param attention_head_dim Integer. Per-head dimension.
#'
#' @return Module whose forward(hidden_states, encoder_hidden_states,
#'   temb, image_rotary_emb) returns
#'   \code{list(encoder_hidden_states, hidden_states)}.
#'
#' @export
flux_double_block <- torch::nn_module(
                                      "flux_double_block",
                                      initialize = function(dim, num_attention_heads, attention_head_dim) {
    self$norm1 <- flux_ada_layer_norm_zero(dim)
    self$norm1_context <- flux_ada_layer_norm_zero(dim)
    self$attn <- flux_attention(dim, num_attention_heads, attention_head_dim,
                                added_kv = TRUE)
    self$norm2 <- torch::nn_layer_norm(dim, eps = 1e-6,
                                       elementwise_affine = FALSE)
    self$ff <- ltx23_feed_forward(dim, mult = 4L)
    self$norm2_context <- torch::nn_layer_norm(dim, eps = 1e-6,
        elementwise_affine = FALSE)
    self$ff_context <- ltx23_feed_forward(dim, mult = 4L)
},
                                      forward = function(hidden_states, encoder_hidden_states, temb,
        image_rotary_emb = NULL, chunk_size = NULL) {
    n1 <- self$norm1(hidden_states, emb = temb)
    n1c <- self$norm1_context(encoder_hidden_states, emb = temb)

    attn_out <- self$attn(
                          hidden_states = n1[[1]],
                          encoder_hidden_states = n1c[[1]],
                          image_rotary_emb = image_rotary_emb,
                          chunk_size = chunk_size
    )

    # Image stream: gated attention + modulated feed-forward
    hidden_states <- hidden_states + n1[[2]]$unsqueeze(2L) * attn_out[[1]]
    norm_h <- self$norm2(hidden_states) * n1[[4]]$unsqueeze(2L)$add(1) +
    n1[[3]]$unsqueeze(2L)
    hidden_states <- hidden_states + n1[[5]]$unsqueeze(2L) * self$ff(norm_h)

    # Text stream mirrors with its own modulation
    encoder_hidden_states <- encoder_hidden_states +
    n1c[[2]]$unsqueeze(2L) * attn_out[[2]]
    norm_c <- self$norm2_context(encoder_hidden_states) *
    n1c[[4]]$unsqueeze(2L)$add(1) + n1c[[3]]$unsqueeze(2L)
    encoder_hidden_states <- encoder_hidden_states +
    n1c[[5]]$unsqueeze(2L) * self$ff_context(norm_c)

    if (encoder_hidden_states$dtype == torch::torch_float16()) {
        encoder_hidden_states <- encoder_hidden_states$clamp(-65504, 65504)
    }
    list(encoder_hidden_states, hidden_states)
}
)

#' FLUX single-stream transformer block
#'
#' Parallel attention + MLP over the joint [text; image] sequence with a
#' shared gate: \code{x + gate * proj_out(cat(attn, gelu(mlp)))}. The
#' reference concatenates the streams inside every block and splits after;
#' here the caller concatenates once before the single-block stack, which
#' is numerically identical. Reference: diffusers
#' FluxSingleTransformerBlock.
#'
#' @param dim Integer. Model dimension.
#' @param num_attention_heads Integer. Attention heads.
#' @param attention_head_dim Integer. Per-head dimension.
#' @param mlp_ratio Numeric. MLP hidden dim multiplier.
#'
#' @return Module whose forward(hidden_states, temb, image_rotary_emb)
#'   returns the joint hidden states.
#'
#' @export
flux_single_block <- torch::nn_module(
                                      "flux_single_block",
                                      initialize = function(dim, num_attention_heads, attention_head_dim,
        mlp_ratio = 4.0) {
    mlp_hidden_dim <- as.integer(dim * mlp_ratio)
    self$norm <- flux_ada_layer_norm_zero_single(dim)
    self$proj_mlp <- torch::nn_linear(dim, mlp_hidden_dim)
    self$proj_out <- torch::nn_linear(dim + mlp_hidden_dim, dim)
    self$attn <- flux_attention(dim, num_attention_heads, attention_head_dim,
                                pre_only = TRUE)
},
                                      forward = function(hidden_states, temb, image_rotary_emb = NULL,
        chunk_size = NULL) {
    residual <- hidden_states
    n <- self$norm(hidden_states, emb = temb)
    mlp <- torch::nnf_gelu(self$proj_mlp(n[[1]]), approximate = "tanh")
    attn_out <- self$attn(
                          hidden_states = n[[1]],
                          image_rotary_emb = image_rotary_emb,
                          chunk_size = chunk_size
    )

    # Attention half first, then the MLP half
    hidden_states <- torch::torch_cat(list(attn_out, mlp), dim = 3L)
    hidden_states <- residual + n[[2]]$unsqueeze(2L) * self$proj_out(hidden_states)
    if (hidden_states$dtype == torch::torch_float16()) {
        hidden_states <- hidden_states$clamp(-65504, 65504)
    }
    hidden_states
}
)

#' Generate an image from a text prompt using SDXL
#'
#' @param prompt A character string prompt describing the image to generate.
#' @param negative_prompt Optional negative prompt to guide the generation.
#' @param img_dim Dimension of the output image (e.g., 512 for 512x512).
#' @param pipeline Optional A pre-loaded diffusion pipeline. If `NULL`, it will be loaded based on the model name and devices.
#' @param devices A named list of devices for each model component (e.g., `list(unet = "cuda", decoder = "cpu", text_encoder = "cpu")`),
#'   or "auto" to use `auto_devices()`, or NULL to use memory_profile devices.
#' @param memory_profile Character or list. Memory profile for GPU-poor optimization:
#'   "auto" for auto-detection, or a profile name ("full_gpu", "balanced", "unet_gpu", "cpu_only"),
#'   or a list from `sdxl_memory_profile()`. When specified, overrides devices parameter.
#' @param unet_dtype_str Optional A character for dtype of the unet component (typically "float16" for cuda and "float32" for cpu; float32 is available for cuda).
#' @param download_models Logical indicating whether to download the model files if they are not found.
#' @param scheduler Scheduler to use (e.g., `"ddim"`, `"euler"`).
#' @param timesteps Optional A vector of timesteps to use.
#' @param initial_latents Optional initial latents for the diffusion process.
#' @param num_inference_steps Number of inference steps to run.
#' @param guidance_scale Scale for classifier-free guidance (typically 7.5).
#' @param seed Optional seed for reproducibility.
#' @param save_file Logical indicating whether to save the generated image.
#' @param filename Optional filename for saving the image. If `NULL`, a default name is generated.
#' @param metadata_path Optional file path to save metadata.
#' @param use_native_decoder Logical; if TRUE, uses native R torch decoder instead of TorchScript.
#'   Native decoder has better GPU compatibility (especially Blackwell).
#' @param use_native_text_encoder Logical; if TRUE, uses native R torch text encoder instead of TorchScript.
#'   Native text encoder has better GPU compatibility (especially Blackwell).
#' @param use_native_unet Logical; if TRUE, uses native R torch UNet instead of TorchScript.
#'   Native UNet has better GPU compatibility (especially Blackwell).
#' @param diffusers_dir Optional path to a diffusers safetensors directory
#'   (\code{unet/}, \code{vae/}, \code{text_encoder/}, \code{text_encoder_2/}).
#'   When supplied, the pipeline is built end-to-end from safetensors via
#'   \code{\link{sdxl_pipeline_from_safetensors}} (native, Blackwell-safe,
#'   no TorchScript) and the other \code{use_native_*} flags are ignored.
#' @param verbose Logical. Print progress and memory status messages.
#' @param ... Additional parameters passed to the diffusion process.
#'
#' @return An image array and metadata
#' @export
#'
#' @examples
#' \dontrun{
#' # Basic usage with auto-detection
#' img <- txt2img_sdxl("a cat wearing sunglasses in space")
#'
#' # GPU-poor mode (8GB VRAM)
#' img <- txt2img_sdxl("a sunset over mountains", memory_profile = "unet_gpu")
#'
#' # Explicit memory profile
#' profile <- sdxl_memory_profile(vram_gb = 8)
#' img <- txt2img_sdxl("a forest path", memory_profile = profile)
#' }
txt2img_sdxl <- function(prompt, negative_prompt = NULL, img_dim = 1024,
                         pipeline = NULL, devices = "auto",
                         memory_profile = NULL, unet_dtype_str = NULL,
                         download_models = FALSE, scheduler = "ddim",
                         timesteps = NULL, initial_latents = NULL,
                         num_inference_steps = 30, guidance_scale = 7.5,
                         seed = NULL, save_file = TRUE, filename = NULL,
                         metadata_path = NULL, use_native_decoder = FALSE,
                         use_native_text_encoder = FALSE,
                         use_native_unet = FALSE, diffusers_dir = NULL,
                         verbose = TRUE, ...) {
    model_name <- "sdxl"

    # Resolve memory profile if provided
    profile <- NULL
    if (!is.null(memory_profile)) {
        if (identical(memory_profile, "auto")) {
            profile <- sdxl_memory_profile()
        } else if (is.character(memory_profile)) {
            # Named profile
            vram <- switch(memory_profile, "full_gpu" = 20, "balanced" = 12,
                           "unet_gpu" = 8, "cpu_only" = 0, NULL)
            if (!is.null(vram)) {
                profile <- sdxl_memory_profile(vram_gb = vram)
            }
        } else if (is.list(memory_profile)) {
            profile <- memory_profile
        }
    }

    # Use profile devices if available, otherwise fall back to devices parameter
    if (!is.null(profile)) {
        devices <- profile$devices
        if (verbose) {
            message(sprintf("Using memory profile: %s", profile$name))
        }
        # Validate resolution
        if (img_dim > profile$max_resolution) {
            warning(sprintf("Resolution %d exceeds profile max %d, reducing",
                            img_dim, profile$max_resolution))
            img_dim <- profile$max_resolution
        }
        # Set dtype from profile if not explicitly provided
        if (is.null(unet_dtype_str)) {
            unet_dtype_str <- profile$dtype
        }
    } else if (identical(devices, "auto")) {
        devices <- auto_devices(model_name)
    }

    m2d <- models2devices(model_name = model_name, devices = devices,
                          unet_dtype_str = unet_dtype_str)
    devices <- m2d$devices
    unet_dtype <- m2d$unet_dtype
    device_cpu <- m2d$device_cpu
    device_cuda <- m2d$device_cuda

    if (is.null(pipeline)) {
        if (!is.null(diffusers_dir)) {
            pipeline <- sdxl_pipeline_from_safetensors(diffusers_dir,
                devices = devices, unet_dtype = unet_dtype, verbose = verbose)
        } else {
            pipeline <- load_pipeline(model_name = model_name, m2d = m2d,
                                      unet_dtype_str = unet_dtype_str,
                                      use_native_decoder = use_native_decoder,
                                      use_native_text_encoder = use_native_text_encoder,
                                      use_native_unet = use_native_unet)
        }
    }

    # Start timing
    start_time <- proc.time()
    # Process text prompt
    if (verbose) {
        message("Processing prompt...")
    }
    ## Tokenizer
    tokens <- CLIPTokenizer(prompt)
    prompt_embed1 <- pipeline$text_encoder(tokens)
    te2_output <- pipeline$text_encoder2(tokens)
    prompt_embed2 <- te2_output[[1]]
    prompt_embed <- torch::torch_cat(list(prompt_embed1, prompt_embed2), dim = 3)

    text_embeds <- te2_output[[2]]
    text_embeds <- text_embeds$to(dtype = unet_dtype,
                                  device = torch::torch_device(devices$unet))
    time_ids <- torch::torch_tensor(c(img_dim, img_dim, 0, 0, img_dim, img_dim), # zero indexed as python
                                    dtype = unet_dtype,
                                    device = torch::torch_device(devices$unet))$unsqueeze(1)
    # clip-vit-large-patch14
    if (is.null(negative_prompt)) {
        empty_tokens <- CLIPTokenizer("")
        empty_prompt_embed <- torch::torch_zeros_like(prompt_embed)
        empty_text_embeds <- torch::torch_zeros_like(text_embeds)
    } else {
        empty_tokens <- CLIPTokenizer(negative_prompt)
        empty_prompt_embed1 <- pipeline$text_encoder(empty_tokens)
        empty_te2_output <- pipeline$text_encoder2(empty_tokens)
        empty_prompt_embed2 <- empty_te2_output[[1]]
        empty_prompt_embed <- torch::torch_cat(list(empty_prompt_embed1,
                empty_prompt_embed2), dim = 3)
        empty_text_embeds <- empty_te2_output[[2]]
    }
    empty_prompt_embed <- empty_prompt_embed$to(dtype = unet_dtype,
        device = torch::torch_device(devices$unet))
    prompt_embed <- prompt_embed$to(dtype = unet_dtype,
                                    device = torch::torch_device(devices$unet))
    empty_text_embeds <- empty_text_embeds$to(dtype = unet_dtype,
        device = torch::torch_device(devices$unet))

    # Phase cleanup after text encoding (for GPU-poor mode)
    if (!is.null(profile) && profile$cleanup == "phase") {
        if (verbose) {
            message("Clearing VRAM after text encoding...")
        }
        clear_vram(verbose = FALSE)
    }

    if (verbose) {
        message("Creating schedule...")
    }
    # Load scheduler
    schedule <- ddim_scheduler_create(num_inference_steps = num_inference_steps,
                                      beta_schedule = "scaled_linear",
                                      beta_start = 0.00085,
                                      beta_end = 0.012,
                                      rescale_betas_zero_snr = FALSE,
                                      device = torch::torch_device(devices$unet))
    if (is.null(timesteps)) {
        timesteps <- schedule$timesteps
    }

    # Run diffusion process
    if (verbose) {
        message("Generating image...")
    }
    if (!is.null(seed)) {
        set.seed(seed)
        torch::torch_manual_seed(seed = seed)
    }

    latent_dim <- img_dim / 8
    if (!is.null(initial_latents)) {
        latents <- initial_latents
        latents <- latents$to(dtype = unet_dtype, device = torch::torch_device(devices$unet))
    } else {
        # Create random latents
        latents <- torch::torch_randn(c(1, 4, latent_dim, latent_dim),
                                      dtype = unet_dtype,
                                      device = torch::torch_device(devices$unet))
    }
    # Denoising loop (no gradients needed for inference)
    pb <- utils::txtProgressBar(min = 0, max = length(timesteps), style = 3)
    torch::with_no_grad({
        for (i in seq_along(timesteps)) {
            timestep <- torch::torch_tensor(timesteps[i],
                dtype = torch::torch_long(),
                device = torch::torch_device(devices$unet))

            # Get both conditional and unconditional predictions
            noise_pred_cond <- pipeline$unet(latents, timestep, prompt_embed,
                text_embeds, time_ids)

            if (guidance_scale != 1) {
                # If guidance scale is not 1, we need to calculate the unconditional prediction
                # with an empty prompt
                noise_pred_uncond <- pipeline$unet(latents, timestep, empty_prompt_embed,
                    empty_text_embeds, time_ids)
                # CFG step
                noise_pred <- noise_pred_uncond + guidance_scale *
                (noise_pred_cond - noise_pred_uncond)
            } else {
                # If guidance scale is 1, we can use the conditional prediction directly
                noise_pred <- noise_pred_cond
            }

            # Calculating latent
            latents <- ddim_scheduler_step(model_output = noise_pred,
                timestep = timestep,
                sample = latents,
                schedule = schedule,
                prediction_type = "epsilon",
                device = devices$unet)
            latents <- latents$to(dtype = unet_dtype, device = torch::torch_device(devices$unet))
            utils::setTxtProgressBar(pb, i)

            # Step cleanup for GPU-poor mode
            if (!is.null(profile) && profile$step_cleanup_interval > 0) {
                if (i %% profile$step_cleanup_interval == 0) {
                    clear_vram(verbose = FALSE)
                }
            }
        }
    })
    close(pb)

    # Phase cleanup before decode (for GPU-poor mode)
    if (!is.null(profile) && profile$cleanup == "phase") {
        if (verbose) {
            message("Clearing VRAM before decode...")
        }
        clear_vram(verbose = FALSE)
    }

    # Decode latents to image. SDXL's VAE scaling_factor is 0.13025 (vs SD
    # 2.1's 0.18215); the safetensors pipeline reads it from the VAE config
    # and its decoder already applies post_quant_conv, so skip the separate
    # TorchScript post_quant_conv in that path.
    vae_scaling <- if (!is.null(pipeline$vae_scaling)) {
        pipeline$vae_scaling
    } else {
        0.18215
    }
    scaled_latent <- latents / vae_scaling
    scaled_latent <- scaled_latent$to(dtype = torch::torch_float32(),
                                      device = torch::torch_device(devices$decoder))

    if (verbose) {
        message("Decoding image...")
    }
    if (isTRUE(pipeline$native_decode)) {
        decoded_output <- pipeline$decoder(scaled_latent)
    } else {
        post_conv_latent <- post_quant_conv(x = scaled_latent,
            dtype = torch::torch_float32(),
            device = devices$decoder)
        decoded_output <- pipeline$decoder(post_conv_latent)
    }
    # Ensure tensor is on CPU
    img <- decoded_output$cpu()

    # Remove batch dimension if present
    if (length(img$shape) == 4) {
        img <- img$squeeze(1)
    }

    # Reorder channels: [3, H, W] → [H, W, 3]
    img <- img$permute(c(2, 3, 1))

    # Normalize
    img <- (img + 1) / 2 # scale from [-1, 1] → [0, 1]
    img <- torch::torch_clamp(img, min = 0, max = 1)

    # Convert to R array
    img_array <- as.array(img)
    # Save if requested
    if (save_file) {
        # Creating filename while we're here
        if (is.null(filename)) {
            filename <- filename_from_prompt(prompt, datetime = TRUE)
        }
        message("Saving image to ", filename)
        save_image(img = img_array, filename)
    } else {
        # Display in RStudio Viewer if interactive
        if (interactive()) {
            grid::grid.raster(img_array)
        }
    }
    # Save metadata
    metadata <- list(
                     prompt = prompt,
                     negative_prompt = negative_prompt,
                     width = img_dim,
                     height = img_dim,
                     num_inference_steps = num_inference_steps,
                     guidance_scale = guidance_scale,
                     seed = seed,
                     scheduler = scheduler,
                     model = model_name,
                     filename = filename
    )
    if (!is.null(metadata_path)) {
        utils::write.csv(metadata, file = metadata_path, row.names = FALSE)
        message("Metadata saved to: ", metadata_path)
    }
    # Report timing
    elapsed <- proc.time() - start_time
    if (verbose) {
        message(sprintf("Image generated in %.2f seconds", elapsed[3]))
    }

    # Return the generated image and metadata
    return(list(image = img_array, metadata = metadata))
}

#' LTX-2.3 Audio-Video Diffusion Transformer
#'
#' Fresh R port of the LTX-2 transformer from the diffusers reference
#' (Apache-2.0, src/diffusers/models/transformers/transformer_ltx2.py),
#' configured for LTX 2.3: gated attention, cross-attention modulation,
#' prompt AdaLN, split RoPE, and connector-projected text embeddings
#' (no in-model caption projection).
#'
#' @name dit_ltx23
NULL

#' LTX-2.3 video transformer model
#'
#' Dual-stream audio/video DiT. Text embeddings arrive already projected
#' to the video (\code{inner_dim}) and audio (\code{audio_inner_dim})
#' dimensions by the connector modules.
#'
#' @param in_channels,out_channels Integers. Video latent channels.
#' @param patch_size,patch_size_t Integers. Video patch sizes.
#' @param num_attention_heads,attention_head_dim Video attention shape.
#' @param cross_attention_dim Integer. Video text embedding dimension.
#' @param vae_scale_factors Integer vector. VAE (time, height, width) scales.
#' @param pos_embed_max_pos,base_height,base_width RoPE base grid.
#' @param audio_in_channels,audio_out_channels Integers. Audio latent channels.
#' @param audio_patch_size,audio_patch_size_t Integers. Audio patch sizes.
#' @param audio_num_attention_heads,audio_attention_head_dim Audio attention shape.
#' @param audio_cross_attention_dim Integer. Audio text embedding dimension.
#' @param audio_scale_factor,audio_pos_embed_max_pos,audio_sampling_rate,audio_hop_length
#'   Audio latent grid parameters.
#' @param num_layers Integer. Transformer block count.
#' @param norm_eps Numeric. Norm epsilon.
#' @param rope_theta,rope_double_precision,causal_offset RoPE parameters.
#' @param timestep_scale_multiplier,cross_attn_timestep_scale_multiplier
#'   Timestep scaling (inputs arrive already scaled; the ratio modulates
#'   the a2v/v2a gates).
#' @param rope_type "split" (LTX-2.3) or "interleaved".
#' @param gated_attn,cross_attn_mod,audio_gated_attn,audio_cross_attn_mod,perturbed_attn
#'   LTX-2.3 feature flags (all TRUE for the 2.3 checkpoints).
#'
#' @return Module whose forward(hidden_states, ...) returns
#'   \code{list(sample, audio_sample)}: the predicted velocity for the
#'   video latent tokens and, when the audio branch is active, for the
#'   audio latent tokens (\code{audio_sample} is NULL otherwise).
#'
#' @export
ltx23_transformer <- torch::nn_module(
                                      "ltx23_transformer",
                                      initialize = function(
        in_channels = 128L,
        out_channels = 128L,
        patch_size = 1L,
        patch_size_t = 1L,
        num_attention_heads = 32L,
        attention_head_dim = 128L,
        cross_attention_dim = 4096L,
        vae_scale_factors = c(8L, 32L, 32L),
        pos_embed_max_pos = 20L,
        base_height = 2048L,
        base_width = 2048L,
        gated_attn = TRUE,
        cross_attn_mod = TRUE,
        audio_in_channels = 128L,
        audio_out_channels = 128L,
        audio_patch_size = 1L,
        audio_patch_size_t = 1L,
        audio_num_attention_heads = 32L,
        audio_attention_head_dim = 64L,
        audio_cross_attention_dim = 2048L,
        audio_scale_factor = 4L,
        audio_pos_embed_max_pos = 20L,
        audio_sampling_rate = 16000L,
        audio_hop_length = 160L,
        audio_gated_attn = TRUE,
        audio_cross_attn_mod = TRUE,
        num_layers = 48L,
        norm_eps = 1e-6,
        rope_theta = 10000.0,
        rope_double_precision = TRUE,
        causal_offset = 1L,
        timestep_scale_multiplier = 1000,
        cross_attn_timestep_scale_multiplier = 1000,
        rope_type = "split",
        perturbed_attn = TRUE
    ) {
    inner_dim <- num_attention_heads * attention_head_dim
    audio_inner_dim <- audio_num_attention_heads * audio_attention_head_dim
    self$inner_dim <- inner_dim
    self$audio_inner_dim <- audio_inner_dim
    self$timestep_scale_multiplier <- timestep_scale_multiplier
    self$cross_attn_timestep_scale_multiplier <- cross_attn_timestep_scale_multiplier
    self$prompt_modulation <- cross_attn_mod || audio_cross_attn_mod

    # 1. Patchified input projections
    self$proj_in <- linear_noinit(in_channels, inner_dim)
    self$audio_proj_in <- linear_noinit(audio_in_channels, audio_inner_dim)

    # 2. Global timestep modulation (9 params each with cross-attn mod)
    video_mod_params <- if (cross_attn_mod) 9L else 6L
    audio_mod_params <- if (audio_cross_attn_mod) 9L else 6L
    self$time_embed <- ltx23_ada_layer_norm_single(inner_dim, video_mod_params)
    self$audio_time_embed <- ltx23_ada_layer_norm_single(audio_inner_dim,
        audio_mod_params)

    # Global a2v/v2a cross-attention modulation
    self$av_cross_attn_video_scale_shift <- ltx23_ada_layer_norm_single(inner_dim, 4L)
    self$av_cross_attn_audio_scale_shift <- ltx23_ada_layer_norm_single(audio_inner_dim, 4L)
    self$av_cross_attn_video_a2v_gate <- ltx23_ada_layer_norm_single(inner_dim, 1L)
    self$av_cross_attn_audio_v2a_gate <- ltx23_ada_layer_norm_single(audio_inner_dim, 1L)

    # Output layer modulation
    self$scale_shift_table <- torch::nn_parameter(
        torch::torch_randn(2L, inner_dim) / sqrt(inner_dim)
    )
    self$audio_scale_shift_table <- torch::nn_parameter(
        torch::torch_randn(2L, audio_inner_dim) / sqrt(audio_inner_dim)
    )

    # Prompt modulation (LTX-2.3)
    if (self$prompt_modulation) {
        self$prompt_adaln <- ltx23_ada_layer_norm_single(inner_dim, 2L)
        self$audio_prompt_adaln <- ltx23_ada_layer_norm_single(audio_inner_dim, 2L)
    }

    # 3. RoPE embedders (self-attention per modality + a2v/v2a cross)
    self$rope <- ltx23_rotary_pos_embed(
                                        dim = inner_dim, patch_size = patch_size, patch_size_t = patch_size_t,
                                        base_num_frames = pos_embed_max_pos, base_height = base_height,
                                        base_width = base_width, scale_factors = vae_scale_factors,
                                        theta = rope_theta, causal_offset = causal_offset, modality = "video",
                                        double_precision = rope_double_precision, rope_type = rope_type,
                                        num_attention_heads = num_attention_heads
    )
    self$audio_rope <- ltx23_rotary_pos_embed(
        dim = audio_inner_dim, patch_size = audio_patch_size,
        patch_size_t = audio_patch_size_t, base_num_frames = audio_pos_embed_max_pos,
        sampling_rate = audio_sampling_rate, hop_length = audio_hop_length,
        scale_factors = audio_scale_factor,
        theta = rope_theta, causal_offset = causal_offset, modality = "audio",
        double_precision = rope_double_precision, rope_type = rope_type,
        num_attention_heads = audio_num_attention_heads
    )
    ca_max_pos <- max(pos_embed_max_pos, audio_pos_embed_max_pos)
    self$cross_attn_rope <- ltx23_rotary_pos_embed(
        dim = audio_cross_attention_dim, patch_size = patch_size,
        patch_size_t = patch_size_t, base_num_frames = ca_max_pos,
        base_height = base_height, base_width = base_width,
        theta = rope_theta, causal_offset = causal_offset, modality = "video",
        double_precision = rope_double_precision, rope_type = rope_type,
        num_attention_heads = num_attention_heads
    )
    self$cross_attn_audio_rope <- ltx23_rotary_pos_embed(
        dim = audio_cross_attention_dim, patch_size = audio_patch_size,
        patch_size_t = audio_patch_size_t, base_num_frames = ca_max_pos,
        sampling_rate = audio_sampling_rate, hop_length = audio_hop_length,
        theta = rope_theta, causal_offset = causal_offset, modality = "audio",
        double_precision = rope_double_precision, rope_type = rope_type,
        num_attention_heads = audio_num_attention_heads
    )

    # 4. Transformer blocks
    self$transformer_blocks <- torch::nn_module_list(lapply(seq_len(num_layers), function(i) {
        ltx23_transformer_block(
                                dim = inner_dim,
                                num_attention_heads = num_attention_heads,
                                attention_head_dim = attention_head_dim,
                                cross_attention_dim = cross_attention_dim,
                                audio_dim = audio_inner_dim,
                                audio_num_attention_heads = audio_num_attention_heads,
                                audio_attention_head_dim = audio_attention_head_dim,
                                audio_cross_attention_dim = audio_cross_attention_dim,
                                video_gated_attn = gated_attn,
                                video_cross_attn_adaln = cross_attn_mod,
                                audio_gated_attn = audio_gated_attn,
                                audio_cross_attn_adaln = audio_cross_attn_mod,
                                eps = norm_eps,
                                elementwise_affine = FALSE,
                                rope_type = rope_type,
                                perturbed_attn = perturbed_attn
        )
    }))

    # 5. Output layers
    self$norm_out <- torch::nn_layer_norm(inner_dim, eps = 1e-6, elementwise_affine = FALSE)
    self$proj_out <- linear_noinit(inner_dim, out_channels)
    self$audio_norm_out <- torch::nn_layer_norm(audio_inner_dim, eps = 1e-6,
        elementwise_affine = FALSE)
    self$audio_proj_out <- linear_noinit(audio_inner_dim, audio_out_channels)
},
                                      forward = function(
        hidden_states,
        audio_hidden_states,
        encoder_hidden_states,
        audio_encoder_hidden_states,
        timestep,
        audio_timestep = NULL,
        sigma = NULL,
        audio_sigma = NULL,
        encoder_attention_mask = NULL,
        audio_encoder_attention_mask = NULL,
        num_frames = NULL,
        height = NULL,
        width = NULL,
        fps = 24.0,
        audio_num_frames = NULL,
        video_coords = NULL,
        audio_coords = NULL,
        isolate_modalities = FALSE,
        spatio_temporal_guidance_blocks = NULL,
        perturbation_mask = NULL,
        use_cross_timestep = FALSE,
        video_self_attention_mask = NULL,
        cond_token_index = NULL
    ) {
    audio_timestep <- audio_timestep %||% timestep
    audio_sigma <- audio_sigma %||% sigma
    if (self$prompt_modulation && is.null(sigma)) {
        stop("sigma is required for LTX-2.3 prompt modulation")
    }

    # Multiplicative [B, S] masks -> additive bias [B, 1, S]
    if (!is.null(encoder_attention_mask) && encoder_attention_mask$ndim == 2L) {
        encoder_attention_mask <- encoder_attention_mask$to(
            dtype = hidden_states$dtype
        )$neg()$add(1)$mul(-10000.0)$unsqueeze(2L)
    }
    if (!is.null(audio_encoder_attention_mask) &&
                                      audio_encoder_attention_mask$ndim == 2L) {
        audio_encoder_attention_mask <- audio_encoder_attention_mask$to(
            dtype = audio_hidden_states$dtype
        )$neg()$add(1)$mul(-10000.0)$unsqueeze(2L)
    }
    if (!is.null(video_self_attention_mask)) {
        video_self_attention_mask <- video_self_attention_mask$to(
            dtype = hidden_states$dtype
        )$neg()$add(1)$mul(-10000.0)
    }

    batch_size <- hidden_states$shape[1]

    # 1. RoPE frequencies
    if (is.null(video_coords)) {
        video_coords <- self$rope$prepare_video_coords(
            batch_size, num_frames, height, width, hidden_states$device, fps = fps
        )
    }
    if (is.null(audio_coords)) {
        audio_coords <- self$audio_rope$prepare_audio_coords(
            batch_size, audio_num_frames, audio_hidden_states$device
        )
    }
    video_rotary_emb <- self$rope(video_coords, device = hidden_states$device)
    audio_rotary_emb <- self$audio_rope(audio_coords, device = audio_hidden_states$device)
    # Cross-modal attention uses the temporal axis only
    video_ca_rotary_emb <- self$cross_attn_rope(
        video_coords$narrow(2L, 1L, 1L), device = hidden_states$device
    )
    audio_ca_rotary_emb <- self$cross_attn_audio_rope(
        audio_coords$narrow(2L, 1L, 1L), device = audio_hidden_states$device
    )

    # 2. Input projections
    hidden_states <- self$proj_in(hidden_states)
    audio_hidden_states <- self$audio_proj_in(audio_hidden_states)

    # 3. Timestep embeddings and global modulation parameters
    gate_scale_factor <- self$cross_attn_timestep_scale_multiplier /
    self$timestep_scale_multiplier

    te <- self$time_embed(timestep$flatten(), hidden_dtype = hidden_states$dtype)
    temb <- te[[1]]$view(c(batch_size, -1L, te[[1]]$shape[2]))
    embedded_timestep <- te[[2]]$view(c(batch_size, -1L, te[[2]]$shape[2]))

    ate <- self$audio_time_embed(audio_timestep$flatten(),
                                 hidden_dtype = audio_hidden_states$dtype)
    temb_audio <- ate[[1]]$view(c(batch_size, -1L, ate[[1]]$shape[2]))
    audio_embedded_timestep <- ate[[2]]$view(c(batch_size, -1L, ate[[2]]$shape[2]))

    if (self$prompt_modulation) {
        pe <- self$prompt_adaln(sigma$flatten(), hidden_dtype = hidden_states$dtype)
        temb_prompt <- pe[[1]]$view(c(batch_size, -1L, pe[[1]]$shape[2]))
        ape <- self$audio_prompt_adaln(audio_sigma$flatten(),
                                       hidden_dtype = audio_hidden_states$dtype)
        temb_prompt_audio <- ape[[1]]$view(c(batch_size, -1L, ape[[1]]$shape[2]))
    } else {
        temb_prompt <- NULL
        temb_prompt_audio <- NULL
    }

    # a2v/v2a modulation; 2.3 modulates each modality by the *other*
    # modality's sigma (use_cross_timestep)
    video_ca_timestep <- if (use_cross_timestep) audio_sigma$flatten() else timestep$flatten()
    vcss <- self$av_cross_attn_video_scale_shift(
        video_ca_timestep, hidden_dtype = hidden_states$dtype
    )[[1]]
    vcg <- self$av_cross_attn_video_a2v_gate(
        video_ca_timestep * gate_scale_factor, hidden_dtype = hidden_states$dtype
    )[[1]]
    video_ca_scale_shift <- vcss$view(c(batch_size, -1L, vcss$shape[2]))
    video_ca_a2v_gate <- vcg$view(c(batch_size, -1L, vcg$shape[2]))

    audio_ca_timestep <- if (use_cross_timestep) sigma$flatten() else audio_timestep$flatten()
    acss <- self$av_cross_attn_audio_scale_shift(
        audio_ca_timestep, hidden_dtype = audio_hidden_states$dtype
    )[[1]]
    acg <- self$av_cross_attn_audio_v2a_gate(
        audio_ca_timestep * gate_scale_factor, hidden_dtype = audio_hidden_states$dtype
    )[[1]]
    audio_ca_scale_shift <- acss$view(c(batch_size, -1L, acss$shape[2]))
    audio_ca_v2a_gate <- acg$view(c(batch_size, -1L, acg$shape[2]))

    # 4. STG perturbation setup
    stg_blocks <- spatio_temporal_guidance_blocks %||% integer(0)
    if (length(stg_blocks) > 0L && is.null(perturbation_mask)) {
        perturbation_mask <- torch::torch_zeros(batch_size, device = hidden_states$device)
    }
    if (!is.null(perturbation_mask) && perturbation_mask$ndim == 1L) {
        perturbation_mask <- perturbation_mask$unsqueeze(2L)$unsqueeze(3L)
    }
    all_perturbed <- if (!is.null(perturbation_mask)) {
        as.logical((perturbation_mask == 0)$all()$item())
    } else {
        FALSE
    }

    # 5. Transformer blocks
    # Compiled path: the entire block stack in one TorchScript call (no
    # per-op R dispatch, no R-side garbage, fused SDPA, no per-block
    # gc). Requires the NF4-quantized 2.3 configuration and none of the
    # STG/masking extras the script doesn't model.
    use_jit <- isTRUE(getOption("diffuseR.jit_blocks", TRUE)) &&
    length(spatio_temporal_guidance_blocks %||% integer(0)) == 0L &&
    is.null(perturbation_mask) &&
    is.null(video_self_attention_mask) &&
    !isolate_modalities &&
    self$prompt_modulation &&
    !is.null(video_rotary_emb) && !is.null(audio_rotary_emb) &&
    !is.null(video_ca_rotary_emb) && !is.null(audio_ca_rotary_emb) &&
    .ltx23_jit_block_ok(self$transformer_blocks[[1]])

    if (use_jit) {
        res <- .ltx23_jit_run_stack(
                                    self$transformer_blocks,
                                    hidden_states, audio_hidden_states,
                                    encoder_hidden_states, audio_encoder_hidden_states,
                                    temb, temb_audio,
                                    video_ca_scale_shift, audio_ca_scale_shift,
                                    video_ca_a2v_gate, audio_ca_v2a_gate,
                                    temb_prompt, temb_prompt_audio,
                                    video_rotary_emb, audio_rotary_emb,
                                    video_ca_rotary_emb, audio_ca_rotary_emb,
                                    encoder_attention_mask,
                                    audio_encoder_attention_mask,
                                    cond_token_index = cond_token_index
        )
        hidden_states <- res[[1]]
        audio_hidden_states <- res[[2]]
    }

    block_gc <- isTRUE(getOption("diffuseR.block_gc"))
    if (!is.null(cond_token_index)) {
        # Expand the compact [B, V, ...] timestep variants per token
        # (1-based index). The output layer always needs the expanded
        # embedded_timestep; the eager blocks also need expanded temb
        # (the compiled stack selects per token internally instead).
        idx1 <- cond_token_index$to(dtype = torch::torch_long())$add(1L)$squeeze(1L)
        embedded_timestep <- embedded_timestep$index_select(2L, idx1)
        if (!use_jit) {
            temb <- temb$index_select(2L, idx1)
        }
    }
    eager_blocks <- if (use_jit) integer(0) else seq_along(self$transformer_blocks)
    for (block_idx in eager_blocks) {
        is_stg_block <- (block_idx - 1L) %in% stg_blocks # reference indices are 0-based
        res <- self$transformer_blocks[[block_idx]](
                                      hidden_states = hidden_states,
                                      audio_hidden_states = audio_hidden_states,
                                      encoder_hidden_states = encoder_hidden_states,
                                      audio_encoder_hidden_states = audio_encoder_hidden_states,
                                      temb = temb,
                                      temb_audio = temb_audio,
                                      temb_ca_scale_shift = video_ca_scale_shift,
                                      temb_ca_audio_scale_shift = audio_ca_scale_shift,
                                      temb_ca_gate = video_ca_a2v_gate,
                                      temb_ca_audio_gate = audio_ca_v2a_gate,
                                      temb_prompt = temb_prompt,
                                      temb_prompt_audio = temb_prompt_audio,
                                      video_rotary_emb = video_rotary_emb,
                                      audio_rotary_emb = audio_rotary_emb,
                                      ca_video_rotary_emb = video_ca_rotary_emb,
                                      ca_audio_rotary_emb = audio_ca_rotary_emb,
                                      encoder_attention_mask = encoder_attention_mask,
                                      audio_encoder_attention_mask = audio_encoder_attention_mask,
                                      self_attention_mask = video_self_attention_mask,
                                      use_a2v_cross_attention = !isolate_modalities,
                                      use_v2a_cross_attention = !isolate_modalities,
                                      perturbation_mask = if (is_stg_block) perturbation_mask else NULL,
                                      all_perturbed = if (is_stg_block) all_perturbed else FALSE
        )
        hidden_states <- res[[1]]
        audio_hidden_states <- res[[2]]
        if (isTRUE(getOption("diffuseR.debug"))) {
            ms <- torch::cuda_memory_stats()
            message(sprintf("    block %d: %.2f GB allocated, %.2f GB reserved",
                            block_idx,
                            ms$allocated_bytes$all$current / 1e9,
                            ms$reserved_bytes$all$current / 1e9))
        }
        if (block_gc) {
            # Only needed when the quantized linears allocate per call
            # (buffered scratch makes per-block garbage negligible)
            gc(verbose = FALSE)
        }
    }

    # 6. Output layers
    scale_shift_values <- self$scale_shift_table$unsqueeze(1L)$unsqueeze(1L) +
    embedded_timestep$unsqueeze(3L)
    shift <- scale_shift_values[,, 1,]
    scale <- scale_shift_values[,, 2,]
    hidden_states <- self$norm_out(hidden_states)
    hidden_states <- hidden_states * scale$add(1) + shift
    output <- self$proj_out(hidden_states)

    audio_scale_shift_values <- self$audio_scale_shift_table$unsqueeze(1L)$unsqueeze(1L) +
    audio_embedded_timestep$unsqueeze(3L)
    audio_shift <- audio_scale_shift_values[,, 1,]
    audio_scale <- audio_scale_shift_values[,, 2,]
    audio_hidden_states <- self$audio_norm_out(audio_hidden_states)
    audio_hidden_states <- audio_hidden_states * audio_scale$add(1) + audio_shift
    audio_output <- self$audio_proj_out(audio_hidden_states)

    list(sample = output, audio_sample = audio_output)
}
)

#' Map an official DiT checkpoint key to the R module name
#'
#' Applies the official-to-diffusers renames for the LTX-2.3 transformer
#' (cf. diffusers scripts/convert_ltx2_to_diffusers.py). Our module tree
#' matches the diffusers names, so this is the full mapping.
#'
#' @param key Character. Checkpoint key (with or without the
#'   \code{model.diffusion_model.} prefix).
#'
#' @return Character. Module parameter/buffer name.
#'
#' @export
ltx23_map_dit_key <- function(key) {
    key <- sub("^model\\.diffusion_model\\.", "", key)

    # Substring renames; *_adaln_single variants must precede the bare
    # adaln_single prefix renames below
    key <- gsub("av_ca_video_scale_shift_adaln_single",
                "av_cross_attn_video_scale_shift", key, fixed = TRUE)
    key <- gsub("av_ca_a2v_gate_adaln_single", "av_cross_attn_video_a2v_gate", key, fixed = TRUE)
    key <- gsub("av_ca_audio_scale_shift_adaln_single", "av_cross_attn_audio_scale_shift", key, fixed = TRUE)
    key <- gsub("av_ca_v2a_gate_adaln_single", "av_cross_attn_audio_v2a_gate", key, fixed = TRUE)
    key <- gsub("scale_shift_table_a2v_ca_video", "video_a2v_cross_attn_scale_shift_table", key, fixed = TRUE)
    key <- gsub("scale_shift_table_a2v_ca_audio", "audio_a2v_cross_attn_scale_shift_table", key, fixed = TRUE)
    key <- gsub("prompt_adaln_single", "prompt_adaln", key, fixed = TRUE)
    key <- sub("^audio_adaln_single\\.", "audio_time_embed.", key)
    key <- sub("^adaln_single\\.", "time_embed.", key)
    key <- gsub("patchify_proj", "proj_in", key, fixed = TRUE)
    key <- gsub("q_norm", "norm_q", key, fixed = TRUE)
    key <- gsub("k_norm", "norm_k", key, fixed = TRUE)
    key
}

# serve.R
# Minimal HTTP server exposing diffuseR over OpenAI-style endpoints.
# Built on base R sockets (serverSocket/socketAccept) so it adds no
# dependencies and runs as a single persistent process: the model loads
# once and stays resident (no fork, so the CUDA context is never
# invalidated). Requests are served one at a time - the natural fit for
# a single-GPU box. Same skeleton as whisper::serve / chatterbox::serve
# (cornball serve ports: whisper 7809, chatterbox 7810, qwen3 7811,
# diffuseR 7812).

# (Per-prompt caches live in server state, created per serve() call:
# a bounded LRU of connector outputs (~9 MB each), never the raw
# Gemma stacks (~0.4 GB each) - a persistent server must not grow
# with prompt count.)

#' Serve diffuseR over HTTP
#'
#' Starts a blocking HTTP server that loads one model and answers
#' OpenAI-style generation requests. Never downloads weights: if the
#' model's artifacts are missing, startup stops with the loader's
#' pointer to the explicit \code{download_*()} function.
#'
#' Endpoints:
#' \itemize{
#'   \item \code{GET /health} - liveness probe, returns
#'     \code{{"status":"ok","model":...}}. The server is
#'     single-threaded, so health only answers between requests.
#'   \item \code{POST /v1/images/generations} - image models. JSON body
#'     \code{{prompt, size, seed, steps}} (\code{size} like
#'     \code{"1024x1024"}; \code{n > 1} is not supported). Returns
#'     \code{{created, data: [{b64_json}]}} with a base64 PNG.
#'   \item \code{POST /v1/videos/generations} - \code{model = "ltx"}
#'     only. JSON body \code{{prompt, width, height, num_frames,
#'     frame_rate, seed}}. Returns raw \code{video/mp4} bytes. Video
#'     generation takes minutes: give your client a matching timeout.
#' }
#'
#' The server is single-threaded and runs until interrupted. Run it
#' under a process supervisor (systemd, tmux); an example unit ships
#' with the package: \code{system.file("diffuser.service",
#' package = "diffuseR")}.
#'
#' Security: base R's \code{serverSocket} binds all interfaces, so the
#' server is reachable by anything that can reach the machine. Keep it
#' behind a firewall or reverse proxy, and/or set \code{token}: when
#' set, every request must carry \code{Authorization: Bearer <token>}
#' or it is refused with 401. Generation size is capped by
#' \code{max_pixels}/\code{max_frames}; oversized requests get 400. A
#' CUDA out-of-memory during a request answers 500 and then exits the
#' process (status 70) so a supervisor restarts it with clean GPU
#' state rather than serving on with stranded components.
#'
#' @param port Integer. TCP port. Default 7812 (cornball serve range: whisper 7809, chatterbox 7810, qwen3 TTS 7811).
#' @param model One of "flux2", "zimage", "flux1" (images) or "ltx"
#'   (video). SD 2.1/SDXL are not served yet.
#' @param device Character. "cuda" or "cpu".
#' @param token Character or NULL. Shared secret; when set, requests
#'   must send \code{Authorization: Bearer <token>}.
#' @param max_pixels Integer. Maximum width x height accepted (images
#'   and video frames). Default 1024^2.
#' @param max_frames Integer. Maximum video frame count. Default 161.
#' @param max_steps Integer. Maximum image inference steps. Default 50.
#' @param max_pixel_frames Numeric. Joint video budget: width x height
#'   x frames must stay under it (NULL = max_pixels x 121, so full-
#'   resolution clips top out at 121 frames and longer clips must
#'   shrink spatially).
#' @param max_prompts Integer. Bound on the per-prompt connector-embed
#'   cache for "ltx" (~9 MB per entry, LRU-evicted). Default 32.
#' @param timeout Integer. Per-connection I/O timeout in seconds.
#' @param max_body Integer. Maximum request body bytes. Default 1 MB
#'   (bodies are JSON).
#' @param warmup Logical. Image models: run one small generation at
#'   startup so the first request doesn't pay tracing and allocator
#'   growth. Ignored for "ltx".
#'
#' @return Does not return normally; runs until interrupted.
#' @export
serve <- function(port = 7812L, model = c("flux2", "zimage", "flux1", "ltx"),
                  device = "cuda", token = NULL, max_pixels = 1024L ^ 2,
                  max_frames = 161L, max_steps = 50L,
                  max_pixel_frames = NULL, max_prompts = 32L, timeout = 300L,
                  max_body = 1024L ^ 2, warmup = TRUE) {
    model <- match.arg(model)
    if (is.null(max_pixel_frames)) {
        # Joint video budget: full max_pixels only up to 121 frames;
        # longer clips must shrink spatially
        max_pixel_frames <- as.numeric(max_pixels) * 121
    }

    message("Loading ", model, " on ", device, " (no downloads; run the ",
            "download_*() function first if artifacts are missing)...")
    state <- .dserve_load(model, device, max_prompts = max_prompts)
    state$token <- token
    state$max_pixels <- as.integer(max_pixels)
    state$max_frames <- as.integer(max_frames)
    state$max_steps <- as.integer(max_steps)
    state$max_pixel_frames <- as.numeric(max_pixel_frames)
    message("Model loaded.")

    if (isTRUE(warmup) && model != "ltx") {
        message("Warming up...")
        tryCatch({
            invisible(state$generate("a red square", width = 256L,
                                     height = 256L, seed = 1L))
            message("Warmup done.")
        }, error = function(e) {
            # A CUDA OOM here strands GPU components before the socket
            # even opens: fail startup so the supervisor sees it
            if (grepl("CUDA out of memory", conditionMessage(e),
                      fixed = TRUE)) {
                stop("warmup hit CUDA out of memory; refusing to serve ",
                     "with stranded GPU state: ", conditionMessage(e),
                     call. = FALSE)
            }
            message("warmup skipped: ", conditionMessage(e))
        })
    }

    srv <- serverSocket(port)
    on.exit(close(srv), add = TRUE)
    message("diffuseR::serve listening on port ", port, " (interrupt to stop)")

    repeat {
        con <- tryCatch(
                        socketAccept(srv, blocking = TRUE, open = "r+b", timeout = timeout),
                        error = function(e) {
            message("accept error: ", conditionMessage(e))
            Sys.sleep(0.5)
            NULL
        }
        )
        if (is.null(con)) {
            next
        }
        resp <- NULL
        tryCatch({
            req <- .dserve_read_request(con, max_body)
            if (!is.null(req)) {
                resp <- tryCatch(
                                 .dserve_route(req, state),
                                 error = function(e) {
                    r <- .dserve_err(500L, conditionMessage(e))
                    if (grepl("CUDA out of memory",
                              conditionMessage(e),
                              fixed = TRUE)) {
                        attr(r, "fatal") <- TRUE
                    }
                    r
                }
                )
                .dserve_send(con, resp$status, resp$content_type, resp$body)
            }
        },
                 error = function(e) message("request error: ", conditionMessage(e)),
                 finally = {
            try(close(con), silent = TRUE)
            gc(verbose = FALSE) # bound dead handles; keep the warm pool
        })
        if (exists("resp", inherits = FALSE) && isTRUE(attr(resp, "fatal"))) {
            message("CUDA out of memory: exiting for a clean supervisor restart")
            quit(save = "no", status = 70L)
        }
    }
}

# Load the served model once; return generate closures + metadata
.dserve_load <- function(model, device, max_prompts = 32L) {
    if (model == "ltx") {
        data_dir <- tools::R_user_dir("diffuseR", "data")
        # Serve a COMPLETE artifact only (a partial/interrupted quantize
        # must not shadow a valid one), and when both precisions are
        # built, follow the machine-aware recommendation - on a card
        # too small for resident nf4, recommend() prescribes streamed
        # fp8 (which is also what download_ltx2() produces by default)
        valid_artifact <- function(d) {
            mp <- file.path(d, "manifest.json")
            file.exists(mp) && isTRUE(tryCatch({
                m <- jsonlite::fromJSON(mp)
                length(m$shards) > 0 &&
                all(file.exists(file.path(d, m$shards)))
            }, error = function(e) FALSE))
        }
        dirs <- c(nf4 = file.path(data_dir, "ltx2.3-nf4"),
                  fp8 = file.path(data_dir, "ltx2.3-fp8"))
        avail <- dirs[vapply(dirs, valid_artifact, logical(1))]
        if (!length(avail)) {
            stop("No complete LTX-2.3 artifact under ", data_dir,
                 "; run download_ltx2() (fp8) or ltx23_quantize_nf4() first.",
                 call. = FALSE)
        }
        pick <- if (length(avail) == 1L) {
            avail[[1]]
        } else {
            prec <- tryCatch(recommend("ltx")$precision,
                             error = function(e) "nf4")
            if (prec %in% names(avail)) {
                avail[[prec]]
            } else {
                avail[[1]]
            }
        }
        pipe <- ltx23_load_pipeline(pick, device = device, verbose = FALSE)
        te_dir <- file.path(data_dir, "gemma3-nf4")
        if (!dir.exists(te_dir)) {
            stop("Gemma3 NF4 artifact not found at ", te_dir,
                 "; run gemma3_quantize_nf4() first (serving loads the ",
                 "quantized encoder, not the raw fp32 weights).",
                 call. = FALSE)
        }
        if (!requireNamespace("hfhub", quietly = TRUE)) {
            stop("Serving ltx requires the hfhub package (tokenizer files).",
                 call. = FALSE)
        }
        tok_dir <- dirname(hfhub::hub_download("Lightricks/LTX-2",
                "tokenizer/tokenizer.json", local_files_only = TRUE))
        te <- load_gemma3_text_encoder(te_dir, device = "cpu", verbose = FALSE)
        tok <- gemma3_tokenizer(tok_dir)
        enc_dev <- if (identical(device, "cuda") &&
            torch::cuda_is_available()) {
            "cuda"
        } else {
            "cpu"
        }
        # Bounded LRU of CONNECTOR outputs (~9 MB each): the raw Gemma
        # stack (~0.4 GB) is encoded, projected, and dropped per prompt
        cache <- new.env(parent = emptyenv())
        order <- character(0)
        embeds <- function(prompt) {
            key <- paste0("p:", prompt)
            e <- cache[[key]]
            if (!is.null(e)) {
                order <<- c(setdiff(order, key), key)
                return(e)
            }
            raw <- encode_with_gemma3(prompt, model = te, tokenizer = tok,
                                      max_sequence_length = 1024L,
                                      device = enc_dev, verbose = FALSE)
            conn <- torch::with_no_grad(pipe$connectors(
                    raw$prompt_embeds$to(device = "cpu",
                        dtype = torch::torch_bfloat16()),
                    raw$prompt_attention_mask$to(device = "cpu")))
            e <- list(video_text_embedding = conn$video_text_embedding,
                      audio_text_embedding = conn$audio_text_embedding,
                      attention_mask = conn$attention_mask)
            rm(raw, conn)
            cache[[key]] <- e
            order <<- c(order, key)
            while (length(order) > max_prompts) {
                rm(list = order[[1]], envir = cache)
                order <<- order[-1]
            }
            gc(verbose = FALSE)
            e
        }
        list(model = model, video = TRUE, pipe = pipe, embeds = embeds,
             device = device)
    } else {
        loader <- switch(model, flux1 = flux_load_pipeline,
                         flux2 = flux2_load_pipeline,
                         zimage = zimage_load_pipeline)
        pipe <- loader(device = device, verbose = FALSE)
        genfn <- switch(model, flux1 = txt2img_flux, flux2 = txt2img_flux2,
                        zimage = txt2img_zimage)
        generate <- function(prompt, width, height, seed = NULL, steps = NULL) {
            args <- list(prompt = prompt, pipeline = pipe,
                         width = as.integer(width),
                         height = as.integer(height), seed = seed,
                         save_file = FALSE, verbose = FALSE)
            if (!is.null(steps)) {
                args$num_inference_steps <- as.integer(steps)
            }
            res <- do.call(genfn, args)
            # The generators return list(image, metadata)
            if (is.list(res) && !is.null(res$image)) {
                res$image
            } else {
                res
            }
        }
        list(model = model, video = FALSE, generate = generate, device = device)
    }
}

.dserve_route <- function(req, state) {
    if (isTRUE(req$too_large)) {
        return(.dserve_err(413L, "request body too large"))
    }
    if (!is.null(state$token)) {
        auth <- req$headers[["authorization"]]
        if (!identical(auth, paste("Bearer", state$token))) {
            return(.dserve_err(401L, "unauthorized"))
        }
    }
    path <- sub("\\?.*$", "", req$path)

    if (identical(req$method, "GET") && path == "/health") {
        return(.dserve_json(list(status = "ok", model = state$model)))
    }
    if (identical(req$method, "POST") && path == "/v1/images/generations") {
        if (isTRUE(state$video)) {
            return(.dserve_err(400L,
                               "this server hosts a video model; POST /v1/videos/generations"))
        }
        return(.dserve_image(req, state))
    }
    if (identical(req$method, "POST") && path == "/v1/videos/generations") {
        if (!isTRUE(state$video)) {
            return(.dserve_err(400L,
                               "this server hosts an image model; POST /v1/images/generations"))
        }
        return(.dserve_video(req, state))
    }
    .dserve_err(404L, "not found")
}

# One length-1 atomic value or the default; JSON arrays/objects sent
# for scalar fields become NA so validation answers 400, not 500
.dserve_scalar <- function(x, default = NULL) {
    if (is.null(x)) {
        return(default)
    }
    if (!is.atomic(x) || length(x) != 1L) {
        return(NA)
    }
    x
}

.dserve_prompt_ok <- function(p) {
    is.atomic(p) && length(p) == 1L && is.character(p) && !is.na(p) &&
    nzchar(p)
}

.dserve_body <- function(req) {
    if (!length(req$body)) {
        return(NULL)
    }
    tryCatch(jsonlite::fromJSON(rawToChar(req$body), simplifyVector = TRUE),
             error = function(e) NULL)
}

.dserve_image <- function(req, state) {
    body <- .dserve_body(req)
    if (is.null(body) || !.dserve_prompt_ok(body$prompt)) {
        return(.dserve_err(400L,
                           "body must be JSON with a single string prompt"))
    }
    n <- .dserve_scalar(body$n, 1L)
    if (is.na(suppressWarnings(as.integer(n))) || as.integer(n) > 1L) {
        return(.dserve_err(400L, "n > 1 is not supported"))
    }
    size <- .dserve_scalar(body$size, "1024x1024")
    if (!is.character(size) || is.na(size)) {
        return(.dserve_err(400L, "size must look like 1024x1024"))
    }
    wh <- suppressWarnings(as.integer(strsplit(size, "x", fixed = TRUE)[[1]]))
    if (length(wh) != 2L || anyNA(wh)) {
        return(.dserve_err(400L, "size must look like 1024x1024"))
    }
    if (anyNA(wh) || any(wh < 16L) || wh[1] * wh[2] > state$max_pixels) {
        return(.dserve_err(400L, sprintf(
                    "request exceeds limits (max %d pixels, min side 16)",
                    state$max_pixels)))
    }
    steps <- .dserve_scalar(body$steps)
    if (!is.null(steps)) {
        steps <- suppressWarnings(as.integer(steps))
        if (is.na(steps) || steps < 1L || steps > state$max_steps) {
            return(.dserve_err(400L, sprintf(
                        "steps must be between 1 and %d", state$max_steps)))
        }
    }
    seed <- .dserve_scalar(body$seed)
    if (!is.null(seed)) {
        seed <- suppressWarnings(as.integer(seed))
        if (is.na(seed)) {
            return(.dserve_err(400L, "seed must be a single integer"))
        }
    }
    img <- state$generate(body$prompt, width = wh[1], height = wh[2],
                          seed = seed, steps = steps)
    png <- png::writePNG(img)
    .dserve_json(list(
                      created = as.integer(Sys.time()),
                      data = list(list(b64_json = jsonlite::base64_enc(png)))
        ))
}

.dserve_video <- function(req, state) {
    body <- .dserve_body(req)
    if (is.null(body) || !.dserve_prompt_ok(body$prompt)) {
        return(.dserve_err(400L,
                           "body must be JSON with a single string prompt"))
    }
    w <- suppressWarnings(as.integer(.dserve_scalar(body$width, 768L)))
    h <- suppressWarnings(as.integer(.dserve_scalar(body$height, 512L)))
    nf <- suppressWarnings(as.integer(.dserve_scalar(body$num_frames, 121L)))
    fr <- suppressWarnings(as.numeric(.dserve_scalar(body$frame_rate, 24)))
    if (anyNA(c(w, h, nf)) || w < 32L || h < 32L || nf < 9L ||
        w * h > state$max_pixels || nf > state$max_frames ||
        as.numeric(w) * h * nf > state$max_pixel_frames) {
        return(.dserve_err(400L, sprintf(
                    "request exceeds limits (max %d pixels, %d frames, %.0f pixel-frames)",
                    state$max_pixels, state$max_frames, state$max_pixel_frames)))
    }
    # frame_rate scales the audio-latent length inversely: bound it
    if (is.na(fr) || fr < 12 || fr > 60) {
        return(.dserve_err(400L, "frame_rate must be between 12 and 60"))
    }
    vseed <- .dserve_scalar(body$seed)
    if (!is.null(vseed)) {
        vseed <- suppressWarnings(as.integer(vseed))
        if (is.na(vseed)) {
            return(.dserve_err(400L, "seed must be a single integer"))
        }
    }
    out <- tempfile(fileext = ".mp4")
    on.exit(unlink(out), add = TRUE)
    txt2vid_ltx2(
                 prompt = body$prompt,
                 pipeline = state$pipe,
                 connector_embeds = state$embeds(body$prompt),
                 width = w,
                 height = h,
                 num_frames = nf,
                 frame_rate = fr,
                 seed = vseed,
                 device = state$device, dtype = "bfloat16",
                 filename = out, verbose = FALSE
    )
    bytes <- readBin(out, "raw", n = file.size(out))
    list(status = 200L, content_type = "video/mp4", body = bytes)
}

# ---- HTTP plumbing (same shape as whisper::serve) ---------------------------

.dserve_read_request <- function(con, max_body) {
    term <- as.raw(c(13L, 10L, 13L, 10L))
    buf <- raw(0)
    max_header <- 65536L
    repeat {
        b <- readBin(con, "raw", n = 1L)
        if (length(b) == 0L) {
            return(NULL)
        }
        buf <- c(buf, b)
        n <- length(buf)
        if (n >= 4L && identical(buf[(n - 3L):n], term)) {
            break
        }
        if (n > max_header) {
            return(NULL)
        }
    }

    lines <- strsplit(rawToChar(buf), "\r\n", fixed = TRUE)[[1]]
    req_line <- strsplit(lines[1], " ", fixed = TRUE)[[1]]
    if (length(req_line) < 2L) {
        return(NULL)
    }
    hdr <- list()
    if (length(lines) > 1L) {
        for (ln in lines[-1L]) {
            if (!nzchar(ln)) {
                next
            }
            pos <- regexpr(":", ln, fixed = TRUE)
            if (pos < 1L) {
                next
            }
            hdr[[tolower(trimws(substr(ln, 1L, pos - 1L)))]] <-
            trimws(substr(ln, pos + 1L, nchar(ln)))
        }
    }
    clen <- suppressWarnings(as.integer(hdr[["content-length"]] %||% "0"))
    if (length(clen) != 1L || is.na(clen) || clen < 0L) {
        clen <- 0L
    }
    if (clen > max_body) {
        return(list(method = req_line[1], path = req_line[2], headers = hdr,
                    body = raw(0), too_large = TRUE))
    }
    body <- raw(0)
    while (length(body) < clen) {
        chunk <- readBin(con, "raw", n = clen - length(body))
        if (length(chunk) == 0L) {
            break
        }
        body <- c(body, chunk)
    }
    list(method = req_line[1], path = req_line[2], headers = hdr, body = body)
}

.dserve_send <- function(con, status, content_type, body) {
    if (is.character(body)) {
        body <- charToRaw(enc2utf8(body))
    }
    reason <- switch(as.character(status), "200" = "OK",
                     "400" = "Bad Request", "401" = "Unauthorized",
                     "404" = "Not Found", "405" = "Method Not Allowed",
                     "413" = "Payload Too Large",
                     "500" = "Internal Server Error", "Unknown")
    head <- paste0(
                   sprintf("HTTP/1.1 %d %s\r\n", status, reason),
                   sprintf("Content-Type: %s\r\n", content_type),
                   sprintf("Content-Length: %d\r\n", length(body)),
                   "Connection: close\r\n\r\n")
    writeBin(c(charToRaw(head), body), con)
    flush(con)
}

.dserve_json <- function(x, status = 200L) {
    list(status = status, content_type = "application/json",
         body = jsonlite::toJSON(x, auto_unbox = TRUE))
}

.dserve_err <- function(status, msg) {
    .dserve_json(list(error = list(message = msg)), status = status)
}

# serve(): HTTP parser, router, auth, and limits - pure R, no sockets,
# no models (the generate closure is faked).

library(diffuseR)

mkreq <- function(method, path, body = "", headers = character(0)) {
  raw_body <- charToRaw(body)
  # NB: paste0(character(0), "\r\n", collapse = "") is "\r\n", not ""
  # (with collapse set, zero-length args become "") - hence the guard
  hdr_block <- if (length(headers)) {
    paste0(paste0(headers, "\r\n"), collapse = "")
  } else {
    ""
  }
  head <- paste0(method, " ", path, " HTTP/1.1\r\n",
                 "Content-Length: ", length(raw_body), "\r\n",
                 hdr_block, "\r\n")
  con <- rawConnection(c(charToRaw(head), raw_body), open = "rb")
  on.exit(close(con))
  diffuseR:::.dserve_read_request(con, max_body = 1024L^2)
}
jbody <- function(resp) jsonlite::fromJSON(as.character(resp$body), simplifyVector = FALSE)

# --- parser ------------------------------------------------------------------------

req <- mkreq("GET", "/health")
expect_equal(req$method, "GET")
expect_equal(req$path, "/health")
expect_equal(length(req$body), 0L)

req <- mkreq("POST", "/x", body = '{"a":1}',
             headers = "Content-Type: application/json")
expect_equal(rawToChar(req$body), '{"a":1}')
expect_equal(req$headers[["content-type"]], "application/json")

# too-large bodies are flagged, not read. (Assign the raw first:
# rawConnection uses deparse(substitute(object)) as the connection
# name, and a long inline expression deparses to length > 1.)
big_bytes <- charToRaw(paste0(
  "POST /x HTTP/1.1\r\nContent-Length: 100\r\n\r\n",
  paste(rep("a", 100), collapse = "")))
con <- rawConnection(big_bytes, open = "rb")
big <- diffuseR:::.dserve_read_request(con, max_body = 4L)
close(con)
expect_true(isTRUE(big$too_large))

# --- router: health, 404, auth -----------------------------------------------------

state <- list(model = "fake", video = FALSE,
              max_pixels = 1024L^2, max_frames = 161L)

r <- diffuseR:::.dserve_route(mkreq("GET", "/health"), state)
expect_equal(r$status, 200L)
expect_equal(jbody(r)$model, "fake")

expect_equal(diffuseR:::.dserve_route(mkreq("GET", "/nope"), state)$status,
             404L)

auth_state <- state
auth_state$token <- "s3cret"
expect_equal(diffuseR:::.dserve_route(mkreq("GET", "/health"),
                                      auth_state)$status, 401L)
ok <- diffuseR:::.dserve_route(
  mkreq("GET", "/health", headers = "Authorization: Bearer s3cret"),
  auth_state)
expect_equal(ok$status, 200L)

# --- image handler: caps, validation, b64 round trip -------------------------------

state$generate <- function(prompt, width, height, seed = NULL, steps = NULL) {
  array(runif(height * width * 3), dim = c(height, width, 3L))
}

bad <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/images/generations", body = '{"size":"64x64"}'), state)
expect_equal(bad$status, 400L)   # missing prompt

over <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/images/generations",
        body = '{"prompt":"x","size":"4096x4096"}'), state)
expect_equal(over$status, 400L)
expect_true(grepl("exceeds limits", jbody(over)$error$message))

good <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/images/generations",
        body = '{"prompt":"x","size":"64x64","seed":1}'), state)
expect_equal(good$status, 200L)
png_bytes <- jsonlite::base64_dec(jbody(good)$data[[1]]$b64_json)
expect_identical(png_bytes[1:4], as.raw(c(0x89, 0x50, 0x4e, 0x47)))
img <- png::readPNG(png_bytes)
expect_equal(dim(img)[1:2], c(64L, 64L))

# wrong endpoint for the hosted model type
expect_equal(diffuseR:::.dserve_route(
  mkreq("POST", "/v1/videos/generations", body = '{"prompt":"x"}'),
  state)$status, 400L)

# --- video caps (no model needed: limits check precedes generation) ----------------

vstate <- list(model = "fake", video = TRUE,
               max_pixels = 1024L^2, max_frames = 161L)
overv <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/videos/generations",
        body = '{"prompt":"x","width":2048,"height":2048}'), vstate)
expect_equal(overv$status, 400L)
overf <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/videos/generations",
        body = '{"prompt":"x","num_frames":9999}'), vstate)
expect_equal(overf$status, 400L)

# --- steps, frame_rate, and joint-budget limits ------------------------------------

state$max_steps <- 50L
overstep <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/images/generations",
        body = '{"prompt":"x","size":"64x64","steps":5000}'), state)
expect_equal(overstep$status, 400L)
expect_true(grepl("steps", jbody(overstep)$error$message))

okstep <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/images/generations",
        body = '{"prompt":"x","size":"64x64","steps":4}'), state)
expect_equal(okstep$status, 200L)

vstate$max_steps <- 50L
vstate$max_pixel_frames <- 1024^2 * 121
badfr <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/videos/generations",
        body = '{"prompt":"x","frame_rate":1}'), vstate)
expect_equal(badfr$status, 400L)
expect_true(grepl("frame_rate", jbody(badfr)$error$message))

# joint budget: each dimension legal alone, product over budget
joint <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/videos/generations",
        body = '{"prompt":"x","width":1024,"height":1024,"num_frames":161}'),
  vstate)
expect_equal(joint$status, 400L)
expect_true(grepl("pixel-frames", jbody(joint)$error$message))

tiny <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/videos/generations",
        body = '{"prompt":"x","width":8,"height":8}'), vstate)
expect_equal(tiny$status, 400L)

# --- scalar validation: JSON arrays for scalar fields are 400s ---------------------

arrprompt <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/images/generations",
        body = '{"prompt":["a","b"],"size":"64x64"}'), state)
expect_equal(arrprompt$status, 400L)

arrsize <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/images/generations",
        body = '{"prompt":"x","size":["64x64","128x128"]}'), state)
expect_equal(arrsize$status, 400L)

arrseed <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/images/generations",
        body = '{"prompt":"x","size":"64x64","seed":[1,2]}'), state)
expect_equal(arrseed$status, 400L)

arrw <- diffuseR:::.dserve_route(
  mkreq("POST", "/v1/videos/generations",
        body = '{"prompt":"x","width":[512,640]}'), vstate)
expect_equal(arrw$status, 400L)

# Structural translation audit for the FLUX port, using treesitR.
#
# Parses the Python reference (diffusers / transformers, Apache-2.0) and
# the R port, extracts numeric literals and torch-op call names per
# pairing, and reports the set differences. The point is a short,
# reviewable drift report - not a zero-diff goal: R-side plumbing
# (seq_len, message, ...) and Python-side plumbing (isinstance, super,
# ...) are filtered, the rest is eyeballed.
#
# Run from the package root:  r tools/compare_translation.R
# Requires: treesitR, the ref/ symlink, and tools/cache/modeling_t5.py
# (curl it from transformers v4.49.0 if absent).

library(treesitR)

DIFFUSERS <- "ref/upstream/diffusers/src/diffusers"

# ---- tree helpers -------------------------------------------------------------

parse_file <- function(path, language) {
    p <- ts_parser_new()
    ts_parser_set_language(p, language)
    src <- paste(readLines(path, warn = FALSE), collapse = "\n")
    ts_tree_root_node(ts_parse(p, src))
}

walk_collect <- function(node, fn) {
    acc <- list()
    recurse <- function(n) {
        r <- fn(n)
        if (!is.null(r)) {
            acc[[length(acc) + 1L]] <<- r
        }
        for (child in ts_node_children(n, named = TRUE)) {
            recurse(child)
        }
    }
    recurse(node)
    acc
}

# Named scopes (class_definition / function_definition) from a Python tree
py_scopes <- function(root, names) {
    walk_collect(root, function(n) {
        if (ts_node_type(n) %in% c("class_definition", "function_definition")) {
            kids <- ts_node_children(n, named = TRUE)
            if (length(kids) && ts_node_type(kids[[1]]) == "identifier" &&
                ts_node_text(kids[[1]]) %in% names) {
                return(n)
            }
        }
        NULL
    })
}

# Numeric literal values under a node (sign-insensitive)
literals <- function(node) {
    vals <- walk_collect(node, function(n) {
        if (ts_node_type(n) %in% c("integer", "float")) {
            txt <- gsub("L$|_", "", ts_node_text(n))
            v <- suppressWarnings(as.numeric(txt))
            if (!is.na(v)) {
                return(v)
            }
        }
        NULL
    })
    sort(unique(unlist(vals)))
}

# Canonical call names under a node: last identifier of the callee,
# stripped of torch_/nnf_ prefixes
call_names <- function(node) {
    names <- walk_collect(node, function(n) {
        if (ts_node_type(n) != "call") {
            return(NULL)
        }
        callee <- ts_node_children(n, named = TRUE)[[1]]
        if (ts_node_type(callee) %in%
            c("attribute", "extract_operator", "namespace_operator")) {
            kids <- ts_node_children(callee, named = TRUE)
            callee <- kids[[length(kids)]]
        }
        if (ts_node_type(callee) != "identifier") {
            return(NULL)
        }
        sub("^torch_|^nnf_", "", ts_node_text(callee))
    })
    sort(unique(unlist(names)))
}

# Plumbing that legitimately exists on only one side
PY_NOISE <- c(
              "super", "range", "len", "isinstance", "hasattr", "getattr", "print",
              "int", "float", "str", "list", "dict", "tuple", "set", "zip",
              "enumerate", "ValueError", "ImportError", "warning", "warn",
              "deprecate", "items", "keys", "values", "pop", "update", "get",
              "append", "join", "startswith", "endswith", "register_to_config",
              "apply_lora_scale", "maybe_allow_in_graph", "ceil", "is_grad_enabled",
              "_gradient_checkpointing_func", "register_buffer", "ModuleList",
              "Module", "signature", "is_torch_npu_available",
              "maybe_adjust_dtype_for_device", "dispatch_attention_fn",
              "set_processor", "processor", "retrieve_timesteps", "register_modules",
              "randn_tensor", "postprocess", "numpy", "maybe_free_model_hooks",
              "MultiPipelineCallbacks", "PipelineCallback", "progress_bar"
)
R_NOISE <- c(
             "c", "list", "seq_len", "seq_along", "lapply", "vapply", "function",
             "paste0", "paste", "message", "stop", "warning", "sprintf", "length",
             "names", "file.path", "is.null", "invisible", "structure", "inherits",
             "getOption", "options", "requireNamespace", "Sys.time", "difftime",
             "Sys.getenv", "Sys.setenv", "close", "txtProgressBar",
             "setTxtProgressBar", "nn_module", "nn_module_list", "gc", "rm",
             "isTRUE", "as.integer", "as.numeric", "nzchar", "nchar", "dirname",
             "path.expand", "file.exists", "dir.exists", "fromJSON", "tryCatch",
             "match.arg", "strrep", "modifyList", "do.call", "Filter", "Negate",
             "grepl", "sub", "gsub", "startsWith", "endsWith", "setdiff", "head",
             "utils", "hub_download", "filename_from_prompt", "save_image",
             "clear_vram", "onload", "offload", "print"
)

report_pair <- function(label, py_nodes, r_files) {
    r_lits <- c()
    r_calls <- c()
    for (f in r_files) {
        root <- parse_file(f, ts_language_r())
        r_lits <- union(r_lits, literals(root))
        r_calls <- union(r_calls, call_names(root))
    }
    py_lits <- sort(unique(unlist(lapply(py_nodes, literals))))
    py_calls <- sort(unique(unlist(lapply(py_nodes, call_names))))

    cat("\n== ", label, " ==\n", sep = "")
    cat("literals only in Python: ",
        paste(setdiff(py_lits, r_lits), collapse = " "), "\n")
    cat("literals only in R:      ",
        paste(setdiff(r_lits, py_lits), collapse = " "), "\n")
    cat("calls only in Python:    ",
        paste(setdiff(setdiff(py_calls, r_calls), PY_NOISE), collapse = " "),
        "\n")
    cat("calls only in R:         ",
        paste(setdiff(setdiff(r_calls, py_calls), R_NOISE), collapse = " "),
        "\n")
}

# ---- pairings -------------------------------------------------------------------

# 1. Transformer stack: blocks, norms, embedders, RoPE
tf <- parse_file(file.path(DIFFUSERS, "models/transformers/transformer_flux.py"),
                 ts_language_python())
norm <- parse_file(file.path(DIFFUSERS, "models/normalization.py"),
                   ts_language_python())
emb <- parse_file(file.path(DIFFUSERS, "models/embeddings.py"),
                  ts_language_python())
act <- parse_file(file.path(DIFFUSERS, "models/attention.py"),
                  ts_language_python())
py_transformer <- c(
                    py_scopes(tf, c("FluxAttnProcessor", "FluxAttention",
                                    "FluxSingleTransformerBlock", "FluxTransformerBlock",
                                    "FluxPosEmbed", "FluxTransformer2DModel",
                                    "_get_qkv_projections")),
                    py_scopes(norm, c("AdaLayerNormZero", "AdaLayerNormZeroSingle",
                                      "AdaLayerNormContinuous")),
                    py_scopes(emb, c("get_1d_rotary_pos_embed", "apply_rotary_emb",
                                     "CombinedTimestepTextProjEmbeddings",
                                     "PixArtAlphaTextProjection", "Timesteps",
                                     "TimestepEmbedding", "get_timestep_embedding")),
                    py_scopes(act, c("FeedForward"))
)
report_pair("transformer stack",
            py_transformer,
            c("R/dit_flux_modules.R", "R/dit_flux.R", "R/rope_flux.R"))

# 2. Pipeline flow
pipe <- parse_file(file.path(DIFFUSERS, "pipelines/flux/pipeline_flux.py"),
                   ts_language_python())
report_pair("pipeline",
            py_scopes(pipe, c("FluxPipeline", "calculate_shift")),
            c("R/txt2img_flux.R"))

# 3. T5 encoder
t5_path <- "tools/cache/modeling_t5.py"
if (file.exists(t5_path)) {
    t5 <- parse_file(t5_path, ts_language_python())
    report_pair("t5 encoder",
                py_scopes(t5, c("T5LayerNorm", "T5DenseGatedActDense",
                                "T5Attention", "T5LayerSelfAttention", "T5LayerFF",
                                "T5Block", "T5Stack", "T5EncoderModel")),
                c("R/t5_text_encoder.R"))
} else {
    cat("\n(t5 encoder pairing skipped: tools/cache/modeling_t5.py missing)\n")
}

# 4. FLUX.2 transformer stack + pipeline
tf2 <- parse_file(file.path(DIFFUSERS, "models/transformers/transformer_flux2.py"),
                  ts_language_python())
report_pair("flux2 transformer",
            py_scopes(tf2, c("Flux2SwiGLU", "Flux2FeedForward", "Flux2AttnProcessor",
                             "Flux2Attention", "Flux2ParallelSelfAttnProcessor",
                             "Flux2ParallelSelfAttention", "Flux2SingleTransformerBlock",
                             "Flux2TransformerBlock", "Flux2Modulation", "Flux2PosEmbed",
                             "Flux2TimestepGuidanceEmbeddings", "Flux2Transformer2DModel")),
            c("R/dit_flux2_modules.R", "R/dit_flux2.R", "R/rope_flux2.R"))

pipe2 <- parse_file(file.path(DIFFUSERS, "pipelines/flux2/pipeline_flux2_klein.py"),
                    ts_language_python())
report_pair("flux2 klein pipeline",
            py_scopes(pipe2, c("Flux2KleinPipeline", "compute_empirical_mu")),
            c("R/txt2img_flux2.R", "R/vae_flux2.R", "R/rope_flux2.R"))

# 5. Z-Image transformer stack + pipeline
tfz <- parse_file(file.path(DIFFUSERS, "models/transformers/transformer_z_image.py"),
                  ts_language_python())
report_pair("zimage transformer",
            py_scopes(tfz, c("TimestepEmbedder", "ZSingleStreamAttnProcessor",
                             "FeedForward", "ZImageTransformerBlock", "FinalLayer",
                             "RopeEmbedder", "ZImageTransformer2DModel")),
            c("R/dit_zimage_modules.R", "R/dit_zimage.R", "R/rope_zimage.R"))

pipez <- parse_file(file.path(DIFFUSERS, "pipelines/z_image/pipeline_z_image.py"),
                    ts_language_python())
report_pair("zimage pipeline",
            py_scopes(pipez, c("ZImagePipeline", "get_default_z_image_sigmas")),
            c("R/txt2img_zimage.R"))

# 6. Qwen3 encoder (modular reference; requires tools/cache/modular_qwen3.py)
q3_path <- "tools/cache/modular_qwen3.py"
if (file.exists(q3_path)) {
    q3 <- parse_file(q3_path, ts_language_python())
    report_pair("qwen3 encoder",
                py_scopes(q3, c("Qwen3MLP", "Qwen3Attention", "Qwen3DecoderLayer",
                                "Qwen3RMSNorm", "Qwen3Model")),
                c("R/qwen3_text_encoder.R"))
} else {
    cat("\n(qwen3 pairing skipped: tools/cache/modular_qwen3.py missing)\n")
}

#' Native SDXL pipeline from diffusers safetensors
#'
#' Assemble and run the native SDXL pipeline directly from a HuggingFace
#' diffusers directory (\code{unet/}, \code{vae/}, \code{text_encoder/},
#' \code{text_encoder_2/}), with no TorchScript \code{.pt} step - so it
#' works on Blackwell and loads the same weights everyone else uses. The
#' counterpart to \code{\link{sd_pipeline_from_safetensors}}, adding the
#' second text encoder (OpenCLIP ViT-bigG) and the added-conditioning
#' embeddings (pooled \code{text_embeds} + \code{time_ids}) SDXL needs.
#'
#' @name sdxl_pipeline_safetensors
NULL

.sdxl_repo <- "cornball-ai/sdxl-R"

# Read a diffusers VAE's scaling_factor from its config.json (SDXL is
# 0.13025, distinct from SD 2.1's 0.18215); fall back to `default` if the
# config is missing the field.
.vae_scaling_factor <- function(vae_dir, default = 0.13025) {
    cfg <- file.path(path.expand(vae_dir), "config.json")
    if (!file.exists(cfg)) {
        return(default)
    }
    j <- jsonlite::fromJSON(cfg, simplifyVector = TRUE)
    sf <- j$scaling_factor
    if (is.null(sf) || !is.numeric(sf)) {
        default
    } else {
        as.numeric(sf)
    }
}

#' Assemble a native SDXL pipeline from a diffusers safetensors directory
#'
#' Builds the two native CLIP text encoders, the native SDXL UNet, and the
#' VAE decode module from a diffusers directory using the
#' \code{*_from_safetensors} constructors, places each on its component
#' device, and returns the list the \code{\link{txt2img_sdxl}} denoise
#' loop expects.
#'
#' Both encoders return their penultimate hidden state (SDXL feeds the
#' UNet the concatenated \code{[text_encoder (768) | text_encoder_2
#' (1280)]} = 2048-dim penultimate embeds); \code{text_encoder} uses
#' \code{quick_gelu} (OpenAI CLIP ViT-L) and \code{text_encoder_2} uses
#' exact GELU (OpenCLIP bigG). The pooled \code{text_embeds} come from
#' \code{text_encoder_2}'s full stack. The VAE decodes in float32 (the
#' SDXL fp16 VAE overflows in fp16) and its \code{scaling_factor} is read
#' from \code{vae/config.json}.
#'
#' @param diffusers_dir Directory with \code{unet/}, \code{vae/},
#'   \code{text_encoder/}, \code{text_encoder_2/} subdirectories.
#' @param devices Named list of component devices (\code{unet},
#'   \code{decoder}, \code{text_encoder}, \code{text_encoder2});
#'   defaults to all-CPU. \code{text_encoder2} defaults to the
#'   \code{text_encoder} device when unset.
#' @param unet_dtype A torch dtype for the UNet (default float16 on CUDA,
#'   float32 on CPU).
#' @param verbose Logical.
#'
#' @return A list with \code{unet}, \code{decoder}, \code{text_encoder},
#'   \code{text_encoder2}, \code{vae_scaling}, and \code{native_decode}
#'   (\code{TRUE}; the decoder already applies \code{post_quant_conv}).
#'
#' @export
sdxl_pipeline_from_safetensors <- function(diffusers_dir, devices = NULL,
    unet_dtype = NULL, verbose = TRUE) {
    diffusers_dir <- path.expand(diffusers_dir)
    if (is.null(devices)) {
        devices <- list(unet = "cpu", decoder = "cpu", text_encoder = "cpu",
                        text_encoder2 = "cpu")
    }
    if (is.null(devices$text_encoder2)) {
        devices$text_encoder2 <- devices$text_encoder
    }
    if (is.null(unet_dtype)) {
        unet_dtype <- if (identical(devices$unet, "cuda")) {
            torch::torch_float16()
        } else {
            torch::torch_float32()
        }
    }

    if (verbose) {
        message("Building text encoder 1 (CLIP ViT-L)...")
    }
    # CLIP ViT-L: quick_gelu, penultimate hidden state, no final LN.
    text_encoder <- text_encoder_native_from_safetensors(
        file.path(diffusers_dir, "text_encoder"),
        apply_final_ln = FALSE, return_penultimate = TRUE,
        gelu_type = "quick", verbose = FALSE)
    text_encoder$to(device = torch::torch_device(devices$text_encoder),
                    dtype = torch::torch_float32())

    if (verbose) {
        message("Building text encoder 2 (OpenCLIP bigG)...")
    }
    # OpenCLIP bigG: penultimate hidden state + pooled text_embeds.
    text_encoder2 <- text_encoder2_native_from_safetensors(
        file.path(diffusers_dir, "text_encoder_2"),
        return_penultimate = TRUE, verbose = FALSE)
    text_encoder2$to(device = torch::torch_device(devices$text_encoder2),
                     dtype = torch::torch_float32())

    if (verbose) {
        message("Building UNet...")
    }
    unet <- unet_sdxl_native_from_safetensors(file.path(diffusers_dir, "unet"),
        verbose = FALSE)
    unet$to(device = torch::torch_device(devices$unet), dtype = unet_dtype)

    if (verbose) {
        message("Building VAE decoder...")
    }
    # decode = post_quant_conv + decoder, in float32 (the SDXL fp16 VAE
    # overflows in fp16; up-casting the stored fp16 weights is lossless).
    decoder <- .sd_vae_decode_from_safetensors(file.path(diffusers_dir, "vae"),
        latent_channels = 4L)
    decoder$to(device = torch::torch_device(devices$decoder),
               dtype = torch::torch_float32())

    vae_scaling <- .vae_scaling_factor(file.path(diffusers_dir, "vae"),
                                       default = 0.13025)

    list(unet = unet, decoder = decoder, text_encoder = text_encoder,
         text_encoder2 = text_encoder2, vae_scaling = vae_scaling,
         native_decode = TRUE)
}

# Hosted on the cornball-ai/sdxl-R dataset under diffusers/. fp16, sub-2 GB
# per file (the 5 GB UNet is re-sharded via reshard_safetensors so it is
# CRAN-safetensors readable). The native UNet constructor uses the SDXL
# defaults; only the two CLIP encoders and the VAE need a config.json.
.sdxl_files <- c("diffusers/vae/config.json",
                 "diffusers/vae/diffusion_pytorch_model.safetensors",
                 "diffusers/text_encoder/config.json",
                 "diffusers/text_encoder/model.safetensors",
                 "diffusers/text_encoder_2/config.json",
                 "diffusers/text_encoder_2/model.safetensors",
                 "diffusers/unet/config.json",
                 "diffusers/unet/diffusion_pytorch_model.safetensors.index.json")

#' Download the Stable Diffusion XL diffusers weights
#'
#' Fetches the UNet (re-sharded to sub-2 GB shards), VAE, and both CLIP
#' text encoders from the \code{cornball-ai/sdxl-R} HuggingFace dataset
#' (fp16 diffusers safetensors, converted from the original
#' \code{stabilityai/stable-diffusion-xl-base-1.0} OpenRAIL++ weights).
#' About 7 GB, one-time. The native tokenizer and DDIM scheduler need no
#' downloads.
#'
#' @param verbose Logical.
#'
#' @return Invisibly, the diffusers directory (the parent of
#'   \code{unet/}, \code{vae/}, \code{text_encoder/},
#'   \code{text_encoder_2/}).
#'
#' @export
download_sdxl <- function(verbose = TRUE) {
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to download model weights.")
    }
    have <- .hub_all_cached(.sdxl_repo, .sdxl_files, repo_type = "dataset")
    if (!have) {
        ok <- .ltx23_consent(
                             "Stable Diffusion XL UNet + VAE + 2 CLIP text encoders (~7 GB)")
        if (!ok) {
            stop("Download cancelled.", call. = FALSE)
        }
        if (verbose) {
            message("Downloading Stable Diffusion XL (diffusers safetensors)...")
        }
    }
    paths <- vapply(.sdxl_files, function(f) {
        hfhub::hub_download(.sdxl_repo, f, repo_type = "dataset")
    }, character(1))
    # The UNet ships sharded: fetch every shard named in its index.
    index_path <- paths[[length(.sdxl_files)]]
    idx <- jsonlite::fromJSON(index_path, simplifyVector = TRUE)
    shards <- unique(unlist(idx$weight_map))
    shard_files <- paste0("diffusers/unet/", shards)
    if (!.hub_all_cached(.sdxl_repo, shard_files, repo_type = "dataset") &&
        !.ltx23_consent("the remaining SDXL UNet shards")) {
        stop("Download cancelled.", call. = FALSE)
    }
    for (s in shards) {
        hfhub::hub_download(.sdxl_repo, paste0("diffusers/unet/", s),
                            repo_type = "dataset")
    }
    # diffusers root = parent of vae/ (paths[[1]] is diffusers/vae/config.json)
    diffusers_dir <- dirname(dirname(paths[[1]]))
    if (verbose) {
        message("SDXL ready: ", diffusers_dir)
    }
    invisible(diffusers_dir)
}

#' LTX-2.3 Video VAE Building Blocks
#'
#' Fresh R port of the LTX-2 causal video autoencoder blocks from the
#' diffusers reference (Apache-2.0,
#' src/diffusers/models/autoencoders/autoencoder_kl_ltx2.py). Training
#' and unused inference branches (noise injection, timestep conditioning,
#' plain-conv downsampling) are intentionally not ported; the 2.3
#' checkpoints carry no such weights.
#'
#' @name vae_ltx23_modules
NULL

#' Per-channel RMS normalization
#'
#' Normalizes by the root-mean-square across the channel dimension
#' (dim 2 of [B, C, F, H, W]); no learned parameters.
#'
#' @param eps Numeric. Stability epsilon.
#'
#' @return Module whose forward(x) returns \code{x} divided by its
#'   per-channel root mean square, a tensor of the same shape.
#'
#' @export
ltx23_per_channel_rms_norm <- torch::nn_module(
    "ltx23_per_channel_rms_norm",
    initialize = function(eps = 1e-8) {
    self$eps <- eps
},
    forward = function(x) {
    mean_sq <- torch::torch_mean(x ^ 2, dim = 2L, keepdim = TRUE)
    x$div(mean_sq$add(self$eps)$sqrt())
}
)

#' Causal 3D convolution
#'
#' Spatial padding is handled by the convolution; temporal padding
#' replicates the first frame (causal) or both edge frames (non-causal),
#' chosen at call time.
#'
#' @param in_channels,out_channels Integers.
#' @param kernel_size Integer or length-3 vector (t, h, w).
#' @param stride Integer or length-3 vector.
#' @param spatial_padding_mode Character. Conv padding mode.
#'
#' @return Module whose forward(hidden_states, causal) returns the 3-D
#'   convolution of the input. With \code{causal = TRUE} the temporal
#'   axis is left-padded by replicating the first frame, so no output
#'   frame sees a later input frame.
#'
#' @export
ltx23_causal_conv3d <- torch::nn_module(
                                        "ltx23_causal_conv3d",
                                        initialize = function(
        in_channels,
        out_channels,
        kernel_size = 3L,
        stride = 1L,
        spatial_padding_mode = "zeros"
    ) {
    if (length(kernel_size) == 1L) kernel_size <- rep(kernel_size, 3L)
    if (length(stride) == 1L) stride <- rep(stride, 3L)
    self$kernel_size <- as.integer(kernel_size)

    height_pad <- kernel_size[2] %/% 2L
    width_pad <- kernel_size[3] %/% 2L

    self$conv <- torch::nn_conv3d(in_channels, out_channels,
                                  self$kernel_size,
                                  stride = as.integer(stride),
                                  padding = c(0L, height_pad, width_pad),
                                  padding_mode = spatial_padding_mode)
},
                                        forward = function(hidden_states, causal = TRUE) {
    time_k <- self$kernel_size[1]
    if (time_k > 1L) {
        if (causal) {
            pad_left <- hidden_states$narrow(3L, 1L, 1L)$`repeat`(c(1L, 1L, time_k - 1L, 1L, 1L))
            hidden_states <- torch::torch_cat(list(pad_left, hidden_states), dim = 3L)
        } else {
            half <- (time_k - 1L) %/% 2L
            pad_left <- hidden_states$narrow(3L, 1L, 1L)$`repeat`(c(1L, 1L, half, 1L, 1L))
            pad_right <- hidden_states$narrow(
                3L, hidden_states$shape[3], 1L
            )$`repeat`(c(1L, 1L, half, 1L, 1L))
            hidden_states <- torch::torch_cat(list(pad_left, hidden_states, pad_right), dim = 3L)
        }
    }
    self$conv(hidden_states)
}
)

#' LTX 3D ResNet block
#'
#' PerChannelRMSNorm -> SiLU -> causal conv, twice, with a LayerNorm +
#' 1x1 Conv3d shortcut when the channel count changes.
#'
#' @param in_channels,out_channels Integers.
#' @param eps Numeric. Shortcut LayerNorm epsilon.
#' @param spatial_padding_mode Character.
#'
#' @return Module whose forward(inputs, causal) returns \code{inputs}
#'   plus the residual branch, a tensor of the same shape.
#'
#' @export
ltx23_video_resnet_block3d <- torch::nn_module(
    "ltx23_video_resnet_block3d",
    initialize = function(
                          in_channels,
                          out_channels = NULL,
                          eps = 1e-6,
                          spatial_padding_mode = "zeros"
    ) {
    out_channels <- out_channels %||% in_channels

    self$norm1 <- ltx23_per_channel_rms_norm()
    self$conv1 <- ltx23_causal_conv3d(
                                      in_channels, out_channels, kernel_size = 3L,
                                      spatial_padding_mode = spatial_padding_mode
    )
    self$norm2 <- ltx23_per_channel_rms_norm()
    self$conv2 <- ltx23_causal_conv3d(
                                      out_channels, out_channels, kernel_size = 3L,
                                      spatial_padding_mode = spatial_padding_mode
    )

    if (in_channels != out_channels) {
        self$norm3 <- torch::nn_layer_norm(in_channels, eps = eps,
            elementwise_affine = TRUE)
        # A plain (non-causal) 1x1 Conv3d, per the reference
        self$conv_shortcut <- torch::nn_conv3d(in_channels, out_channels,
            kernel_size = 1L, stride = 1L)
    }
},
    forward = function(inputs, causal = TRUE) {
    hidden_states <- self$norm1(inputs)
    hidden_states <- torch::nnf_silu(hidden_states)
    hidden_states <- self$conv1(hidden_states, causal = causal)

    hidden_states <- self$norm2(hidden_states)
    hidden_states <- torch::nnf_silu(hidden_states)
    hidden_states <- self$conv2(hidden_states, causal = causal)

    if (!is.null(self$norm3)) {
        # LayerNorm over channels: move C last, norm, move back
        inputs <- self$norm3(inputs$permute(c(1L, 3L, 4L, 5L, 2L)))$permute(c(1L, 5L, 2L, 3L, 4L))
    }
    if (!is.null(self$conv_shortcut)) {
        inputs <- self$conv_shortcut(inputs)
    }
    hidden_states + inputs
}
)

#' Pixel-unshuffle 3D downsampler
#'
#' Conv followed by space/time-to-channel rearrangement, plus a grouped
#' channel-mean residual of the same rearrangement.
#'
#' @param in_channels,out_channels Integers.
#' @param stride Length-3 integer vector (t, h, w).
#' @param spatial_padding_mode Character.
#'
#' @return Module whose forward(hidden_states, causal) returns the
#'   space-to-depth downsampled tensor plus its residual: spatial and
#'   temporal extents shrink by \code{stride}, channels grow to match.
#'
#' @export
ltx23_video_downsampler3d <- torch::nn_module(
    "ltx23_video_downsampler3d",
    initialize = function(
                          in_channels,
                          out_channels,
                          stride = c(1L, 1L, 1L),
                          spatial_padding_mode = "zeros"
    ) {
    if (length(stride) == 1L) stride <- rep(stride, 3L)
    self$stride <- as.integer(stride)
    self$group_size <- (in_channels * prod(stride)) %/% out_channels

    conv_out <- out_channels %/% prod(stride)
    self$conv <- ltx23_causal_conv3d(
                                     in_channels, conv_out, kernel_size = 3L, stride = 1L,
                                     spatial_padding_mode = spatial_padding_mode
    )
},
    forward = function(hidden_states, causal = TRUE) {
    s <- self$stride
    if (s[1] > 1L) {
        pad <- hidden_states$narrow(3L, 1L, 1L)$`repeat`(c(1L, 1L, s[1] - 1L, 1L, 1L))
        hidden_states <- torch::torch_cat(list(pad, hidden_states), dim = 3L)
    }

    # Space/time-to-channel rearrangement shared by both paths
    to_channels <- function(x) {
        x <- x$unflatten(5L, c(-1L, s[3]))$unflatten(4L, c(-1L, s[2]))$unflatten(3L,
            c(-1L, s[1]))
        # [B, C, F', st, H', sh, W', sw] -> [B, C, st, sh, sw, F', H', W']
        x$permute(c(1L, 2L, 4L, 6L, 8L, 3L, 5L, 7L))$flatten(start_dim = 2L,
            end_dim = 5L)
    }

    residual <- to_channels(hidden_states)
    residual <- residual$unflatten(2L, c(-1L, self$group_size))$mean(dim = 3L)

    hidden_states <- self$conv(hidden_states, causal = causal)
    hidden_states <- to_channels(hidden_states)
    hidden_states + residual
}
)

#' Pixel-shuffle 3D upsampler
#'
#' Conv followed by channel-to-space/time rearrangement, with an optional
#' channel-repeat residual and an upscale factor that divides the conv
#' output channels.
#'
#' @param in_channels Integer.
#' @param stride Length-3 integer vector (t, h, w).
#' @param residual Logical. Add the rearranged input as a residual.
#' @param upscale_factor Integer.
#' @param spatial_padding_mode Character.
#'
#' @return Module whose forward(hidden_states, causal) returns the
#'   depth-to-space upsampled tensor: spatial and temporal extents grow
#'   by \code{stride}, channels shrink to match.
#'
#' @export
ltx23_video_upsampler3d <- torch::nn_module(
    "ltx23_video_upsampler3d",
    initialize = function(
                          in_channels,
                          stride = c(1L, 1L, 1L),
                          residual = FALSE,
                          upscale_factor = 1L,
                          spatial_padding_mode = "zeros"
    ) {
    if (length(stride) == 1L) stride <- rep(stride, 3L)
    self$stride <- as.integer(stride)
    self$residual <- residual
    self$upscale_factor <- as.integer(upscale_factor)

    out_channels <- (in_channels * prod(stride)) %/% upscale_factor
    self$conv <- ltx23_causal_conv3d(
                                     in_channels, out_channels, kernel_size = 3L, stride = 1L,
                                     spatial_padding_mode = spatial_padding_mode
    )
},
    forward = function(hidden_states, causal = TRUE) {
    s <- self$stride
    dims <- hidden_states$shape
    num_frames <- dims[3]; height <- dims[4]; width <- dims[5]

    # Channel-to-space/time rearrangement shared by both paths
    to_space <- function(x) {
        x <- x$reshape(c(dims[1], -1L, s[1], s[2], s[3], num_frames,
                         height, width))
        # [B, C', st, sh, sw, F, H, W] -> [B, C', F, st, H, sh, W, sw]
        x <- x$permute(c(1L, 2L, 6L, 3L, 7L, 4L, 8L, 5L))
        x <- x$flatten(start_dim = 7L, end_dim = 8L)$
        flatten(start_dim = 5L, end_dim = 6L)$
        flatten(start_dim = 3L, end_dim = 4L)
        # Drop the causally duplicated leading frames
        if (s[1] > 1L) x <- x$narrow(3L, s[1], x$shape[3] - s[1] + 1L) else x
    }

    residual <- NULL
    if (self$residual) {
        residual <- to_space(hidden_states)
        repeats <- prod(s) %/% self$upscale_factor
        residual <- residual$`repeat`(c(1L, repeats, 1L, 1L, 1L))
    }

    hidden_states <- self$conv(hidden_states, causal = causal)
    hidden_states <- to_space(hidden_states)

    if (!is.null(residual)) hidden_states <- hidden_states + residual
    hidden_states
}
)

#' LTX video down block
#'
#' ResNet stack (at the input channel count) followed by a
#' pixel-unshuffle downsampler that also changes the channel count.
#'
#' @param in_channels,out_channels Integers.
#' @param num_layers Integer. ResNet count.
#' @param resnet_eps Numeric.
#' @param spatio_temporal_scale Logical. Whether to downsample at all.
#' @param downsample_type "spatial", "temporal", or "spatiotemporal".
#' @param spatial_padding_mode Character.
#'
#' @return Module whose forward(hidden_states, causal) returns the
#'   stage output: the resnet stack applied in sequence, then the
#'   optional downsampler.
#'
#' @export
ltx23_video_down_block3d <- torch::nn_module(
    "ltx23_video_down_block3d",
    initialize = function(
                          in_channels,
                          out_channels = NULL,
                          num_layers = 1L,
                          resnet_eps = 1e-6,
                          spatio_temporal_scale = TRUE,
                          downsample_type = "spatiotemporal",
                          spatial_padding_mode = "zeros"
    ) {
    out_channels <- out_channels %||% in_channels

    self$resnets <- torch::nn_module_list(lapply(seq_len(num_layers), function(i) {
        ltx23_video_resnet_block3d(in_channels, in_channels,
                                   eps = resnet_eps,
                                   spatial_padding_mode = spatial_padding_mode)
    }))

    if (spatio_temporal_scale) {
        stride <- switch(downsample_type,
                         spatial = c(1L, 2L, 2L),
                         temporal = c(2L, 1L, 1L),
                         spatiotemporal = c(2L, 2L, 2L),
                         stop("Unknown downsample_type: ", downsample_type)
        )
        self$downsamplers <- torch::nn_module_list(list(
                ltx23_video_downsampler3d(
                    in_channels, out_channels, stride = stride,
                    spatial_padding_mode = spatial_padding_mode
                )
            ))
    }
},
    forward = function(hidden_states, causal = TRUE) {
    for (i in seq_along(self$resnets)) {
        hidden_states <- self$resnets[[i]](hidden_states, causal = causal)
    }
    if (!is.null(self$downsamplers)) {
        for (i in seq_along(self$downsamplers)) {
            hidden_states <- self$downsamplers[[i]](hidden_states, causal = causal)
        }
    }
    hidden_states
}
)

#' LTX video mid block
#'
#' A plain ResNet stack at a fixed channel count.
#'
#' @param in_channels Integer.
#' @param num_layers Integer.
#' @param resnet_eps Numeric.
#' @param spatial_padding_mode Character.
#'
#' @return Module whose forward(hidden_states, causal) returns the
#'   bottleneck output, a tensor of the same shape as the input.
#'
#' @export
ltx23_video_mid_block3d <- torch::nn_module(
    "ltx23_video_mid_block3d",
    initialize = function(
                          in_channels,
                          num_layers = 1L,
                          resnet_eps = 1e-6,
                          spatial_padding_mode = "zeros"
    ) {
    self$resnets <- torch::nn_module_list(lapply(seq_len(num_layers), function(i) {
        ltx23_video_resnet_block3d(in_channels, in_channels,
                                   eps = resnet_eps,
                                   spatial_padding_mode = spatial_padding_mode)
    }))
},
    forward = function(hidden_states, causal = TRUE) {
    for (i in seq_along(self$resnets)) {
        hidden_states <- self$resnets[[i]](hidden_states, causal = causal)
    }
    hidden_states
}
)

#' LTX video up block
#'
#' Optional channel-changing conv-in ResNet, pixel-shuffle upsampler,
#' then a ResNet stack at the output channel count.
#'
#' @param in_channels,out_channels Integers.
#' @param num_layers Integer.
#' @param resnet_eps Numeric.
#' @param spatio_temporal_scale Logical.
#' @param upsample_type "spatial", "temporal", or "spatiotemporal".
#' @param upsample_residual Logical.
#' @param upscale_factor Integer.
#' @param spatial_padding_mode Character.
#'
#' @return Module whose forward(hidden_states, causal) returns the
#'   stage output: the optional input projection and upsampler, then the
#'   resnet stack applied in sequence.
#'
#' @export
ltx23_video_up_block3d <- torch::nn_module(
    "ltx23_video_up_block3d",
    initialize = function(
                          in_channels,
                          out_channels = NULL,
                          num_layers = 1L,
                          resnet_eps = 1e-6,
                          spatio_temporal_scale = TRUE,
                          upsample_type = "spatiotemporal",
                          upsample_residual = FALSE,
                          upscale_factor = 1L,
                          spatial_padding_mode = "zeros"
    ) {
    out_channels <- out_channels %||% in_channels

    if (in_channels != out_channels) {
        self$conv_in <- ltx23_video_resnet_block3d(in_channels, out_channels,
            eps = resnet_eps, spatial_padding_mode = spatial_padding_mode)
    }

    if (spatio_temporal_scale) {
        stride <- switch(upsample_type,
                         spatial = c(1L, 2L, 2L),
                         temporal = c(2L, 1L, 1L),
                         spatiotemporal = c(2L, 2L, 2L),
                         stop("Unknown upsample_type: ", upsample_type)
        )
        self$upsamplers <- torch::nn_module_list(list(
                ltx23_video_upsampler3d(
                                        in_channels = out_channels * upscale_factor,
                                        stride = stride,
                                        residual = upsample_residual,
                                        upscale_factor = upscale_factor,
                                        spatial_padding_mode = spatial_padding_mode
                )
            ))
    }

    self$resnets <- torch::nn_module_list(lapply(seq_len(num_layers), function(i) {
        ltx23_video_resnet_block3d(
                                   out_channels, out_channels, eps = resnet_eps,
                                   spatial_padding_mode = spatial_padding_mode
        )
    }))
},
    forward = function(hidden_states, causal = TRUE) {
    if (!is.null(self$conv_in)) {
        hidden_states <- self$conv_in(hidden_states, causal = causal)
    }
    if (!is.null(self$upsamplers)) {
        for (i in seq_along(self$upsamplers)) {
            hidden_states <- self$upsamplers[[i]](hidden_states, causal = causal)
        }
    }
    for (i in seq_along(self$resnets)) {
        hidden_states <- self$resnets[[i]](hidden_states, causal = causal)
    }
    hidden_states
}
)

#' Create a FlowMatch Euler Discrete Scheduler
#'
#' Creates a FlowMatch scheduler for use with flow-matching diffusion models
#' like LTX-2. FlowMatch schedulers use Euler integration for sampling,
#' which is simpler and often faster than DDIM-style schedulers.
#'
#' @param num_train_timesteps Integer. The number of diffusion steps used to
#'   train the model. Default: 1000
#' @param shift Numeric. The shift value for the timestep schedule. Default: 1.0
#' @param use_dynamic_shifting Logical. Whether to apply timestep shifting
#'   on-the-fly based on the image/video resolution. Default: FALSE
#' @param base_shift Numeric. Value to stabilize generation. Increasing
#'   reduces variation. Default: 0.5
#' @param max_shift Numeric. Maximum shift allowed. Increasing encourages
#'   more variation. Default: 1.15
#' @param base_seq_len Integer. Base sequence length for dynamic shifting. Default: 256
#' @param max_seq_len Integer. Maximum sequence length for dynamic shifting. Default: 4096
#' @param invert_sigmas Logical. Whether to invert the sigmas (used by some models
#'   like Mochi). Default: FALSE
#' @param shift_terminal Numeric or NULL. End value of shifted schedule. Default: NULL
#' @param time_shift_type Character. Type of dynamic shifting: "exponential" or
#'   "linear". Default: "exponential"
#'
#' @return A FlowMatch scheduler object (list) containing:
#'   \describe{
#'     \item{sigmas}{The noise schedule}
#'     \item{timesteps}{The timestep schedule}
#'     \item{num_train_timesteps}{Training timesteps}
#'     \item{config}{All configuration parameters}
#'   }
#'
#' @details
#' FlowMatch (Flow Matching) is a framework for training continuous normalizing
#' flows by regressing onto target probability paths. The Euler discrete scheduler
#' implements simple Euler integration for sampling from trained flow models.
#'
#' The core update rule is:
#' \code{prev_sample = sample + dt * model_output}
#' where \code{dt = sigma_next - sigma_current}.
#'
#' @references
#' Lipman, Y., Chen, R. T. Q., Ben-Hamu, H., Nickel, M., & Le, M. (2022).
#' "Flow Matching for Generative Modeling."
#' \url{https://arxiv.org/abs/2210.02747}
#'
#' @examples
#' if (torch::torch_is_installed()) {
#'   scheduler <- flowmatch_scheduler_create(
#'     num_train_timesteps = 1000,
#'     shift = 1.0
#'   )
#'
#'   # Set timesteps for inference
#'   scheduler <- flowmatch_set_timesteps(scheduler, num_inference_steps = 8)
#'   scheduler$timesteps
#' }
#' @export
flowmatch_scheduler_create <- function(num_train_timesteps = 1000L,
                                       shift = 1.0,
                                       use_dynamic_shifting = FALSE,
                                       base_shift = 0.5, max_shift = 1.15,
                                       base_seq_len = 256L,
                                       max_seq_len = 4096L,
                                       invert_sigmas = FALSE,
                                       shift_terminal = NULL,
                                       time_shift_type = c("exponential", "linear")) {
    time_shift_type <- match.arg(time_shift_type)

    # Initial timesteps (reversed, from max to min)
    timesteps <- seq(from = 1, to = num_train_timesteps,
                     length.out = num_train_timesteps)
    timesteps <- rev(timesteps)

    # Initial sigmas (normalized timesteps)
    sigmas <- timesteps / num_train_timesteps

    # Apply shift if not using dynamic shifting
    if (!use_dynamic_shifting) {
        sigmas <- shift * sigmas / (1 + (shift - 1) * sigmas)
    }

    timesteps <- sigmas * num_train_timesteps

    list(
         sigmas = sigmas,
         timesteps = timesteps,
         num_train_timesteps = num_train_timesteps,
         num_inference_steps = NULL,
         step_index = NULL,
         config = list(
                       num_train_timesteps = num_train_timesteps,
                       shift = shift,
                       use_dynamic_shifting = use_dynamic_shifting,
                       base_shift = base_shift,
                       max_shift = max_shift,
                       base_seq_len = base_seq_len,
                       max_seq_len = max_seq_len,
                       invert_sigmas = invert_sigmas,
                       shift_terminal = shift_terminal,
                       time_shift_type = time_shift_type
        )
    )
}

#' Calculate shift for dynamic shifting
#'
#' Computes the shift parameter (mu) based on sequence length for
#' resolution-dependent timestep shifting.
#'
#' @param seq_len Integer. The sequence length (num_patches).
#' @param base_seq_len Integer. Base sequence length. Default: 256
#' @param max_seq_len Integer. Maximum sequence length. Default: 4096
#' @param base_shift Numeric. Base shift value. Default: 0.5
#' @param max_shift Numeric. Maximum shift value. Default: 1.15
#'
#' @return Numeric. The computed shift value (mu).
#' @export
flowmatch_calculate_shift <- function(seq_len, base_seq_len = 256L,
                                      max_seq_len = 4096L, base_shift = 0.5,
                                      max_shift = 1.15) {
    m <- (max_shift - base_shift) / (max_seq_len - base_seq_len)
    b <- base_shift - m * base_seq_len
    mu <- seq_len * m + b
    mu
}

#' Set timesteps for inference
#'
#' Configures the scheduler timesteps for a specific number of inference steps.
#' This must be called before using the scheduler for denoising.
#'
#' @param schedule List. The FlowMatch scheduler object.
#' @param num_inference_steps Integer. Number of denoising steps. Default: 50
#' @param device Character or torch device. Device for tensors. Default: "cpu"
#' @param mu Numeric or NULL. Shift parameter for dynamic shifting. Required
#'   if use_dynamic_shifting is TRUE. Default: NULL
#' @param sigmas Numeric vector or NULL. Custom sigma values. Default: NULL
#' @param timesteps Numeric vector or NULL
#'
#' @return Updated scheduler with configured timesteps and sigmas.
#' @export
flowmatch_set_timesteps <- function(schedule, num_inference_steps = 50L,
                                    device = "cpu", mu = NULL, sigmas = NULL,
                                    timesteps = NULL) {
    config <- schedule$config

    if (config$use_dynamic_shifting && is.null(mu)) {
        stop("`mu` must be provided when use_dynamic_shifting is TRUE")
    }

    schedule$num_inference_steps <- num_inference_steps

    # Get sigma_max and sigma_min
    sigma_max <- max(schedule$sigmas)
    sigma_min <- min(schedule$sigmas)

    # Create sigmas if not provided
    if (is.null(sigmas)) {
        if (is.null(timesteps)) {
            # Linear spacing from sigma_max to sigma_min
            timesteps <- seq(from = sigma_max * config$num_train_timesteps,
                             to = sigma_min * config$num_train_timesteps,
                             length.out = num_inference_steps)
        }
        sigmas <- timesteps / config$num_train_timesteps
    }

    # Apply timestep shifting
    if (config$use_dynamic_shifting) {
        sigmas <- .flowmatch_time_shift(
                                        mu = mu,
                                        sigma = 1.0,
                                        t = sigmas,
                                        shift_type = config$time_shift_type
        )
    } else {
        shift <- config$shift
        sigmas <- shift * sigmas / (1 + (shift - 1) * sigmas)
    }

    # Stretch to terminal if configured
    if (!is.null(config$shift_terminal)) {
        sigmas <- .flowmatch_stretch_shift_to_terminal(sigmas, config$shift_terminal)
    }

    # Convert to tensors
    sigmas <- torch::torch_tensor(sigmas, dtype = torch::torch_float32())
    timesteps <- sigmas * config$num_train_timesteps

    # Append terminal sigma
    if (config$invert_sigmas) {
        sigmas <- 1.0 - sigmas
        timesteps <- sigmas * config$num_train_timesteps
        sigmas <- torch::torch_cat(list(sigmas, torch::torch_ones(1)))
    } else {
        sigmas <- torch::torch_cat(list(sigmas, torch::torch_zeros(1)))
    }

    # Move to device
    sigmas <- sigmas$to(device = device)
    timesteps <- timesteps$to(device = device)

    schedule$sigmas <- sigmas
    schedule$timesteps <- timesteps
    schedule$step_index <- NULL

    schedule
}

#' Perform a FlowMatch scheduler step
#'
#' Performs a single denoising step using Euler integration. This is the
#' core sampling function for FlowMatch models.
#'
#' @param model_output torch tensor. The output from the diffusion model
#'   (velocity prediction).
#' @param timestep Numeric. The current timestep.
#' @param sample torch tensor. The current noisy sample.
#' @param schedule List. The FlowMatch scheduler object.
#' @param generator torch generator or NULL. Random generator for reproducibility.
#'
#' @return A list containing:
#'   \describe{
#'     \item{prev_sample}{The denoised sample at the previous timestep}
#'     \item{schedule}{The updated scheduler with incremented step_index}
#'   }
#'
#' @details
#' The FlowMatch Euler step is remarkably simple:
#' \code{prev_sample = sample + dt * model_output}
#' where \code{dt = sigma_next - sigma_current}.
#'
#' This implements the Euler method for solving the probability flow ODE
#' in continuous normalizing flows.
#'
#' @export
flowmatch_scheduler_step <- function(model_output, timestep, sample,
                                     schedule, generator = NULL) {
    # Initialize step index if needed
    if (is.null(schedule$step_index)) {
        schedule$step_index <- .flowmatch_init_step_index(timestep, schedule)
    }

    # Upcast sample for precision
    sample <- sample$to(dtype = torch::torch_float32())

    # Get current and next sigma
    sigma_idx <- schedule$step_index
    sigma <- schedule$sigmas[sigma_idx]$item()
    sigma_next <- schedule$sigmas[sigma_idx + 1]$item()

    # Compute dt (step size)
    dt <- sigma_next - sigma

    # Euler step: x_{t-1} = x_t + dt * v
    prev_sample <- sample + dt * model_output

    # Cast back to model dtype
    prev_sample <- prev_sample$to(dtype = model_output$dtype)

    # Increment step index
    schedule$step_index <- schedule$step_index + 1L

    list(prev_sample = prev_sample, schedule = schedule)
}

#' Scale noise for flow matching forward process
#'
#' Applies the forward process in flow-matching: interpolates between
#' the clean sample and noise.
#'
#' @param sample torch tensor. The clean sample.
#' @param timestep torch tensor. The current timestep.
#' @param noise torch tensor. The noise tensor.
#' @param schedule List. The FlowMatch scheduler object.
#'
#' @return torch tensor. The noisy sample at timestep t.
#' @export
flowmatch_scale_noise <- function(sample, timestep, noise, schedule) {
    # Get sigma for this timestep
    sigmas <- schedule$sigmas$to(device = sample$device, dtype = sample$dtype)

    step_indices <- .flowmatch_index_for_timestep(timestep, schedule)
    sigma <- sigmas[step_indices]

    # Expand sigma dimensions to match sample
    while (length(sigma$shape) < length(sample$shape)) {
        sigma <- sigma$unsqueeze(-1)
    }

    # Flow matching interpolation: x_t = (1 - sigma) * x_0 + sigma * noise
    noisy_sample <- (1.0 - sigma) * sample + sigma * noise

    noisy_sample
}

# Internal helper functions

.flowmatch_time_shift <- function(mu, sigma, t, shift_type) {
    if (shift_type == "exponential") {
        exp(mu) / (exp(mu) + (1 / t - 1) ^ sigma)
    } else {
        # linear
        mu / (mu + (1 / t - 1) ^ sigma)
    }
}

.flowmatch_stretch_shift_to_terminal <- function(t, shift_terminal) {
    one_minus_z <- 1 - t
    scale_factor <- one_minus_z[length(one_minus_z)] / (1 - shift_terminal)
    stretched_t <- 1 - (one_minus_z / scale_factor)
    stretched_t
}

.flowmatch_init_step_index <- function(timestep, schedule) {
    # Find index for this timestep
    idx <- .flowmatch_index_for_timestep(timestep, schedule)
    idx
}

.flowmatch_index_for_timestep <- function(timestep, schedule) {
    timesteps <- schedule$timesteps

    if (inherits(timestep, "torch_tensor")) {
        timestep <- timestep$item()
    }

    # Find matching index
    indices <- which(abs(as.numeric(timesteps) - timestep) < 1e-6)

    if (length(indices) == 0) {
        # Find closest
        indices <- which.min(abs(as.numeric(timesteps) - timestep))
    }

    # Return first match (or second if multiple, for numerical stability)
    if (length(indices) > 1) {
        indices[2]
    } else {
        indices[1]
    }
}

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

# Validation tests for FlowMatch scheduler against Python diffusers
# These tests ensure numerical equivalence with HuggingFace implementation

# Load Python test cases
test_cases_path <- system.file("validation/flowmatch_test_cases.json",
                                package = "diffuseR")

if (!file.exists(test_cases_path)) {
  # Skip if validation file not found (CI/CRAN)
  cat("Skipping Python validation tests - test cases file not found\n")
} else {
  test_cases <- jsonlite::fromJSON(test_cases_path)

  # Tolerance for numerical comparison
  tol <- 1e-4

  # Test 1: Basic 8-step scheduler sigmas match Python
  cat("Test 1: Basic 8-step scheduler sigmas\n")
  schedule <- flowmatch_scheduler_create(
    num_train_timesteps = 1000,
    shift = 1.0
  )
  schedule <- flowmatch_set_timesteps(schedule, num_inference_steps = 8)

  python_sigmas <- test_cases$basic_8_steps$sigmas
  r_sigmas <- as.numeric(schedule$sigmas)

  expect_equal(length(r_sigmas), length(python_sigmas),
               info = "Sigma lengths should match")

  max_diff <- max(abs(r_sigmas - python_sigmas))
  expect_true(max_diff < tol,
              info = sprintf("Max sigma diff %.6f exceeds tolerance %.6f",
                           max_diff, tol))
  cat(sprintf("  Max sigma diff: %.6e (tolerance: %.6e)\n", max_diff, tol))

  # Test 2: Shifted scheduler (shift=3.0, 20 steps)
  cat("Test 2: Shifted scheduler (shift=3.0)\n")
  schedule_shifted <- flowmatch_scheduler_create(
    num_train_timesteps = 1000,
    shift = 3.0
  )
  schedule_shifted <- flowmatch_set_timesteps(schedule_shifted,
                                               num_inference_steps = 20)

  python_sigmas_shifted <- test_cases$shifted_20_steps$sigmas
  r_sigmas_shifted <- as.numeric(schedule_shifted$sigmas)

  max_diff_shifted <- max(abs(r_sigmas_shifted - python_sigmas_shifted))
  expect_true(max_diff_shifted < tol,
              info = sprintf("Max shifted sigma diff %.6f exceeds tolerance",
                           max_diff_shifted))
  cat(sprintf("  Max sigma diff: %.6e\n", max_diff_shifted))

  # Test 3: Timesteps match
  cat("Test 3: Timesteps match\n")
  python_timesteps <- test_cases$basic_8_steps$timesteps
  r_timesteps <- as.numeric(schedule$timesteps)

  max_timestep_diff <- max(abs(r_timesteps - python_timesteps))
  expect_true(max_timestep_diff < tol,
              info = sprintf("Max timestep diff %.6f exceeds tolerance",
                           max_timestep_diff))
  cat(sprintf("  Max timestep diff: %.6e\n", max_timestep_diff))

  # Test 4: Step function output matches
  cat("Test 4: Step function numerical match\n")
  schedule <- flowmatch_scheduler_create(num_train_timesteps = 1000, shift = 1.0)
  schedule <- flowmatch_set_timesteps(schedule, num_inference_steps = 8)

  # Use same seed as Python
  torch::torch_manual_seed(42)
  sample <- torch::torch_randn(c(1, 4, 8, 16, 16))
  model_output <- torch::torch_randn(c(1, 4, 8, 16, 16))
  timestep <- schedule$timesteps[1]$item()

  result <- flowmatch_scheduler_step(
    model_output = model_output,
    timestep = timestep,
    sample = sample,
    schedule = schedule
  )
  prev_sample <- result$prev_sample

  python_step <- test_cases$step_test

  r_mean <- as.numeric(prev_sample$mean())
  r_std <- as.numeric(prev_sample$std())
  r_min <- as.numeric(prev_sample$min())
  r_max <- as.numeric(prev_sample$max())

  # Note: Due to different RNG implementations, exact match is unlikely
  # But dt and sigma values should match exactly
  expect_true(abs(python_step$sigma_current - 1.0) < tol,
              info = "Sigma current should be 1.0")
  expect_true(abs(python_step$sigma_next - as.numeric(schedule$sigmas[2])) < tol,
              info = "Sigma next should match")

  cat(sprintf("  Python mean: %.6f, R mean: %.6f\n",
              python_step$prev_sample_mean, r_mean))
  cat(sprintf("  dt: %.6f (expected: %.6f)\n",
              as.numeric(schedule$sigmas[2]) - as.numeric(schedule$sigmas[1]),
              python_step$dt))

  # Test 5: Full loop with deterministic scaling
  cat("Test 5: Full denoising loop pattern\n")
  schedule <- flowmatch_scheduler_create(num_train_timesteps = 1000, shift = 1.0)
  schedule <- flowmatch_set_timesteps(schedule, num_inference_steps = 8)

  # Initialize with seed
  torch::torch_manual_seed(123)
  latents <- torch::torch_randn(c(1, 4, 4, 16, 16))

  # Run full loop
  for (i in seq_along(schedule$timesteps)) {
    t <- schedule$timesteps[i]$item()
    model_output <- latents * 0.1  # Same as Python test
    result <- flowmatch_scheduler_step(
      model_output = model_output,
      timestep = t,
      sample = latents,
      schedule = schedule
    )
    latents <- result$prev_sample
    schedule <- result$schedule
  }

  final_mean <- as.numeric(latents$mean())
  final_std <- as.numeric(latents$std())

  python_final <- test_cases$full_loop

  # The pattern should be similar even if RNG differs
  # Check that values are in reasonable range
  expect_true(abs(final_std) < 2.0,
              info = "Final std should be reasonable")
  cat(sprintf("  Python final: mean=%.6f, std=%.6f\n",
              python_final$final_mean, python_final$final_std))
  cat(sprintf("  R final:      mean=%.6f, std=%.6f\n", final_mean, final_std))

  cat("\nAll Python validation tests completed\n")
}

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

# Tests for FlowMatch Euler Discrete Scheduler

# Test scheduler creation
expect_silent({
  schedule <- flowmatch_scheduler_create()
})

expect_equal(schedule$config$num_train_timesteps, 1000L)
expect_equal(schedule$config$shift, 1.0)
expect_false(schedule$config$use_dynamic_shifting)
expect_equal(length(schedule$sigmas), 1000)

# Test with custom parameters
schedule2 <- flowmatch_scheduler_create(
  num_train_timesteps = 500,
  shift = 2.0,
  use_dynamic_shifting = FALSE
)
expect_equal(schedule2$config$num_train_timesteps, 500)
expect_equal(schedule2$config$shift, 2.0)

# Test calculate_shift
mu <- flowmatch_calculate_shift(
  seq_len = 1024,
  base_seq_len = 256,
  max_seq_len = 4096,
  base_shift = 0.5,
  max_shift = 1.15
)
expect_true(is.numeric(mu))
expect_true(mu > 0.5 && mu < 1.15)

# Test set_timesteps
schedule <- flowmatch_scheduler_create()
schedule <- flowmatch_set_timesteps(schedule, num_inference_steps = 8)

expect_equal(schedule$num_inference_steps, 8)
expect_true(inherits(schedule$sigmas, "torch_tensor"))
expect_true(inherits(schedule$timesteps, "torch_tensor"))
# Should have num_inference_steps + 1 sigmas (with terminal)
expect_equal(as.numeric(schedule$sigmas$shape[1]), 9)

# Test scheduler step
if (torch::cuda_is_available()) {
  device <- "cuda"
} else {
  device <- "cpu"
}

schedule <- flowmatch_scheduler_create()
schedule <- flowmatch_set_timesteps(schedule, num_inference_steps = 8, device = device)

# Create dummy sample and model output
sample <- torch::torch_randn(c(1, 4, 8, 16, 16), device = device)
model_output <- torch::torch_randn(c(1, 4, 8, 16, 16), device = device)
timestep <- schedule$timesteps[1]$item()

# Perform step
result <- flowmatch_scheduler_step(
  model_output = model_output,
  timestep = timestep,
  sample = sample,
  schedule = schedule
)
prev_sample <- result$prev_sample
schedule <- result$schedule

expect_true(inherits(prev_sample, "torch_tensor"))
expect_equal(as.numeric(prev_sample$shape), c(1, 4, 8, 16, 16))

# Test that step_index advances
expect_equal(schedule$step_index, 2L)

# Test scale_noise (forward process)
clean_sample <- torch::torch_randn(c(1, 4, 8, 16, 16), device = device)
noise <- torch::torch_randn(c(1, 4, 8, 16, 16), device = device)
timestep_tensor <- torch::torch_tensor(schedule$timesteps[1]$item(), device = device)

noisy_sample <- flowmatch_scale_noise(
  sample = clean_sample,
  timestep = timestep_tensor,
  noise = noise,
  schedule = schedule
)

expect_true(inherits(noisy_sample, "torch_tensor"))
expect_equal(as.numeric(noisy_sample$shape), c(1, 4, 8, 16, 16))

# Test that shifted schedules differ from unshifted
schedule_shifted <- flowmatch_scheduler_create(shift = 3.0)
schedule_unshifted <- flowmatch_scheduler_create(shift = 1.0)

expect_false(all(schedule_shifted$sigmas == schedule_unshifted$sigmas))

cat("All FlowMatch scheduler tests passed\n")

#' models2devices
#' @description This function sets up the model directory, device configuration, and data types for diffusion models.
#' It checks the validity of the model name and devices, detects model type, and downloads the model if necessary.
#'
#' @param model_name A character string representing the name of the model to be used.
#' @param devices A character string or a named list specifying the devices for different components of the model.
#' @param unet_dtype_str A character string specifying the data type for the UNet model.
#' @param download_models Logical indicating whether to download models if they are not found.
#'
#' @return A list containing the device configuration, UNet data type, and CPU/CUDA devices.
#' @export
models2devices <- function(model_name, devices = "cpu",
                           unet_dtype_str = NULL, download_models = FALSE) {
    # Validation (same as before)
    if (is.null(model_name) || !is.character(model_name)) {
        stop("Invalid model name")
    }

    if (is.null(devices) || (!is.character(devices) && !is.list(devices))) {
        stop("Invalid devices parameter")
    }

    if (!is.null(unet_dtype_str) && !is.character(unet_dtype_str)) {
        stop("Invalid unet_dtype_str parameter")
    }

    # Get required components for this model type
    required_components <- get_required_components(model_name)

    # Standardize the 'devices' parameter
    devices <- standardize_devices(devices, required_components)

    # Set up dtype
    unet_dtype <- setup_dtype(devices, unet_dtype_str)

    device_cpu <- torch::torch_device("cpu")
    device_cuda <- torch::torch_device("cuda")

    # Verify model files are available (downloads if allowed)
    download_model(model_name, devices, unet_dtype_str,
                   download_models = download_models)

    return(list(
                devices = devices,
                unet_dtype = unet_dtype,
                device_cpu = device_cpu,
                device_cuda = device_cuda
        ))
}

#' Get required components for each model type
#' @description This function returns a list of required components for each supported model type.
#' @param model_name A character string representing the name of the model.
#' @return A character vector of required components for the specified model.
get_required_components <- function(model_name) {
    components <- list(
                       # "sd15" = c("unet", "decoder", "text_encoder", "encoder"),
                       "sd21" = c("unet", "decoder", "text_encoder", "encoder"),
                       "sdxl" = c("unet", "decoder", "text_encoder", "text_encoder2",
                                  "encoder"),
                       "flux1" = c("transformer", "decoder", "text_encoder", "text_encoder2"),
                       "flux2" = c("transformer", "decoder", "text_encoder"),
                       "zimage" = c("transformer", "decoder", "text_encoder")
                       # "sd3" = c("transformer", "decoder", "text_encoder", "text_encoder2", "text_encoder3", "encoder"),
                       # "cascade" = c("prior", "decoder", "text_encoder", "vqgan")
    )

    if (!model_name %in% names(components)) {
        stop("Unsupported model type: ", model_name)
    }

    return(components[[model_name]])
}

#' Standardize devices configuration
#' @description This function standardizes the device configuration for model components.
#' It checks if the devices parameter is a single string or a named list, and fills in missing components with reasonable defaults.
#' @param devices A character string or a named list specifying the devices for model components.
#' @param required_components A character vector of required components for the model.
#' @return A named list of devices for each required component.
standardize_devices <- function(devices, required_components) {
    if (is.character(devices) && length(devices) == 1) {
        # Single device string - apply to all components
        device_list <- as.list(rep(devices, length(required_components)))
        names(device_list) <- required_components
        return(device_list)
    }

    if (is.list(devices)) {
        # Check for required components
        missing_components <- setdiff(required_components, names(devices))

        if (length(missing_components) > 0) {
            # Try to fill in missing components with reasonable defaults
            for (component in missing_components) {
                if (component == "encoder" && "decoder" %in% names(devices)) {
                    devices$encoder <- devices$decoder
                } else if (component == "text_encoder2" &&
                    "text_encoder" %in% names(devices)) {
                    devices$text_encoder2 <- devices$text_encoder
                    message("text_encoder2 is set to text_encoder")
                } else if (component == "text_encoder3" &&
                    "text_encoder" %in% names(devices)) {
                    devices$text_encoder3 <- devices$text_encoder
                } else if (component == "transformer" &&
                    "unet" %in% names(devices)) {
                    devices$transformer <- devices$unet
                } else {
                    stop("Missing required component: ", component)
                }
            }
        }

        return(devices)
    }
    stop("'devices' must be either a single device string or a named list")
}

#' Set up dtype based on device configuration
#' @param devices A character string or a named list specifying the devices for model components.
#' @param unet_dtype_str A character string specifying the data type for the UNet model.
#' @return A torch dtype object based on the main computation device.
setup_dtype <- function(devices, unet_dtype_str) {
    # Find the main computation device (unet or transformer)
    main_device <- if ("unet" %in% names(devices)) {
        devices$unet
    } else if ("transformer" %in% names(devices)) {
        devices$transformer
    } else {
        stop("No main computation component found")
    }

    if (main_device == "cpu") {
        return(torch::torch_float32())
    } else if (main_device == "cuda") {
        if (is.null(unet_dtype_str)) {
            return(torch::torch_float16())
        } else if (unet_dtype_str == "float16") {
            return(torch::torch_float16())
        } else if (unet_dtype_str == "float32") {
            return(torch::torch_float32())
        } else {
            stop("Invalid dtype: ", unet_dtype_str)
        }
    } else {
        stop("Invalid device: ", main_device)
    }
}

# Test utility functions

# --- filename_from_prompt ---

# Test basic filename generation without datetime
result <- filename_from_prompt("test prompt", datetime = FALSE)
expect_equal(result, "prompt_test_prompt.png")

# Test special characters are replaced with underscores
result <- filename_from_prompt("hello! world?", datetime = FALSE)
expect_equal(result, "prompt_hello__world_.png")

# Test prompt truncation at 50 characters
long_prompt <- paste(rep("a", 100), collapse = "")
result <- filename_from_prompt(long_prompt, datetime = FALSE)
expect_true(nchar(result) <= 62)  # "prompt_" (7) + 50 + ".png" (4)

# Test with datetime (just check format, not exact value)
result <- filename_from_prompt("test", datetime = TRUE)
expect_true(grepl("^\\d{8}_\\d{6}_test\\.png$", result))

# --- get_required_components (internal function) ---

# Test sd21 components
components <- diffuseR:::get_required_components("sd21")
expect_true(is.character(components))
expect_true("unet" %in% components)
expect_true("decoder" %in% components)
expect_true("text_encoder" %in% components)

# Test sdxl components
components <- diffuseR:::get_required_components("sdxl")
expect_true("text_encoder2" %in% components)

# Test invalid model
expect_error(diffuseR:::get_required_components("invalid_model"))

# --- standardize_devices (internal function) ---

# Test single device string expands to all components
required <- c("unet", "decoder", "text_encoder")
result <- diffuseR:::standardize_devices("cpu", required)
expect_true(is.list(result))
expect_equal(result$unet, "cpu")
expect_equal(result$decoder, "cpu")
expect_equal(result$text_encoder, "cpu")

# Test list passthrough
devices <- list(unet = "cuda", decoder = "cpu", text_encoder = "cpu")
result <- diffuseR:::standardize_devices(devices, required)
expect_equal(result$unet, "cuda")

# Test text_encoder2 inherits from text_encoder
devices <- list(unet = "cuda", decoder = "cpu", text_encoder = "cpu", encoder = "cpu")
required_xl <- c("unet", "decoder", "text_encoder", "text_encoder2", "encoder")
result <- diffuseR:::standardize_devices(devices, required_xl)
expect_equal(result$text_encoder2, "cpu")

# Test encoder inherits from decoder
devices <- list(unet = "cuda", decoder = "cpu", text_encoder = "cpu")
required_enc <- c("unet", "decoder", "text_encoder", "encoder")
result <- diffuseR:::standardize_devices(devices, required_enc)
expect_equal(result$encoder, "cpu")

#' Generate an image from a text prompt using a diffusion pipeline
#'
#' This function generates an image based on a text prompt using the Stable Diffusion model.
#' It allows for various configurations such as model name, device, scheduler, and more.
#' @param prompt A character string prompt describing the image to generate.
#' @param negative_prompt Optional negative prompt to guide the generation.
#' @param img_dim Dimension of the output image (e.g., 512 for 512x512).
#' @param pipeline Optional A pre-loaded diffusion pipeline. If `NULL`, it will be loaded based on the model name and devices.
#' @param devices A named list of devices for each model component (e.g., `list(unet = "cuda", decoder = "cpu", text_encoder = "cpu")`).
#' @param unet_dtype_str Optional A character for dtype of the unet component (typically "float16" for cuda and "float32" for cpu; float32 is available for cuda).
#' @param download_models Logical indicating whether to download the model files if they are not found.
#' @param scheduler Scheduler to use (e.g., `"ddim"`, `"euler"`).
#' @param timesteps Optional A vector of timesteps to use.
#' @param initial_latents Optional initial latents for the diffusion process.
#' @param num_inference_steps Number of inference steps to run.
#' @param guidance_scale Scale for classifier-free guidance (typically 7.5).
#' @param seed Optional seed for reproducibility.
#' @param save_file Logical indicating whether to save the generated image.
#' @param filename Optional filename for saving the image. If `NULL`, a default name is generated.
#' @param metadata_path Optional file path to save metadata.
#' @param use_native_decoder Logical; if TRUE, uses native R torch decoder instead of TorchScript.
#'   Native decoder has better GPU compatibility (especially Blackwell).
#' @param use_native_text_encoder Logical; if TRUE, uses native R torch text encoder instead of TorchScript.
#'   Native text encoder has better GPU compatibility (especially Blackwell).
#' @param use_native_unet Logical; if TRUE, uses native R torch UNet instead of TorchScript.
#'   Native UNet has better GPU compatibility (especially Blackwell).
#' @param diffusers_dir Optional path to a HuggingFace diffusers
#'   directory (with `unet/`, `vae/`, `text_encoder/`). When set, the
#'   pipeline is built natively from safetensors (no TorchScript), via
#'   [sd_pipeline_from_safetensors()]. See [download_sd21()].
#' @param ... Additional parameters passed to the diffusion process.
#'
#' @return An image array and metadata
#' @export
#'
#' @examples
#' \dontrun{
#' img <- txt2img("a cat wearing sunglasses in space", device = "cuda")
#' }
txt2img_sd21 <- function(prompt, negative_prompt = NULL, img_dim = 768,
                         pipeline = NULL, devices = "auto",
                         unet_dtype_str = NULL, download_models = FALSE,
                         scheduler = "ddim", timesteps = NULL,
                         initial_latents = NULL, num_inference_steps = 50,
                         guidance_scale = 7.5, seed = NULL, save_file = TRUE,
                         filename = NULL, metadata_path = NULL,
                         use_native_decoder = FALSE,
                         use_native_text_encoder = FALSE,
                         use_native_unet = FALSE, diffusers_dir = NULL, ...) {
    model_name <- "sd21"

    # Handle "auto" devices
    if (identical(devices, "auto")) {
        devices <- auto_devices(model_name)
    }

    device_cpu <- torch::torch_device("cpu")
    device_cuda <- torch::torch_device("cuda")
    if (!is.null(diffusers_dir)) {
        # Native safetensors path: resolve devices/dtype with the pure
        # helpers, skipping the .pt model verification models2devices runs.
        # SD 2.1 attention overflows in float16 (all-NaN output), so
        # default this path to float32 unless float16 is asked for.
        devices <- standardize_devices(devices,
                                       get_required_components(model_name))
        if (is.null(unet_dtype_str)) {
            unet_dtype_str <- "float32"
        }
        unet_dtype <- setup_dtype(devices, unet_dtype_str)
    } else {
        m2d <- models2devices(model_name = model_name, devices = devices,
                              unet_dtype_str = unet_dtype_str,
                              download_models = download_models)
        devices <- m2d$devices
        unet_dtype <- m2d$unet_dtype
        device_cpu <- m2d$device_cpu
        device_cuda <- m2d$device_cuda
    }

    if (is.null(pipeline)) {
        if (!is.null(diffusers_dir)) {
            # Native pipeline straight from HuggingFace diffusers
            # safetensors (no TorchScript .pt); see download_sd21().
            pipeline <- sd_pipeline_from_safetensors(diffusers_dir,
                model_name = model_name, devices = devices,
                unet_dtype = unet_dtype)
        } else {
            pipeline <- load_pipeline(model_name = model_name, m2d = m2d,
                                      unet_dtype_str = unet_dtype_str,
                                      use_native_decoder = use_native_decoder,
                                      use_native_text_encoder = use_native_text_encoder,
                                      use_native_unet = use_native_unet)
        }
    }

    # Start timing
    start_time <- proc.time()

    # Process text prompt
    message("Processing prompt...")
    ## Tokenizer
    tokens <- CLIPTokenizer(prompt)
    # clip-vit-large-patch14
    prompt_embed <- pipeline$text_encoder(tokens)

    if (is.null(negative_prompt)) {
        empty_tokens <- CLIPTokenizer("")
    } else {
        empty_tokens <- CLIPTokenizer(negative_prompt)
    }
    empty_prompt_embed <- pipeline$text_encoder(empty_tokens)

    empty_prompt_embed <- empty_prompt_embed$to(dtype = unet_dtype,
        device = torch::torch_device(devices$unet))
    prompt_embed <- prompt_embed$to(dtype = unet_dtype,
                                    device = torch::torch_device(devices$unet))

    message("Creating schedule...")
    # Load scheduler
    schedule <- ddim_scheduler_create(num_inference_steps = num_inference_steps,
                                      beta_schedule = "scaled_linear",
                                      device = torch::torch_device(devices$unet))
    if (is.null(timesteps)) {
        timesteps <- schedule$timesteps
    }

    # Run diffusion process
    message("Generating image...")
    if (!is.null(seed)) {
        set.seed(seed)
        torch::torch_manual_seed(seed = seed)
    }

    latent_dim <- img_dim / 8
    if (!is.null(initial_latents)) {
        latents <- initial_latents
        latents <- latents$to(dtype = unet_dtype, device = torch::torch_device(devices$unet))
    } else {
        # Create random latents
        latents <- torch::torch_randn(c(1, 4, latent_dim, latent_dim),
                                      dtype = unet_dtype,
                                      device = torch::torch_device(devices$unet))
    }
    # Denoising loop (no gradients needed for inference)
    pb <- utils::txtProgressBar(min = 0, max = length(timesteps), style = 3)
    torch::with_no_grad({
        for (i in seq_along(timesteps)) {
            timestep <- torch::torch_tensor(timesteps[i],
                dtype = torch::torch_long(),
                device = torch::torch_device(devices$unet))

            # Get both conditional and unconditional predictions
            noise_pred_uncond <- pipeline$unet(latents, timestep, empty_prompt_embed)
            noise_pred_cond <- pipeline$unet(latents, timestep, prompt_embed)

            # CFG step
            noise_pred <- noise_pred_uncond + guidance_scale *
            (noise_pred_cond - noise_pred_uncond)

            # Calculating latent
            latents <- ddim_scheduler_step(model_output = noise_pred,
                timestep = timestep,
                sample = latents,
                schedule = schedule,
                prediction_type = "v_prediction",
                device = devices$unet)
            latents <- latents$to(dtype = unet_dtype, device = torch::torch_device(devices$unet))
            utils::setTxtProgressBar(pb, i)
        }
    })
    close(pb)

    # Decode latents to image
    scaled_latent <- latents / 0.18215
    scaled_latent <- scaled_latent$to(dtype = torch::torch_float32(),
                                      device = torch::torch_device(devices$decoder))
    message("Decoding image...")
    decoded_output <- pipeline$decoder(scaled_latent)
    # Ensure tensor is on CPU
    img <- decoded_output$cpu()

    # Remove batch dimension if present
    if (length(img$shape) == 4) {
        img <- img$squeeze(1)
    }

    # Reorder channels: [3, H, W] → [H, W, 3]
    img <- img$permute(c(2, 3, 1))

    # Normalize if needed
    img <- (img + 1) / 2 # scale from [-1, 1] → [0, 1]
    img <- torch::torch_clamp(img, min = 0, max = 1)

    # Convert to R array
    img_array <- as.array(img)
    # Save if requested
    if (save_file) {
        # Creating filename while we're here
        if (is.null(filename)) {
            filename <- filename_from_prompt(prompt, datetime = TRUE)
        }
        message("Saving image to ", filename)
        save_image(img = img_array, filename)
    } else {
        # Display in RStudio Viewer if interactive
        if (interactive()) {
            grid::grid.raster(img_array)
        }
    }
    # Save metadata
    metadata <- list(
                     prompt = prompt,
                     negative_prompt = negative_prompt,
                     width = img_dim,
                     height = img_dim,
                     num_inference_steps = num_inference_steps,
                     guidance_scale = guidance_scale,
                     seed = seed,
                     scheduler = scheduler,
                     model = model_name
    )
    if (!is.null(metadata_path)) {
        utils::write.csv(metadata, file = metadata_path, row.names = FALSE)
        message("Metadata saved to: ", metadata_path)
    }
    # Report timing
    elapsed <- proc.time() - start_time
    message(sprintf("Image generated in %.2f seconds", elapsed[3]))

    # Return the generated image and metadata
    return(list(image = img_array, metadata = metadata))
}

#!/usr/bin/env r
#
# Test INT4 Memory Usage
#
# Verifies that INT4 weights fit on GPU with room for activations.
#

library(torch)
library(diffuseR)

cat("=== INT4 Memory Test ===\n\n")

# Helper to get GPU memory via nvidia-smi
get_gpu_memory <- function() {
  out <- system("nvidia-smi --query-gpu=memory.used --format=csv,noheader,nounits", intern = TRUE)
  as.numeric(out[1]) / 1024  # Convert MB to GB
}

# Clear GPU
gc(full = TRUE)
torch::cuda_empty_cache()

# Initial memory
mem_start <- get_gpu_memory()
cat(sprintf("Initial GPU memory: %.2f GB\n", mem_start))

# Load INT4 weights from disk
cat("\nLoading INT4 weights...\n")
t0 <- Sys.time()
int4_path <- path.expand("~/.cache/diffuseR/ltx2_transformer_int4.safetensors")
q <- load_int4_weights(int4_path, verbose = TRUE)

# Move all INT4 weights to GPU
cat("\nMoving INT4 weights to GPU...\n")
t0 <- Sys.time()
for (name in names(q)) {
  q[[name]]$packed <- q[[name]]$packed$to(device = "cuda")
  q[[name]]$scales <- q[[name]]$scales$to(device = "cuda")
}
elapsed <- as.numeric(Sys.time() - t0)
cat(sprintf("Moved to GPU in %.2fs\n", elapsed))

mem_after_load <- get_gpu_memory()
cat(sprintf("GPU memory after INT4 load: %.2f GB\n", mem_after_load))
cat(sprintf("INT4 weights on GPU: %.2f GB\n", mem_after_load - mem_start))

# Simulate activations for a typical forward pass
cat("\nSimulating activations...\n")

# Typical video generation: 720p, 121 frames, batch 1
# Latent size: 768/32 x 512/32 x (121-1)/8+1 = 24 x 16 x 16
batch_size <- 1L
seq_len <- 24L * 16L * 16L  # 6144 patches
hidden_dim <- 4096L

# Create activation tensors (hidden states, attention, etc.)
hidden_states <- torch::torch_randn(c(batch_size, seq_len, hidden_dim),
                                     dtype = torch::torch_float16(), device = "cuda")
# Attention needs Q, K, V plus output
attn_q <- torch::torch_randn(c(batch_size, 32L, seq_len, 128L),
                              dtype = torch::torch_float16(), device = "cuda")
attn_k <- torch::torch_randn(c(batch_size, 32L, seq_len, 128L),
                              dtype = torch::torch_float16(), device = "cuda")
attn_v <- torch::torch_randn(c(batch_size, 32L, seq_len, 128L),
                              dtype = torch::torch_float16(), device = "cuda")

mem_with_activations <- get_gpu_memory()
cat(sprintf("GPU memory with activations: %.2f GB\n", mem_with_activations))
cat(sprintf("Activation memory: %.2f GB\n", mem_with_activations - mem_after_load))

# Test a forward pass through one layer
cat("\nTesting INT4 forward pass...\n")
weight_name <- "transformer_blocks.0.attn1.to_q.weight"
q_weight <- q[[weight_name]]

# Create INT4 layer and do forward
int4_layer <- int4_linear_from_quantized(q_weight, device = "cuda",
                                          dtype = torch::torch_float16())

x_in <- hidden_states[, 1:128, ]  # Small batch for testing
t0 <- Sys.time()
y_out <- int4_layer(x_in)
elapsed <- as.numeric(Sys.time() - t0) * 1000
cat(sprintf("Forward pass: %.1fms\n", elapsed))

mem_during_forward <- get_gpu_memory()
cat(sprintf("GPU memory during forward: %.2f GB\n", mem_during_forward))

# Cleanup test tensors
rm(hidden_states, attn_q, attn_k, attn_v, x_in, y_out, int4_layer)
gc()
torch::cuda_empty_cache()

mem_after_cleanup <- get_gpu_memory()
cat(sprintf("\nGPU memory after cleanup: %.2f GB\n", mem_after_cleanup))

# Summary
cat("\n=== Summary ===\n")
cat(sprintf("INT4 weights: %.2f GB\n", mem_after_load - mem_start))
cat(sprintf("Peak activations: %.2f GB\n", mem_with_activations - mem_after_load))
cat(sprintf("Peak total: %.2f GB\n", mem_with_activations))
cat(sprintf("Available headroom: %.2f GB (on 16GB GPU)\n", 16 - mem_with_activations))

if (mem_with_activations < 14) {
  cat("\nVERDICT: INT4 model should fit on 16GB GPU with room to spare\n")
} else if (mem_with_activations < 15.5) {
  cat("\nVERDICT: INT4 model fits on 16GB GPU (tight)\n")
} else {
  cat("\nVERDICT: INT4 model may not fit - need optimization\n")
}

cat("\n=== Test Complete ===\n")

#' Download a single TorchScript component via hfhub
#'
#' Internal helper that wraps \code{hfhub::hub_download()} with the cornball-ai
#' dataset repo naming convention. Falls back to legacy \code{R_user_dir()} paths
#' if the file exists there but not yet in the hfhub cache.
#'
#' @param model_name Model name (e.g., \code{"sd21"}, \code{"sdxl"}).
#' @param filename Filename within the repo (e.g., \code{"unet-cpu.pt"}).
#' @param download If \code{TRUE}, download from HuggingFace when not cached.
#'
#' @return The local file path (character string).
#' @keywords internal
hf_download_pt <- function(model_name, filename, download = TRUE) {
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("Package 'hfhub' is required. Install it from CRAN before ",
             "calling download_model() again.")
    }

    repo_id <- paste0("cornball-ai/", model_name, "-R")

    # 1. Try hfhub cache first (no network)
    path <- tryCatch(
                     hfhub::hub_download(repo_id, filename, repo_type = "dataset",
            local_files_only = TRUE),
                     error = function(e) NULL
    )
    if (!is.null(path) && file.exists(path)) {
        return(path)
    }

    # 2. Check legacy R_user_dir location
    legacy_path <- file.path(tools::R_user_dir("diffuseR", "data"),
                             model_name, filename)
    if (file.exists(legacy_path)) {
        return(legacy_path)
    }

    # 3. Download via hfhub
    if (!download) {
        stop("Component '", filename, "' for model '", model_name,
             "' not found locally. Set download=TRUE to download.")
    }

    hfhub::hub_download(repo_id, filename, repo_type = "dataset")
}

#' Download TorchScript model files for Stable Diffusion
#'
#' Downloads the required model files (e.g., UNet, decoder, text encoder)
#' for a given Stable Diffusion model using \code{hfhub::hub_download()}.
#'
#' Files are cached by hfhub (typically \code{~/.cache/huggingface/hub/}).
#' Legacy files in the old \code{R_user_dir()} location are also recognized.
#'
#' @param model_name Name of the model (e.g., "sd21" for stable-diffusion-2-1)
#' @param devices Either a single device string or a named list with elements 'unet', 'decoder', 'text_encoder'; optionally 'encoder'
#' @param unet_dtype_str Optional: "float16" or "float32" (only applies if unet uses CUDA)
#' @param overwrite If TRUE, force re-download of model files
#' @param show_progress Show download progress messages
#' @param download_models If TRUE, download the model files from HuggingFace
#'
#' @return A named list of full file paths, keyed by component name.
#' @export
#'
#' @examples
#' \dontrun{
#' paths <- download_model("sd21")
#' }
#'
download_model <- function(model_name = "sd21",
                           devices = list(unet = "cpu", decoder = "cpu", text_encoder = "cpu"),
                           unet_dtype_str = NULL, overwrite = FALSE,
                           show_progress = TRUE, download_models = FALSE) {
    # Normalize 'devices'
    if (is.character(devices) && length(devices) == 1) {
        if (model_name == "sdxl") {
            devices <- list(unet = devices, decoder = devices,
                            text_encoder = devices, text_encoder2 = devices,
                            encoder = devices)
        } else {
            devices <- list(
                            unet = devices,
                            decoder = devices,
                            text_encoder = devices,
                            encoder = devices
            )
        }
    } else if (is.list(devices)) {
        if (model_name == "sdxl") {
            required_keys <- c("unet", "decoder", "text_encoder", "text_encoder2")
            if ("text_encoder" %in% names(devices) &&
                !("text_encoder2" %in% names(devices))) {
                devices$text_encoder2 <- devices$text_encoder
            }
        } else {
            required_keys <- c("unet", "decoder", "text_encoder")
        }

        missing_keys <- setdiff(required_keys, names(devices))
        if (length(missing_keys) > 0) {
            stop(paste0("Missing required devices: ", paste(missing_keys, collapse = ", ")))
        }

        if (!"encoder" %in% names(devices)) {
            devices$encoder <- devices$decoder
        }
    } else {
        stop("'devices' must be a device string or a named list with required components")
    }

    # Determine component names and filenames
    if (model_name == "sdxl") {
        model_names <- c("unet", "decoder", "text_encoder", "text_encoder2", "encoder")
    } else if (model_name == "sd21") {
        model_names <- c("unet", "decoder", "text_encoder", "encoder")
    } else {
        stop("Unsupported model name: ", model_name)
    }

    filenames <- character(length(model_names))
    for (i in seq_along(model_names)) {
        name <- model_names[i]
        device <- devices[[name]]
        if (is.null(device)) {
            stop("Device not specified for component: ", name)
        }
        if (name == "unet" && device != "cpu") {
            if (is.null(unet_dtype_str)) {
                dtype <- "float16"
            } else {
                dtype <- unet_dtype_str
            }
            if (!dtype %in% c("float16", "float32")) {
                stop("Invalid unet_dtype_str: must be 'float16' or 'float32'")
            }
            filenames[i] <- paste0("unet-cuda-", dtype, ".pt")
        } else {
            filenames[i] <- paste0(name, "-", device, ".pt")
        }
    }

    # Check which files are already available (hfhub cache or legacy)
    already_available <- vapply(filenames, function(f) {
        # Check hfhub cache
        path <- tryCatch(
                         hfhub::hub_download(paste0("cornball-ai/", model_name, "-R"), f,
                repo_type = "dataset", local_files_only = TRUE),
                         error = function(e) NULL
        )
        if (!is.null(path) && file.exists(path)) return(TRUE)
        # Check legacy location
        legacy <- file.path(tools::R_user_dir("diffuseR", "data"), model_name, f)
        file.exists(legacy)
    }, logical(1))

    needs_download <- !already_available | overwrite

    # Consent when files actually need downloading: interactive prompt,
    # or the diffuseR.consent option in non-interactive sessions
    if (download_models && any(needs_download)) {
        if (!.ltx23_consent(paste0("'", model_name,
                                   "' model files from HuggingFace"))) {
            stop("Download cancelled.", call. = FALSE)
        }
    }

    # Resolve all paths (download if needed and allowed)
    model_paths <- character(length(model_names))
    names(model_paths) <- model_names
    for (i in seq_along(model_names)) {
        if (needs_download[i] && !download_models) {
            # Not available and not allowed to download — will error
            model_paths[i] <- tryCatch(
                                       hf_download_pt(model_name, filenames[i], download = FALSE),
                                       error = function(e) NA_character_
            )
        } else {
            model_paths[i] <- tryCatch(
                                       hf_download_pt(model_name, filenames[i], download = download_models),
                                       error = function(e) {
                warning("Download failed for ", filenames[i], ": ", e$message)
                NA_character_
            }
            )
        }
    }

    missing <- model_names[is.na(model_paths)]
    if (length(missing) > 0) {
        stop("Missing model files: ", paste(filenames[is.na(model_paths)], collapse = ", "))
    }

    model_paths
}

#' Native Stable Diffusion pipelines from diffusers safetensors
#'
#' Assemble and run the native SD pipeline directly from a HuggingFace
#' diffusers directory (\code{unet/}, \code{vae/}, \code{text_encoder/}),
#' with no TorchScript \code{.pt} step - so it works on Blackwell and
#' loads the same weights everyone else uses. SD21 is wired end to end
#' here; SDXL still needs its second text encoder and added-conditioning
#' embeddings (tracked in tasks/todo.md).
#'
#' @name sd_pipeline_safetensors
NULL

.sd21_repo <- "cornball-ai/sd21-R"

# The SD/SDXL AutoencoderKL applies a 1x1 post_quant_conv to the latent
# before the decoder (decode(z) = decoder(post_quant_conv(z))). The
# native vae_decoder_native is the decoder submodule only (the FLUX VAE
# has no post_quant_conv), so the SD decode path must apply it - without
# it the decode is badly wrong (cos ~0.5 vs the reference). This wrapper
# restores it; its weights come from the artifact's own
# post_quant_conv.{weight,bias}.
sd_vae_decode <- torch::nn_module(
                                  "SDVAEDecode",
                                  initialize = function(decoder, post_quant_conv) {
    self$post_quant_conv <- post_quant_conv
    self$decoder <- decoder
},
                                  forward = function(z) {
    self$decoder(self$post_quant_conv(z))
}
)

# Build the SD decode module (post_quant_conv + native decoder) from a
# diffusers VAE directory.
.sd_vae_decode_from_safetensors <- function(vae_dir, latent_channels = 4L) {
    decoder <- vae_decoder_native_from_safetensors(vae_dir,
        latent_channels = latent_channels, verbose = FALSE)
    path <- file.path(path.expand(vae_dir),
                      "diffusion_pytorch_model.safetensors")
    h <- safetensors::safetensors$new(path, framework = "torch")
    pqc <- torch::nn_conv2d(latent_channels, latent_channels, kernel_size = 1L)
    torch::with_no_grad({
        pqc$weight$copy_(h$get_tensor("post_quant_conv.weight"))
        pqc$bias$copy_(h$get_tensor("post_quant_conv.bias"))
    })
    m <- sd_vae_decode(decoder, pqc)
    m$eval()
    m
}

# Hosted on the cornball-ai/sd21-R dataset under diffusers/. fp16, sub-2 GB
# per file (CRAN-safetensors readable). No unet/vae config.json: the native
# constructors use the SD 2.1 defaults; only the CLIP encoder needs one.
.sd21_files <- c("diffusers/unet/diffusion_pytorch_model.safetensors",
                 "diffusers/vae/diffusion_pytorch_model.safetensors",
                 "diffusers/text_encoder/config.json",
                 "diffusers/text_encoder/model.safetensors")

#' Download the Stable Diffusion 2.1 diffusers weights
#'
#' Fetches the UNet, VAE, and CLIP text encoder from the
#' \code{cornball-ai/sd21-R} HuggingFace dataset (fp16 diffusers
#' safetensors, converted from the original OpenRAIL weights; the
#' upstream \code{stabilityai} repo was deprecated). About 2.5 GB,
#' one-time. The native tokenizer and DDIM scheduler need no downloads.
#'
#' @param verbose Logical.
#'
#' @return Invisibly, the diffusers directory (the parent of
#'   \code{unet/}, \code{vae/}, \code{text_encoder/}).
#'
#' @export
download_sd21 <- function(verbose = TRUE) {
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to download model weights.")
    }
    have <- .hub_all_cached(.sd21_repo, .sd21_files, repo_type = "dataset")
    if (!have) {
        ok <- .ltx23_consent(
                             "Stable Diffusion 2.1 UNet + VAE + CLIP text encoder (~2.5 GB)"
        )
        if (!ok) {
            stop("Download cancelled.", call. = FALSE)
        }
        if (verbose) {
            message("Downloading Stable Diffusion 2.1 (diffusers safetensors)...")
        }
    }
    paths <- vapply(.sd21_files, function(f) {
        hfhub::hub_download(.sd21_repo, f, repo_type = "dataset")
    }, character(1))
    # diffusers root = parent of unet/ (paths[[1]] is diffusers/unet/*.safetensors)
    diffusers_dir <- dirname(dirname(paths[[1]]))
    if (verbose) {
        message("SD 2.1 ready: ", diffusers_dir)
    }
    invisible(diffusers_dir)
}

#' Assemble a native SD pipeline from a diffusers safetensors directory
#'
#' Builds the native UNet, VAE decoder, and CLIP text encoder from a
#' diffusers directory using the \code{*_from_safetensors} constructors,
#' places each on its component device, and returns the \code{$unet /
#' $decoder / $text_encoder} list the \code{txt2img_*} denoise loop
#' expects.
#'
#' @param diffusers_dir Directory with \code{unet/}, \code{vae/},
#'   \code{text_encoder/} subdirectories.
#' @param model_name Currently "sd21" (SDXL pending its second encoder).
#' @param devices Named list of component devices (\code{unet},
#'   \code{decoder}, \code{text_encoder}); defaults to all-CPU.
#' @param unet_dtype A torch dtype for the UNet (default float16 on CUDA,
#'   float32 on CPU).
#' @param verbose Logical.
#'
#' @return A list with \code{unet}, \code{decoder}, \code{text_encoder}.
#'
#' @export
sd_pipeline_from_safetensors <- function(diffusers_dir, model_name = "sd21",
    devices = NULL, unet_dtype = NULL,
    verbose = TRUE) {
    if (!identical(model_name, "sd21")) {
        stop("sd_pipeline_from_safetensors currently supports \"sd21\"; ",
             "SDXL needs its second text encoder and added-conditioning ",
             "embeddings (see tasks/todo.md).")
    }
    diffusers_dir <- path.expand(diffusers_dir)
    if (is.null(devices)) {
        devices <- list(unet = "cpu", decoder = "cpu", text_encoder = "cpu")
    }
    if (is.null(unet_dtype)) {
        unet_dtype <- if (identical(devices$unet, "cuda")) {
            torch::torch_float16()
        } else {
            torch::torch_float32()
        }
    }

    if (verbose) {
        message("Building text encoder...")
    }
    # SD 2.1 uses the final-LN last hidden state (v-prediction path). The
    # text encoder and VAE decoder compute in float32 (the denoise loop
    # casts prompt embeds to unet_dtype and the latent to float32 before
    # decode); only the UNet runs at unet_dtype. Up-casting a float16
    # artifact to float32 for these is lossless.
    text_encoder <- text_encoder_native_from_safetensors(
        file.path(diffusers_dir, "text_encoder"), apply_final_ln = TRUE,
        verbose = FALSE)
    text_encoder$to(device = torch::torch_device(devices$text_encoder),
                    dtype = torch::torch_float32())

    if (verbose) {
        message("Building UNet...")
    }
    unet <- unet_native_from_safetensors(file.path(diffusers_dir, "unet"),
        verbose = FALSE)
    unet$to(device = torch::torch_device(devices$unet), dtype = unet_dtype)

    if (verbose) {
        message("Building VAE decoder...")
    }
    # decode = post_quant_conv + decoder (the SD VAE needs the 1x1
    # post_quant_conv the FLUX-derived native decoder omits)
    decoder <- .sd_vae_decode_from_safetensors(file.path(diffusers_dir, "vae"),
        latent_channels = 4L)
    decoder$to(device = torch::torch_device(devices$decoder),
               dtype = torch::torch_float32())

    list(unet = unet, decoder = decoder, text_encoder = text_encoder)
}

#' FLUX.1 Text-to-Image Pipeline
#'
#' FLUX latent packing helpers and (in later phases) the schnell
#' text-to-image pipeline. Ported from the diffusers reference
#' implementation (Apache-2.0, src/diffusers/pipelines/flux/
#' pipeline_flux.py).
#'
#' @name txt2img_flux
NULL

#' Pack FLUX latents into a patch sequence
#'
#' Packs a [B, C, H, W] latent into 2x2 patches, giving a sequence
#' [B, (H/2) * (W/2), C * 4]. Reference: FluxPipeline._pack_latents.
#'
#' @param latents Tensor of shape [B, C, H, W]; H and W must be even.
#'
#' @return Tensor of shape [B, (H/2) * (W/2), C * 4].
#'
#' @export
flux_pack_latents <- function(latents) {
    shape <- latents$shape
    b <- shape[1]
    ch <- shape[2]
    h <- shape[3]
    w <- shape[4]
    latents <- latents$view(c(b, ch, h %/% 2L, 2L, w %/% 2L, 2L))
    # Python permute (0, 2, 4, 1, 3, 5), 1-indexed here
    latents <- latents$permute(c(1L, 3L, 5L, 2L, 4L, 6L))
    latents$reshape(c(b, (h %/% 2L) * (w %/% 2L), ch * 4L))
}

#' Unpack a FLUX patch sequence back into latents
#'
#' Inverse of \code{flux_pack_latents}. Height and width are the target
#' image dimensions in pixels; the latent grid is derived via the VAE
#' scale factor and the 2x2 patch size. Reference:
#' FluxPipeline._unpack_latents.
#'
#' @param latents Tensor of shape [B, S, C_packed].
#' @param height,width Integers. Image height/width in pixels.
#' @param vae_scale_factor Integer. Spatial downsampling of the VAE (8).
#'
#' @return Tensor of shape [B, C_packed / 4, height / 8, width / 8].
#'
#' @export
flux_unpack_latents <- function(latents, height, width, vae_scale_factor = 8L) {
    shape <- latents$shape
    b <- shape[1]
    ch <- shape[3]
    h <- 2L * (as.integer(height) %/% (vae_scale_factor * 2L))
    w <- 2L * (as.integer(width) %/% (vae_scale_factor * 2L))
    latents <- latents$view(c(b, h %/% 2L, w %/% 2L, ch %/% 4L, 2L, 2L))
    # Python permute (0, 3, 1, 4, 2, 5), 1-indexed here
    latents <- latents$permute(c(1L, 4L, 2L, 5L, 3L, 6L))
    latents$reshape(c(b, ch %/% 4L, h, w))
}

# Resolve a FLUX.1-schnell support file from the HuggingFace cache
.flux1_cached <- function(file) {
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to locate model files.")
    }
    tryCatch(
             hfhub::hub_download(.flux1_repo, file, local_files_only = TRUE),
             error = function(e) {
        stop("Missing ", file, " in the HuggingFace cache; ",
             "run download_flux1() first.", call. = FALSE)
    }
    )
}

#' Load the FLUX.1-schnell pipeline
#'
#' Loads the quantized transformer artifact plus the VAE decoder, CLIP
#' and T5 text encoders, tokenizer, and scheduler config (from the
#' HuggingFace cache populated by \code{\link{download_flux1}}).
#' Components load to the CPU when \code{phase_offload} is on and move
#' to the GPU only for their phase of the generation.
#'
#' @param model_dir Quantized artifact directory (default: the
#'   \code{download_flux1} location for \code{precision}), or a raw
#'   diffusers transformer directory for full-precision loading.
#' @param device Character. Compute device.
#' @param precision "nf4" or "fp8"; NULL picks the
#'   \code{\link{flux_memory_profile}} recommendation.
#' @param text_device Where the text encoders compute. NULL (default)
#'   takes the \code{\link{flux_memory_profile}} recommendation: "cuda"
#'   on 14 GB+ cards (T5-XXL onloads in bfloat16 for its encode) and
#'   "cpu" below that (T5-XXL runs float32 in place, since its ~9.8 GB
#'   GPU encode phase would not fit).
#' @param attn_chunk Integer or NULL. Attention query-chunk override.
#' @param phase_offload Logical. One GPU tenant per phase.
#' @param pin Logical or NULL. Page-lock the phase-swapped weights for
#'   DMA-rate transfer (see \code{\link{staging}}). NULL (default)
#'   resolves via \code{options(diffuseR.pin_staging)} then the
#'   host-RAM-aware \code{\link{recommend}} decision.
#' @param verbose Logical.
#'
#' @return A \code{flux_pipeline} list.
#'
#' @export
flux_load_pipeline <- function(model_dir = NULL, device = "cuda",
                               precision = NULL, text_device = NULL,
                               attn_chunk = NULL, phase_offload = TRUE,
                               pin = NULL, verbose = TRUE) {
    profile <- flux_memory_profile()
    if (verbose && !is.null(profile$note)) {
        message(profile$note)
    }
    if (is.null(attn_chunk)) {
        attn_chunk <- profile$attn_chunk
    }
    prefix <- file.path(tools::R_user_dir("diffuseR", "data"), "flux1-schnell-")
    if (is.null(precision)) {
        # Reconcile the VRAM recommendation with what is on disk: prefer
        # a built artifact (fp8 first when readable), else nf4. Keeps a
        # widened fp8 recommendation from pointing at an unbuilt artifact.
        precision <- .flux_resolve_precision("auto", prefix)
    }
    if (is.null(model_dir)) {
        model_dir <- .flux_model_dir("flux1", precision, prefix)
    }

    ckpt <- if (file.exists(file.path(model_dir, "manifest.json"))) {
        flux_open_quantized(model_dir)
    } else {
        flux_open_checkpoint(model_dir)
    }
    pin <- .resolve_pin(pin, "flux1", ckpt$format %||% "full")
    if (is.null(text_device)) {
        # The recommend tier GPU-encodes T5 only where its ~9.8 GB phase
        # fits (fp8/bf16 tiers); nf4 tiers keep the CPU-fp32 encode.
        text_device <- if (device == "cuda") {
            profile$text_device %||% "cpu"
        } else {
            "cpu"
        }
    }

    if (!nzchar(Sys.getenv("PYTORCH_CUDA_ALLOC_CONF"))) {
        # Must be set before the first CUDA allocation (see
        # ltx23_load_pipeline for the per-format rationale)
        conf <- if (identical(ckpt$format, "nf4")) {
            "backend:native"
        } else {
            "expandable_segments:True"
        }
        Sys.setenv(PYTORCH_CUDA_ALLOC_CONF = conf)
    }
    if (device == "cuda") {
        if (identical(ckpt$format, "nf4")) {
            footprint <- 8
        } else {
            footprint <- 4
        }
        if (identical(text_device, "cuda")) {
            # GPU T5 encode is the largest phase (~9.8 GB); size the gc
            # gate to it, not the smaller transformer footprint
            footprint <- max(footprint, 10)
        }
        .flux_gc_gates(footprint_gb = footprint)
    }

    if (phase_offload) {
        component_device <- "cpu"
    } else {
        component_device <- device
    }

    pipe <- list(
                 format = ckpt$format %||% "full",
                 device = device,
                 text_device = text_device,
                 phase_offload = phase_offload,
                 attn_chunk = if (is.null(attn_chunk)) NULL else as.integer(attn_chunk),
                 config = ckpt$config
    )

    if (verbose) {
        message("Loading transformer (", pipe$format, ")...")
    }
    pipe$transformer <- flux_load_transformer(ckpt, device = component_device,
        dtype = if (device == "cpu") "float32" else "bfloat16",
        pin = pin && device == "cuda", verbose = verbose)

    vae_config <- jsonlite::fromJSON(.flux1_cached("vae/config.json"))
    pipe$vae_scaling_factor <- vae_config$scaling_factor %||% 0.3611
    pipe$vae_shift_factor <- vae_config$shift_factor %||% 0.1159
    if (verbose) {
        message("Loading VAE decoder...")
    }
    dec <- vae_decoder_native(
                              latent_channels = as.integer(vae_config$latent_channels %||% 16L)
    )
    load_decoder_safetensors(
                             dec, .flux1_cached("vae/diffusion_pytorch_model.safetensors"),
                             verbose = verbose
    )
    dec$to(device = component_device)
    if (device == "cuda" && phase_offload) {
        # Cast to compute dtype once so the per-generation onload is
        # device-only and the pinned copy is bf16 (see flux2 loader).
        dec$to(dtype = torch::torch_bfloat16())
    }
    dec$eval()
    pipe$decoder <- dec

    # Text encoders load to the host when phase-offloading (so they can
    # stage and swap in for the encode); text_device controls where they
    # compute. On the CPU-encode tiers they stay resident and compute in
    # place (CLIP fp32, T5 fp32); when they GPU-encode, T5 loads bf16.
    if (phase_offload) {
        te_device <- "cpu"
    } else {
        te_device <- text_device
    }
    if (verbose) {
        message("Loading CLIP text encoder...")
    }
    clip <- text_encoder_native(gelu_type = "quick")
    load_text_encoder_safetensors(
                                  clip, .flux1_cached("text_encoder/model.safetensors"),
                                  verbose = verbose
    )
    clip$to(device = te_device)
    clip$eval()
    pipe$text_encoder <- clip

    if (verbose) {
        message("Loading T5 text encoder...")
    }
    t5_dir <- dirname(.flux1_cached("text_encoder_2/config.json"))
    pipe$text_encoder2 <- load_t5_text_encoder(
        t5_dir, device = te_device,
        dtype = if (text_device == "cpu") "float32" else "bfloat16",
        verbose = verbose
    )
    pipe$tokenizer2 <- unigram_tokenizer(.flux1_cached("tokenizer_2/tokenizer.json"))

    sched_cfg <- tryCatch(
                          jsonlite::fromJSON(.flux1_cached("scheduler/scheduler_config.json")),
                          error = function(e) NULL
    )
    pipe$scheduler_shift <- sched_cfg$shift %||% 1.0

    components <- c("transformer", "decoder")
    if (!identical(text_device, "cpu")) {
        components <- c(components, "text_encoder", "text_encoder2")
    }
    pipe$staging <- .flux_build_staging(pipe, pin, phase_offload, device,
                                        components, verbose = verbose)

    structure(pipe, class = "flux_pipeline")
}

# Flow-matching Euler loop over the packed latent sequence. Latents stay
# float32; the transformer runs in compute_dtype. schnell is CFG-free:
# one forward per step.
.flux_denoise <- function(transformer, latents, schedule, prompt_embeds,
                          pooled_prompt_embeds, image_rotary_emb,
                          compute_dtype, chunk_size = NULL, verbose = TRUE) {
    timesteps <- as.numeric(schedule$timesteps$cpu())
    n <- length(timesteps)
    pb <- .denoise_progress(n, NULL, verbose)
    f32 <- torch::torch_float32()

    torch::with_no_grad({
        for (i in seq_len(n)) {
            t <- timesteps[[i]]
            # The transformer takes sigma-space time (it rescales by 1000
            # internally, matching the reference pipeline's t/1000)
            t_model <- torch::torch_tensor(t / 1000, dtype = compute_dtype,
                device = latents$device)$reshape(1L)

            noise_pred <- transformer(
                                      hidden_states = latents$to(dtype = compute_dtype),
                                      encoder_hidden_states = prompt_embeds,
                                      pooled_projections = pooled_prompt_embeds,
                                      timestep = t_model,
                                      image_rotary_emb = image_rotary_emb,
                                      chunk_size = chunk_size
            )

            step <- flowmatch_scheduler_step(
                noise_pred$to(dtype = f32), t, latents, schedule
            )
            latents <- step$prev_sample
            schedule <- step$schedule
            rm(noise_pred, step)
            pb$tick(i)
        }
    })
    pb$done()
    latents
}

#' Generate an image with FLUX.1-schnell
#'
#' 4-step distilled text-to-image generation (no classifier-free
#' guidance): T5 + CLIP prompt encoding, flow-matching Euler denoising
#' over the packed latent sequence, and 16-channel VAE decode. With
#' phase offloading each component is the sole GPU tenant for its phase.
#'
#' @param prompt Character. The prompt.
#' @param pipeline A \code{flux_pipeline} from
#'   \code{\link{flux_load_pipeline}}; NULL loads one (passing
#'   \code{...} through).
#' @param width,height Integers, divisible by 16.
#' @param num_inference_steps Integer. Denoising steps (schnell: 4).
#' @param max_sequence_length Integer. T5 token length (schnell: 256).
#' @param seed Integer or NULL. Initial latents are drawn on the CPU, so
#'   a seed matches a Python diffusers run with a CPU generator.
#' @param prompt_embeds,pooled_prompt_embeds Optional precomputed text
#'   embeddings (skip the text encoders).
#' @param save_file Logical. Write a PNG.
#' @param filename Output path (default derived from the prompt).
#' @param verbose Logical, or one of "silent", "progress", "steps".
#'   TRUE = "steps" (full per-phase chatter), FALSE = "silent".
#'   "progress" prints a one-line generation summary plus a denoise
#'   progress bar (interactive) or periodic step ticks (captured logs).
#' @param ... Passed to \code{\link{flux_load_pipeline}} when
#'   \code{pipeline} is NULL.
#'
#' @return Invisibly, \code{list(image, metadata)} where \code{image} is
#'   an [H, W, 3] array in [0, 1].
#'
#' @export
txt2img_flux <- function(prompt, pipeline = NULL, width = 1024L,
                         height = 1024L, num_inference_steps = 4L,
                         max_sequence_length = 256L, seed = NULL,
                         prompt_embeds = NULL, pooled_prompt_embeds = NULL,
                         save_file = TRUE, filename = NULL, verbose = TRUE,
                         ...) {
    level <- .verbosity(verbose)
    verbose <- level == "steps"
    if (level != "silent") {
        message(sprintf("FLUX.1 schnell: %dx%d, %d steps", as.integer(width),
                        as.integer(height), as.integer(num_inference_steps)))
    }
    if (is.null(pipeline)) {
        pipeline <- flux_load_pipeline(..., verbose = verbose)
    }
    device <- pipeline$device
    width <- as.integer(width)
    height <- as.integer(height)
    if (width %% 16L != 0L || height %% 16L != 0L) {
        stop("width and height must be divisible by 16")
    }

    f32 <- torch::torch_float32()
    compute_dtype <- if (device == "cpu") {
        f32
    } else {
        torch::torch_bfloat16()
    }

    phase_offload <- isTRUE(pipeline$phase_offload) && device != "cpu"
    staging <- pipeline$staging %||% list()
    # Components move by name so pinned staging (see staging.R) can carry
    # the CPU<->GPU transfer when the loader prepared it; otherwise the
    # pageable module$to() path runs.
    onload <- function(what) {
        module <- pipeline[[what]]
        if (phase_offload) {
            st <- staging[[what]]
            if (is.null(st)) {
                module$to(device = device)
            } else {
                .staged_onload(st, device)
            }
        }
        module
    }
    offload <- function(what) {
        module <- pipeline[[what]]
        if (phase_offload) {
            st <- staging[[what]]
            if (is.null(st)) {
                module$to(device = "cpu")
            } else {
                .staged_offload(st)
            }
            clear_vram()
        }
        invisible(module)
    }

    t0 <- Sys.time()

    # --- Phase 1: text encoding ------------------------------------------------
    # On the CPU-encode tiers (text_device == "cpu") the encoders compute
    # in place; otherwise each swaps to the GPU for its own encode. T5 and
    # CLIP are independent (either embed may be supplied precomputed), so
    # each onloads inside its own guard.
    gpu_encode <- !identical(pipeline$text_device, "cpu")
    torch::with_no_grad({
        if (is.null(prompt_embeds)) {
            if (verbose) {
                message("Encoding prompt (T5)...")
            }
            if (gpu_encode) {
                onload("text_encoder2")
            }
            prompt_embeds <- encode_with_t5(prompt, pipeline$text_encoder2,
                pipeline$tokenizer2, max_sequence_length = max_sequence_length)
            if (gpu_encode) {
                offload("text_encoder2")
            }
        }
        if (is.null(pooled_prompt_embeds)) {
            if (gpu_encode) {
                onload("text_encoder")
            }
            tokens <- CLIPTokenizer(prompt)
            clip_device <- pipeline$text_encoder$token_embedding$weight$device
            tokens <- tokens$to(device = clip_device)
            hidden <- pipeline$text_encoder(tokens)
            pooled_prompt_embeds <- clip_pooled_output(hidden, tokens)
            if (gpu_encode) {
                offload("text_encoder")
            }
        }
    })
    prompt_embeds <- prompt_embeds$to(device = device, dtype = compute_dtype)
    pooled_prompt_embeds <- pooled_prompt_embeds$to(device = device,
        dtype = compute_dtype)
    txt_len <- prompt_embeds$shape[2]

    # --- Phase 2: latents, rotary embeddings, schedule --------------------------
    lat_ch <- as.integer((pipeline$config$in_channels %||% 64L) %/% 4L)
    h_lat <- 2L * (height %/% 16L)
    w_lat <- 2L * (width %/% 16L)
    if (!is.null(seed)) {
        torch::torch_manual_seed(seed)
    }
    # Drawn on the CPU in the unpacked diffusers shape for seed parity
    latents <- torch::torch_randn(c(1L, lat_ch, h_lat, w_lat), dtype = f32)
    latents <- flux_pack_latents(latents)$to(device = device)

    img_ids <- flux_prepare_latent_image_ids(h_lat %/% 2L, w_lat %/% 2L)
    txt_ids <- torch::torch_zeros(txt_len, 3L)
    ids <- torch::torch_cat(list(txt_ids, img_ids), dim = 1L)
    rope <- flux_pos_embed(ids,
                           axes_dim = pipeline$transformer$axes_dims_rope %||% c(16L, 56L, 56L))
    rope <- list(rope[[1]]$to(device = device), rope[[2]]$to(device = device))

    sched <- flowmatch_scheduler_create(
                                        shift = pipeline$scheduler_shift %||% 1.0,
                                        use_dynamic_shifting = FALSE
    )
    n_steps <- as.integer(num_inference_steps)
    sched <- flowmatch_set_timesteps(
                                     sched, n_steps,
                                     sigmas = seq(1, 1 / n_steps, length.out = n_steps)
    )

    # --- Phase 3: denoise --------------------------------------------------------
    transformer <- onload("transformer")
    if (verbose) {
        message(sprintf("Denoising: %d steps at %dx%d...", n_steps, width,
                        height))
    }
    latents <- .flux_denoise(
                             transformer, latents, sched, prompt_embeds,
                             pooled_prompt_embeds, rope, compute_dtype,
                             chunk_size = pipeline$attn_chunk, verbose = level
    )
    offload("transformer")
    ltx23_release_dequant_buffers()

    # --- Phase 4: decode -----------------------------------------------------------
    if (verbose) {
        message("Decoding...")
    }
    latents <- flux_unpack_latents(latents, height, width)
    latents <- latents$div(pipeline$vae_scaling_factor %||% 0.3611)$
    add(pipeline$vae_shift_factor %||% 0.1159)

    decoder <- onload("decoder")
    torch::with_no_grad({
        dec_param <- decoder$conv_in$weight
        img <- decoder(latents$to(device = dec_param$device,
                                  dtype = dec_param$dtype))
        img <- img$to(dtype = f32)$cpu()
    })
    offload("decoder")

    img <- img$squeeze(1)$permute(c(2L, 3L, 1L))
    img <- img$add(1)$div(2)$clamp(0, 1)
    img_array <- as.array(img)

    gen_seconds <- as.numeric(difftime(Sys.time(), t0, units = "secs"))
    if (verbose) {
        message(sprintf("Generated in %.1f s", gen_seconds))
    }

    if (save_file) {
        if (is.null(filename)) {
            filename <- filename_from_prompt(prompt)
        }
        save_image(img_array, filename)
        if (verbose) {
            message("Saved to ", filename)
        }
    }

    metadata <- list(
                     prompt = prompt, width = width, height = height,
                     steps = n_steps, seed = seed, model = "flux1-schnell",
                     precision = pipeline$format, seconds = gen_seconds
    )
    invisible(list(image = img_array, metadata = metadata))
}

#' LTX-2.3 Causal Video VAE
#'
#' Fresh R port of the LTX-2 video autoencoder from the diffusers
#' reference (Apache-2.0, autoencoder_kl_ltx2.py), with LTX 2.3 defaults:
#' encoder blocks (256, 512, 1024, 1024), a 4-up-block decoder with mixed
#' (spatiotemporal, spatiotemporal, temporal, spatial) upsampling, no
#' upsample residuals, and zeros spatial padding throughout. The encoder
#' is causal; the decoder is not.
#'
#' @name vae_ltx23
NULL

#' LTX-2.3 video encoder
#'
#' Pixel video [B, 3, F, H, W] -> latent statistics
#' [B, 2 * latent_channels, F/8, H/32, W/32] (mean and a uniform log-var
#' channel broadcast across the latent channels).
#'
#' @param in_channels,out_channels Integers. Pixel and latent channels.
#' @param block_out_channels Integer vector. Per-block output channels.
#' @param spatio_temporal_scaling Logical vector per block.
#' @param layers_per_block Integer vector (blocks then mid).
#' @param downsample_type Character vector per block.
#' @param patch_size,patch_size_t Integers. Pixel patchification.
#' @param resnet_norm_eps Numeric.
#' @param is_causal Logical.
#' @param spatial_padding_mode Character.
#'
#' @return Module whose forward(hidden_states, causal) returns the
#'   encoded video latent [B, C, F, H, W], with the last channel
#'   repeated to carry the per-channel scale expected downstream.
#'
#' @export
ltx23_video_encoder3d <- torch::nn_module(
    "ltx23_video_encoder3d",
    initialize = function(
                          in_channels = 3L,
                          out_channels = 128L,
                          block_out_channels = c(256L, 512L, 1024L, 1024L),
                          spatio_temporal_scaling = c(TRUE, TRUE, TRUE, TRUE),
                          layers_per_block = c(4L, 6L, 4L, 2L, 2L),
                          downsample_type = NULL,
                          patch_size = 4L,
                          patch_size_t = 1L,
                          resnet_norm_eps = 1e-6,
                          is_causal = TRUE,
                          spatial_padding_mode = "zeros"
    ) {
    if (is.null(downsample_type)) {
        downsample_type <- c("spatial", "temporal", "spatiotemporal",
                             "spatiotemporal")
    }
    self$patch_size <- as.integer(patch_size)
    self$patch_size_t <- as.integer(patch_size_t)
    self$in_channels <- in_channels * patch_size ^ 2
    self$is_causal <- is_causal

    output_channel <- out_channels
    self$conv_in <- ltx23_causal_conv3d(
                                        self$in_channels, output_channel, kernel_size = 3L, stride = 1L,
                                        spatial_padding_mode = spatial_padding_mode
    )

    down_blocks <- list()
    for (i in seq_along(block_out_channels)) {
        input_channel <- output_channel
        output_channel <- block_out_channels[i]
        down_blocks[[i]] <- ltx23_video_down_block3d(
            in_channels = input_channel,
            out_channels = output_channel,
            num_layers = layers_per_block[i],
            resnet_eps = resnet_norm_eps,
            spatio_temporal_scale = spatio_temporal_scaling[i],
            downsample_type = downsample_type[i],
            spatial_padding_mode = spatial_padding_mode
        )
    }
    self$down_blocks <- torch::nn_module_list(down_blocks)

    self$mid_block <- ltx23_video_mid_block3d(
        in_channels = output_channel,
        num_layers = layers_per_block[length(layers_per_block)],
        resnet_eps = resnet_norm_eps,
        spatial_padding_mode = spatial_padding_mode
    )

    self$norm_out <- ltx23_per_channel_rms_norm()
    self$conv_out <- ltx23_causal_conv3d(
        output_channel, out_channels + 1L, kernel_size = 3L, stride = 1L,
        spatial_padding_mode = spatial_padding_mode
    )
},
    forward = function(hidden_states, causal = NULL) {
    causal <- causal %||% self$is_causal
    p <- self$patch_size
    p_t <- self$patch_size_t

    dims <- hidden_states$shape
    batch_size <- dims[1]; num_channels <- dims[2]
    post_f <- dims[3] %/% p_t; post_h <- dims[4] %/% p; post_w <- dims[5] %/% p

    # Pixel patchification: space-to-channel with the LTX ordering
    hidden_states <- hidden_states$reshape(c(
            batch_size, num_channels, post_f, p_t, post_h, p, post_w, p
        ))
    # [B, C, F', pt, H', p, W', p] -> [B, C, pt, pw, ph, F', H', W']
    hidden_states <- hidden_states$permute(c(1L, 2L, 4L, 8L, 6L, 3L, 5L, 7L))$
    flatten(start_dim = 2L, end_dim = 5L)
    hidden_states <- self$conv_in(hidden_states, causal = causal)

    for (i in seq_along(self$down_blocks)) {
        hidden_states <- self$down_blocks[[i]](hidden_states, causal = causal)
    }
    hidden_states <- self$mid_block(hidden_states, causal = causal)

    hidden_states <- self$norm_out(hidden_states)
    hidden_states <- torch::nnf_silu(hidden_states)
    hidden_states <- self$conv_out(hidden_states, causal = causal)

    # Broadcast the single log-var channel across all latent channels
    n_ch <- hidden_states$shape[2]
    last_channel <- hidden_states$narrow(2L, n_ch, 1L)$`repeat`(c(1L, n_ch - 2L, 1L, 1L, 1L))
    torch::torch_cat(list(hidden_states, last_channel), dim = 2L)
}
)

#' LTX-2.3 video decoder
#'
#' Latents [B, 128, F, H, W] -> pixel video [B, 3, 8F - 7, 32H, 32W].
#' Block channel lists are given encoder-side (as in the config) and
#' reversed internally; \code{upsample_type} is indexed directly.
#'
#' @param in_channels,out_channels Integers. Latent and pixel channels.
#' @param block_out_channels Integer vector (config order).
#' @param spatio_temporal_scaling Logical vector per up block.
#' @param layers_per_block Integer vector (config order; first entry is
#'   the mid block after reversal).
#' @param upsample_type Character vector per up block (not reversed).
#' @param patch_size,patch_size_t Integers.
#' @param resnet_norm_eps Numeric.
#' @param is_causal Logical. FALSE for LTX (symmetric temporal padding).
#' @param upsample_residual Logical vector per up block.
#' @param upsample_factor Integer vector per up block.
#' @param spatial_padding_mode Character.
#'
#' @return Module whose forward(hidden_states, causal) returns the
#'   decoded pixel tensor [B, 3, F, H, W], with the final patch axes
#'   flattened back into height and width.
#'
#' @export
ltx23_video_decoder3d <- torch::nn_module(
    "ltx23_video_decoder3d",
    initialize = function(
                          in_channels = 128L,
                          out_channels = 3L,
                          block_out_channels = c(256L, 512L, 512L, 1024L),
                          spatio_temporal_scaling = c(TRUE, TRUE, TRUE, TRUE),
                          layers_per_block = c(4L, 6L, 4L, 2L, 2L),
                          upsample_type = NULL,
                          patch_size = 4L,
                          patch_size_t = 1L,
                          resnet_norm_eps = 1e-6,
                          is_causal = FALSE,
                          upsample_residual = c(FALSE, FALSE, FALSE, FALSE),
                          upsample_factor = c(2L, 2L, 1L, 2L),
                          spatial_padding_mode = "zeros"
    ) {
    if (is.null(upsample_type)) {
        upsample_type <- c("spatiotemporal", "spatiotemporal", "temporal",
                           "spatial")
    }
    self$patch_size <- as.integer(patch_size)
    self$patch_size_t <- as.integer(patch_size_t)
    self$out_channels <- out_channels * patch_size ^ 2
    self$is_causal <- is_causal

    block_out_channels <- rev(block_out_channels)
    spatio_temporal_scaling <- rev(spatio_temporal_scaling)
    layers_per_block <- rev(layers_per_block)
    upsample_residual <- rev(upsample_residual)
    upsample_factor <- rev(upsample_factor)
    # NOTE: upsample_type is deliberately NOT reversed (reference behavior)

    output_channel <- block_out_channels[1]
    self$conv_in <- ltx23_causal_conv3d(
                                        in_channels, output_channel, kernel_size = 3L, stride = 1L,
                                        spatial_padding_mode = spatial_padding_mode
    )

    self$mid_block <- ltx23_video_mid_block3d(
        in_channels = output_channel,
        num_layers = layers_per_block[1],
        resnet_eps = resnet_norm_eps,
        spatial_padding_mode = spatial_padding_mode
    )

    up_blocks <- list()
    for (i in seq_along(block_out_channels)) {
        input_channel <- output_channel %/% upsample_factor[i]
        output_channel <- block_out_channels[i] %/% upsample_factor[i]
        up_blocks[[i]] <- ltx23_video_up_block3d(
            in_channels = input_channel,
            out_channels = output_channel,
            num_layers = layers_per_block[i + 1L],
            resnet_eps = resnet_norm_eps,
            spatio_temporal_scale = spatio_temporal_scaling[i],
            upsample_type = upsample_type[i],
            upsample_residual = upsample_residual[i],
            upscale_factor = upsample_factor[i],
            spatial_padding_mode = spatial_padding_mode
        )
    }
    self$up_blocks <- torch::nn_module_list(up_blocks)

    self$norm_out <- ltx23_per_channel_rms_norm()
    self$conv_out <- ltx23_causal_conv3d(
        output_channel, self$out_channels, kernel_size = 3L, stride = 1L,
        spatial_padding_mode = spatial_padding_mode
    )
},
    forward = function(hidden_states, causal = NULL) {
    causal <- causal %||% self$is_causal

    hidden_states <- self$conv_in(hidden_states, causal = causal)
    hidden_states <- self$mid_block(hidden_states, causal = causal)
    for (i in seq_along(self$up_blocks)) {
        hidden_states <- self$up_blocks[[i]](hidden_states, causal = causal)
    }

    hidden_states <- self$norm_out(hidden_states)
    hidden_states <- torch::nnf_silu(hidden_states)
    hidden_states <- self$conv_out(hidden_states, causal = causal)

    # Un-patchify: channel-to-space with the LTX ordering
    p <- self$patch_size
    p_t <- self$patch_size_t
    dims <- hidden_states$shape
    hidden_states <- hidden_states$reshape(c(
            dims[1], -1L, p_t, p, p, dims[3], dims[4], dims[5]
        ))
    # [B, C, pt, ph, pw, F, H, W] -> [B, C, F, pt, H, pw, W, ph]
    hidden_states <- hidden_states$permute(c(1L, 2L, 6L, 3L, 7L, 5L, 8L, 4L))
    hidden_states$flatten(start_dim = 7L, end_dim = 8L)$
    flatten(start_dim = 5L, end_dim = 6L)$
    flatten(start_dim = 3L, end_dim = 4L)
}
)

#' LTX-2.3 video VAE
#'
#' Encoder + decoder + per-channel latent statistics (loaded from the
#' checkpoint's \code{per_channel_statistics}). The checkpoint's
#' \code{scaling_factor} is 1.0, so latent (de)normalization is purely
#' the per-channel affine map.
#'
#' @param in_channels,out_channels Integers. Pixel channels.
#' @param latent_channels Integer.
#' @param block_out_channels,layers_per_block,spatio_temporal_scaling,downsample_type
#'   Encoder configuration (see \code{\link{ltx23_video_encoder3d}}).
#' @param decoder_block_out_channels,decoder_layers_per_block,decoder_spatio_temporal_scaling,upsample_type,upsample_residual,upsample_factor
#'   Decoder configuration (see \code{\link{ltx23_video_decoder3d}}).
#' @param patch_size,patch_size_t Integers. Pixel patchification.
#' @param resnet_norm_eps Numeric.
#' @param encoder_causal,decoder_causal Logicals. Temporal padding modes.
#' @param encoder_spatial_padding_mode,decoder_spatial_padding_mode Characters.
#'
#' @return Module bundling the video encoder and decoder. Its
#'   forward(z) is \code{decode(z)}, returning pixels for a latent;
#'   \code{$encode()} and \code{$decode()} are callable separately.
#'
#' @export
ltx23_video_vae <- torch::nn_module(
                                    "ltx23_video_vae",
                                    initialize = function(
        in_channels = 3L,
        out_channels = 3L,
        latent_channels = 128L,
        block_out_channels = c(256L, 512L, 1024L, 1024L),
        decoder_block_out_channels = c(256L, 512L, 512L, 1024L),
        layers_per_block = c(4L, 6L, 4L, 2L, 2L),
        decoder_layers_per_block = c(4L, 6L, 4L, 2L, 2L),
        spatio_temporal_scaling = c(TRUE, TRUE, TRUE, TRUE),
        decoder_spatio_temporal_scaling = c(TRUE, TRUE, TRUE, TRUE),
        downsample_type = NULL,
        upsample_type = NULL,
        upsample_residual = c(FALSE, FALSE, FALSE, FALSE),
        upsample_factor = c(2L, 2L, 1L, 2L),
        patch_size = 4L,
        patch_size_t = 1L,
        resnet_norm_eps = 1e-6,
        encoder_causal = TRUE,
        decoder_causal = FALSE,
        encoder_spatial_padding_mode = "zeros",
        decoder_spatial_padding_mode = "zeros"
    ) {
    if (is.null(downsample_type)) {
        downsample_type <- c("spatial", "temporal", "spatiotemporal",
                             "spatiotemporal")
    }
    if (is.null(upsample_type)) {
        upsample_type <- c("spatiotemporal", "spatiotemporal", "temporal",
                           "spatial")
    }
    self$latent_channels <- as.integer(latent_channels)

    self$encoder <- ltx23_video_encoder3d(
        in_channels = in_channels,
        out_channels = latent_channels,
        block_out_channels = block_out_channels,
        spatio_temporal_scaling = spatio_temporal_scaling,
        layers_per_block = layers_per_block,
        downsample_type = downsample_type,
        patch_size = patch_size,
        patch_size_t = patch_size_t,
        resnet_norm_eps = resnet_norm_eps,
        is_causal = encoder_causal,
        spatial_padding_mode = encoder_spatial_padding_mode
    )
    self$decoder <- ltx23_video_decoder3d(
        in_channels = latent_channels,
        out_channels = out_channels,
        block_out_channels = decoder_block_out_channels,
        spatio_temporal_scaling = decoder_spatio_temporal_scaling,
        layers_per_block = decoder_layers_per_block,
        upsample_type = upsample_type,
        patch_size = patch_size,
        patch_size_t = patch_size_t,
        resnet_norm_eps = resnet_norm_eps,
        is_causal = decoder_causal,
        upsample_residual = upsample_residual,
        upsample_factor = upsample_factor,
        spatial_padding_mode = decoder_spatial_padding_mode
    )

    self$latents_mean <- torch::nn_buffer(torch::torch_zeros(latent_channels))
    self$latents_std <- torch::nn_buffer(torch::torch_ones(latent_channels))

    # Tiled decoding (reference defaults, in sample/pixel space)
    self$use_tiling <- FALSE
    self$use_framewise_decoding <- FALSE
    self$tile_sample_min_height <- 512L
    self$tile_sample_min_width <- 512L
    self$tile_sample_min_num_frames <- 16L
    self$tile_sample_stride_height <- 448L
    self$tile_sample_stride_width <- 448L
    self$tile_sample_stride_num_frames <- 8L
},
                                    encode = function(x, causal = NULL) {
    moments <- self$encoder(x, causal = causal)
    n <- self$latent_channels
    list(
         mean = moments$narrow(2L, 1L, n),
         logvar = moments$narrow(2L, n + 1L, n)
    )
},
                                    enable_tiling = function(spatial = TRUE, temporal = TRUE) {
    self$use_tiling <- spatial
    self$use_framewise_decoding <- temporal
    invisible(self)
},
                                    decode = function(z, causal = NULL) {
    num_frames <- z$shape[3]
    height <- z$shape[4]
    width <- z$shape[5]
    tile_lat_h <- self$tile_sample_min_height %/% 32L
    tile_lat_w <- self$tile_sample_min_width %/% 32L
    tile_lat_f <- self$tile_sample_min_num_frames %/% 8L

    wants_temporal <- self$use_framewise_decoding && num_frames > tile_lat_f
    wants_spatial <- self$use_tiling &&
    (width > tile_lat_w || height > tile_lat_h)
    if ((wants_temporal || wants_spatial) && .ltx23_untiled_fits(z)) {
        # Tiling exists to bound VRAM, not for correctness: when the
        # whole sample fits, one full-latent forward skips the tile
        # overlap recompute and seam blending (2.7 s vs 4.4 s at
        # 768x512x49)
        return(.ltx23_decode_tile(self, z, causal))
    }
    if (wants_temporal) {
        return(.ltx23_temporal_tiled_decode(self, z, causal = causal))
    }
    if (wants_spatial) {
        return(.ltx23_tiled_decode(self, z, causal = causal))
    }
    .ltx23_decode_tile(self, z, causal)
},
                                    forward = function(z) {
    self$decode(z)
}
)

# Linear crossfade of tile b's leading rows/columns/frames with tile a's
# trailing ones (in-place on b), per the diffusers reference
.ltx23_blend_v <- function(a, b, blend_extent) {
    blend_extent <- min(a$shape[4], b$shape[4], blend_extent)
    if (blend_extent <= 0L) {
        return(b)
    }
    a_h <- a$shape[4]
    for (y in seq_len(blend_extent)) {
        w <- (y - 1) / blend_extent
        b[,,, y,] <- a[,,, a_h - blend_extent + y,]$mul(1 - w) +
        b[,,, y,]$mul(w)
    }
    b
}

.ltx23_blend_h <- function(a, b, blend_extent) {
    blend_extent <- min(a$shape[5], b$shape[5], blend_extent)
    if (blend_extent <= 0L) {
        return(b)
    }
    a_w <- a$shape[5]
    for (x in seq_len(blend_extent)) {
        w <- (x - 1) / blend_extent
        b[,,,, x] <- a[,,,, a_w - blend_extent + x]$mul(1 - w) +
        b[,,,, x]$mul(w)
    }
    b
}

.ltx23_blend_t <- function(a, b, blend_extent) {
    blend_extent <- min(a$shape[3], b$shape[3], blend_extent)
    if (blend_extent <= 0L) {
        return(b)
    }
    a_f <- a$shape[3]
    for (x in seq_len(blend_extent)) {
        w <- (x - 1) / blend_extent
        b[,, x,,] <- a[,, a_f - blend_extent + x,,]$mul(1 - w) +
        b[,, x,,]$mul(w)
    }
    b
}

# Tiling exists to bound decoder activation VRAM. When the whole
# sample's untiled decode fits in what is left of the card, one
# full-latent forward beats the tiles. Activation cost is linear in
# output pixel-frames: ~1 GB + ~350 B per pixel-frame measured on the
# 2.3 decoder across 2.5M-19.3M pixel-frames; 360 B/pxf plus 15%
# headroom below rounds up. options(diffuseR.vae_untiled = TRUE/FALSE)
# forces a path; the "auto" default applies the fit test on CUDA and
# keeps tiled behavior elsewhere.
.ltx23_untiled_env <- new.env(parent = emptyenv())

.ltx23_untiled_fits <- function(z) {
    pref <- getOption("diffuseR.vae_untiled", "auto")
    if (isTRUE(pref)) {
        return(TRUE)
    }
    if (!identical(pref, "auto")) {
        return(FALSE)
    }
    if (z$device$type != "cuda") {
        return(FALSE)
    }
    total <- .ltx23_untiled_env$total_bytes
    if (is.null(total)) {
        gb <- tryCatch(.detect_vram(use_free = FALSE), error = function(e) NULL)
        if (!isTRUE(gb > 0)) {
            return(FALSE)
        }
        total <- gb * 1e9
        .ltx23_untiled_env$total_bytes <- total
    }
    alloc <- tryCatch(
                      as.numeric(torch::cuda_memory_stats()$allocated_bytes$all$current),
                      error = function(e) NULL
    )
    if (is.null(alloc)) {
        return(FALSE)
    }
    pxf <- ((z$shape[3] - 1) * 8 + 1) * z$shape[4] * 32 * z$shape[5] * 32
    est <- 1e9 + 360 * pxf
    est < 0.85 * (total - alloc)
}

# Spatially tiled decode: overlapping latent tiles, decoded separately,
# crossfaded at the seams (diffusers AutoencoderKLLTX2Video.tiled_decode)
.ltx23_tiled_decode <- function(vae, z, causal = NULL) {
    height <- z$shape[4]
    width <- z$shape[5]
    sample_height <- height * 32L
    sample_width <- width * 32L

    tile_lat_h <- vae$tile_sample_min_height %/% 32L
    tile_lat_w <- vae$tile_sample_min_width %/% 32L
    stride_lat_h <- vae$tile_sample_stride_height %/% 32L
    stride_lat_w <- vae$tile_sample_stride_width %/% 32L

    blend_height <- vae$tile_sample_min_height - vae$tile_sample_stride_height
    blend_width <- vae$tile_sample_min_width - vae$tile_sample_stride_width

    rows <- list()
    for (i in seq(0L, height - 1L, by = stride_lat_h)) {
        row <- list()
        for (j in seq(0L, width - 1L, by = stride_lat_w)) {
            h_len <- min(tile_lat_h, height - i)
            w_len <- min(tile_lat_w, width - j)
            tile <- .ltx23_decode_tile(
                                       vae,
                                       z$narrow(4L, i + 1L, h_len)$narrow(5L, j + 1L, w_len),
                                       causal
            )
            # Dead tile handles are left to the allocator callback:
            # with the gc gates live (ltx23_tune_gc) that no longer
            # storms, and an explicit per-tile gc() here measured
            # 1.9 s per decode
            row[[length(row) + 1L]] <- tile
        }
        rows[[length(rows) + 1L]] <- row
    }

    result_rows <- list()
    for (i in seq_along(rows)) {
        result_row <- list()
        for (j in seq_along(rows[[i]])) {
            tile <- rows[[i]][[j]]
            if (i > 1L) {
                tile <- .ltx23_blend_v(rows[[i - 1L]][[j]], tile, blend_height)
            }
            if (j > 1L) {
                tile <- .ltx23_blend_h(rows[[i]][[j - 1L]], tile, blend_width)
            }
            keep_h <- min(vae$tile_sample_stride_height, tile$shape[4])
            keep_w <- min(vae$tile_sample_stride_width, tile$shape[5])
            result_row[[length(result_row) + 1L]] <-
            tile$narrow(4L, 1L, keep_h)$narrow(5L, 1L, keep_w)
        }
        result_rows[[length(result_rows) + 1L]] <- torch::torch_cat(result_row,
            dim = 5L)
    }

    dec <- torch::torch_cat(result_rows, dim = 4L)
    dec$narrow(4L, 1L, min(sample_height, dec$shape[4]))$
    narrow(5L, 1L, min(sample_width, dec$shape[5]))
}

# Temporally tiled decode: overlapping latent frame windows (spatially
# tiled inside when the tile is large), crossfaded over time
# (diffusers AutoencoderKLLTX2Video._temporal_tiled_decode)
.ltx23_temporal_tiled_decode <- function(vae, z, causal = NULL) {
    num_frames <- z$shape[3]
    num_sample_frames <- (num_frames - 1L) * 8L + 1L

    tile_lat_h <- vae$tile_sample_min_height %/% 32L
    tile_lat_w <- vae$tile_sample_min_width %/% 32L
    tile_lat_f <- vae$tile_sample_min_num_frames %/% 8L
    stride_lat_f <- vae$tile_sample_stride_num_frames %/% 8L
    blend_num_frames <- vae$tile_sample_min_num_frames - vae$tile_sample_stride_num_frames

    row <- list()
    for (i in seq(0L, num_frames - 1L, by = stride_lat_f)) {
        f_len <- min(tile_lat_f + 1L, num_frames - i)
        tile <- z$narrow(3L, i + 1L, f_len)
        if (vae$use_tiling &&
            (tile$shape[5] > tile_lat_w || tile$shape[4] > tile_lat_h)) {
            decoded <- .ltx23_tiled_decode(vae, tile, causal = causal)
        } else {
            decoded <- .ltx23_decode_tile(vae, tile, causal)
        }
        if (i > 0L) {
            decoded <- decoded$narrow(3L, 1L, decoded$shape[3] - 1L)
        }
        row[[length(row) + 1L]] <- decoded
    }

    result_row <- list()
    for (i in seq_along(row)) {
        tile <- row[[i]]
        if (i > 1L) {
            tile <- .ltx23_blend_t(row[[i - 1L]], tile, blend_num_frames)
            keep <- min(vae$tile_sample_stride_num_frames, tile$shape[3])
            result_row[[length(result_row) + 1L]] <- tile$narrow(3L, 1L, keep)
        } else {
            keep <- min(vae$tile_sample_stride_num_frames + 1L, tile$shape[3])
            result_row[[length(result_row) + 1L]] <- tile$narrow(3L, 1L, keep)
        }
    }

    dec <- torch::torch_cat(result_row, dim = 3L)
    dec$narrow(3L, 1L, min(num_sample_frames, dec$shape[3]))
}

#' Normalize latents with the VAE's per-channel statistics
#'
#' @param latents Tensor [B, C, F, H, W].
#' @param latents_mean,latents_std Tensors [C].
#'
#' @return Normalized latents.
#'
#' @export
ltx23_normalize_latents <- function(latents, latents_mean, latents_std) {
    mean <- latents_mean$view(c(1L, -1L, 1L, 1L, 1L))$to(device = latents$device,
        dtype = latents$dtype)
    std <- latents_std$view(c(1L, -1L, 1L, 1L, 1L))$to(device = latents$device,
        dtype = latents$dtype)
    (latents - mean) / std
}

#' Denormalize latents with the VAE's per-channel statistics
#'
#' @param latents Tensor [B, C, F, H, W].
#' @param latents_mean,latents_std Tensors [C].
#'
#' @return Denormalized latents ready for the decoder.
#'
#' @export
ltx23_denormalize_latents <- function(latents, latents_mean, latents_std) {
    mean <- latents_mean$view(c(1L, -1L, 1L, 1L, 1L))$to(device = latents$device,
        dtype = latents$dtype)
    std <- latents_std$view(c(1L, -1L, 1L, 1L, 1L))$to(device = latents$device,
        dtype = latents$dtype)
    latents * std + mean
}

#' Map an official VAE checkpoint key to the R module name
#'
#' The official checkpoint stores the encoder/decoder as flat block lists
#' (down_blocks.0-8 / up_blocks.0-8) where downsamplers/upsamplers and
#' the mid block are separate entries; diffusers (and this port) nest
#' them. Index mapping per diffusers convert_ltx2_to_diffusers.py.
#'
#' @param key Character. Checkpoint key (with or without "vae." prefix).
#'
#' @return Character. Module parameter/buffer name.
#'
#' @export
ltx23_map_vae_key <- function(key) {
    key <- sub("^vae\\.", "", key)

    key <- sub("^per_channel_statistics\\.mean-of-means$", "latents_mean", key)
    key <- sub("^per_channel_statistics\\.std-of-means$", "latents_std", key)

    down_map <- c("0" = "down_blocks.0", "1" = "down_blocks.0.downsamplers.0",
                  "2" = "down_blocks.1", "3" = "down_blocks.1.downsamplers.0",
                  "4" = "down_blocks.2", "5" = "down_blocks.2.downsamplers.0",
                  "6" = "down_blocks.3", "7" = "down_blocks.3.downsamplers.0",
                  "8" = "mid_block")
    up_map <- c(
                "0" = "mid_block",
                "1" = "up_blocks.0.upsamplers.0",
                "2" = "up_blocks.0",
                "3" = "up_blocks.1.upsamplers.0",
                "4" = "up_blocks.1",
                "5" = "up_blocks.2.upsamplers.0",
                "6" = "up_blocks.2",
                "7" = "up_blocks.3.upsamplers.0",
                "8" = "up_blocks.3"
    )

    m <- regmatches(key, regexec("^(encoder\\.down_blocks|decoder\\.up_blocks)\\.([0-9]+)\\.(.*)$", key))[[1]]
    if (length(m) == 4L) {
        if (startsWith(m[2], "encoder")) {
            section <- "encoder"
        } else {
            section <- "decoder"
        }
        if (section == "encoder") {
            map <- down_map
        } else {
            map <- up_map
        }
        idx <- m[3]
        if (is.na(map[idx])) {
            return(NA_character_)
        }
        rest <- m[4]
        # Sampler entries map to the nested module directly (their inner
        # structure is just .conv.conv / norms)
        key <- paste0(section, ".", map[[idx]], ".", rest)
    }

    key <- gsub("res_blocks", "resnets", key, fixed = TRUE)
    key
}

# Parity tests for the LTX-2.3 video VAE port against diffusers
# reference fixtures (tools/gen_fixtures_vae.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "vae_ltx23.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/vae_ltx23.safetensors"
if (!file.exists(fixture_path)) exit_file("vae fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

load_group <- function(module, prefix) {
  keys <- grep(paste0("^", prefix, "\\."), names(fx), value = TRUE)
  w <- fx[keys]
  names(w) <- sub(paste0("^", prefix, "\\."), "", keys)
  dests <- c(module$named_parameters(), module$named_buffers())
  if (!setequal(names(w), names(dests))) {
    stop(
      prefix, ": name mismatch. missing dest: ",
      paste(utils::head(setdiff(names(w), names(dests)), 3), collapse = ", "),
      " | unfilled: ",
      paste(utils::head(setdiff(names(dests), names(w)), 3), collapse = ", ")
    )
  }
  torch::with_no_grad({
    for (name in names(w)) dests[[name]]$copy_(w[[name]])
  })
  module$eval()
  module
}

# --- Causal conv -----------------------------------------------------------------

cc <- load_group(ltx23_causal_conv3d(4L, 6L, kernel_size = 3L), "cc")
torch::with_no_grad({
  out_c <- cc(fx$cc_x, causal = TRUE)
  out_nc <- cc(fx$cc_x, causal = FALSE)
})
expect_true(max_abs_diff(out_c, fx$cc_out_causal) < 1e-5)
expect_true(max_abs_diff(out_nc, fx$cc_out_noncausal) < 1e-5)

# --- Resnet block with channel-change shortcut -------------------------------------

rb <- load_group(ltx23_video_resnet_block3d(4L, 8L), "rb")
torch::with_no_grad(out_rb <- rb(fx$cc_x, causal = TRUE))
expect_true(max_abs_diff(out_rb, fx$rb_out) < 1e-5)

# --- Down/upsamplers ---------------------------------------------------------------

ds <- load_group(
  ltx23_video_downsampler3d(8L, 16L, stride = c(2L, 1L, 1L)), "ds"
)
torch::with_no_grad(out_ds <- ds(fx$ds_x, causal = TRUE))
expect_equal(as.integer(out_ds$shape), as.integer(fx$ds_out$shape))
expect_true(max_abs_diff(out_ds, fx$ds_out) < 1e-5)

us <- load_group(
  ltx23_video_upsampler3d(16L, stride = c(2L, 2L, 2L), residual = TRUE, upscale_factor = 2L),
  "us"
)
torch::with_no_grad(out_us <- us(fx$us_x, causal = FALSE))
expect_equal(as.integer(out_us$shape), as.integer(fx$us_out$shape))
expect_true(max_abs_diff(out_us, fx$us_out) < 1e-5)

# --- Tiny encoder (2.3 structure) --------------------------------------------------

enc <- load_group(
  ltx23_video_encoder3d(
    in_channels = 3L,
    out_channels = 4L,
    block_out_channels = c(8L, 16L, 32L, 32L),
    layers_per_block = c(1L, 1L, 1L, 1L, 1L),
    downsample_type = c("spatial", "temporal", "spatiotemporal", "spatiotemporal"),
    patch_size = 4L
  ),
  "enc"
)
torch::with_no_grad(out_enc <- enc(fx$enc_x))
expect_equal(as.integer(out_enc$shape), as.integer(fx$enc_out$shape))
expect_true(max_abs_diff(out_enc, fx$enc_out) < 1e-5)

# --- Tiny decoder (2.3 structure) --------------------------------------------------

dec <- load_group(
  ltx23_video_decoder3d(
    in_channels = 4L,
    out_channels = 3L,
    block_out_channels = c(16L, 32L, 32L, 64L),
    layers_per_block = c(1L, 1L, 1L, 1L, 1L),
    upsample_type = c("spatiotemporal", "spatiotemporal", "temporal", "spatial"),
    upsample_residual = c(FALSE, FALSE, FALSE, FALSE),
    upsample_factor = c(2L, 2L, 1L, 2L),
    patch_size = 4L
  ),
  "dec"
)
torch::with_no_grad(out_dec <- dec(fx$dec_x))
expect_equal(as.integer(out_dec$shape), c(1L, 3L, 17L, 64L, 64L))
expect_true(max_abs_diff(out_dec, fx$dec_out) < 1e-5)

# --- Latent (de)normalization round trip -------------------------------------------

lat <- torch::torch_randn(1L, 4L, 2L, 2L, 2L)
mean <- torch::torch_randn(4L)
std <- torch::torch_rand(4L) + 0.5
lat_n <- ltx23_normalize_latents(lat, mean, std)
lat_rt <- ltx23_denormalize_latents(lat_n, mean, std)
expect_true(max_abs_diff(lat, lat_rt) < 1e-5)

# --- Key mapper --------------------------------------------------------------------

expect_equal(
  ltx23_map_vae_key("vae.decoder.up_blocks.1.conv.conv.weight"),
  "decoder.up_blocks.0.upsamplers.0.conv.conv.weight"
)
expect_equal(
  ltx23_map_vae_key("vae.decoder.up_blocks.0.res_blocks.0.conv1.conv.weight"),
  "decoder.mid_block.resnets.0.conv1.conv.weight"
)
expect_equal(
  ltx23_map_vae_key("vae.decoder.up_blocks.8.res_blocks.1.norm3.weight"),
  "decoder.up_blocks.3.resnets.1.norm3.weight"
)
expect_equal(
  ltx23_map_vae_key("vae.encoder.down_blocks.8.res_blocks.0.conv2.conv.bias"),
  "encoder.mid_block.resnets.0.conv2.conv.bias"
)
expect_equal(
  ltx23_map_vae_key("vae.encoder.down_blocks.3.conv.conv.weight"),
  "encoder.down_blocks.1.downsamplers.0.conv.conv.weight"
)
expect_equal(
  ltx23_map_vae_key("vae.per_channel_statistics.mean-of-means"),
  "latents_mean"
)
expect_equal(ltx23_map_vae_key("vae.encoder.conv_in.conv.weight"), "encoder.conv_in.conv.weight")

# --- Tiled decode ------------------------------------------------------------------

# Reuse the tiny decoder inside a vae wrapper with small tile settings
vae_t <- ltx23_video_vae(
  latent_channels = 4L,
  block_out_channels = c(8L, 8L, 8L, 8L),
  decoder_block_out_channels = c(16L, 32L, 32L, 64L),
  layers_per_block = c(1L, 1L, 1L, 1L, 1L),
  decoder_layers_per_block = c(1L, 1L, 1L, 1L, 1L)
)
# Adopt the fixture decoder weights (same architecture)
dests_t <- vae_t$decoder$named_parameters()
w_dec <- fx[grep("^dec\\.", names(fx))]
names(w_dec) <- sub("^dec\\.", "", names(w_dec))
torch::with_no_grad({
  for (name in names(w_dec)) dests_t[[name]]$copy_(w_dec[[name]])
})
vae_t$eval()

z_big <- torch::torch_randn(1L, 4L, 5L, 4L, 6L)
torch::with_no_grad(ref_full <- vae_t$decode(z_big))
expect_equal(as.integer(ref_full$shape), c(1L, 3L, 33L, 128L, 192L))

# Spatial tiling: tiles of 2x2 latent with stride 1
vae_t$enable_tiling(spatial = TRUE, temporal = FALSE)
vae_t$tile_sample_min_height <- 64L
vae_t$tile_sample_min_width <- 64L
vae_t$tile_sample_stride_height <- 32L
vae_t$tile_sample_stride_width <- 32L
torch::with_no_grad(tiled <- vae_t$decode(z_big))
expect_equal(as.integer(tiled$shape), as.integer(ref_full$shape))
expect_true(as.logical(tiled$isfinite()$all()$item()))
# With random weights, tile borders dominate; value similarity to the
# full decode is only meaningful with trained weights. Instead check the
# degenerate case: a tile that covers the whole input must be identical.
vae_t$tile_sample_min_height <- 128L
vae_t$tile_sample_min_width <- 192L
vae_t$tile_sample_stride_height <- 96L
vae_t$tile_sample_stride_width <- 160L
torch::with_no_grad(tiled_one <- vae_t$decode(z_big))
expect_true(max_abs_diff(tiled_one, ref_full) < 1e-5)
vae_t$tile_sample_min_height <- 64L
vae_t$tile_sample_min_width <- 64L
vae_t$tile_sample_stride_height <- 32L
vae_t$tile_sample_stride_width <- 32L

# Temporal + spatial tiling
vae_t$enable_tiling(spatial = TRUE, temporal = TRUE)
vae_t$tile_sample_min_num_frames <- 16L
vae_t$tile_sample_stride_num_frames <- 8L
torch::with_no_grad(tiled_t <- vae_t$decode(z_big))
expect_equal(as.integer(tiled_t$shape), as.integer(ref_full$shape))
expect_true(as.logical(tiled_t$isfinite()$all()$item()))

# Small inputs bypass tiling entirely (identical output)
z_small <- torch::torch_randn(1L, 4L, 2L, 2L, 2L)
torch::with_no_grad({
  a <- vae_t$decode(z_small)
  vae_t$enable_tiling(spatial = FALSE, temporal = FALSE)
  b <- vae_t$decode(z_small)
})
expect_true(as.numeric((a - b)$abs()$max()) == 0)

# --- Untiled-when-fits dispatch ----------------------------------------------------

# Force-untiled: with tiling enabled and tiles smaller than the input,
# the option must route to the single full-latent forward (identical
# to the untiled reference, no seam blending)
vae_t$enable_tiling(spatial = TRUE, temporal = TRUE)
vae_t$tile_sample_min_height <- 64L
vae_t$tile_sample_min_width <- 64L
vae_t$tile_sample_stride_height <- 32L
vae_t$tile_sample_stride_width <- 32L
vae_t$tile_sample_min_num_frames <- 16L
vae_t$tile_sample_stride_num_frames <- 8L
options(diffuseR.vae_untiled = TRUE)
torch::with_no_grad(forced <- vae_t$decode(z_big))
expect_true(max_abs_diff(forced, ref_full) < 1e-6)

# Force-tiled: FALSE must reproduce the tiled path exactly
options(diffuseR.vae_untiled = FALSE)
torch::with_no_grad(forced_tiled <- vae_t$decode(z_big))
expect_true(max_abs_diff(forced_tiled, tiled_t) < 1e-6)

# "auto" on CPU keeps the tiled path (no VRAM to test against)
options(diffuseR.vae_untiled = NULL)
torch::with_no_grad(auto_cpu <- vae_t$decode(z_big))
expect_true(max_abs_diff(auto_cpu, tiled_t) < 1e-6)

# Traced-decode parity (R/jit_vae_ltx23.R): the per-shape jit_trace
# cache must reproduce the eager video decoder, audio decoder, and
# vocoder exactly, re-trace on new shapes, and honor the off switch.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)
torch::torch_manual_seed(11)

old_opt <- options(diffuseR.jit_vae = TRUE)
on.exit(options(old_opt), add = TRUE)

max_abs_diff <- function(a, b) as.numeric((a - b)$abs()$max())
trace_count <- function() length(ls(diffuseR:::.ltx23_vae_traces))

diffuseR:::.ltx23_release_vae_traces()

# --- Video decoder through the vae wrapper (tiled + direct paths) -------------------

vae <- ltx23_video_vae(
  latent_channels = 4L,
  block_out_channels = c(8L, 8L, 8L, 8L),
  decoder_block_out_channels = c(16L, 32L, 32L, 64L),
  layers_per_block = c(1L, 1L, 1L, 1L, 1L),
  decoder_layers_per_block = c(1L, 1L, 1L, 1L, 1L)
)
vae$eval()

z <- torch::torch_randn(1L, 4L, 3L, 4L, 6L)
torch::with_no_grad({
  op <- options(diffuseR.jit_vae = FALSE)
  ref <- vae$decode(z)
  options(op)
  op <- options(diffuseR.jit_vae = TRUE)
  out <- vae$decode(z)
  options(op)
})
expect_equal(as.integer(out$shape), as.integer(ref$shape))
expect_true(max_abs_diff(out, ref) < 1e-6)
expect_equal(trace_count(), 1L)

# New shape -> new trace, still exact
z2 <- torch::torch_randn(1L, 4L, 3L, 6L, 4L)
torch::with_no_grad({
  op <- options(diffuseR.jit_vae = FALSE)
  ref2 <- vae$decode(z2)
  options(op)
  op <- options(diffuseR.jit_vae = TRUE)
  out2 <- vae$decode(z2)
  options(op)
})
expect_true(max_abs_diff(out2, ref2) < 1e-6)
expect_equal(trace_count(), 2L)

# Same shape again reuses the cache (no growth)
torch::with_no_grad({
  op <- options(diffuseR.jit_vae = TRUE)
  invisible(vae$decode(z))
  options(op)
})
expect_equal(trace_count(), 2L)

# Tiled decode: traced tiles must equal eager tiles exactly
vae$enable_tiling(spatial = TRUE, temporal = TRUE)
vae$tile_sample_min_height <- 64L
vae$tile_sample_min_width <- 64L
vae$tile_sample_stride_height <- 32L
vae$tile_sample_stride_width <- 32L
vae$tile_sample_min_num_frames <- 16L
vae$tile_sample_stride_num_frames <- 8L
z_big <- torch::torch_randn(1L, 4L, 5L, 4L, 6L)
torch::with_no_grad({
  op <- options(diffuseR.jit_vae = FALSE)
  ref_t <- vae$decode(z_big)
  options(op)
  op <- options(diffuseR.jit_vae = TRUE)
  out_t <- vae$decode(z_big)
  options(op)
})
expect_true(max_abs_diff(out_t, ref_t) < 1e-6)

# Release drops every cached trace
diffuseR:::.ltx23_release_vae_traces()
expect_equal(trace_count(), 0L)

# --- Audio decoder ---------------------------------------------------------------------

adec <- ltx23_audio_decoder(
  base_channels = 8L, output_channels = 2L, num_res_blocks = 1L,
  latent_channels = 4L, ch_mult = c(1L, 2L), mel_bins = 8L
)
adec$eval()
za <- torch::torch_randn(1L, 4L, 5L, 2L)
torch::with_no_grad({
  ref_a <- adec(za)
  out_a <- diffuseR:::.ltx23_traced_call(adec, za)
})
expect_true(max_abs_diff(out_a, ref_a) < 1e-6)

# --- Vocoder with BWE ------------------------------------------------------------------

bwe <- ltx23_vocoder_with_bwe(
  in_channels = 8L, hidden_channels = 16L, out_channels = 2L,
  upsample_kernel_sizes = c(4L, 4L), upsample_factors = c(2L, 2L),
  resnet_kernel_sizes = c(3L), resnet_dilations = list(c(1L, 3L)),
  bwe_in_channels = 16L, bwe_hidden_channels = 8L,
  bwe_upsample_kernel_sizes = c(8L, 4L), bwe_upsample_factors = c(4L, 2L),
  bwe_resnet_kernel_sizes = c(3L), bwe_resnet_dilations = list(c(1L, 3L)),
  filter_length = 8L, hop_length = 2L, window_length = 8L,
  num_mel_channels = 8L,
  input_sampling_rate = 100L, output_sampling_rate = 400L
)
bwe$eval()
# Vocoder input is [B, C, T, M] with C * M == in_channels
mel <- torch::torch_randn(1L, 2L, 6L, 4L)
torch::with_no_grad({
  ref_v <- bwe(mel)
  out_v <- diffuseR:::.ltx23_traced_call(bwe, mel)
})
expect_true(max_abs_diff(out_v, ref_v) < 1e-6)

diffuseR:::.ltx23_release_vae_traces()

# progress.R
#
# Shared denoise-loop progress reporting (issue #30). Three verbosity
# levels: "silent" (nothing), "progress" (a one-line generation summary
# plus a progress bar interactively, or periodic step ticks in captured
# logs), "steps" (the full per-phase chatter the generators already
# emit). Logicals map for backwards compatibility: TRUE = "steps",
# FALSE = "silent".

#' Normalize a verbosity flag
#'
#' @param verbose Logical, or one of "silent", "progress", "steps".
#'   TRUE maps to "steps" and FALSE to "silent".
#'
#' @return One of "silent", "progress", "steps".
#'
#' @keywords internal
.verbosity <- function(verbose) {
    if (isTRUE(verbose)) {
        return("steps")
    }
    if (isFALSE(verbose)) {
        return("silent")
    }
    if (is.character(verbose) && length(verbose) == 1L &&
        verbose %in% c("silent", "progress", "steps")) {
        return(verbose)
    }
    stop("'verbose' must be TRUE, FALSE, \"silent\", \"progress\", or \"steps\"",
         call. = FALSE)
}

# Progress reporter for a denoise loop of n steps: returns tick(i, info)
# and done(). At "progress", interactive sessions get a txtProgressBar
# and non-interactive ones a tick line every ~n/10 steps, so captured
# logs stay readable; label (if any) prints once before the loop. With
# per_step = TRUE the caller supplies its own per-step detail through
# tick(info = ) at the "steps" level instead of the bar/ticks.
.denoise_progress <- function(n, label, verbose, per_step = FALSE) {
    level <- .verbosity(verbose)
    if (level == "silent") {
        return(list(tick = function(i, info = NULL) invisible(NULL),
                    done = function() invisible(NULL)))
    }
    if (level == "progress" && !is.null(label) && nzchar(label)) {
        message(label)
    }
    stepwise <- level == "steps" && per_step
    use_bar <- interactive() && !stepwise
    pb <- if (use_bar) {
        utils::txtProgressBar(min = 0, max = n, style = 3)
    } else {
        NULL
    }
    every <- max(1L, as.integer(ceiling(n / 10)))
    tick <- function(i, info = NULL) {
        if (stepwise) {
            message(sprintf("  step %d/%d%s", i, n,
                    if (is.null(info)) "" else paste0(" ", info)))
        } else if (use_bar) {
            utils::setTxtProgressBar(pb, i)
        } else if (i %% every == 0L || i == n) {
            message(sprintf("  step %d/%d", i, n))
        }
        invisible(NULL)
    }
    done <- function() {
        if (use_bar) {
            close(pb)
        }
        invisible(NULL)
    }
    list(tick = tick, done = done)
}

# .verbosity normalization and .denoise_progress output. These tests
# run non-interactively, so the "progress" level takes the periodic
# tick path (the txtProgressBar branch needs interactive()).

verbosity <- diffuseR:::.verbosity
progress <- diffuseR:::.denoise_progress

expect_equal(verbosity(TRUE), "steps")
expect_equal(verbosity(FALSE), "silent")
expect_equal(verbosity("silent"), "silent")
expect_equal(verbosity("progress"), "progress")
expect_equal(verbosity("steps"), "steps")
expect_error(verbosity("loud"))
expect_error(verbosity(1))
expect_error(verbosity(c("progress", "steps")))

run_loop <- function(n, verbose, per_step = FALSE, label = NULL,
                     info = NULL) {
    capture.output({
        pb <- progress(n, label, verbose, per_step = per_step)
        for (i in seq_len(n)) {
            pb$tick(i, info)
        }
        pb$done()
    }, type = "message")
}

# silent: nothing at all
expect_equal(length(run_loop(20, FALSE)), 0L)
expect_equal(length(run_loop(20, "silent", label = "denoise: 20 steps")), 0L)

# progress, non-interactive: label plus a tick every ceiling(n/10) steps
out <- run_loop(20, "progress", label = "denoise: 20 steps")
expect_equal(out[1], "denoise: 20 steps")
expect_equal(out[-1], sprintf("  step %d/20", seq(2, 20, by = 2)))

# the final step always ticks, even off-cadence
out <- run_loop(7, "progress")
expect_true("  step 7/7" %in% out)

# steps with per_step: one line per step carrying the caller's info
out <- run_loop(3, "steps", per_step = TRUE, info = "(sigma 0.9)")
expect_equal(out, sprintf("  step %d/3 (sigma 0.9)", 1:3))

# steps with per_step: label is progress-only, so it does not print
out <- run_loop(3, "steps", per_step = TRUE, label = "stage 1")
expect_equal(out, sprintf("  step %d/3", 1:3))

# steps without per_step (flux-family loops): same ticks as progress
out <- run_loop(20, TRUE)
expect_equal(out, sprintf("  step %d/20", seq(2, 20, by = 2)))

#' LTX-2.3 Spatial Latent Upsampler
#'
#' Fresh R port of the LTX latent upsampler from the diffusers reference
#' (Apache-2.0, pipelines/ltx2/latent_upsampler.py and
#' pipeline_ltx2_latent_upsample.py), with the LTX 2.3 configuration:
#' Conv3d ResBlock stages around a per-frame 2x pixel-shuffle spatial
#' upsampler (no rational resampler). Operates on unnormalized latents.
#'
#' @name upsampler_ltx23
NULL

# Conv -> GroupNorm(32) twice with the activation applied to the residual sum
ltx23_upsampler_res_block <- torch::nn_module(
    "ltx23_upsampler_res_block",
    initialize = function(channels, mid_channels = NULL) {
    mid_channels <- mid_channels %||% channels
    self$conv1 <- torch::nn_conv3d(channels, mid_channels, kernel_size = 3L,
                                   padding = 1L)
    self$norm1 <- torch::nn_group_norm(32L, mid_channels)
    self$conv2 <- torch::nn_conv3d(mid_channels, channels, kernel_size = 3L, padding = 1L)
    self$norm2 <- torch::nn_group_norm(32L, channels)
},
    forward = function(x) {
    residual <- x
    x <- torch::nnf_silu(self$norm1(self$conv1(x)))
    x <- self$norm2(self$conv2(x))
    torch::nnf_silu(x + residual)
}
)

#' LTX-2.3 latent upsampler model
#'
#' Latents [B, 128, F, H, W] -> [B, 128, F, 2H, 2W].
#'
#' @param in_channels Integer. Latent channels.
#' @param mid_channels Integer.
#' @param num_blocks_per_stage Integer.
#'
#' @return Module whose forward(hidden_states) returns the 2x
#'   spatially upscaled latent, a tensor with the same batch, channel
#'   and frame counts and doubled height and width.
#'
#' @export
ltx23_latent_upsampler <- torch::nn_module(
    "ltx23_latent_upsampler",
    initialize = function(
                          in_channels = 128L,
                          mid_channels = 1024L,
                          num_blocks_per_stage = 4L
    ) {
    self$initial_conv <- torch::nn_conv3d(in_channels, mid_channels,
        kernel_size = 3L, padding = 1L)
    self$initial_norm <- torch::nn_group_norm(32L, mid_channels)

    self$res_blocks <- torch::nn_module_list(lapply(seq_len(num_blocks_per_stage),
            function(i) ltx23_upsampler_res_block(mid_channels)))

    # Per-frame 2D conv + pixel shuffle (upsampler.0 in the checkpoint)
    self$upsampler <- torch::nn_module_list(list(
            torch::nn_conv2d(mid_channels, 4L * mid_channels, kernel_size = 3L, padding = 1L)
        ))

    self$post_upsample_res_blocks <- torch::nn_module_list(
        lapply(seq_len(num_blocks_per_stage),
               function(i) ltx23_upsampler_res_block(mid_channels))
    )
    self$final_conv <- torch::nn_conv3d(mid_channels, in_channels,
                                        kernel_size = 3L, padding = 1L)
},
    forward = function(hidden_states) {
    batch_size <- hidden_states$shape[1]

    hidden_states <- torch::nnf_silu(self$initial_norm(self$initial_conv(hidden_states)))
    for (i in seq_along(self$res_blocks)) {
        hidden_states <- self$res_blocks[[i]](hidden_states)
    }

    # [B, C, F, H, W] -> per-frame [B*F, C, H, W], 2x pixel shuffle, back
    hidden_states <- hidden_states$permute(c(1L, 3L, 2L, 4L, 5L))$
    flatten(start_dim = 1L, end_dim = 2L)
    hidden_states <- self$upsampler[[1]](hidden_states)
    hidden_states <- torch::nnf_pixel_shuffle(hidden_states, 2L)
    hidden_states <- hidden_states$unflatten(1L, c(batch_size, -1L))$
    permute(c(1L, 3L, 2L, 4L, 5L))

    for (i in seq_along(self$post_upsample_res_blocks)) {
        hidden_states <- self$post_upsample_res_blocks[[i]](hidden_states)
    }
    self$final_conv(hidden_states)
}
)

#' Load the LTX-2.3 spatial upscaler weights
#'
#' The checkpoint keys match this module tree directly.
#'
#' @param path Path to e.g. \code{ltx-2.3-spatial-upscaler-x2-1.1.safetensors}.
#' @param device,dtype Placement for the loaded model.
#' @param verbose Logical.
#'
#' @return The loaded \code{ltx23_latent_upsampler}.
#'
#' @export
ltx23_load_upsampler <- function(path, device = "cuda", dtype = "bfloat16",
                                 verbose = TRUE) {
    torch_dtype <- switch(dtype, bfloat16 = torch::torch_bfloat16(),
                          float16 = torch::torch_float16(),
                          float32 = torch::torch_float32(),
                          stop("Unsupported dtype: ", dtype))
    handle <- safetensors::safetensors$new(path.expand(path), framework = "torch")
    keys <- setdiff(handle$keys(), "__metadata__")

    model <- ltx23_latent_upsampler()
    model$to(dtype = torch_dtype)
    dests <- c(model$named_parameters(), model$named_buffers())

    missing_dest <- setdiff(keys, names(dests))
    unfilled <- setdiff(names(dests), keys)
    if (length(missing_dest) || length(unfilled)) {
        stop("Upsampler load mismatch: ", length(missing_dest), " unmapped, ",
             length(unfilled), " unfilled")
    }
    torch::with_no_grad({
        for (key in keys) {
            dests[[key]]$copy_(handle$get_tensor(key))
        }
    })
    model$to(device = device)
    model$eval()
    if (verbose) {
        message("Loaded latent upsampler (", length(keys), " tensors)")
    }
    model
}

#' Adaptive instance normalization between latent tensors
#'
#' Matches each (batch, channel) slice's mean/std to the reference
#' latents, blended by \code{factor} (cf. diffusers
#' \code{LTX2LatentUpsamplePipeline.adain_filter_latent}).
#'
#' @param latents Tensor [B, C, F, H, W].
#' @param reference_latents Tensor with the target statistics.
#' @param factor Numeric blend in [-10, 10]; 0 is identity.
#'
#' @return Filtered latents.
#'
#' @export
ltx23_adain_filter_latent <- function(latents, reference_latents,
                                      factor = 1.0) {
    if (factor == 0) {
        return(latents)
    }
    dims <- c(3L, 4L, 5L)
    r_mean <- reference_latents$mean(dim = dims, keepdim = TRUE)
    r_sd <- reference_latents$std(dim = dims, keepdim = TRUE)
    i_mean <- latents$mean(dim = dims, keepdim = TRUE)
    i_sd <- latents$std(dim = dims, keepdim = TRUE)

    result <- (latents - i_mean)$div(i_sd)$mul(r_sd) + r_mean
    torch::torch_lerp(latents, result, factor)
}

#' Sigmoid tone mapping for latents
#'
#' Compresses the latent dynamic range (cf. diffusers
#' \code{tone_map_latents}). \code{compression} 0 is identity, 1 is the
#' full effect.
#'
#' @param latents Tensor.
#' @param compression Numeric in [0, 1].
#'
#' @return Tone-mapped latents.
#'
#' @export
ltx23_tone_map_latents <- function(latents, compression) {
    if (compression == 0) {
        return(latents)
    }
    stopifnot(compression >= 0, compression <= 1)
    scale_factor <- compression * 0.75
    abs_latents <- torch::torch_abs(latents)
    sigmoid_term <- torch::torch_sigmoid(abs_latents$add(-1)$mul(4 * scale_factor))
    scales <- sigmoid_term$mul(-0.8 * scale_factor)$add(1)
    latents * scales
}

#' SentencePiece Unigram Tokenizer
#'
#' Pure R implementation of HuggingFace tokenizer.json files with a
#' Unigram model (SentencePiece), as used by T5 - FLUX's second text
#' encoder. Segmentation is Viterbi best-path over the vocab log
#' probabilities (Kudo 2018, arXiv:1804.10959). The normalizer and
#' Metaspace pre-tokenizer settings are read from the file.
#'
#' Limitation: the Precompiled charsmap normalizer (NFKC-style unicode
#' mapping) is approximated by control-whitespace substitution only;
#' ASCII and common latin text tokenizes identically to the reference,
#' exotic unicode may differ.
#'
#' @name tokenizer_unigram
NULL

#' Load a Unigram tokenizer from tokenizer.json
#'
#' @param tokenizer_path Path to a HuggingFace tokenizer.json with a
#'   Unigram model, or a directory containing one.
#'
#' @return A \code{unigram_tokenizer} object.
#'
#' @export
unigram_tokenizer <- function(tokenizer_path) {
    path <- path.expand(tokenizer_path)
    if (dir.exists(path)) {
        path <- file.path(path, "tokenizer.json")
    }
    if (!file.exists(path)) {
        stop("tokenizer.json not found: ", path)
    }

    tj <- jsonlite::fromJSON(path, simplifyVector = FALSE)
    model <- tj$model
    if (is.null(model) || !identical(model$type, "Unigram")) {
        stop("Only Unigram tokenizers are supported (got ",
             model$type %||% "none", "); use bpe_tokenizer() for BPE.")
    }

    pieces <- vapply(model$vocab, function(p) p[[1]], character(1))
    scores <- vapply(model$vocab, function(p) as.numeric(p[[2]]), numeric(1))
    ids0 <- seq_along(pieces) - 1L

    vocab_env <- new.env(parent = emptyenv(), size = length(pieces))
    for (i in seq_along(pieces)) {
        assign(pieces[i], c(ids0[i], scores[i]), envir = vocab_env)
    }

    # Normalizer settings (Sequence or single normalizer)
    norms <- tj$normalizer
    if (!is.null(norms) && identical(norms$type, "Sequence")) {
        norms <- norms$normalizers
    } else if (!is.null(norms)) {
        norms <- list(norms)
    } else {
        norms <- list()
    }
    strip_right <- FALSE
    space_collapse <- NULL
    for (n in norms) {
        if (identical(n$type, "Strip") && isTRUE(n$strip_right)) {
            strip_right <- TRUE
        }
        if (identical(n$type, "Replace") &&
            identical(n$pattern$Regex, " {2,}")) {
            space_collapse <- n$content
        }
    }

    # Metaspace pre-tokenizer (possibly inside a Sequence)
    pre <- tj$pre_tokenizer
    pres <- if (!is.null(pre) && identical(pre$type, "Sequence")) {
        pre$pretokenizers
    } else if (!is.null(pre)) {
        list(pre)
    } else {
        list()
    }
    prepend_scheme <- "never"
    replacement <- "\u2581"
    for (p in pres) {
        if (identical(p$type, "Metaspace")) {
            replacement <- p$replacement %||% "\u2581"
            prepend_scheme <- p$prepend_scheme %||%
            (if (isTRUE(p$add_prefix_space)) "always" else "never")
        }
    }

    unk_id <- as.integer(model$unk_id %||% 2L)
    eos <- get0("</s>", envir = vocab_env)
    pad <- get0("<pad>", envir = vocab_env)

    structure(
              list(
                   vocab = vocab_env,
                   n_pieces = length(pieces),
                   max_piece_chars = max(nchar(pieces)),
                   unk_id = unk_id,
                   unk_score = min(scores) - 10.0,
                   eos_id = if (!is.null(eos)) as.integer(eos[1]) else 1L,
                   pad_id = if (!is.null(pad)) as.integer(pad[1]) else 0L,
                   strip_right = strip_right,
                   space_collapse = space_collapse,
                   prepend_scheme = prepend_scheme,
                   replacement = replacement,
                   path = path
        ),
              class = "unigram_tokenizer"
    )
}

#' @export
print.unigram_tokenizer <- function(x, ...) {
    cat("<unigram_tokenizer>\n")
    cat("  pieces:  ", x$n_pieces, "\n")
    cat("  unk/eos/pad:", x$unk_id, x$eos_id, x$pad_id, "\n")
    cat("  path:    ", x$path, "\n")
    invisible(x)
}

# Viterbi best-path segmentation of one pre-token (0-based ids)
.unigram_viterbi <- function(word, tokenizer) {
    n <- nchar(word)
    if (n == 0L) {
        return(integer(0))
    }
    vocab <- tokenizer$vocab
    max_len <- tokenizer$max_piece_chars

    best <- c(0, rep(-Inf, n))
    back_len <- integer(n)
    back_id <- integer(n)

    for (end in seq_len(n)) {
        for (len in seq_len(min(max_len, end))) {
            start <- end - len + 1L
            piece <- substr(word, start, end)
            entry <- get0(piece, envir = vocab)
            if (is.null(entry)) {
                if (len > 1L) {
                    next
                }
                id <- tokenizer$unk_id
                score <- tokenizer$unk_score
            } else {
                id <- entry[1]
                score <- entry[2]
            }
            cand <- best[start] + score
            if (cand > best[end + 1L]) {
                best[end + 1L] <- cand
                back_len[end] <- len
                back_id[end] <- id
            }
        }
    }

    ids <- integer(0)
    pos <- n
    while (pos > 0L) {
        ids <- c(back_id[pos], ids)
        pos <- pos - back_len[pos]
    }
    ids
}

#' Encode text with a Unigram tokenizer
#'
#' Normalizes (strip-right, multi-space collapse, control whitespace to
#' space), applies the Metaspace pre-tokenizer, segments each pre-token
#' by Viterbi over the Unigram scores, fuses consecutive unknowns, and
#' appends EOS. T5 semantics: right padding with \code{<pad>} (id 0),
#' truncation to \code{max_length - 1} before the EOS.
#'
#' @param tokenizer A \code{\link{unigram_tokenizer}}.
#' @param texts Character vector of prompts.
#' @param max_length Integer. Fixed sequence length (NULL for no
#'   truncation/padding).
#' @param add_eos Logical. Append the EOS token.
#' @param pad Logical. Right-pad to \code{max_length}.
#'
#' @return List with \code{input_ids} and \code{attention_mask}, each an
#'   integer matrix [length(texts), max_length] (or ragged lists when
#'   \code{max_length} is NULL). Ids are 0-based (HuggingFace
#'   convention); add 1 for R torch embedding lookups.
#'
#' @export
encode_unigram <- function(tokenizer, texts, max_length = 256L,
                           add_eos = TRUE, pad = TRUE) {
    stopifnot(inherits(tokenizer, "unigram_tokenizer"))
    rep_char <- tokenizer$replacement

    encode_one <- function(text) {
        # Control whitespace to space (charsmap approximation), then the
        # file's normalizer chain
        text <- enc2utf8(text)
        text <- gsub("[\t\n\r\f\v]", " ", text)
        if (tokenizer$strip_right) {
            text <- sub("[ ]+$", "", text)
        }
        if (!is.null(tokenizer$space_collapse)) {
            text <- gsub(" {2,}", tokenizer$space_collapse, text)
        }
        # Metaspace: spaces to the replacement, optional prefix.
        # Empty input yields no pre-tokens (before any prepend).
        text <- gsub(" ", rep_char, text, fixed = TRUE)
        if (nchar(text) == 0L) {
            return(integer(0))
        }
        if (tokenizer$prepend_scheme == "always" &&
            !startsWith(text, rep_char)) {
            text <- paste0(rep_char, text)
        }
        # Split before each replacement char, keeping it attached.
        # (strsplit with a zero-width lookahead detaches the char, so
        # mark boundaries with a control byte instead.)
        marked <- gsub(rep_char, paste0("\u0001", rep_char), text, fixed = TRUE)
        words <- strsplit(marked, "\u0001", fixed = TRUE)[[1]]
        words <- words[nzchar(words)]

        ids <- unlist(lapply(words, .unigram_viterbi, tokenizer = tokenizer))
        ids <- as.integer(ids %||% integer(0))

        # Fuse consecutive unknowns (HF fuse_unk)
        if (length(ids) > 1L) {
            is_unk <- ids == tokenizer$unk_id
            drop <- is_unk & c(FALSE, is_unk[-length(is_unk)])
            ids <- ids[!drop]
        }
        ids
    }

    all_ids <- lapply(as.character(texts), encode_one)

    if (add_eos) {
        if (is.null(max_length)) {
            keep <- Inf
        } else {
            keep <- max_length - 1L
        }
        all_ids <- lapply(all_ids, function(ids) {
            if (length(ids) > keep) {
                ids <- ids[seq_len(keep)]
            }
            c(ids, tokenizer$eos_id)
        })
    } else if (!is.null(max_length)) {
        all_ids <- lapply(all_ids, function(ids) {
            if (length(ids) > max_length) ids[seq_len(max_length)] else ids
        })
    }

    if (is.null(max_length) || !pad) {
        masks <- lapply(all_ids, function(ids) rep(1L, length(ids)))
        return(list(input_ids = all_ids, attention_mask = masks))
    }

    n <- length(all_ids)
    input_ids <- matrix(tokenizer$pad_id, nrow = n, ncol = max_length)
    attention_mask <- matrix(0L, nrow = n, ncol = max_length)
    for (i in seq_len(n)) {
        len <- length(all_ids[[i]])
        if (len > 0L) {
            input_ids[i, seq_len(len)] <- all_ids[[i]]
            attention_mask[i, seq_len(len)] <- 1L
        }
    }
    list(input_ids = input_ids, attention_mask = attention_mask)
}

# Exact token-id parity for the Unigram (T5/SentencePiece) tokenizer
# against HuggingFace tokenizers reference cases (generated by
# tools/gen_t5_tokenizer_cases.py, checked in).
#
# The tokenizer.json itself is not shipped (2.4 MB, and FLUX's copy is
# behind the gated repo); the test locates one and skips if absent.

library(diffuseR)

find_t5_tokenizer <- function() {
  # 1. Explicit override
  p <- Sys.getenv("DIFFUSER_T5_TOKENIZER", "")
  if (nzchar(p) && file.exists(p)) {
    return(p)
  }
  # 2. FLUX.1-schnell HF cache (post-download_flux1)
  if (requireNamespace("hfhub", quietly = TRUE)) {
    p <- tryCatch(
      suppressMessages(hfhub::hub_download(
        "black-forest-labs/FLUX.1-schnell",
        "tokenizer_2/tokenizer.json", local_files_only = TRUE
      )),
      error = function(e) ""
    )
    if (nzchar(p) && file.exists(p)) {
      return(p)
    }
  }
  # 3. Dev copy in the source tree (tools/gen_t5_tokenizer_cases.py)
  p <- "../../tools/cache/tokenizer_t5.json"
  if (file.exists(p)) {
    return(p)
  }
  ""
}

tok_path <- find_t5_tokenizer()
if (!nzchar(tok_path)) exit_file("no T5 tokenizer.json available")

cases_path <- system.file("tinytest", "fixtures", "t5_tokenizer_cases.json",
  package = "diffuseR")
if (cases_path == "") cases_path <- "fixtures/t5_tokenizer_cases.json"
if (!file.exists(cases_path)) exit_file("t5 tokenizer cases missing")

cases <- jsonlite::fromJSON(cases_path, simplifyVector = FALSE)

tok <- unigram_tokenizer(tok_path)
expect_equal(tok$n_pieces, 32100L)
expect_equal(tok$eos_id, 1L)
expect_equal(tok$pad_id, 0L)
expect_equal(tok$unk_id, 2L)

# --- raw encodings (no padding/truncation) --------------------------------------

for (case in cases$cases) {
  expected <- as.integer(unlist(case$ids))
  got <- encode_unigram(tok, case$text, max_length = NULL, pad = FALSE)
  expect_equal(got$input_ids[[1]], expected,
    info = sprintf("text: %s", substr(case$text, 1, 60)))
}

# --- padding + truncation at max_length -------------------------------------------

for (case in cases$padded) {
  got <- encode_unigram(tok, case$text, max_length = case$max_length,
    pad = TRUE)
  expect_equal(as.integer(got$input_ids[1, ]), as.integer(unlist(case$ids)),
    info = sprintf("padded text: %s", substr(case$text, 1, 60)))
  expect_equal(as.integer(got$attention_mask[1, ]),
    as.integer(unlist(case$mask)))
}

# --- batch shape -------------------------------------------------------------------

batch <- encode_unigram(tok, c("a photo of a cat", "hello"),
  max_length = 32L)
expect_equal(dim(batch$input_ids), c(2L, 32L))
expect_equal(dim(batch$attention_mask), c(2L, 32L))

#' Load a specific component of a diffusion model
#'
#' Loads a TorchScript model component (UNet, decoder, or text encoder) from the
#' hfhub cache or legacy model directory, downloading it first if necessary.
#'
#' @param component Character string, the component to load: "unet", "decoder", or "text_encoder".
#' @param model_name Character string, the name of the model to use.
#' @param device Character string, the torch device to load the model onto ("cpu" or "cuda").
#' @param unet_dtype_str Optional; the data type for the UNet model. If `NULL`, defaults to `float32` for CPU and `float16` for CUDA.
#' @param download Logical; if `TRUE` (default), downloads the model if it doesn't exist locally.
#' @param use_native Logical; if `TRUE`, uses native R torch modules instead of TorchScript.
#'   Supported for unet, decoder, text_encoder, and text_encoder2. Native modules have
#'   better GPU compatibility (especially on Blackwell/RTX 50xx).
#'
#' @return A torch model object.
#' @export
#'
#' @examples
#' \dontrun{
#' unet <- load_model_component("unet", "sd21", "cpu")
#' }
load_model_component <- function(component, model_name = "sd21",
                                 device = "cpu", unet_dtype_str = NULL,
                                 download = TRUE, use_native = FALSE) {
    # Set valid components based on model
    if (model_name == "sdxl") {
        valid_components <- c("unet", "decoder", "text_encoder",
                              "text_encoder2", "encoder")
    } else {
        valid_components <- c("unet", "decoder", "text_encoder", "encoder")
    }

    if (!component %in% valid_components) {
        stop("Invalid component name for model '", model_name, "'. Must be one of: ",
             paste(valid_components, collapse = ", "))
    }

    # Determine filename for this component
    if (use_native) {
        # For SDXL UNet, always use CUDA file (CPU file is often corrupted/truncated)
        if (component == "unet" && model_name == "sdxl") {
            filename <- component_filename("unet", "cuda", unet_dtype_str)
        } else {
            filename <- component_filename(component, "cpu", unet_dtype_str)
        }
    } else {
        filename <- component_filename(component, device, unet_dtype_str)
    }

    # Resolve file path via hfhub (cache, legacy, or download)
    file_path <- hf_download_pt(model_name, filename, download = download)

    # Load the model
    if (use_native && component == "unet") {
        if (model_name == "sdxl") {
            model <- unet_sdxl_native_from_torchscript(file_path, verbose = FALSE)
        } else {
            model <- unet_native_from_torchscript(file_path, verbose = FALSE)
        }
        if (device == "cuda" &&
            (is.null(unet_dtype_str) || unet_dtype_str == "float16")) {
            model$to(device = torch::torch_device(device), dtype = torch::torch_float16())
        } else {
            model$to(device = torch::torch_device(device))
        }
    } else if (use_native && component == "decoder") {
        model <- vae_decoder_native()
        load_decoder_weights(model, file_path, verbose = FALSE)
        model$to(device = torch::torch_device(device))
    } else if (use_native && component == "text_encoder") {
        arch <- detect_text_encoder_architecture(file_path)
        model <- text_encoder_native(
                                     vocab_size = arch$vocab_size,
                                     context_length = arch$context_length,
                                     embed_dim = arch$embed_dim,
                                     num_layers = arch$num_layers,
                                     num_heads = arch$num_heads,
                                     mlp_dim = arch$mlp_dim,
                                     apply_final_ln = arch$apply_final_ln
        )
        load_text_encoder_weights(model, file_path, verbose = FALSE)
        model$to(device = torch::torch_device(device))
    } else if (use_native && component == "text_encoder2") {
        arch <- detect_text_encoder_architecture(file_path)
        model <- text_encoder2_native(
                                      vocab_size = arch$vocab_size,
                                      context_length = arch$context_length,
                                      embed_dim = arch$embed_dim,
                                      num_layers = arch$num_layers,
                                      num_heads = arch$num_heads,
                                      mlp_dim = arch$mlp_dim
        )
        load_text_encoder2_weights(model, file_path, verbose = FALSE)
        model$to(device = torch::torch_device(device))
    } else {
        model <- torch::jit_load(file_path, map_location = torch::torch_device(device))
    }

    model
}

# Build the expected filename for a component
component_filename <- function(component, device, unet_dtype_str = NULL) {
    if (component == "unet" && device != "cpu") {
        dtype <- if (is.null(unet_dtype_str) || unet_dtype_str == "float16") {
            "float16"
        } else if (unet_dtype_str == "float32") {
            "float32"
        } else {
            stop("Invalid unet_dtype_str: must be 'float16' or 'float32'")
        }
        paste0("unet-cuda-", dtype, ".pt")
    } else {
        paste0(component, "-", device, ".pt")
    }
}

# Convenience function to load both text encoders for SDXL
load_text_encoders <- function(model_name = "sdxl", device = "cpu",
                               download = TRUE) {
    if (model_name != "sdxl") {
        return(list(
                    text_encoder = load_model_component("text_encoder",
                    model_name, device, download = download)
            ))
    }

    text_encoder <- load_model_component("text_encoder", model_name, device, download = download)
    text_encoder2 <- load_model_component("text_encoder2", model_name, device, download = download)

    list(text_encoder = text_encoder, text_encoder2 = text_encoder2)
}

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

# Test native VAE decoder

# Skip if no GPU or models not available
skip_if_not <- function(cond, msg) {
  if (!cond) exit_file(msg)
}

# Test module architecture
decoder <- vae_decoder_native()
expect_true(inherits(decoder, "nn_module"))

# Check parameters exist
params <- names(decoder$parameters)
expect_true(length(params) > 0)
expect_true("conv_in.weight" %in% params)
expect_true("conv_out.weight" %in% params)

# Test forward pass on CPU with random weights
library(torch)
with_no_grad({
  test_input <- torch_randn(c(1, 4, 8, 8))
  out <- decoder(test_input)
  expect_equal(out$shape[1], 1L)
  expect_equal(out$shape[2], 3L)  # RGB output
  expect_equal(out$shape[3], 64L)  # 8x upscaling
  expect_equal(out$shape[4], 64L)
})

# Test load_model_component with use_native
model_dir <- tools::R_user_dir("diffuseR", "data")
decoder_file <- file.path(model_dir, "sdxl", "decoder-cpu.pt")
skip_if_not(file.exists(decoder_file), "SDXL decoder not available")

native_dec <- load_model_component("decoder", "sdxl", "cpu", use_native = TRUE)
expect_true(inherits(native_dec, "VAEDecoderNative"))

# Verify weight loading (parameter count)
expect_equal(length(native_dec$parameters), 138L)

#' LTX-2.3 Rotary Positional Embeddings
#'
#' Fresh R port of the LTX rotary positional embedding scheme from the
#' diffusers reference implementation (Apache-2.0,
#' src/diffusers/models/transformers/transformer_ltx2.py). LTX 2.3 uses
#' the "split" RoPE layout everywhere; "interleaved" is kept for
#' completeness. Frequencies are computed in float64 per the checkpoint
#' config (frequencies_precision) and applied in float32.
#'
#' @name rope_ltx23
NULL

#' Apply interleaved rotary embeddings
#'
#' Rotates adjacent element pairs of the last dimension:
#' \code{out = x * cos + rotate_half(x) * sin} with pairs interleaved
#' (elements 1,2 form the first complex pair).
#'
#' @param x Tensor of shape [B, S, C].
#' @param freqs List of two tensors (cos, sin), each [B, S, C].
#'
#' @return Tensor with the same shape and dtype as \code{x}.
#'
#' @export
ltx23_apply_interleaved_rotary_emb <- function(x, freqs) {
    cos <- freqs[[1]]
    sin <- freqs[[2]]

    # Split last dim into (C/2, 2) pairs: real = even positions, imag = odd
    pairs <- x$unflatten(3L, c(-1L, 2L))
    x_real <- pairs[,,, 1] # [B, S, C/2]
    x_imag <- pairs[,,, 2]
    x_rotated <- torch::torch_stack(list(-x_imag, x_real), dim = -1L)$flatten(start_dim = 3L)

    out <- x$to(dtype = torch::torch_float32()) * cos +
    x_rotated$to(dtype = torch::torch_float32()) * sin
    out$to(dtype = x$dtype)
}

#' Apply split rotary embeddings
#'
#' Rotates element pairs formed by splitting the last dimension in half:
#' element i pairs with element i + d/2. The cos/sin tensors carry half
#' the head dimension.
#'
#' @param x Tensor of shape [B, H, T, D] (per-head layout), or [B, T, H*D]
#'   which is reshaped per-head when \code{freqs} is 4D.
#' @param freqs List of two tensors (cos, sin), each [B, H, T, D/2].
#'
#' @return Tensor with the same shape and dtype as \code{x}.
#'
#' @export
ltx23_apply_split_rotary_emb <- function(x, freqs) {
    cos <- freqs[[1]]
    sin <- freqs[[2]]

    x_dtype <- x$dtype
    needs_reshape <- FALSE
    if (x$ndim != 4L && cos$ndim == 4L) {
        # cos is [B, H, T, r] -> reshape x [B, T, H*D] to [B, H, T, D]
        b <- cos$shape[1]
        h <- cos$shape[2]
        t <- cos$shape[3]
        x <- x$reshape(c(b, t, h, -1L))$transpose(2L, 3L)
        needs_reshape <- TRUE
    }

    d <- x$shape[length(x$shape)]
    if (d %% 2L != 0L) {
        stop("Expected the last dimension of x to be even for split rotary, got ",
             d)
    }
    r <- d %/% 2L

    # First/second halves of the last dim form the rotation pairs
    x_first <- x$narrow(-1L, 1L, r)$to(dtype = torch::torch_float32())
    x_second <- x$narrow(-1L, r + 1L, r)$to(dtype = torch::torch_float32())

    out_first <- x_first * cos - x_second * sin
    out_second <- x_second * cos + x_first * sin
    out <- torch::torch_cat(list(out_first, out_second), dim = -1L)

    if (needs_reshape) {
        out <- out$transpose(2L, 3L)$reshape(c(b, t, -1L))
    }

    out$to(dtype = x_dtype)
}

#' LTX-2.3 audio/video rotary position embedder
#'
#' Computes RoPE cos/sin frequency tensors from spatiotemporal patch
#' coordinates. Video coordinates are 3D (frames scaled to seconds via
#' fps, height, width in pixel space); audio coordinates are 1D
#' (seconds). Coordinates are patch boundaries [start, end); the midpoint
#' is used as the position.
#'
#' @param dim Integer. Rotary dimension (attention head dim x heads for
#'   split type at model level; see reference).
#' @param patch_size,patch_size_t Integers. Spatial/temporal patch sizes.
#' @param base_num_frames,base_height,base_width Integers. Base grid the
#'   coordinates are normalized against.
#' @param sampling_rate,hop_length Integers. Audio spectrogram params.
#' @param scale_factors Integer vector. VAE (time, height, width) scale
#'   factors.
#' @param theta Numeric. RoPE theta.
#' @param causal_offset Integer. Temporal offset for the causal VAE
#'   (first frame has stride 1).
#' @param modality "video" or "audio".
#' @param double_precision Logical. Compute base frequencies in float64.
#' @param rope_type "split" (LTX 2.3) or "interleaved".
#' @param num_attention_heads Integer. Needed for the split layout.
#'
#' @return Module whose forward(coords, device) returns
#'   \code{list(cos_freqs, sin_freqs)}, the two rotary tables to apply
#'   to queries and keys.
#'
#' @export
ltx23_rotary_pos_embed <- torch::nn_module(
    "ltx23_rotary_pos_embed",
    initialize = function(
                          dim,
                          patch_size = 1L,
                          patch_size_t = 1L,
                          base_num_frames = 20L,
                          base_height = 2048L,
                          base_width = 2048L,
                          sampling_rate = 16000L,
                          hop_length = 160L,
                          scale_factors = c(8L, 32L, 32L),
                          theta = 10000.0,
                          causal_offset = 1L,
                          modality = "video",
                          double_precision = TRUE,
                          rope_type = "split",
                          num_attention_heads = 32L
    ) {
    if (!rope_type %in% c("interleaved", "split")) {
        stop("rope_type must be 'interleaved' or 'split', got: ", rope_type)
    }
    if (!modality %in% c("video", "audio")) {
        stop("modality must be 'video' or 'audio', got: ", modality)
    }
    self$dim <- as.integer(dim)
    self$patch_size <- as.integer(patch_size)
    self$patch_size_t <- as.integer(patch_size_t)
    self$base_num_frames <- base_num_frames
    self$base_height <- base_height
    self$base_width <- base_width
    self$sampling_rate <- sampling_rate
    self$hop_length <- hop_length
    self$scale_factors <- as.integer(scale_factors)
    self$theta <- theta
    self$causal_offset <- causal_offset
    self$modality <- modality
    self$double_precision <- double_precision
    self$rope_type <- rope_type
    self$num_attention_heads <- as.integer(num_attention_heads)
},

    # Patch boundaries [B, 3, num_patches, 2] in pixel/second space
    prepare_video_coords = function(batch_size, num_frames, height, width,
                                    device, fps = 24.0) {
    f32 <- torch::torch_float32()
    # torch_arange has an inclusive end; end = n - 1 matches Python's
    # exclusive arange for integer grids
    grid_f <- torch::torch_arange(start = 0, end = num_frames - 1,
                                  step = self$patch_size_t, dtype = f32,
                                  device = device)
    grid_h <- torch::torch_arange(
                                  start = 0, end = height - 1, step = self$patch_size,
                                  dtype = f32, device = device
    )
    grid_w <- torch::torch_arange(
                                  start = 0, end = width - 1, step = self$patch_size,
                                  dtype = f32, device = device
    )
    grid <- torch::torch_meshgrid(list(grid_f, grid_h, grid_w), indexing = "ij")
    grid <- torch::torch_stack(grid, dim = 1L) # [3, NF, NH, NW]

    patch_size <- c(self$patch_size_t, self$patch_size, self$patch_size)
    patch_delta <- torch::torch_tensor(patch_size, dtype = grid$dtype, device = grid$device)
    patch_ends <- grid + patch_delta$view(c(3L, 1L, 1L, 1L))

    latent_coords <- torch::torch_stack(list(grid, patch_ends), dim = -1L) # [3,NF,NH,NW,2]
    latent_coords <- latent_coords$flatten(start_dim = 2L, end_dim = 4L) # [3,N,2]
    latent_coords <- latent_coords$unsqueeze(1L)$`repeat`(c(batch_size, 1L, 1L, 1L))

    scale_tensor <- torch::torch_tensor(self$scale_factors, device = latent_coords$device)
    pixel_coords <- latent_coords * scale_tensor$view(c(1L, -1L, 1L, 1L))

    # First latent frame has temporal stride 1 in the causal VAE: shift and
    # clamp so timestamps stay causal and non-negative
    pixel_coords[, 1,,] <- (pixel_coords[, 1,,] + self$causal_offset -
        self$scale_factors[1])$clamp(min = 0)
    # Temporal coordinates in seconds
    pixel_coords[, 1,,] <- pixel_coords[, 1,,] / fps

    pixel_coords
},

    # Patch boundaries [B, 1, num_patches, 2] in seconds
    prepare_audio_coords = function(batch_size, num_frames, device, shift = 0L) {
    f32 <- torch::torch_float32()
    grid_f <- torch::torch_arange(
                                  start = shift, end = num_frames + shift - 1, step = self$patch_size_t,
                                  dtype = f32, device = device
    )

    audio_scale_factor <- self$scale_factors[1]
    grid_start_mel <- grid_f * audio_scale_factor
    grid_start_mel <- (grid_start_mel + self$causal_offset - audio_scale_factor)$clamp(min = 0)
    grid_start_s <- grid_start_mel * self$hop_length / self$sampling_rate

    grid_end_mel <- (grid_f + self$patch_size_t) * audio_scale_factor
    grid_end_mel <- (grid_end_mel + self$causal_offset - audio_scale_factor)$clamp(min = 0)
    grid_end_s <- grid_end_mel * self$hop_length / self$sampling_rate

    audio_coords <- torch::torch_stack(list(grid_start_s, grid_end_s), dim = -1L) # [N, 2]
    audio_coords <- audio_coords$unsqueeze(1L)$expand(c(batch_size, -1L, -1L))
    audio_coords$unsqueeze(2L) # [B, 1, N, 2]
},

    forward = function(coords, device = NULL) {
    device <- device %||% coords$device
    num_pos_dims <- coords$shape[2]

    # Patch boundaries [start, end) -> midpoint position
    if (coords$ndim == 4L) {
        chunks <- torch::torch_chunk(coords, 2L, dim = -1L)
        coords <- ((chunks[[1]] + chunks[[2]]) / 2.0)$squeeze(-1L) # [B, dims, N]
    }

    max_positions <- if (self$modality == "video") {
        c(self$base_num_frames, self$base_height, self$base_width)
    } else {
        self$base_num_frames
    }
    # [B, dims, N] -> [B, N, dims], each dim normalized to its base size
    grid <- torch::torch_stack(
                               lapply(seq_len(num_pos_dims), function(i) coords[, i,] / max_positions[i]),
                               dim = -1L
    )$to(device = device)

    num_rope_elems <- num_pos_dims * 2L

    freqs_dtype <- if (self$double_precision) torch::torch_float64() else torch::torch_float32()
    pow_indices <- torch::torch_pow(
                                    self$theta,
                                    torch::torch_linspace(
            start = 0.0, end = 1.0, steps = self$dim %/% num_rope_elems,
            dtype = freqs_dtype, device = device
        )
    )
    freqs <- (pow_indices * pi / 2.0)$to(dtype = torch::torch_float32())

    # Outer product of normalized positions (mapped to [-1, 1]) and freqs:
    # [B, N, dims, dim / num_rope_elems]
    freqs <- (grid$unsqueeze(-1L) * 2 - 1) * freqs
    freqs <- freqs$transpose(-1L, -2L)$flatten(start_dim = 3L) # [B, N, dim/2]

    if (self$rope_type == "interleaved") {
        cos_freqs <- freqs$cos()$repeat_interleave(2L, dim = -1L)
        sin_freqs <- freqs$sin()$repeat_interleave(2L, dim = -1L)

        pad <- self$dim %% num_rope_elems
        if (pad != 0L) {
            cos_padding <- torch::torch_ones_like(cos_freqs[,, 1:pad])
            sin_padding <- torch::torch_zeros_like(cos_freqs[,, 1:pad])
            cos_freqs <- torch::torch_cat(list(cos_padding, cos_freqs), dim = -1L)
            sin_freqs <- torch::torch_cat(list(sin_padding, sin_freqs), dim = -1L)
        }
    } else {
        expected_freqs <- self$dim %/% 2L
        current_freqs <- freqs$shape[length(freqs$shape)]
        pad_size <- expected_freqs - current_freqs
        cos_freqs <- freqs$cos()
        sin_freqs <- freqs$sin()

        if (pad_size != 0L) {
            cos_padding <- torch::torch_ones_like(cos_freqs[,, 1:pad_size])
            sin_padding <- torch::torch_zeros_like(sin_freqs[,, 1:pad_size])
            cos_freqs <- torch::torch_cat(list(cos_padding, cos_freqs), dim = -1L)
            sin_freqs <- torch::torch_cat(list(sin_padding, sin_freqs), dim = -1L)
        }

        # Per-head layout for split application: [B, H, N, dim/(2H)]
        b <- cos_freqs$shape[1]
        t <- cos_freqs$shape[2]
        cos_freqs <- cos_freqs$reshape(c(b, t, self$num_attention_heads, -1L))$transpose(2L, 3L)
        sin_freqs <- sin_freqs$reshape(c(b, t, self$num_attention_heads, -1L))$transpose(2L, 3L)
    }

    list(cos_freqs, sin_freqs)
}
)

# Parity tests for the LTX-2.3 RoPE port against diffusers reference
# fixtures (generated by tools/gen_fixtures_rope.py, checked in).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "rope_ltx23.safetensors",
  package = "diffuseR")
if (fixture_path == "") {
  # Running from the source tree (e.g. during development)
  fixture_path <- "fixtures/rope_ltx23.safetensors"
}
if (!file.exists(fixture_path)) exit_file("rope fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# --- apply_split_rotary_emb ---------------------------------------------------

out4 <- ltx23_apply_split_rotary_emb(fx$split4_x, list(fx$split4_cos, fx$split4_sin))
expect_equal(as.integer(out4$shape), as.integer(fx$split4_out$shape))
expect_true(max_abs_diff(out4, fx$split4_out) < 1e-5)

# 3D input with 4D freqs reshapes per-head and back
out3 <- ltx23_apply_split_rotary_emb(fx$split3_x, list(fx$split4_cos, fx$split4_sin))
expect_equal(as.integer(out3$shape), as.integer(fx$split3_out$shape))
expect_true(max_abs_diff(out3, fx$split3_out) < 1e-5)

# --- apply_interleaved_rotary_emb ----------------------------------------------

outi <- ltx23_apply_interleaved_rotary_emb(fx$inter_x, list(fx$inter_cos, fx$inter_sin))
expect_true(max_abs_diff(outi, fx$inter_out) < 1e-5)

# --- video coords + split freqs -------------------------------------------------

rope_v <- ltx23_rotary_pos_embed(
  dim = 64L, scale_factors = c(8L, 32L, 32L), modality = "video",
  rope_type = "split", num_attention_heads = 4L
)
vc <- rope_v$prepare_video_coords(
  batch_size = 2L, num_frames = 3L, height = 4L, width = 6L,
  device = "cpu", fps = 24.0
)
expect_equal(as.integer(vc$shape), as.integer(fx$video_coords$shape))
expect_true(max_abs_diff(vc, fx$video_coords) < 1e-6)

vf <- rope_v(vc)
expect_true(max_abs_diff(vf[[1]], fx$video_cos) < 1e-6)
expect_true(max_abs_diff(vf[[2]], fx$video_sin) < 1e-6)

# fp16 input stays fp16 and matches reference values
outv16 <- ltx23_apply_split_rotary_emb(fx$video_x_f16, vf)
expect_equal(outv16$dtype$.type(), "Half")
expect_true(max_abs_diff(outv16, fx$video_out_f16) < 1e-3)

# --- audio coords + split freqs -------------------------------------------------

rope_a <- ltx23_rotary_pos_embed(
  dim = 32L, scale_factors = c(8L, 32L, 32L), modality = "audio",
  rope_type = "split", num_attention_heads = 4L
)
ac <- rope_a$prepare_audio_coords(batch_size = 2L, num_frames = 5L, device = "cpu")
expect_equal(as.integer(ac$shape), as.integer(fx$audio_coords$shape))
expect_true(max_abs_diff(ac, fx$audio_coords) < 1e-6)

af <- rope_a(ac)
expect_true(max_abs_diff(af[[1]], fx$audio_cos) < 1e-6)
expect_true(max_abs_diff(af[[2]], fx$audio_sin) < 1e-6)

# --- interleaved freqs -----------------------------------------------------------

rope_i <- ltx23_rotary_pos_embed(
  dim = 64L, modality = "video", rope_type = "interleaved",
  num_attention_heads = 4L
)
ifr <- rope_i(vc)
expect_true(max_abs_diff(ifr[[1]], fx$video_inter_cos) < 1e-6)
expect_true(max_abs_diff(ifr[[2]], fx$video_inter_sin) < 1e-6)

#' LTX-2.3 Text Embedding Connectors
#'
#' Fresh R port of the LTX text connectors from the diffusers reference
#' (Apache-2.0, src/diffusers/pipelines/ltx2/connectors.py). The
#' connectors take raw stacked per-layer Gemma3 hidden states
#' [batch, seq, caption_channels, num_layers + 1], normalize and project
#' them per modality, replace padding with learnable registers, and run a
#' small 1D transformer per modality to produce the DiT text embeddings.
#'
#' @name connectors_ltx23
NULL

# Largest finite value for a float dtype; R torch's torch_finfo rejects
# bfloat16, so fall back to the known constants
.ltx23_finfo_max <- function(dtype) {
    tryCatch(
             torch::torch_finfo(dtype)$max,
             error = function(e) {
        switch(dtype$.type(), BFloat16 = 3.3895313892515355e38, Half = 65504,
               3.402823466e38)
    }
    )
}

#' Per-token RMS normalization over the channel axis
#'
#' @param x Tensor [B, S, C, L] of stacked per-layer hidden states.
#' @param eps Numeric. Stability epsilon.
#'
#' @return Tensor of the same shape.
#'
#' @export
ltx23_per_token_rms_norm <- function(x, eps = 1e-6) {
    variance <- torch::torch_mean(x ^ 2, dim = 3L, keepdim = TRUE)
    x * torch::torch_rsqrt(variance$add(eps))
}

#' 1D rotary embeddings for the text connectors
#'
#' @param dim Integer. Rotary dimension (connector inner dim).
#' @param base_seq_len Integer. Base sequence length for normalization.
#' @param theta Numeric. RoPE theta.
#' @param double_precision Logical. Compute base frequencies in float64.
#' @param rope_type "split" (LTX-2.3) or "interleaved".
#' @param num_attention_heads Integer. For the split per-head layout.
#'
#' @return Module whose forward(batch_size, pos, device) returns
#'   \code{list(cos_freqs, sin_freqs)}, the 1-D rotary tables for a
#'   sequence of length \code{pos}.
#'
#' @export
ltx23_rotary_pos_embed_1d <- torch::nn_module(
    "ltx23_rotary_pos_embed_1d",
    initialize = function(
                          dim,
                          base_seq_len = 4096L,
                          theta = 10000.0,
                          double_precision = TRUE,
                          rope_type = "split",
                          num_attention_heads = 32L
    ) {
    if (!rope_type %in% c("interleaved", "split")) {
        stop("rope_type must be 'interleaved' or 'split', got: ", rope_type)
    }
    self$dim <- as.integer(dim)
    self$base_seq_len <- base_seq_len
    self$theta <- theta
    self$double_precision <- double_precision
    self$rope_type <- rope_type
    self$num_attention_heads <- as.integer(num_attention_heads)
},
    forward = function(batch_size, pos, device) {
    grid_1d <- torch::torch_arange(start = 0, end = pos - 1,
                                   dtype = torch::torch_float32(),
                                   device = device) / self$base_seq_len
    grid <- grid_1d$unsqueeze(1L)$`repeat`(c(batch_size, 1L)) # [B, S]

    freqs_dtype <- if (self$double_precision) torch::torch_float64() else torch::torch_float32()
    pow_indices <- torch::torch_pow(
                                    self$theta,
                                    torch::torch_linspace(
            start = 0.0, end = 1.0, steps = self$dim %/% 2L,
            dtype = freqs_dtype, device = device
        )
    )
    freqs <- (pow_indices * pi / 2.0)$to(dtype = torch::torch_float32())
    freqs <- (grid$unsqueeze(-1L) * 2 - 1) * freqs # [B, S, dim/2]

    if (self$rope_type == "interleaved") {
        cos_freqs <- freqs$cos()$repeat_interleave(2L, dim = -1L)
        sin_freqs <- freqs$sin()$repeat_interleave(2L, dim = -1L)
    } else {
        cos_freqs <- freqs$cos()
        sin_freqs <- freqs$sin()
        b <- cos_freqs$shape[1]
        t <- cos_freqs$shape[2]
        cos_freqs <- cos_freqs$reshape(c(b, t, self$num_attention_heads, -1L))$transpose(2L, 3L)
        sin_freqs <- sin_freqs$reshape(c(b, t, self$num_attention_heads, -1L))$transpose(2L, 3L)
    }
    list(cos_freqs, sin_freqs)
}
)

# Pre-norm 1D transformer block: RMS norm -> attention -> RMS norm -> FF,
# both with plain residual connections.
ltx23_transformer_block_1d <- torch::nn_module(
    "ltx23_transformer_block_1d",
    initialize = function(
                          dim,
                          num_attention_heads,
                          attention_head_dim,
                          eps = 1e-6,
                          rope_type = "split",
                          apply_gated_attention = TRUE
    ) {
    self$norm1 <- ltx23_rms_norm(dim, eps = eps, elementwise_affine = FALSE)
    self$attn1 <- ltx23_attention(query_dim = dim,
                                  heads = num_attention_heads,
                                  dim_head = attention_head_dim,
                                  rope_type = rope_type,
                                  apply_gated_attention = apply_gated_attention)
    self$norm2 <- ltx23_rms_norm(dim, eps = eps, elementwise_affine = FALSE)
    self$ff <- ltx23_feed_forward(dim)
},
    forward = function(hidden_states, attention_mask = NULL, rotary_emb = NULL) {
    norm_hidden_states <- self$norm1(hidden_states)
    hidden_states <- hidden_states + self$attn1(
        norm_hidden_states, attention_mask = attention_mask, query_rotary_emb = rotary_emb
    )
    norm_hidden_states <- self$norm2(hidden_states)
    hidden_states + self$ff(norm_hidden_states)
}
)

#' 1D connector transformer
#'
#' Replaces padded positions with learnable registers (valid tokens are
#' front-aligned in their original order; the tail is filled with
#' registers indexed by absolute position, after which the attention mask
#' is cleared), then runs 1D transformer blocks with rotary embeddings.
#'
#' @param num_attention_heads,attention_head_dim,num_layers Transformer shape.
#' @param num_learnable_registers Integer or NULL. Register count (the
#'   sequence length must be divisible by it).
#' @param rope_base_seq_len,rope_theta,rope_double_precision,rope_type RoPE config.
#' @param eps Numeric. Norm epsilon.
#' @param gated_attention Logical. Per-head attention output gates.
#'
#' @return Module whose forward(hidden_states, attention_mask,
#'   attn_mask_binarize_threshold) returns
#'   \code{list(hidden_states, attention_mask)}: the transformed
#'   sequence and the (possibly binarized) mask that accompanies it.
#'
#' @export
ltx23_connector_transformer_1d <- torch::nn_module(
    "ltx23_connector_transformer_1d",
    initialize = function(
                          num_attention_heads = 32L,
                          attention_head_dim = 128L,
                          num_layers = 8L,
                          num_learnable_registers = 128L,
                          rope_base_seq_len = 4096L,
                          rope_theta = 10000.0,
                          rope_double_precision = TRUE,
                          eps = 1e-6,
                          rope_type = "split",
                          gated_attention = TRUE
    ) {
    self$num_attention_heads <- as.integer(num_attention_heads)
    self$inner_dim <- as.integer(num_attention_heads * attention_head_dim)
    self$num_learnable_registers <- num_learnable_registers

    if (!is.null(num_learnable_registers)) {
        self$learnable_registers <- torch::nn_parameter(
            torch::torch_rand(num_learnable_registers, self$inner_dim) * 2 - 1
        )
    }

    self$rope <- ltx23_rotary_pos_embed_1d(self$inner_dim,
        base_seq_len = rope_base_seq_len, theta = rope_theta,
        double_precision = rope_double_precision, rope_type = rope_type,
        num_attention_heads = num_attention_heads)

    self$transformer_blocks <- torch::nn_module_list(lapply(seq_len(num_layers), function(i) {
        ltx23_transformer_block_1d(
                                   dim = self$inner_dim,
                                   num_attention_heads = num_attention_heads,
                                   attention_head_dim = attention_head_dim,
                                   eps = eps,
                                   rope_type = rope_type,
                                   apply_gated_attention = gated_attention
        )
    }))

    self$norm_out <- ltx23_rms_norm(self$inner_dim, eps = eps, elementwise_affine = FALSE)
},
    forward = function(hidden_states, attention_mask = NULL,
                       attn_mask_binarize_threshold = -9000.0) {
    batch_size <- hidden_states$shape[1]
    seq_len <- hidden_states$shape[2]

    if (!is.null(self$num_learnable_registers)) {
        if (seq_len %% self$num_learnable_registers != 0L) {
            stop(
                 "Sequence length ", seq_len, " must be divisible by the number of ",
                 "learnable registers ", self$num_learnable_registers
            )
        }
        num_repeats <- seq_len %/% self$num_learnable_registers
        registers <- self$learnable_registers$unsqueeze(1L)$
        expand(c(num_repeats, -1L, -1L))$reshape(c(seq_len, -1L)) # [S, D]

        binary_attn_mask <- (attention_mask >= attn_mask_binarize_threshold)$to(
            dtype = torch::torch_int()
        )
        if (binary_attn_mask$ndim == 4L) {
            binary_attn_mask <- binary_attn_mask$squeeze(2L)$squeeze(2L) # [B, 1, 1, S] -> [B, S]
        }

        # Front-align valid tokens (stable sort keeps their order), fill the
        # tail with registers indexed by absolute position
        order <- torch::torch_sort(1L - binary_attn_mask, dim = 2L, stable = TRUE)[[2]] # [B, S]
        front_aligned <- torch::torch_gather(
            hidden_states, 2L,
            order$unsqueeze(-1L)$expand(c(-1L, -1L, hidden_states$shape[3]))
        )
        num_valid <- binary_attn_mask$sum(dim = 2L, keepdim = TRUE) # [B, 1]
        positions <- torch::torch_arange(
            start = 0, end = seq_len - 1, device = hidden_states$device
        )$unsqueeze(1L) # [1, S]
        front_mask <- (positions < num_valid)$unsqueeze(-1L) # [B, S, 1]
        registers_expanded <- registers$unsqueeze(1L)$expand(c(batch_size, -1L, -1L))
        hidden_states <- torch::torch_where(
            front_mask, front_aligned, registers_expanded$to(dtype = hidden_states$dtype)
        )

        # All positions are valid once registers replace the padding
        attention_mask <- torch::torch_zeros_like(attention_mask)
    }

    rotary_emb <- self$rope(batch_size, seq_len, device = hidden_states$device)

    for (i in seq_along(self$transformer_blocks)) {
        hidden_states <- self$transformer_blocks[[i]](
            hidden_states, attention_mask = attention_mask, rotary_emb = rotary_emb
        )
    }
    hidden_states <- self$norm_out(hidden_states)

    list(hidden_states, attention_mask)
}
)

#' LTX-2.3 text connectors
#'
#' Takes raw stacked per-layer text encoder hidden states and produces
#' the video and audio text embeddings for the DiT: per-token RMS norm,
#' per-modality sqrt(dim ratio) rescaling and projection, then a
#' per-modality 1D connector transformer.
#'
#' @param caption_channels Integer. Text encoder hidden size (3840 for
#'   Gemma3-12B).
#' @param text_proj_in_factor Integer. Number of stacked hidden states
#'   (num_layers + 1 = 49 for Gemma3-12B).
#' @param video_connector_num_attention_heads,video_connector_attention_head_dim,video_connector_num_layers
#'   Video connector shape (LTX-2.3: 32 x 128, 8 layers).
#' @param video_connector_num_learnable_registers Integer or NULL.
#' @param video_gated_attn Logical.
#' @param audio_connector_num_attention_heads,audio_connector_attention_head_dim,audio_connector_num_layers
#'   Audio connector shape (LTX-2.3: 32 x 64, 8 layers).
#' @param audio_connector_num_learnable_registers Integer or NULL.
#' @param audio_gated_attn Logical.
#' @param connector_rope_base_seq_len,rope_theta,rope_double_precision,rope_type RoPE config.
#' @param video_hidden_dim,audio_hidden_dim Integers. Projection targets
#'   (DiT inner dims: 4096 / 2048).
#' @param proj_bias Logical. Projection bias (TRUE for LTX-2.3).
#'
#' @return Module whose forward(text_encoder_hidden_states,
#'   attention_mask) returns \code{list(video_text_embedding,
#'   audio_text_embedding, attention_mask)}: the caption states adapted
#'   for the video and audio cross-attention streams, plus the binary
#'   mask to use with them.
#'
#' @export
ltx23_text_connectors <- torch::nn_module(
    "ltx23_text_connectors",
    initialize = function(
                          caption_channels = 3840L,
                          text_proj_in_factor = 49L,
                          video_connector_num_attention_heads = 32L,
                          video_connector_attention_head_dim = 128L,
                          video_connector_num_layers = 8L,
                          video_connector_num_learnable_registers = 128L,
                          video_gated_attn = TRUE,
                          audio_connector_num_attention_heads = 32L,
                          audio_connector_attention_head_dim = 64L,
                          audio_connector_num_layers = 8L,
                          audio_connector_num_learnable_registers = 128L,
                          audio_gated_attn = TRUE,
                          connector_rope_base_seq_len = 4096L,
                          rope_theta = 10000.0,
                          rope_double_precision = TRUE,
                          rope_type = "split",
                          video_hidden_dim = 4096L,
                          audio_hidden_dim = 2048L,
                          proj_bias = TRUE
    ) {
    self$caption_channels <- as.integer(caption_channels)
    self$video_hidden_dim <- as.integer(video_hidden_dim)
    self$audio_hidden_dim <- as.integer(audio_hidden_dim)

    text_encoder_dim <- caption_channels * text_proj_in_factor
    self$video_text_proj_in <- linear_noinit(text_encoder_dim,
        video_hidden_dim, bias = proj_bias)
    self$audio_text_proj_in <- linear_noinit(text_encoder_dim, audio_hidden_dim, bias = proj_bias)

    self$video_connector <- ltx23_connector_transformer_1d(
        num_attention_heads = video_connector_num_attention_heads,
        attention_head_dim = video_connector_attention_head_dim,
        num_layers = video_connector_num_layers,
        num_learnable_registers = video_connector_num_learnable_registers,
        rope_base_seq_len = connector_rope_base_seq_len,
        rope_theta = rope_theta,
        rope_double_precision = rope_double_precision,
        rope_type = rope_type,
        gated_attention = video_gated_attn
    )
    self$audio_connector <- ltx23_connector_transformer_1d(
        num_attention_heads = audio_connector_num_attention_heads,
        attention_head_dim = audio_connector_attention_head_dim,
        num_layers = audio_connector_num_layers,
        num_learnable_registers = audio_connector_num_learnable_registers,
        rope_base_seq_len = connector_rope_base_seq_len,
        rope_theta = rope_theta,
        rope_double_precision = rope_double_precision,
        rope_type = rope_type,
        gated_attention = audio_gated_attn
    )
},
    forward = function(text_encoder_hidden_states, attention_mask) {
    if (text_encoder_hidden_states$ndim == 3L) {
        text_encoder_hidden_states <- text_encoder_hidden_states$unflatten(
            3L, c(self$caption_channels, -1L)
        )
    }

    norm_states <- ltx23_per_token_rms_norm(text_encoder_hidden_states)
    norm_states <- norm_states$flatten(start_dim = 3L, end_dim = 4L)
    bool_mask <- attention_mask$to(dtype = torch::torch_bool())$unsqueeze(-1L)
    norm_states <- torch::torch_where(
                                      bool_mask, norm_states, torch::torch_zeros_like(norm_states)
    )

    # Rescale per modality by sqrt(target_dim / caption_channels)
    video_norm <- norm_states$mul(sqrt(self$video_hidden_dim / self$caption_channels))
    audio_norm <- norm_states$mul(sqrt(self$audio_hidden_dim / self$caption_channels))

    video_proj <- self$video_text_proj_in(video_norm)
    audio_proj <- self$audio_text_proj_in(audio_norm)

    # Multiplicative [B, S] mask -> additive [B, 1, 1, S] with -finfo max
    text_dtype <- video_proj$dtype
    add_mask <- (attention_mask$to(dtype = torch::torch_int64()) - 1L)$to(dtype = text_dtype)
    add_mask <- add_mask$reshape(c(add_mask$shape[1], 1L, -1L, add_mask$shape[2]))
    add_mask <- add_mask$mul(.ltx23_finfo_max(text_dtype))

    video_res <- self$video_connector(video_proj, add_mask)
    video_text_embedding <- video_res[[1]]
    video_attn_mask <- video_res[[2]]

    # Post-connector mask back to binary; mask the video embedding
    binary_mask <- (video_attn_mask < 1e-6)$to(dtype = torch::torch_int64())
    binary_mask <- binary_mask$reshape(c(
            video_text_embedding$shape[1], video_text_embedding$shape[2], 1L
        ))
    video_text_embedding <- video_text_embedding * binary_mask

    audio_res <- self$audio_connector(audio_proj, add_mask)
    audio_text_embedding <- audio_res[[1]]

    list(
         video_text_embedding = video_text_embedding,
         audio_text_embedding = audio_text_embedding,
         attention_mask = binary_mask$squeeze(-1L)
    )
}
)

#' Map an official connectors checkpoint key to the R module name
#'
#' @param key Character. Checkpoint key.
#'
#' @return Character. Module parameter name.
#'
#' @export
ltx23_map_connector_key <- function(key) {
    key <- sub("^model\\.diffusion_model\\.", "", key)
    key <- gsub("video_embeddings_connector", "video_connector", key,
                fixed = TRUE)
    key <- gsub("audio_embeddings_connector", "audio_connector", key, fixed = TRUE)
    key <- gsub("transformer_1d_blocks", "transformer_blocks", key, fixed = TRUE)
    key <- gsub("text_embedding_projection.video_aggregate_embed", "video_text_proj_in", key, fixed = TRUE)
    key <- gsub("text_embedding_projection.audio_aggregate_embed", "audio_text_proj_in", key, fixed = TRUE)
    key <- gsub("q_norm", "norm_q", key, fixed = TRUE)
    key <- gsub("k_norm", "norm_k", key, fixed = TRUE)
    key
}

# Parity tests for the LTX-2.3 text connectors against diffusers
# reference fixtures (tools/gen_fixtures_connectors.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "connectors_ltx23.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/connectors_ltx23.safetensors"
if (!file.exists(fixture_path)) exit_file("connector fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

conn <- ltx23_text_connectors(
  caption_channels = 8L,
  text_proj_in_factor = 3L,
  video_connector_num_attention_heads = 2L,
  video_connector_attention_head_dim = 8L,
  video_connector_num_layers = 2L,
  video_connector_num_learnable_registers = 4L,
  video_gated_attn = TRUE,
  audio_connector_num_attention_heads = 2L,
  audio_connector_attention_head_dim = 4L,
  audio_connector_num_layers = 2L,
  audio_connector_num_learnable_registers = 4L,
  audio_gated_attn = TRUE,
  rope_type = "split",
  video_hidden_dim = 16L,
  audio_hidden_dim = 8L,
  proj_bias = TRUE
)
conn$eval()

weights <- fx[grep("^conn\\.", names(fx))]
names(weights) <- sub("^conn\\.", "", names(weights))
dests <- c(conn$named_parameters(), conn$named_buffers())
expect_equal(sort(names(weights)), sort(names(dests)))
torch::with_no_grad({
  for (name in names(weights)) dests[[name]]$copy_(weights[[name]])
})

torch::with_no_grad({
  res <- conn(fx$c_states, fx$c_mask)
})
expect_equal(as.integer(res$video_text_embedding$shape), as.integer(fx$c_video_emb$shape))
video_diff <- max_abs_diff(res$video_text_embedding, fx$c_video_emb)
expect_true(
  video_diff < 1e-4,
  info = sprintf("video max abs diff: %.9g", video_diff)
)
expect_true(max_abs_diff(res$audio_text_embedding, fx$c_audio_emb) < 1e-4)
expect_true(max_abs_diff(
  res$attention_mask$to(dtype = torch::torch_float32()), fx$c_out_mask
) < 1e-6)

# Flattened 3D input gives identical output
torch::with_no_grad({
  res3 <- conn(fx$c_states$flatten(start_dim = 3L), fx$c_mask)
})
video3_diff <- max_abs_diff(res3$video_text_embedding, fx$c_video_emb3)
expect_true(
  video3_diff < 1e-4,
  info = sprintf("video3 max abs diff: %.9g", video3_diff)
)

# Key mapper
expect_equal(
  ltx23_map_connector_key("model.diffusion_model.video_embeddings_connector.transformer_1d_blocks.0.attn1.q_norm.weight"),
  "video_connector.transformer_blocks.0.attn1.norm_q.weight"
)
expect_equal(
  ltx23_map_connector_key("text_embedding_projection.video_aggregate_embed.weight"),
  "video_text_proj_in.weight"
)
expect_equal(
  ltx23_map_connector_key("text_embedding_projection.audio_aggregate_embed.bias"),
  "audio_text_proj_in.bias"
)
expect_equal(
  ltx23_map_connector_key("model.diffusion_model.audio_embeddings_connector.learnable_registers"),
  "audio_connector.learnable_registers"
)

#' FLUX.2 Klein Text-to-Image Pipeline
#'
#' Klein text-to-image, ported from the diffusers reference
#' (Apache-2.0, src/diffusers/pipelines/flux2/pipeline_flux2_klein.py).
#' Klein-4B is step-distilled: no classifier-free guidance and no
#' guidance embedding; the FlowMatch schedule uses dynamic shifting with
#' the BFL empirical mu. Latent noise is drawn directly in the packed
#' 128-channel space; only VAE-encoded latents get the BatchNorm
#' normalization (txt2img never encodes).
#'
#' @name txt2img_flux2
NULL

# Resolve a FLUX.2-klein support file from the HuggingFace cache
.flux2_cached <- function(file) {
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to locate model files.")
    }
    tryCatch(
             hfhub::hub_download(.flux2_repo, file, local_files_only = TRUE),
             error = function(e) {
        stop("Missing ", file, " in the HuggingFace cache; ",
             "run download_flux2_klein() first.", call. = FALSE)
    }
    )
}

#' Load the FLUX.2 klein pipeline
#'
#' Loads the quantized transformer artifact plus the FLUX.2 VAE decoder,
#' Qwen3 text encoder, and tokenizer from the HuggingFace cache
#' populated by \code{\link{download_flux2_klein}}. With fp8 precision
#' the ~4 GB transformer stays GPU-resident.
#'
#' @param model_dir Quantized artifact directory (default: the
#'   \code{download_flux2_klein} location for \code{precision}), or a
#'   raw diffusers transformer directory.
#' @param device Character. Compute device.
#' @param precision "auto" (default: reuse an existing artifact, else
#'   fp8 when safetensors supports float8, else nf4), "fp8", or "nf4".
#' @param text_device Device for the Qwen3 encoder (default:
#'   \code{device}; it encodes in its own phase and offloads).
#' @param attn_chunk Integer or NULL. Attention query-chunk override.
#' @param phase_offload Logical. One GPU tenant per phase.
#' @param pin Logical or NULL. Page-lock the phase-swapped weights for
#'   DMA-rate transfer (see \code{\link{staging}}). NULL (default)
#'   resolves via \code{options(diffuseR.pin_staging)} then the
#'   host-RAM-aware \code{\link{recommend}} decision.
#' @param verbose Logical.
#'
#' @return A \code{flux2_pipeline} list.
#'
#' @export
flux2_load_pipeline <- function(model_dir = NULL, device = "cuda",
                                precision = c("auto", "fp8", "nf4", "bf16"),
                                text_device = NULL, attn_chunk = NULL,
                                phase_offload = TRUE, pin = NULL,
                                verbose = TRUE) {
    precision <- .flux_resolve_precision(match.arg(precision),
        file.path(tools::R_user_dir("diffuseR", "data"), "flux2-klein-4b-"))
    if (is.null(text_device)) {
        if (device == "cuda") {
            text_device <- "cuda"
        } else {
            text_device <- "cpu"
        }
    }
    if (is.null(model_dir)) {
        model_dir <- .flux_model_dir("flux2", precision,
            file.path(tools::R_user_dir("diffuseR", "data"),
                      "flux2-klein-4b-"))
    }

    ckpt <- if (file.exists(file.path(model_dir, "manifest.json"))) {
        flux_open_quantized(model_dir)
    } else {
        flux_open_checkpoint(model_dir)
    }
    pin <- .resolve_pin(pin, "flux2", ckpt$format %||% "full")

    if (!nzchar(Sys.getenv("PYTORCH_CUDA_ALLOC_CONF"))) {
        # Resident fp8 has an NF4-like stable footprint: the native
        # backend avoids expandable_segments' page-unmap cost on the
        # per-step activation churn (see ltx23_load_pipeline)
        Sys.setenv(PYTORCH_CUDA_ALLOC_CONF = "backend:native")
    }
    if (device == "cuda") {
        # Footprint sized to the LARGEST phase (the 8 GB Qwen3 encode),
        # not the transformer: a low footprint puts the allocator's
        # R-gc callback threshold under the working set, and every
        # callback walks the ~300k-object tokenizer heap (measured:
        # 13-20 s forwards at footprint 6 vs sub-second at 12)
        .flux_gc_gates(footprint_gb = 12)
    }

    if (phase_offload) {
        component_device <- "cpu"
    } else {
        component_device <- device
    }

    pipe <- list(
                 format = ckpt$format %||% "full",
                 device = device,
                 text_device = text_device,
                 phase_offload = phase_offload,
                 attn_chunk = if (is.null(attn_chunk)) NULL else as.integer(attn_chunk),
                 config = ckpt$config
    )

    if (verbose) {
        message("Loading transformer (", pipe$format, ")...")
    }
    pipe$transformer <- flux_load_transformer(
        ckpt, device = component_device,
        dtype = if (device == "cpu") "float32" else "bfloat16",
        pin = FALSE,
        fp8_resident = FALSE,
        verbose = verbose
    )
    # Resident fp8 happens at onload time (the weights ride to the GPU
    # with the phase and back off after)
    pipe$fp8_resident <- identical(pipe$format, "fp8") && device == "cuda"

    if (verbose) {
        message("Loading FLUX.2 VAE decoder...")
    }
    vae_config <- jsonlite::fromJSON(.flux2_cached("vae/config.json"))
    pipe$vae_bn_eps <- vae_config$batch_norm_eps %||% 1e-4
    pipe$decoder <- load_flux2_vae_decoder(
        .flux2_cached("vae/diffusion_pytorch_model.safetensors"),
        latent_channels = as.integer(vae_config$latent_channels %||% 32L),
        verbose = verbose
    )
    pipe$decoder$to(device = component_device)
    if (device == "cuda" && phase_offload) {
        # Cast to the compute dtype once here so the per-generation
        # onload is device-only (and the pinned copy is bf16, not fp32);
        # the decode reads the decoder's own dtype and follows.
        pipe$decoder$to(dtype = torch::torch_bfloat16())
    }

    if (verbose) {
        message("Loading Qwen3 text encoder...")
    }
    te_dir <- dirname(.flux2_cached("text_encoder/config.json"))
    pipe$text_encoder <- load_qwen3_text_encoder(
        te_dir, device = if (phase_offload) "cpu" else text_device,
        dtype = if (text_device == "cpu") "float32" else "bfloat16",
        verbose = verbose
    )
    pipe$tokenizer <- qwen_bpe_tokenizer(.flux2_cached("tokenizer/tokenizer.json"))

    components <- c("transformer", "decoder")
    if (!identical(text_device, "cpu")) {
        components <- c(components, "text_encoder")
    }
    pipe$staging <- .flux_build_staging(pipe, pin, phase_offload, device,
                                        components, verbose = verbose)

    structure(pipe, class = "flux2_pipeline")
}

# Flow-matching Euler loop; klein is CFG-free (one forward per step)
.flux2_denoise <- function(transformer, latents, schedule, prompt_embeds,
                           image_rotary_emb, compute_dtype,
                           chunk_size = NULL, verbose = TRUE) {
    timesteps <- as.numeric(schedule$timesteps$cpu())
    n <- length(timesteps)
    pb <- .denoise_progress(n, NULL, verbose)
    f32 <- torch::torch_float32()

    torch::with_no_grad({
        for (i in seq_len(n)) {
            t <- timesteps[[i]]
            t_model <- torch::torch_tensor(t / 1000, dtype = compute_dtype,
                device = latents$device)$reshape(1L)

            noise_pred <- transformer(
                                      hidden_states = latents$to(dtype = compute_dtype),
                                      encoder_hidden_states = prompt_embeds,
                                      timestep = t_model,
                                      image_rotary_emb = image_rotary_emb,
                                      chunk_size = chunk_size
            )

            step <- flowmatch_scheduler_step(
                noise_pred$to(dtype = f32), t, latents, schedule
            )
            latents <- step$prev_sample
            schedule <- step$schedule
            rm(noise_pred, step)
            pb$tick(i)
        }
    })
    pb$done()
    latents
}

#' Generate an image with FLUX.2 klein
#'
#' Step-distilled text-to-image (klein-4B: 4 steps, no guidance): Qwen3
#' prompt encoding (chat template, mid-stack hidden states), FlowMatch
#' denoising with the empirical dynamic shift, and 32-channel VAE decode
#' through the BatchNorm latent statistics.
#'
#' @param prompt Character. The prompt.
#' @param pipeline A \code{flux2_pipeline} from
#'   \code{\link{flux2_load_pipeline}}; NULL loads one (passing
#'   \code{...} through).
#' @param width,height Integers, divisible by 16.
#' @param num_inference_steps Integer. Denoising steps (klein-4B: 4).
#' @param max_sequence_length Integer. Qwen3 token length (512).
#' @param seed Integer or NULL. Latents are drawn on the CPU in the
#'   packed shape, so a seed matches a Python diffusers run with a CPU
#'   generator.
#' @param prompt_embeds Optional precomputed [B, S, 7680] embeddings.
#' @param save_file Logical. Write a PNG.
#' @param filename Output path (default derived from the prompt).
#' @param verbose Logical, or one of "silent", "progress", "steps".
#'   TRUE = "steps" (full per-phase chatter), FALSE = "silent".
#'   "progress" prints a one-line generation summary plus a denoise
#'   progress bar (interactive) or periodic step ticks (captured logs).
#' @param ... Passed to \code{\link{flux2_load_pipeline}} when
#'   \code{pipeline} is NULL.
#'
#' @return Invisibly, \code{list(image, metadata)} where \code{image} is
#'   an [H, W, 3] array in [0, 1].
#'
#' @export
txt2img_flux2 <- function(prompt, pipeline = NULL, width = 1024L,
                          height = 1024L, num_inference_steps = 4L,
                          max_sequence_length = 512L, seed = NULL,
                          prompt_embeds = NULL, save_file = TRUE,
                          filename = NULL, verbose = TRUE, ...) {
    level <- .verbosity(verbose)
    verbose <- level == "steps"
    if (level != "silent") {
        message(sprintf("FLUX.2 klein: %dx%d, %d steps", as.integer(width),
                        as.integer(height), as.integer(num_inference_steps)))
    }
    if (is.null(pipeline)) {
        pipeline <- flux2_load_pipeline(..., verbose = verbose)
    }
    device <- pipeline$device
    width <- as.integer(width)
    height <- as.integer(height)
    if (width %% 16L != 0L || height %% 16L != 0L) {
        stop("width and height must be divisible by 16")
    }

    f32 <- torch::torch_float32()
    compute_dtype <- if (device == "cpu") {
        f32
    } else {
        torch::torch_bfloat16()
    }

    phase_offload <- isTRUE(pipeline$phase_offload) && device != "cpu"
    staging <- pipeline$staging %||% list()
    # Components move by name so pinned staging (see staging.R) can carry
    # the CPU<->GPU transfer when the loader prepared it; otherwise the
    # pageable module$to() path runs.
    onload <- function(what) {
        module <- pipeline[[what]]
        if (phase_offload) {
            st <- staging[[what]]
            if (is.null(st)) {
                module$to(device = device)
            } else {
                .staged_onload(st, device)
            }
        }
        module
    }
    offload <- function(what) {
        module <- pipeline[[what]]
        if (phase_offload) {
            st <- staging[[what]]
            if (is.null(st)) {
                module$to(device = "cpu")
            } else {
                .staged_offload(st)
            }
            clear_vram()
        }
        invisible(module)
    }

    t0 <- Sys.time()

    # --- Phase 1: text encoding --------------------------------------------------
    if (is.null(prompt_embeds)) {
        if (verbose) {
            message("Encoding prompt (Qwen3)...")
        }
        # An explicit text_device = "cpu" keeps Qwen3 fp32 on the host
        # (it computes in place); only onload it when it phase-swaps.
        gpu_encode <- !identical(pipeline$text_device, "cpu")
        if (gpu_encode) {
            onload("text_encoder")
        }
        te_device <- pipeline$text_encoder$model$embed_tokens$weight$device
        prompt_embeds <- encode_with_qwen3(prompt, pipeline$text_encoder,
            pipeline$tokenizer, max_sequence_length = max_sequence_length,
            device = te_device)
        if (gpu_encode) {
            offload("text_encoder")
        }
    }
    prompt_embeds <- prompt_embeds$to(device = device, dtype = compute_dtype)
    txt_len <- prompt_embeds$shape[2]

    # --- Phase 2: latents, rotary embeddings, schedule -----------------------------
    h2 <- height %/% 16L
    w2 <- width %/% 16L
    if (!is.null(seed)) {
        torch::torch_manual_seed(seed)
    }
    # Noise directly in the packed/patchified space (no BN for txt2img)
    latents <- torch::torch_randn(c(1L, 128L, h2, w2), dtype = f32)
    latents <- flux2_pack_latents(latents)$to(device = device)
    seq_img <- latents$shape[2]

    txt_ids <- flux2_prepare_text_ids(txt_len)
    latent_ids <- flux2_prepare_latent_ids(h2, w2)
    ids <- torch::torch_cat(list(txt_ids, latent_ids), dim = 1L)
    rope <- flux_pos_embed(
                           ids,
                           axes_dim = pipeline$transformer$axes_dims_rope %||% c(32L, 32L, 32L, 32L),
                           theta = pipeline$transformer$rope_theta %||% 2000
    )
    rope <- list(rope[[1]]$to(device = device), rope[[2]]$to(device = device))

    n_steps <- as.integer(num_inference_steps)
    mu <- flux2_empirical_mu(seq_img, n_steps)
    sched <- flowmatch_scheduler_create(
                                        use_dynamic_shifting = TRUE,
                                        time_shift_type = "exponential"
    )
    sched <- flowmatch_set_timesteps(
                                     sched, n_steps, mu = mu,
                                     sigmas = seq(1, 1 / n_steps, length.out = n_steps)
    )

    # --- Phase 3: denoise ------------------------------------------------------------
    transformer <- onload("transformer")
    # Resident fp8's plain weight fields ride to the GPU with the phase.
    # When staging covers the transformer it already moved them (and the
    # field reassignment here would orphan the staged pairs), so skip it.
    fp8_manual <- isTRUE(pipeline$fp8_resident) &&
    is.null(staging[["transformer"]])
    if (fp8_manual) {
        .flux_fp8_to_device(transformer, device)
    }
    if (verbose) {
        message(sprintf("Denoising: %d steps at %dx%d...", n_steps, width,
                        height))
    }
    latents <- .flux2_denoise(
                              transformer, latents, sched, prompt_embeds, rope,
                              compute_dtype, chunk_size = pipeline$attn_chunk,
                              verbose = level
    )
    if (fp8_manual && phase_offload) {
        .flux_fp8_to_device(pipeline$transformer, "cpu")
    }
    offload("transformer")
    ltx23_release_dequant_buffers()

    # --- Phase 4: decode ---------------------------------------------------------------
    if (verbose) {
        message("Decoding...")
    }
    latents <- flux2_unpack_latents_with_ids(latents, latent_ids, h2, w2)
    latents <- flux2_bn_normalize(
                                  latents, pipeline$decoder$bn$running_mean,
                                  pipeline$decoder$bn$running_var,
                                  eps = pipeline$vae_bn_eps %||% 1e-4, inverse = TRUE
    )
    latents <- flux2_unpatchify_latents(latents)

    decoder <- onload("decoder")
    torch::with_no_grad({
        dec_param <- decoder$post_quant_conv$weight
        img <- decoder(latents$to(device = dec_param$device,
                                  dtype = dec_param$dtype))
        img <- img$to(dtype = f32)$cpu()
    })
    offload("decoder")

    img <- img$squeeze(1)$permute(c(2L, 3L, 1L))
    img <- img$add(1)$div(2)$clamp(0, 1)
    img_array <- as.array(img)

    gen_seconds <- as.numeric(difftime(Sys.time(), t0, units = "secs"))
    if (verbose) {
        message(sprintf("Generated in %.1f s", gen_seconds))
    }

    if (save_file) {
        if (is.null(filename)) {
            filename <- filename_from_prompt(prompt)
        }
        save_image(img_array, filename)
        if (verbose) {
            message("Saved to ", filename)
        }
    }

    metadata <- list(
                     prompt = prompt, width = width, height = height,
                     steps = n_steps, seed = seed, model = "flux2-klein-4b",
                     precision = pipeline$format, seconds = gen_seconds
    )
    invisible(list(image = img_array, metadata = metadata))
}

# End-to-end smoke test for the FLUX.2 klein pipeline wiring on the CPU
# with tiny random-init components: pack -> denoise (2 steps, dynamic
# shifting) -> unpack-with-ids -> BN denorm -> unpatchify -> decode.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)

torch::torch_manual_seed(8)

transformer <- flux2_transformer(
  in_channels = 128L,
  num_layers = 1L,
  num_single_layers = 1L,
  attention_head_dim = 8L,
  num_attention_heads = 2L,
  joint_attention_dim = 24L,
  mlp_ratio = 3.0,
  axes_dims_rope = c(2L, 2L, 2L, 2L),
  rope_theta = 2000
)
transformer$eval()

decoder <- flux2_vae_decoder(
  latent_channels = 32L,
  block_channels = c(32L, 32L, 16L, 8L),
  norm_groups = 8L
)
decoder$eval()

pipeline <- structure(
  list(
    transformer = transformer,
    decoder = decoder,
    device = "cpu",
    text_device = "cpu",
    phase_offload = FALSE,
    fp8_resident = FALSE,
    format = "full",
    attn_chunk = NULL,
    config = list(in_channels = 128L),
    vae_bn_eps = 1e-4
  ),
  class = "flux2_pipeline"
)

res <- txt2img_flux2(
  "tiny smoke test",
  pipeline = pipeline,
  width = 64L, height = 64L,
  num_inference_steps = 2L,
  seed = 42L,
  prompt_embeds = torch::torch_randn(1L, 7L, 24L),
  save_file = FALSE,
  verbose = FALSE
)

expect_equal(dim(res$image), c(64L, 64L, 3L))
expect_true(all(is.finite(res$image)))
expect_true(all(res$image >= 0) && all(res$image <= 1))
expect_equal(res$metadata$steps, 2L)
expect_equal(res$metadata$model, "flux2-klein-4b")

#' Qwen3 Text Encoder
#'
#' Fresh R port of the Qwen3 decoder stack from HuggingFace transformers
#' (Apache-2.0, src/transformers/models/qwen3/), used by FLUX.2 klein as
#' its text encoder (Qwen3-4B: 36 layers, hidden 2560, 32 query / 8 KV
#' heads, head_dim 128, SwiGLU 9728, RoPE theta 1e6). The pipeline
#' consumes mid-stack hidden states (layers 9, 18, 27 for klein-4B)
#' concatenated per token, so the forward runs only as deep as the last
#' requested layer; the LM head is never needed (embeddings are tied).
#' Causal attention with the tokenizer's padding mask, matching the
#' reference exactly.
#'
#' @name qwen3_text_encoder
NULL

# Llama-convention rotary tables: cos/sin [S, head_dim/2] at the given
# theta, applied with the split-half kernel (x1*cos - x2*sin | x2*cos +
# x1*sin), which matches rotate_half exactly.
.qwen3_rope_tables <- function(seq_len, head_dim, theta, device) {
    f32 <- torch::torch_float32()
    inv_freq <- 1.0 / torch::torch_pow(
                                       theta,
                                       torch::torch_arange(start = 0, end = head_dim - 2, step = 2,
            dtype = f32, device = device) / head_dim
    )
    pos <- torch::torch_arange(start = 0, end = seq_len - 1, dtype = f32,
                               device = device)
    freqs <- pos$unsqueeze(2L) * inv_freq$unsqueeze(1L) # [S, D/2]
    # [1, 1, S, D/2] for broadcasting against [B, H, S, D/2]
    list(
         freqs$cos()$unsqueeze(1L)$unsqueeze(1L),
         freqs$sin()$unsqueeze(1L)$unsqueeze(1L)
    )
}

.qwen3_attention <- torch::nn_module(
                                     "qwen3_attention",
                                     initialize = function(hidden_size, num_heads, num_kv_heads, head_dim,
        eps = 1e-6) {
    self$num_heads <- num_heads
    self$num_kv_heads <- num_kv_heads
    self$head_dim <- head_dim
    self$q_proj <- torch::nn_linear(hidden_size, num_heads * head_dim,
                                    bias = FALSE)
    self$k_proj <- torch::nn_linear(hidden_size, num_kv_heads * head_dim,
                                    bias = FALSE)
    self$v_proj <- torch::nn_linear(hidden_size, num_kv_heads * head_dim,
                                    bias = FALSE)
    self$o_proj <- torch::nn_linear(num_heads * head_dim, hidden_size,
                                    bias = FALSE)
    self$q_norm <- ltx23_rms_norm(head_dim, eps = eps)
    self$k_norm <- ltx23_rms_norm(head_dim, eps = eps)
},
                                     forward = function(x, rope, mask = NULL) {
    shape <- x$shape
    b <- shape[1]
    s <- shape[2]

    q <- self$q_proj(x)$view(c(b, s, self$num_heads, self$head_dim))
    k <- self$k_proj(x)$view(c(b, s, self$num_kv_heads, self$head_dim))
    v <- self$v_proj(x)$view(c(b, s, self$num_kv_heads, self$head_dim))

    q <- self$q_norm(q)$transpose(2L, 3L) # [B, H, S, D]
    k <- self$k_norm(k)$transpose(2L, 3L)
    v <- v$transpose(2L, 3L)

    q <- ltx23_apply_split_rotary_emb(q, rope)
    k <- ltx23_apply_split_rotary_emb(k, rope)

    # GQA: repeat KV heads to match the query heads
    groups <- self$num_heads %/% self$num_kv_heads
    if (groups > 1L) {
        k <- k$repeat_interleave(groups, dim = 2L)
        v <- v$repeat_interleave(groups, dim = 2L)
    }

    out <- .ltx23_sdpa(q, k, v, attention_mask = mask)
    out <- out$transpose(2L, 3L)$reshape(c(b, s, -1L))
    self$o_proj(out)
}
)

.qwen3_mlp <- torch::nn_module(
                               "qwen3_mlp",
                               initialize = function(hidden_size, intermediate_size) {
    self$gate_proj <- torch::nn_linear(hidden_size, intermediate_size,
                                       bias = FALSE)
    self$up_proj <- torch::nn_linear(hidden_size, intermediate_size,
                                     bias = FALSE)
    self$down_proj <- torch::nn_linear(intermediate_size, hidden_size,
                                       bias = FALSE)
},
                               forward = function(x) {
    self$down_proj(torch::nnf_silu(self$gate_proj(x)) * self$up_proj(x))
}
)

.qwen3_layer <- torch::nn_module(
                                 "qwen3_layer",
                                 initialize = function(hidden_size, num_heads, num_kv_heads, head_dim,
        intermediate_size, eps = 1e-6) {
    self$self_attn <- .qwen3_attention(hidden_size, num_heads, num_kv_heads,
                                       head_dim, eps = eps)
    self$mlp <- .qwen3_mlp(hidden_size, intermediate_size)
    self$input_layernorm <- ltx23_rms_norm(hidden_size, eps = eps)
    self$post_attention_layernorm <- ltx23_rms_norm(hidden_size, eps = eps)
},
                                 forward = function(x, rope, mask = NULL) {
    x <- x + self$self_attn(self$input_layernorm(x), rope, mask)
    x + self$mlp(self$post_attention_layernorm(x))
}
)

#' Qwen3 encoder stack
#'
#' Defaults are the Qwen3-4B configuration used by FLUX.2 klein. The
#' module tree mirrors the checkpoint keys (\code{model.embed_tokens},
#' \code{model.layers.*}, \code{model.norm}); the tied LM head carries
#' no weights of its own and is not implemented.
#'
#' @param vocab_size,hidden_size,intermediate_size,num_hidden_layers
#'   Integers.
#' @param num_attention_heads,num_key_value_heads,head_dim Integers.
#' @param rope_theta Numeric.
#' @param rms_norm_eps Numeric.
#'
#' @return Module whose forward(input_ids, attention_mask = NULL,
#'   out_layers) returns a list of hidden-state tensors [B, S, hidden],
#'   one per requested layer (a value of k means the state after k
#'   layers, matching HF \code{output.hidden_states[k]}). Runs only to
#'   \code{max(out_layers)}. \code{input_ids} are 1-based.
#'
#' @export
qwen3_encoder <- torch::nn_module(
                                  "qwen3_encoder",
                                  initialize = function(vocab_size = 151936L, hidden_size = 2560L,
        intermediate_size = 9728L,
        num_hidden_layers = 36L,
        num_attention_heads = 32L,
        num_key_value_heads = 8L, head_dim = 128L,
        rope_theta = 1e6, rms_norm_eps = 1e-6) {
    self$head_dim <- head_dim
    self$rope_theta <- rope_theta

    inner <- torch::nn_module(
                              "qwen3_model",
                              initialize = function() {
        self$embed_tokens <- torch::nn_embedding(vocab_size, hidden_size)
        self$layers <- torch::nn_module_list(
            lapply(seq_len(num_hidden_layers), function(i) {
            .qwen3_layer(hidden_size, num_attention_heads,
                         num_key_value_heads, head_dim, intermediate_size,
                         eps = rms_norm_eps)
        })
        )
        self$norm <- ltx23_rms_norm(hidden_size, eps = rms_norm_eps)
    }
    )
    self$model <- inner()
},
                                  forward = function(input_ids, attention_mask = NULL,
        out_layers = c(9L, 18L, 27L)) {
    x <- self$model$embed_tokens(input_ids)
    b <- input_ids$shape[1]
    s <- input_ids$shape[2]
    device <- x$device
    dt <- x$dtype

    rope <- .qwen3_rope_tables(s, self$head_dim, self$rope_theta, device)

    # Additive causal (+ padding) mask [B, 1, S, S] in the compute dtype:
    # the fused SDPA kernel requires the bias dtype to match the query.
    # -10000 (the dit_ltx23 convention) stays finite in every float
    # dtype, including where -3.4e38 would overflow bfloat16 to -Inf.
    # The $neg()$add(1) chain (not `1 - x`) keeps torch Scalars, since
    # the Ops dispatch materializes R scalars as Float and promotes.
    neg <- -10000
    causal <- torch::torch_full(c(s, s), neg, dtype = dt,
                                device = device)$triu(diagonal = 1L)
    mask <- causal$unsqueeze(1L)$unsqueeze(1L)$expand(c(b, 1L, s, s))
    if (!is.null(attention_mask)) {
        pad <- attention_mask$to(dtype = dt, device = device)$neg()$add(1)$mul(neg)
        mask <- mask + pad$unsqueeze(2L)$unsqueeze(2L)
    }

    out_layers <- sort(as.integer(out_layers))
    states <- vector("list", length(out_layers))
    for (i in seq_len(max(out_layers))) {
        x <- self$model$layers[[i]](x, rope, mask)
        hit <- which(out_layers == i)
        if (length(hit)) {
            states[[hit]] <- x
        }
    }
    states
}
)

#' Load a Qwen3 encoder from a transformers directory
#'
#' Streams the (possibly sharded) safetensors weights into
#' \code{\link{qwen3_encoder}}. The LM head is tied to the embeddings
#' and skipped.
#'
#' @param model_path Directory with \code{config.json} and
#'   \code{model*.safetensors} (FLUX.2-klein's \code{text_encoder}).
#' @param device Character. Target device.
#' @param dtype Character. "bfloat16" (GPU) or "float32" (CPU).
#' @param verbose Logical.
#' @param ... Overrides for \code{\link{qwen3_encoder}} arguments.
#'
#' @return The loaded \code{qwen3_encoder} in eval mode.
#'
#' @export
load_qwen3_text_encoder <- function(model_path, device = "cpu",
                                    dtype = "float32", verbose = TRUE, ...) {
    model_path <- path.expand(model_path)
    config <- NULL
    config_path <- file.path(model_path, "config.json")
    if (file.exists(config_path)) {
        config <- jsonlite::fromJSON(config_path, simplifyVector = TRUE)
    }

    args <- list(vocab_size = config$vocab_size,
                 hidden_size = config$hidden_size,
                 intermediate_size = config$intermediate_size,
                 num_hidden_layers = config$num_hidden_layers,
                 num_attention_heads = config$num_attention_heads,
                 num_key_value_heads = config$num_key_value_heads,
                 head_dim = config$head_dim)
    args <- Filter(function(x) !is.null(x) && length(x) > 0L, args)
    args <- lapply(args, as.integer)
    if (!is.null(config$rope_theta)) {
        args$rope_theta <- as.numeric(config$rope_theta)
    }
    if (!is.null(config$rms_norm_eps)) {
        args$rms_norm_eps <- as.numeric(config$rms_norm_eps)
    }
    args <- utils::modifyList(args, list(...))
    model <- do.call(qwen3_encoder, args)
    model$to(dtype = .flux_dtype(dtype))

    opened <- .flux_open_sharded_dir(model_path, "model")
    ckpt <- structure(
                      list(handle = opened$handle, keys = opened$keys,
                           version = NULL, config = config, path = model_path),
                      class = "ltx23_checkpoint"
    )
    map_key <- function(key) {
        if (identical(key, "lm_head.weight")) {
            return(NA_character_) # tied to embed_tokens
        }
        key
    }
    res <- ltx23_load_group(ckpt, ckpt$keys, model, map_key = map_key,
                            verbose = verbose)
    if (length(res$unmapped) || length(res$unfilled)) {
        stop("Qwen3 encoder load: ", length(res$unmapped),
             " unmapped keys, ", length(res$unfilled), " unfilled params")
    }

    model$to(device = device)
    model$eval()
    model
}

#' Encode prompts with the Qwen3 encoder for FLUX.2
#'
#' Tokenizes with the chat template, runs the encoder with the padding
#' mask, and concatenates the requested mid-stack hidden states per
#' token, matching Flux2KleinPipeline._get_qwen3_prompt_embeds.
#'
#' @param prompts Character vector.
#' @param model A \code{\link{qwen3_encoder}}.
#' @param tokenizer A \code{\link{qwen_bpe_tokenizer}}.
#' @param max_sequence_length Integer. Fixed token length (klein: 512).
#' @param out_layers Integer vector. Hidden-state layers (klein-4B:
#'   9, 18, 27).
#' @param device Device for the input ids (defaults to the model's).
#'
#' @return Tensor [length(prompts), max_sequence_length,
#'   3 * hidden_size].
#'
#' @export
encode_with_qwen3 <- function(prompts, model, tokenizer,
                              max_sequence_length = 512L,
                              out_layers = c(9L, 18L, 27L), device = NULL) {
    enc <- encode_qwen(tokenizer, prompts, max_length = max_sequence_length,
                       chat_template = TRUE)
    device <- device %||% model$model$embed_tokens$weight$device
    long <- torch::torch_long()
    ids <- torch::torch_tensor(enc$input_ids + 1L, dtype = long,
                               device = device)
    mask <- torch::torch_tensor(enc$attention_mask, dtype = long,
                                device = device)

    states <- torch::with_no_grad(model(ids, attention_mask = mask,
                                        out_layers = out_layers))
    # stack(dim=1) + permute + reshape == per-token concatenation
    torch::torch_cat(states, dim = -1L)
}

# Parity tests for the Qwen3 encoder port against HF transformers
# reference fixtures (generated by tools/gen_fixtures_qwen3.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "qwen3_flux2.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/qwen3_flux2.safetensors"
if (!file.exists(fixture_path)) exit_file("qwen3 fixtures missing")

ckpt_dir <- system.file("tinytest", "fixtures", "qwen3_tiny_ckpt",
  package = "diffuseR")
if (ckpt_dir == "") ckpt_dir <- "fixtures/qwen3_tiny_ckpt"
if (!dir.exists(ckpt_dir)) exit_file("qwen3 tiny checkpoint missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

model <- load_qwen3_text_encoder(ckpt_dir, device = "cpu",
  dtype = "float32", verbose = FALSE)

ids <- fx$input_ids$to(dtype = torch::torch_long()) + 1L
mask <- fx$attention_mask$to(dtype = torch::torch_long())

states <- torch::with_no_grad(model(ids, attention_mask = mask,
  out_layers = c(1L, 2L, 3L)))
expect_equal(as.integer(states[[1]]$shape), as.integer(fx$h1$shape))
expect_true(max_abs_diff(states[[1]], fx$h1) < 1e-4)
expect_true(max_abs_diff(states[[2]], fx$h2) < 1e-4)
expect_true(max_abs_diff(states[[3]], fx$h3) < 1e-4)

# Per-token concatenation matches the pipeline's stack/permute/reshape
stacked <- torch::torch_cat(states, dim = -1L)
expect_equal(as.integer(stacked$shape), as.integer(fx$stacked$shape))
expect_true(max_abs_diff(stacked, fx$stacked) < 1e-4)

# The mask matters: dropping it must change the padded batch's states
states_nomask <- torch::with_no_grad(model(ids, out_layers = c(3L)))
expect_true(max_abs_diff(states_nomask[[1]], fx$h3) > 1e-4)

#' LTX-2.3 Vocoder with Bandwidth Extension
#'
#' Fresh R port of the LTX-2 BigVGAN-style vocoder from the diffusers
#' reference (Apache-2.0, pipelines/ltx2/vocoder.py). The 2.3 vocoder
#' runs a 16 kHz stage (hidden 1536, snakebeta activations with
#' anti-aliased up/downsampling), re-analyzes its output into a causal
#' log-mel spectrogram, and feeds a bandwidth-extension vocoder whose
#' residual is added to a Hann-resampled skip path for 48 kHz output.
#' The Kaiser sinc / Hann filters and STFT bases are checkpoint buffers.
#' Runs in float32 (small model; snakebeta is precision-sensitive).
#'
#' @name vocoder_ltx23
NULL

# Kaiser window (periodic = FALSE), base-R besselI implementation used
# when building filters at init; checkpoint buffers override them.
.ltx23_kaiser_window <- function(n, beta) {
    if (n == 1L) {
        return(torch::torch_ones(1L))
    }
    k <- seq(0L, n - 1L)
    ratio <- (2 * k / (n - 1)) - 1
    vals <- besselI(beta * sqrt(pmax(1 - ratio ^ 2, 0)), 0) / besselI(beta, 0)
    torch::torch_tensor(vals, dtype = torch::torch_float32())
}

#' Kaiser sinc low-pass filter kernel
#'
#' @param cutoff Numeric. Normalized cutoff in (0, 0.5].
#' @param half_width Numeric. Transition band half width.
#' @param kernel_size Integer.
#'
#' @return Tensor [kernel_size].
#'
#' @export
ltx23_kaiser_sinc_filter1d <- function(cutoff, half_width, kernel_size) {
    delta_f <- 4 * half_width
    half_size <- kernel_size %/% 2L
    amplitude <- 2.285 * (half_size - 1) * pi * delta_f + 7.95
    beta <- if (amplitude > 50) {
        0.1102 * (amplitude - 8.7)
    } else if (amplitude >= 21) {
        0.5842 * (amplitude - 21) ^ 0.4 + 0.07886 * (amplitude - 21)
    } else {
        0
    }

    window <- .ltx23_kaiser_window(kernel_size, beta)

    even <- kernel_size %% 2L == 0L
    time <- if (even) {
        torch::torch_arange(start = -half_size, end = half_size - 1,
                            dtype = torch::torch_float32()) + 0.5
    } else {
        torch::torch_arange(start = 0, end = kernel_size - 1,
                            dtype = torch::torch_float32()) - half_size
    }

    if (cutoff == 0) {
        return(torch::torch_zeros_like(time))
    }
    time <- 2 * cutoff * time
    sinc <- torch::torch_where(
                               time == 0,
                               torch::torch_ones_like(time),
                               torch::torch_sin(pi * time) / pi / time
    )
    filter <- 2 * cutoff * window * sinc
    filter / filter$sum()
}

#' Anti-aliasing 1D downsampler (low-pass then stride)
#'
#' @param ratio Integer. Downsampling ratio.
#' @param kernel_size Integer or NULL (default 6*ratio rounded even).
#'
#' @return Module whose forward(x) returns \code{x} low-pass filtered
#'   and decimated by \code{ratio} along the time axis.
#'
#' @export
ltx23_downsample1d <- torch::nn_module(
                                       "ltx23_downsample1d",
                                       initialize = function(ratio = 2L, kernel_size = NULL) {
    self$ratio <- as.integer(ratio)
    self$kernel_size <- as.integer(kernel_size %||% (as.integer(6 * ratio %/% 2) * 2L))
    self$pad_left <- self$kernel_size %/% 2L + (self$kernel_size %% 2L) - 1L
    self$pad_right <- self$kernel_size %/% 2L

    lp <- ltx23_kaiser_sinc_filter1d(0.5 / ratio, 0.6 / ratio, self$kernel_size)
    self$filter <- torch::nn_buffer(lp$view(c(1L, 1L, self$kernel_size)))
},
                                       forward = function(x) {
    num_channels <- x$shape[2]
    x <- torch::nnf_pad(x, c(self$pad_left, self$pad_right), mode = "replicate")
    torch::nnf_conv1d(x, self$filter$expand(c(num_channels, -1L, -1L)),
                      stride = self$ratio, groups = num_channels)
}
)

#' Anti-aliasing 1D upsampler (transposed low-pass)
#'
#' @param ratio Integer. Upsampling ratio.
#' @param kernel_size Integer or NULL.
#' @param window_type "kaiser" (BigVGAN default) or "hann" (final resampler).
#' @param persistent Logical. Register the filter as a buffer (present in
#'   checkpoints); FALSE stores the computed filter as a plain field.
#'
#' @return Module whose forward(x) returns \code{x} interpolated up by
#'   \code{ratio} along the time axis, with the filter padding trimmed
#'   off.
#'
#' @export
ltx23_upsample1d <- torch::nn_module(
                                     "ltx23_upsample1d",
                                     initialize = function(ratio = 2L, kernel_size = NULL, window_type = "kaiser",
        persistent = TRUE) {
    self$ratio <- as.integer(ratio)

    if (window_type == "hann") {
        rolloff <- 0.99
        lowpass_filter_width <- 6L
        width <- as.integer(ceiling(lowpass_filter_width / rolloff))
        self$kernel_size <- 2L * width * self$ratio + 1L
        self$pad <- width
        self$pad_left <- 2L * width * self$ratio
        self$pad_right <- self$kernel_size - self$ratio

        time_axis <- (torch::torch_arange(start = 0,
                end = self$kernel_size - 1, dtype = torch::torch_float32()) / ratio - width) * rolloff
        time_clamped <- time_axis$clamp(-lowpass_filter_width, lowpass_filter_width)
        window <- torch::torch_cos(time_clamped * pi / lowpass_filter_width / 2) ^ 2
        # sinc(x) = sin(pi x) / (pi x), 1 at x = 0
        sinc <- torch::torch_where(
                                   time_axis == 0,
                                   torch::torch_ones_like(time_axis),
                                   torch::torch_sin(pi * time_axis) / (pi * time_axis)
        )
        filt <- (sinc * window * rolloff / ratio)$view(c(1L, 1L, -1L))
    } else {
        self$kernel_size <- as.integer(kernel_size %||% (as.integer(6 * ratio %/% 2) * 2L))
        self$pad <- self$kernel_size %/% self$ratio - 1L
        self$pad_left <- self$pad * self$ratio + (self$kernel_size - self$ratio) %/% 2L
        self$pad_right <- self$pad * self$ratio + (self$kernel_size - self$ratio + 1L) %/% 2L
        filt <- ltx23_kaiser_sinc_filter1d(
            0.5 / ratio, 0.6 / ratio, self$kernel_size
        )$view(c(1L, 1L, -1L))
    }
    if (persistent) {
        self$filter <- torch::nn_buffer(filt)
    } else {
        # Computed filter, absent from checkpoints (moved at use time)
        self$filter <- filt
    }
},
                                     forward = function(x) {
    num_channels <- x$shape[2]
    x <- torch::nnf_pad(x, c(self$pad, self$pad), mode = "replicate")
    lp <- self$filter$to(dtype = x$dtype, device = x$device)$expand(c(num_channels, -1L, -1L))
    x <- torch::nnf_conv_transpose1d(x, lp, stride = self$ratio,
                                     groups = num_channels)$mul(self$ratio)
    n <- x$shape[3]
    x$narrow(3L, self$pad_left + 1L, n - self$pad_left - self$pad_right)
}
)

#' SnakeBeta activation
#'
#' \code{x + (1 / (beta + eps)) * sin(x * alpha)^2} with per-channel
#' log-scale alpha/beta parameters.
#'
#' @param channels Integer.
#' @param eps Numeric.
#'
#' @return Module whose forward(hidden_states) returns the Snake
#'   activation \code{x + sin(alpha * x)^2 / beta}, a tensor of the same
#'   shape as the input.
#'
#' @export
ltx23_snake_beta <- torch::nn_module(
                                     "ltx23_snake_beta",
                                     initialize = function(channels, eps = 1e-9) {
    self$eps <- eps
    self$alpha <- torch::nn_parameter(torch::torch_zeros(channels))
    self$beta <- torch::nn_parameter(torch::torch_zeros(channels))
},
                                     forward = function(hidden_states) {
    shape <- rep(1L, hidden_states$ndim)
    shape[2] <- -1L
    alpha <- torch::torch_exp(self$alpha$view(shape))
    beta <- torch::torch_exp(self$beta$view(shape))
    hidden_states + torch::torch_sin(hidden_states * alpha)$pow(2) *
    beta$add(self$eps)$reciprocal()
}
)

#' Anti-aliased activation
#'
#' Upsample 2x, apply the activation, downsample 2x.
#'
#' @param channels Integer. Channels for the SnakeBeta activation.
#' @param ratio,kernel_size Integers. Resampling config.
#'
#' @return Module whose forward(x) returns the activation applied at 2x
#'   rate (upsample, activate, downsample), a tensor of the same shape
#'   as \code{x}, with the aliasing the raw activation would introduce
#'   filtered out.
#'
#' @export
ltx23_antialias_act1d <- torch::nn_module(
    "ltx23_antialias_act1d",
    initialize = function(channels, ratio = 2L, kernel_size = 12L) {
    self$upsample <- ltx23_upsample1d(ratio = ratio, kernel_size = kernel_size)
    self$act <- ltx23_snake_beta(channels)
    self$downsample <- ltx23_downsample1d(ratio = ratio,
        kernel_size = kernel_size)
},
    forward = function(x) {
    self$downsample(self$act(self$upsample(x)))
}
)

#' Vocoder ResNet block (AMP)
#'
#' Dilated conv pairs, each preceded by an anti-aliased SnakeBeta
#' activation, with residual connections.
#'
#' @param channels Integer.
#' @param kernel_size Integer.
#' @param dilations Integer vector.
#' @param antialias_ratio,antialias_kernel_size Integers.
#'
#' @return Module whose forward(x) returns \code{x} after the dilated
#'   convolution pairs have been added back as residuals, a tensor of
#'   the same shape.
#'
#' @export
ltx23_vocoder_resblock <- torch::nn_module(
    "ltx23_vocoder_resblock",
    initialize = function(
                          channels,
                          kernel_size = 3L,
                          dilations = c(1L, 3L, 5L),
                          antialias_ratio = 2L,
                          antialias_kernel_size = 12L
    ) {
    self$convs1 <- torch::nn_module_list(lapply(dilations, function(d) {
        torch::nn_conv1d(channels, channels, kernel_size, stride = 1L,
                         dilation = d, padding = d * (kernel_size - 1L) %/% 2L)
    }))
    self$acts1 <- torch::nn_module_list(lapply(dilations, function(d) {
        ltx23_antialias_act1d(channels, antialias_ratio, antialias_kernel_size)
    }))
    self$convs2 <- torch::nn_module_list(lapply(dilations, function(d) {
        torch::nn_conv1d(channels, channels, kernel_size, stride = 1L,
                         dilation = 1L, padding = (kernel_size - 1L) %/% 2L)
    }))
    self$acts2 <- torch::nn_module_list(lapply(dilations, function(d) {
        ltx23_antialias_act1d(channels, antialias_ratio, antialias_kernel_size)
    }))
},
    forward = function(x) {
    for (i in seq_along(self$convs1)) {
        xt <- self$acts1[[i]](x)
        xt <- self$convs1[[i]](xt)
        xt <- self$acts2[[i]](xt)
        xt <- self$convs2[[i]](xt)
        x <- x + xt
    }
    x
}
)

#' LTX-2.3 vocoder stage
#'
#' Mel spectrogram [B, C, T, M] -> waveform [B, out_channels, samples].
#' Channel and mel dims are flattened into conv channels; each upsample
#' stage halves the channel count and averages three parallel ResNet
#' branches.
#'
#' @param in_channels Integer. Flattened input channels (C * mel bins / 1).
#' @param hidden_channels Integer.
#' @param out_channels Integer.
#' @param upsample_kernel_sizes,upsample_factors Integer vectors.
#' @param resnet_kernel_sizes Integer vector.
#' @param resnet_dilations List of integer vectors.
#' @param antialias_ratio,antialias_kernel_size Integers.
#' @param final_bias Logical.
#'
#' @return Module whose forward(hidden_states, time_last) returns the
#'   synthesized waveform [B, 1, samples] for a mel spectrogram.
#'
#' @export
ltx23_vocoder <- torch::nn_module(
                                  "ltx23_vocoder",
                                  initialize = function(
        in_channels = 128L,
        hidden_channels = 1536L,
        out_channels = 2L,
        upsample_kernel_sizes = c(11L, 4L, 4L, 4L, 4L, 4L),
        upsample_factors = c(5L, 2L, 2L, 2L, 2L, 2L),
        resnet_kernel_sizes = c(3L, 7L, 11L),
        resnet_dilations = list(c(1L, 3L, 5L), c(1L, 3L, 5L), c(1L, 3L, 5L)),
        antialias_ratio = 2L,
        antialias_kernel_size = 12L,
        final_bias = FALSE
    ) {
    self$num_upsample_layers <- length(upsample_kernel_sizes)
    self$resnets_per_upsample <- length(resnet_kernel_sizes)

    self$conv_in <- torch::nn_conv1d(in_channels, hidden_channels,
                                     kernel_size = 7L, stride = 1L,
                                     padding = 3L)

    upsamplers <- list()
    resnets <- list()
    input_channels <- hidden_channels
    for (i in seq_along(upsample_factors)) {
        output_channels <- input_channels %/% 2L
        upsamplers[[i]] <- torch::nn_conv_transpose1d(
            input_channels, output_channels, upsample_kernel_sizes[i],
            stride = upsample_factors[i],
            padding = (upsample_kernel_sizes[i] - upsample_factors[i]) %/% 2L
        )
        for (j in seq_along(resnet_kernel_sizes)) {
            resnets[[length(resnets) + 1L]] <- ltx23_vocoder_resblock(
                channels = output_channels,
                kernel_size = resnet_kernel_sizes[j],
                dilations = resnet_dilations[[j]],
                antialias_ratio = antialias_ratio,
                antialias_kernel_size = antialias_kernel_size
            )
        }
        input_channels <- output_channels
    }
    self$upsamplers <- torch::nn_module_list(upsamplers)
    self$resnets <- torch::nn_module_list(resnets)

    self$act_out <- ltx23_antialias_act1d(output_channels, antialias_ratio,
        antialias_kernel_size)
    self$conv_out <- torch::nn_conv1d(output_channels, out_channels, 7L,
                                      stride = 1L, padding = 3L, bias = final_bias)
},
                                  forward = function(hidden_states, time_last = FALSE) {
    if (!time_last) {
        hidden_states <- hidden_states$transpose(3L, 4L)
    }
    hidden_states <- hidden_states$flatten(start_dim = 2L, end_dim = 3L)

    hidden_states <- self$conv_in(hidden_states)
    for (i in seq_len(self$num_upsample_layers)) {
        hidden_states <- self$upsamplers[[i]](hidden_states)
        start <- (i - 1L) * self$resnets_per_upsample
        branch_outputs <- lapply(seq_len(self$resnets_per_upsample), function(j) {
            self$resnets[[start + j]](hidden_states)
        })
        hidden_states <- torch::torch_mean(
            torch::torch_stack(branch_outputs, dim = 1L), dim = 1L
        )
    }
    hidden_states <- self$act_out(hidden_states)
    self$conv_out(hidden_states)
}
)

# Causal STFT with checkpoint-loaded DFT bases
ltx23_causal_stft <- torch::nn_module(
                                      "ltx23_causal_stft",
                                      initialize = function(filter_length = 512L, hop_length = 80L, window_length = 512L) {
    self$hop_length <- as.integer(hop_length)
    self$window_length <- as.integer(window_length)
    n_freqs <- filter_length %/% 2L + 1L
    self$forward_basis <- torch::nn_buffer(torch::torch_zeros(n_freqs * 2L,
            1L, filter_length))
    self$inverse_basis <- torch::nn_buffer(torch::torch_zeros(n_freqs * 2L, 1L, filter_length))
},
                                      forward = function(waveform) {
    if (waveform$ndim == 2L) waveform <- waveform$unsqueeze(2L)
    left_pad <- max(0L, self$window_length - self$hop_length)
    waveform <- torch::nnf_pad(waveform, c(left_pad, 0L))

    spec <- torch::nnf_conv1d(waveform, self$forward_basis, stride = self$hop_length)
    n_freqs <- spec$shape[2] %/% 2L
    real <- spec$narrow(2L, 1L, n_freqs)
    imag <- spec$narrow(2L, n_freqs + 1L, n_freqs)
    magnitude <- torch::torch_sqrt(real ^ 2 + imag ^ 2)
    list(magnitude = magnitude)
}
)

#' Causal log-mel spectrogram with checkpoint-loaded bases
#'
#' @param filter_length,hop_length,window_length,num_mel_channels Integers.
#'
#' @return Module whose forward(waveform) returns the log-mel
#'   spectrogram [B, n_mels, frames], clamped at 1e-5 before the log.
#'
#' @export
ltx23_mel_stft <- torch::nn_module(
                                   "ltx23_mel_stft",
                                   initialize = function(
        filter_length = 512L,
        hop_length = 80L,
        window_length = 512L,
        num_mel_channels = 64L
    ) {
    self$stft_fn <- ltx23_causal_stft(filter_length, hop_length, window_length)
    num_freqs <- filter_length %/% 2L + 1L
    self$mel_basis <- torch::nn_buffer(torch::torch_zeros(num_mel_channels,
            num_freqs))
},
                                   forward = function(waveform) {
    magnitude <- self$stft_fn(waveform)$magnitude
    mel <- torch::torch_matmul(self$mel_basis$to(dtype = magnitude$dtype), magnitude)
    torch::torch_log(torch::torch_clamp(mel, min = 1e-5))
}
)

#' LTX-2.3 vocoder with bandwidth extension
#'
#' Full mel [B, 2, T, 64] -> 48 kHz stereo waveform pipeline: 16 kHz
#' vocoder, causal mel re-analysis, BWE vocoder residual added to a
#' Hann-resampled skip path, clamped to [-1, 1].
#'
#' @param in_channels,bwe_in_channels Integers. Flattened mel input channels.
#' @param hidden_channels,bwe_hidden_channels Integers.
#' @param out_channels Integer. Audio channels.
#' @param upsample_kernel_sizes,upsample_factors,bwe_upsample_kernel_sizes,bwe_upsample_factors
#'   Integer vectors. Per-stage transposed-conv configs.
#' @param resnet_kernel_sizes,bwe_resnet_kernel_sizes Integer vectors.
#' @param resnet_dilations,bwe_resnet_dilations Lists of integer vectors.
#' @param filter_length,window_length,num_mel_channels Integers. Mel
#'   re-analysis configuration.
#' @param input_sampling_rate,output_sampling_rate Integers.
#' @param hop_length Integer. Mel analysis hop.
#'
#' @return Module whose forward(mel_spec) returns the
#'   bandwidth-extended waveform [B, 1, samples], trimmed to the sample
#'   count implied by the input frames and the rate ratio.
#'
#' @export
ltx23_vocoder_with_bwe <- torch::nn_module(
    "ltx23_vocoder_with_bwe",
    initialize = function(
                          in_channels = 128L,
                          hidden_channels = 1536L,
                          out_channels = 2L,
                          upsample_kernel_sizes = c(11L, 4L, 4L, 4L, 4L, 4L),
                          upsample_factors = c(5L, 2L, 2L, 2L, 2L, 2L),
                          resnet_kernel_sizes = c(3L, 7L, 11L),
                          resnet_dilations = NULL,
                          bwe_in_channels = 128L,
                          bwe_hidden_channels = 512L,
                          bwe_upsample_kernel_sizes = c(12L, 11L, 4L, 4L, 4L),
                          bwe_upsample_factors = c(6L, 5L, 2L, 2L, 2L),
                          bwe_resnet_kernel_sizes = c(3L, 7L, 11L),
                          bwe_resnet_dilations = NULL,
                          filter_length = 512L,
                          hop_length = 80L,
                          window_length = 512L,
                          num_mel_channels = 64L,
                          input_sampling_rate = 16000L,
                          output_sampling_rate = 48000L
    ) {
    if (is.null(resnet_dilations)) {
        resnet_dilations <- list(c(1L, 3L, 5L), c(1L, 3L, 5L), c(1L, 3L, 5L))
    }
    if (is.null(bwe_resnet_dilations)) {
        bwe_resnet_dilations <- list(c(1L, 3L, 5L), c(1L, 3L, 5L),
                                     c(1L, 3L, 5L))
    }
    self$hop_length <- as.integer(hop_length)
    self$rate_ratio <- as.integer(output_sampling_rate %/% input_sampling_rate)

    self$vocoder <- ltx23_vocoder(
                                  in_channels = in_channels,
                                  hidden_channels = hidden_channels,
                                  out_channels = out_channels,
                                  upsample_kernel_sizes = upsample_kernel_sizes,
                                  upsample_factors = upsample_factors,
                                  resnet_kernel_sizes = resnet_kernel_sizes,
                                  resnet_dilations = resnet_dilations
    )
    self$bwe_generator <- ltx23_vocoder(
                                        in_channels = bwe_in_channels,
                                        hidden_channels = bwe_hidden_channels,
                                        out_channels = out_channels,
                                        upsample_kernel_sizes = bwe_upsample_kernel_sizes,
                                        upsample_factors = bwe_upsample_factors,
                                        resnet_kernel_sizes = bwe_resnet_kernel_sizes,
                                        resnet_dilations = bwe_resnet_dilations
    )
    self$mel_stft <- ltx23_mel_stft(
                                    filter_length = filter_length,
                                    hop_length = hop_length,
                                    window_length = window_length,
                                    num_mel_channels = num_mel_channels
    )
    self$resampler <- ltx23_upsample1d(ratio = self$rate_ratio, window_type = "hann",
                                       persistent = FALSE)
},
    forward = function(mel_spec) {
    x <- self$vocoder(mel_spec)
    num_channels <- x$shape[2]
    num_samples <- x$shape[3]

    remainder <- num_samples %% self$hop_length
    if (remainder != 0L) {
        x <- torch::nnf_pad(x, c(0L, self$hop_length - remainder))
    }

    mel <- self$mel_stft(x$flatten(start_dim = 1L, end_dim = 2L))
    mel <- mel$unflatten(1L, c(-1L, num_channels))

    # [B, C, mel_bins, frames] -> [B, C, frames, mel_bins]
    residual <- self$bwe_generator(mel$transpose(3L, 4L))

    skip <- self$resampler(x)
    waveform <- torch::torch_clamp(residual + skip, -1, 1)
    output_samples <- num_samples * self$rate_ratio
    waveform$narrow(3L, 1L, min(output_samples, waveform$shape[3]))
}
)

#' Map an official vocoder checkpoint key to the R module name
#'
#' @param key Character. Checkpoint key.
#'
#' @return Character. Module parameter/buffer name.
#'
#' @export
ltx23_map_vocoder_key <- function(key) {
    key <- sub("^vocoder\\.", "", key)
    key <- gsub("resblocks", "resnets", key, fixed = TRUE)
    key <- gsub("conv_pre", "conv_in", key, fixed = TRUE)
    key <- gsub("conv_post", "conv_out", key, fixed = TRUE)
    key <- gsub("act_post", "act_out", key, fixed = TRUE)
    key <- gsub("downsample.lowpass", "downsample", key, fixed = TRUE)
    key <- gsub(".ups.", ".upsamplers.", key, fixed = TRUE)
    key <- sub("^ups\\.", "upsamplers.", key)
    key
}

# Parity tests for the LTX-2.3 audio VAE decoder and vocoder port
# against diffusers reference fixtures (tools/gen_fixtures_audio.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "audio_ltx23.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/audio_ltx23.safetensors"
if (!file.exists(fixture_path)) exit_file("audio fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

load_group <- function(module, prefix) {
  keys <- grep(paste0("^", prefix, "\\."), names(fx), value = TRUE)
  w <- fx[keys]
  names(w) <- sub(paste0("^", prefix, "\\."), "", keys)
  dests <- c(module$named_parameters(), module$named_buffers())
  if (!setequal(names(w), names(dests))) {
    stop(
      prefix, ": name mismatch. missing dest: ",
      paste(utils::head(setdiff(names(w), names(dests)), 4), collapse = ", "),
      " | unfilled: ",
      paste(utils::head(setdiff(names(dests), names(w)), 4), collapse = ", ")
    )
  }
  torch::with_no_grad({
    for (name in names(w)) dests[[name]]$copy_(w[[name]])
  })
  module$eval()
  module
}

# --- Audio VAE decoder --------------------------------------------------------

adec <- load_group(
  ltx23_audio_decoder(
    base_channels = 8L, output_channels = 2L, num_res_blocks = 1L,
    latent_channels = 4L, ch_mult = c(1L, 2L), mel_bins = 8L
  ),
  "adec"
)
torch::with_no_grad(out_adec <- adec(fx$adec_x))
expect_equal(as.integer(out_adec$shape), c(1L, 2L, 17L, 8L))# 5 * 4 - 3 frames
expect_true(max_abs_diff(out_adec, fx$adec_out) < 1e-5)

# --- Kaiser sinc filter + window ------------------------------------------------

expect_true(max_abs_diff(
  ltx23_kaiser_sinc_filter1d(0.25, 0.3, 12L), fx$kaiser_filt_12
) < 1e-5)
expect_true(max_abs_diff(
  ltx23_kaiser_sinc_filter1d(0.1, 0.12, 13L), fx$kaiser_filt_13
) < 1e-5)
expect_true(max_abs_diff(
  diffuseR:::.ltx23_kaiser_window(12L, 4.7), fx$kaiser_window_12
) < 1e-6)

# --- Up/DownSample1d ---------------------------------------------------------------

ds1 <- ltx23_downsample1d(ratio = 2L, kernel_size = 12L)
torch::with_no_grad(out_down <- ds1(fx$w_x))
expect_true(max_abs_diff(out_down, fx$w_down) < 1e-5)

us1 <- ltx23_upsample1d(ratio = 2L, kernel_size = 12L)
torch::with_no_grad(out_up <- us1(fx$w_x))
expect_equal(as.integer(out_up$shape), as.integer(fx$w_up$shape))
expect_true(max_abs_diff(out_up, fx$w_up) < 1e-5)

ush <- ltx23_upsample1d(ratio = 4L, window_type = "hann")
torch::with_no_grad(out_uph <- ush(fx$w_x))
expect_equal(as.integer(out_uph$shape), as.integer(fx$w_up_hann$shape))
expect_true(max_abs_diff(out_uph, fx$w_up_hann) < 1e-5)

# --- SnakeBeta -----------------------------------------------------------------------

sb <- ltx23_snake_beta(3L)
torch::with_no_grad({
  sb$alpha$copy_(fx$sb.alpha)
  sb$beta$copy_(fx$sb.beta)
  out_sb <- sb(fx$w_x)
})
expect_true(max_abs_diff(out_sb, fx$sb_out) < 1e-5)

# --- Tiny vocoder stage -----------------------------------------------------------------

voc <- load_group(
  ltx23_vocoder(
    in_channels = 8L, hidden_channels = 16L, out_channels = 2L,
    upsample_kernel_sizes = c(4L, 4L), upsample_factors = c(2L, 2L),
    resnet_kernel_sizes = c(3L), resnet_dilations = list(c(1L, 3L)),
    final_bias = FALSE
  ),
  "voc"
)
torch::with_no_grad(out_voc <- voc(fx$voc_x))
expect_equal(as.integer(out_voc$shape), c(1L, 2L, 24L))
expect_true(max_abs_diff(out_voc, fx$voc_out) < 1e-4)

# --- Tiny BWE wrapper --------------------------------------------------------------------

bwe <- load_group(
  ltx23_vocoder_with_bwe(
    in_channels = 8L, hidden_channels = 16L, out_channels = 2L,
    upsample_kernel_sizes = c(4L, 4L), upsample_factors = c(2L, 2L),
    resnet_kernel_sizes = c(3L), resnet_dilations = list(c(1L, 3L)),
    bwe_in_channels = 16L, bwe_hidden_channels = 8L,
    bwe_upsample_kernel_sizes = c(8L, 4L), bwe_upsample_factors = c(4L, 2L),
    bwe_resnet_kernel_sizes = c(3L), bwe_resnet_dilations = list(c(1L, 3L)),
    filter_length = 8L, hop_length = 2L, window_length = 8L,
    num_mel_channels = 8L,
    input_sampling_rate = 100L, output_sampling_rate = 400L
  ),
  "bwe"
)
torch::with_no_grad(out_bwe <- bwe(fx$voc_x))
expect_equal(as.integer(out_bwe$shape), c(1L, 2L, 96L))
expect_true(max_abs_diff(out_bwe, fx$bwe_out) < 1e-4)

# --- Key mappers ----------------------------------------------------------------------------

expect_equal(
  ltx23_map_vocoder_key("vocoder.vocoder.conv_pre.weight"),
  "vocoder.conv_in.weight"
)
expect_equal(
  ltx23_map_vocoder_key("vocoder.vocoder.resblocks.0.acts1.1.downsample.lowpass.filter"),
  "vocoder.resnets.0.acts1.1.downsample.filter"
)
expect_equal(
  ltx23_map_vocoder_key("vocoder.vocoder.ups.2.weight"),
  "vocoder.upsamplers.2.weight"
)
expect_equal(
  ltx23_map_vocoder_key("vocoder.bwe_generator.act_post.act.alpha"),
  "bwe_generator.act_out.act.alpha"
)
expect_equal(
  ltx23_map_vocoder_key("vocoder.mel_stft.stft_fn.forward_basis"),
  "mel_stft.stft_fn.forward_basis"
)
expect_equal(
  ltx23_map_audio_vae_key("audio_vae.decoder.up.1.upsample.conv.conv.weight"),
  "decoder.up.1.upsample.conv.conv.weight"
)
expect_equal(
  ltx23_map_audio_vae_key("audio_vae.per_channel_statistics.mean-of-means"),
  "latents_mean"
)
expect_equal(
  ltx23_map_audio_vae_key("audio_vae.encoder.conv_in.conv.weight"),
  "encoder.conv_in.conv.weight"
)
expect_equal(
  ltx23_map_audio_vae_key("audio_vae.encoder.down.1.downsample.conv.weight"),
  "encoder.down.1.downsample.conv.weight"
)

# --- Encoder: shapes and key-name census -------------------------------------------

enc <- ltx23_audio_encoder(
  base_channels = 8L, in_channels = 2L, num_res_blocks = 2L,
  latent_channels = 4L, ch_mult = c(1L, 2L)
)
enc$eval()
mel_in <- torch::torch_randn(1L, 2L, 12L, 8L)
torch::with_no_grad(out_enc <- enc(mel_in))
# One downsample level: time and mel halved; moments = 2 * latent
expect_equal(as.integer(out_enc$shape), c(1L, 8L, 6L, 4L))

# Parameter names line up with the checkpoint layout
pn <- names(enc$named_parameters())
expect_true("conv_in.conv.weight" %in% pn)
expect_true("down.0.block.0.conv1.conv.weight" %in% pn)
expect_true("down.0.block.1.conv2.conv.bias" %in% pn)
expect_true("down.0.downsample.conv.weight" %in% pn)
expect_true("down.1.block.0.nin_shortcut.conv.weight" %in% pn)
expect_true("mid.block_1.conv1.conv.weight" %in% pn)
expect_true("mid.block_2.conv2.conv.bias" %in% pn)
expect_true("conv_out.conv.weight" %in% pn)
# Pixel norms are parameterless: exactly the conv tensors
expect_true(all(grepl("conv", pn)))

# encode() through the vae wrapper returns mean/logvar of latent size
avae <- ltx23_audio_vae(
  base_channels = 8L, output_channels = 2L, num_res_blocks = 1L,
  latent_channels = 4L, ch_mult = c(1L, 2L), mel_bins = 8L
)
avae$eval()
torch::with_no_grad(m <- avae$encode(torch::torch_randn(1L, 2L, 12L, 8L)))
expect_equal(as.integer(m$mean$shape)[2], 4L)
expect_equal(as.integer(m$logvar$shape)[2], 4L)

#!/usr/bin/env r
#
# End-to-end test for LTX-2 pipeline
#
# Tests the pipeline structure without full model weights.
# Use text_backend="random" to skip text encoder loading.
#

library(torch)

# Load diffuseR
if (requireNamespace("diffuseR", quietly = TRUE)) {
  library(diffuseR)
} else {
  rhydrogen::load_all()
}

cat("=== LTX-2 End-to-End Test ===\n\n")

# Test 1: Text encoding + connectors (skip DiT for now - too large for single GPU)
cat("--- Test 1: Text encoding + connectors ---\n")
tryCatch({
  library(diffuseR)

  # Test encode_text_ltx2 + connectors
  t0 <- Sys.time()

  # Random packed embeddings
  packed_embeds <- encode_text_ltx2(
    prompt = "Test prompt for connector pipeline",
    backend = "random",
    max_sequence_length = 128L,
    caption_channels = 3840L
  )
  cat(sprintf("Packed embeds shape: %s\n",
              paste(as.integer(packed_embeds$prompt_embeds$shape), collapse = "x")))

  # Load and apply connectors
  connector_path <- Sys.glob("~/.cache/huggingface/hub/models--Lightricks--LTX-2/snapshots/*/connectors/diffusion_pytorch_model.safetensors")[1]
  connectors <- load_ltx2_connectors(connector_path, verbose = FALSE)

  result <- connectors(packed_embeds$prompt_embeds, packed_embeds$prompt_attention_mask)
  video_embeds <- result[[1]]
  audio_embeds <- result[[2]]

  elapsed <- as.numeric(Sys.time() - t0)
  cat(sprintf("Video embeds: %s\n", paste(as.integer(video_embeds$shape), collapse = "x")))
  cat(sprintf("Audio embeds: %s\n", paste(as.integer(audio_embeds$shape), collapse = "x")))
  cat(sprintf("Pipeline (random->connectors) in %.2fs\n", elapsed))
  cat("Test 1: PASSED\n\n")
}, error = function(e) {
  cat(sprintf("Test 1: FAILED - %s\n\n", e$message))
})

# Test 2: INT4 weight loading
cat("--- Test 2: INT4 weight loading ---\n")
tryCatch({
  int4_path <- path.expand("~/.cache/diffuseR/ltx2_transformer_int4.safetensors")
  t0 <- Sys.time()
  weights <- load_int4_weights(int4_path, verbose = TRUE)
  load_time <- as.numeric(Sys.time() - t0)
  cat(sprintf("Loaded %d quantized parameters in %.2fs\n", length(weights), load_time))
  cat("Test 2: PASSED\n\n")
}, error = function(e) {
  cat(sprintf("Test 2: FAILED - %s\n\n", e$message))
})

# Test 3: Connector loading
cat("--- Test 3: Connector loading ---\n")
tryCatch({
  connector_path <- Sys.glob("~/.cache/huggingface/hub/models--Lightricks--LTX-2/snapshots/*/connectors/diffusion_pytorch_model.safetensors")[1]
  t0 <- Sys.time()
  connectors <- load_ltx2_connectors(connector_path, verbose = TRUE)
  load_time <- as.numeric(Sys.time() - t0)
  cat(sprintf("Connectors loaded in %.2fs\n", load_time))

  # Test forward pass
  test_embeds <- torch_randn(c(1L, 128L, 188160L))  # [B, seq_len, 3840*49]
  test_mask <- torch_ones(c(1L, 128L), dtype = torch_int64())
  result <- connectors(test_embeds, test_mask)
  cat(sprintf("Video embeds: %s\n", paste(as.integer(result[[1]]$shape), collapse = "x")))
  cat(sprintf("Audio embeds: %s\n", paste(as.integer(result[[2]]$shape), collapse = "x")))
  cat("Test 3: PASSED\n\n")
}, error = function(e) {
  cat(sprintf("Test 3: FAILED - %s\n\n", e$message))
})

# Test 4: INT4Linear layer
cat("--- Test 4: INT4Linear layer ---\n")
tryCatch({
  # Create a standard linear layer
  linear <- torch::nn_linear(4096L, 4096L, bias = TRUE)
  linear <- linear$to(dtype = torch::torch_float16(), device = "cuda")

  # Test input
  x <- torch::torch_randn(c(1L, 128L, 4096L), dtype = torch::torch_float16(), device = "cuda")

  # Standard forward
  t0 <- Sys.time()
  y_std <- linear(x)
  elapsed_std <- as.numeric(Sys.time() - t0) * 1000

  # Convert to INT4
  int4_layer <- linear_to_int4(linear, device = "cuda", dtype = torch::torch_float16())

  # INT4 forward
  t0 <- Sys.time()
  y_int4 <- int4_layer(x)
  elapsed_int4 <- as.numeric(Sys.time() - t0) * 1000

  # Check output similarity
  diff <- (y_std - y_int4)$abs()$mean()$item()
  cat(sprintf("  Standard linear: %.1fms\n", elapsed_std))
  cat(sprintf("  INT4 linear: %.1fms\n", elapsed_int4))
  cat(sprintf("  Mean abs diff: %.6f\n", diff))

  # Memory comparison
  std_mem <- prod(linear$weight$shape) * 2 / 1e6  # float16
  int4_mem <- (int4_layer$weight_packed$numel() + int4_layer$weight_scales$numel() * 4) / 1e6
  cat(sprintf("  Memory: %.2f MB (std) vs %.2f MB (INT4) = %.1fx\n",
              std_mem, int4_mem, std_mem / int4_mem))

  # Cleanup
  rm(linear, int4_layer, x, y_std, y_int4)
  gc()
  torch::cuda_empty_cache()

  cat("Test 4: PASSED\n\n")
}, error = function(e) {
  cat(sprintf("Test 4: FAILED - %s\n\n", e$message))
})

# Test 5: INT4Linear from pre-quantized disk weights
cat("--- Test 5: INT4Linear from pre-quantized disk weights ---\n")
tryCatch({
  # Load INT4 weights
  int4_path <- path.expand("~/.cache/diffuseR/ltx2_transformer_int4.safetensors")
  q <- load_int4_weights(int4_path, verbose = FALSE)

  # Clear GPU memory first
  gc(full = TRUE)
  torch::cuda_empty_cache()

  # Create INT4Linear from a real DiT weight
  weight_name <- "transformer_blocks.0.attn1.to_q.weight"
  if (weight_name %in% names(q)) {
    q_weight <- q[[weight_name]]

    # Create layer directly from quantized weights
    t0 <- Sys.time()
    int4_layer <- int4_linear_from_quantized(q_weight, device = "cuda",
                                              dtype = torch::torch_float16())
    create_time <- as.numeric(Sys.time() - t0) * 1000

    # Test forward pass
    out_f <- q_weight$orig_shape[1]
    in_f <- q_weight$orig_shape[2]
    x <- torch::torch_randn(c(1L, 128L, in_f), dtype = torch::torch_float16(), device = "cuda")

    t0 <- Sys.time()
    y <- int4_layer(x)
    forward_time <- as.numeric(Sys.time() - t0) * 1000

    cat(sprintf("  Layer: %dx%d\n", out_f, in_f))
    cat(sprintf("  Create from INT4: %.1fms\n", create_time))
    cat(sprintf("  Forward pass: %.1fms\n", forward_time))
    cat(sprintf("  Output shape: %s\n", paste(as.integer(y$shape), collapse = "x")))

    # Cleanup
    rm(int4_layer, x, y)
    gc()
    torch::cuda_empty_cache()
  }

  cat("Test 5: PASSED\n\n")
}, error = function(e) {
  cat(sprintf("Test 5: FAILED - %s\n\n", e$message))
})

cat("=== Tests Complete ===\n")

#' Auto-Configure Device Assignment
#'
#' Automatically determines optimal device configuration for diffusion model
#' components based on available VRAM (via nvidia-smi) and GPU architecture.
#'
#' @param model Character. Model type: "sd21" or "sdxl".
#' @param strategy Character. Memory strategy: "auto" (default), "full_gpu",
#'   "unet_gpu", or "cpu_only". See Details.
#'
#' @return A named list of device assignments suitable for `models2devices()`.
#'
#' @details
#' Strategies:
#' \describe{
#'   \item{"auto"}{Detect free VRAM and choose the best strategy}
#'   \item{"full_gpu"}{All components on CUDA (10GB+ free VRAM for SDXL)}
#'   \item{"unet_gpu"}{Only unet on CUDA, rest on CPU (6GB+ for SDXL)}
#'   \item{"cpu_only"}{All components on CPU}
#' }
#'
#' On Blackwell GPUs (RTX 50xx), "unet_gpu" is forced due to TorchScript
#' compatibility issues, regardless of available VRAM. The native modules
#' (`use_native_unet` and friends) do not have this restriction.
#'
#' @export
#'
#' @examples
#' # Force a strategy: no GPU or nvidia-smi needed.
#' str(auto_devices("sdxl", strategy = "cpu_only"))
#'
#' str(auto_devices("sd21", strategy = "unet_gpu"))
#'
#' # Auto-detect free VRAM and pick a strategy for this machine.
#' str(auto_devices("sdxl"))
auto_devices <- function(model = "sdxl", strategy = "auto") {
    # Free-VRAM requirements in GB (float16 component sizes + overhead)
    requirements <- list(sd21 = list(full_gpu = 4, unet_gpu = 3),
                         sdxl = list(full_gpu = 10, unet_gpu = 6))
    req <- requirements[[model]]
    if (is.null(req)) {
        stop("Unsupported model: ", model, ". Supported: ",
             paste(names(requirements), collapse = ", "))
    }

    if (strategy == "auto") {
        vram <- .detect_vram(use_free = TRUE)
        strategy <- if (is_blackwell_gpu()) {
            "unet_gpu"
        } else if (vram >= req$full_gpu) {
            "full_gpu"
        } else if (vram >= req$unet_gpu) {
            "unet_gpu"
        } else {
            "cpu_only"
        }
        message(sprintf("auto_devices: %s (%.1f GB free VRAM)", strategy, vram))
    } else if (identical(strategy, "full_gpu") && is_blackwell_gpu()) {
        # TorchScript workaround: full_gpu is not supported on Blackwell
        message("Blackwell GPU detected - overriding full_gpu to unet_gpu")
        strategy <- "unet_gpu"
    }

    .build_fallback_devices(model, strategy)
}

#' Build fallback device configuration
#'
#' @param model Character. Model type.
#' @param strategy Character. Memory strategy.
#' @return Named list of device assignments.
#' @keywords internal
.build_fallback_devices <- function(model, strategy) {
    # Components by model
    components <- list(
                       sd21 = c("unet", "decoder", "text_encoder", "encoder"),
                       sdxl = c("unet", "decoder", "text_encoder", "text_encoder2", "encoder")
    )

    if (!model %in% names(components)) {
        stop("Unsupported model: ", model)
    }

    comp <- components[[model]]

    if (strategy == "full_gpu") {
        devices <- as.list(rep("cuda", length(comp)))
        names(devices) <- comp
    } else if (strategy == "unet_gpu") {
        devices <- as.list(rep("cpu", length(comp)))
        names(devices) <- comp
        devices$unet <- "cuda"
    } else {
        # cpu_only
        devices <- as.list(rep("cpu", length(comp)))
        names(devices) <- comp
    }

    devices
}

# Test auto_devices function

# Test auto_devices returns correct structure for SDXL
devices <- auto_devices("sdxl", strategy = "unet_gpu")
expect_true(is.list(devices))
expect_true("unet" %in% names(devices))
expect_true("decoder" %in% names(devices))
expect_true("text_encoder" %in% names(devices))
expect_true("text_encoder2" %in% names(devices))
expect_true("encoder" %in% names(devices))

# Test unet_gpu strategy
expect_equal(devices$unet, "cuda")
expect_equal(devices$decoder, "cpu")
expect_equal(devices$text_encoder, "cpu")

# Test SD21 (no text_encoder2)
devices_sd21 <- auto_devices("sd21", strategy = "cpu_only")
expect_true("unet" %in% names(devices_sd21))
expect_false("text_encoder2" %in% names(devices_sd21))
expect_equal(devices_sd21$unet, "cpu")

# Test full_gpu strategy (may be overridden on Blackwell)
devices_full <- auto_devices("sdxl", strategy = "full_gpu")
expect_equal(devices_full$unet, "cuda")
# On Blackwell, full_gpu gets overridden to unet_gpu
if (is_blackwell_gpu()) {
  expect_equal(devices_full$decoder, "cpu")
} else {
  expect_equal(devices_full$decoder, "cuda")
}

# Test cpu_only strategy
devices_cpu <- auto_devices("sdxl", strategy = "cpu_only")
expect_equal(devices_cpu$unet, "cpu")
expect_equal(devices_cpu$decoder, "cpu")

# Test invalid model
expect_error(auto_devices("invalid_model"))

# Test fallback function
fallback <- diffuseR:::.build_fallback_devices("sdxl", "unet_gpu")
expect_equal(fallback$unet, "cuda")
expect_equal(fallback$decoder, "cpu")

#' FLUX.2 Transformer Building Blocks
#'
#' Fresh R port of the FLUX.2 MMDiT blocks from the diffusers reference
#' implementation (Apache-2.0,
#' src/diffusers/models/transformers/transformer_flux2.py). Key
#' differences from FLUX.1: modulation is computed ONCE at model level
#' by shared \code{flux2_modulation} projections and passed into the
#' blocks (block norms are parameterless), feed-forwards use SwiGLU with
#' the gate fused into \code{linear_in}, the single-stream block is a
#' ViT-22B-style parallel block with fully fused projections, and every
#' linear is bias-free. Module field names mirror the diffusers
#' state-dict keys 1:1. Reuses \code{flux_attention} (bias = FALSE),
#' \code{ltx23_rms_norm}, \code{.ltx23_sdpa}, and
#' \code{flux_apply_rotary_emb}.
#'
#' @name dit_flux2_modules
NULL

#' FLUX.2 shared modulation projection
#'
#' \code{linear(silu(temb))} producing \code{mod_param_sets} triples of
#' (shift, scale, gate). Computed once per forward at model level and
#' broadcast to every block. Reference: Flux2Modulation.
#'
#' @param dim Integer. Model dimension.
#' @param mod_param_sets Integer. Number of (shift, scale, gate) triples.
#' @param bias Logical.
#'
#' @return Module whose forward(temb) returns the modulation tensor
#'   \code{linear(silu(temb))}, holding \code{mod_param_sets} triples of
#'   (shift, scale, gate) along the last axis.
#'
#' @export
flux2_modulation <- torch::nn_module(
                                     "flux2_modulation",
                                     initialize = function(dim, mod_param_sets = 2L, bias = FALSE) {
    self$mod_param_sets <- mod_param_sets
    self$linear <- torch::nn_linear(dim, dim * 3L * mod_param_sets, bias = bias)
},
                                     forward = function(temb) {
    self$linear(torch::nnf_silu(temb))
}
)

# Split a modulation tensor into mod_param_sets triples of
# (shift, scale, gate), each [N, 1, dim]. Reference: Flux2Modulation.split.
.flux2_mod_split <- function(mod, mod_param_sets) {
    if (mod$ndim == 2L) {
        mod <- mod$unsqueeze(2L)
    }
    parts <- mod$chunk(3L * mod_param_sets, dim = -1L)
    lapply(seq_len(mod_param_sets), function(i) {
        parts[(3L * (i - 1L) + 1L):(3L * i)]
    })
}

#' FLUX.2 feed-forward (fused SwiGLU)
#'
#' \code{linear_in} projects to twice the inner dim; SwiGLU gates the
#' first half with SiLU and multiplies by the second half;
#' \code{linear_out} projects back. Reference: Flux2FeedForward +
#' Flux2SwiGLU.
#'
#' @param dim Integer. Input dimension.
#' @param dim_out Integer. Output dimension (defaults to \code{dim}).
#' @param mult Numeric. Inner dim multiplier (FLUX.2: 3.0).
#' @param bias Logical.
#'
#' @return Module whose forward(x) returns the SwiGLU-gated projection
#'   of \code{x}, a tensor with the last axis of width \code{dim_out}.
#'
#' @export
flux2_feed_forward <- torch::nn_module(
                                       "flux2_feed_forward",
                                       initialize = function(dim, dim_out = NULL, mult = 3.0, bias = FALSE) {
    inner_dim <- as.integer(dim * mult)
    self$linear_in <- torch::nn_linear(dim, inner_dim * 2L, bias = bias)
    self$linear_out <- torch::nn_linear(inner_dim, dim_out %||% dim,
                                        bias = bias)
},
                                       forward = function(x) {
    x <- self$linear_in(x)
    half <- x$shape[length(x$shape)] %/% 2L
    x <- torch::nnf_silu(x$narrow(-1L, 1L, half)) *
    x$narrow(-1L, half + 1L, half)
    self$linear_out(x)
}
)

#' FLUX.2 parallel self-attention (single-stream)
#'
#' ViT-22B-style parallel block internals: one fused projection produces
#' QKV and the SwiGLU MLP input; one fused projection consumes
#' cat(attention output, MLP output). Reference:
#' Flux2ParallelSelfAttention + Flux2ParallelSelfAttnProcessor.
#'
#' @param query_dim Integer. Model dimension.
#' @param heads Integer. Attention heads.
#' @param dim_head Integer. Per-head dimension.
#' @param mlp_ratio Numeric. MLP hidden multiplier (FLUX.2: 3.0).
#' @param eps Numeric. RMS norm epsilon.
#' @param bias Logical.
#'
#' @return Module whose forward(hidden_states, image_rotary_emb,
#'   chunk_size) returns the block output [B, S, query_dim]: attention
#'   and MLP branches computed in parallel from one fused projection,
#'   concatenated, and projected back by a second fused layer.
#'
#' @export
flux2_parallel_self_attention <- torch::nn_module(
    "flux2_parallel_self_attention",
    initialize = function(query_dim, heads, dim_head, mlp_ratio = 3.0,
                          eps = 1e-6, bias = FALSE) {
    inner_dim <- heads * dim_head
    self$heads <- heads
    self$inner_dim <- inner_dim
    self$mlp_hidden_dim <- as.integer(query_dim * mlp_ratio)

    self$to_qkv_mlp_proj <- torch::nn_linear(query_dim,
        inner_dim * 3L + self$mlp_hidden_dim * 2L, bias = bias)
    self$norm_q <- ltx23_rms_norm(dim_head, eps = eps)
    self$norm_k <- ltx23_rms_norm(dim_head, eps = eps)
    self$to_out <- torch::nn_linear(inner_dim + self$mlp_hidden_dim,
                                    query_dim, bias = bias)
},
    forward = function(hidden_states, image_rotary_emb = NULL,
                       chunk_size = NULL) {
    proj <- self$to_qkv_mlp_proj(hidden_states)
    qkv <- proj$narrow(-1L, 1L, 3L * self$inner_dim)
    mlp <- proj$narrow(-1L, 3L * self$inner_dim + 1L, self$mlp_hidden_dim * 2L)

    parts <- qkv$chunk(3L, dim = -1L)
    query <- parts[[1]]$unflatten(3L, c(self$heads, -1L))
    key <- parts[[2]]$unflatten(3L, c(self$heads, -1L))
    value <- parts[[3]]$unflatten(3L, c(self$heads, -1L))

    query <- self$norm_q(query)
    key <- self$norm_k(key)

    # [B, S, H, D] -> [B, H, S, D]
    query <- query$transpose(2L, 3L)
    key <- key$transpose(2L, 3L)
    value <- value$transpose(2L, 3L)
    if (!is.null(image_rotary_emb)) {
        query <- flux_apply_rotary_emb(query, image_rotary_emb)
        key <- flux_apply_rotary_emb(key, image_rotary_emb)
    }
    attn <- .ltx23_sdpa(query, key, value, chunk_size = chunk_size)
    attn <- attn$transpose(2L, 3L)$flatten(start_dim = 3L)
    attn <- attn$to(dtype = hidden_states$dtype)

    # SwiGLU on the fused MLP half
    half <- self$mlp_hidden_dim
    mlp <- torch::nnf_silu(mlp$narrow(-1L, 1L, half)) *
    mlp$narrow(-1L, half + 1L, half)

    # Attention half first, then the MLP half
    self$to_out(torch::torch_cat(list(attn, mlp), dim = -1L))
}
)

#' FLUX.2 double-stream (MMDiT) block
#'
#' Image and text streams with externally supplied (shift, scale, gate)
#' modulation triples, joint attention (txt first), and SwiGLU
#' feed-forwards. Reference: Flux2TransformerBlock.
#'
#' @param dim Integer. Model dimension.
#' @param num_attention_heads Integer. Attention heads.
#' @param attention_head_dim Integer. Per-head dimension.
#' @param mlp_ratio Numeric. FF multiplier (FLUX.2: 3.0).
#' @param eps Numeric. Norm epsilon.
#' @param bias Logical.
#'
#' @return Module whose forward(hidden_states, encoder_hidden_states,
#'   temb_mod_img, temb_mod_txt, image_rotary_emb) returns
#'   \code{list(encoder_hidden_states, hidden_states)}.
#'
#' @export
flux2_double_block <- torch::nn_module(
                                       "flux2_double_block",
                                       initialize = function(dim, num_attention_heads, attention_head_dim,
        mlp_ratio = 3.0, eps = 1e-6, bias = FALSE) {
    self$norm1 <- torch::nn_layer_norm(dim, eps = eps,
                                       elementwise_affine = FALSE)
    self$norm1_context <- torch::nn_layer_norm(dim, eps = eps,
        elementwise_affine = FALSE)
    self$attn <- flux_attention(dim, num_attention_heads,
                                attention_head_dim, added_kv = TRUE,
                                eps = eps, bias = bias)
    self$norm2 <- torch::nn_layer_norm(dim, eps = eps,
                                       elementwise_affine = FALSE)
    self$ff <- flux2_feed_forward(dim, dim, mult = mlp_ratio, bias = bias)
    self$norm2_context <- torch::nn_layer_norm(dim, eps = eps,
        elementwise_affine = FALSE)
    self$ff_context <- flux2_feed_forward(dim, dim, mult = mlp_ratio,
        bias = bias)
},
                                       forward = function(hidden_states, encoder_hidden_states, temb_mod_img,
        temb_mod_txt, image_rotary_emb = NULL,
        chunk_size = NULL) {
    mi <- .flux2_mod_split(temb_mod_img, 2L)
    mt <- .flux2_mod_split(temb_mod_txt, 2L)
    # Each triple is (shift, scale, gate)
    msa <- mi[[1]]
    mlp <- mi[[2]]
    c_msa <- mt[[1]]
    c_mlp <- mt[[2]]

    norm_h <- self$norm1(hidden_states) * msa[[2]]$add(1) + msa[[1]]
    norm_c <- self$norm1_context(encoder_hidden_states) *
    c_msa[[2]]$add(1) + c_msa[[1]]

    attn_out <- self$attn(
                          hidden_states = norm_h,
                          encoder_hidden_states = norm_c,
                          image_rotary_emb = image_rotary_emb,
                          chunk_size = chunk_size
    )

    hidden_states <- hidden_states + msa[[3]] * attn_out[[1]]
    norm_h <- self$norm2(hidden_states) * mlp[[2]]$add(1) + mlp[[1]]
    hidden_states <- hidden_states + mlp[[3]] * self$ff(norm_h)

    encoder_hidden_states <- encoder_hidden_states + c_msa[[3]] * attn_out[[2]]
    norm_c <- self$norm2_context(encoder_hidden_states) *
    c_mlp[[2]]$add(1) + c_mlp[[1]]
    encoder_hidden_states <- encoder_hidden_states +
    c_mlp[[3]] * self$ff_context(norm_c)

    if (encoder_hidden_states$dtype == torch::torch_float16()) {
        encoder_hidden_states <- encoder_hidden_states$clamp(-65504, 65504)
    }
    list(encoder_hidden_states, hidden_states)
}
)

#' FLUX.2 single-stream block (parallel)
#'
#' Parameterless LayerNorm with external modulation, then the fused
#' parallel attention+MLP. Operates on the pre-concatenated [text; image]
#' sequence (the reference model concatenates once before the stack).
#' Reference: Flux2SingleTransformerBlock.
#'
#' @param dim Integer. Model dimension.
#' @param num_attention_heads Integer. Attention heads.
#' @param attention_head_dim Integer. Per-head dimension.
#' @param mlp_ratio Numeric. MLP multiplier (FLUX.2: 3.0).
#' @param eps Numeric. Norm epsilon.
#' @param bias Logical.
#'
#' @return Module whose forward(hidden_states, temb_mod,
#'   image_rotary_emb) returns the joint hidden states.
#'
#' @export
flux2_single_block <- torch::nn_module(
                                       "flux2_single_block",
                                       initialize = function(dim, num_attention_heads, attention_head_dim,
        mlp_ratio = 3.0, eps = 1e-6, bias = FALSE) {
    self$norm <- torch::nn_layer_norm(dim, eps = eps,
                                      elementwise_affine = FALSE)
    self$attn <- flux2_parallel_self_attention(
        dim, num_attention_heads, attention_head_dim,
        mlp_ratio = mlp_ratio, eps = eps, bias = bias
    )
},
                                       forward = function(hidden_states, temb_mod, image_rotary_emb = NULL,
        chunk_size = NULL) {
    mod <- .flux2_mod_split(temb_mod, 1L)[[1]]
    norm_h <- self$norm(hidden_states) * mod[[2]]$add(1) + mod[[1]]
    attn_out <- self$attn(
                          hidden_states = norm_h,
                          image_rotary_emb = image_rotary_emb,
                          chunk_size = chunk_size
    )
    hidden_states <- hidden_states + mod[[3]] * attn_out
    if (hidden_states$dtype == torch::torch_float16()) {
        hidden_states <- hidden_states$clamp(-65504, 65504)
    }
    hidden_states
}
)

#' FLUX Rotary Positional Embeddings
#'
#' Fresh R port of the FLUX rotary positional embedding scheme from the
#' diffusers reference implementation (Apache-2.0,
#' src/diffusers/models/transformers/transformer_flux.py FluxPosEmbed and
#' src/diffusers/models/embeddings.py get_1d_rotary_pos_embed /
#' apply_rotary_emb). FLUX uses the interleaved adjacent-pair convention
#' (use_real_unbind_dim = -1) with per-axis frequencies computed in
#' float64 and applied in float32. Text tokens carry all-zero ids, so
#' they receive the identity rotation.
#'
#' @name rope_flux
NULL

#' Build FLUX latent image position ids
#'
#' Position ids over the packed latent grid (latent height/2 x width/2).
#' Channel 1 is always zero, channel 2 holds the row index, channel 3 the
#' column index. Reference: FluxPipeline._prepare_latent_image_ids.
#'
#' @param height Integer. Packed grid height (latent height / 2).
#' @param width Integer. Packed grid width (latent width / 2).
#' @param device Device for the resulting tensor.
#'
#' @return Float tensor of shape [height * width, 3].
#'
#' @export
flux_prepare_latent_image_ids <- function(height, width, device = "cpu") {
    f32 <- torch::torch_float32()
    ids <- torch::torch_zeros(height, width, 3L, dtype = f32, device = device)
    # torch_arange has an inclusive end; end = n - 1 matches Python arange
    rows <- torch::torch_arange(start = 0, end = height - 1, dtype = f32,
                                device = device)
    cols <- torch::torch_arange(start = 0, end = width - 1, dtype = f32,
                                device = device)
    ids[,, 2] <- ids[,, 2] + rows$unsqueeze(2L)
    ids[,, 3] <- ids[,, 3] + cols$unsqueeze(1L)
    ids$reshape(c(height * width, 3L))
}

#' Compute FLUX rotary frequencies from position ids
#'
#' Per-axis 1D rotary frequencies (interleaved-real convention), computed
#' in float64 on CPU and concatenated over the axes. Reference:
#' FluxPosEmbed with get_1d_rotary_pos_embed(repeat_interleave_real=TRUE,
#' use_real=TRUE, freqs_dtype=float64).
#'
#' @param ids Tensor of shape [S, 3]: concatenated text ids (all zero)
#'   and image ids from \code{flux_prepare_latent_image_ids}.
#' @param axes_dim Integer vector of per-axis rotary dims; must sum to
#'   the attention head dim. FLUX uses c(16, 56, 56).
#' @param theta Numeric. RoPE base frequency.
#'
#' @return List of two tensors (cos, sin), each [S, sum(axes_dim)],
#'   float32, on the device of \code{ids}.
#'
#' @export
flux_pos_embed <- function(ids, axes_dim = c(16L, 56L, 56L), theta = 10000) {
    n_axes <- ids$shape[2]
    device <- ids$device
    f64 <- torch::torch_float64()
    # Frequencies in float64 on CPU: Blackwell fp64 throughput is 1/64,
    # and the tensors are tiny
    pos <- ids$to(dtype = f64)$cpu()

    cos_out <- vector("list", n_axes)
    sin_out <- vector("list", n_axes)
    for (i in seq_len(n_axes)) {
        d <- axes_dim[i]
        # freqs = 1 / theta^(seq(0, d - 2, by = 2) / d), length d / 2
        exponents <- torch::torch_arange(start = 0, end = d - 2, step = 2,
            dtype = f64)
        freqs <- 1.0 / torch::torch_pow(theta, exponents / d)
        # Outer product [S, d/2]
        freqs <- pos[, i]$unsqueeze(2L) * freqs$unsqueeze(1L)
        cos_out[[i]] <- freqs$cos()$repeat_interleave(2L, dim = 2L)
        sin_out[[i]] <- freqs$sin()$repeat_interleave(2L, dim = 2L)
    }

    f32 <- torch::torch_float32()
    list(
         torch::torch_cat(cos_out, dim = -1L)$to(dtype = f32, device = device),
         torch::torch_cat(sin_out, dim = -1L)$to(dtype = f32, device = device)
    )
}

#' Apply FLUX rotary embeddings to a per-head tensor
#'
#' Rotates adjacent element pairs of the last dimension:
#' \code{out = x * cos + rotate_half(x) * sin} with pairs interleaved
#' (elements 1,2 form the first complex pair). Math in float32, result
#' cast back to the input dtype. Reference: apply_rotary_emb with
#' use_real_unbind_dim = -1.
#'
#' @param x Tensor of shape [B, H, S, D] (per-head layout).
#' @param freqs List of two tensors (cos, sin), each [S, D], from
#'   \code{flux_pos_embed}.
#'
#' @return Tensor with the same shape and dtype as \code{x}.
#'
#' @export
flux_apply_rotary_emb <- function(x, freqs) {
    cos <- freqs[[1]]$unsqueeze(1L)$unsqueeze(1L) # [1, 1, S, D]
    sin <- freqs[[2]]$unsqueeze(1L)$unsqueeze(1L)

    pairs <- x$unflatten(4L, c(-1L, 2L)) # [B, H, S, D/2, 2]
    x_real <- pairs[,,,, 1]
    x_imag <- pairs[,,,, 2]
    x_rotated <- torch::torch_stack(list(-x_imag, x_real), dim = -1L)$flatten(start_dim = 4L)

    out <- x$to(dtype = torch::torch_float32()) * cos +
    x_rotated$to(dtype = torch::torch_float32()) * sin
    out$to(dtype = x$dtype)
}

# Parity test for the full FLUX transformer against a tiny random-init
# diffusers FluxTransformer2DModel (fixture generated by
# tools/gen_fixtures_flux_model.py, checked in).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "flux_model.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/flux_model.safetensors"
if (!file.exists(fixture_path)) exit_file("flux model fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

model <- flux_transformer(
  in_channels = 4L,
  num_layers = 1L,
  num_single_layers = 1L,
  attention_head_dim = 8L,
  num_attention_heads = 2L,
  joint_attention_dim = 16L,
  pooled_projection_dim = 12L,
  axes_dims_rope = c(2L, 2L, 4L)
)
model$eval()

# Strict both ways: every fixture key must land, every param must fill.
# This proves the R module tree matches the diffusers state dict exactly.
weights <- fx[grep("^model\\.", names(fx))]
names(weights) <- sub("^model\\.", "", names(weights))
dests <- c(model$named_parameters(), model$named_buffers())
expect_equal(character(0), setdiff(names(weights), names(dests)))
expect_equal(character(0), setdiff(names(dests), names(weights)))
torch::with_no_grad({
  for (name in names(weights)) dests[[name]]$copy_(weights[[name]])
})

ids <- torch::torch_cat(list(fx$txt_ids, fx$img_ids), dim = 1L)
rope <- flux_pos_embed(ids, axes_dim = c(2L, 2L, 4L))

out <- torch::with_no_grad(model(
  hidden_states = fx$hidden,
  encoder_hidden_states = fx$encoder,
  pooled_projections = fx$pooled,
  timestep = fx$timestep,
  image_rotary_emb = rope
))
expect_equal(as.integer(out$shape), as.integer(fx$out$shape))
expect_true(max_abs_diff(out, fx$out) < 1e-4)

# Parity tests for the FLUX RoPE port and latent pack/unpack against
# diffusers reference fixtures (generated by tools/gen_fixtures_flux.py,
# checked in).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "rope_flux.safetensors",
  package = "diffuseR")
if (fixture_path == "") {
  # Running from the source tree (e.g. during development)
  fixture_path <- "fixtures/rope_flux.safetensors"
}
if (!file.exists(fixture_path)) exit_file("flux rope fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# --- latent image ids ---------------------------------------------------------

ids_r <- flux_prepare_latent_image_ids(8L, 12L)
expect_equal(as.integer(ids_r$shape), as.integer(fx$img_ids$shape))
expect_true(max_abs_diff(ids_r, fx$img_ids) == 0)

# --- flux_pos_embed: full-size axes (16, 56, 56) -------------------------------

pf <- flux_pos_embed(fx$ids, axes_dim = c(16L, 56L, 56L))
expect_equal(as.integer(pf[[1]]$shape), as.integer(fx$pos_full_cos$shape))
expect_true(max_abs_diff(pf[[1]], fx$pos_full_cos) < 1e-6)
expect_true(max_abs_diff(pf[[2]], fx$pos_full_sin) < 1e-6)

# Text ids (all zero) get the identity rotation: cos 1, sin 0
txt_cos <- pf[[1]][1:7, ]
txt_sin <- pf[[2]][1:7, ]
expect_true(max_abs_diff(txt_cos, torch::torch_ones_like(txt_cos)) == 0)
expect_true(max_abs_diff(txt_sin, torch::torch_zeros_like(txt_sin)) == 0)

# --- flux_pos_embed: tiny axes (2, 2, 4) ---------------------------------------

pt <- flux_pos_embed(fx$ids, axes_dim = c(2L, 2L, 4L))
expect_equal(as.integer(pt[[1]]$shape), as.integer(fx$pos_tiny_cos$shape))
expect_true(max_abs_diff(pt[[1]], fx$pos_tiny_cos) < 1e-6)
expect_true(max_abs_diff(pt[[2]], fx$pos_tiny_sin) < 1e-6)

# --- flux_apply_rotary_emb ------------------------------------------------------

rot <- flux_apply_rotary_emb(fx$rot_x, list(fx$pos_tiny_cos, fx$pos_tiny_sin))
expect_equal(as.integer(rot$shape), as.integer(fx$rot_out$shape))
expect_true(max_abs_diff(rot, fx$rot_out) < 1e-5)

# fp16 input stays fp16 and matches reference values
rot16 <- flux_apply_rotary_emb(fx$rot_x_f16, list(fx$pos_tiny_cos, fx$pos_tiny_sin))
expect_equal(rot16$dtype$.type(), "Half")
expect_true(max_abs_diff(rot16, fx$rot_out_f16) < 1e-3)

# --- latent pack / unpack -------------------------------------------------------

packed <- flux_pack_latents(fx$pack_in)
expect_equal(as.integer(packed$shape), as.integer(fx$pack_out$shape))
expect_true(max_abs_diff(packed, fx$pack_out) == 0)

unpacked <- flux_unpack_latents(fx$pack_out, height = 64L, width = 96L,
  vae_scale_factor = 8L)
expect_equal(as.integer(unpacked$shape), as.integer(fx$unpack_out$shape))
expect_true(max_abs_diff(unpacked, fx$unpack_out) == 0)

# R-side roundtrip identity: unpack(pack(x)) == x
expect_true(max_abs_diff(unpacked, fx$pack_in) == 0)

#' Audio Conditioning Frontend for LTX-2.3
#'
#' Turns user audio into the normalized, packed audio latents the joint
#' denoiser conditions on (lip sync): decode to 16 kHz stereo PCM,
#' log-mel via a causal STFT (filter 1024, hop 160, 64 slaney-normed
#' mel bins to 8 kHz — the checkpoint's preprocessing spec), then the
#' audio VAE encoder in argmax mode. The STFT and mel-filterbank
#' constructors were verified against the checkpoint's stored vocoder
#' bases (identical up to bf16 rounding), so the convention matches
#' training.
#'
#' @name audio_encode_ltx23
NULL

# Periodic Hann window
.ltx23_hann <- function(n) {
    0.5 - 0.5 * cos(2 * pi * (0:(n - 1L)) / n)
}

# Windowed Fourier basis in the layout ltx23_causal_stft expects:
# [n_freqs * 2, 1, filter_length], cos rows then -sin rows, each times
# the periodic Hann window (matches vocoder.mel_stft.stft_fn.forward_basis)
.ltx23_stft_basis <- function(filter_length) {
    nf <- filter_length %/% 2L + 1L
    n <- 0:(filter_length - 1L)
    win <- .ltx23_hann(filter_length)
    basis <- matrix(0, nf * 2L, filter_length)
    for (k in 0:(nf - 1L)) {
        basis[k + 1L,] <- cos(2 * pi * k * n / filter_length) * win
        basis[nf + k + 1L,] <- -sin(2 * pi * k * n / filter_length) * win
    }
    torch::torch_tensor(basis, dtype = torch::torch_float32())$unsqueeze(2L)
}

# Slaney-scale, slaney-normed mel filterbank (matches
# vocoder.mel_stft.mel_basis up to bf16 rounding)
.ltx23_mel_filterbank <- function(sample_rate, n_fft, n_mels, fmin, fmax) {
    hz2mel <- function(f) {
        ifelse(f < 1000, f * 3 / 200, 15 + log(f / 1000) / (log(6.4) / 27))
    }
    mel2hz <- function(m) {
        ifelse(m < 15, m * 200 / 3, 1000 * exp((m - 15) * log(6.4) / 27))
    }
    nf <- n_fft %/% 2L + 1L
    fft_freqs <- (0:(nf - 1L)) * sample_rate / n_fft
    hz <- mel2hz(seq(hz2mel(fmin), hz2mel(fmax), length.out = n_mels + 2L))
    W <- matrix(0, n_mels, nf)
    for (i in seq_len(n_mels)) {
        lower <- (fft_freqs - hz[i]) / (hz[i + 1] - hz[i])
        upper <- (hz[i + 2] - fft_freqs) / (hz[i + 2] - hz[i + 1])
        W[i,] <- pmax(0, pmin(lower, upper)) * 2 / (hz[i + 2] - hz[i])
    }
    torch::torch_tensor(W, dtype = torch::torch_float32())
}

#' Build the 16 kHz log-mel frontend for audio conditioning
#'
#' An \code{\link{ltx23_mel_stft}} whose STFT and mel bases are
#' constructed (not checkpoint-loaded) with the audio VAE's
#' preprocessing spec.
#'
#' @param filter_length,hop_length,n_mels,sample_rate,fmin,fmax The
#'   checkpoint preprocessing parameters (defaults are LTX-2.3's).
#'
#' @return An \code{ltx23_mel_stft} module.
#'
#' @export
ltx23_audio_mel_frontend <- function(filter_length = 1024L,
                                     hop_length = 160L, n_mels = 64L,
                                     sample_rate = 16000L, fmin = 0,
                                     fmax = 8000) {
    frontend <- ltx23_mel_stft(filter_length = filter_length,
                               hop_length = hop_length,
                               window_length = filter_length,
                               num_mel_channels = n_mels)
    torch::with_no_grad({
        frontend$stft_fn$forward_basis$copy_(.ltx23_stft_basis(filter_length))
        frontend$mel_basis$copy_(.ltx23_mel_filterbank(sample_rate,
                filter_length, n_mels, fmin, fmax))
    })
    frontend$eval()
    frontend
}

#' Read an audio file as 16 kHz stereo PCM
#'
#' Decodes MP3/WAV/etc. via \code{av} to 16-bit PCM at the target rate
#' and parses the RIFF container in base R.
#'
#' @param path Audio file.
#' @param sample_rate Integer.
#'
#' @return Matrix [2, samples] in [-1, 1].
#'
#' @export
ltx23_read_audio <- function(path, sample_rate = 16000L) {
    if (!requireNamespace("av", quietly = TRUE)) {
        stop("Reading audio requires the 'av' package.")
    }
    wav <- tempfile(fileext = ".wav")
    on.exit(unlink(wav), add = TRUE)
    av::av_audio_convert(path, wav, format = NULL, channels = 2L,
                         sample_rate = sample_rate, verbose = FALSE)

    con <- file(wav, "rb")
    on.exit(close(con), add = TRUE)
    riff <- readBin(con, "raw", 12L)
    stopifnot(rawToChar(riff[1:4]) == "RIFF")
    n_channels <- 2L
    repeat {
        hdr <- readBin(con, "raw", 8L)
        if (length(hdr) < 8L) {
            stop("No data chunk found in ", wav)
        }
        id <- rawToChar(hdr[1:4])
        size <- readBin(hdr[5:8], "integer", 1L, size = 4L, endian = "little")
        if (id == "fmt ") {
            fmt <- readBin(con, "raw", size)
            n_channels <- readBin(fmt[3:4], "integer", 1L, size = 2L,
                                  endian = "little")
        } else if (id == "data") {
            pcm <- readBin(con, "integer", size %/% 2L, size = 2L,
                           signed = TRUE, endian = "little")
            break
        } else {
            invisible(readBin(con, "raw", size + size %% 2L))
        }
    }
    m <- matrix(pcm / 32768, nrow = n_channels)
    if (n_channels == 1L) {
        m <- rbind(m, m)
    }
    m
}

#' Encode audio into normalized, packed conditioning latents
#'
#' Pads or trims the waveform so the latent length equals
#' \code{audio_num_frames} (mel frames \code{4L - 3}, mirroring the
#' decoder's \code{target_frames}), computes the log-mel, encodes in
#' argmax mode, packs, and normalizes with the checkpoint statistics.
#'
#' @param audio_vae An \code{ltx23_audio_vae} (with encoder weights).
#' @param wav Matrix [2, samples] in [-1, 1] at 16 kHz (see
#'   \code{\link{ltx23_read_audio}}).
#' @param audio_num_frames Integer. Target latent length.
#' @param frontend Optional prebuilt \code{\link{ltx23_audio_mel_frontend}}.
#'
#' @return Packed normalized latents [1, audio_num_frames, 128] (float32).
#'
#' @export
ltx23_encode_audio <- function(audio_vae, wav, audio_num_frames,
                               frontend = NULL) {
    frontend <- frontend %||% ltx23_audio_mel_frontend()
    target_mel_frames <- 4L * audio_num_frames - 3L

    dev <- audio_vae$latents_mean$device
    enc_dtype <- audio_vae$encoder$conv_in$conv$weight$dtype
    frontend$to(device = dev)

    x <- torch::torch_tensor(wav, dtype = torch::torch_float32(), device = dev)
    torch::with_no_grad({
        # Stereo channels through the STFT as a batch of mono signals
        mel <- frontend(x$unsqueeze(2L)) # [2, 64, T]
        # -> [1, 2, T, 64]
        mel <- mel$permute(c(1L, 3L, 2L))$unsqueeze(1L)
        t_cur <- mel$shape[3]
        if (t_cur < target_mel_frames) {
            mel <- torch::nnf_pad(mel, c(0L, 0L, 0L, target_mel_frames - t_cur))
        } else if (t_cur > target_mel_frames) {
            mel <- mel$narrow(3L, 1L, target_mel_frames)
        }
        moments <- audio_vae$encode(mel$to(dtype = enc_dtype))
        latents <- moments$mean$to(dtype = torch::torch_float32())
    })
    packed <- ltx23_pack_audio_latents(latents)
    mean <- audio_vae$latents_mean$to(dtype = torch::torch_float32())
    std <- audio_vae$latents_std$to(dtype = torch::torch_float32())
    (packed - mean) / std
}

# Audio conditioning frontend (R/audio_encode_ltx23.R): STFT/mel basis
# constructors, mel geometry, and encode length math.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)
torch::torch_manual_seed(31)

# --- Basis constructors ---------------------------------------------------------------

fb <- diffuseR:::.ltx23_stft_basis(64L)
expect_equal(as.integer(fb$shape), c(66L, 1L, 64L))
# Row 0 is the DC row: pure window (cos(0) = 1)
w <- diffuseR:::.ltx23_hann(64L)
expect_true(as.numeric((fb[1, 1, ] - torch::torch_tensor(w))$abs()$max()) < 1e-6)
# The imaginary DC row is all zeros (-sin(0))
expect_true(as.numeric(fb[34, 1, ]$abs()$max()) < 1e-6)

mel <- diffuseR:::.ltx23_mel_filterbank(16000, 512L, 64L, 0, 8000)
expect_equal(as.integer(mel$shape), c(64L, 257L))
# Every filter is nonnegative with positive mass
expect_true(as.numeric(mel$min()) >= 0)
expect_true(all(as.numeric(mel$sum(dim = 2L)) > 0))

# --- Frontend geometry ---------------------------------------------------------------

frontend <- ltx23_audio_mel_frontend(filter_length = 64L, hop_length = 16L,
  n_mels = 8L, sample_rate = 1600L, fmin = 0, fmax = 800)
x <- torch::torch_randn(2L, 1L, 320L)
torch::with_no_grad(m <- frontend(x))
expect_equal(as.integer(m$shape)[1:2], c(2L, 8L))
expect_true(all(is.finite(as.numeric(m$min()))))
# Log-clamped: never below log(1e-5)
expect_true(as.numeric(m$min()) >= log(1e-5) - 1e-4)

# --- Encode length math ---------------------------------------------------------------

# base_channels 128 so the stats buffer matches the packed feature dim
# (latent_channels * latent_mel_bins = 128), as in the real checkpoint
avae <- ltx23_audio_vae(
  base_channels = 128L, output_channels = 2L, num_res_blocks = 1L,
  latent_channels = 8L, ch_mult = c(1L, 1L, 1L), mel_bins = 64L
)
avae$eval()
small_frontend <- ltx23_audio_mel_frontend()
wav <- matrix(runif(2L * 8000L, -0.5, 0.5), nrow = 2L)
lat <- ltx23_encode_audio(avae, wav, audio_num_frames = 9L,
  frontend = small_frontend)
# Packed [1, L, latent_channels * latent_mel_bins]; L == audio_num_frames
expect_equal(as.integer(lat$shape), c(1L, 9L, 8L * 16L))

# Longer target than the audio covers: right-padded, still exact length
lat2 <- ltx23_encode_audio(avae, wav, audio_num_frames = 25L,
  frontend = small_frontend)
expect_equal(as.integer(lat2$shape)[2], 25L)

#' FLUX Transformer (MMDiT)
#'
#' Fresh R port of FluxTransformer2DModel from the diffusers reference
#' implementation (Apache-2.0,
#' src/diffusers/models/transformers/transformer_flux.py). The module
#' tree mirrors the diffusers state-dict keys 1:1, so checkpoints load
#' without remapping. FLUX.1-schnell has no guidance embedder
#' (guidance_embeds = FALSE); the guidance-distilled dev variant is not
#' implemented.
#'
#' @name dit_flux
NULL

# Combined timestep + pooled-text conditioning, matching diffusers
# CombinedTimestepTextProjEmbeddings state-dict names. Both embedders are
# linear_1 -> silu -> linear_2, which ltx23_timestep_embedding provides
# (PixArtAlphaTextProjection with act_fn = "silu" is the same function).
flux_time_text_embed <- torch::nn_module(
    "flux_time_text_embed",
    initialize = function(embedding_dim, pooled_projection_dim) {
    self$timestep_embedder <- ltx23_timestep_embedding(256L, embedding_dim)
    self$text_embedder <- ltx23_timestep_embedding(pooled_projection_dim,
        embedding_dim)
},
    forward = function(timestep, pooled_projection) {
    proj <- ltx23_get_timestep_embedding(timestep, 256L,
        flip_sin_to_cos = TRUE, downscale_freq_shift = 0)
    temb <- self$timestep_embedder(proj$to(dtype = pooled_projection$dtype))
    temb + self$text_embedder(pooled_projection)
}
)

#' FLUX transformer model
#'
#' 19 double-stream (MMDiT) blocks followed by 38 single-stream blocks
#' over the joint [text; image] sequence, with adaLN-Zero conditioning on
#' timestep + pooled CLIP text. Rotary embeddings are precomputed by the
#' caller with \code{flux_pos_embed} (they are static across denoise
#' steps). Defaults are the FLUX.1-schnell configuration.
#'
#' @param in_channels Integer. Packed latent channels (64).
#' @param num_layers Integer. Double-stream block count.
#' @param num_single_layers Integer. Single-stream block count.
#' @param attention_head_dim Integer. Per-head dimension.
#' @param num_attention_heads Integer. Attention heads.
#' @param joint_attention_dim Integer. T5 embedding dim (4096).
#' @param pooled_projection_dim Integer. CLIP pooled dim (768).
#' @param axes_dims_rope Integer vector. Per-axis rotary dims.
#' @param out_channels Integer or NULL. Output channels (defaults to
#'   \code{in_channels}).
#'
#' @return Module whose forward(hidden_states, encoder_hidden_states,
#'   pooled_projections, timestep, image_rotary_emb) returns the
#'   predicted velocity for the image tokens [B, S_img, out_channels].
#'   \code{timestep} is in sigma space (0-1); it is scaled by 1000
#'   internally, matching the reference.
#'
#' @export
flux_transformer <- torch::nn_module(
                                     "flux_transformer",
                                     initialize = function(in_channels = 64L,
        num_layers = 19L,
        num_single_layers = 38L,
        attention_head_dim = 128L,
        num_attention_heads = 24L,
        joint_attention_dim = 4096L,
        pooled_projection_dim = 768L,
        axes_dims_rope = c(16L, 56L, 56L),
        out_channels = NULL) {
    inner_dim <- num_attention_heads * attention_head_dim
    self$inner_dim <- inner_dim
    self$axes_dims_rope <- as.integer(axes_dims_rope)
    self$out_channels <- as.integer(out_channels %||% in_channels)

    self$time_text_embed <- flux_time_text_embed(inner_dim,
        pooled_projection_dim)
    self$context_embedder <- torch::nn_linear(joint_attention_dim, inner_dim)
    self$x_embedder <- torch::nn_linear(in_channels, inner_dim)

    self$transformer_blocks <- torch::nn_module_list(
        lapply(seq_len(num_layers), function(i) {
        flux_double_block(inner_dim, num_attention_heads, attention_head_dim)
    })
    )
    self$single_transformer_blocks <- torch::nn_module_list(
        lapply(seq_len(num_single_layers), function(i) {
        flux_single_block(inner_dim, num_attention_heads, attention_head_dim)
    })
    )

    self$norm_out <- flux_ada_layer_norm_continuous(inner_dim, inner_dim)
    self$proj_out <- torch::nn_linear(inner_dim, self$out_channels, bias = TRUE)
},
                                     forward = function(hidden_states, encoder_hidden_states,
        pooled_projections, timestep, image_rotary_emb,
        chunk_size = NULL) {
    hidden_states <- self$x_embedder(hidden_states)
    timestep <- timestep$to(dtype = hidden_states$dtype)$mul(1000)
    temb <- self$time_text_embed(timestep, pooled_projections)
    encoder_hidden_states <- self$context_embedder(encoder_hidden_states)

    block_gc <- isTRUE(getOption("diffuseR.block_gc"))
    debug <- isTRUE(getOption("diffuseR.debug"))

    for (i in seq_along(self$transformer_blocks)) {
        res <- self$transformer_blocks[[i]](
                                     hidden_states = hidden_states,
                                     encoder_hidden_states = encoder_hidden_states,
                                     temb = temb,
                                     image_rotary_emb = image_rotary_emb,
                                     chunk_size = chunk_size
        )
        encoder_hidden_states <- res[[1]]
        hidden_states <- res[[2]]
        if (debug && torch::cuda_is_available()) {
            ms <- torch::cuda_memory_stats()
            message(sprintf("    double block %d: %.2f GB allocated", i,
                            ms$allocated_bytes$all$current / 1e9))
        }
        if (block_gc) {
            gc(verbose = FALSE)
        }
    }

    # The reference concatenates [text; image] inside every single block
    # and splits after; concatenating once here is numerically identical
    txt_len <- encoder_hidden_states$shape[2]
    hidden_states <- torch::torch_cat(
                                      list(encoder_hidden_states, hidden_states),
                                      dim = 2L
    )
    for (i in seq_along(self$single_transformer_blocks)) {
        hidden_states <- self$single_transformer_blocks[[i]](
                                     hidden_states = hidden_states,
                                     temb = temb,
                                     image_rotary_emb = image_rotary_emb,
                                     chunk_size = chunk_size
        )
        if (block_gc) {
            gc(verbose = FALSE)
        }
    }
    hidden_states <- hidden_states$narrow(
        2L, txt_len + 1L,
        hidden_states$shape[2] - txt_len
    )

    hidden_states <- self$norm_out(hidden_states, temb)
    self$proj_out(hidden_states)
}
)

#' safetensors read-capability probes and fork messaging
#'
#' The CRAN build of safetensors 0.2.1 reads bfloat16 but cannot write
#' it, and has no float8 support at all. Both fixes merged upstream on
#' 2026-07-31 (mlverse/safetensors#11 for bfloat16 write, #13 for
#' float8) without a version bump, so the installed version number
#' cannot tell you which build you have. That is why every gate here is
#' a runtime probe: write a tiny tensor, read it back, cache the answer.
#' Two capabilities matter and they differ:
#'
#' \itemize{
#'   \item \emph{write} (\code{\link{flux_quantize}}'s internal
#'     \code{.st_can_write}, in quantize_flux.R): needed to BUILD a
#'     quantized artifact in that dtype.
#'   \item \emph{read} (\code{.st_can_read}, here): needed to LOAD a
#'     hosted artifact in that dtype. This is the capability that gates
#'     user-facing recommendations. It is strictly weaker than write:
#'     CRAN safetensors reads bfloat16 it cannot write, so the write
#'     probe is the wrong signal for whether a hosted bf16 artifact will
#'     load.
#' }
#'
#' Both are capability-probed, never version-pinned, so the fork
#' requirement self-heals the day the fixes reach CRAN.
#'
#' @name st_caps
NULL

# Read-probe cache, keyed by dtype. Separate from quantize_flux.R's
# `.st_caps` write cache: the same dtype can be readable but not
# writable (bfloat16 on CRAN), so the two must not share entries.
.st_read_caps <- new.env(parent = emptyenv())

# Write a minimal 2-element safetensors file by hand: a u64
# little-endian header length, the JSON header, then the raw tensor
# bytes. Deliberately does NOT go through safetensors::safe_save_file -
# that is the whole point, since a CRAN safetensors cannot WRITE
# bfloat16 yet can READ it. Lets `.st_can_read` test read capability in
# isolation from write capability.
.st_write_min <- function(path, dtype_name, payload) {
    hdr <- sprintf('{"w":{"dtype":"%s","shape":[2],"data_offsets":[0,%d]}}',
                   dtype_name, length(payload))
    hb <- charToRaw(hdr)
    n <- length(hb)
    # u64 header length, little-endian (headers here are < 256 bytes)
    len_bytes <- as.raw(c(n %% 256L, (n %/% 256L) %% 256L, 0L, 0L, 0L, 0L,
                          0L, 0L))
    con <- file(path, "wb")
    on.exit(close(con), add = TRUE)
    writeBin(len_bytes, con)
    writeBin(hb, con)
    writeBin(payload, con)
    invisible(path)
}

# The [0, 1] byte patterns for the two dtypes the recommender gates on.
# bfloat16 is the top 16 bits of float32: 1.0f = 0x3F800000 -> 0x3F80
# (little-endian 0x80 0x3F); 0.0 -> 0x0000. float8_e4m3fn: 1.0 = exp
# bias 7, mantissa 0 = 0x38; 0.0 = 0x00.
.st_read_probe_spec <- list(
                            bfloat16 = list(name = "BF16", bytes = as.raw(c(0x00, 0x00, 0x80, 0x3f))),
                            float8_e4m3fn = list(name = "F8_E4M3", bytes = as.raw(c(0x00, 0x38)))
)

#' Probe whether the installed safetensors can READ a dtype
#'
#' Hand-builds a tiny safetensors file of the dtype (via
#' \code{.st_write_min}, no safetensors writer involved) and tries to
#' load it back. Cached per session;
#' \code{options(diffuseR.st_read_caps = list(bfloat16 = TRUE, ...))}
#' overrides the probe for tests and for forcing a tier.
#'
#' @param dtype "bfloat16" or "float8_e4m3fn".
#' @return Logical.
#' @keywords internal
.st_can_read <- function(dtype = c("bfloat16", "float8_e4m3fn")) {
    dtype <- match.arg(dtype)
    override <- getOption("diffuseR.st_read_caps")
    if (!is.null(override) && !is.null(override[[dtype]])) {
        return(isTRUE(override[[dtype]]))
    }
    cached <- .st_read_caps[[dtype]]
    if (!is.null(cached)) {
        return(cached)
    }
    ok <- requireNamespace("safetensors", quietly = TRUE) && tryCatch({
        spec <- .st_read_probe_spec[[dtype]]
        tmp <- tempfile(fileext = ".safetensors")
        on.exit(unlink(tmp), add = TRUE)
        .st_write_min(tmp, spec$name, spec$bytes)
        y <- safetensors::safe_load_file(tmp, framework = "torch")
        !is.null(y$w) && identical(as.integer(y$w$shape), 2L)
    }, error = function(e) FALSE)
    .st_read_caps[[dtype]] <- ok
    ok
}

# The standard "install the fork, or press on with nf4" message. Shared
# by the recommender (read side, fit = TRUE: "best fit for your card")
# and the download graceful-fallback path (write side, fit = FALSE, since
# the user asked for it outright) so the wording stays identical
# everywhere. No em dashes (house style).
.st_fork_note <- function(precision, fit = TRUE) {
    precision <- as.character(precision)
    detail <- switch(precision,
                     fp8 = "float8 support (mlverse/safetensors#13)",
                     float8_e4m3fn = "float8 support (mlverse/safetensors#13)",
                     bf16 = "bfloat16 write support (mlverse/safetensors#11)",
                     bfloat16 = "bfloat16 write support (mlverse/safetensors#11)",
                     paste0(precision, " support (mlverse/safetensors)"))
    lead <- if (fit) {
        sprintf("%s is the best fit for your card but needs", precision)
    } else {
        sprintf("%s needs", precision)
    }
    sprintf(paste0(
                   "%s a safetensors newer than the one on CRAN: %s is ",
                   "merged upstream but not yet released. Install the ",
                   "development version from the mlverse/safetensors ",
                   "repository on GitHub, or press on with nf4: same ",
                   "weights, slightly lower precision, and it just works."),
            lead, detail)
}

# When a user explicitly asks for fp8/bf16 but the needed safetensors
# capability is missing, print the fork suggestion and fall back to nf4
# instead of letting a downstream builder or loader fail. nf4, fp16,
# fp32 and anything unrecognized pass through untouched. `mode` selects
# the capability that matters: "write" when about to BUILD an artifact,
# "read" when about to LOAD one.
.st_graceful_precision <- function(precision, mode = c("write", "read"),
                                   verbose = TRUE) {
    mode <- match.arg(mode)
    cap <- switch(precision, fp8 = "float8_e4m3fn", bf16 = "bfloat16",
                  bfloat16 = "bfloat16", NULL)
    if (is.null(cap)) {
        return(precision)
    }
    if (mode == "write") {
        ok <- .st_can_write(cap)
    } else {
        ok <- .st_can_read(cap)
    }
    if (ok) {
        return(precision)
    }
    if (verbose) {
        message(.st_fork_note(precision, fit = FALSE),
                "\nFalling back to nf4 for now.")
    }
    "nf4"
}

# Actionable message for the >2 GB shard-read overflow. Split out from
# .st_read_or_breadcrumb so it can be unit-tested without a real 2 GB
# file.
.st_overflow_message <- function(file_path, size_bytes, underlying) {
    sprintf(paste0("Could not read %s (%.1f GB). Stock CRAN safetensors ",
                   "overflows a 32-bit offset on files at or above 2^31 ",
                   "bytes (~2.15 GB). Rebuild the artifact with smaller ",
                   "shards (the quantizers now default to ",
                   "shard_bytes = 1.9e9), or install the development ",
                   "version of safetensors from the mlverse/safetensors ",
                   "repository on GitHub, where the fix is merged. ",
                   "Underlying error: %s"),
            basename(file_path), size_bytes / 1e9, underlying)
}

# Run a safetensors read; if it fails AND the backing shard is at/above
# the 2^31-byte ceiling, translate the cryptic overflow into the
# fork-or-smaller-shards breadcrumb. A read that succeeds (fork, or a
# sub-2 GB shard) is untouched; a failure on a small shard rethrows
# verbatim. Reactive by design, so it never false-alarms on a machine
# that can read large files.
.st_read_or_breadcrumb <- function(read_fn, file_path = NULL) {
    tryCatch(read_fn(), error = function(e) {
        sz <- if (!is.null(file_path)) {
            tryCatch(file.size(file_path), error = function(...) NA_real_)
        } else {
            NA_real_
        }
        if (!is.na(sz) && sz >= 2 ^ 31) {
            stop(.st_overflow_message(file_path, sz, conditionMessage(e)),
                 call. = FALSE)
        }
        stop(e)
    })
}

#' FLUX Memory Profiles
#'
#' VRAM-based execution profiles for the FLUX.1-schnell pipeline. The
#' 12B transformer runs NF4 (~7 GB) or fp8 (~12 GB), phase-onloaded to
#' the GPU for denoise; the T5-XXL text encoder phase-onloads to the GPU
#' (bfloat16, pinned) on 14 GB+ cards and computes on the CPU (float32)
#' below that, where its ~9.8 GB encode phase does not fit.
#'
#' @name memory_flux
NULL

# Allocator gc gates for the resident-weight image pipelines
# (backend:native). The allocated/reserved ratio gate is chronically
# over its 0.95 threshold under a lazily-reserving caching allocator
# (measured 0.957 at steady state), so the R gc callback fired on
# nearly every allocation - ~2,200 gcs x 62 ms = 89% of wall time on a
# Z-Image generation. Disabling the ratio gate alone OOMs (dead tensor
# handles accumulate until finalizers run), so the absolute
# allocated_rate gate takes over garbage-accumulation duty at 0.65 of
# total VRAM, above the ~8.3 GB live phase peak with headroom under the
# card limit. Measured (RTX 5060 Ti 16 GB): Z-Image 512^2 142 -> 12 s,
# 1024^2 143 -> 25 s; klein 1024^2 48 -> 13 s, no OOM, reserved peak
# 12.9 GB. ltx23_tune_gc's reserved_rate stays as the fragmentation
# net; the LTX video pipelines keep their own separately measured
# gates.
.flux_gc_gates <- function(footprint_gb = 12) {
    if (is.null(getOption("torch.cuda_allocator_allocated_reserved_rate"))) {
        options(torch.cuda_allocator_allocated_reserved_rate = 1.0)
    }
    if (is.null(getOption("torch.cuda_allocator_allocated_rate"))) {
        options(torch.cuda_allocator_allocated_rate = 0.65)
    }
    ltx23_tune_gc(footprint_gb = footprint_gb)
    # start_torch() reads the three gate options ONCE at torch init and
    # pushes them into the C++ allocator; option changes after that are
    # inert. Torch is long started by the time a loader runs, so push
    # the current values into the live allocator directly.
    push <- get0("cpp_set_cuda_allocator_allocator_thresholds",
                 envir = asNamespace("torch"))
    if (is.function(push)) {
        try(push(
                 getOption("torch.cuda_allocator_reserved_rate", 0.2),
                 getOption("torch.cuda_allocator_allocated_rate", 0.8),
                 getOption("torch.cuda_allocator_allocated_reserved_rate", 0.8)
            ), silent = TRUE)
    }
    invisible(NULL)
}

#' Resolve a FLUX memory profile
#'
#' A thin adapter over \code{\link{recommend}} for the FLUX.1 pipeline,
#' kept for back-compatibility. \code{recommend("flux1")} is the policy;
#' this reshapes it into the legacy profile fields the loader consumes.
#' Precision now rises with VRAM (nf4 default, fp8 GPU-resident on 14 GB+
#' cards when safetensors can read float8, bf16 on 24 GB+); the old
#' bands, which put fp8 in a narrow low-VRAM slot it can no longer fit,
#' were backwards.
#'
#' @param vram_gb Numeric or NULL. Available VRAM; auto-detected when
#'   NULL (via nvidia-smi).
#'
#' @return List with \code{name}, \code{precision} ("nf4"/"fp8"/"bf16"),
#'   \code{attn_chunk}, \code{text_device}, \code{phase_offload},
#'   \code{max_pixels}, and (advisory) \code{fork_suggested} and
#'   \code{note}.
#'
#' @export
flux_memory_profile <- function(vram_gb = NULL) {
    r <- recommend("flux1", vram_gb = vram_gb)
    name <- if (identical(r$devices$transformer, "cpu")) {
        "cpu_only"
    } else if (r$precision %in% c("bf16", "fp8")) {
        "high"
    } else if (r$max_pixels >= 1024L * 1024L) {
        "medium"
    } else {
        "low"
    }
    list(name = name, precision = r$precision, attn_chunk = r$attn_chunk,
         text_device = r$text_device, phase_offload = r$offload,
         max_pixels = r$max_pixels, fork_suggested = r$fork_suggested,
         note = r$note)
}

# In-process model residency: pinned host weights, disposable GPU copies.
#
# A resident pipeline keeps its canonical weights as page-locked (pinned)
# CPU tensors for the life of the handle. Activation creates the GPU
# representation with a DMA copy from pinned memory; deactivation
# destroys only the GPU representation and re-points the modules at the
# pinned host storage. Reactivation never touches the disk, so handing
# a small GPU between models is a sub-second operation instead of a full
# pipeline reload.
#
# Same contract as whisper's and chatterbox's R/resident.R (the three
# packages present one interface to a residency broker), adapted to
# diffuseR's shape: a pipeline is a classed list holding SEVERAL
# nn_modules (transformer, decoder, text encoder(s), and for LTX a video
# VAE, audio VAE and vocoder), so components are discovered by scanning
# the pipeline for nn_module fields and the staging set is keyed by
# field name. Non-tensor fields (config, tokenizer, scheduler) ride
# along untouched.
#
# This layer sits ABOVE the per-generation phase offloading already in
# txt2img_flux()/txt2img_flux2()/txt2img_zimage()/txt2vid_ltx23(). Those
# swap one component at a time WITHIN a render; residency is about who
# owns the GPU BETWEEN renders. The two compose: an active handle renders
# with its normal internal phase behaviour, and deactivation releases
# whatever is still resident so a sibling model can take the card.
#
# Mechanics rest on two torch behaviours (verified in whisper's port,
# re-verified by this package's tests):
# - nn_module$to() REBINDS parameter/buffer objects, so pinned host
#   tensors held in res$staging survive activation, and any tensor handle
#   taken before a transition is stale after it. All re-binding therefore
#   resolves the modules' CURRENT tensors by name, every time.
# - Tensor$set_data() works across devices: a CUDA parameter can be
#   re-pointed directly at a pinned CPU tensor. That is the evict
#   mechanism; the orphaned CUDA storage is reclaimed by gc() +
#   cuda_empty_cache().
#
# States: inactive -> activating -> active -> deactivating -> inactive.
# Failed transitions roll back to pinned host state; a rollback that
# cannot be verified leaves the handle "broken" (fail-closed: only status
# and unload work). "unloaded" is terminal.

# Families that ship a pinned/staged loader. Keyed by the `model` name
# used everywhere else in the package (see recommend()).
.resident_families <- c("flux1", "flux2", "zimage", "ltx")

#' Every nn_module field of a pipeline, by name
#'
#' Discovery beats a hard-coded list: the families disagree on which
#' components exist (FLUX.1 has two text encoders, LTX adds a video VAE,
#' an audio VAE and a vocoder), and a field added later is picked up
#' without touching this file.
#'
#' @param pipeline A loaded diffuseR pipeline.
#'
#' @return A named list of the pipeline's \code{nn_module} fields,
#'   possibly empty.
#'
#' @keywords internal
.resident_components <- function(pipeline) {
    keep <- vapply(pipeline, function(x) inherits(x, "nn_module"), logical(1))
    pipeline[keep]
}

#' Pin every component of a pipeline for fast transfer
#'
#' Re-uses any staging the loader already built (the phase-offload path
#' pins as part of loading), and pins the rest. Pinning a component that
#' is currently on the GPU also evicts it, since \code{.pin_component}
#' copies into page-locked host memory and re-points the live tensors at
#' it, so this doubles as the initial offload.
#'
#' @param pipeline A loaded diffuseR pipeline.
#' @param verbose Print progress.
#'
#' @return A named list of staging sets, one per component that could be
#'   pinned. Components that fail to page-lock are absent, and fall back
#'   to the pageable \code{$to()} path.
#'
#' @keywords internal
.resident_pin <- function(pipeline, verbose = TRUE) {
    existing <- pipeline$staging %||% list()
    comps <- .resident_components(pipeline)
    if (verbose && length(comps)) {
        message("Pinning ", length(comps), " components for residency...")
    }
    staging <- list()
    for (nm in names(comps)) {
        if (!is.null(existing[[nm]])) {
            staging[[nm]] <- existing[[nm]]
            next
        }
        extra <- if (identical(nm, "transformer") &&
            isTRUE(pipeline$fp8_resident)) {
            .flux_fp8_collect(comps[[nm]])
        } else {
            NULL
        }
        st <- .pin_component(comps[[nm]], extra = extra)
        if (!is.null(st)) {
            staging[[nm]] <- st
        }
    }
    staging
}

#' Total pinned host bytes across a staging set
#'
#' @param staging A named list of staging sets.
#'
#' @return Numeric. Bytes of page-locked host memory held.
#'
#' @keywords internal
.resident_pinned_bytes <- function(staging) {
    total <- 0
    for (st in staging) {
        for (pair in st) {
            total <- total + prod(as.numeric(pair$pinned$shape)) *
            .dtype_bytes(pair$pinned$dtype)
        }
    }
    total
}

# Bytes per element, keyed by the libtorch dtype name that
# as.character() on a torch_dtype returns ("Float", "Half", "Byte",
# "Long", ...), NOT the R constructor alias. Unknown dtypes fall back to
# 4, which only affects a reported number.
.dtype_widths <- c(double = 8, long = 8, complexfloat = 8,
                   float = 4, int = 4,
                   half = 2, bfloat16 = 2, short = 2,
                   byte = 1, char = 1, bool = 1,
                   float8_e4m3fn = 1, float8_e5m2 = 1)

.dtype_bytes <- function(dtype) {
    nm <- tolower(tryCatch(as.character(dtype), error = function(e) ""))
    w <- .dtype_widths[[nm, exact = TRUE]]
    if (is.null(w)) 4 else w
}

#' TRUE when every staged tensor sits on the expected device type
#'
#' @param staging A named list of staging sets.
#' @param type "cpu" or "cuda".
#'
#' @return Logical.
#'
#' @keywords internal
.resident_all_on <- function(staging, type) {
    for (st in staging) {
        for (pair in st) {
            dev <- tryCatch(pair$live$device$type, error = function(e) NA_character_)
            if (!identical(dev, type)) {
                return(FALSE)
            }
        }
    }
    TRUE
}

#' How many components actually have their tensors on the GPU
#'
#' Ground truth, as opposed to the handle's declared state. The two can
#' disagree: a pipeline built with \code{phase_offload = TRUE} swaps each
#' component back to pinned host memory as its phase finishes, so after a
#' render the handle is still "active" while the card holds nothing. A
#' broker deciding who to evict needs the measurement, not the claim.
#'
#' @param staging A named list of staging sets.
#'
#' @return Integer. Number of components whose live tensors are on CUDA.
#'
#' @keywords internal
.resident_on_gpu_count <- function(staging) {
    sum(vapply(staging, function(st) {
        isTRUE(length(st) > 0) &&
        identical(tryCatch(st[[1]]$live$device$type,
                           error = function(e) NA_character_), "cuda")
    }, logical(1)))
}

#' Refuse operations that the current state cannot serve
#'
#' @param res A resident handle.
#' @param verb What the caller is attempting, for the message.
#'
#' @return Invisibly TRUE, or an error.
#'
#' @keywords internal
.resident_guard <- function(res, verb) {
    if (identical(res$state, "unloaded")) {
        stop("cannot ", verb, ": this handle is unloaded", call. = FALSE)
    }
    if (identical(res$state, "broken")) {
        stop("cannot ", verb, ": this handle is broken (", res$last_error %||%
                 "no detail recorded", "). Only resident_status() and ",
             "resident_unload() work from here.", call. = FALSE)
    }
    invisible(TRUE)
}

#' Load a diffusion pipeline as a resident handle
#'
#' Loads a pipeline once and keeps its weights page-locked on the host
#' for the life of the handle. The GPU representation is created by
#' \code{\link{resident_activate}} and destroyed by
#' \code{\link{resident_deactivate}}, so a 16 GB card can hand itself
#' between models without either one re-reading its weights from disk.
#'
#' The handle is bound to one explicit GPU at load: a bare \code{"cuda"}
#' resolves to the current device now, and every later transition uses
#' that index, so the handle cannot drift to whichever GPU happens to be
#' current at transition time.
#'
#' One caveat on multi-GPU hosts: the family loader itself runs on the
#' \emph{current} device, and only the residency handle is bound to
#' \code{device}. Loading with \code{device = "cuda:1"} from a session
#' whose current device is 0 therefore stages through GPU 0 before the
#' first activation lands on GPU 1. Wrap the call in
#' \code{torch::with_device(device = "cuda:1", ...)} when that matters.
#'
#' The pipeline is left \emph{inactive} (weights pinned on the host, no
#' VRAM held). Call \code{\link{resident_activate}} before generating.
#'
#' @param model One of "flux1", "flux2", "zimage", "ltx".
#' @param device Target CUDA device, e.g. "cuda" or "cuda:1".
#' @param ... Passed to the family loader (\code{\link{flux_load_pipeline}},
#'   \code{\link{flux2_load_pipeline}}, \code{\link{zimage_load_pipeline}}
#'   or \code{\link{ltx23_load_pipeline}}). \code{ltx} requires
#'   \code{checkpoint_path}.
#' @param verbose Print progress messages.
#'
#' @return A \code{diffuseR_resident} handle (an environment). Inspect it
#'   with \code{\link{resident_status}}; the fields of interest are the
#'   state, the bound device, the component names, and the pinned host
#'   byte count.
#'
#' @seealso \code{\link{resident_activate}}, \code{\link{resident_status}}
#'
#' @examples
#' \dontrun{
#' res <- resident_load("flux2")
#' resident_activate(res)
#' img <- resident_generate(res, "a cat in a spacesuit", seed = 7)
#' resident_deactivate(res) # VRAM freed, weights stay pinned in RAM
#' resident_activate(res) # fast: DMA copy, no disk
#' resident_unload(res)
#' }
#'
#' @export
resident_load <- function(model = c("flux2", "flux1", "zimage", "ltx"),
                          device = "cuda", ..., verbose = TRUE) {
    model <- match.arg(model)
    if (!torch::cuda_is_available()) {
        stop("resident_load() requires CUDA", call. = FALSE)
    }
    if (!grepl("^cuda", device)) {
        stop("resident_load() requires a CUDA device, got '", device, "'",
             call. = FALSE)
    }
    # Bind to one explicit GPU now, so later transitions cannot drift.
    bound <- if (identical(device, "cuda")) {
        paste0("cuda:", torch::cuda_current_device())
    } else {
        device
    }

    loader <- switch(model,
                     flux1 = flux_load_pipeline,
                     flux2 = flux2_load_pipeline,
                     zimage = zimage_load_pipeline,
                     ltx = ltx23_load_pipeline)
    # Capture the phase-offload choice here rather than reading it back
    # off the pipeline: the FLUX family stores it as a field, LTX takes
    # it again at generate time and stores nothing, so the field is
    # absent there and a NULL would be misread as "stays resident".
    # Every family loader defaults it TRUE.
    dots <- list(...)
    phase_offload <- if (is.null(dots$phase_offload)) {
        TRUE
    } else {
        isTRUE(dots$phase_offload)
    }
    pipeline <- loader(device = "cuda", verbose = verbose, ...)

    staging <- .resident_pin(pipeline, verbose = verbose)
    # Pinning also evicted anything the loader had left resident, so the
    # handle starts inactive with no VRAM held.
    .resident_release_vram()

    res <- new.env(parent = emptyenv())
    res$model <- model
    res$device <- bound
    res$pipeline <- pipeline
    res$phase_offload <- phase_offload
    res$staging <- staging
    res$components <- names(.resident_components(pipeline))
    res$pinned_bytes <- .resident_pinned_bytes(staging)
    res$state <- "inactive"
    res$last_error <- NULL
    res$loaded_at <- Sys.time()
    structure(res, class = "diffuseR_resident")
}

#' Refuse a bulk activation that cannot fit
#'
#' Fails before the transfer rather than part-way through it. A partial
#' onload that OOMs is recoverable (activation rolls back), but it wastes
#' the transfer and reports a libtorch allocator error instead of the
#' actual problem, which is that this model does not fit this card.
#'
#' @param res A resident handle.
#' @param free_gb Free VRAM in GB. NULL measures it. Pass a value to
#'   make the decision deterministic: with no GPU the measurement is 0,
#'   which means "cannot tell" and never refuses, so a test that wants
#'   the refusal has to state the budget rather than depend on the
#'   machine having a card.
#'
#' @return Invisibly TRUE, or an error naming both figures.
#'
#' @keywords internal
.resident_check_fits <- function(res, free_gb = NULL) {
    if (is.null(free_gb)) {
        free_gb <- tryCatch(.detect_vram(use_free = TRUE),
                            error = function(e) NA_real_)
    }
    need_gb <- res$pinned_bytes / 1024^3
    if (!is.na(free_gb) && free_gb > 0 && need_gb > free_gb) {
        stop(sprintf(paste0("%s needs %.2f GB resident but only %.2f GB of ",
                            "VRAM is free. Load the pipeline with ",
                            "phase_offload = TRUE so components move on ",
                            "one phase at a time."),
                     res$model, need_gb, free_gb), call. = FALSE)
    }
    invisible(TRUE)
}

# gc() then empty the caching allocator. Split out so every transition
# releases VRAM the same way.
.resident_release_vram <- function() {
    gc()
    tryCatch(torch::cuda_empty_cache(), error = function(e) NULL)
    invisible(NULL)
}

#' Bring a resident pipeline onto the GPU
#'
#' Copies every pinned component to the handle's bound device by DMA and
#' verifies the result tensor-by-tensor. A failure rolls back to the
#' pinned host state; a rollback that cannot itself be verified leaves
#' the handle broken.
#'
#' What activation does depends on how the pipeline was loaded:
#'
#' \itemize{
#'   \item \code{phase_offload = TRUE} (the default, and what the
#'     \code{txt2img_*} functions expect): no bulk transfer. The render
#'     moves each component on as its phase begins and back off as it
#'     ends, from these same pinned copies, so pre-loading them would be
#'     undone within one phase. Activation is the ownership claim.
#'   \item \code{phase_offload = FALSE}: every component is copied to the
#'     card up front and stays there across renders. This is the fast
#'     path, and it is checked against free VRAM first.
#' }
#'
#' The distinction is not cosmetic. FLUX.1's pinned set is 15.73 GB,
#' which does not fit a 15.47 GiB card -- bulk-onloading it OOMs even
#' though the phased render fits comfortably. So \code{state} is a claim
#' about who owns the card, not a measurement of what is on it; read
#' \code{components_on_gpu} from \code{\link{resident_status}} for the
#' measurement.
#'
#' @param res A \code{diffuseR_resident} handle.
#'
#' @return Invisibly the handle, with state "active".
#'
#' @export
resident_activate <- function(res) {
    stopifnot(inherits(res, "diffuseR_resident"))
    .resident_guard(res, "activate")
    if (identical(res$state, "active")) {
        return(invisible(res))
    }
    if (!identical(res$state, "inactive")) {
        stop("cannot activate from state '", res$state, "'", call. = FALSE)
    }
    res$state <- "activating"
    # A phase-offloading pipeline moves each component onto the card as
    # its phase begins and straight back off as it ends, from these same
    # pinned copies. Bulk-onloading here is therefore redundant -- the
    # render undoes it within one phase -- and actively harmful: FLUX.1's
    # pinned set is 15.73 GB, which does not fit a 15.47 GiB card even
    # though the phased render does. For those pipelines, activation is
    # the ownership claim; the transfers stay per-phase.
    # The loader's own field wins when it kept one (the FLUX family
    # downgrades phase_offload to FALSE on a CPU device); otherwise fall
    # back to what resident_load() was asked for.
    bulk <- !isTRUE(res$pipeline$phase_offload %||% res$phase_offload)
    ok <- tryCatch({
        if (bulk) {
            .resident_check_fits(res)
            for (nm in names(res$staging)) {
                .staged_onload(res$staging[[nm]], res$device)
            }
        }
        TRUE
    }, error = function(e) {
        res$last_error <- conditionMessage(e)
        FALSE
    })
    if (ok) {
        res$state <- "active"
        return(invisible(res))
    }
    # Roll back to pinned host state and verify it.
    rolled <- tryCatch({
        for (nm in names(res$staging)) {
            .staged_offload(res$staging[[nm]])
        }
        .resident_release_vram()
        .resident_all_on(res$staging, "cpu")
    }, error = function(e) FALSE)
    if (isTRUE(rolled)) {
        res$state <- "inactive"
        stop("resident_activate() failed (rolled back to pinned host ",
             "state): ", res$last_error, call. = FALSE)
    }
    res$state <- "broken"
    stop("resident_activate() failed and the rollback could not be ",
         "verified: ", res$last_error, call. = FALSE)
}

#' Release a resident pipeline's VRAM
#'
#' Re-points every component at its pinned host copy and drops the GPU
#' storage. Weights are immutable during inference, so the pinned copies
#' are still current and this moves no bytes: it is a pointer swap plus a
#' cache release. The handle stays loaded and can be reactivated without
#' touching the disk.
#'
#' @param res A \code{diffuseR_resident} handle.
#' @param release Empty the CUDA caching allocator afterwards. Leave TRUE
#'   unless another handle on the same device is about to reuse the pool.
#'
#' @return Invisibly the handle, with state "inactive".
#'
#' @export
resident_deactivate <- function(res, release = TRUE) {
    stopifnot(inherits(res, "diffuseR_resident"))
    .resident_guard(res, "deactivate")
    if (identical(res$state, "inactive")) {
        return(invisible(res))
    }
    if (!identical(res$state, "active")) {
        stop("cannot deactivate from state '", res$state, "'", call. = FALSE)
    }
    res$state <- "deactivating"
    verified <- tryCatch({
        for (nm in names(res$staging)) {
            .staged_offload(res$staging[[nm]])
        }
        if (isTRUE(release)) {
            .resident_release_vram()
        }
        .resident_all_on(res$staging, "cpu")
    }, error = function(e) {
        res$last_error <- conditionMessage(e)
        FALSE
    })
    if (isTRUE(verified)) {
        res$state <- "inactive"
        return(invisible(res))
    }
    res$state <- "broken"
    stop("resident_deactivate() could not verify the pinned host state; ",
         "the handle is broken and holds no usable GPU copy. ",
         res$last_error %||% "", call. = FALSE)
}

#' Generate from an active resident pipeline
#'
#' Dispatches to the family's generator with the resident pipeline
#' supplied, so no weights are re-read. The handle must be active.
#'
#' @param res A \code{diffuseR_resident} handle.
#' @param prompt Character. The text prompt.
#' @param ... Passed to \code{\link{txt2img_flux}},
#'   \code{\link{txt2img_flux2}}, \code{\link{txt2img_zimage}} or
#'   \code{\link{txt2vid_ltx2}}.
#'
#' @return Whatever the family generator returns: an image array for the
#'   image families, a video array for \code{ltx}.
#'
#' @export
resident_generate <- function(res, prompt, ...) {
    stopifnot(inherits(res, "diffuseR_resident"))
    .resident_guard(res, "generate")
    if (!identical(res$state, "active")) {
        stop("cannot generate from state '", res$state,
             "'; call resident_activate() first", call. = FALSE)
    }
    gen <- switch(res$model,
                  flux1 = txt2img_flux,
                  flux2 = txt2img_flux2,
                  zimage = txt2img_zimage,
                  ltx = txt2vid_ltx2)
    gen(prompt, pipeline = res$pipeline, ...)
}

#' Status of a resident handle
#'
#' @param res A \code{diffuseR_resident} handle.
#'
#' @return A list with \code{model}, \code{state}, \code{device},
#'   \code{components} (character vector of pinned component names),
#'   \code{pinned_bytes} (page-locked host bytes held),
#'   \code{gpu_allocated} and \code{gpu_reserved} (bytes the CUDA
#'   caching allocator reports live and held for this process, NA
#'   without CUDA), \code{components_on_gpu} (how many components are
#'   *actually* resident right now), \code{loaded_at}, and
#'   \code{last_error} (NULL unless a transition failed).
#'
#'   \code{state} is the handle's claim on the card;
#'   \code{components_on_gpu} is the measurement. They disagree by
#'   design after a render on a \code{phase_offload = TRUE} pipeline,
#'   which returns each component to pinned host memory as its phase
#'   finishes: the handle stays "active" (it still owns the card's
#'   budget and can render again without touching disk) while
#'   \code{components_on_gpu} is 0. Schedule on the measurement.
#'
#' @export
resident_status <- function(res) {
    stopifnot(inherits(res, "diffuseR_resident"))
    mem <- .cuda_bytes()
    list(model = res$model,
         state = res$state,
         device = res$device,
         components = names(res$staging),
         components_on_gpu = .resident_on_gpu_count(res$staging),
         pinned_bytes = res$pinned_bytes,
         gpu_allocated = mem$allocated,
         gpu_reserved = mem$reserved,
         loaded_at = res$loaded_at,
         last_error = res$last_error)
}

# Live and reserved CUDA bytes for this process. torch has no
# cuda_memory_allocated(); the numbers live under cuda_memory_stats(),
# which itself errors without a CUDA build, hence the tryCatch.
.cuda_bytes <- function() {
    s <- tryCatch(torch::cuda_memory_stats(), error = function(e) NULL)
    if (is.null(s)) {
        return(list(allocated = NA_real_, reserved = NA_real_))
    }
    list(allocated = s$allocated_bytes$all$current %||% NA_real_,
         reserved = s$reserved_bytes$all$current %||% NA_real_)
}

#' Drop a resident handle entirely
#'
#' Releases the GPU copy if any, drops the pipeline and the pinned host
#' storage, and marks the handle unloaded. Terminal: nothing but
#' \code{\link{resident_status}} works afterwards.
#'
#' @param res A \code{diffuseR_resident} handle.
#'
#' @return Invisibly the handle, with state "unloaded".
#'
#' @export
resident_unload <- function(res) {
    stopifnot(inherits(res, "diffuseR_resident"))
    if (identical(res$state, "unloaded")) {
        return(invisible(res))
    }
    # Best effort: a broken handle still gets its memory back.
    tryCatch({
        for (nm in names(res$staging)) {
            .staged_offload(res$staging[[nm]])
        }
    }, error = function(e) NULL)
    res$pipeline <- NULL
    res$staging <- list()
    res$components <- character(0)
    res$pinned_bytes <- 0
    res$state <- "unloaded"
    .resident_release_vram()
    invisible(res)
}

#' Print a resident handle
#'
#' @param x A \code{diffuseR_resident} handle.
#' @param ... Ignored.
#'
#' @return Invisibly \code{x}. Called for the side effect of printing a
#'   one-block summary to the console.
#'
#' @export
print.diffuseR_resident <- function(x, ...) {
    s <- resident_status(x)
    cat("<diffuseR resident>\n")
    cat("  model:      ", s$model, "\n", sep = "")
    cat("  state:      ", s$state, "\n", sep = "")
    cat("  device:     ", s$device, "\n", sep = "")
    cat("  components: ",
        if (length(s$components)) paste(s$components, collapse = ", ") else "-",
        "\n", sep = "")
    cat("  on gpu:     ", s$components_on_gpu, " of ", length(s$components),
        "\n", sep = "")
    cat("  pinned:     ", .fmt_gb(s$pinned_bytes), "\n", sep = "")
    if (!is.na(s$gpu_allocated)) {
        cat("  gpu:        ", .fmt_gb(s$gpu_allocated), " allocated, ",
            .fmt_gb(s$gpu_reserved), " reserved\n", sep = "")
    }
    if (!is.null(s$last_error)) {
        cat("  last error: ", s$last_error, "\n", sep = "")
    }
    invisible(x)
}

# Byte count as GB, for the print method and messages.
.fmt_gb <- function(b) {
    if (is.null(b) || is.na(b) || b <= 0) {
        return("0 GB")
    }
    sprintf("%.2f GB", b / 1024^3)
}

# Residency contract: component discovery, byte accounting, the state
# machine, and the pinned<->GPU round trip.
#
# The state-machine tests build a handle by hand around synthetic
# nn_modules, so they exercise the transitions without loading a real
# multi-GB pipeline. The round-trip test needs CUDA and is skipped
# without it.

library(tinytest)
library(diffuseR)

fake_pipeline <- function() {
    structure(list(transformer = torch::nn_linear(8, 8),
                   decoder = torch::nn_linear(8, 4),
                   text_encoder = torch::nn_linear(4, 8),
                   config = list(family = "test"),
                   scheduler = "flowmatch",
                   phase_offload = TRUE),
              class = "test_pipeline")
}

# Everything from here to the "state machine" section builds real
# nn_modules or torch dtypes, so it needs a working lantern. torch
# installed WITHOUT its lantern binaries is the normal state on
# win-builder and CRAN, and there `torch::nn_linear()` errors rather
# than returning. The state-machine sections below are pure R and run
# everywhere.
have_torch <- requireNamespace("torch", quietly = TRUE) &&
    torch::torch_is_installed()

# --- component discovery ----------------------------------------------------------

if (have_torch) {

pipe <- fake_pipeline()
comps <- diffuseR:::.resident_components(pipe)
expect_equal(sort(names(comps)), c("decoder", "text_encoder", "transformer"))
# Non-module fields are not components.
expect_false("config" %in% names(comps))
expect_false("scheduler" %in% names(comps))
expect_false("phase_offload" %in% names(comps))

# A pipeline with no modules yields an empty set, not an error.
expect_equal(length(diffuseR:::.resident_components(list(a = 1, b = "x"))), 0L)

# --- dtype byte table -------------------------------------------------------------

expect_equal(diffuseR:::.dtype_bytes(torch::torch_float32()), 4)
expect_equal(diffuseR:::.dtype_bytes(torch::torch_float16()), 2)
expect_equal(diffuseR:::.dtype_bytes(torch::torch_bfloat16()), 2)
expect_equal(diffuseR:::.dtype_bytes(torch::torch_uint8()), 1)
expect_equal(diffuseR:::.dtype_bytes(torch::torch_int64()), 8)

} # end have_torch

# --- byte formatting --------------------------------------------------------------

expect_equal(diffuseR:::.fmt_gb(0), "0 GB")
expect_equal(diffuseR:::.fmt_gb(NA), "0 GB")
expect_equal(diffuseR:::.fmt_gb(NULL), "0 GB")
expect_equal(diffuseR:::.fmt_gb(1024^3), "1.00 GB")

# --- state guard ------------------------------------------------------------------

mk <- function(state, staging = list(), phase_offload = FALSE,
               pinned_bytes = 0, store_field = TRUE) {
    e <- new.env(parent = emptyenv())
    e$model <- "flux2"
    e$device <- "cuda:0"
    e$state <- state
    e$staging <- staging
    e$components <- character(0)
    e$pinned_bytes <- pinned_bytes
    e$last_error <- NULL
    e$loaded_at <- Sys.time()
    # store_field = FALSE mimics LTX, which never records phase_offload
    # on the pipeline object.
    e$pipeline <- if (store_field) list(phase_offload = phase_offload) else list()
    e$phase_offload <- phase_offload
    structure(e, class = "diffuseR_resident")
}

# A pipeline that does not store the field must still be read as
# phase-offloading (this is what sent LTX down the bulk path and made it
# demand 18.41 GB on a 15 GB card).
expect_true(isTRUE(mk("inactive", phase_offload = TRUE,
                      store_field = FALSE)$phase_offload))

expect_error(diffuseR:::.resident_guard(mk("unloaded"), "activate"),
             pattern = "unloaded")
expect_error(diffuseR:::.resident_guard(mk("broken"), "activate"),
             pattern = "broken")
expect_true(diffuseR:::.resident_guard(mk("inactive"), "activate"))
expect_true(diffuseR:::.resident_guard(mk("active"), "generate"))

# --- illegal transitions ----------------------------------------------------------

# Deactivating something that was never activated is an error, not a
# silent no-op, unless it is already inactive.
expect_silent(resident_deactivate(mk("inactive")))
expect_error(resident_deactivate(mk("activating")), pattern = "cannot deactivate")
expect_error(resident_activate(mk("deactivating")), pattern = "cannot activate")

# Activate on an already-active handle is idempotent.
expect_silent(resident_activate(mk("active")))

# Generation is refused unless active.
expect_error(resident_generate(mk("inactive"), "a cat"),
             pattern = "resident_activate")
expect_error(resident_generate(mk("broken"), "a cat"), pattern = "broken")

# --- a set that cannot fit is refused before any transfer -------------------------

# Named figures beat a libtorch allocator error: FLUX.1's 15.73 GB pinned
# set does not fit a 15.47 GiB card, and the message should say so.
#
# free_gb is stated rather than measured. Measured, it is 0 on any
# machine without a GPU -- which means "cannot tell", so the check
# declines to refuse and these assertions saw no error. That is correct
# behaviour and a broken test: it passed on the GPU box and failed CI on
# both ubuntu-latest and macos-latest.
huge <- mk("inactive", pinned_bytes = 16 * 1024^3)
expect_error(diffuseR:::.resident_check_fits(huge, free_gb = 15.47),
             pattern = "needs")
expect_error(diffuseR:::.resident_check_fits(huge, free_gb = 15.47),
             pattern = "phase_offload")
# A set that fits is not refused.
expect_true(diffuseR:::.resident_check_fits(huge, free_gb = 24))
# A trivial set is never refused.
expect_true(diffuseR:::.resident_check_fits(mk("inactive", pinned_bytes = 0),
                                            free_gb = 15.47))
# Undetectable VRAM (0, as on a GPU-less machine) must not block: the
# check refuses only when it can prove the set does not fit.
expect_true(diffuseR:::.resident_check_fits(huge, free_gb = 0))
expect_true(diffuseR:::.resident_check_fits(huge, free_gb = NA_real_))

# --- status and print -------------------------------------------------------------

h <- mk("inactive")
s <- resident_status(h)
expect_true(all(c("model", "state", "device", "components",
                  "components_on_gpu", "pinned_bytes", "gpu_allocated",
                  "gpu_reserved", "loaded_at", "last_error") %in% names(s)))
# An empty staging set has nothing resident.
expect_equal(s$components_on_gpu, 0L)
# The allocator numbers must come from a torch function that actually
# exists: cuda_memory_allocated() does not (R CMD check catches it, but
# only as a WARNING buried in the dependencies step).
expect_true("cuda_memory_stats" %in% getNamespaceExports("torch"))
mem <- diffuseR:::.cuda_bytes()
expect_true(all(c("allocated", "reserved") %in% names(mem)))
expect_equal(s$state, "inactive")
expect_null(s$last_error)

out <- capture.output(print(h))
expect_true(any(grepl("diffuseR resident", out)))
expect_true(any(grepl("flux2", out)))

# --- unload is terminal and idempotent --------------------------------------------

u <- mk("inactive")
resident_unload(u)
expect_equal(u$state, "unloaded")
expect_silent(resident_unload(u))
expect_error(resident_activate(u), pattern = "unloaded")
# Status still works on an unloaded handle.
expect_equal(resident_status(u)$state, "unloaded")

# --- CUDA round trip --------------------------------------------------------------

if (have_torch && at_home() && torch::cuda_is_available()) {
    pipe <- fake_pipeline()
    staging <- diffuseR:::.resident_pin(pipe, verbose = FALSE)
    expect_equal(sort(names(staging)),
                 c("decoder", "text_encoder", "transformer"))

    # Pinning leaves everything on the host.
    expect_true(diffuseR:::.resident_all_on(staging, "cpu"))
    expect_true(diffuseR:::.resident_pinned_bytes(staging) > 0)

    res <- mk("inactive", staging)
    res$components <- names(staging)

    resident_activate(res)
    expect_equal(res$state, "active")
    expect_true(diffuseR:::.resident_all_on(staging, "cuda"))
    expect_equal(resident_status(res)$components_on_gpu, 3L)

    resident_deactivate(res)
    expect_equal(res$state, "inactive")
    expect_true(diffuseR:::.resident_all_on(staging, "cpu"))
    expect_equal(resident_status(res)$components_on_gpu, 0L)

    # components_on_gpu is ground truth, not the declared state: evict
    # behind the handle's back and status must notice while state still
    # says "active". This is what a phase-offloading render does.
    resident_activate(res)
    for (nm in names(staging)) diffuseR:::.staged_offload(staging[[nm]])
    expect_equal(res$state, "active")
    expect_equal(resident_status(res)$components_on_gpu, 0L)
    resident_deactivate(res)

    # A second round trip reuses the same pinned buffers.
    resident_activate(res)
    expect_true(diffuseR:::.resident_all_on(staging, "cuda"))
    resident_deactivate(res)
    expect_true(diffuseR:::.resident_all_on(staging, "cpu"))

    # The module still computes correctly after the round trip.
    y <- torch::with_no_grad(pipe$transformer(torch::torch_randn(c(2, 8))))
    expect_equal(as.integer(y$shape), c(2L, 8L))

    resident_unload(res)
    expect_equal(res$state, "unloaded")
}

#' filename_from_prompt
#' @title Generate a filename from a prompt
#' @description This function generates a filename from a prompt by removing all non-alphanumeric characters and replacing them with underscores. The filename is limited to 50 characters. If `datetime` is set to TRUE, the current date and time are prepended to the filename.
#'
#' @param prompt A character string representing the prompt.
#' @param datetime Logical indicating whether to prepend the current date and time to the filename. Default is TRUE.
#' @return A character string representing the generated filename.
#'
#' @examples
#' filename_from_prompt("A beautiful sunset over the mountains")
#' filename_from_prompt("A beautiful sunset over the mountains", datetime = FALSE)
#' @export
filename_from_prompt <- function(prompt, datetime = TRUE) {
    # Remove all non-alphanumeric characters from the prompt
    prompt_strs <- gsub("[^a-zA-Z0-9]", "_", prompt)
    # limit prompt length to 50 characters
    prompt_strs <- substr(prompt_strs, 1, 50)
    if (datetime == FALSE) {
        # Create a filename using the modified prompt
        file_name <- paste0("prompt_", prompt_strs, ".png")
    } else {
        # date time stamp
        datetime <- format(Sys.time(), "%Y%m%d_%H%M%S")
        file_name <- paste0(datetime, "_", prompt_strs, ".png")
    }
    return(file_name)
}

#' Z-Image Transformer
#'
#' Fresh R port of ZImageTransformer2DModel from the diffusers reference
#' (Apache-2.0, src/diffusers/models/transformers/transformer_z_image.py).
#' Single-stream DiT: image tokens pass through a modulated noise
#' refiner, caption tokens through an unmodulated context refiner, then
#' both are concatenated (image first) and run through the main trunk.
#' The module tree mirrors the reference state-dict keys 1:1
#' (all_x_embedder.2-1, noise_refiner.N, context_refiner.N, layers.N,
#' all_final_layer.2-1, t_embedder, cap_embedder, x_pad_token,
#' cap_pad_token).
#'
#' This port is batch-of-1: \code{x} is a single latent [C, F, H, W] and
#' \code{cap_feats} a single caption [L, cap_feat_dim], so sub-sequences
#' are uniform and no attention mask is needed. Padding to a multiple of
#' 32 tokens uses the learned pad parameters, appended after embedding
#' (the reference pads raw features with repeats, embeds pointwise, then
#' overwrites the pad rows with the same learned tokens).
#'
#' @param in_channels Integer. Latent channels. Default 16.
#' @param dim Integer. Model width. Default 3840.
#' @param n_layers Integer. Main trunk depth. Default 30.
#' @param n_refiner_layers Integer. Refiner depth. Default 2.
#' @param n_heads Integer. Attention heads. Default 30.
#' @param norm_eps Numeric. RMSNorm epsilon. Default 1e-5.
#' @param cap_feat_dim Integer. Caption embedding width. Default 2560.
#' @param rope_theta Numeric. RoPE base frequency. Default 256.
#' @param t_scale Numeric. Timestep scale. Default 1000.
#' @param axes_dims Integer vector. Per-axis rotary dims. Default
#'   c(32, 48, 48).
#' @param patch_size Integer. Spatial patch size. Default 2.
#' @param f_patch_size Integer. Temporal patch size. Default 1.
#'
#' @return Module whose forward(x, t, cap_feats, chunk_size) returns the
#'   predicted velocity for the single latent, a tensor [C, F, H, W]
#'   matching the shape of \code{x}. Note that the checkpoint negates
#'   this output and consumes a reversed timestep; see
#'   \code{\link{txt2img_zimage}}.
#'
#' @export
zimage_transformer <- torch::nn_module(
                                       "zimage_transformer",
                                       initialize = function(in_channels = 16L, dim = 3840L, n_layers = 30L,
        n_refiner_layers = 2L, n_heads = 30L, norm_eps = 1e-5,
        cap_feat_dim = 2560L, rope_theta = 256, t_scale = 1000,
        axes_dims = c(32L, 48L, 48L), patch_size = 2L, f_patch_size = 1L) {
    stopifnot(dim %/% n_heads == sum(axes_dims))
    self$in_channels <- in_channels
    self$out_channels <- in_channels
    self$dim <- dim
    self$rope_theta <- rope_theta
    self$t_scale <- t_scale
    self$axes_dims <- axes_dims
    self$patch_size <- patch_size
    self$f_patch_size <- f_patch_size
    patch_key <- paste0(patch_size, "-", f_patch_size)
    patch_dim <- f_patch_size * patch_size * patch_size * in_channels

    embedders <- list(torch::nn_linear(patch_dim, dim, bias = TRUE))
    names(embedders) <- patch_key
    self$all_x_embedder <- torch::nn_module_dict(embedders)

    finals <- list(zimage_final_layer(dim, patch_dim))
    names(finals) <- patch_key
    self$all_final_layer <- torch::nn_module_dict(finals)

    self$noise_refiner <- torch::nn_module_list(lapply(
            seq_len(n_refiner_layers),
            function(i) zimage_block(dim, n_heads, norm_eps, modulation = TRUE)
        ))
    self$context_refiner <- torch::nn_module_list(lapply(
            seq_len(n_refiner_layers),
            function(i) zimage_block(dim, n_heads, norm_eps, modulation = FALSE)
        ))
    self$layers <- torch::nn_module_list(lapply(seq_len(n_layers),
            function(i) zimage_block(dim, n_heads, norm_eps, modulation = TRUE)))

    self$t_embedder <- zimage_t_embedder(out_size = min(dim, 256L),
        mid_size = 1024L)
    self$cap_embedder <- torch::nn_sequential(
        ltx23_rms_norm(cap_feat_dim, eps = norm_eps),
        torch::nn_linear(cap_feat_dim, dim, bias = TRUE)
    )
    self$x_pad_token <- torch::nn_parameter(torch::torch_zeros(1L, dim))
    self$cap_pad_token <- torch::nn_parameter(torch::torch_zeros(1L, dim))
},
                                       forward = function(x, t, cap_feats, chunk_size = NULL) {
    # x: [C, F, H, W] latent; t: [1] in [0, 1]; cap_feats: [L, cap_feat_dim]
    device <- x$device
    p <- self$patch_size
    pf <- self$f_patch_size
    patch_key <- paste0(p, "-", pf)
    size <- x$shape[2:4]
    f_tokens <- size[1] %/% pf
    h_tokens <- size[2] %/% p
    w_tokens <- size[3] %/% p

    adaln <- self$t_embedder(t * self$t_scale)$to(dtype = x$dtype) # [1, 256]

    # Position ids and rotary frequencies
    cap_len <- cap_feats$shape[1]
    cap_padded <- cap_len + zimage_pad_len(cap_len)
    cap_freqs <- zimage_pos_embed(
                                  zimage_cap_pos_ids(cap_padded, device = device),
                                  axes_dim = self$axes_dims, theta = self$rope_theta
    )
    img_freqs <- zimage_pos_embed(
                                  zimage_img_pos_ids(h_tokens, w_tokens, start0 = cap_padded + 1L,
            f_tokens = f_tokens, device = device),
                                  axes_dim = self$axes_dims, theta = self$rope_theta
    )

    # Image tokens: patchify, embed, pad, refine
    tokens <- zimage_patchify(x, p, pf)
    img_len <- tokens$shape[1]
    x_emb <- self$all_x_embedder[[patch_key]](tokens)$unsqueeze(1L)
    img_pad <- zimage_pad_len(img_len)
    if (img_pad > 0L) {
        pad_rows <- self$x_pad_token$unsqueeze(1L)$expand(c(1L, img_pad, self$dim))
        x_emb <- torch::torch_cat(list(x_emb, pad_rows), dim = 2L)
    }
    for (i in seq_len(length(self$noise_refiner))) {
        x_emb <- self$noise_refiner[[i]](x_emb, img_freqs, adaln_input = adaln,
                                       chunk_size = chunk_size)
    }

    # Caption tokens: embed, pad, refine
    cap_emb <- self$cap_embedder(cap_feats)$unsqueeze(1L)
    if (cap_padded > cap_len) {
        pad_rows <- self$cap_pad_token$unsqueeze(1L)$expand(
            c(1L, cap_padded - cap_len, self$dim)
        )
        cap_emb <- torch::torch_cat(list(cap_emb, pad_rows), dim = 2L)
    }
    for (i in seq_len(length(self$context_refiner))) {
        cap_emb <- self$context_refiner[[i]](cap_emb, cap_freqs,
                                       chunk_size = chunk_size)
    }

    # Unified sequence, image first
    unified <- torch::torch_cat(list(x_emb, cap_emb), dim = 2L)
    unified_freqs <- list(
                          torch::torch_cat(list(img_freqs[[1]], cap_freqs[[1]]), dim = 1L),
                          torch::torch_cat(list(img_freqs[[2]], cap_freqs[[2]]), dim = 1L)
    )
    for (i in seq_len(length(self$layers))) {
        unified <- self$layers[[i]](unified, unified_freqs,
                                       adaln_input = adaln,
                                       chunk_size = chunk_size)
    }

    out <- self$all_final_layer[[patch_key]](unified, adaln)
    zimage_unpatchify(out$squeeze(1L), size, p, pf, self$out_channels)
}
)

#' Re-shard a large safetensors file into sub-2 GB shards
#'
#' Splits a single \code{.safetensors} file into diffusers-style shards
#' (\code{<base>-00001-of-000NN.safetensors} plus a
#' \code{<base>.safetensors.index.json} weight map) so it loads on stock
#' CRAN safetensors, which overflows a 32-bit offset on any file at or
#' above 2^31 bytes. Reading the oversize source requires a
#' fork-patched safetensors (a build-machine step); the shards it writes
#' are fork-free to read. Used to host large fp16 diffusers weights
#' (e.g. the 5 GB SDXL UNet) unchanged, without quantization.
#'
#' @param input Path to the source \code{.safetensors} file, or a
#'   directory containing \code{<base>.safetensors}.
#' @param output_dir Output directory for the shards + index.
#' @param base Shard basename (default
#'   \code{"diffusion_pytorch_model"}).
#' @param shard_bytes Target shard size; the default 1.9e9 keeps each
#'   shard under the ~2.15 GB ceiling.
#' @param verbose Logical.
#'
#' @return Invisibly, the path to the written index.json.
#'
#' @export
reshard_safetensors <- function(input, output_dir,
                                base = "diffusion_pytorch_model",
                                shard_bytes = 1.9e9, verbose = TRUE) {
    if (!requireNamespace("safetensors", quietly = TRUE)) {
        stop("The safetensors package is required to reshard.")
    }
    input <- path.expand(input)
    if (dir.exists(input)) {
        input <- file.path(input, paste0(base, ".safetensors"))
    }
    if (!file.exists(input)) {
        stop("No safetensors file at ", input)
    }
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
    handle <- safetensors::safetensors$new(input, framework = "torch")
    keys <- setdiff(handle$keys(), "__metadata__")

    shard <- list()
    shard_size <- 0
    shard_idx <- 0L
    key_to_shard <- integer(0) # 1-based shard index per key (named)
    total_size <- 0

    flush <- function() {
        if (!length(shard)) {
            return()
        }
        shard_idx <<- shard_idx + 1L
        # temporary name; renamed once the shard total is known
        fname <- sprintf("%s-%05d.tmp.safetensors", base, shard_idx)
        safetensors::safe_save_file(shard, file.path(output_dir, fname))
        if (verbose) {
            message(sprintf("  shard %d: %.2f GB, %d tensors", shard_idx,
                            shard_size / 1e9, length(shard)))
        }
        shard <<- list()
        shard_size <<- 0
        gc(verbose = FALSE)
    }

    torch::with_no_grad({
        for (key in keys) {
            t <- handle$get_tensor(key)
            bytes <- prod(t$shape) *
            as.integer(switch(as.character(t$dtype),
                              Float = 4L, Double = 8L, Half = 2L,
                              BFloat16 = 2L, Byte = 1L, Char = 1L,
                              Long = 8L, Int = 4L, 4L))
            if (shard_size > 0 && shard_size + bytes > shard_bytes) {
                flush()
            }
            shard[[key]] <- t
            shard_size <- shard_size + bytes
            total_size <- total_size + bytes
            key_to_shard[[key]] <- shard_idx + 1L
        }
    })
    flush()

    n <- shard_idx
    # rename tmp shards to the diffusers -of- convention, build the map
    weight_map <- list()
    final_names <- character(n)
    for (i in seq_len(n)) {
        final_names[i] <- sprintf("%s-%05d-of-%05d.safetensors", base, i, n)
        file.rename(file.path(output_dir, sprintf("%s-%05d.tmp.safetensors", base, i)),
                    file.path(output_dir, final_names[i]))
    }
    for (key in names(key_to_shard)) {
        weight_map[[key]] <- final_names[key_to_shard[[key]]]
    }
    index <- list(metadata = list(total_size = total_size),
                  weight_map = weight_map)
    index_path <- file.path(output_dir, paste0(base, ".safetensors.index.json"))
    jsonlite::write_json(index, index_path, auto_unbox = TRUE, pretty = TRUE)
    if (verbose) {
        message(sprintf("Re-sharded %d tensors into %d shards (%.2f GB): %s",
                        length(keys), n, total_size / 1e9, output_dir))
    }
    invisible(index_path)
}

#' Z-Image Rotary Positional Embeddings and Patchify Helpers
#'
#' Fresh R port of the Z-Image position scheme from the diffusers
#' reference (Apache-2.0,
#' src/diffusers/models/transformers/transformer_z_image.py RopeEmbedder,
#' create_coordinate_grid, _patchify_image, _pad_with_ids, unpatchify).
#' Z-Image uses 3-axis interleaved RoPE with theta 256; frequencies are
#' built in float64 but the angles are cast to float32 before cos/sin
#' (torch.polar on a .float() tensor), which differs measurably from the
#' FLUX convention at large positions. Every sub-sequence is padded to a
#' multiple of 32 (SEQ_MULTI_OF); caption positions are a 1-based ramp on
#' axis 1 built over the padded length, image positions sit on axes 2/3
#' with axis 1 offset just past the caption.
#'
#' @name rope_zimage
NULL

# Sub-sequences are padded to a multiple of this many tokens.
.zimage_seq_multi_of <- 32L

#' Padding length to the next multiple of 32
#'
#' @param n Integer token count.
#' @return Integer pad length in [0, 31].
#' @keywords internal
zimage_pad_len <- function(n) {
    (-n) %% .zimage_seq_multi_of
}

#' Build Z-Image caption position ids
#'
#' Caption tokens ramp 1..cap_padded_len on the first axis (axes 2 and 3
#' zero). The reference builds the grid over the already-padded length,
#' so pad tokens continue the ramp rather than sitting at the origin
#' (the (0,0,0) pad ids it also emits are truncated away in
#' _prepare_sequence and never reach RoPE).
#'
#' @param cap_padded_len Integer. Caption length after padding to a
#'   multiple of 32.
#' @param device Device for the resulting tensor.
#'
#' @return Float tensor of shape [cap_padded_len, 3].
#'
#' @export
zimage_cap_pos_ids <- function(cap_padded_len, device = "cpu") {
    f32 <- torch::torch_float32()
    ids <- torch::torch_zeros(cap_padded_len, 3L, dtype = f32, device = device)
    ids[, 1] <- torch::torch_arange(start = 1, end = cap_padded_len,
                                    dtype = f32, device = device)
    ids
}

#' Build Z-Image latent image position ids
#'
#' Image tokens use axis 1 for the frame index offset past the caption
#' (start0 = cap_padded_len + 1), axis 2 for the token row and axis 3 for
#' the token column. Trailing pad tokens (token count not a multiple of
#' 32) sit at (0, 0, 0). Reference: patchify_and_embed / _pad_with_ids.
#'
#' @param h_tokens Integer. Token grid height (latent height / patch).
#' @param w_tokens Integer. Token grid width (latent width / patch).
#' @param start0 Integer. First-axis start, cap_padded_len + 1.
#' @param f_tokens Integer. Token grid frames; 1 for txt2img.
#' @param device Device for the resulting tensor.
#'
#' @return Float tensor of shape [padded token count, 3].
#'
#' @export
zimage_img_pos_ids <- function(h_tokens, w_tokens, start0, f_tokens = 1L,
                               device = "cpu") {
    f32 <- torch::torch_float32()
    ids <- torch::torch_zeros(f_tokens, h_tokens, w_tokens, 3L,
                              dtype = f32, device = device)
    frames <- torch::torch_arange(start = start0,
                                  end = start0 + f_tokens - 1,
                                  dtype = f32, device = device)
    rows <- torch::torch_arange(start = 0, end = h_tokens - 1, dtype = f32,
                                device = device)
    cols <- torch::torch_arange(start = 0, end = w_tokens - 1, dtype = f32,
                                device = device)
    ids[,,, 1] <- ids[,,, 1] + frames$reshape(c(-1L, 1L, 1L))
    ids[,,, 2] <- ids[,,, 2] + rows$reshape(c(1L, -1L, 1L))
    ids[,,, 3] <- ids[,,, 3] + cols$reshape(c(1L, 1L, -1L))
    ids <- ids$reshape(c(f_tokens * h_tokens * w_tokens, 3L))

    pad <- zimage_pad_len(ids$shape[1])
    if (pad > 0L) {
        ids <- torch::torch_cat(list(
                                     ids,
                                     torch::torch_zeros(pad, 3L, dtype = f32, device = device)
            ))
    }
    ids
}

#' Compute Z-Image rotary frequencies from position ids
#'
#' Per-axis 1D rotary frequencies in the interleaved-real convention.
#' Frequencies and angles are built in float64, then the angles are cast
#' to float32 before cos/sin — matching the reference torch.polar call on
#' a .float() tensor. Output format matches \code{flux_pos_embed} so
#' \code{flux_apply_rotary_emb} applies unchanged.
#'
#' @param ids Tensor of shape [S, 3] from \code{zimage_cap_pos_ids} /
#'   \code{zimage_img_pos_ids}.
#' @param axes_dim Integer vector of per-axis rotary dims; must sum to
#'   the attention head dim. Z-Image uses c(32, 48, 48).
#' @param theta Numeric. RoPE base frequency. Z-Image uses 256.
#'
#' @return List of two tensors (cos, sin), each [S, sum(axes_dim)],
#'   float32, on the device of \code{ids}.
#'
#' @export
zimage_pos_embed <- function(ids, axes_dim = c(32L, 48L, 48L), theta = 256) {
    n_axes <- ids$shape[2]
    device <- ids$device
    f64 <- torch::torch_float64()
    f32 <- torch::torch_float32()
    pos <- ids$to(dtype = f64)$cpu()

    cos_out <- vector("list", n_axes)
    sin_out <- vector("list", n_axes)
    for (i in seq_len(n_axes)) {
        d <- axes_dim[i]
        exponents <- torch::torch_arange(start = 0, end = d - 2, step = 2,
            dtype = f64)
        freqs <- 1.0 / torch::torch_pow(theta, exponents / d)
        # Angle in float64, cast to float32 BEFORE cos/sin (reference:
        # torch.outer(...).float() then torch.polar)
        angles <- (pos[, i]$unsqueeze(2L) * freqs$unsqueeze(1L))$to(dtype = f32)
        cos_out[[i]] <- angles$cos()$repeat_interleave(2L, dim = 2L)
        sin_out[[i]] <- angles$sin()$repeat_interleave(2L, dim = 2L)
    }

    list(
         torch::torch_cat(cos_out, dim = -1L)$to(device = device),
         torch::torch_cat(sin_out, dim = -1L)$to(device = device)
    )
}

#' Patchify a latent image to Z-Image tokens
#'
#' (C, F, H, W) -> [F/pF * H/p * W/p, pF * p * p * C], matching
#' _patchify_image. No padding is applied here.
#'
#' @param image Tensor of shape [C, F, H, W].
#' @param patch_size Integer spatial patch size. Default 2.
#' @param f_patch_size Integer temporal patch size. Default 1.
#'
#' @return Tensor of shape [num_tokens, patch_dim].
#'
#' @export
zimage_patchify <- function(image, patch_size = 2L, f_patch_size = 1L) {
    p <- patch_size
    pf <- f_patch_size
    shape <- image$shape
    c_in <- shape[1]
    f_tokens <- shape[2] %/% pf
    h_tokens <- shape[3] %/% p
    w_tokens <- shape[4] %/% p
    image <- image$view(c(c_in, f_tokens, pf, h_tokens, p, w_tokens, p))
    # Python permute(1, 3, 5, 2, 4, 6, 0) -> R 1-indexed
    image <- image$permute(c(2L, 4L, 6L, 3L, 5L, 7L, 1L))
    image$reshape(c(f_tokens * h_tokens * w_tokens, pf * p * p * c_in))
}

#' Unpatchify Z-Image tokens back to a latent image
#'
#' Takes the first F/pF * H/p * W/p tokens (the image span of the
#' unified sequence) and reassembles [C, F, H, W], matching unpatchify.
#'
#' @param tokens Tensor of shape [S, pF * p * p * C] with the image
#'   tokens first.
#' @param size Integer vector c(F, H, W) of the target latent size.
#' @param patch_size Integer spatial patch size. Default 2.
#' @param f_patch_size Integer temporal patch size. Default 1.
#' @param out_channels Integer number of latent channels. Default 16.
#'
#' @return Tensor of shape [C, F, H, W].
#'
#' @export
zimage_unpatchify <- function(tokens, size, patch_size = 2L,
                              f_patch_size = 1L, out_channels = 16L) {
    p <- patch_size
    pf <- f_patch_size
    f_tokens <- size[1] %/% pf
    h_tokens <- size[2] %/% p
    w_tokens <- size[3] %/% p
    ori_len <- f_tokens * h_tokens * w_tokens
    x <- tokens[1:ori_len,]
    x <- x$view(c(f_tokens, h_tokens, w_tokens, pf, p, p, out_channels))
    # Python permute(6, 0, 3, 1, 4, 2, 5) -> R 1-indexed
    x <- x$permute(c(7L, 1L, 4L, 2L, 5L, 3L, 6L))
    x$reshape(c(out_channels, size[1], size[2], size[3]))
}

# Parity tests for the Z-Image position ids, 3-axis RoPE, patchify chain,
# timestep sinusoid, and static-shift scheduler against diffusers
# reference fixtures (generated by tools/gen_fixtures_zimage.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "rope_zimage.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/rope_zimage.safetensors"
if (!file.exists(fixture_path)) exit_file("zimage rope fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# --- padding lengths ------------------------------------------------------------

pad_len <- diffuseR:::zimage_pad_len
expect_equal(pad_len(37L), 27L)
expect_equal(pad_len(60L), 4L)
expect_equal(pad_len(64L), 0L)
expect_equal(pad_len(4096L), 0L)

# --- position ids ---------------------------------------------------------------

# Caption: 37 tokens pad to 64; ramp 1..64 on axis 1 built over the
# padded length
cap_ids <- zimage_cap_pos_ids(37L + pad_len(37L))
expect_equal(cap_ids$shape, c(64L, 3L))
expect_true(max_abs_diff(cap_ids, fx$cap_pos_ids) == 0)

# Image: 6x10 grid = 60 tokens pad to 64; axis-1 start just past the
# padded caption, pads at the origin
img_ids <- zimage_img_pos_ids(6L, 10L, start0 = 65L)
expect_equal(img_ids$shape, c(64L, 3L))
expect_true(max_abs_diff(img_ids, fx$img_pos_ids) == 0)

# --- RoPE frequencies (theta 256, f32-angle cast) --------------------------------

freqs <- zimage_pos_embed(fx$rope_ids)
expect_equal(freqs[[1]]$shape, c(9L, 128L))
expect_true(max_abs_diff(freqs[[1]], fx$rope_cos) < 1e-6)
expect_true(max_abs_diff(freqs[[2]], fx$rope_sin) < 1e-6)

# Position (0,0,0) is the identity rotation
expect_true(max_abs_diff(freqs[[1]][9, ],
  torch::torch_ones(128L)) == 0)
expect_true(max_abs_diff(freqs[[2]][9, ],
  torch::torch_zeros(128L)) == 0)

# --- RoPE application (reuses the FLUX interleaved apply) -------------------------

# Fixture x is [B, S, H, D] (processor layout); flux_apply_rotary_emb
# takes [B, H, S, D]
x_bhsd <- fx$rope_x$permute(c(1L, 3L, 2L, 4L))
out <- flux_apply_rotary_emb(x_bhsd, freqs)
expect_true(max_abs_diff(out, fx$rope_out$permute(c(1L, 3L, 2L, 4L))) < 1e-5)

# --- patchify / unpatchify --------------------------------------------------------

tokens <- zimage_patchify(fx$img)
expect_equal(tokens$shape, c(60L, 64L))
expect_true(max_abs_diff(tokens, fx$img_tokens_padded[1:60, ]) == 0)

unpat <- zimage_unpatchify(fx$unpat_tokens, c(1L, 12L, 20L))
expect_equal(unpat$shape, c(16L, 1L, 12L, 20L))
expect_true(max_abs_diff(unpat, fx$unpat_out) == 0)

# Roundtrip identity
rt <- zimage_unpatchify(zimage_patchify(fx$img), c(1L, 12L, 20L))
expect_true(max_abs_diff(rt, fx$img) == 0)

# --- timestep sinusoid -------------------------------------------------------------

t_emb <- ltx23_get_timestep_embedding(fx$t_emb_in, 256L,
  flip_sin_to_cos = TRUE, downscale_freq_shift = 0)
# 1e-4, not 1e-5: sin(t * exp(...)) at t ~ 1000 amplifies a 1-ULP
# libtorch difference in exp() to ~6e-5 (CRAN torch vs the fixture's
# libtorch); flag/math bugs are O(1) so the assertion still catches them
expect_true(max_abs_diff(t_emb, fx$t_emb_out) < 1e-4)

# --- scheduler: static shift 3.0 on linspace(1, 1/N, N) ---------------------------

for (n in c(4L, 8L)) {
  sched <- flowmatch_scheduler_create(shift = 3.0, use_dynamic_shifting = FALSE)
  sched <- flowmatch_set_timesteps(sched, num_inference_steps = n,
    sigmas = seq(1, 1 / n, length.out = n))
  expect_true(max_abs_diff(sched$sigmas,
    fx[[paste0("sched_sigmas_", n)]]) < 1e-6)
  expect_true(max_abs_diff(sched$timesteps,
    fx[[paste0("sched_timesteps_", n)]]) < 1e-3)
}

#' JIT-Traced Decode for the LTX-2.3 VAEs and Vocoder
#'
#' The video/audio decoders and the vocoder are static feed-forward
#' graphs, so \code{torch::jit_trace} converts them wholesale: one
#' R-to-libtorch crossing per forward, intermediates freed eagerly by
#' libtorch instead of accumulating as R handles until gc. Traces are
#' shape-specialized (runtime sizes bake into the graph as constants;
#' a mismatched input errors), so they are cached per instance, input
#' shape, dtype, device, and call tag, and re-traced on a miss.
#'
#' A trace captures the module's weight tensors, which would pin them
#' on the GPU across phase offloads; the pipeline releases all traces
#' whenever a component offloads (\code{.ltx23_release_vae_traces}).
#'
#' Tracing hazard on this lantern build: if the allocator callback
#' runs R's gc \emph{during} trace recording (memory pressure), the
#' recorded graph can capture garbage argument values (observed as
#' corrupted \code{narrow} starts on the full-size decoder; verified
#' 5/5 clean once gc cannot fire mid-trace). Defenses, in order: a
#' gc + cache flush right before each trace so pressure starts near
#' zero, \code{tryCatch} around trace and replay, and a one-time
#' validation of every fresh trace against the eager output — any
#' mismatch permanently blacklists that shape and runs eager. With
#' those in place the traced path cannot corrupt output — but per
#' render it measured slower than eager (traces are released on phase
#' offload, so every render re-pays trace + validation), so it stays
#' opt-in: \code{options(diffuseR.jit_vae = TRUE)}.
#'
#' @name jit_vae_ltx23
NULL

.ltx23_vae_traces <- new.env(parent = emptyenv())

#' Run a module forward through a shape-specialized trace
#'
#' @param module The nn_module (identity for the cache key; also the
#'   default callable).
#' @param x Input tensor.
#' @param forward Optional closure wrapping the call (for extra fixed
#'   arguments like \code{causal}); must be pure in \code{x}.
#' @param tag Character. Distinguishes call variants of one module.
#'
#' @return The forward result.
#'
#' @keywords internal
.ltx23_traced_call <- function(module, x, forward = NULL, tag = "") {
    run <- forward %||% module
    if (!isTRUE(getOption("diffuseR.jit_vae", FALSE))) {
        return(run(x))
    }
    key <- paste(format(environment(module)), tag,
                 paste(as.integer(x$shape), collapse = "x"), x$dtype$.type(),
                 x$device$type, x$device$index %||% 0L, sep = "|")
    tr <- .ltx23_vae_traces[[key]]
    if (isFALSE(tr)) {
        # This shape failed tracing or validation before: eager only
        return(run(x))
    }
    if (is.null(tr)) {
        # Start the trace from minimal memory pressure so the
        # allocator callback cannot run R gc mid-recording (which
        # corrupts captured argument values on this lantern build)
        gc(verbose = FALSE)
        if (x$device$type == "cuda") {
            torch::cuda_empty_cache()
        }
        # Tracing captures parameters as constants, which requires
        # grad-free tensors; these modules are inference-only. Always
        # trace through a plain closure: jit_trace treats a bare
        # nn_module via a separate (and here broken) code path.
        for (p in module$parameters) {
            p$requires_grad_(FALSE)
        }
        out <- tryCatch({
            tr <- torch::jit_trace(function(z) run(z), x)
            replay <- tr(x)
            ref <- run(x)
            d <- (replay$to(dtype = torch::torch_float32()) -
                        ref$to(dtype = torch::torch_float32()))$abs()$max()
            if (as.numeric(d) > 0) {
                warning("diffuseR: traced decode failed validation for ",
                        "shape [", paste(as.integer(x$shape), collapse = ", "),
                        "]; using the eager module for it", call. = FALSE)
                .ltx23_vae_traces[[key]] <- FALSE
                ref
            } else {
                .ltx23_vae_traces[[key]] <- tr
                replay
            }
        }, error = function(e) {
            .ltx23_vae_traces[[key]] <- FALSE
            run(x)
        })
        return(out)
    }
    tr(x)
}

#' Release all cached decode traces
#'
#' Traces hold references to the weight tensors they captured; drop
#' them when a component leaves the GPU so its memory actually frees.
#'
#' @return Invisibly, NULL.
#'
#' @keywords internal
.ltx23_release_vae_traces <- function() {
    rm(list = ls(.ltx23_vae_traces), envir = .ltx23_vae_traces)
    invisible(NULL)
}

# One choke point for the three video-decoder call sites (direct,
# spatially tiled, temporally tiled)
.ltx23_decode_tile <- function(vae, x, causal) {
    dec <- vae$decoder
    .ltx23_traced_call(dec, x, forward = function(z) dec(z, causal = causal),
                       tag = paste0("causal:", format(causal)))
}

#' Image-to-Image Generation with Stable Diffusion
#'
#' This function generates an image based on an input image and a text prompt using the Stable Diffusion model.
#' It allows for various configurations such as model name, device, scheduler, and more.
#' @param input_image Path to the input image or a tensor representing the image.
#' @param prompt Text prompt to guide the image generation.
#' @param negative_prompt Optional negative prompt to guide the image generation.`
#' @param img_dim Dimension of the output image (default: 512).
#' @param model_name Name of the Stable Diffusion model to use (default: "sd21").
#' @param pipeline Optional pre-loaded pipeline. If `NULL`, it will be loaded based on `model_name`.
#' @param devices A named list of devices for each model component (e.g., `list(unet = "cuda", decoder = "cpu", text_encoder = "cpu", encoder = "cpu")`).
#' @param unet_dtype_str Optional A character for dtype of the unet component (typically "torch_float16" for cuda and "torch_float32" for cpu).
#' @param download_models Logical indicating whether to download models if not found (default: FALSE).
#' @param scheduler Scheduler to use for the diffusion process (default: "ddim").
#' @param num_inference_steps Number of diffusion steps (default: 50).
#' @param strength Strength of the image-to-image transformation (default: 0.8).
#' @param guidance_scale Scale for classifier-free guidance (default: 7.5).
#' @param seed Random seed for reproducibility (default: NULL).
#' @param save_file Logical indicating whether to save the generated image.
#' @param filename Optional filename for saving the image. If `NULL`, a default name is generated.
#' @param metadata_path Path to save metadata (default: NULL).
#' @param use_native_decoder Logical; if TRUE, uses native R torch decoder instead of TorchScript.
#'   Native decoder has better GPU compatibility (especially Blackwell).
#' @param use_native_text_encoder Logical; if TRUE, uses native R torch text encoder instead of TorchScript.
#'   Native text encoder has better GPU compatibility (especially Blackwell).
#' @param use_native_unet Logical; if TRUE, uses native R torch UNet instead of TorchScript.
#'   Native UNet has better GPU compatibility (especially Blackwell).
#' @param ... Additional arguments for future use.
#' @return An image array and metadata
#' @export

img2img <- function(input_image, prompt, negative_prompt = NULL,
                    img_dim = 512, model_name = c("sd21", "sdxl"),
                    pipeline = NULL, devices = "auto",
                    unet_dtype_str = "float16", download_models = FALSE,
                    scheduler = "ddim", num_inference_steps = 50,
                    strength = 0.8, guidance_scale = 7.5, seed = NULL,
                    save_file = TRUE, filename = NULL, metadata_path = NULL,
                    use_native_decoder = FALSE,
                    use_native_text_encoder = FALSE, use_native_unet = FALSE,
                    ...) {
    if (model_name %in% c("sd21", "sdxl")) {
        num_train_timesteps <- 1000
    } else {
        stop("Model not supported")
    }

    # Handle "auto" devices
    if (identical(devices, "auto")) {
        devices <- auto_devices(model_name)
    }

    # 1. Get models
    m2d <- models2devices(model_name, devices = devices,
                          unet_dtype_str = NULL,
                          download_models = download_models)
    devices <- m2d$devices
    unet_dtype <- m2d$unet_dtype
    device_cpu <- m2d$device_cpu
    device_cuda <- m2d$device_cuda

    if (is.null(pipeline)) {
        pipeline <- load_pipeline(model_name = model_name, m2d = m2d,
                                  i2i = TRUE,
                                  unet_dtype_str = unet_dtype_str,
                                  use_native_decoder = use_native_decoder,
                                  use_native_text_encoder = use_native_text_encoder,
                                  use_native_unet = use_native_unet)
    }

    # 2. Encode input image to latents
    image_tensor <- preprocess_image(input_image, width = img_dim, height = img_dim,
                                     device = torch::torch_device(devices$encoder)) # Resize & normalize
    message("Encoding image...")
    encoded <- pipeline$encoder(image_tensor)
    # message("Loading quant_conv...")
    conv_latents <- quant_conv(encoded, dtype = unet_dtype,
                               device = devices$unet)

    latents_mean <- conv_latents[, 1:4,,] # First 4 channels
    latents_log_var <- conv_latents[, 5:8,,] # Last 4 channels
    init_latents <- latents_mean$to(dtype = unet_dtype,
                                    device = torch::torch_device(devices$unet)) * 0.18215
    # Need to FIX
    # Sample from the distribution (reparameterization trick)
    # if(eps > 0){
    #   std <- torch_exp(0.5 * latents_log_var)
    #   eps <- torch_randn_like(std)
    #   sampled_latents <- mean + eps * std
    # }

    # 3. Compute noise timestep from strength
    t_strength <- as.integer(strength * num_train_timesteps)
    schedule <- ddim_scheduler_create(num_train_timesteps = 1000,
                                      num_inference_steps = num_inference_steps,
                                      beta_schedule = "scaled_linear",
                                      device = torch::torch_device(devices$unet))

    all_inference_timesteps <- schedule$timesteps
    timestep_idx <- which.min(abs(all_inference_timesteps - t_strength))
    timestep_start <- all_inference_timesteps[timestep_idx]
    timesteps <- all_inference_timesteps[timestep_idx:length(all_inference_timesteps)]

    # 4. Add noise to latents
    message("Adding noise to latent image...")
    if (!is.null(seed)) {
        set.seed(seed)
    }
    noised_latents <- scheduler_add_noise(original_latents = init_latents,
        noise = torch::torch_randn_like(init_latents),
        timestep = timestep_start,
        scheduler_obj = schedule)
    noised_latents <- noised_latents$to(dtype = unet_dtype,
                                        device = torch::torch_device(devices$unet))

    txt2img(
            prompt = prompt,
            negative_prompt = negative_prompt,
            img_dim = img_dim,
            model_name = model_name,
            pipeline = pipeline,
            devices = devices,
            unet_dtype_str = unet_dtype_str,
            scheduler = "ddim",
            timesteps = timesteps,
            initial_latents = noised_latents,
            num_inference_steps = num_inference_steps,
            guidance_scale = guidance_scale,
            seed = seed,
            save_file = save_file,
            filename = filename,
            metadata_path = metadata_path
    )
}

#' Preprocess image for Stable Diffusion
#'
#' @param input File path to .jpg or .png, or a 3D array
#' @param device Target device for torch ("cpu" or "cuda")
#' @param width Desired width of the output image
#' @param height Desired height of the output image
#'
#' @return Torch tensor of shape c(1, 3, 512, 512), scaled to c(-1, 1)
#' @export
preprocess_image <- function(input, device = "cpu", width = 512, height = 512) {
    # Load JPEG/PNG if path
    if (is.character(input)) {
        if (grepl("\\.jpg$|\\.jpeg$", input, ignore.case = TRUE)) {
            img <- jpeg::readJPEG(input)
        } else if (grepl("\\.png$", input, ignore.case = TRUE)) {
            img <- png::readPNG(input)
        } else {
            stop("Unsupported image format: only .jpg/.jpeg/.png allowed")
        }
    } else if (is.array(input)) {
        img <- input
    } else {
        stop("Input must be a file path or 3D array")
    }

    # Handle alpha channel if present (RGBA -> RGB)
    if (length(dim(img)) == 3 && dim(img)[3] == 4) {
        img <- img[,, 1:3] # Drop alpha channel
    }

    # Handle grayscale images (add channels dimension)
    if (length(dim(img)) == 2) {
        img <- array(img, dim = c(dim(img), 1))
        # Optionally repeat to make it RGB:
        img <- array(rep(img, 3), dim = c(dim(img)[1], dim(img)[2], 3))
    }

    # Convert to torch tensor
    img_tensor <- torch::torch_tensor(img)

    # Add batch dimension and rearrange to [batch, channel, height, width]
    if (length(dim(img_tensor)) == 3) {
        # [H, W, C] -> [1, C, H, W]
        img_tensor <- img_tensor$permute(c(3, 1, 2))$unsqueeze(1)
    } else if (length(dim(img_tensor)) == 4) {
        # Assume [B, H, W, C] -> [B, C, H, W]
        img_tensor <- img_tensor$permute(c(1, 4, 2, 3))
    }

    # Use torch's interpolation for resizing
    img_tensor <- torch::nnf_interpolate(img_tensor, size = c(height, width),
        mode = "bilinear", align_corners = FALSE) # Usually FALSE for stable diffusion

    # Convert to float and normalize to [-1, 1]
    img_tensor <- img_tensor$to(dtype = torch::torch_float())$to(device = device)
    img_tensor <- (img_tensor * 2) - 1

    return(img_tensor)
}

#' Save and Display an Image from a Torch Tensor
#'
#' Converts a Torch tensor to a normalized RGB image array, saves it as a PNG file,
#' and optionally displays it in the RStudio Viewer pane using `grid::grid.raster()`.
#'
#' @param img A numeric with shape `[3, H, W]`.
#' @param save_to File path for the PNG image (default is `"output.png"`).
#' @param normalize Logical; whether to normalize pixel values to `[0, 1]`. Default is `TRUE`.
#'
#' @return Invisibly returns the saved file path.
#' @export
#'
#' @examples
#' img <- array(runif(32 * 32 * 3), dim = c(32, 32, 3))
#' out <- file.path(tempdir(), "sample.png")
#' save_image(img, out)
#' file.exists(out)
#' unlink(out)
save_image <- function(img, save_to = "output.png", normalize = TRUE) {
    # img_array <- tensor2image(img, normalize = normalize)
    dims <- dim(img)

    grDevices::png(filename = save_to, width = dims[2], height = dims[1])
    grid::grid.raster(img)
    grDevices::dev.off()

    if (interactive()) {
        grid::grid.raster(img)
    }

    cat("Image saved to", save_to, "\n")
    invisible(save_to)
}

#!/usr/bin/env r
#
# Quantize LTX-2 Transformer and VAE to INT4 format
#
# Usage:
#   r inst/scripts/quantize_ltx2.R
#   r inst/scripts/quantize_ltx2.R --transformer-only
#   r inst/scripts/quantize_ltx2.R --vae-only
#   r inst/scripts/quantize_ltx2.R --force  # Re-quantize even if cached
#
# Output:
#   ~/.cache/diffuseR/ltx2_transformer_int4-NNNNN-of-NNNNN.safetensors
#   ~/.cache/diffuseR/ltx2_vae_int4.safetensors
#
# Requirements:
#   - LTX-2 model downloaded via: huggingface-cli download Lightricks/LTX-2
#   - ~40GB disk space for processing (final output ~11GB)
#

library(torch)

# Parse args
args <- commandArgs(trailingOnly = TRUE)
do_transformer <- !("--vae-only" %in% args)
do_vae <- !("--transformer-only" %in% args)
force <- "--force" %in% args

# Load diffuseR (installed or via rhydrogen)
if (requireNamespace("diffuseR", quietly = TRUE)) {
  library(diffuseR)
} else {
  # Development mode - load from source
  if (file.exists("DESCRIPTION")) {
    rhydrogen::load_all()
  } else {
    stop("diffuseR not installed. Run from package root or install first.")
  }
}

cat("=== LTX-2 INT4 Quantization ===\n\n")

# Quantize transformer
if (do_transformer) {
  cat("--- Transformer ---\n")
  t0 <- Sys.time()
  transformer_path <- quantize_ltx2_transformer(force = force, verbose = TRUE)
  elapsed <- as.numeric(Sys.time() - t0, units = "mins")
  cat(sprintf("Transformer quantization complete in %.1f minutes\n", elapsed))
  cat(sprintf("Output: %s\n\n", transformer_path))
}

# Quantize VAE
if (do_vae) {
  cat("--- VAE ---\n")
  t0 <- Sys.time()
  vae_path <- quantize_ltx2_vae(force = force, verbose = TRUE)
  elapsed <- as.numeric(Sys.time() - t0, units = "secs")
  cat(sprintf("VAE quantization complete in %.1f seconds\n", elapsed))
  cat(sprintf("Output: %s\n\n", vae_path))
}

# Summary
cat("=== Summary ===\n")
cache_dir <- path.expand("~/.cache/diffuseR")
int4_files <- list.files(cache_dir, pattern = "int4.*\\.safetensors$", full.names = TRUE)
if (length(int4_files) > 0) {
  total_gb <- sum(file.info(int4_files)$size) / 1e9
  cat(sprintf("Cached INT4 weights: %.2f GB total\n", total_gb))
  for (f in int4_files) {
    cat(sprintf("  %s: %.2f GB\n", basename(f), file.info(f)$size / 1e9))
  }
}

cat("\nTo load these weights in R:\n")
cat("  q <- load_int4_weights('~/.cache/diffuseR/ltx2_transformer_int4.safetensors')\n")
cat("  w <- dequantize_int4(q[['transformer_blocks.0.attn1.to_q.weight']], device = 'cuda')\n")

#' Recommend a precision and device configuration for a model
#'
#' One VRAM-and-capability-aware recommendation for every diffuseR
#' model. The policy:
#'
#' \itemize{
#'   \item nf4 is the default tier for the quantized families (flux1,
#'     flux2, zimage, ltx). Its weights are packed uint8 plus float32
#'     blocks in sub-2 GB shards, which every safetensors reads, so it
#'     always loads. The SD models ship no quantized weights; their
#'     floor is fp16 with placement varying by VRAM.
#'   \item When the card has room for a higher-quality tier (fp8 or
#'     bf16) AND the installed safetensors can \emph{read} that dtype
#'     (\code{\link{.st_can_read}}), that tier is recommended instead.
#'   \item When the card has room but safetensors cannot read the tier,
#'     nf4 is recommended and the fork suggestion is surfaced in
#'     \code{note} (never an error).
#' }
#'
#' This is the policy engine; it does no disk I/O and does not know which
#' artifacts are built. Loaders reconcile the recommendation with what is
#' on disk (see \code{\link{flux_load_pipeline}}). Thresholds are
#' validated on an RTX 5060 Ti (16 GB) and are deliberately conservative
#' elsewhere. Video sizing for \code{"ltx"} is coarse here; the LTX
#' pipeline uses \code{\link{ltx23_memory_profile}} for frame-aware
#' placement.
#'
#' The pinning decision: phase-swapped weights are page-locked host
#' copies (see \code{\link{staging}}) that transfer at DMA rate -
#' but pinned pages are unswappable, so on small-RAM machines they turn
#' memory pressure into OOM kills. \code{pin} is TRUE when available
#' host RAM covers the model's pinned set twice over, FALSE below that,
#' FALSE on the cpu tier (nothing stages), and TRUE when RAM cannot be
#' detected (page-locking already fails soft per component). The LTX
#' pipeline, the Gemma3 encoder, and the FLUX-family image loaders
#' (flux1, flux2, zimage) take \code{pin} arguments and stage pinned
#' weights (see \code{\link{staging}}); the SD-family loaders place
#' components statically and do not phase-swap, so \code{pin} is inert
#' for them. \code{options(diffuseR.pin_staging)} is the global switch.
#'
#' @param model "sd21", "sdxl", "flux1", "flux2", "zimage", or "ltx".
#' @param vram_gb Numeric or NULL. Free VRAM in GB; auto-detected via
#'   nvidia-smi when NULL.
#' @param st_caps NULL or a named logical list with \code{bfloat16}
#'   and/or \code{float8_e4m3fn} - the safetensors READ capabilities.
#'   NULL probes the installed safetensors.
#' @param host_ram_gb Numeric or NULL. Available host RAM in GB;
#'   auto-detected (Linux \code{MemAvailable}) when NULL, NA where
#'   undetectable.
#'
#' @return A list with \code{model}, \code{precision}, \code{devices}
#'   (named component -> device map), \code{offload} (phase-offloading
#'   logical), \code{max_pixels}, \code{text_device}, \code{attn_chunk},
#'   \code{vram_gb}, \code{pin} (page-lock the phase-swapped host
#'   copies), \code{pinned_set_gb} (estimated pinned bytes),
#'   \code{host_ram_gb}, \code{fork_suggested} (logical), and
#'   \code{note} (the fork suggestion string, or NULL).
#'
#' @export
#'
#' @examples
#' # Stating vram_gb and st_caps makes the policy deterministic: no GPU
#' # and no installed safetensors needed.
#' r <- recommend("flux1", vram_gb = 16,
#'                st_caps = list(bfloat16 = TRUE, float8_e4m3fn = FALSE))
#' r$precision       # "nf4": fp8 fits the card, but cannot be read
#' r$fork_suggested  # TRUE
#' cat(r$note)       # the fork-or-nf4 message
#'
#' # Same card, once safetensors can read float8
#' recommend("flux1", vram_gb = 16,
#'           st_caps = list(bfloat16 = TRUE, float8_e4m3fn = TRUE))$precision
#'
#' # Auto-detect VRAM and probe the installed safetensors
#' str(recommend("flux2"))
recommend <- function(model = c("sd21", "sdxl", "flux1", "flux2", "zimage",
                                "ltx"),
                      vram_gb = NULL, st_caps = NULL, host_ram_gb = NULL) {
    model <- match.arg(model)
    if (is.null(vram_gb)) {
        vram_gb <- .detect_vram(use_free = TRUE)
    }
    if (is.null(vram_gb) || is.na(vram_gb) || vram_gb < 0) {
        vram_gb <- 0
    }
    if (is.null(st_caps)) {
        st_caps <- list(bfloat16 = .st_can_read("bfloat16"),
                        float8_e4m3fn = .st_can_read("float8_e4m3fn"))
    }

    tiers <- .recommend_specs()[[model]]

    chosen <- NULL
    want <- NULL # first VRAM-eligible tier blocked by a missing read cap
    for (tier in tiers) {
        if (vram_gb < tier$min_vram) {
            next
        }
        need <- tier$needs
        if (is.null(need) || isTRUE(st_caps[[need]])) {
            chosen <- tier
            break
        }
        if (is.null(want)) {
            want <- tier
        }
    }
    if (is.null(chosen)) {
        chosen <- tiers[[length(tiers)]] # terminal cpu tier
    }

    fork <- !is.null(want) && !identical(want$precision, chosen$precision)

    if (is.null(host_ram_gb)) {
        host_ram_gb <- .detect_host_ram()
    }
    pinned_set <- .pinned_set_gb(model, chosen$precision)
    pin <- .pin_decision(model, chosen$precision, host_ram_gb,
                         cpu = isTRUE(chosen$cpu))

    list(
         model = model,
         precision = chosen$precision,
         devices = chosen$devices,
         offload = isTRUE(chosen$offload),
         max_pixels = chosen$max_pixels,
         text_device = chosen$text_device %||% "cpu",
         attn_chunk = chosen$attn_chunk,
         vram_gb = vram_gb,
         pin = pin,
         pinned_set_gb = pinned_set,
         host_ram_gb = host_ram_gb,
         fork_suggested = fork,
         note = if (fork) {
             .st_fork_note(want$precision)
         } else {
             .bf16_note(model, chosen$precision)
         }
    )
}

# bf16 is not a built artifact: it is the unquantized source the
# quantizers read. Recommending the tier without saying that leaves the
# user with a precision they cannot act on, so spell out the one call
# that makes it available.
.bf16_note <- function(model, precision) {
    if (!identical(precision, "bf16")) {
        return(NULL)
    }
    fn <- switch(model, flux1 = "download_flux1",
                 flux2 = "download_flux2_klein",
                 zimage = "download_zimage_turbo", NULL)
    if (is.null(fn)) {
        return(NULL)
    }
    sprintf(paste0("bf16 loads the unquantized source rather than a ",
                   "quantized artifact. Keep it with %s(quantize = FALSE), ",
                   "then load with precision = \"bf16\"."), fn)
}

# Available host RAM in GB (Linux MemAvailable); NA where undetectable
# (macOS, Windows). NA feeds a keep-pinning decision: page-locking
# already falls back silently per component, and platforms without
# /proc rarely pair with CUDA.
.detect_host_ram <- function() {
    if (!file.exists("/proc/meminfo")) {
        return(NA_real_)
    }
    tryCatch({
        line <- grep("^MemAvailable:", readLines("/proc/meminfo", n = 20L),
                     value = TRUE)
        if (!length(line)) {
            return(NA_real_)
        }
        kb <- as.numeric(strsplit(trimws(sub("^MemAvailable:", "", line)),
                                  "[[:space:]]+")[[1]][1])
        kb / 1024 ^ 2
    }, error = function(e) NA_real_)
}

# Host GB that pinned staging would page-lock for a model at a given
# precision: the quantized checkpoint plus the text encoder's host
# copy. Coarse artifact-scale estimates - the pin decision needs the
# order of magnitude, not the byte.
.pinned_set_gb <- function(model, precision) {
    sets <- list(
                 sd21 = c(fp16 = 3),
                 sdxl = c(fp16 = 8),
                 flux1 = c(bf16 = 34, fp8 = 22, nf4 = 17), # DiT + T5 bf16 host copy
                 flux2 = c(bf16 = 17, fp8 = 12, nf4 = 10), # DiT + Qwen3 bf16
                 zimage = c(bf16 = 21, fp8 = 14, nf4 = 12), # DiT + Qwen3 bf16
                 ltx = c(fp8 = 34, nf4 = 28) # checkpoint + Gemma3 NF4
    )
    v <- unname(sets[[model]][precision])
    if (length(v) != 1L || is.na(v)) {
        0
    } else {
        v
    }
}

# Host-RAM-aware pinning decision for a model at a resolved precision.
# TRUE when available host RAM covers the pinned set twice over; FALSE
# on the cpu tier (nothing stages, so nothing to page-lock); TRUE when
# RAM is undetectable (page-locking already fails soft per component).
# Shared by recommend() and the phase-offloading loaders.
.pin_decision <- function(model, precision, host_ram_gb = NULL, cpu = FALSE) {
    if (isTRUE(cpu)) {
        return(FALSE)
    }
    if (is.null(host_ram_gb)) {
        host_ram_gb <- .detect_host_ram()
    }
    if (is.na(host_ram_gb)) {
        return(TRUE)
    }
    host_ram_gb >= 2 * .pinned_set_gb(model, precision)
}

# Resolve a loader's pin argument. Precedence: an explicit pin= wins,
# then options(diffuseR.pin_staging) (read with no default so an unset
# option falls through), then the RAM-aware .pin_decision. A "full"
# checkpoint is raw bf16, so it maps to the "bf16" pinned-set row.
.resolve_pin <- function(pin, model, format) {
    if (!is.null(pin)) {
        return(isTRUE(pin))
    }
    opt <- getOption("diffuseR.pin_staging")
    if (!is.null(opt)) {
        return(isTRUE(opt))
    }
    if (identical(format, "full")) {
        precision <- "bf16"
    } else {
        precision <- format
    }
    .pin_decision(model, precision)
}

# flux-family component placement: the big DiT and the VAE compute on
# the GPU (or all on CPU for the cpu tier); the text encoder is resident
# on the CPU and phase-onloaded during its own phase.
.dev_flux <- function(gpu = TRUE) {
    if (gpu) {
        list(transformer = "cuda", text = "cpu", vae = "cuda")
    } else {
        list(transformer = "cpu", text = "cpu", vae = "cpu")
    }
}

# One bf16/fp8/nf4/nf4-tight/cpu ladder for the flux-family image models.
# Precision rises with VRAM (nf4 default, fp8/bf16 as upgrades) - the
# inverse of the old flux_memory_profile, which had fp8 in a narrow
# low-VRAM band it can no longer fit now that fp8 is GPU-resident.
#
# text_device is the phase where the encoder computes: "cuda" onloads it
# to the GPU for the encode (fast, pinned), "cpu" keeps it resident on
# the host. flux1's T5-XXL needs ~9.8 GB for its encode phase, so it
# only GPU-encodes above gpu_encode_vram; the small flux2/zimage Qwen3
# encodes on the GPU on every GPU tier (gpu_encode_vram = 0).
.flux_family_tiers <- function(bf16_vram, fp8_vram, nf4_vram, nf4_tight_vram,
                               max_hi, max_mid, max_lo, max_cpu,
                               attn_tight = NULL, gpu_encode_vram = 0) {
    gpu_text <- function(min_vram) {
        if (min_vram >= gpu_encode_vram) {
            "cuda"
        } else {
            "cpu"
        }
    }
    list(
         list(precision = "bf16", min_vram = bf16_vram, needs = "bfloat16",
              devices = .dev_flux(TRUE), offload = TRUE, max_pixels = max_hi,
              text_device = gpu_text(bf16_vram), attn_chunk = NULL),
         list(precision = "fp8", min_vram = fp8_vram, needs = "float8_e4m3fn",
              devices = .dev_flux(TRUE), offload = TRUE, max_pixels = max_mid,
              text_device = gpu_text(fp8_vram), attn_chunk = NULL),
         list(precision = "nf4", min_vram = nf4_vram, needs = NULL,
              devices = .dev_flux(TRUE), offload = TRUE, max_pixels = max_mid,
              text_device = gpu_text(nf4_vram), attn_chunk = NULL),
         list(precision = "nf4", min_vram = nf4_tight_vram, needs = NULL,
              devices = .dev_flux(TRUE), offload = TRUE, max_pixels = max_lo,
              text_device = gpu_text(nf4_tight_vram), attn_chunk = attn_tight),
         list(precision = "nf4", min_vram = 0, needs = NULL, cpu = TRUE,
              devices = .dev_flux(FALSE), offload = FALSE, max_pixels = max_cpu,
              text_device = "cpu", attn_chunk = NULL)
    )
}

# SD-family ladder: the SD models ship no quantized weights, so every
# tier is fp16 - what varies is placement (all-GPU, UNet-only-GPU, or
# CPU) and the area budget. Device maps reuse the auto_devices
# strategy builder.
.sd_tiers <- function(model, fp16_vram, nf4_vram, max_fp16, max_nf4, max_cpu) {
    list(
         list(precision = "fp16", min_vram = fp16_vram, needs = NULL,
              devices = .build_fallback_devices(model, "full_gpu"),
              offload = FALSE, max_pixels = max_fp16),
         list(precision = "fp16", min_vram = nf4_vram, needs = NULL,
              devices = .build_fallback_devices(model, "unet_gpu"),
              offload = TRUE, max_pixels = max_nf4),
         list(precision = "fp16", min_vram = 0, needs = NULL, cpu = TRUE,
              devices = .build_fallback_devices(model, "cpu_only"),
              offload = FALSE, max_pixels = max_cpu)
    )
}

# Per-model tier ladders. Built lazily (device maps call helpers) so the
# thresholds live in one auditable place.
.recommend_specs <- function() {
    px <- function(n) as.integer(n) * as.integer(n)
    list(
         sd21 = .sd_tiers("sd21", fp16_vram = 6, nf4_vram = 3,
                          max_fp16 = px(1024), max_nf4 = px(768),
                          max_cpu = px(512)),
         sdxl = .sd_tiers("sdxl", fp16_vram = 12, nf4_vram = 6,
                          max_fp16 = px(1024), max_nf4 = px(1024),
                          max_cpu = px(768)),
         # 12B: nf4 peaks ~9.6 GB at 1024^2, fp8 ~12 GB resident,
         # bf16 needs a 24 GB card. attn-chunk the tight nf4 tier.
         flux1 = .flux_family_tiers(bf16_vram = 24, fp8_vram = 14,
                                    nf4_vram = 10, nf4_tight_vram = 8,
                                    max_hi = px(1536), max_mid = px(1024),
                                    max_lo = px(768), max_cpu = px(512),
                                    attn_tight = 2048L, gpu_encode_vram = 14),
         # 4B but activation-heavy: 1024^2 peaks ~12.5 GB regardless of
         # weight precision, so the 1024^2 tiers want ~13 GB free.
         flux2 = .flux_family_tiers(bf16_vram = 16, fp8_vram = 14,
                                    nf4_vram = 13, nf4_tight_vram = 8,
                                    max_hi = px(1024), max_mid = px(1024),
                                    max_lo = px(768), max_cpu = px(512)),
         # 6B: 1024^2 peaks ~13.1 GB, 512^2 ~half that.
         zimage = .flux_family_tiers(bf16_vram = 18, fp8_vram = 14,
                                     nf4_vram = 13, nf4_tight_vram = 8,
                                     max_hi = px(1024), max_mid = px(1024),
                                     max_lo = px(512), max_cpu = px(512)),
         # 22B video. fp8 is CPU-resident and streamed here (unlike the
         # image models); nf4 keeps the transformer resident. Coarse -
         # ltx23_memory_profile does the frame-aware sizing.
         ltx = list(
                    list(precision = "nf4", min_vram = 14, needs = NULL,
                         devices = .dev_flux(TRUE), offload = TRUE,
                         max_pixels = px(1280), text_device = "cpu",
                         attn_chunk = NULL),
                    list(precision = "fp8", min_vram = 10,
                         needs = "float8_e4m3fn", devices = .dev_flux(TRUE),
                         offload = TRUE, max_pixels = px(1024),
                         text_device = "cpu", attn_chunk = 4096L),
                    list(precision = "nf4", min_vram = 0, needs = NULL,
                         cpu = TRUE, devices = .dev_flux(FALSE),
                         offload = FALSE, max_pixels = px(512),
                         text_device = "cpu", attn_chunk = NULL)
        )
    )
}

# recommend(): VRAM + safetensors-read-capability policy, the read
# probe, and the flux_memory_profile adapter.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
library(diffuseR)

fork <- list(bfloat16 = TRUE, float8_e4m3fn = TRUE)   # can read everything
cran <- list(bfloat16 = TRUE, float8_e4m3fn = FALSE)  # reads bf16, not fp8

# --- structure --------------------------------------------------------------------

r <- recommend("flux1", vram_gb = 16, st_caps = fork)
expect_true(is.list(r))
expect_equal(r$model, "flux1")
expect_true(all(c("precision", "devices", "offload", "max_pixels",
  "text_device", "attn_chunk", "vram_gb", "fork_suggested", "note") %in%
  names(r)))
expect_true(is.integer(r$max_pixels) || is.numeric(r$max_pixels))

# --- floors: nf4 for the quantized families, fp16 for SD -------------------------

# No GPU: everything runs on CPU at its floor precision, no fork nag.
# SD ships no quantized weights, so its floor is fp16.
for (m in c("sd21", "sdxl", "flux1", "flux2", "zimage")) {
  z <- recommend(m, vram_gb = 0, st_caps = cran)
  expect_equal(z$precision, if (m %in% c("sd21", "sdxl")) "fp16" else "nf4")
  expect_false(z$fork_suggested)
  expect_true(all(unlist(z$devices) == "cpu"))
}

# Mid VRAM with no float8: nf4, and no nag when the card could not fit
# fp8 anyway (fp8 tier not VRAM-eligible).
expect_equal(recommend("flux1", 12, cran)$precision, "nf4")
expect_false(recommend("flux1", 12, cran)$fork_suggested)

# --- fp8 upgrade + fork gating ----------------------------------------------------

# 16 GB with float8 read: fp8 recommended, no nag.
f16 <- recommend("flux1", 16, fork)
expect_equal(f16$precision, "fp8")
expect_false(f16$fork_suggested)
expect_null(f16$note)
expect_equal(f16$devices$transformer, "cuda")

# 16 GB WITHOUT float8 read: fp8 wanted but unreadable -> nf4 + fork nag.
c16 <- recommend("flux1", 16, cran)
expect_equal(c16$precision, "nf4")
expect_true(c16$fork_suggested)
expect_true(is.character(c16$note) && grepl("safetensors#13", c16$note))
expect_false(grepl("—", c16$note))   # no em dash (house style)

# --- bf16 top tier (CRAN-readable, no fork needed) --------------------------------

# 24 GB reads bf16 on CRAN too -> bf16, no nag.
b24 <- recommend("flux1", 24, cran)
expect_equal(b24$precision, "bf16")
expect_false(b24$fork_suggested)

# But if bf16 read were somehow missing, it is gated like fp8.
nob <- recommend("flux1", 24, list(bfloat16 = FALSE, float8_e4m3fn = FALSE))
expect_true(nob$fork_suggested)
expect_equal(nob$precision, "nf4")

# --- SD ladder: fp16 for cards that fit, nf4 default, cpu -------------------------

expect_equal(recommend("sdxl", 16, cran)$precision, "fp16")
expect_equal(recommend("sdxl", 8, cran)$precision, "fp16")
expect_equal(recommend("sdxl", 16, cran)$devices$unet, "cuda")
expect_true("text_encoder2" %in% names(recommend("sdxl", 16, cran)$devices))
expect_false("text_encoder2" %in% names(recommend("sd21", 16, cran)$devices))
# SD is CRAN-readable at every tier: never a fork nag.
for (v in c(0, 4, 8, 16, 24)) expect_false(recommend("sdxl", v, cran)$fork_suggested)

# --- precision rises monotonically with VRAM (quality never drops) ----------------

rank <- c(nf4 = 1L, fp16 = 2L, fp8 = 3L, bf16 = 4L)
for (m in c("flux1", "flux2", "zimage")) {
  precs <- vapply(c(0, 8, 12, 16, 24),
    function(v) rank[[recommend(m, v, fork)$precision]], integer(1))
  expect_true(!is.unsorted(precs))
}

# --- read probe: override option wins, and self-consistency -----------------------

expect_true(is.logical(diffuseR:::.st_can_read("bfloat16")))
options(diffuseR.st_read_caps = list(bfloat16 = FALSE, float8_e4m3fn = FALSE))
expect_false(diffuseR:::.st_can_read("bfloat16"))
expect_false(diffuseR:::.st_can_read("float8_e4m3fn"))
# recommend() with st_caps = NULL now reads the (forced) probe.
expect_true(recommend("flux1", 24)$fork_suggested)
options(diffuseR.st_read_caps = NULL)

# On a machine whose safetensors CAN read a dtype, the hand-built probe
# file must decode (guards the writeBin byte patterns).
if (requireNamespace("safetensors", quietly = TRUE)) {
  # These reflect the installed safetensors; just assert they are logical
  # and, when TRUE, that recommend trusts them.
  expect_true(is.logical(diffuseR:::.st_can_read("float8_e4m3fn")))
}

# --- flux_memory_profile adapter follows recommend --------------------------------

options(diffuseR.st_read_caps = fork)
p <- flux_memory_profile(vram_gb = 16)
expect_equal(p$precision, "fp8")
expect_equal(p$phase_offload, TRUE)
expect_true(p$max_pixels >= 1024L * 1024L)
options(diffuseR.st_read_caps = cran)
p2 <- flux_memory_profile(vram_gb = 16)
expect_equal(p2$precision, "nf4")
expect_true(isTRUE(p2$fork_suggested))
options(diffuseR.st_read_caps = NULL)

# --- graceful fallback for an explicitly requested precision ----------------------

grc <- diffuseR:::.st_graceful_precision
# nf4/fp16 pass through untouched
expect_equal(grc("nf4", "write"), "nf4")
expect_equal(grc("fp16", "write"), "fp16")
# fp8 without float8 WRITE support -> nf4 + fork message
options(diffuseR.st_caps = list(float8_e4m3fn = FALSE))
expect_message(rr <- grc("fp8", "write"), pattern = "safetensors#13")
expect_equal(rr, "nf4")
options(diffuseR.st_caps = list(float8_e4m3fn = TRUE))
expect_equal(grc("fp8", "write"), "fp8")   # present -> passes through
options(diffuseR.st_caps = NULL)
# bf16 gate goes through the READ probe in read mode
options(diffuseR.st_read_caps = list(bfloat16 = FALSE))
expect_message(rr2 <- grc("bf16", "read"), pattern = "safetensors#11")
expect_equal(rr2, "nf4")
options(diffuseR.st_read_caps = NULL)

# --- fork note fit parameter ------------------------------------------------------

fn <- diffuseR:::.st_fork_note
expect_true(grepl("best fit for your card", fn("fp8", fit = TRUE)))
expect_false(grepl("best fit for your card", fn("fp8", fit = FALSE)))
expect_false(grepl("—", fn("fp8")))   # no em dash, either variant

# --- multi-GB read breadcrumb -----------------------------------------------------

msg <- diffuseR:::.st_overflow_message("shard-00001.safetensors", 3.4e9, "boom")
expect_true(grepl("3.4 GB", msg))
expect_true(grepl("2\\^31", msg))
expect_true(grepl("shard-00001", msg))

brc <- diffuseR:::.st_read_or_breadcrumb
# a read that succeeds is returned untouched
expect_equal(brc(function() 42L, NULL), 42L)
# a failure on a small (or absent) file rethrows verbatim, no breadcrumb
small <- tempfile()
writeLines("x", small)
e <- tryCatch(brc(function() stop("plain boom"), small),
  error = function(err) conditionMessage(err))
expect_true(grepl("plain boom", e))
expect_false(grepl("2\\^31", e))
unlink(small)

# --- pin decision -----------------------------------------------------------------

r <- recommend("ltx", vram_gb = 16, st_caps = cran, host_ram_gb = 125)
expect_true(all(c("pin", "pinned_set_gb", "host_ram_gb") %in% names(r)))
expect_true(r$pin)                   # 125 GB >> 2 x ~28 GB pinned set
expect_true(r$pinned_set_gb > 0)

expect_false(recommend("ltx", vram_gb = 16, st_caps = cran,
                       host_ram_gb = 32)$pin)   # 32 < 2 x pinned set

expect_false(recommend("ltx", vram_gb = 0, st_caps = cran,
                       host_ram_gb = 125)$pin)  # cpu tier: nothing stages

expect_true(recommend("sdxl", vram_gb = 16, st_caps = cran,
                      host_ram_gb = NA)$pin)    # undetectable: fail-soft on

hr <- diffuseR:::.detect_host_ram()
expect_true(is.na(hr) || (is.numeric(hr) && hr > 0))
expect_equal(diffuseR:::.pinned_set_gb("ltx", "bf16"), 0)  # unknown tier -> 0

# --- every recommended tier must be reachable -------------------------------------

# recommend() used to return "bf16" for big cards while no loader
# accepted it and no builder produced it, with note = NULL. bf16 is the
# unquantized source, so the tier is real; the note has to say so.
fork <- list(bfloat16 = TRUE, float8_e4m3fn = TRUE)
b <- recommend("flux2", vram_gb = 24, st_caps = fork)
expect_equal(b$precision, "bf16")
expect_true(is.character(b$note))
expect_true(grepl("download_flux2_klein", b$note))
expect_true(grepl("quantize = FALSE", b$note))
expect_false(b$fork_suggested)

# The flux-family loaders must accept it as a precision.
for (fn in c("flux2_load_pipeline", "zimage_load_pipeline")) {
  choices <- eval(formals(getExportedValue("diffuseR", fn))$precision)
  expect_true("bf16" %in% choices)
}

# A quantized tier still resolves to its artifact directory, and bf16
# resolves elsewhere (the hub cache), not to a "<prefix>bf16" dir.
expect_equal(diffuseR:::.flux_model_dir("flux2", "nf4", "/tmp/x-"), "/tmp/x-nf4")
expect_false(identical(
  tryCatch(diffuseR:::.flux_model_dir("flux2", "bf16", "/tmp/x-"),
           error = function(e) "errored"),
  "/tmp/x-bf16"))

# An unknown model has no bf16 source and must say so rather than
# silently building a bogus path.
expect_error(diffuseR:::.flux_source_dir("nosuch"), pattern = "No bf16 source")

# --- LTX's recommended tier must be downloadable ----------------------------------

# recommend("ltx") is nf4 on any card >= 14 GB, so download_ltx2() has
# to be able to build nf4, not just fp8.
expect_equal(recommend("ltx", vram_gb = 16)$precision, "nf4")
expect_true("precision" %in% names(formals(download_ltx2)))
expect_equal(eval(formals(download_ltx2)$precision), c("nf4", "fp8"))

# Flux-family pinned staging surface + gate logic. The CPU/no-torch
# parts (formals, .resolve_pin precedence, .flux_build_staging opt-out
# paths) run everywhere; the pin round trip is CUDA-gated below.

library(diffuseR)

# --- loader surface: pin + text_device knobs --------------------------------------

for (fn in list(flux_load_pipeline, flux2_load_pipeline,
                zimage_load_pipeline)) {
    expect_true("pin" %in% names(formals(fn)))
    expect_true("text_device" %in% names(formals(fn)))
}
# flux1's text_device default moved from "cpu" to NULL (profile-resolved)
expect_null(eval(formals(flux_load_pipeline)$text_device))

# --- .resolve_pin precedence: explicit > option > decision ------------------------

# Explicit argument wins over everything.
old <- getOption("diffuseR.pin_staging")
on.exit(options(diffuseR.pin_staging = old), add = TRUE)
options(diffuseR.pin_staging = FALSE)
expect_true(diffuseR:::.resolve_pin(TRUE, "flux1", "nf4"))
expect_false(diffuseR:::.resolve_pin(FALSE, "flux1", "nf4"))
# NULL falls through to the option.
expect_false(diffuseR:::.resolve_pin(NULL, "flux1", "nf4"))
options(diffuseR.pin_staging = TRUE)
expect_true(diffuseR:::.resolve_pin(NULL, "flux1", "nf4"))
# Unset option falls through to the RAM-aware decision (logical result).
options(diffuseR.pin_staging = NULL)
expect_true(is.logical(diffuseR:::.resolve_pin(NULL, "flux2", "fp8")))

# --- .pin_decision: cpu -> FALSE, NA RAM -> TRUE, else 2x rule ---------------------

expect_false(diffuseR:::.pin_decision("flux1", "nf4", host_ram_gb = 64,
                                      cpu = TRUE))
expect_true(diffuseR:::.pin_decision("flux1", "nf4", host_ram_gb = NA))
expect_true(diffuseR:::.pin_decision("flux1", "nf4", host_ram_gb = 128))
expect_false(diffuseR:::.pin_decision("flux1", "nf4", host_ram_gb = 8))

# --- .pinned_set_gb flux1 row: T5 host copy now bf16 (was fp32) --------------------

expect_equal(diffuseR:::.pinned_set_gb("flux1", "nf4"), 17)
expect_equal(diffuseR:::.pinned_set_gb("flux1", "fp8"), 22)
expect_equal(diffuseR:::.pinned_set_gb("flux1", "bf16"), 34)

# --- flux tier text_device: flux1 GPU-encodes only where T5 fits ------------------

# fp8/bf16 tiers (>=14 GB) onload T5; nf4 tiers keep CPU-fp32 encode.
#
# st_caps is pinned explicitly here. Left NULL, recommend() probes the
# *installed* safetensors, so the fp8 tier only exists on a machine whose
# build can read float8 -- which made this block pass on a dev box with
# the fork and fail on Windows R-devel, and would have failed on CRAN,
# where flux1 at 16 GB resolves to nf4 and its CPU encode instead.
st_fork <- list(bfloat16 = TRUE, float8_e4m3fn = TRUE)
st_cran <- list(bfloat16 = TRUE, float8_e4m3fn = FALSE)
expect_equal(recommend("flux1", vram_gb = 24, st_caps = st_fork)$text_device,
             "cuda")
expect_equal(recommend("flux1", vram_gb = 16, st_caps = st_fork)$text_device,
             "cuda") # fp8 tier
expect_equal(recommend("flux1", vram_gb = 16, st_caps = st_cran)$text_device,
             "cpu") # same card, stock safetensors: nf4 tier
expect_equal(recommend("flux1", vram_gb = 10, st_caps = st_fork)$text_device,
             "cpu") # nf4 tier
expect_equal(recommend("flux1", vram_gb = 0, st_caps = st_fork)$text_device,
             "cpu") # cpu tier
# The small flux2/zimage encoder rides the GPU on every GPU tier,
# whichever safetensors is installed.
expect_equal(recommend("flux2", vram_gb = 16, st_caps = st_fork)$text_device,
             "cuda")
expect_equal(recommend("flux2", vram_gb = 16, st_caps = st_cran)$text_device,
             "cuda")

# --- .flux_build_staging opt-out paths return NULL (no torch needed) --------------

dummy <- list(format = "nf4")
expect_null(diffuseR:::.flux_build_staging(dummy, pin = FALSE,
    phase_offload = TRUE, device = "cuda", components = "transformer"))
expect_null(diffuseR:::.flux_build_staging(dummy, pin = TRUE,
    phase_offload = FALSE, device = "cuda", components = "transformer"))
expect_null(diffuseR:::.flux_build_staging(dummy, pin = TRUE,
    phase_offload = TRUE, device = "cpu", components = "transformer"))

# --- fp8 field collector + pin round trip (torch / CUDA gated) --------------------

if (!requireNamespace("torch", quietly = TRUE) ||
    !torch::torch_is_installed()) {
    exit_file("torch not fully installed")
}

# .flux_fp8_collect walks children and returns weight_fp8 + weight_scale
# pairs. Build two fp8 linears under a parent and check the count.
parent <- torch::nn_module(
    "fp8_parent",
    initialize = function() {
        self$a <- ltx23_fp8_linear(4L, 4L, bias = FALSE)
        self$b <- ltx23_fp8_linear(4L, 4L, bias = FALSE)
    },
    forward = function(x) x
)()
w <- torch::torch_randn(4L, 4L)$to(dtype = torch::torch_float8_e4m3fn())
s <- torch::torch_tensor(0.5)
parent$a$set_fp8_weight(w, s)
parent$b$set_fp8_weight(w, s)
collected <- diffuseR:::.flux_fp8_collect(parent)
expect_equal(length(collected), 4L)   # 2 linears x (weight_fp8, weight_scale)

if (!torch::cuda_is_available()) {
    exit_file("no CUDA")
}

# .pin_component with extra fields pins params/buffers AND the fp8 fields;
# staged onload/offload moves them all and preserves the forward output.
m <- ltx23_feed_forward(16L)
m$eval()
extra_t <- torch::torch_randn(8L, 8L)$to(dtype = torch::torch_float8_e4m3fn())
scale0 <- torch::torch_tensor(2.0)   # 0-dim scale must pin too
st <- diffuseR:::.pin_component(m, extra = list(extra_t, scale0))
expect_false(is.null(st))
# params + buffers + 2 extras all became pinned live/pinned pairs
expect_true(length(st) >= length(c(m$parameters, m$buffers)) + 2L)
expect_true(suppressWarnings(
    st[[length(st)]]$live$is_pinned(device = torch::torch_device("cuda"))
))

x <- torch::torch_randn(2L, 5L, 16L)
torch::with_no_grad(ref <- m(x))
diffuseR:::.staged_onload(st, "cuda")
expect_equal(st[[1]]$live$device$type, "cuda")
torch::with_no_grad(out_gpu <- m(x$to(device = "cuda"))$cpu())
expect_true(as.numeric((out_gpu - ref)$abs()$max()) < 1e-5)
diffuseR:::.staged_offload(st)
expect_equal(st[[1]]$live$device$type, "cpu")
torch::with_no_grad(out_back <- m(x))
expect_true(as.numeric((out_back - ref)$abs()$max()) == 0)

#' LTX-2.3 Prefix Conditioning (image-to-video, video continuation)
#'
#' Fresh R port of the frame-conditioning mechanics from the diffusers
#' reference (Apache-2.0, pipelines/ltx2/pipeline_ltx2_image2video.py
#' and pipeline_ltx2_condition.py), restricted to prefix conditioning
#' at latent index 0 with strength 1: a single start image (i2v) or the
#' leading pixel frames of a previous clip (continuation). Conditioned
#' latent tokens are initialized from the VAE-encoded pixels, see a
#' per-token timestep of zero, and are frozen through the Euler loop.
#'
#' @name condition_ltx23
NULL

#' Preprocess an image (or frame stack) for VAE encoding
#'
#' Mirrors the diffusers VideoProcessor: bilinear resize so the shorter
#' relative side matches, center-crop to the exact target, and scale to
#' [-1, 1].
#'
#' @param x Path to a PNG/JPEG, or an array [H, W, 3] (values in
#'   [0, 1]), or a [F, H, W, 3] array of frames.
#' @param height,width Integers. Target size (multiples of 32).
#'
#' @return Float32 tensor [1, 3, F, height, width] in [-1, 1].
#'
#' @export
ltx23_preprocess_frames <- function(x, height, width) {
    if (is.character(x)) {
        x <- if (grepl("\\.png$", x, ignore.case = TRUE)) {
            png::readPNG(x)
        } else {
            jpeg::readJPEG(x)
        }
    }
    if (length(dim(x)) == 3L) {
        dim(x) <- c(1L, dim(x))
    }
    if (dim(x)[4] > 3L) {
        # Drop alpha
        x <- x[,,, 1:3, drop = FALSE]
    }
    frames <- torch::torch_tensor(x, dtype = torch::torch_float32())
    # [F, H, W, C] -> [F, C, H, W]
    frames <- frames$permute(c(1L, 4L, 2L, 3L))

    src_h <- frames$shape[3]
    src_w <- frames$shape[4]
    scale <- max(height / src_h, width / src_w)
    new_h <- as.integer(round(src_h * scale))
    new_w <- as.integer(round(src_w * scale))
    frames <- torch::nnf_interpolate(frames, size = c(new_h, new_w),
                                     mode = "bilinear", align_corners = FALSE)
    top <- (new_h - height) %/% 2L
    left <- (new_w - width) %/% 2L
    frames <- frames$narrow(3L, top + 1L, height)$narrow(4L, left + 1L, width)

    # [F, C, H, W] -> [1, C, F, H, W], [0, 1] -> [-1, 1]
    frames$permute(c(2L, 1L, 3L, 4L))$unsqueeze(1L)$mul(2)$sub(1)
}

#' Read the trailing frames of a video file
#'
#' Extracts the last \code{n} frames of an MP4 (via \code{av}) for use
#' as continuation conditioning.
#'
#' @param path Video file.
#' @param n Integer. Trailing frame count.
#'
#' @return Array [n, H, W, 3] in [0, 1].
#'
#' @export
ltx23_read_tail_frames <- function(path, n = 9L) {
    if (!requireNamespace("av", quietly = TRUE)) {
        stop("Reading video frames requires the 'av' package.")
    }
    info <- av::av_media_info(path)
    total <- info$video$frames
    if (is.null(total) || is.na(total)) {
        # Some containers omit the frame count; derive from duration
        total <- floor(info$duration * info$video$framerate)
    }
    out_dir <- tempfile("ltx23_tail_")
    dir.create(out_dir)
    on.exit(unlink(out_dir, recursive = TRUE), add = TRUE)
    av::av_video_images(path, destdir = out_dir, format = "png")
    files <- sort(list.files(out_dir, pattern = "\\.png$", full.names = TRUE))
    files <- utils::tail(files, n)
    frames <- lapply(files, png::readPNG)
    arr <- array(0, dim = c(length(frames), dim(frames[[1]])[1],
                            dim(frames[[1]])[2], 3L))
    for (i in seq_along(frames)) {
        f <- frames[[i]]
        arr[i,,,] <- f[,, 1:3]
    }
    arr
}

#' Encode pixel frames to normalized video latents
#'
#' VAE encode in "argmax" mode (the distribution mean), then normalize
#' with the checkpoint's per-channel statistics — the exact inverse of
#' the decode path.
#'
#' @param vae An \code{ltx23_video_vae}.
#' @param frames Tensor [1, 3, F, H, W] in [-1, 1] (see
#'   \code{\link{ltx23_preprocess_frames}}).
#'
#' @return Normalized latents [1, 128, F', H/32, W/32] (float32).
#'
#' @export
ltx23_encode_video_frames <- function(vae, frames) {
    enc_dtype <- vae$encoder$conv_in$conv$weight$dtype
    enc_device <- vae$encoder$conv_in$conv$weight$device
    moments <- torch::with_no_grad(
                                   vae$encode(frames$to(device = enc_device, dtype = enc_dtype))
    )
    latents <- moments$mean$to(dtype = torch::torch_float32())
    ltx23_normalize_latents(latents, vae$latents_mean$to(dtype = torch::torch_float32()),
                            vae$latents_std$to(dtype = torch::torch_float32()))
}

#' Slice the trailing latent frames of a generation for chaining
#'
#' Cuts the last \code{k} latent frames out of a result's video
#' latents, in the [1, 128, k, H', W'] layout that
#' \code{txt2vid_ltx2(condition_latents = )} consumes, so one chunk
#' can seed the next without leaving latent space: no decode, no
#' re-encode, no video round-trip.
#'
#' Semantics caveat: a latent frame sliced from inside a sequence
#' represents 8 pixel frames, while a fresh VAE encode of a k-frame
#' tail represents 1 + 8(k - 1) pixel frames with its first latent in
#' first-frame form. The frozen prefix the next generation sees is
#' therefore not identical to the pixel-path prefix; compare both on
#' real content before relying on latent-only joins.
#'
#' @param result A \code{\link{txt2vid_ltx2}} result list (uses its
#'   \code{latents} and \code{latent_shape}), or the packed latents
#'   tensor [1, S, 128] itself (then \code{latent_shape} is
#'   required).
#' @param k Integer. Trailing latent frames to keep (default 2 = the
#'   standard 9-pixel-frame conditioning prefix).
#' @param latent_shape Integer vector c(frames, height, width) of the
#'   latent geometry; only needed when \code{result} is a raw tensor.
#'
#' @return Normalized latents [1, 128, k, H', W'] (float32), ready
#'   for \code{txt2vid_ltx2(condition_latents = )}.
#'
#' @export
ltx23_tail_latents <- function(result, k = 2L, latent_shape = NULL) {
    if (is.list(result)) {
        latents <- result$latents
        latent_shape <- latent_shape %||% result$latent_shape
    } else {
        latents <- result
    }
    if (is.null(latent_shape)) {
        stop("latent_shape is required when result is a raw latents tensor")
    }
    lf <- latent_shape[1]
    if (k < 1L || k > lf) {
        stop("k must be between 1 and the latent frame count")
    }
    video <- ltx23_unpack_video_latents(latents, lf, latent_shape[2],
                                        latent_shape[3])
    video$narrow(3L, lf - k + 1L, k)
}

#' Build conditioned initial latents and the conditioning mask
#'
#' i2v (\code{cond_latents} has one latent frame): the encoded frame is
#' repeated across all latent frames and only latent frame 0 is marked
#' conditioned. Continuation (k latent frames): the prefix tokens are
#' replaced and marked. Unconditioned positions start as pure noise.
#'
#' @param cond_latents Normalized condition latents
#'   [1, 128, k, H', W'] from \code{\link{ltx23_encode_video_frames}}.
#' @param latent_frames,latent_height,latent_width Integers. Full
#'   latent geometry of the generation.
#' @param noise Tensor [1, 128, F', H', W'] of standard noise (caller
#'   provides so seeding stays in one place).
#' @param cond_noise_scale Numeric. Optional partial noising of the
#'   conditioned tokens (diffusers \code{noise_scale}, default 0).
#'
#' @return list(latents [1, S, 128] float32 packed,
#'   conditioning_mask [1, S] float32 packed).
#'
#' @export
ltx23_prepare_conditioned_latents <- function(cond_latents, latent_frames,
    latent_height, latent_width,
    noise, cond_noise_scale = 0) {
    k <- cond_latents$shape[3]
    stopifnot(k <= latent_frames, cond_latents$shape[4] == latent_height,
              cond_latents$shape[5] == latent_width)

    init <- if (k == 1L) {
        cond_latents$`repeat`(c(1L, 1L, latent_frames, 1L, 1L))
    } else {
        torch::torch_cat(list(
                              cond_latents,
                              noise$narrow(3L, k + 1L, latent_frames - k)
            ), dim = 3L)
    }

    mask <- torch::torch_zeros(1L, 1L, latent_frames, latent_height,
                               latent_width, dtype = torch::torch_float32(),
                               device = noise$device)
    mask$narrow(3L, 1L, k)$fill_(1)

    latents <- init$to(device = noise$device) * mask + noise * (1 - mask)
    if (cond_noise_scale > 0) {
        # Partially re-noise the conditioned tokens (reference
        # _create_noised_state with per-token scale)
        scale <- mask * cond_noise_scale
        latents <- noise * scale + latents * (1 - scale)
    }

    list(
         latents = ltx23_pack_video_latents(latents),
         conditioning_mask = ltx23_pack_video_latents(mask)$squeeze(3L)
    )
}

# Unit tests for the prefix-conditioning helpers (R/condition_ltx23.R):
# preprocessing geometry, i2v repeat semantics, prefix token placement,
# and the packed conditioning mask.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)
torch::torch_manual_seed(21)

# --- Preprocessing: resize + center-crop + range ------------------------------------

img <- array(runif(48 * 80 * 3), dim = c(48L, 80L, 3L))
fr <- ltx23_preprocess_frames(img, height = 32L, width = 32L)
expect_equal(as.integer(fr$shape), c(1L, 3L, 1L, 32L, 32L))
expect_true(as.numeric(fr$max()) <= 1)
expect_true(as.numeric(fr$min()) >= -1)

# Multi-frame input keeps frame count
vid <- array(runif(5 * 40 * 40 * 3), dim = c(5L, 40L, 40L, 3L))
frv <- ltx23_preprocess_frames(vid, height = 32L, width = 32L)
expect_equal(as.integer(frv$shape), c(1L, 3L, 5L, 32L, 32L))

# RGBA alpha is dropped
rgba <- array(runif(32 * 32 * 4), dim = c(32L, 32L, 4L))
fra <- ltx23_preprocess_frames(rgba, height = 32L, width = 32L)
expect_equal(as.integer(fra$shape)[2], 3L)

# A [0,1] gray image maps to ~0 in [-1,1]
gray <- array(0.5, dim = c(32L, 32L, 3L))
frg <- ltx23_preprocess_frames(gray, height = 32L, width = 32L)
expect_true(as.numeric(frg$abs()$max()) < 1e-6)

# --- Conditioned latent init + mask -------------------------------------------------

lf <- 4L; lh <- 2L; lw <- 3L
noise <- torch::torch_randn(1L, 128L, lf, lh, lw)

# i2v: one conditioning latent frame, repeated then masked to frame 0
cond1 <- torch::torch_randn(1L, 128L, 1L, lh, lw)
prep1 <- ltx23_prepare_conditioned_latents(cond1, lf, lh, lw, noise)
expect_equal(as.integer(prep1$latents$shape), c(1L, lf * lh * lw, 128L))
expect_equal(as.integer(prep1$conditioning_mask$shape), c(1L, lf * lh * lw))
n_tok <- lh * lw
m <- as.numeric(prep1$conditioning_mask[1, ])
expect_true(all(m[1:n_tok] == 1))
expect_true(all(m[(n_tok + 1):(lf * n_tok)] == 0))
# Conditioned tokens equal the packed condition; the rest equal noise
packed_cond <- diffuseR:::ltx23_pack_video_latents(cond1)
expect_true(as.numeric((prep1$latents$narrow(2L, 1L, n_tok) -
  packed_cond)$abs()$max()) < 1e-6)
packed_noise <- diffuseR:::ltx23_pack_video_latents(noise)
expect_true(as.numeric((prep1$latents$narrow(2L, n_tok + 1L, (lf - 1L) * n_tok) -
  packed_noise$narrow(2L, n_tok + 1L, (lf - 1L) * n_tok))$abs()$max()) < 1e-6)

# Continuation: two conditioning latent frames
cond2 <- torch::torch_randn(1L, 128L, 2L, lh, lw)
prep2 <- ltx23_prepare_conditioned_latents(cond2, lf, lh, lw, noise)
m2 <- as.numeric(prep2$conditioning_mask[1, ])
expect_true(all(m2[1:(2L * n_tok)] == 1))
expect_true(all(m2[(2L * n_tok + 1L):(lf * n_tok)] == 0))
expect_true(as.numeric((prep2$latents$narrow(2L, 1L, 2L * n_tok) -
  diffuseR:::ltx23_pack_video_latents(cond2))$abs()$max()) < 1e-6)

# cond_noise_scale partially noises ONLY the conditioned tokens
prep3 <- ltx23_prepare_conditioned_latents(cond1, lf, lh, lw, noise,
  cond_noise_scale = 0.5)
d_cond <- as.numeric((prep3$latents$narrow(2L, 1L, n_tok) -
  packed_cond)$abs()$max())
expect_true(d_cond > 1e-3)
d_free <- as.numeric((prep3$latents$narrow(2L, n_tok + 1L, (lf - 1L) * n_tok) -
  packed_noise$narrow(2L, n_tok + 1L, (lf - 1L) * n_tok))$abs()$max())
expect_true(d_free < 1e-6)

# Expected blend value: noise*0.5 + cond*0.5 on conditioned tokens
expected <- packed_noise$narrow(2L, 1L, n_tok)$mul(0.5) +
packed_cond$mul(0.5)
expect_true(as.numeric((prep3$latents$narrow(2L, 1L, n_tok) -
  expected)$abs()$max()) < 1e-6)

# --- Encoding: argmax + normalization round trip ------------------------------------

vae <- ltx23_video_vae(
  latent_channels = 4L,
  block_out_channels = c(8L, 8L, 8L, 8L),
  decoder_block_out_channels = c(4L, 8L, 8L, 16L),
  layers_per_block = c(1L, 1L, 1L, 1L, 1L),
  decoder_layers_per_block = c(1L, 1L, 1L, 1L, 1L)
)
vae$eval()
torch::with_no_grad({
  vae$latents_mean$copy_(torch::torch_randn(4L))
  vae$latents_std$copy_(torch::torch_rand(4L) + 0.5)
})
px <- torch::torch_rand(1L, 3L, 1L, 32L, 32L)$mul(2)$sub(1)
lat <- ltx23_encode_video_frames(vae, px)
expect_equal(as.integer(lat$shape), c(1L, 4L, 1L, 1L, 1L))
# Matches manual encode + normalize
torch::with_no_grad(ref_moments <- vae$encode(px))
ref <- ltx23_normalize_latents(ref_moments$mean, vae$latents_mean,
  vae$latents_std)
expect_true(as.numeric((lat - ref)$abs()$max()) < 1e-5)

#!/usr/bin/env r
#
# Full LTX-2 Pipeline Test with Real Text Encoding and VAE
#

library(torch)
library(diffuseR)

cat("=== LTX-2 Full Pipeline (Real Text + VAE) ===\n\n")

get_gpu_memory <- function() {
  out <- system("nvidia-smi --query-gpu=memory.used --format=csv,noheader,nounits", intern = TRUE)
  as.numeric(out[1]) / 1024
}

gc(full = TRUE)
torch::cuda_empty_cache()

# Test parameters
prompt <- "A Bigfoot dj transforming into a humanoid robot with lifelike realistic human facial features and hair with rubbery synthetic skin, uncanny valley style exaggerated features"
width <- 512L
height <- 320L
num_frames <- 17L
num_steps <- 25L  # Proper quality test

cat(sprintf("Prompt: %s\n", prompt))
cat(sprintf("Resolution: %dx%d, %d frames, %d steps\n\n", width, height, num_frames, num_steps))

# LTX-2 compression ratios
spatial_ratio <- 32L
temporal_ratio <- 8L
latent_height <- height %/% spatial_ratio
latent_width <- width %/% spatial_ratio
latent_frames <- (num_frames - 1L) %/% temporal_ratio + 1L

cat(sprintf("Latent dims: %dx%dx%d\n\n", latent_width, latent_height, latent_frames))

torch::with_no_grad({

  # ---- Step 1: Text Encoding with Gemma3 ----
  cat("=== Step 1: Text Encoding (Gemma3) ===\n")
  t0 <- Sys.time()

  # Find Gemma3 model
  gemma_path <- Sys.glob("~/.cache/huggingface/hub/models--google--gemma-3-12b-it/snapshots/*/")[1]
  if (is.na(gemma_path)) {
    stop("Gemma3 model not found. Download with: huggingface-cli download google/gemma-3-12b-it")
  }
  cat(sprintf("Using Gemma3 from: %s\n", gemma_path))

  text_result <- encode_text_ltx2(
    prompt = prompt,
    backend = "gemma3",
    model_path = gemma_path,
    tokenizer_path = gemma_path,
    max_sequence_length = 128L,
    caption_channels = 3840L,
    device = "cpu",
    dtype = torch::torch_float32()
  )

  # Negative prompt (empty)
  neg_result <- encode_text_ltx2(
    prompt = "",
    backend = "gemma3",
    model_path = gemma_path,
    tokenizer_path = gemma_path,
    max_sequence_length = 128L,
    caption_channels = 3840L,
    device = "cpu",
    dtype = torch::torch_float32()
  )

  text_time <- as.numeric(Sys.time() - t0)
  cat(sprintf("Text encoding: %.1f seconds\n", text_time))
  cat(sprintf("Prompt embeds shape: %s\n", paste(dim(text_result$prompt_embeds), collapse = "x")))

  # Save embeddings for future use
  embed_cache <- list(
    prompt_embeds = text_result$prompt_embeds,
    prompt_attention_mask = text_result$prompt_attention_mask,
    neg_prompt_embeds = neg_result$prompt_embeds,
    neg_attention_mask = neg_result$prompt_attention_mask,
    prompt = prompt
  )
  cache_path <- path.expand("~/.cache/diffuseR/text_embeds_test.rds")
  saveRDS(embed_cache, cache_path)
  cat(sprintf("Cached embeddings to: %s\n\n", cache_path))

  # ---- Step 2: Apply Connectors ----
  cat("=== Step 2: Text Connectors ===\n")
  t0 <- Sys.time()

  connector_path <- Sys.glob("~/.cache/huggingface/hub/models--Lightricks--LTX-2/snapshots/*/connectors/diffusion_pytorch_model.safetensors")[1]
  connectors <- load_ltx2_connectors(connector_path, device = "cpu", dtype = "float32", verbose = FALSE)

  conn_result <- connectors(text_result$prompt_embeds, text_result$prompt_attention_mask)
  video_embeds <- conn_result[[1]]
  audio_embeds <- conn_result[[2]]

  neg_conn_result <- connectors(neg_result$prompt_embeds, neg_result$prompt_attention_mask)
  neg_video_embeds <- neg_conn_result[[1]]
  neg_audio_embeds <- neg_conn_result[[2]]

  conn_time <- as.numeric(Sys.time() - t0)
  cat(sprintf("Connectors: %.1f seconds\n", conn_time))
  cat(sprintf("Video embeds shape: %s\n\n", paste(dim(video_embeds), collapse = "x")))

  # Move to GPU
  latent_dtype <- torch::torch_float16()
  video_embeds <- video_embeds$to(device = "cuda", dtype = latent_dtype)
  audio_embeds <- audio_embeds$to(device = "cuda", dtype = latent_dtype)
  neg_video_embeds <- neg_video_embeds$to(device = "cuda", dtype = latent_dtype)
  neg_audio_embeds <- neg_audio_embeds$to(device = "cuda", dtype = latent_dtype)

  # ---- Step 3: Load INT4 DiT ----
  cat("=== Step 3: INT4 DiT Loading ===\n")
  t0 <- Sys.time()
  mem_before <- get_gpu_memory()

  options(diffuseR.use_int4 = TRUE)
  options(diffuseR.int4_device = "cuda")
  options(diffuseR.int4_dtype = latent_dtype)

  dit <- ltx2_video_transformer_3d_model(
    in_channels = 128L,
    out_channels = 128L,
    num_attention_heads = 32L,
    attention_head_dim = 128L,
    cross_attention_dim = 4096L,
    audio_in_channels = 128L,
    audio_out_channels = 128L,
    audio_num_attention_heads = 32L,
    audio_attention_head_dim = 64L,
    audio_cross_attention_dim = 2048L,
    num_layers = 48L,
    caption_channels = 3840L,
    vae_scale_factors = c(temporal_ratio, spatial_ratio, spatial_ratio)
  )

  int4_weights <- load_int4_weights("~/.cache/diffuseR/ltx2_transformer_int4.safetensors", verbose = FALSE)
  load_int4_weights_into_model(dit, int4_weights, verbose = FALSE)
  dit <- dit$to(device = "cuda", dtype = latent_dtype)

  dit_time <- as.numeric(Sys.time() - t0)
  mem_after <- get_gpu_memory()
  cat(sprintf("DiT loading: %.1f seconds\n", dit_time))
  cat(sprintf("GPU memory: %.2f GB -> %.2f GB\n\n", mem_before, mem_after))

  # ---- Step 4: Initialize and Denoise ----
  cat("=== Step 4: Denoising ===\n")
  t0 <- Sys.time()

  # Initialize latents
  latents <- torch::torch_randn(
    c(1L, 128L, latent_frames, latent_height, latent_width),
    device = "cuda",
    dtype = latent_dtype
  )

  # Pack for transformer
  num_patches <- latent_frames * latent_height * latent_width
  latents <- latents$permute(c(1L, 3L, 4L, 5L, 2L))
  latents <- latents$reshape(c(1L, num_patches, 128L))

  # Audio placeholder
  audio_latents <- torch::torch_zeros(c(1L, 50L, 128L), device = "cuda", dtype = latent_dtype)

  # Scheduler
  schedule <- flowmatch_set_timesteps(
    flowmatch_scheduler_create(shift = 9.0),
    num_inference_steps = num_steps,
    device = "cuda"
  )

  guidance_scale <- 4.0

  for (i in seq_len(num_steps)) {
    sigma <- schedule$sigmas[i]
    sigma_next <- if (i < num_steps) schedule$sigmas[i + 1L] else 0

    timestep <- torch::torch_tensor(c(as.numeric(schedule$timesteps[i])))$unsqueeze(2L)
    timestep <- timestep$to(device = "cuda", dtype = latent_dtype)

    # Unconditional forward
    out_uncond <- dit(
      hidden_states = latents,
      audio_hidden_states = audio_latents,
      encoder_hidden_states = neg_video_embeds,
      audio_encoder_hidden_states = neg_audio_embeds,
      timestep = timestep,
      num_frames = latent_frames,
      height = latent_height,
      width = latent_width,
      fps = 24.0,
      audio_num_frames = 50L
    )

    # Conditional forward
    out_cond <- dit(
      hidden_states = latents,
      audio_hidden_states = audio_latents,
      encoder_hidden_states = video_embeds,
      audio_encoder_hidden_states = audio_embeds,
      timestep = timestep,
      num_frames = latent_frames,
      height = latent_height,
      width = latent_width,
      fps = 24.0,
      audio_num_frames = 50L
    )

    # CFG
    noise_uncond <- out_uncond$sample
    noise_cond <- out_cond$sample
    noise_pred <- noise_uncond + (noise_cond - noise_uncond)$mul(guidance_scale)

    # FlowMatch step
    dt <- torch::torch_tensor(sigma_next - as.numeric(sigma), dtype = latent_dtype, device = "cuda")
    latents <- latents + dt * noise_pred

    cat(sprintf("  Step %d/%d done\n", i, num_steps))
  }

  denoise_time <- as.numeric(Sys.time() - t0)
  mem_peak <- get_gpu_memory()
  cat(sprintf("Denoising: %.1f seconds (%.1f s/step)\n", denoise_time, denoise_time / num_steps))
  cat(sprintf("Peak GPU memory: %.2f GB\n\n", mem_peak))

  # Unpack latents
  latents <- latents$reshape(c(1L, latent_frames, latent_height, latent_width, 128L))
  latents <- latents$permute(c(1L, 5L, 2L, 3L, 4L))

  # ---- Step 5: VAE Decode on CPU ----
  cat("=== Step 5: VAE Decode (CPU) ===\n")
  t0 <- Sys.time()

  # Clear GPU for VAE
  rm(dit, int4_weights)
  gc(full = TRUE)
  torch::cuda_empty_cache()

  # Move latents to CPU
  latents_cpu <- latents$to(device = "cpu", dtype = torch::torch_float32())
  rm(latents)
  gc()

  cat(sprintf("Latents moved to CPU, shape: %s\n", paste(dim(latents_cpu), collapse = "x")))

  # Load VAE
  vae_path <- Sys.glob("~/.cache/huggingface/hub/models--Lightricks--LTX-2/snapshots/*/vae/")[1]
  if (!is.na(vae_path) && dir.exists(vae_path)) {
    cat("Loading VAE...\n")
    vae <- load_ltx2_vae(vae_path, device = "cpu", dtype = "float32", verbose = TRUE)

    cat("Decoding latents...\n")
    video_tensor <- vae$decode(latents_cpu)

    vae_time <- as.numeric(Sys.time() - t0)
    cat(sprintf("VAE decode: %.1f seconds\n", vae_time))
    cat(sprintf("Video tensor shape: %s\n", paste(dim(video_tensor), collapse = "x")))

    # Convert to array
    video_array <- as.array(video_tensor$squeeze(1L)$permute(c(2L, 3L, 4L, 1L))$cpu())
    video_array <- pmax(pmin(video_array, 1), 0)
    cat(sprintf("Video array shape: %s\n", paste(dim(video_array), collapse = "x")))
    cat(sprintf("Value range: [%.3f, %.3f]\n", min(video_array), max(video_array)))

    # Save video
    output_path <- path.expand("~/cornball_media/bigfoot_dj_25steps.mp4")
    dir.create(dirname(output_path), showWarnings = FALSE, recursive = TRUE)
    cat(sprintf("\nSaving video to: %s\n", output_path))
    save_video(video_array, output_path, fps = 24)
    cat("Video saved.\n")
  } else {
    cat("VAE not found - skipping decode\n")
    vae_time <- 0
  }

})

# ---- Summary ----
cat("\n=== TIMING SUMMARY ===\n")
cat(sprintf("Text encoding (Gemma3):  %.1f s\n", text_time))
cat(sprintf("Connectors:              %.1f s\n", conn_time))
cat(sprintf("DiT loading:             %.1f s\n", dit_time))
cat(sprintf("Denoising (%d steps):    %.1f s\n", num_steps, denoise_time))
cat(sprintf("VAE decode (CPU):        %.1f s\n", vae_time))
cat(sprintf("TOTAL:                   %.1f s\n", text_time + conn_time + dit_time + denoise_time + vae_time))
cat(sprintf("\nPeak GPU memory: %.2f GB\n", mem_peak))

#' FLUX.2 Transformer (MMDiT)
#'
#' Fresh R port of Flux2Transformer2DModel from the diffusers reference
#' implementation (Apache-2.0,
#' src/diffusers/models/transformers/transformer_flux2.py). Defaults are
#' the klein-4B configuration (5 double + 20 single blocks). Guidance
#' embeddings (FLUX.2-dev) are not implemented; klein is step-distilled
#' with \code{guidance_embeds = false}. Timestep conditioning has no
#' pooled-text component, and the three modulation projections are
#' shared across all blocks.
#'
#' @name dit_flux2
NULL

# Timestep-only conditioning: sinusoid(256) -> MLP, bias-free. Matches
# diffusers Flux2TimestepGuidanceEmbeddings state-dict names (klein has
# no guidance_embedder).
flux2_time_guidance_embed <- torch::nn_module(
    "flux2_time_guidance_embed",
    initialize = function(embedding_dim, in_channels = 256L) {
    self$in_channels <- in_channels
    self$timestep_embedder <- ltx23_timestep_embedding(in_channels,
        embedding_dim, bias = FALSE)
},
    forward = function(timestep) {
    proj <- ltx23_get_timestep_embedding(timestep, self$in_channels,
        flip_sin_to_cos = TRUE, downscale_freq_shift = 0)
    self$timestep_embedder(proj$to(dtype = self$timestep_embedder$linear_1$weight$dtype))
}
)

#' FLUX.2 transformer model
#'
#' Shared modulation computed once per forward; double blocks over
#' separate text/image streams, then single (parallel) blocks over the
#' concatenated [text; image] sequence. Rotary embeddings are
#' precomputed by the caller with \code{\link{flux_pos_embed}}
#' (\code{axes_dim = c(32, 32, 32, 32)}, \code{theta = 2000}) over the
#' concatenated [text; image] 4-axis position ids.
#'
#' @param in_channels Integer. Packed latent channels (128).
#' @param num_layers Integer. Double-stream block count (klein-4B: 5).
#' @param num_single_layers Integer. Single-stream block count (20).
#' @param attention_head_dim Integer. Per-head dimension.
#' @param num_attention_heads Integer. Attention heads.
#' @param joint_attention_dim Integer. Text embedding dim (7680).
#' @param mlp_ratio Numeric. Feed-forward multiplier (3.0).
#' @param timestep_guidance_channels Integer. Sinusoid width (256).
#' @param axes_dims_rope Integer vector. Per-axis rotary dims.
#' @param rope_theta Numeric. Rotary base frequency (2000).
#' @param eps Numeric. Norm epsilon.
#' @param out_channels Integer or NULL. Defaults to \code{in_channels}.
#'
#' @return Module whose forward(hidden_states, encoder_hidden_states,
#'   timestep, image_rotary_emb) returns the predicted velocity for the
#'   image tokens [B, S_img, out_channels]. \code{timestep} is in sigma
#'   space (0-1); it is scaled by 1000 internally.
#'
#' @export
flux2_transformer <- torch::nn_module(
                                      "flux2_transformer",
                                      initialize = function(in_channels = 128L,
        num_layers = 5L,
        num_single_layers = 20L,
        attention_head_dim = 128L,
        num_attention_heads = 24L,
        joint_attention_dim = 7680L,
        mlp_ratio = 3.0,
        timestep_guidance_channels = 256L,
        axes_dims_rope = c(32L, 32L, 32L, 32L),
        rope_theta = 2000,
        eps = 1e-6,
        out_channels = NULL) {
    inner_dim <- num_attention_heads * attention_head_dim
    self$inner_dim <- inner_dim
    self$axes_dims_rope <- as.integer(axes_dims_rope)
    self$rope_theta <- rope_theta
    self$out_channels <- as.integer(out_channels %||% in_channels)

    self$time_guidance_embed <- flux2_time_guidance_embed(inner_dim,
        as.integer(timestep_guidance_channels))
    self$double_stream_modulation_img <- flux2_modulation(inner_dim, 2L)
    self$double_stream_modulation_txt <- flux2_modulation(inner_dim, 2L)
    self$single_stream_modulation <- flux2_modulation(inner_dim, 1L)

    self$x_embedder <- torch::nn_linear(in_channels, inner_dim, bias = FALSE)
    self$context_embedder <- torch::nn_linear(joint_attention_dim,
        inner_dim, bias = FALSE)

    self$transformer_blocks <- torch::nn_module_list(
        lapply(seq_len(num_layers), function(i) {
        flux2_double_block(inner_dim, num_attention_heads,
                           attention_head_dim, mlp_ratio = mlp_ratio,
                           eps = eps)
    })
    )
    self$single_transformer_blocks <- torch::nn_module_list(
        lapply(seq_len(num_single_layers), function(i) {
        flux2_single_block(inner_dim, num_attention_heads,
                           attention_head_dim, mlp_ratio = mlp_ratio,
                           eps = eps)
    })
    )

    self$norm_out <- flux_ada_layer_norm_continuous(inner_dim, inner_dim,
        bias = FALSE)
    self$proj_out <- torch::nn_linear(inner_dim, self$out_channels,
                                      bias = FALSE)
},
                                      forward = function(hidden_states, encoder_hidden_states, timestep,
        image_rotary_emb, chunk_size = NULL) {
    hidden_states <- self$x_embedder(hidden_states)
    timestep <- timestep$to(dtype = hidden_states$dtype)$mul(1000)
    temb <- self$time_guidance_embed(timestep)

    mod_img <- self$double_stream_modulation_img(temb)
    mod_txt <- self$double_stream_modulation_txt(temb)
    mod_single <- self$single_stream_modulation(temb)

    encoder_hidden_states <- self$context_embedder(encoder_hidden_states)

    block_gc <- isTRUE(getOption("diffuseR.block_gc"))
    for (i in seq_along(self$transformer_blocks)) {
        res <- self$transformer_blocks[[i]](
                                      hidden_states = hidden_states,
                                      encoder_hidden_states = encoder_hidden_states,
                                      temb_mod_img = mod_img,
                                      temb_mod_txt = mod_txt,
                                      image_rotary_emb = image_rotary_emb,
                                      chunk_size = chunk_size
        )
        encoder_hidden_states <- res[[1]]
        hidden_states <- res[[2]]
        if (block_gc) {
            gc(verbose = FALSE)
        }
    }

    txt_len <- encoder_hidden_states$shape[2]
    hidden_states <- torch::torch_cat(
                                      list(encoder_hidden_states, hidden_states),
                                      dim = 2L
    )
    for (i in seq_along(self$single_transformer_blocks)) {
        hidden_states <- self$single_transformer_blocks[[i]](
                                      hidden_states = hidden_states,
                                      temb_mod = mod_single,
                                      image_rotary_emb = image_rotary_emb,
                                      chunk_size = chunk_size
        )
        if (block_gc) {
            gc(verbose = FALSE)
        }
    }
    hidden_states <- hidden_states$narrow(
        2L, txt_len + 1L,
        hidden_states$shape[2] - txt_len
    )

    hidden_states <- self$norm_out(hidden_states, temb)
    self$proj_out(hidden_states)
}
)

#` load_pipeline` function to load a diffusion model pipeline
#' Load a diffusion model pipeline
#'
#' This function loads a diffusion model pipeline consisting of a UNet, VAE decoder, and text encoder.
#' It initializes the models and sets up the environment for inference.
#'
#' @param model_name The name of the model to load.
#' @param m2d A list containing model-to-device mappings and configurations.
#' @param i2i Logical indicating whether to load the encoder for img2img().
#' @param unet_dtype_str A string representing the data type for the UNet model (e.g., "float32", "float16").
#' @param use_native_decoder Logical; if TRUE, uses native R torch decoder instead of TorchScript.
#'   Native decoder has better GPU compatibility (especially Blackwell).
#' @param use_native_text_encoder Logical; if TRUE, uses native R torch text encoder instead of TorchScript.
#'   Native text encoder has better GPU compatibility (especially Blackwell).
#' @param use_native_unet Logical; if TRUE, uses native R torch UNet instead of TorchScript.
#'   Native UNet has better GPU compatibility (especially Blackwell).
#' @param ... Additional arguments passed to the model loading functions.
#'
#' @return An environment containing the loaded models and configuration.
#' @export
#'
#' @examples
#' \dontrun{
#' pipeline <- load_pipeline("my_model", device = "cuda")
#' }
#'
load_pipeline <- function(model_name, m2d, i2i = FALSE, unet_dtype_str,
                          use_native_decoder = FALSE,
                          use_native_text_encoder = FALSE,
                          use_native_unet = FALSE, ...) {
    # Create an environment to store the pipeline components
    # pipeline <- new.env(parent = emptyenv())
    devices <- m2d$devices
    unet_dtype <- m2d$unet_dtype
    device_cpu <- m2d$device_cpu
    device_cuda <- m2d$device_cuda

    if (any(unlist(devices) == "cuda")) {
        # Same chronic-allocator-gate fix as the FLUX/Z-Image pipelines
        # (see .flux_gc_gates). Measured at 50 steps (native modules,
        # decoder on CPU, RTX 5060 Ti 16 GB), warm generation:
        # SDXL 1024^2 234 -> 100 s (gc 209 -> 75 s); SD 2.1 768^2
        # 101 -> 26 s (gc 88 -> 15 s). Note the reduced gc cadence lets
        # the garbage sawtooth ride higher (SD21 peak 5.4 -> 13.3 GB);
        # cards much smaller than 16 GB may need different gate sizing.
        .flux_gc_gates(footprint_gb = 12)
    }

    pipeline <- list()
    # Load models into the environment
    if (i2i) {
        message("Loading image encoder...")
        pipeline$encoder <- load_model_component(component = "encoder",
            model_name, devices$encoder)
    }
    message("Loading text_encoder...")
    pipeline$text_encoder <- load_model_component("text_encoder", model_name,
        devices$text_encoder,
        use_native = use_native_text_encoder)
    if (model_name == "sdxl") {
        message("Loading text_encoder2...")
        pipeline$text_encoder2 <- load_model_component(component = "text_encoder2",
            model_name,
            devices$text_encoder2,
            use_native = use_native_text_encoder)
    }
    message("Loading unet...")
    pipeline$unet <- load_model_component("unet", model_name,
        device = devices$unet,
        unet_dtype_str = unet_dtype_str,
        use_native = use_native_unet)

    message("Loading image decoder...")
    pipeline$decoder <- load_model_component("decoder", model_name,
        devices$decoder,
        use_native = use_native_decoder)

    # Store configuration
    pipeline$devices <- devices

    # Add a class for S3 method dispatch if needed
    # class(pipeline) <- c("diffusion_pipeline", "environment")

    return(pipeline)
}

#' Pinned Staging for Phase-Sequential Components
#'
#' Phase offloading moves each large component (transformer,
#' connectors, VAEs, vocoder, text encoders) between CPU and GPU every
#' render. From pageable memory those copies run through the driver's
#' bounce buffer at a fraction of PCIe speed; page-locked (pinned) host
#' memory transfers by DMA at full rate. Each component's parameters
#' and buffers are pinned once at load; onload swaps every tensor to
#' a non-blocking GPU copy of its pinned source, and offload simply
#' re-points the tensors at the still-valid pinned copies — weights
#' are immutable during inference, so offload moves no bytes at all.
#'
#' Costs: the model's host copies become non-swappable for the life
#' of the pipeline (no extra RAM - set_data repoints the same
#' tensors), and page-locking adds ~9s to pipeline load. Measured
#' post byte-LUT (768x512x49, NF4, RTX 5060 Ti): ~7s saved per render
#' (warm renders 64-66s pageable vs 57-59s pinned; denoise and decode
#' identical, the delta is pure transfer), so pinning breaks even on
#' the second render and costs a single-render session ~2s net. On by
#' default; page-locking failure falls back silently per component,
#' and \code{options(diffuseR.pin_staging = FALSE)} before the loader
#' opts out (e.g. under host memory pressure, where unswappable pages
#' turn thrashing into OOM). The LTX pipeline, the Gemma3 encoder, and
#' the FLUX-family image loaders (flux1, flux2, zimage) all stage
#' pinned weights; \code{\link{recommend}} computes the RAM-aware
#' \code{pin} default per model.
#'
#' @name staging
#' @aliases staging_ltx23
NULL

# Allocate pinned host memory without Tensor$pin_memory(): this torch
# build's binding requires the deprecated device argument (omitting it
# errors with "Expected a torch_device"; passing it prints two libtorch
# deprecation warnings per tensor - thousands of lines per pipeline
# load). torch_empty_strided is the one creation op that exposes
# pin_memory, so pin by allocating and copying; fall back to the noisy
# path on builds where that fails.
.pin_host <- function(p) {
    tryCatch({
        sz <- as.integer(p$shape)
        st <- if (length(sz)) {
            as.integer(rev(cumprod(c(1, rev(sz)[-length(sz)]))))
        } else {
            integer(0)
        }
        buf <- torch::torch_empty_strided(sz, st, dtype = p$dtype,
            pin_memory = TRUE)
        torch::with_no_grad(buf$copy_(p))
        buf
    }, error = function(e) {
        suppressWarnings(p$pin_memory(device = torch::torch_device("cuda")))
    })
}

#' Pin a component's tensors for fast phase transfer
#'
#' @param module An nn_module on the CPU.
#' @param extra Optional list of additional plain-field tensors to pin
#'   alongside the module's parameters and buffers (e.g. an fp8
#'   linear's \code{weight_fp8}/\code{weight_scale} fields, which live
#'   outside \code{parameters}/\code{buffers}). \code{set_data} mutates
#'   each tensor in place, so the field reference stays valid.
#'
#' @return A list of \code{list(live, pinned)} tensor pairs, or NULL
#'   if pinning is unavailable (no CUDA, or page-locking failed).
#'
#' @keywords internal
.pin_component <- function(module, extra = NULL) {
    if (!torch::cuda_is_available()) {
        return(NULL)
    }
    tryCatch({
        tensors <- c(module$parameters, module$buffers, extra)
        lapply(tensors, function(p) {
            pinned <- .pin_host(p)
            p$set_data(pinned)
            list(live = p, pinned = pinned)
        })
    }, error = function(e) NULL)
}

#' Move a pinned component onto the compute device
#'
#' Non-blocking copies from pinned memory share the default stream,
#' so later kernels are ordered after them; no explicit sync needed.
#'
#' @keywords internal
.staged_onload <- function(staging, device) {
    for (pair in staging) {
        pair$live$set_data(pair$pinned$to(device = device, non_blocking = TRUE))
    }
    invisible(NULL)
}

#' Return a pinned component to the CPU
#'
#' Weights are immutable during inference, so the pinned host copies
#' are still current: offload is a pointer swap, no transfer.
#'
#' @keywords internal
.staged_offload <- function(staging) {
    for (pair in staging) {
        pair$live$set_data(pair$pinned)
    }
    invisible(NULL)
}

# Pinned staging round trip (R/staging.R): pin -> onload ->
# offload -> onload must preserve outputs exactly. CUDA-only.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!torch::cuda_is_available()) exit_file("no CUDA")

library(diffuseR)
torch::torch_manual_seed(13)

m <- ltx23_feed_forward(16L)
m$eval()
x <- torch::torch_randn(2L, 5L, 16L)
torch::with_no_grad(ref <- m(x))

st <- diffuseR:::.pin_component(m)
expect_false(is.null(st))
expect_true(suppressWarnings(
  st[[1]]$live$is_pinned(device = torch::torch_device("cuda"))
))

# Pinning must not change CPU outputs
torch::with_no_grad(out_pinned <- m(x))
expect_true(as.numeric((out_pinned - ref)$abs()$max()) == 0)

# Onload: GPU forward matches
diffuseR:::.staged_onload(st, "cuda")
expect_equal(st[[1]]$live$device$type, "cuda")
torch::with_no_grad(
  out_gpu <- m(x$to(device = "cuda"))$cpu()
)
expect_true(as.numeric((out_gpu - ref)$abs()$max()) < 1e-5)

# Offload: pointer swap back to the pinned copies, exact outputs
diffuseR:::.staged_offload(st)
expect_equal(st[[1]]$live$device$type, "cpu")
torch::with_no_grad(out_back <- m(x))
expect_true(as.numeric((out_back - ref)$abs()$max()) == 0)

# Second round trip still exact
diffuseR:::.staged_onload(st, "cuda")
torch::with_no_grad(out_gpu2 <- m(x$to(device = "cuda"))$cpu())
expect_true(as.numeric((out_gpu2 - out_gpu)$abs()$max()) == 0)
diffuseR:::.staged_offload(st)

#' Download and Prepare FLUX.2 Klein 4B Weights
#'
#' Downloads FLUX.2-klein-4B from HuggingFace (Apache-2.0, ungated) and
#' quantizes the 4B transformer to a local fp8 (~4 GB) or NF4 (~2.3 GB)
#' artifact.
#'
#' @name download_flux2
NULL

.flux2_repo <- "black-forest-labs/FLUX.2-klein-4B"

.flux2_transformer_files <- c("transformer/config.json",
                              "transformer/diffusion_pytorch_model.safetensors")

.flux2_support_files <- c(
                          "vae/config.json",
                          "vae/diffusion_pytorch_model.safetensors",
                          "text_encoder/config.json",
                          "text_encoder/model.safetensors.index.json",
                          sprintf("text_encoder/model-%05d-of-00002.safetensors", 1:2),
                          "tokenizer/tokenizer.json",
                          "tokenizer/tokenizer_config.json",
                          "tokenizer/special_tokens_map.json",
                          "scheduler/scheduler_config.json"
)

#' Download FLUX.2-klein-4B and build the quantized artifact
#'
#' Skips work already done: a valid quantized manifest short-circuits
#' the transformer download; cached files are not re-fetched. No token
#' is needed (the repo is ungated). The bf16 transformer source
#' (~7.8 GB in the HuggingFace cache) may be deleted after quantization.
#'
#' @param quantize Logical. Build the quantized artifact.
#' @param precision "auto" (default: fp8 when safetensors supports
#'   float8, else nf4), "fp8" (~4 GB, GPU-resident; near-bf16 quality),
#'   or "nf4" (~2.3 GB).
#' @param output_dir Directory for the quantized artifact.
#' @param text_encoders Logical. Also fetch the Qwen3 text encoder,
#'   tokenizer, VAE, and scheduler config (~8.3 GB).
#' @param verbose Logical.
#'
#' @return Invisibly, a list with \code{transformer_dir},
#'   \code{artifact_dir}, and \code{support} (named file paths).
#'
#' @export
download_flux2_klein <- function(quantize = TRUE,
                                 precision = c("auto", "fp8", "nf4"),
                                 output_dir = NULL, text_encoders = TRUE,
                                 verbose = TRUE) {
    precision <- match.arg(precision)
    precision <- .flux_resolve_precision(precision,
        file.path(tools::R_user_dir("diffuseR", "data"), "flux2-klein-4b-"))
    # Explicit fp8 without float8 support: warn + build nf4 rather than fail.
    precision <- .st_graceful_precision(precision, mode = "write")
    if (is.null(output_dir)) {
        output_dir <- file.path(tools::R_user_dir("diffuseR", "data"),
                                paste0("flux2-klein-4b-", precision))
    }
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("The hfhub package is required to download model weights.")
    }
    result <- list(transformer_dir = NULL, artifact_dir = output_dir,
                   support = character(0))

    manifest_path <- file.path(output_dir, "manifest.json")
    have_artifact <- file.exists(manifest_path) && {
        m <- jsonlite::fromJSON(manifest_path)
        all(file.exists(file.path(output_dir, m$shards)))
    }

    if (!have_artifact || !quantize) {
        cached <- tryCatch(
                           hfhub::hub_download(.flux2_repo, .flux2_transformer_files[[2]],
                local_files_only = TRUE),
                           error = function(e) NULL
        )
        if (is.null(cached) ||
            !.hub_all_cached(.flux2_repo, .flux2_transformer_files)) {
            free <- .ltx23_disk_free_gb(path.expand("~"))
            if (!is.na(free) && free < 25) {
                warning(sprintf(
                                "Only %.0f GB free; the download + %s artifact need ~25 GB.",
                                free, precision
                    ))
            }
            ok <- .ltx23_consent(paste0(
                                        "FLUX.2-klein-4B: the 7.8 GB bf16 transformer plus a local ",
                                        precision, " artifact (Apache-2.0, ungated)"
                ))
            if (!ok) {
                stop("Download cancelled.", call. = FALSE)
            }
            if (verbose) {
                message("Downloading the FLUX.2-klein-4B transformer (7.8 GB)...")
            }
        }
        paths <- vapply(.flux2_transformer_files, function(f) {
            hfhub::hub_download(.flux2_repo, f)
        }, character(1))
        result$transformer_dir <- dirname(paths[[1]])

        if (quantize && !have_artifact) {
            if (verbose) {
                message("Quantizing transformer linears to ", precision,
                        " (one-time)...")
            }
            flux_quantize(result$transformer_dir, output_dir,
                          format = precision, verbose = verbose)
            if (verbose) {
                message(
                        toupper(precision), " artifact ready: ", output_dir, "\n",
                        "The 7.8 GB source in the HuggingFace cache may be ",
                        "deleted if you do not need bf16 weights."
                )
            }
        }
    } else if (verbose) {
        message(toupper(precision), " artifact already present: ", output_dir)
    }

    if (text_encoders) {
        have_te <- .hub_all_cached(.flux2_repo, .flux2_support_files)
        if (!have_te) {
            ok <- .ltx23_consent(
                                 "the Qwen3-4B text encoder, tokenizer, and VAE (~8.3 GB)"
            )
            if (!ok) {
                stop("Download cancelled.", call. = FALSE)
            }
            if (verbose) {
                message("Downloading text encoder + VAE...")
            }
        }
        result$support <- vapply(.flux2_support_files, function(f) {
            hfhub::hub_download(.flux2_repo, f)
        }, character(1))
    }

    invisible(result)
}

# Native R BPE Tokenizer
#
# Implements Byte-Pair Encoding tokenization in pure R.
# Compatible with HuggingFace tokenizer.json format.

# -----------------------------------------------------------------------------
# BPE Tokenizer Class
# -----------------------------------------------------------------------------
#' BPE Tokenizer
#'
#' Native R implementation of Byte-Pair Encoding tokenizer.
#' Loads from HuggingFace tokenizer.json format.
#'
#' @param tokenizer_path Path to tokenizer.json or directory containing it.
#' @return A bpe_tokenizer object.
#' @export
bpe_tokenizer <- function(tokenizer_path) {
    # Find tokenizer.json

    if (dir.exists(tokenizer_path)) {
        json_path <- file.path(tokenizer_path, "tokenizer.json")
    } else {
        json_path <- tokenizer_path
    }

    if (!file.exists(json_path)) {
        stop("tokenizer.json not found at: ", json_path)
    }

    # Load JSON
    config <- jsonlite::fromJSON(json_path, simplifyVector = FALSE)

    # Extract model configuration
    model <- config$model
    if (is.null(model) || model$type != "BPE") {
        stop("Only BPE tokenizers are supported")
    }

    # Build vocabulary lookup (token -> id)
    vocab <- model$vocab
    if (is.null(vocab)) {
        stop("No vocabulary found in tokenizer.json")
    }

    # Convert vocab list to named vector for fast lookup
    vocab_ids <- unlist(vocab)
    names(vocab_ids) <- names(vocab)

    # Build reverse vocabulary (id -> token)
    id_to_token <- names(vocab_ids)
    names(id_to_token) <- as.character(vocab_ids)

    # Parse merges into a priority map
    merges <- model$merges
    if (is.null(merges)) {
        merges <- character(0)
    }

    # Create merge priority lookup (pair -> priority)
    # Lower priority number = merge first
    merge_priority <- seq_along(merges)
    names(merge_priority) <- merges

    # Extract special tokens
    added_tokens <- config$added_tokens
    special_tokens <- list()
    if (!is.null(added_tokens)) {
        for (tok in added_tokens) {
            special_tokens[[tok$content]] <- list(id = tok$id,
                special = tok$special %||% FALSE)
        }
    }

    # Extract configuration
    byte_fallback <- model$byte_fallback %||% FALSE
    fuse_unk <- model$fuse_unk %||% FALSE
    unk_token <- model$unk_token

    # Pre-tokenizer configuration
    pre_tokenizer <- config$pre_tokenizer
    add_prefix_space <- FALSE
    if (!is.null(pre_tokenizer)) {
        # Check for Metaspace pre-tokenizer (adds space prefix)
        if (!is.null(pre_tokenizer$type) && pre_tokenizer$type == "Metaspace") {
            add_prefix_space <- pre_tokenizer$add_prefix_space %||% TRUE
        }
    }

    structure(
              list(
                   vocab = vocab_ids,
                   id_to_token = id_to_token,
                   merges = merges,
                   merge_priority = merge_priority,
                   special_tokens = special_tokens,
                   byte_fallback = byte_fallback,
                   fuse_unk = fuse_unk,
                   unk_token = unk_token,
                   add_prefix_space = add_prefix_space,
                   vocab_size = length(vocab_ids)
        ),
              class = "bpe_tokenizer"
    )
}

#' Print BPE Tokenizer
#' @param x A bpe_tokenizer object.
#' @param ... Additional arguments (ignored).
#'
#' @return Invisibly returns \code{x}. Called for the side effect of
#'   printing a summary of the tokenizer to the console.
#'
#' @export
print.bpe_tokenizer <- function(x, ...) {
    cat("BPE Tokenizer\n")
    cat("  Vocabulary size:", x$vocab_size, "\n")
    cat("  Merge rules:", length(x$merges), "\n")
    cat("  Special tokens:", length(x$special_tokens), "\n")
    cat("  Byte fallback:", x$byte_fallback, "\n")
    invisible(x)
}

# -----------------------------------------------------------------------------
# Encoding
# -----------------------------------------------------------------------------

#' Encode text to token IDs
#'
#' @param tokenizer A bpe_tokenizer object.
#' @param text Character string or vector to encode.
#' @param add_special_tokens Logical. Add BOS/EOS tokens.
#' @param max_length Integer. Maximum sequence length (NULL for no limit).
#' @param padding Character. Padding strategy: "none", "max_length", or "longest".
#' @param truncation Logical. Truncate to max_length.
#' @param return_tensors Character. Return type: "list" or "pt" (torch tensors).
#' @return List with input_ids and attention_mask.
#' @export
encode_bpe <- function(tokenizer, text, add_special_tokens = TRUE,
                       max_length = NULL, padding = "none",
                       truncation = FALSE, return_tensors = "list") {
    if (!inherits(tokenizer, "bpe_tokenizer")) {
        stop("tokenizer must be a bpe_tokenizer object")
    }

    # Handle single string
    if (length(text) == 1) {
        text <- list(text)
    }

    # Encode each text
    encoded <- lapply(text, function(t) {
        encode_single(tokenizer, t, add_special_tokens = add_special_tokens)
    })

    # Get lengths
    lengths <- vapply(encoded, length, integer(1))
    max_len <- max(lengths)

    # Apply max_length constraint
    if (!is.null(max_length)) {
        if (truncation) {
            encoded <- lapply(encoded, function(ids) {
                if (length(ids) > max_length) ids[1:max_length] else ids
            })
        }
        if (padding == "max_length") {
            max_len <- max_length
        }
    }

    # Apply padding
    if (padding != "none") {
        pad_id <- get_pad_id(tokenizer)
        encoded <- lapply(encoded, function(ids) {
            if (length(ids) < max_len) {
                # Left padding (Gemma style)
                c(rep(pad_id, max_len - length(ids)), ids)
            } else {
                ids
            }
        })
    }

    # Create attention mask
    attention_mask <- lapply(encoded, function(ids) {
        pad_id <- get_pad_id(tokenizer)
        as.integer(ids != pad_id)
    })

    # Convert to matrix
    input_ids <- do.call(rbind, encoded)
    attention_mask <- do.call(rbind, attention_mask)

    # Convert to tensors if requested
    if (return_tensors == "pt") {
        input_ids <- torch::torch_tensor(input_ids, dtype = torch::torch_long())
        attention_mask <- torch::torch_tensor(attention_mask,
            dtype = torch::torch_long())
    }

    list(input_ids = input_ids, attention_mask = attention_mask)
}

#' Encode a single text string
#' @keywords internal
encode_single <- function(tokenizer, text, add_special_tokens = TRUE) {
    # Add prefix space if configured (Gemma style)
    if (tokenizer$add_prefix_space && !startsWith(text, " ")) {
        text <- paste0(" ", text)
    }

    # Pre-tokenize: split on whitespace while keeping track of spaces
    # Replace spaces with special character (SentencePiece style)
    text <- gsub(" ", "\u2581", text) # LOWER ONE EIGHTH BLOCK

    # Check for special tokens first
    for (tok_name in names(tokenizer$special_tokens)) {
        if (text == tok_name) {
            return(tokenizer$special_tokens[[tok_name]]$id)
        }
    }

    # Use greedy longest match tokenization (for large vocabs like Gemma)
    # This is more efficient than BPE merging for pre-trained vocabs
    ids <- greedy_tokenize(text, tokenizer$vocab, tokenizer$byte_fallback,
                           tokenizer$unk_token)

    # Add special tokens
    if (add_special_tokens) {
        bos_id <- get_bos_id(tokenizer)
        if (!is.null(bos_id)) {
            ids <- c(bos_id, ids)
        }
    }

    ids
}

#' Greedy longest match tokenization
#' @keywords internal
greedy_tokenize <- function(text, vocab, byte_fallback = FALSE,
                            unk_token = NULL) {
    ids <- integer(0)
    i <- 1
    n <- nchar(text)
    vocab_names <- names(vocab)

    while (i <= n) {
        # Try to match the longest token starting at position i
        matched <- FALSE

        # Try decreasing lengths
        for (len in min(n - i + 1, 50):1) { # Cap at 50 chars max
            candidate <- substr(text, i, i + len - 1)

            if (candidate %in% vocab_names) {
                ids <- c(ids, vocab[[candidate]])
                i <- i + len
                matched <- TRUE
                break
            }
        }

        if (!matched) {
            # No match found - use byte fallback or UNK
            char <- substr(text, i, i)
            if (byte_fallback) {
                # Try single character in vocab first
                if (char %in% vocab_names) {
                    ids <- c(ids, vocab[[char]])
                } else {
                    # Byte fallback: convert to <0xHH> format
                    bytes <- charToRaw(char)
                    for (b in bytes) {
                        byte_token <- sprintf("<0x%02X>", as.integer(b))
                        if (byte_token %in% vocab_names) {
                            ids <- c(ids, vocab[[byte_token]])
                        } else if (!is.null(unk_token) &&
                            unk_token %in% vocab_names) {
                            ids <- c(ids, vocab[[unk_token]])
                        } else {
                            ids <- c(ids, 3L) # Default UNK
                        }
                    }
                }
            } else if (!is.null(unk_token) && unk_token %in% vocab_names) {
                ids <- c(ids, vocab[[unk_token]])
            } else {
                ids <- c(ids, 3L) # Default UNK
            }
            i <- i + 1
        }
    }

    ids
}

#' Apply BPE merge rules
#' @keywords internal
apply_bpe_merges <- function(tokens, merge_priority, vocab) {
    if (length(tokens) <= 1) {
        return(tokens)
    }

    # Iteratively merge token pairs
    changed <- TRUE
    while (changed && length(tokens) > 1) {
        changed <- FALSE
        best_idx <- NULL
        best_priority <- Inf

        # Find the highest priority merge
        for (i in seq_len(length(tokens) - 1)) {
            pair <- paste(tokens[i], tokens[i + 1])
            if (pair %in% names(merge_priority)) {
                priority <- merge_priority[[pair]]
                if (priority < best_priority) {
                    best_priority <- priority
                    best_idx <- i
                }
            }
        }

        # Apply the best merge
        if (!is.null(best_idx)) {
            merged <- paste0(tokens[best_idx], tokens[best_idx + 1])
            # Only merge if result is in vocabulary
            if (merged %in% names(vocab)) {
                tokens <- c(
                    if (best_idx > 1) {
                        tokens[1:(best_idx - 1)]
                    } else {
                        character(0)
                    },
                            merged,
                    if (best_idx + 2 <= length(tokens)) {
                        tokens[(best_idx + 2):length(tokens)]
                    } else {
                        character(0)
                    }
                )
                changed <- TRUE
            }
        }
    }

    tokens
}

# -----------------------------------------------------------------------------
# Decoding
# -----------------------------------------------------------------------------

#' Decode token IDs to text
#'
#' @param tokenizer A bpe_tokenizer object.
#' @param ids Integer vector or matrix of token IDs.
#' @param skip_special_tokens Logical. Skip special tokens in output.
#' @return Character string or vector.
#' @export
decode_bpe <- function(tokenizer, ids, skip_special_tokens = TRUE) {
    if (!inherits(tokenizer, "bpe_tokenizer")) {
        stop("tokenizer must be a bpe_tokenizer object")
    }

    # Handle matrix input
    if (is.matrix(ids)) {
        return(apply(ids, 1, function(row) {
            decode_bpe(tokenizer, row,
                       skip_special_tokens = skip_special_tokens)
        }))
    }

    # Handle torch tensor
    if (inherits(ids, "torch_tensor")) {
        ids <- as.integer(ids$cpu()$numpy())
    }

    # Get special token IDs to skip
    special_ids <- integer(0)
    if (skip_special_tokens) {
        special_ids <- vapply(tokenizer$special_tokens, function(tok) tok$id, integer(1))
    }

    # Convert IDs to tokens
    tokens <- vapply(ids, function(id) {
        if (id %in% special_ids) {
            ""
        } else {
            id_str <- as.character(id)
            if (id_str %in% names(tokenizer$id_to_token)) {
                tokenizer$id_to_token[[id_str]]
            } else {
                ""
            }
        }
    }, character(1))

    # Join tokens and decode
    text <- paste(tokens, collapse = "")

    # Replace SentencePiece space marker with actual space
    text <- gsub("\u2581", " ", text)

    # Decode byte tokens like <0x41>
    # Find all byte tokens and replace them
    byte_pattern <- "<0x([0-9A-Fa-f]{2})>"
    while (grepl(byte_pattern, text)) {
        match <- regmatches(text, regexpr(byte_pattern, text))
        if (length(match) > 0) {
            hex_str <- sub("<0x([0-9A-Fa-f]{2})>", "\\1", match)
            byte_val <- strtoi(hex_str, base = 16)
            char <- rawToChar(as.raw(byte_val))
            text <- sub(byte_pattern, char, text, fixed = FALSE)
        } else {
            break
        }
    }

    # Trim leading space if prefix was added
    text <- sub("^ ", "", text)

    text
}

# -----------------------------------------------------------------------------
# Helper Functions
# -----------------------------------------------------------------------------

#' Get padding token ID
#' @keywords internal
get_pad_id <- function(tokenizer) {
    if ("<pad>" %in% names(tokenizer$special_tokens)) {
        tokenizer$special_tokens[["<pad>"]]$id
    } else if ("<pad>" %in% names(tokenizer$vocab)) {
        tokenizer$vocab[["<pad>"]]
    } else {
        0L
    }
}

#' Get BOS token ID
#' @keywords internal
get_bos_id <- function(tokenizer) {
    if ("<bos>" %in% names(tokenizer$special_tokens)) {
        tokenizer$special_tokens[["<bos>"]]$id
    } else if ("<bos>" %in% names(tokenizer$vocab)) {
        tokenizer$vocab[["<bos>"]]
    } else {
        NULL
    }
}

#' Get EOS token ID
#' @keywords internal
get_eos_id <- function(tokenizer) {
    if ("<eos>" %in% names(tokenizer$special_tokens)) {
        tokenizer$special_tokens[["<eos>"]]$id
    } else if ("<eos>" %in% names(tokenizer$vocab)) {
        tokenizer$vocab[["<eos>"]]
    } else {
        NULL
    }
}

#' Get vocabulary size
#' @param tokenizer A bpe_tokenizer object.
#' @return Integer vocabulary size.
#' @export
vocab_size <- function(tokenizer) {
    tokenizer$vocab_size
}

# Null-coalescing operator
if (!exists("%||%", mode = "function")) {
    `%||%` <- function(x, y) if (is.null(x)) {
        y
    } else {
        x
    }
}

#' Download a single TorchScript model component
#'
#' Downloads a specific model component file (e.g., UNet, decoder, text encoder)
#' using \code{hfhub::hub_download()} from the cornball-ai dataset repos.
#'
#' @param model_name Character string, the name of the model (e.g., \code{"sd21"}).
#' @param component Character string, the component to download (e.g., \code{"unet"}, \code{"decoder"}).
#' @param device Character string, the device type (e.g., \code{"cpu"} or \code{"cuda"}).
#' @param overwrite Logical; if \code{TRUE}, force re-download even if cached.
#' @param show_progress Logical; if \code{TRUE} (default), displays progress during download.
#'
#' @return The local file path to the downloaded component (character string).
#' @export
#'
#' @examples
#' \dontrun{
#' path <- download_component("sd21", "text_encoder", "cpu")
#' }
download_component <- function(model_name = "sd21", component,
                               device = "cpu", overwrite = FALSE,
                               show_progress = TRUE) {
    filename <- paste0(component, "-", device, ".pt")
    if (!requireNamespace("hfhub", quietly = TRUE)) {
        stop("Package 'hfhub' is required. Install it from CRAN before ",
             "calling download_component() again.")
    }
    repo_id <- paste0("cornball-ai/", model_name, "-R")

    cached <- !overwrite && !is.null(tryCatch(
            hfhub::hub_download(repo_id, filename, repo_type = "dataset",
                                local_files_only = TRUE),
            error = function(e) NULL))
    if (!cached &&
        !.ltx23_consent(paste0(model_name, " ", component, " (", filename, ")"))) {
        stop("Download cancelled.", call. = FALSE)
    }

    if (overwrite) {
        return(hfhub::hub_download(repo_id, filename, repo_type = "dataset",
                                   force_download = TRUE))
    }
    hf_download_pt(model_name, filename, download = TRUE)
}

#' LTX-2.3 Text-to-Video Pipeline
#'
#' Fresh R port of the LTX-2 text-to-video flow from the diffusers
#' reference (Apache-2.0, pipelines/ltx2/pipeline_ltx2.py), specialized
#' for the distilled LTX 2.3 checkpoints: 8-step official sigma schedule,
#' no classifier-free guidance, joint audio-video denoising with an Euler
#' velocity step, and audio decoding through the audio VAE and BWE
#' vocoder to 48 kHz stereo.
#'
#' @name txt2vid_ltx23
NULL

#' Official distilled sigma schedule
#'
#' The distilled LTX sigma values (with terminal zero appended), as
#' published in the Apache-2.0 diffusers reference
#' (pipelines/ltx2/utils.py).
#'
#' @return Numeric vector of length 9.
#'
#' @export
ltx23_distilled_sigmas <- function() {
    c(1.0, 0.99375, 0.9875, 0.98125, 0.975, 0.909375, 0.725, 0.421875, 0.0)
}

#' Stage-2 distilled sigma schedule (two-stage refinement)
#'
#' @return Numeric vector of length 4.
#'
#' @export
ltx23_stage2_distilled_sigmas <- function() {
    c(0.909375, 0.725, 0.421875, 0.0)
}

# Pack video latents [B, C, F, H, W] -> [B, F*H*W, C] (patch sizes are 1
# for the LTX transformer; kept general per the reference)
ltx23_pack_video_latents <- function(latents, patch_size = 1L,
                                     patch_size_t = 1L) {
    d <- latents$shape
    pf <- d[3] %/% patch_size_t
    ph <- d[4] %/% patch_size
    pw <- d[5] %/% patch_size
    latents <- latents$reshape(c(d[1], -1L, pf, patch_size_t, ph,
                                 patch_size, pw, patch_size))
    # [B, C, F', pt, H', p, W', p] -> [B, F', H', W', C, pt, p, p]
    latents <- latents$permute(c(1L, 3L, 5L, 7L, 2L, 4L, 6L, 8L))
    latents$flatten(start_dim = 5L, end_dim = 8L)$flatten(start_dim = 2L, end_dim = 4L)
}

# Unpack video latents [B, S, D] -> [B, C, F, H, W]
ltx23_unpack_video_latents <- function(latents, num_frames, height, width,
                                       patch_size = 1L, patch_size_t = 1L) {
    b <- latents$shape[1]
    latents <- latents$reshape(c(b, num_frames, height, width, -1L,
                                 patch_size_t, patch_size, patch_size))
    latents <- latents$permute(c(1L, 5L, 2L, 6L, 3L, 7L, 4L, 8L))
    latents$flatten(start_dim = 7L, end_dim = 8L)$
    flatten(start_dim = 5L, end_dim = 6L)$
    flatten(start_dim = 3L, end_dim = 4L)
}

# Pack audio latents [B, C, L, M] -> [B, L, C * M]
ltx23_pack_audio_latents <- function(latents) {
    latents$transpose(2L, 3L)$flatten(start_dim = 3L, end_dim = 4L)
}

# Unpack audio latents [B, L, C * M] -> [B, C, L, M]
ltx23_unpack_audio_latents <- function(latents, num_mel_bins) {
    latents$unflatten(3L, c(-1L, num_mel_bins))$transpose(2L, 3L)
}

# Audio statistics apply to the PACKED [B, L, C*M] representation
# (broadcast over the trailing feature dim)
.ltx23_denormalize_audio <- function(latents, latents_mean, latents_std) {
    mean <- latents_mean$to(device = latents$device, dtype = latents$dtype)
    std <- latents_std$to(device = latents$device, dtype = latents$dtype)
    latents * std + mean
}

# Joint audio/video Euler denoise loop over a sigma schedule (CFG-free).
# With a conditioning_mask [B, S] (prefix conditioning), conditioned
# video tokens see a per-token timestep of zero and are frozen through
# the Euler updates (reference i2v semantics at strength 1).
.ltx23_denoise <- function(transformer, latents, audio_latents, sigmas,
                           video_text_embeds, audio_text_embeds, text_mask,
                           latent_frames, latent_height, latent_width,
                           audio_num_frames, frame_rate, device,
                           compute_dtype, verbose = TRUE, stage = "",
                           conditioning_mask = NULL,
                           audio_conditioned = FALSE) {
    f32 <- torch::torch_float32()
    keep_mask <- if (!is.null(conditioning_mask)) {
        conditioning_mask$unsqueeze(3L) # [B, S, 1] for latent blending
    } else {
        NULL
    }
    t_zero <- if (audio_conditioned) {
        torch::torch_tensor(0, device = device, dtype = f32)
    } else {
        NULL
    }
    # Compact per-token timestep: the conditioned prefix sees exactly
    # two values (t and 0), so the transformer gets a 2-entry timestep
    # plus a per-token selector instead of a [B, S] vector -- the
    # modulation tensors stay [B, 2, ...] until selected on demand
    cond_idx <- if (!is.null(conditioning_mask)) {
        conditioning_mask$squeeze(1L)$to(dtype = torch::torch_long())
    } else {
        NULL
    }
    video_coords <- transformer$rope$prepare_video_coords(latents$shape[1],
        latent_frames, latent_height, latent_width, device,
        fps = frame_rate)
    audio_coords <- transformer$audio_rope$prepare_audio_coords(
        audio_latents$shape[1], audio_num_frames, device
    )

    n_steps <- length(sigmas) - 1L
    scale_mult <- transformer$timestep_scale_multiplier
    pb <- .denoise_progress(
                            n_steps,
        if (nzchar(stage)) {
            sprintf("%sdenoise: %d steps", stage, n_steps)
        } else {
            NULL
        },
                            verbose, per_step = TRUE
    )
    step_t0 <- Sys.time()
    torch::with_no_grad({
        for (i in seq_len(n_steps)) {
            sigma <- sigmas[i]
            t <- torch::torch_tensor(sigma * scale_mult, device = device, dtype = f32)
            t_video <- if (is.null(conditioning_mask)) {
                t
            } else {
                # [t, 0]: index 0 = free tokens, index 1 = conditioned
                torch::torch_tensor(c(sigma * scale_mult, 0),
                                    device = device, dtype = f32)
            }

            out <- transformer(
                               hidden_states = latents$to(dtype = compute_dtype),
                               audio_hidden_states = audio_latents$to(dtype = compute_dtype),
                               encoder_hidden_states = video_text_embeds,
                               audio_encoder_hidden_states = audio_text_embeds,
                               timestep = t_video,
                               audio_timestep = if (audio_conditioned) t_zero else t,
                               sigma = t,
                               audio_sigma = if (audio_conditioned) t_zero else t,
                               encoder_attention_mask = text_mask,
                               audio_encoder_attention_mask = text_mask,
                               num_frames = latent_frames,
                               height = latent_height,
                               width = latent_width,
                               fps = frame_rate,
                               audio_num_frames = audio_num_frames,
                               video_coords = video_coords,
                               audio_coords = audio_coords,
                               use_cross_timestep = TRUE,
                               cond_token_index = cond_idx
            )

            # Euler velocity step in float32; dt is negative (sigma decreasing)
            dt <- torch::torch_tensor(sigmas[i + 1L] - sigma, device = device, dtype = f32)
            stepped <- latents + dt * out$sample$to(dtype = f32)
            latents <- if (is.null(keep_mask)) {
                stepped
            } else {
                # Conditioned tokens stay clean
                latents * keep_mask + stepped * (1 - keep_mask)
            }
            if (!audio_conditioned) {
                audio_latents <- audio_latents + dt * out$audio_sample$to(dtype = f32)
            }
            rm(out)
            gc(verbose = FALSE)
            pb$tick(i, sprintf("%s(sigma %.4f, %.1fs)", stage, sigma,
                               as.numeric(difftime(Sys.time(), step_t0, units = "secs"))))
            step_t0 <- Sys.time()
        }
    })
    pb$done()
    list(latents = latents, audio_latents = audio_latents)
}

#' Load the LTX-2.3 generation components from a single-file checkpoint
#'
#' Builds the transformer, connectors, video VAE, audio VAE, and vocoder
#' with the LTX 2.3 configuration and streams the checkpoint weights into
#' them. The Gemma3 text encoder ships separately (see
#' \code{\link{load_gemma3_text_encoder}}).
#'
#' @param checkpoint_path Path to the single-file checkpoint (e.g.
#'   \code{ltx-2.3-22b-distilled-1.1.safetensors}) or to an fp8 artifact
#'   directory produced by \code{\link{ltx23_quantize_fp8}}. With the fp8
#'   artifact, the transformer loads with CPU-resident fp8 weights that
#'   stream to \code{device} during the forward pass.
#' @param device Character. Device for the small components (VAEs,
#'   vocoder, connectors) and, with fp8, the transformer residents.
#' @param dtype Character. "bfloat16" (checkpoint native) or "float32".
#' @param transformer_device Character. Device for the transformer
#'   weights when loading the plain (non-fp8) checkpoint.
#' @param components Character vector. Which components to load.
#' @param pin Logical. Pin fp8 host memory (fp8 artifact only).
#' @param attn_chunk Integer or NULL. Query-chunk size for attention
#'   (see \code{\link{ltx23_set_attn_chunk}}).
#' @param phase_offload Logical. Load the small components (connectors,
#'   VAEs, vocoder) to the CPU; the pipeline moves each onto the compute
#'   device only for its phase.
#' @param verbose Logical.
#'
#' @return A list with the loaded modules and the checkpoint config,
#'   class \code{ltx23_pipeline}.
#'
#' @export
ltx23_load_pipeline <- function(checkpoint_path, device = "cuda",
                                dtype = "bfloat16",
                                transformer_device = "cpu",
                                components = c("dit", "connectors", "vae", "audio_vae", "vocoder"),
                                pin = TRUE, attn_chunk = NULL,
                                phase_offload = TRUE, verbose = TRUE) {
    if (phase_offload) {
        component_device <- "cpu"
    } else {
        component_device <- device
    }
    fp8 <- dir.exists(checkpoint_path)
    ckpt <- if (fp8) {
        ltx23_open_fp8_checkpoint(checkpoint_path)
    } else {
        ltx23_open_checkpoint(checkpoint_path)
    }
    if (!nzchar(Sys.getenv("PYTORCH_CUDA_ALLOC_CONF"))) {
        # Must be set before the first CUDA allocation. NF4 runs the
        # compiled block stack whose intermediates never become R
        # handles: the native backend is faster there (expandable
        # frees are page-unmaps, which is what made the decode gc
        # storm expensive) and its 1280x704 fit is validated. The
        # fp8/eager paths were only ever validated under
        # expandable_segments (fragmentation from eager handle churn),
        # so they keep it.
        conf <- if (identical(ckpt$format, "nf4")) {
            "backend:native"
        } else {
            "expandable_segments:True"
        }
        Sys.setenv(PYTORCH_CUDA_ALLOC_CONF = conf)
    }
    if (device == "cuda") {
        # Stop the allocator gc storm before the first CUDA op: the
        # decode phases measured 86% of wall time in callback-driven
        # R gc at the 0.20 default reserved rate. Only-if-unset, so an
        # explicit user option wins.
        if (identical(ckpt$format, "nf4")) {
            footprint <- 12
        } else {
            footprint <- 8
        }
        ltx23_tune_gc(footprint_gb = footprint)
        # Pre-warm the caching allocator with one large allocation
        # freed straight into the pool: the phase onloads then carve
        # from cached blocks instead of growing the pool one
        # cudaMalloc per tensor (83.5s -> 4.6s for the NF4
        # transformer's first onload, measured). This is also the
        # first CUDA op, so it runs after the options above are set.
        tryCatch({
            warm <- torch::torch_empty(as.integer(footprint + 1) * 1e9,
                                       dtype = torch::torch_uint8(),
                                       device = "cuda")
            rm(warm)
            gc(verbose = FALSE)
        }, error = function(e) invisible(NULL))
    }
    groups <- ltx23_split_keys(ckpt$keys)
    torch_dtype <- switch(dtype, bfloat16 = torch::torch_bfloat16(),
                          float16 = torch::torch_float16(),
                          float32 = torch::torch_float32(),
                          stop("Unsupported dtype: ", dtype))

    pipe <- list(config = ckpt$config, checkpoint_path = ckpt$path)

    load_component <- function(name, module, map_key, dev) {
        if (verbose) {
            message("Loading ", name, " (", length(groups[[name]]), " tensors)...")
        }
        # Modules arrive as skeletons already at torch_dtype
        res <- ltx23_load_group(ckpt, groups[[name]], module,
                                map_key = map_key, verbose = verbose)
        if (length(res$unmapped) || length(res$unfilled)) {
            stop(name, ": incomplete load (", length(res$unmapped), " unmapped, ",
                 length(res$unfilled), " unfilled)")
        }
        module$to(device = dev)
        module$eval()
        module
    }

    if ("dit" %in% components) {
        if (fp8 && identical(ckpt$format, "nf4")) {
            pipe$transformer <- ltx23_load_transformer_nf4(
                ckpt, device = component_device, verbose = verbose
            )
        } else if (fp8) {
            pipe$transformer <- ltx23_load_transformer_fp8(
                ckpt, device = component_device, pin = pin, verbose = verbose
            )
        } else {
            pipe$transformer <- load_component(
                "dit", .construct_skeleton(ltx23_transformer, dtype = torch_dtype),
                ltx23_map_dit_key, transformer_device
            )
        }
        if (!is.null(attn_chunk)) {
            ltx23_set_attn_chunk(pipe$transformer, as.integer(attn_chunk))
        }
    }
    if ("connectors" %in% components) {
        pipe$connectors <- load_component(
            "connectors", .construct_skeleton(ltx23_text_connectors, dtype = torch_dtype),
            ltx23_map_connector_key, component_device
        )
    }
    if ("vae" %in% components) {
        pipe$vae <- load_component(
                                   "vae", .construct_skeleton(ltx23_video_vae, dtype = torch_dtype),
                                   ltx23_map_vae_key, component_device
        )
        pipe$vae$enable_tiling()
    }
    if ("audio_vae" %in% components) {
        pipe$audio_vae <- load_component(
            "audio_vae", .construct_skeleton(ltx23_audio_vae, dtype = torch_dtype),
            ltx23_map_audio_vae_key, component_device
        )
    }
    if ("vocoder" %in% components) {
        # The vocoder is small and precision-sensitive: run in float32
        voc <- ltx23_vocoder_with_bwe()
        voc$to(dtype = torch_dtype)
        res <- ltx23_load_group(ckpt, groups$vocoder, voc,
                                map_key = ltx23_map_vocoder_key, verbose = verbose)
        if (length(res$unmapped) || length(res$unfilled)) {
            stop("vocoder: incomplete load")
        }
        voc$to(device = component_device, dtype = torch::torch_float32())
        voc$eval()
        pipe$vocoder <- voc
    }

    if (phase_offload && device == "cuda" &&
        isTRUE(getOption("diffuseR.pin_staging", TRUE))) {
        # Page-lock every phase-offloaded component once so the
        # per-render CPU<->GPU moves run at full PCIe rate (offload
        # becomes a pointer swap; see staging.R). Falls back
        # silently per component if page-locking fails.
        if (verbose) {
            message("Pinning host staging buffers...")
        }
        staging <- list()
        for (nm in intersect(c("transformer", "connectors", "vae",
                               "audio_vae", "vocoder"), names(pipe))) {
            st <- .pin_component(pipe[[nm]])
            if (!is.null(st)) {
                staging[[nm]] <- st
            }
        }
        if (length(staging)) {
            pipe$staging <- staging
        }
    }

    structure(pipe, class = "ltx23_pipeline")
}

#' Generate video (and audio) with LTX-2.3
#'
#' Distilled text-to-video generation: encodes the prompt with Gemma3 +
#' connectors, denoises joint audio/video latents over the official
#' 8-step distilled schedule (no classifier-free guidance), decodes the
#' video with the causal VAE and the audio with the audio VAE + BWE
#' vocoder, and optionally muxes both into an MP4.
#'
#' @param prompt Character. The prompt.
#' @param pipeline An \code{ltx23_pipeline} from
#'   \code{\link{ltx23_load_pipeline}}.
#' @param text_encoder,tokenizer Gemma3 model and tokenizer (or paths;
#'   see \code{\link{load_gemma3_text_encoder}} and
#'   \code{\link{gemma3_tokenizer}}). Ignored when \code{prompt_embeds}
#'   is supplied.
#' @param prompt_embeds Optional precomputed list with
#'   \code{prompt_embeds} (raw stacked Gemma3 states) and
#'   \code{prompt_attention_mask}; bypasses the text encoder.
#' @param connector_embeds Optional precomputed text-connector outputs:
#'   a list with \code{video_text_embedding},
#'   \code{audio_text_embedding}, and \code{attention_mask} (the result
#'   of \code{pipeline$connectors} on the Gemma3 states). The prompt is
#'   constant across a chained track, so compute this once and pass it
#'   to every chunk: it skips the per-call connectors phase, whose
#'   GPU-side handling of the raw hidden-state stack does not fit next
#'   to a resident transformer.
#' @param width,height Integers. Output resolution (multiples of 32).
#' @param num_frames Integer. 8k + 1 frames (e.g. 121).
#' @param frame_rate Numeric. Frames per second.
#' @param sigmas Numeric vector. Denoising schedule (default: official
#'   distilled schedule; must end in 0).
#' @param guidance_scale Numeric. Only 1 (no CFG) is supported; the
#'   distilled checkpoints are trained for CFG-free sampling.
#' @param seed Integer or NULL.
#' @param device Character. Compute device for the denoising loop.
#' @param dtype Character. Model compute dtype ("bfloat16" or "float32").
#' @param filename Character or NULL. Output video path (.mp4). Audio is
#'   muxed in when the av package is available.
#' @param max_sequence_length Integer. Text token length (multiple of 128).
#' @param decode_video,decode_audio Logicals. Decode the respective
#'   latents (disable for latent-space work).
#' @param two_stage Logical. Generate at half resolution, upsample the
#'   latents 2x spatially, and refine over the stage-2 schedule
#'   (resolution must then be a multiple of 64; requires
#'   \code{upsampler}).
#' @param upsampler An \code{\link{ltx23_latent_upsampler}} (see
#'   \code{\link{ltx23_load_upsampler}}).
#' @param adain_factor Numeric. AdaIN blend of the upsampled latents
#'   toward the stage-1 statistics (0 disables).
#' @param tone_map_compression Numeric in [0, 1]. Optional latent tone
#'   mapping before stage 2.
#' @param phase_offload Logical. Move each small component to the compute
#'   device only for its phase (text encoding, upsampling, decoding) and
#'   back to the CPU afterwards, keeping the denoise phase as the sole
#'   GPU tenant.
#' @param image Optional start image for image-to-video: a PNG/JPEG path
#'   or an [H, W, 3] array in [0, 1]. The image conditions the first
#'   frame; the rest of the video is generated (reference i2v).
#' @param condition_video Optional continuation source: a video path
#'   (its trailing \code{conditioning_frames} frames are used) or an
#'   [F, H, W, 3] array. The clip's tail becomes the frozen prefix of
#'   the new video, so the output's first \code{conditioning_frames}
#'   frames overlap the source (trim or crossfade when concatenating).
#' @param conditioning_frames Integer. Trailing pixel frames taken from
#'   \code{condition_video} (8k + 1, default 9 = 2 latent frames).
#' @param cond_noise_scale Numeric in [0, 1]. Optional partial noising
#'   of the conditioned tokens (0 = keep them exactly).
#' @param condition_latents Optional continuation source already in
#'   latent space: normalized video latents [1, 128, k, height/32,
#'   width/32] (e.g. \code{\link{ltx23_tail_latents}} on a previous
#'   result), used directly as the frozen prefix with no VAE encode.
#'   Mutually exclusive with \code{image} and \code{condition_video}.
#' @param resident Character vector of pipeline component names
#'   ("transformer", "vae", "audio_vae", "connectors", "vocoder") to
#'   keep on the compute device after their phase instead of
#'   offloading, for callers running several generations back to back
#'   (chained chunks). Components already on the device are not
#'   re-copied on later calls.
#' @param trim_frames Integer. Drop this many leading pixel frames
#'   from the decoded video (and the saved file), e.g. the
#'   conditioning-head overlap of a continuation. The returned
#'   \code{latents} keep the full sequence (tail slicing for chaining
#'   needs it). Audio is muxed exactly as supplied, so drop any head
#'   padding from the conditioning audio when trimming.
#' @param audio Optional conditioning audio for audio-driven generation
#'   (lip sync): a file path (decoded via \code{av}) or a matrix
#'   [2, samples] in [-1, 1] at 16 kHz. The audio is encoded into
#'   clean, frozen audio latents that the video attends to while
#'   denoising, and the original samples are muxed into the output
#'   (audio decoding is skipped).
#' @param verbose Logical, or one of "silent", "progress", "steps".
#'   TRUE = "steps" (full per-phase chatter, per-step sigma/timing
#'   lines), FALSE = "silent". "progress" prints a one-line generation
#'   summary plus a denoise progress bar (interactive) or periodic
#'   step ticks (captured logs).
#'
#' @return Invisibly, a list with \code{video} (array
#'   [frames, height, width, 3] in [0, 1]), \code{audio} (matrix
#'   [2, samples] in [-1, 1]), \code{sample_rate}, the raw
#'   \code{latents} and \code{audio_latents}, and \code{latent_shape}
#'   (c(frames, height, width) of the latent geometry, for
#'   \code{\link{ltx23_tail_latents}}).
#'
#' @export
txt2vid_ltx2 <- function(prompt, pipeline, text_encoder = NULL,
                         tokenizer = NULL, prompt_embeds = NULL,
                         connector_embeds = NULL, width = 768L,
                         height = 512L, num_frames = 121L, frame_rate = 24,
                         sigmas = ltx23_distilled_sigmas(),
                         guidance_scale = 1, seed = NULL, device = "cuda",
                         dtype = "bfloat16", filename = NULL,
                         max_sequence_length = 1024L, decode_video = TRUE,
                         decode_audio = TRUE, two_stage = FALSE,
                         upsampler = NULL, adain_factor = 1.0,
                         tone_map_compression = 0, phase_offload = TRUE,
                         image = NULL, condition_video = NULL,
                         conditioning_frames = 9L, cond_noise_scale = 0,
                         condition_latents = NULL, resident = character(),
                         trim_frames = 0L, audio = NULL, verbose = TRUE) {
    level <- .verbosity(verbose)
    verbose <- level == "steps"
    if (level != "silent") {
        message(sprintf("LTX-2.3: %d frames @ %dx%d, %d steps%s", num_frames,
                        width, height, length(sigmas) - 1L,
                if (two_stage) " (two-stage)" else ""))
    }
    stopifnot(inherits(pipeline, "ltx23_pipeline"))
    if (guidance_scale != 1) {
        stop("Only guidance_scale = 1 is supported (distilled checkpoints).")
    }
    n_sources <- sum(!is.null(image), !is.null(condition_video),
                     !is.null(condition_latents))
    if (n_sources > 1L) {
        stop("Provide only one of image (i2v), condition_video, or condition_latents.")
    }
    conditioned <- n_sources > 0L
    known_components <- c("transformer", "vae", "audio_vae", "connectors",
                          "vocoder")
    if (length(resident) && !all(resident %in% known_components)) {
        stop("resident must name pipeline components: ",
             paste(known_components, collapse = ", "))
    }
    trim_frames <- as.integer(trim_frames)
    if (trim_frames < 0L || trim_frames >= num_frames) {
        stop("trim_frames must be in [0, num_frames)")
    }
    if (conditioned && two_stage) {
        stop("Prefix conditioning with two_stage = TRUE is not supported yet.")
    }
    if (two_stage) {
        if (is.null(upsampler)) {
            stop("two_stage = TRUE requires an upsampler (see ltx23_load_upsampler).")
        }
        if (width %% 64L != 0L || height %% 64L != 0L) {
            stop("width and height must be multiples of 64 for two-stage generation")
        }
    }
    if (width %% 32L != 0L || height %% 32L != 0L) {
        stop("width and height must be multiples of 32")
    }
    if ((num_frames - 1L) %% 8L != 0L) {
        stop("num_frames must be 8k + 1 (e.g. 121)")
    }
    if (utils::tail(sigmas, 1) != 0) {
        stop("sigmas must end with 0")
    }
    if (!is.null(seed)) {
        torch::torch_manual_seed(seed)
    }

    compute_dtype <- switch(dtype, bfloat16 = torch::torch_bfloat16(),
                            float16 = torch::torch_float16(),
                            float32 = torch::torch_float32(),
                            stop("Unsupported dtype: ", dtype))
    f32 <- torch::torch_float32()

    # --- Phase 1: text encoding -------------------------------------------------
    if (is.null(prompt_embeds) && is.null(connector_embeds)) {
        if (is.null(text_encoder) || is.null(tokenizer)) {
            stop("Supply text_encoder + tokenizer, or precomputed prompt_embeds.")
        }
        if (verbose) {
            message("Encoding prompt...")
        }
        prompt_embeds <- encode_with_gemma3(
            prompt,
            model = text_encoder, tokenizer = tokenizer,
            max_sequence_length = max_sequence_length,
            device = if (is.character(text_encoder)) device else "cpu",
            verbose = verbose
        )
    }

    # Each phase is the sole GPU tenant: components move on for their
    # phase and back off afterwards. Pipeline components are referred
    # to by name so pinned staging (see staging.R) can be used
    # when the loader prepared it; plain modules (the upsampler) take
    # the pageable path.
    phase_offload <- phase_offload && device != "cpu"
    staging <- pipeline$staging %||% list()
    target_type <- torch::torch_device(device)$type
    onload <- function(what) {
        if (is.character(what)) {
            module <- pipeline[[what]]
        } else {
            module <- what
        }
        if (phase_offload) {
            # Idempotent: a resident component is already in place on
            # the second and later calls of a chained run. Probe the
            # staging pair's live tensor when staging exists - custom
            # module classes (the NF4 transformer) may not expose
            # $parameters, and a failed probe must not degrade into a
            # re-onload: re-transferring resident weights over
            # themselves fragments the allocator pool until the next
            # large allocation OOMs.
            if (is.character(what)) {
                st_probe <- staging[[what]]
            } else {
                st_probe <- NULL
            }
            cur <- tryCatch({
                if (!is.null(st_probe)) {
                    st_probe[[1]]$live$device$type
                } else {
                    module$parameters[[1]]$device$type
                }
            }, error = function(e) NULL)
            if (identical(cur, target_type)) {
                return(module)
            }
            if (is.character(what)) {
                st <- staging[[what]]
            } else {
                st <- NULL
            }
            if (is.null(st)) {
                module$to(device = device)
            } else {
                .staged_onload(st, device)
            }
        }
        module
    }
    offload <- function(what) {
        if (is.character(what)) {
            module <- pipeline[[what]]
        } else {
            module <- what
        }
        if (phase_offload && !(is.character(what) && what %in% resident)) {
            # Decode traces capture weight tensors; drop them so the
            # module's GPU memory actually frees
            .ltx23_release_vae_traces()
            if (is.character(what)) {
                st <- staging[[what]]
            } else {
                st <- NULL
            }
            if (is.null(st)) {
                module$to(device = "cpu")
            } else {
                .staged_offload(st)
            }
            # gc only -- NO cuda_empty_cache between phases: returning
            # blocks to the driver forces the next phase to regrow the
            # pool through cudaMalloc at ~15ms per allocation (~83s
            # for the transformer's 5530 tensors, measured); the
            # caching allocator reuses the freed blocks directly
            gc(verbose = FALSE)
        }
        invisible(module)
    }

    if (!is.null(connector_embeds)) {
        # Precomputed connector outputs (constant per prompt): skip the
        # per-call connectors phase entirely. Chained generation with a
        # resident transformer needs this - the raw Gemma3 hidden-state
        # stack moving to the GPU next to 11 GB of parked weights is
        # what OOMed every resident chunk (geometry-independent 14.97
        # GiB signature).
        video_text_embeds <- connector_embeds$video_text_embedding$
        to(device = device, dtype = compute_dtype)
        audio_text_embeds <- connector_embeds$audio_text_embedding$
        to(device = device, dtype = compute_dtype)
        text_mask <- connector_embeds$attention_mask$to(device = device)
    } else {
        onload("connectors")
        connectors_device <- pipeline$connectors$video_text_proj_in$weight$device
        torch::with_no_grad({
            conn <- pipeline$connectors(
                                        prompt_embeds$prompt_embeds$to(device = connectors_device, dtype = compute_dtype),
                                        prompt_embeds$prompt_attention_mask$to(device = connectors_device)
            )
            video_text_embeds <- conn$video_text_embedding$to(device = device)
            audio_text_embeds <- conn$audio_text_embedding$to(device = device)
            text_mask <- conn$attention_mask$to(device = device)
        })
        rm(conn)
        offload("connectors")
    }
    gc(verbose = FALSE)

    # --- Phase 2: latent preparation ---------------------------------------------
    latent_frames <- (num_frames - 1L) %/% 8L + 1L
    latent_height <- height %/% 32L
    latent_width <- width %/% 32L
    # Two-stage: stage 1 generates at half resolution
    if (two_stage) {
        s1_height <- latent_height %/% 2L
    } else {
        s1_height <- latent_height
    }
    if (two_stage) {
        s1_width <- latent_width %/% 2L
    } else {
        s1_width <- latent_width
    }

    # 25 audio latent frames per second: sampling_rate / hop / downsample
    audio_num_frames <- as.integer(round(num_frames / frame_rate * 25))
    latent_mel_bins <- 16L # 64 mel bins / 4

    # Prefix conditioning: encode the start image (i2v) or the tail of
    # a previous clip (continuation) into normalized latents
    cond_latents <- NULL
    if (!is.null(condition_latents)) {
        # Already-encoded prefix (e.g. ltx23_tail_latents of a previous
        # chunk): no VAE touch at all
        cond_latents <- condition_latents$to(device = device, dtype = f32)
        if (cond_latents$ndim != 5L || cond_latents$shape[2] != 128L ||
            cond_latents$shape[4] != s1_height ||
            cond_latents$shape[5] != s1_width) {
            stop(sprintf(
                         "condition_latents must be [1, 128, k, %d, %d] for %dx%d generation",
                         s1_height, s1_width, width, height))
        }
    } else if (conditioned) {
        vae <- onload("vae")
        cond_frames <- if (!is.null(image)) {
            ltx23_preprocess_frames(image, height, width)
        } else {
            arr <- if (is.character(condition_video)) {
                ltx23_read_tail_frames(condition_video, conditioning_frames)
            } else {
                condition_video
            }
            ltx23_preprocess_frames(arr, height, width)
        }
        if (verbose) {
            message(sprintf("Encoding %d conditioning frame(s)...",
                            cond_frames$shape[3]))
        }
        cond_latents <- ltx23_encode_video_frames(vae, cond_frames)$
        to(device = device)
        offload("vae")
        gc(verbose = FALSE)
    }

    noise <- torch::torch_randn(
                                c(1L, 128L, latent_frames, s1_height, s1_width),
                                device = device, dtype = f32
    )
    conditioning_mask <- NULL
    if (conditioned) {
        prep <- ltx23_prepare_conditioned_latents(
            cond_latents, latent_frames, s1_height, s1_width,
            noise, cond_noise_scale = cond_noise_scale
        )
        latents <- prep$latents
        conditioning_mask <- prep$conditioning_mask
        rm(prep, cond_latents)
    } else {
        latents <- ltx23_pack_video_latents(noise)
    }
    rm(noise)

    audio_conditioned <- !is.null(audio)
    input_audio <- NULL
    if (audio_conditioned) {
        # Audio-driven generation: the user audio becomes clean,
        # frozen audio latents; the video denoises while attending to
        # them (lip sync). The output carries the original audio.
        input_audio <- if (is.character(audio)) {
            ltx23_read_audio(audio)
        } else {
            audio
        }
        audio_vae <- onload("audio_vae")
        if (verbose) {
            message("Encoding conditioning audio...")
        }
        audio_latents <- ltx23_encode_audio(audio_vae, input_audio,
            audio_num_frames)$
        to(device = device)
        offload("audio_vae")
        gc(verbose = FALSE)
        decode_audio <- FALSE
    } else {
        audio_latents <- torch::torch_randn(
            c(1L, 8L, audio_num_frames, latent_mel_bins),
            device = device, dtype = f32
        )
        audio_latents <- ltx23_pack_audio_latents(audio_latents)
    }

    # --- Phase 3: denoising -------------------------------------------------------
    transformer <- onload("transformer")
    if (verbose) message(sprintf("Denoising: %d steps at %dx%dx%d...",
                                 length(sigmas) - 1L, width %/% (if (two_stage) 2L else 1L),
                                 height %/% (if (two_stage) 2L else 1L), num_frames))

    denoised <- .ltx23_denoise(
                               transformer, latents, audio_latents, sigmas,
                               video_text_embeds, audio_text_embeds, text_mask,
                               latent_frames, s1_height, s1_width,
                               audio_num_frames, frame_rate,
                               device, compute_dtype, verbose = level,
                               stage = if (two_stage) "stage 1 " else "",
                               conditioning_mask = conditioning_mask,
                               audio_conditioned = audio_conditioned
    )
    latents <- denoised$latents
    audio_latents <- denoised$audio_latents

    if (two_stage) {
        vae <- pipeline$vae
        if (verbose) {
            message("Upsampling latents 2x...")
        }
        onload(upsampler)
        torch::with_no_grad({
            up_device <- upsampler$final_conv$weight$device
            up_dtype <- upsampler$final_conv$weight$dtype
            s1_latents <- ltx23_unpack_video_latents(
                latents, latent_frames, s1_height, s1_width
            )$to(device = up_device)
            # The upsampler operates on unnormalized latents
            s1_latents <- ltx23_denormalize_latents(
                s1_latents, vae$latents_mean, vae$latents_std
            )
            up_latents <- upsampler(s1_latents$to(dtype = up_dtype))$to(dtype = f32)
            if (adain_factor != 0) {
                up_latents <- ltx23_adain_filter_latent(
                    up_latents, s1_latents$to(dtype = f32), adain_factor
                )
            }
            if (tone_map_compression > 0) {
                up_latents <- ltx23_tone_map_latents(up_latents, tone_map_compression)
            }
            up_latents <- ltx23_normalize_latents(
                up_latents, vae$latents_mean, vae$latents_std
            )
            latents <- ltx23_pack_video_latents(up_latents$to(device = device))
            rm(s1_latents, up_latents)
        })
        offload(upsampler)
        gc(verbose = FALSE)

        # Re-noise BOTH modalities at the stage-2 entry sigma
        s2_sigmas <- ltx23_stage2_distilled_sigmas()
        noise_scale <- s2_sigmas[1]
        torch::with_no_grad({
            latents <- torch::torch_randn_like(latents)$mul(noise_scale) +
            latents$mul(1 - noise_scale)
            audio_latents <- torch::torch_randn_like(audio_latents)$mul(noise_scale) +
            audio_latents$mul(1 - noise_scale)
        })

        if (verbose) message(sprintf("Refining: %d steps at %dx%d...",
                                     length(s2_sigmas) - 1L, width, height))
        denoised <- .ltx23_denoise(
                                   transformer, latents, audio_latents, s2_sigmas,
                                   video_text_embeds, audio_text_embeds, text_mask,
                                   latent_frames, latent_height, latent_width,
                                   audio_num_frames, frame_rate,
                                   device, compute_dtype, verbose = level, stage = "stage 2 "
        )
        latents <- denoised$latents
        audio_latents <- denoised$audio_latents
    }

    # The transformer and its dequant buffers are not needed past
    # denoising; free the VRAM before the decoders claim it. A
    # resident transformer keeps its buffers for the next chunk.
    offload("transformer")
    if (!("transformer" %in% resident)) {
        ltx23_release_dequant_buffers()
    }

    result <- list(
                   latents = latents,
                   audio_latents = audio_latents,
                   latent_shape = c(latent_frames, latent_height, latent_width),
                   sample_rate = 48000L
    )
    if (audio_conditioned) {
        # The conditioning audio IS the soundtrack: carry the original
        # samples through to the result and the mux
        result$audio <- as.matrix(input_audio)
        result$sample_rate <- 16000L
    }

    # --- Phase 4: decoding -----------------------------------------------------------
    if (decode_video) {
        if (verbose) {
            message("Decoding video...")
        }
        phase_t0 <- Sys.time()
        vae <- onload("vae")
        vae_device <- vae$latents_mean$device
        vae_dtype <- vae$decoder$conv_in$conv$weight$dtype
        torch::with_no_grad({
            video_latents <- ltx23_unpack_video_latents(
                latents, latent_frames, latent_height, latent_width
            )$to(device = vae_device)
            video_latents <- ltx23_denormalize_latents(
                video_latents, vae$latents_mean, vae$latents_std
            )
            video <- vae$decode(video_latents$to(dtype = vae_dtype))
            # [-1, 1] -> [0, 1], [B, 3, F, H, W] -> [F, H, W, 3]
            video <- ((video$to(dtype = f32) / 2 + 0.5)$clamp(0, 1))[1,,,,]
            video <- video$permute(c(2L, 3L, 4L, 1L))$cpu()
        })
        if (verbose) {
            message(sprintf("  video decode: %.1fs",
                            as.numeric(difftime(Sys.time(), phase_t0, units = "secs"))))
            phase_t0 <- Sys.time()
        }
        result$video <- as.array(video)
        rm(video, video_latents)
        offload("vae")
        gc(verbose = FALSE)
        if (verbose) {
            message(sprintf("  video to R array: %.1fs",
                            as.numeric(difftime(Sys.time(), phase_t0, units = "secs"))))
        }
        if (trim_frames > 0L) {
            # Deliver head-free pixels (the conditioning overlap of a
            # continuation); result$latents keeps the full sequence
            result$video <- result$video[-seq_len(trim_frames),,,,
                drop = FALSE]
        }
    }

    if (decode_audio) {
        if (verbose) {
            message("Decoding audio...")
        }
        phase_t0 <- Sys.time()
        audio_vae <- onload("audio_vae")
        onload("vocoder")
        av_device <- audio_vae$latents_mean$device
        av_dtype <- audio_vae$decoder$conv_in$conv$weight$dtype
        torch::with_no_grad({
            audio_packed <- .ltx23_denormalize_audio(
                audio_latents$to(device = av_device),
                audio_vae$latents_mean, audio_vae$latents_std
            )
            audio_lat <- ltx23_unpack_audio_latents(audio_packed, latent_mel_bins)
            mel <- audio_vae$decode(audio_lat$to(dtype = av_dtype))
            waveform <- .ltx23_traced_call(
                pipeline$vocoder,
                mel$to(dtype = torch::torch_float32())
            )
            waveform <- waveform[1,,]$cpu()
        })
        result$audio <- as.matrix(as.array(waveform))
        rm(waveform, mel, audio_lat, audio_packed)
        offload("audio_vae")
        offload("vocoder")
        gc(verbose = FALSE)
        if (verbose) {
            message(sprintf("  audio chain: %.1fs",
                            as.numeric(difftime(Sys.time(), phase_t0, units = "secs"))))
        }
    }

    if (!is.null(filename) && decode_video) {
        phase_t0 <- Sys.time()
        # [[ ]] not $: with decode_audio = FALSE there is no result$audio
        # and partial matching silently hands write_wav the raw
        # audio_latents tensor
        save_video_ltx23(result[["video"]], filename,
                         fps = frame_rate,
                         audio = result[["audio"]],
                         sample_rate = result[["sample_rate"]],
                         verbose = verbose
        )
        result$filename <- filename
        if (verbose) {
            message(sprintf("  encode + mux: %.1fs",
                            as.numeric(difftime(Sys.time(), phase_t0, units = "secs"))))
        }
    }

    invisible(result)
}

#' Write a 16-bit PCM WAV file
#'
#' Minimal RIFF writer in base R.
#'
#' @param audio Numeric matrix [channels, samples] in [-1, 1].
#' @param path Output path.
#' @param sample_rate Integer.
#'
#' @return Invisibly, the path.
#'
#' @export
write_wav <- function(audio, path, sample_rate = 48000L) {
    if (is.null(dim(audio))) {
        audio <- matrix(audio, nrow = 1L)
    }
    n_channels <- nrow(audio)
    n_samples <- ncol(audio)

    # Interleave channels, scale to int16
    pcm <- as.integer(round(pmax(pmin(as.vector(audio), 1), -1) * 32767))
    byte_rate <- sample_rate * n_channels * 2L
    data_size <- n_samples * n_channels * 2L

    con <- file(path, "wb")
    on.exit(close(con), add = TRUE)
    writeChar("RIFF", con, eos = NULL)
    writeBin(as.integer(36L + data_size), con, size = 4, endian = "little")
    writeChar("WAVEfmt ", con, eos = NULL)
    writeBin(16L, con, size = 4, endian = "little")
    writeBin(1L, con, size = 2, endian = "little") # PCM
    writeBin(n_channels, con, size = 2, endian = "little")
    writeBin(as.integer(sample_rate), con, size = 4, endian = "little")
    writeBin(byte_rate, con, size = 4, endian = "little")
    writeBin(n_channels * 2L, con, size = 2, endian = "little") # block align
    writeBin(16L, con, size = 2, endian = "little") # bits per sample
    writeChar("data", con, eos = NULL)
    writeBin(data_size, con, size = 4, endian = "little")
    writeBin(pcm, con, size = 2, endian = "little")
    invisible(path)
}

#' Save an LTX video (optionally with audio) to MP4
#'
#' Uses the av package (Suggests) to encode frames and mux the audio
#' track.
#'
#' @param video Array [frames, height, width, 3] in [0, 1].
#' @param filename Output path (.mp4).
#' @param fps Numeric.
#' @param audio Optional numeric matrix [channels, samples] in [-1, 1].
#' @param sample_rate Integer.
#' @param verbose Logical.
#'
#' @return Invisibly, the filename.
#'
#' @export
save_video_ltx23 <- function(video, filename, fps = 24, audio = NULL,
                             sample_rate = 48000L, verbose = TRUE) {
    if (!requireNamespace("av", quietly = TRUE)) {
        stop("The av package is required to write video files.")
    }
    tmp_dir <- tempfile("ltx23_frames_")
    dir.create(tmp_dir)
    on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

    n_frames <- dim(video)[1]
    frame_paths <- character(n_frames)
    for (i in seq_len(n_frames)) {
        frame_paths[i] <- file.path(tmp_dir, sprintf("frame_%05d.png", i))
        png::writePNG(video[i,,,], frame_paths[i])
    }

    audio_path <- NULL
    if (!is.null(audio)) {
        audio_path <- file.path(tmp_dir, "audio.wav")
        write_wav(audio, audio_path, sample_rate = sample_rate)
    }

    av::av_encode_video(frame_paths, output = filename, framerate = fps,
                        audio = audio_path, verbose = verbose)
    if (verbose) {
        message("Wrote ", filename)
    }
    invisible(filename)
}

# End-to-end plumbing test for the LTX-2.3 pipeline with tiny
# random-weight components (no model downloads; CPU-safe).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)

torch::torch_manual_seed(99)

# Tiny components wired to the pipeline's structural constants:
# 128 video latent channels, 8 audio channels x 16 latent mel bins.
transformer <- ltx23_transformer(
  in_channels = 128L, out_channels = 128L,
  num_attention_heads = 2L, attention_head_dim = 8L,
  cross_attention_dim = 16L,
  audio_in_channels = 128L, audio_out_channels = 128L,
  audio_num_attention_heads = 2L, audio_attention_head_dim = 4L,
  audio_cross_attention_dim = 8L,
  num_layers = 1L
)
transformer$eval()

connectors <- ltx23_text_connectors(
  caption_channels = 8L, text_proj_in_factor = 3L,
  video_connector_num_attention_heads = 2L,
  video_connector_attention_head_dim = 8L,
  video_connector_num_layers = 1L,
  video_connector_num_learnable_registers = 4L,
  audio_connector_num_attention_heads = 2L,
  audio_connector_attention_head_dim = 4L,
  audio_connector_num_layers = 1L,
  audio_connector_num_learnable_registers = 4L,
  video_hidden_dim = 16L, audio_hidden_dim = 8L
)
connectors$eval()

vae <- ltx23_video_vae(
  latent_channels = 128L,
  block_out_channels = c(8L, 8L, 8L, 8L),
  decoder_block_out_channels = c(4L, 8L, 8L, 16L),
  layers_per_block = c(1L, 1L, 1L, 1L, 1L),
  decoder_layers_per_block = c(1L, 1L, 1L, 1L, 1L)
)
vae$eval()

audio_vae <- ltx23_audio_vae(
  base_channels = 128L, ch_mult = c(1L, 1L, 1L), num_res_blocks = 1L,
  latent_channels = 8L, mel_bins = 64L
)
audio_vae$eval()

vocoder <- ltx23_vocoder_with_bwe(
  in_channels = 128L, hidden_channels = 16L,
  upsample_kernel_sizes = c(4L, 4L), upsample_factors = c(2L, 2L),
  resnet_kernel_sizes = c(3L), resnet_dilations = list(c(1L, 3L)),
  bwe_in_channels = 128L, bwe_hidden_channels = 8L,
  bwe_upsample_kernel_sizes = c(8L, 4L), bwe_upsample_factors = c(4L, 2L),
  bwe_resnet_kernel_sizes = c(3L), bwe_resnet_dilations = list(c(1L, 3L)),
  filter_length = 8L, hop_length = 2L, window_length = 8L,
  num_mel_channels = 64L,
  input_sampling_rate = 100L, output_sampling_rate = 400L
)
vocoder$eval()

pipe <- structure(
  list(
    transformer = transformer,
    connectors = connectors,
    vae = vae,
    audio_vae = audio_vae,
    vocoder = vocoder
  ),
  class = "ltx23_pipeline"
)

stub_embeds <- list(
  prompt_embeds = torch::torch_randn(1L, 8L, 8L, 3L),
  prompt_attention_mask = torch::torch_ones(1L, 8L)
)

res <- txt2vid_ltx2(
  prompt = "a tiny test",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 9L, frame_rate = 24,
  seed = 7L,
  device = "cpu",
  dtype = "float32",
  verbose = FALSE
)

# Video: [frames, height, width, 3] in [0, 1]
expect_equal(dim(res$video), c(9L, 64L, 64L, 3L))
expect_true(all(res$video >= 0 & res$video <= 1))
expect_true(all(is.finite(res$video)))

# Audio: [2, samples] in [-1, 1]; latent L = round(9/24*25) = 9 ->
# mel frames 4*9-3 = 33 (padded to hop multiple 34) -> vocoder x4 -> BWE x4
expect_equal(nrow(res$audio), 2L)
expect_true(ncol(res$audio) > 100L)
expect_true(all(abs(res$audio) <= 1))
expect_true(all(is.finite(res$audio)))
expect_equal(res$sample_rate, 48000L)

# Guardrails
expect_error(
  txt2vid_ltx2("x", pipe, prompt_embeds = stub_embeds, guidance_scale = 3,
    device = "cpu", dtype = "float32"),
  pattern = "guidance_scale"
)
expect_error(
  txt2vid_ltx2("x", pipe, prompt_embeds = stub_embeds, width = 50L,
    device = "cpu", dtype = "float32"),
  pattern = "multiples of 32"
)
expect_error(
  txt2vid_ltx2("x", pipe, prompt_embeds = stub_embeds, num_frames = 10L,
    device = "cpu", dtype = "float32"),
  pattern = "8k"
)

# WAV writer round trip (header + size)
wav_path <- tempfile(fileext = ".wav")
write_wav(res$audio, wav_path, sample_rate = 48000L)
expect_true(file.exists(wav_path))
expect_equal(readBin(wav_path, "raw", 4), charToRaw("RIFF"))
expect_equal(file.size(wav_path), 44 + 2 * 2 * ncol(res$audio))
unlink(wav_path)

# MP4 mux when av is available
if (requireNamespace("av", quietly = TRUE)) {
  mp4 <- tempfile(fileext = ".mp4")
  save_video_ltx23(res$video, mp4, fps = 24, audio = res$audio,
    sample_rate = 48000L, verbose = FALSE)
  expect_true(file.exists(mp4) && file.size(mp4) > 0)
  unlink(mp4)
}

# --- Prefix conditioning smoke tests -------------------------------------------------

# i2v: start image conditions frame 0; pipeline runs end to end
start_img <- array(runif(64 * 64 * 3), dim = c(64L, 64L, 3L))
res_i2v <- txt2vid_ltx2(
  prompt = "tiny i2v",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 9L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  image = start_img,
  decode_audio = FALSE,
  verbose = FALSE
)
expect_equal(dim(res_i2v$video), c(9L, 64L, 64L, 3L))
expect_true(all(is.finite(res_i2v$video)))

# Same seed without conditioning gives a different video (mask engaged)
res_t2v <- txt2vid_ltx2(
  prompt = "tiny i2v",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 9L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  decode_audio = FALSE,
  verbose = FALSE
)
expect_true(max(abs(res_i2v$video - res_t2v$video)) > 1e-4)

# Continuation: 9-frame tail array as the frozen prefix
tail_arr <- array(runif(9 * 64 * 64 * 3), dim = c(9L, 64L, 64L, 3L))
res_cont <- txt2vid_ltx2(
  prompt = "tiny continuation",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 17L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  condition_video = tail_arr, conditioning_frames = 9L,
  decode_audio = FALSE,
  verbose = FALSE
)
expect_equal(dim(res_cont$video), c(17L, 64L, 64L, 3L))
expect_true(all(is.finite(res_cont$video)))

# Guardrails
expect_error(
  txt2vid_ltx2("x", pipe, prompt_embeds = stub_embeds,
    image = start_img, condition_video = tail_arr,
    device = "cpu", dtype = "float32"),
  pattern = "only one"
)

# --- Latent-space chaining seams -----------------------------------------------------

# latent_shape rides in the result: 17 frames @ 64x64 -> [3, 2, 2]
expect_equal(res_cont$latent_shape, c(3L, 2L, 2L))

# condition_latents reproduces the condition_video path exactly when
# fed the same encoded tail (the pixel path is just one producer)
enc <- ltx23_encode_video_frames(
  vae, ltx23_preprocess_frames(tail_arr, 64L, 64L))
res_lat <- txt2vid_ltx2(
  prompt = "tiny continuation",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 17L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  condition_latents = enc,
  decode_audio = FALSE,
  verbose = FALSE
)
expect_true(max(abs(res_lat$video - res_cont$video)) < 1e-6)

# Tail slicing: last 2 latent frames, matching a manual unpack + narrow
tail_lat <- ltx23_tail_latents(res_cont, k = 2L)
expect_equal(as.integer(tail_lat$shape), c(1L, 128L, 2L, 2L, 2L))
manual <- diffuseR:::ltx23_unpack_video_latents(res_cont$latents, 3L, 2L, 2L)$
  narrow(3L, 2L, 2L)
expect_equal(as.numeric((tail_lat - manual)$abs()$max()), 0)

# Feeding the sliced tail back runs end to end (chunk N -> chunk N+1)
res_next <- txt2vid_ltx2(
  prompt = "tiny chunk 3",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 17L, frame_rate = 24,
  seed = 8L, device = "cpu", dtype = "float32",
  condition_latents = tail_lat,
  decode_audio = FALSE,
  verbose = FALSE
)
expect_equal(dim(res_next$video), c(17L, 64L, 64L, 3L))
expect_true(all(is.finite(res_next$video)))

# trim_frames delivers head-free pixels; latents keep the full sequence
res_trim <- txt2vid_ltx2(
  prompt = "tiny continuation",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 17L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  condition_video = tail_arr, trim_frames = 9L,
  decode_audio = FALSE,
  verbose = FALSE
)
expect_equal(dim(res_trim$video), c(8L, 64L, 64L, 3L))
expect_equal(max(abs(res_trim$video -
                     res_cont$video[-(1:9), , , , drop = FALSE])), 0)
expect_equal(res_trim$latent_shape, c(3L, 2L, 2L))
expect_equal(as.integer(ltx23_tail_latents(res_trim)$shape[3]), 2L)

# resident= is accepted (inert on CPU, where phase_offload is off)
res_res <- txt2vid_ltx2(
  prompt = "tiny resident",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 9L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  resident = "transformer",
  decode_audio = FALSE,
  verbose = FALSE
)
expect_true(all(is.finite(res_res$video)))

# Guardrails
expect_error(
  txt2vid_ltx2("x", pipe, prompt_embeds = stub_embeds,
    condition_latents = enc, image = start_img,
    device = "cpu", dtype = "float32"),
  pattern = "only one"
)
expect_error(
  txt2vid_ltx2("x", pipe, prompt_embeds = stub_embeds,
    condition_latents = torch::torch_randn(1L, 128L, 2L, 3L, 3L),
    device = "cpu", dtype = "float32"),
  pattern = "condition_latents"
)
expect_error(
  txt2vid_ltx2("x", pipe, prompt_embeds = stub_embeds, num_frames = 9L,
    trim_frames = 9L, device = "cpu", dtype = "float32"),
  pattern = "trim_frames"
)
expect_error(
  txt2vid_ltx2("x", pipe, prompt_embeds = stub_embeds,
    resident = "flux_capacitor", device = "cpu", dtype = "float32"),
  pattern = "resident"
)
expect_error(ltx23_tail_latents(res_cont$latents), pattern = "latent_shape")
expect_error(ltx23_tail_latents(res_cont, k = 5L), pattern = "k must be")

# --- connector_embeds bypass ---------------------------------------------------------

# Precomputed connector outputs reproduce the internal connectors path
# exactly, and satisfy the text requirement without prompt_embeds
conn_out <- torch::with_no_grad(connectors(
  stub_embeds$prompt_embeds, stub_embeds$prompt_attention_mask))
cemb <- list(video_text_embedding = conn_out$video_text_embedding,
             audio_text_embedding = conn_out$audio_text_embedding,
             attention_mask = conn_out$attention_mask)
res_ce <- txt2vid_ltx2(
  prompt = "tiny bypass",
  pipeline = pipe,
  connector_embeds = cemb,
  width = 64L, height = 64L, num_frames = 9L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  decode_audio = FALSE,
  verbose = FALSE
)
res_full <- txt2vid_ltx2(
  prompt = "tiny bypass",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 9L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  decode_audio = FALSE,
  verbose = FALSE
)
expect_equal(dim(res_ce$video), c(9L, 64L, 64L, 3L))
expect_true(max(abs(res_ce$video - res_full$video)) < 1e-6)


# --- Audio-conditioned generation (lip-sync plumbing) --------------------------------

wav_in <- matrix(runif(2L * 6000L, -0.5, 0.5), nrow = 2L)
res_audio <- txt2vid_ltx2(
  prompt = "tiny audio-driven",
  pipeline = pipe,
  prompt_embeds = stub_embeds,
  width = 64L, height = 64L, num_frames = 9L, frame_rate = 24,
  seed = 7L, device = "cpu", dtype = "float32",
  image = start_img,
  audio = wav_in,
  verbose = FALSE
)
expect_equal(dim(res_audio$video), c(9L, 64L, 64L, 3L))
expect_true(all(is.finite(res_audio$video)))
# The original audio rides through untouched
expect_equal(res_audio$sample_rate, 16000L)
expect_equal(res_audio$audio, wav_in)

#' FLUX.2 Position Ids and Empirical Shift
#'
#' Fresh R port of the FLUX.2 position-id builders and the empirical
#' timestep-shift formula from the diffusers reference (Apache-2.0,
#' src/diffusers/pipelines/flux2/pipeline_flux2_klein.py). FLUX.2 uses
#' 4-axis rotary position ids (T, H, W, L): text tokens carry only the
#' L axis (sequence position), image latents carry H and W, and the T
#' axis distinguishes reference images (unused for txt2img). Frequencies
#' come from \code{\link{flux_pos_embed}} with
#' \code{axes_dim = c(32, 32, 32, 32)} and \code{theta = 2000}.
#'
#' @name rope_flux2
NULL

#' Build FLUX.2 text position ids
#'
#' Columns (T, H, W, L) with only L varying: 0..len-1. Reference:
#' Flux2KleinPipeline._prepare_text_ids.
#'
#' @param len Integer. Text sequence length.
#' @param device Device for the resulting tensor.
#'
#' @return Float tensor [len, 4].
#'
#' @export
flux2_prepare_text_ids <- function(len, device = "cpu") {
    f32 <- torch::torch_float32()
    ids <- torch::torch_zeros(len, 4L, dtype = f32, device = device)
    ids[, 4] <- torch::torch_arange(start = 0, end = len - 1, dtype = f32,
                                    device = device)
    ids
}

#' Build FLUX.2 latent position ids
#'
#' Columns (T, H, W, L) with H and W carrying the packed-grid position
#' (row-major: H varies slowest), T = L = 0. Reference:
#' Flux2KleinPipeline._prepare_latent_ids.
#'
#' @param height Integer. Packed grid height (pixel height / 16).
#' @param width Integer. Packed grid width (pixel width / 16).
#' @param device Device for the resulting tensor.
#'
#' @return Float tensor [height * width, 4].
#'
#' @export
flux2_prepare_latent_ids <- function(height, width, device = "cpu") {
    f32 <- torch::torch_float32()
    ids <- torch::torch_zeros(height, width, 4L, dtype = f32, device = device)
    rows <- torch::torch_arange(start = 0, end = height - 1, dtype = f32,
                                device = device)
    cols <- torch::torch_arange(start = 0, end = width - 1, dtype = f32,
                                device = device)
    ids[,, 2] <- ids[,, 2] + rows$unsqueeze(2L)
    ids[,, 3] <- ids[,, 3] + cols$unsqueeze(1L)
    ids$reshape(c(height * width, 4L))
}

#' Empirical timestep shift for FLUX.2
#'
#' BFL's piecewise-linear fit of the dynamic-shifting mu as a function of
#' image sequence length and step count; replaces FLUX.1's
#' calculate_shift. Reference: compute_empirical_mu (adapted from BFL
#' sampling.py).
#'
#' @param image_seq_len Integer. Packed image token count.
#' @param num_steps Integer. Inference steps.
#'
#' @return Numeric mu for \code{\link{flowmatch_set_timesteps}}.
#'
#' @export
flux2_empirical_mu <- function(image_seq_len, num_steps) {
    a1 <- 8.73809524e-05
    b1 <- 1.89833333
    a2 <- 0.00016927
    b2 <- 0.45666666

    if (image_seq_len > 4300) {
        return(a2 * image_seq_len + b2)
    }
    m_200 <- a2 * image_seq_len + b2
    m_10 <- a1 * image_seq_len + b1
    a <- (m_200 - m_10) / 190.0
    b <- m_200 - 200.0 * a
    a * num_steps + b
}

#' FLUX.2 Latent Layout and VAE Helpers
#'
#' Fresh R port of the FLUX.2 latent packing chain from the diffusers
#' reference (Apache-2.0, src/diffusers/pipelines/flux2/
#' pipeline_flux2_klein.py). The 32-channel VAE latent is patchified
#' 2x2 into 128 channels, normalized with the VAE's BatchNorm running
#' statistics (there is no scalar scaling/shift factor in FLUX.2), and
#' flattened to channels-last tokens for the transformer.
#'
#' @name vae_flux2
NULL

#' Patchify FLUX.2 latents (2x2 -> channels)
#'
#' [B, C, H, W] -> [B, 4C, H/2, W/2], channel order (C, ph, pw).
#' Reference: Flux2KleinPipeline._patchify_latents.
#'
#' @param latents Tensor [B, C, H, W]; H and W must be even.
#'
#' @return Tensor [B, C * 4, H / 2, W / 2].
#'
#' @export
flux2_patchify_latents <- function(latents) {
    shape <- latents$shape
    b <- shape[1]
    ch <- shape[2]
    h <- shape[3]
    w <- shape[4]
    latents <- latents$view(c(b, ch, h %/% 2L, 2L, w %/% 2L, 2L))
    # Python permute (0, 1, 3, 5, 2, 4), 1-indexed here
    latents <- latents$permute(c(1L, 2L, 4L, 6L, 3L, 5L))
    latents$reshape(c(b, ch * 4L, h %/% 2L, w %/% 2L))
}

#' Unpatchify FLUX.2 latents (channels -> 2x2)
#'
#' Inverse of \code{\link{flux2_patchify_latents}}. Reference:
#' Flux2KleinPipeline._unpatchify_latents.
#'
#' @param latents Tensor [B, 4C, H, W].
#'
#' @return Tensor [B, C, H * 2, W * 2].
#'
#' @export
flux2_unpatchify_latents <- function(latents) {
    shape <- latents$shape
    b <- shape[1]
    ch <- shape[2]
    h <- shape[3]
    w <- shape[4]
    latents <- latents$reshape(c(b, ch %/% 4L, 2L, 2L, h, w))
    # Python permute (0, 1, 4, 2, 5, 3), 1-indexed here
    latents <- latents$permute(c(1L, 2L, 5L, 3L, 6L, 4L))
    latents$reshape(c(b, ch %/% 4L, h * 2L, w * 2L))
}

#' Pack patchified FLUX.2 latents into tokens
#'
#' [B, C, H, W] -> [B, H * W, C] (row-major spatial flatten,
#' channels-last). Reference: Flux2KleinPipeline._pack_latents.
#'
#' @param latents Tensor [B, C, H, W].
#'
#' @return Tensor [B, H * W, C].
#'
#' @export
flux2_pack_latents <- function(latents) {
    shape <- latents$shape
    b <- shape[1]
    ch <- shape[2]
    latents <- latents$reshape(c(b, ch, shape[3] * shape[4]))
    latents$permute(c(1L, 3L, 2L))
}

#' Unpack FLUX.2 tokens back to a latent grid via position ids
#'
#' Scatters tokens to (H, W) positions taken from the id columns
#' (H = column 2, W = column 3, 0-based values). Reference:
#' Flux2KleinPipeline._unpack_latents_with_ids.
#'
#' @param x Tensor [B, S, C] of tokens.
#' @param ids Tensor [S, 4] (or [B, S, 4]) of position ids.
#' @param height,width Integers. Packed grid dimensions.
#'
#' @return Tensor [B, C, height, width].
#'
#' @export
flux2_unpack_latents_with_ids <- function(x, ids, height, width) {
    if (ids$ndim == 3L) {
        ids <- ids[1,,]
    }
    ids <- ids$to(device = x$device)
    long <- torch::torch_long()
    h_ids <- ids[, 2]$to(dtype = long)
    w_ids <- ids[, 3]$to(dtype = long)
    flat <- h_ids$mul(width)$add(w_ids)$add(1L) # 1-based scatter index

    shape <- x$shape
    b <- shape[1]
    ch <- shape[3]
    out <- torch::torch_zeros(b, height * width, ch, dtype = x$dtype,
                              device = x$device)
    index <- flat$unsqueeze(1L)$unsqueeze(3L)$expand(c(b, -1L, ch))
    out$scatter_(2L, index, x)
    out$view(c(b, height, width, ch))$permute(c(1L, 4L, 2L, 3L))
}

#' Normalize patchified latents with the VAE BatchNorm statistics
#'
#' FLUX.2 has no scalar scaling/shift factor; latents are standardized
#' per packed channel with the VAE's \code{bn.running_mean} /
#' \code{bn.running_var} (eps 1e-4). Reference: encode/decode paths of
#' Flux2KleinPipeline.
#'
#' @param latents Tensor [B, 128, H, W] (patchified).
#' @param bn_mean,bn_var Float tensors [128].
#' @param eps Numeric. BatchNorm epsilon.
#' @param inverse Logical. De-normalize (decode path) instead.
#'
#' @return Tensor like \code{latents}.
#'
#' @export
flux2_bn_normalize <- function(latents, bn_mean, bn_var, eps = 1e-4,
                               inverse = FALSE) {
    mean <- bn_mean$view(c(1L, -1L, 1L, 1L))$to(device = latents$device,
        dtype = latents$dtype)
    std <- bn_var$add(eps)$sqrt()$view(c(1L, -1L, 1L, 1L))$
    to(device = latents$device, dtype = latents$dtype)
    if (inverse) {
        latents$mul(std)$add(mean)
    } else {
        latents$sub(mean)$div(std)
    }
}

#' FLUX.2 VAE decoder
#'
#' The AutoencoderKLFlux2 decode path: post_quant_conv (1x1, 32
#' channels) followed by the standard AutoencoderKL decoder body
#' (reused from \code{\link{vae_decoder_native}}), plus the BatchNorm
#' running statistics used for latent (de)normalization. Reference:
#' src/diffusers/models/autoencoders/autoencoder_kl_flux2.py.
#'
#' @param latent_channels Integer (32 for FLUX.2).
#' @param block_channels Decoder block channels (reversed encoder
#'   block_out_channels).
#' @param norm_groups Integer. Group norm groups.
#'
#' @return Module whose forward(z) decodes [B, 32, H, W] latents to
#'   [B, 3, 8H, 8W] images; \code{$bn$running_mean} /
#'   \code{$bn$running_var} carry the normalization statistics.
#'
#' @export
flux2_vae_decoder <- torch::nn_module(
                                      "flux2_vae_decoder",
                                      initialize = function(latent_channels = 32L,
        block_channels = c(512L, 512L, 256L, 128L),
        norm_groups = 32L) {
    self$post_quant_conv <- torch::nn_conv2d(latent_channels, latent_channels,
        kernel_size = 1)
    self$decoder <- vae_decoder_native(
                                       latent_channels = latent_channels,
                                       block_channels = block_channels,
                                       norm_groups = norm_groups
    )
    bn_stats <- torch::nn_module(
                                 "flux2_bn_stats",
                                 initialize = function(n) {
        self$running_mean <- torch::nn_buffer(torch::torch_zeros(n))
        self$running_var <- torch::nn_buffer(torch::torch_ones(n))
    },
                                 forward = function(x) x
    )
    self$bn <- bn_stats(latent_channels * 4L)
},
                                      forward = function(z) {
    self$decoder(self$post_quant_conv(z))
}
)

#' Load the FLUX.2 VAE decoder from safetensors
#'
#' Loads the decoder half plus post_quant_conv and the BatchNorm running
#' statistics; encoder and quant_conv keys are skipped (txt2img needs no
#' encoder).
#'
#' @param path Path to the VAE .safetensors file (or a directory
#'   containing diffusion_pytorch_model.safetensors).
#' @param latent_channels,block_channels,norm_groups Constructor
#'   arguments for \code{\link{flux2_vae_decoder}}.
#' @param verbose Logical.
#'
#' @return The loaded \code{flux2_vae_decoder} in eval mode.
#'
#' @export
load_flux2_vae_decoder <- function(path, latent_channels = 32L,
                                   block_channels = c(512L, 512L, 256L, 128L),
                                   norm_groups = 32L, verbose = TRUE) {
    path <- path.expand(path)
    if (dir.exists(path)) {
        path <- file.path(path, "diffusion_pytorch_model.safetensors")
    }
    dec <- flux2_vae_decoder(latent_channels = latent_channels,
                             block_channels = block_channels,
                             norm_groups = norm_groups)

    handle <- safetensors::safetensors$new(path, framework = "torch")
    keys <- setdiff(handle$keys(), "__metadata__")
    keep <- keys[!startsWith(keys, "encoder.") &
        !startsWith(keys, "quant_conv.") &
        keys != "bn.num_batches_tracked"]

    dests <- c(dec$named_parameters(), dec$named_buffers())
    filled <- character(0)
    unmapped <- character(0)
    torch::with_no_grad({
        for (key in keep) {
            dest <- dests[[key]]
            if (is.null(dest)) {
                unmapped <- c(unmapped, key)
                next
            }
            dest$copy_(handle$get_tensor(key))
            filled <- c(filled, key)
        }
    })
    unfilled <- setdiff(names(dests), filled)
    if (length(unmapped)) {
        stop("FLUX.2 VAE load: ", length(unmapped), " unmapped keys, e.g. ",
             paste(utils::head(unmapped, 3), collapse = ", "))
    }
    if (length(unfilled)) {
        stop("FLUX.2 VAE load: ", length(unfilled),
             " unfilled params, e.g. ",
             paste(utils::head(unfilled, 3), collapse = ", "))
    }
    if (verbose) {
        message("Loaded ", length(filled), " FLUX.2 VAE tensors from ", path)
    }
    dec$eval()
    dec
}

# Parity tests for the FLUX.2 position ids, 4-axis RoPE, latent
# patchify/pack/unpack chain, and empirical mu against diffusers
# reference fixtures (generated by tools/gen_fixtures_flux2.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "rope_flux2.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/rope_flux2.safetensors"
if (!file.exists(fixture_path)) exit_file("flux2 rope fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# --- position ids ---------------------------------------------------------------

txt_ids <- flux2_prepare_text_ids(7L)
expect_equal(as.integer(txt_ids$shape), as.integer(fx$txt_ids$shape))
expect_true(max_abs_diff(txt_ids, fx$txt_ids) == 0)

lat_ids <- flux2_prepare_latent_ids(6L, 10L)
expect_equal(as.integer(lat_ids$shape), as.integer(fx$latent_ids$shape))
expect_true(max_abs_diff(lat_ids, fx$latent_ids) == 0)

# --- 4-axis RoPE through flux_pos_embed (theta 2000) ---------------------------

pf <- flux_pos_embed(fx$ids, axes_dim = c(32L, 32L, 32L, 32L), theta = 2000)
expect_equal(as.integer(pf[[1]]$shape), as.integer(fx$pos_full_cos$shape))
expect_true(max_abs_diff(pf[[1]], fx$pos_full_cos) < 1e-6)
expect_true(max_abs_diff(pf[[2]], fx$pos_full_sin) < 1e-6)

pt <- flux_pos_embed(fx$ids, axes_dim = c(2L, 2L, 2L, 2L), theta = 2000)
expect_true(max_abs_diff(pt[[1]], fx$pos_tiny_cos) < 1e-6)
expect_true(max_abs_diff(pt[[2]], fx$pos_tiny_sin) < 1e-6)

# --- patchify / pack / unpack chain ----------------------------------------------

patched <- flux2_patchify_latents(fx$z32)
expect_equal(as.integer(patched$shape), as.integer(fx$patched$shape))
expect_true(max_abs_diff(patched, fx$patched) == 0)

expect_true(max_abs_diff(flux2_unpatchify_latents(fx$patched), fx$unpatched) == 0)
expect_true(max_abs_diff(flux2_unpatchify_latents(fx$patched), fx$z32) == 0)

packed <- flux2_pack_latents(fx$patched)
expect_equal(as.integer(packed$shape), as.integer(fx$packed$shape))
expect_true(max_abs_diff(packed, fx$packed) == 0)

unpacked <- flux2_unpack_latents_with_ids(fx$packed,
  flux2_prepare_latent_ids(6L, 10L), height = 6L, width = 10L)
expect_equal(as.integer(unpacked$shape), as.integer(fx$unpacked$shape))
expect_true(max_abs_diff(unpacked, fx$unpacked) == 0)
expect_true(max_abs_diff(unpacked, fx$patched) == 0)

# --- empirical mu ------------------------------------------------------------------

seqs <- as.numeric(fx$mu_seq)
steps <- as.numeric(fx$mu_steps)
vals <- as.numeric(fx$mu_vals)
for (i in seq_along(seqs)) {
  expect_true(abs(flux2_empirical_mu(seqs[i], steps[i]) - vals[i]) < 1e-12,
    info = sprintf("seq %d steps %d", seqs[i], steps[i]))
}

# Parity tests for the FLUX.2 VAE decoder (post_quant_conv + decoder
# body + BatchNorm latent statistics) against diffusers reference
# fixtures (generated by tools/gen_fixtures_flux2_vae.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture <- function(name) {
  p <- system.file("tinytest", "fixtures", name, package = "diffuseR")
  if (p == "") p <- file.path("fixtures", name)
  p
}
vae_path <- fixture("vae_flux2_tiny.safetensors")
io_path <- fixture("vae_flux2_io.safetensors")
if (!file.exists(io_path)) exit_file("flux2 vae fixtures missing")

fx <- safetensors::safe_load_file(io_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

dec <- load_flux2_vae_decoder(vae_path, latent_channels = 32L,
  block_channels = c(32L, 32L, 16L, 8L), norm_groups = 8L,
  verbose = FALSE)

# BN statistics loaded from the checkpoint
expect_true(max_abs_diff(dec$bn$running_mean, fx$bn_mean) == 0)
expect_true(max_abs_diff(dec$bn$running_var, fx$bn_var) == 0)

# Decode parity (post_quant_conv + decoder body)
img <- torch::with_no_grad(dec(fx$latent))
expect_equal(as.integer(img$shape), as.integer(fx$image$shape))
expect_true(max_abs_diff(img, fx$image) < 1e-5)

# BN normalize / denormalize parity + roundtrip
norm <- flux2_bn_normalize(fx$patched, dec$bn$running_mean,
  dec$bn$running_var)
expect_true(max_abs_diff(norm, fx$normalized) < 1e-5)
denorm <- flux2_bn_normalize(fx$normalized, dec$bn$running_mean,
  dec$bn$running_var, inverse = TRUE)
expect_true(max_abs_diff(denorm, fx$denormalized) < 1e-5)
expect_true(max_abs_diff(denorm, fx$patched) < 1e-5)

# Gemma3 Text Encoder for LTX-2
#
# Native R torch implementation of Gemma3 for text encoding.
# Used by LTX-2 video generation pipeline.
#
# Architecture based on HuggingFace transformers Gemma3TextModel:
# - 48 hidden layers
# - hidden_size: 3840
# - 16 attention heads, 8 KV heads (GQA)
# - Sliding window attention (1024 tokens) on most layers
# - RoPE positional embeddings

# -----------------------------------------------------------------------------
# Gemma3 RMS Normalization
# -----------------------------------------------------------------------------

#' Gemma3 RMS Normalization
#'
#' RMSNorm with optional addition of 1 to weights (Gemma-style).
#'
#' @param dim Integer. Hidden dimension.
#' @param eps Numeric. Epsilon for numerical stability.
#' @keywords internal
gemma3_rms_norm <- torch::nn_module(
                                    "Gemma3RMSNorm",
                                    initialize = function(
        dim,
        eps = 1e-6
    ) {
    self$eps <- eps
    self$weight <- torch::nn_parameter(torch::torch_zeros(dim))
},
                                    forward = function(x) {
    input_dtype <- x$dtype
    x <- x$to(dtype = torch::torch_float32())
    variance <- x$pow(2)$mean(dim = -1L, keepdim = TRUE)
    x <- x * torch::torch_rsqrt(variance + self$eps)
    # Gemma adds 1 to the weight: (1 + weight) * x
    x$to(dtype = input_dtype) * (self$weight$add(1))
}
)

# -----------------------------------------------------------------------------
# Gemma3 Rotary Position Embeddings
# -----------------------------------------------------------------------------

#' Gemma3 Rotary Position Embeddings
#'
#' Standard RoPE with optional scaling factor for extended context.
#'
#' @param dim Integer. Head dimension.
#' @param max_position_embeddings Integer. Maximum sequence length.
#' @param base Numeric. RoPE base frequency.
#' @param scaling_factor Numeric. Optional scaling factor for extended context.
#' @keywords internal
gemma3_rotary_embedding <- torch::nn_module(
    "Gemma3RotaryEmbedding",
    initialize = function(
                          dim,
                          max_position_embeddings = 8192L,
                          base = 10000.0,
                          scaling_factor = 1.0
    ) {
    self$dim <- dim
    self$max_position_embeddings <- max_position_embeddings
    self$base <- base
    self$scaling_factor <- scaling_factor

    # Precompute inverse frequencies
    # For linear scaling, divide inv_freq by the scaling factor (HuggingFace convention)
    inv_freq <- 1.0 / (base ^ (torch::torch_arange(0, dim - 1L, 2L)$to(dtype = torch::torch_float32()) / dim))
    inv_freq <- inv_freq / scaling_factor
    self$inv_freq <- torch::nn_buffer(inv_freq, persistent = FALSE)
},

    forward = function(
                       x,
                       position_ids
    ) {
    # x: [batch, seq_len, ...]
    # position_ids: [batch, seq_len] or [1, seq_len]

    # Compute frequencies: [batch, seq_len, dim/2]
    position_ids_float <- position_ids$to(dtype = torch::torch_float32())

    # HuggingFace uses matmul: inv_freq[None,:,None] @ position_ids[:,None,:]
    # We use einsum which is equivalent
    freqs <- torch::torch_einsum("bs,d->bsd",
                                 list(position_ids_float, self$inv_freq$to(device = x$device)))

    # Concatenate for full dimension: [batch, seq_len, dim]
    # This duplicates freqs so that first half and second half are identical
    emb <- torch::torch_cat(list(freqs, freqs), dim = -1L)

    cos_emb <- torch::torch_cos(emb)$to(dtype = x$dtype)
    sin_emb <- torch::torch_sin(emb)$to(dtype = x$dtype)

    list(cos_emb, sin_emb)
}
)

#' Apply rotary position embeddings
#'
#' @param q Query tensor [batch, heads, seq, head_dim]
#' @param k Key tensor [batch, heads, seq, head_dim]
#' @param cos Cosine embeddings [batch, seq, head_dim]
#' @param sin Sine embeddings [batch, seq, head_dim]
#' @keywords internal
apply_rotary_pos_emb <- function(q, k, cos, sin) {
    # Reshape cos/sin for broadcasting: [batch, 1, seq, head_dim]
    cos <- cos$unsqueeze(2L)
    sin <- sin$unsqueeze(2L)

    # Apply rotation
    q_embed <- (q * cos) + (rotate_half(q) * sin)
    k_embed <- (k * cos) + (rotate_half(k) * sin)

    list(q_embed, k_embed)
}

#' Rotate half of the hidden dims
#' @keywords internal
rotate_half <- function(x) {
    # x: [..., dim]
    # Split into two halves and rotate
    dim <- x$shape[length(x$shape)]
    half_dim <- dim %/% 2L
    x1 <- x[.., 1:half_dim]
    x2 <- x[.., (half_dim + 1L):dim]
    torch::torch_cat(list(-x2, x1), dim = -1L)
}

#' Repeat KV heads for GQA (Grouped Query Attention)
#'
#' Interleaves KV heads to match query heads: [k0,k0,k1,k1,...] not [k0,k1,...,k0,k1,...]
#' @param hidden_states Tensor of shape [batch, num_kv_heads, seq_len, head_dim]
#' @param n_rep Number of repetitions per KV head
#' @return Tensor of shape [batch, num_kv_heads * n_rep, seq_len, head_dim]
#' @keywords internal
repeat_kv <- function(hidden_states, n_rep) {
    if (n_rep == 1L) {
        return(hidden_states)
    }
    batch <- hidden_states$shape[1]
    num_kv_heads <- hidden_states$shape[2]
    seq_len <- hidden_states$shape[3]
    head_dim <- hidden_states$shape[4]

    # Add dimension and expand: [batch, kv_heads, 1, seq, dim] -> [batch, kv_heads, n_rep, seq, dim]
    hidden_states <- hidden_states$unsqueeze(3L)$expand(c(batch, num_kv_heads,
            n_rep, seq_len, head_dim))
    # Reshape to interleave: [batch, kv_heads * n_rep, seq, dim]
    hidden_states$reshape(c(batch, num_kv_heads * n_rep, seq_len, head_dim))
}

# -----------------------------------------------------------------------------
# Gemma3 MLP (Gated Linear Unit)
# -----------------------------------------------------------------------------

#' Gemma3 MLP
#'
#' Feed-forward network with gated linear units and GELU activation.
#'
#' @param config List with hidden_size and intermediate_size.
#' @keywords internal
gemma3_mlp <- torch::nn_module(
                               "Gemma3MLP",
                               initialize = function(config) {
    self$hidden_size <- config$hidden_size
    self$intermediate_size <- config$intermediate_size

    self$gate_proj <- linear_noinit(self$hidden_size, self$intermediate_size,
                                    bias = FALSE)
    self$up_proj <- linear_noinit(self$hidden_size, self$intermediate_size, bias = FALSE)
    self$down_proj <- linear_noinit(self$intermediate_size, self$hidden_size, bias = FALSE)

    # Gemma uses approximate GELU
    self$act_fn <- function(x) torch::nnf_gelu(x, approximate = "tanh")
},
                               forward = function(x) {
    # Gated Linear Unit: down(act(gate(x)) * up(x))
    self$down_proj(self$act_fn(self$gate_proj(x)) * self$up_proj(x))
}
)

# -----------------------------------------------------------------------------
# Gemma3 Attention
# -----------------------------------------------------------------------------

#' Gemma3 Attention
#'
#' Multi-head attention with Grouped Query Attention (GQA) and optional
#' sliding window attention.
#'
#' @param config Model configuration.
#' @param layer_idx Integer. Layer index for layer-specific settings.
#' @keywords internal
gemma3_attention <- torch::nn_module(
                                     "Gemma3Attention",
                                     initialize = function(
        config,
        layer_idx = 0L
    ) {
    self$config <- config
    self$layer_idx <- layer_idx

    self$hidden_size <- config$hidden_size
    self$num_heads <- config$num_attention_heads
    self$head_dim <- config$head_dim %||% (config$hidden_size %/% config$num_attention_heads)
    self$num_key_value_heads <- config$num_key_value_heads %||% config$num_attention_heads
    self$num_key_value_groups <- self$num_heads %/% self$num_key_value_heads

    # Sliding window (layer-specific)
    self$sliding_window <- config$sliding_window
    self$is_sliding <- self$get_is_sliding(layer_idx)

    # Attention softcapping (Gemma3 feature)
    self$attn_logit_softcapping <- config$attn_logit_softcapping %||% NULL

    # Projections
    self$q_proj <- linear_noinit(self$hidden_size,
                                 self$num_heads * self$head_dim, bias = FALSE)
    self$k_proj <- linear_noinit(self$hidden_size, self$num_key_value_heads * self$head_dim, bias = FALSE)
    self$v_proj <- linear_noinit(self$hidden_size, self$num_key_value_heads * self$head_dim, bias = FALSE)
    self$o_proj <- linear_noinit(self$num_heads * self$head_dim, self$hidden_size, bias = FALSE)

    # Q/K normalization (Gemma3 feature)
    self$q_norm <- gemma3_rms_norm(self$head_dim, eps = config$rms_norm_eps %||% 1e-6)
    self$k_norm <- gemma3_rms_norm(self$head_dim, eps = config$rms_norm_eps %||% 1e-6)

    # Scaling factor
    self$scaling <- config$query_pre_attn_scalar %||% self$head_dim
    self$scaling <- 1.0 / sqrt(self$scaling)
},

                                     get_is_sliding = function(layer_idx) {
    # Gemma3 uses alternating sliding window pattern
    # Default: every 6th layer is global attention
    sliding_window_pattern <- self$config$sliding_window_pattern %||% 6L
    return((layer_idx + 1L) %% sliding_window_pattern != 0L)
},

                                     forward = function(
        hidden_states,
        attention_mask = NULL,
        position_embeddings_global = NULL,
        position_embeddings_local = NULL
    ) {
    batch_size <- hidden_states$shape[1]
    seq_len <- hidden_states$shape[2]

    # Project Q, K, V
    query_states <- self$q_proj(hidden_states)
    key_states <- self$k_proj(hidden_states)
    value_states <- self$v_proj(hidden_states)

    # Reshape to [batch, heads, seq, head_dim]
    query_states <- query_states$view(c(batch_size, seq_len, self$num_heads, self$head_dim))$transpose(2L, 3L)
    key_states <- key_states$view(c(batch_size, seq_len, self$num_key_value_heads, self$head_dim))$transpose(2L, 3L)
    value_states <- value_states$view(c(batch_size, seq_len, self$num_key_value_heads, self$head_dim))$transpose(2L, 3L)

    # Apply Q/K normalization
    query_states <- self$q_norm(query_states)
    key_states <- self$k_norm(key_states)

    # Apply rotary embeddings - use local for sliding window layers, global otherwise
    if (self$is_sliding) {
        position_embeddings <- position_embeddings_local
    } else {
        position_embeddings <- position_embeddings_global
    }
    if (!is.null(position_embeddings)) {
        cos <- position_embeddings[[1]]
        sin <- position_embeddings[[2]]
        rope_result <- apply_rotary_pos_emb(query_states, key_states, cos, sin)
        query_states <- rope_result[[1]]
        key_states <- rope_result[[2]]
    }

    # Repeat K/V for GQA (interleaved: [k0,k0,k1,k1,...] not [k0,k1,...,k0,k1,...])
    if (self$num_key_value_groups > 1L) {
        key_states <- repeat_kv(key_states, self$num_key_value_groups)
        value_states <- repeat_kv(value_states, self$num_key_value_groups)
    }

    # Compute attention scores
    attn_weights <- torch::torch_matmul(query_states, key_states$transpose(-2L, -1L)) * self$scaling

    # Apply softcapping if configured
    if (!is.null(self$attn_logit_softcapping)) {
        attn_weights <- attn_weights / self$attn_logit_softcapping
        attn_weights <- torch::torch_tanh(attn_weights)
        attn_weights <- attn_weights * self$attn_logit_softcapping
    }

    # Apply attention mask (includes causal mask from layer)
    # Note: sliding window constraint is handled at the model level, not here
    if (!is.null(attention_mask)) {
        attn_weights <- attn_weights + attention_mask
    }

    # Softmax
    attn_weights <- torch::nnf_softmax(attn_weights, dim = -1L)
    attn_weights <- attn_weights$to(dtype = value_states$dtype)

    # Apply attention to values
    attn_output <- torch::torch_matmul(attn_weights, value_states)

    # Reshape back: [batch, heads, seq, head_dim] -> [batch, seq, hidden]
    attn_output <- attn_output$transpose(2L, 3L)$contiguous()
    attn_output <- attn_output$reshape(c(batch_size, seq_len, self$num_heads * self$head_dim))

    # Output projection
    attn_output <- self$o_proj(attn_output)

    attn_output
}
)

#' Create sliding window causal attention mask
#' @keywords internal
create_sliding_window_mask <- function(seq_len, window_size, device = "cpu") {
    # Create causal mask
    mask <- torch::torch_ones(c(seq_len, seq_len), device = device)
    mask <- torch::torch_triu(mask, diagonal = 1L)

    # Create sliding window mask (tokens beyond window are masked)
    row_idx <- torch::torch_arange(0, seq_len - 1L, device = device)$unsqueeze(2L)
    col_idx <- torch::torch_arange(0, seq_len - 1L, device = device)$unsqueeze(1L)
    window_mask <- (row_idx - col_idx) > window_size

    # Combine: mask future tokens and tokens outside window
    mask <- mask$logical_or(window_mask)

    # Convert to additive mask
    mask <- mask$to(dtype = torch::torch_float32()) * (-1e9)
    mask
}

# -----------------------------------------------------------------------------
# Gemma3 Decoder Layer
# -----------------------------------------------------------------------------

#' Gemma3 Decoder Layer
#'
#' Single transformer block with pre-norm attention and MLP.
#'
#' @param config Model configuration.
#' @param layer_idx Integer. Layer index.
#' @keywords internal
gemma3_decoder_layer <- torch::nn_module(
    "Gemma3DecoderLayer",
    initialize = function(
                          config,
                          layer_idx = 0L
    ) {
    self$hidden_size <- config$hidden_size

    self$self_attn <- gemma3_attention(config, layer_idx = layer_idx)
    self$mlp <- gemma3_mlp(config)

    self$input_layernorm <- gemma3_rms_norm(config$hidden_size,
        eps = config$rms_norm_eps %||% 1e-6)
    self$post_attention_layernorm <- gemma3_rms_norm(config$hidden_size, eps = config$rms_norm_eps %||% 1e-6)

    # Gemma3 has additional pre-feedforward norm
    self$pre_feedforward_layernorm <- gemma3_rms_norm(config$hidden_size, eps = config$rms_norm_eps %||% 1e-6)
    self$post_feedforward_layernorm <- gemma3_rms_norm(config$hidden_size, eps = config$rms_norm_eps %||% 1e-6)
},

    forward = function(
                       hidden_states,
                       attention_mask = NULL,
                       position_embeddings_global = NULL,
                       position_embeddings_local = NULL
    ) {
    residual <- hidden_states

    # Pre-norm self-attention
    hidden_states <- self$input_layernorm(hidden_states)
    hidden_states <- self$self_attn(hidden_states,
                                    attention_mask = attention_mask,
                                    position_embeddings_global = position_embeddings_global,
                                    position_embeddings_local = position_embeddings_local)
    hidden_states <- self$post_attention_layernorm(hidden_states)
    hidden_states <- residual + hidden_states

    # Pre-norm MLP
    residual <- hidden_states
    hidden_states <- self$pre_feedforward_layernorm(hidden_states)
    hidden_states <- self$mlp(hidden_states)
    hidden_states <- self$post_feedforward_layernorm(hidden_states)
    hidden_states <- residual + hidden_states

    hidden_states
}
)

# -----------------------------------------------------------------------------
# Gemma3 Text Model
# -----------------------------------------------------------------------------

#' Gemma3 Text Model
#'
#' Full Gemma3 text encoder model.
#'
#' @param config Model configuration list.
#' @return Module whose forward(input_ids, ...) returns
#'   \code{list(last_hidden_state, hidden_states)}: the final hidden
#'   state [B, S, hidden_size] and the list of per-layer hidden states.
#'
#' @export
gemma3_text_model <- torch::nn_module(
                                      "Gemma3TextModel",
                                      initialize = function(config) {
    self$config <- config
    self$hidden_size <- config$hidden_size
    self$num_layers <- config$num_hidden_layers

    # Token embeddings (scaled by sqrt(hidden_size))
    self$embed_tokens <- embedding_noinit(config$vocab_size, config$hidden_size)
    self$embed_scale <- sqrt(config$hidden_size)

    # Rotary embeddings - Gemma3 uses different frequencies for global vs local layers
    head_dim <- config$head_dim %||% (config$hidden_size %/% config$num_attention_heads)
    # Global rotary embedding (for every 6th layer - layers 5, 11, 17, etc.)
    self$rotary_emb <- gemma3_rotary_embedding(
        dim = head_dim,
        max_position_embeddings = config$max_position_embeddings %||% 8192L,
        base = config$rope_theta %||% 1000000.0, # Global default
        scaling_factor = config$rope_scaling$factor %||% 1.0
    )
    # Local rotary embedding (for sliding window layers)
    self$rotary_emb_local <- gemma3_rotary_embedding(
        dim = head_dim,
        max_position_embeddings = config$max_position_embeddings %||% 8192L,
        base = config$rope_local_base_freq %||% 10000.0, # Local default
        scaling_factor = 1.0 # No scaling for local
    )

    # Decoder layers
    self$layers <- torch::nn_module_list(lapply(seq_len(self$num_layers), function(i) {
        gemma3_decoder_layer(config, layer_idx = i - 1L) # 0-indexed
    }))

    # Final norm
    self$norm <- gemma3_rms_norm(config$hidden_size,
                                 eps = config$rms_norm_eps %||% 1e-6)
},

                                      forward = function(
        input_ids,
        attention_mask = NULL,
        position_ids = NULL,
        output_hidden_states = FALSE
    ) {
    batch_size <- input_ids$shape[1]
    seq_len <- input_ids$shape[2]

    # Token embeddings (scaled by sqrt(hidden_size) - Gemma style)
    # R torch uses 1-based indexing, HF uses 0-based, so add 1
    hidden_states <- self$embed_tokens(input_ids + 1L) * self$embed_scale

    # Position IDs
    if (is.null(position_ids)) {
        position_ids <- torch::torch_arange(0, seq_len - 1L, device = input_ids$device)
        position_ids <- position_ids$unsqueeze(1L)$expand(c(batch_size, -1L))
    }

    # Compute rotary embeddings (both global and local)
    position_embeddings_global <- self$rotary_emb(hidden_states, position_ids)
    position_embeddings_local <- self$rotary_emb_local(hidden_states, position_ids)

    # Prepare causal attention mask
    if (is.null(attention_mask)) {
        # Create causal mask
        causal_mask <- torch::torch_ones(c(seq_len, seq_len), device = hidden_states$device)
        causal_mask <- torch::torch_triu(causal_mask, diagonal = 1L)
        causal_mask <- causal_mask$to(dtype = hidden_states$dtype) * (-1e9)
        causal_mask <- causal_mask$unsqueeze(1L)$unsqueeze(1L)
    } else {
        # Convert attention mask to additive mask
        # attention_mask: [batch, seq_len] with 1 for valid, 0 for padding
        causal_mask <- torch::torch_ones(c(seq_len, seq_len), device = hidden_states$device)
        causal_mask <- torch::torch_triu(causal_mask, diagonal = 1L)

        # Expand padding mask: [batch, 1, 1, seq_len]
        padding_mask <- (1 - attention_mask)$unsqueeze(2L)$unsqueeze(2L)

        # Combine masks
        causal_mask <- (causal_mask$unsqueeze(1L)$unsqueeze(1L) + padding_mask)$to(dtype = hidden_states$dtype) * (-1e9)
    }

    # HF transformers hidden-state semantics: record the state BEFORE
    # each layer, then the post-final-norm output — num_layers + 1
    # entries total. (The un-normed last-layer output never appears;
    # the LTX connectors expect exactly this 49-state stack.)
    all_hidden_states <- NULL

    # Apply decoder layers
    for (i in seq_along(self$layers)) {
        if (output_hidden_states) {
            all_hidden_states <- c(all_hidden_states, list(hidden_states))
        }
        layer <- self$layers[[i]]
        hidden_states <- layer(hidden_states,
                               attention_mask = causal_mask,
                               position_embeddings_global = position_embeddings_global,
                               position_embeddings_local = position_embeddings_local)
    }

    # Final normalization
    hidden_states <- self$norm(hidden_states)

    if (output_hidden_states) {
        all_hidden_states <- c(all_hidden_states, list(hidden_states))
    }

    list(last_hidden_state = hidden_states, hidden_states = all_hidden_states)
}
)

# -----------------------------------------------------------------------------
# Gemma3 Config
# -----------------------------------------------------------------------------

#' Create Gemma3 configuration for LTX-2
#'
#' Returns the default configuration used by LTX-2's text encoder.
#'
#' @return List with model configuration parameters.
#' @export
gemma3_config_ltx2 <- function() {
    list(
         vocab_size = 262208L,
         hidden_size = 3840L,
         intermediate_size = 15360L,
         num_hidden_layers = 48L,
         num_attention_heads = 16L,
         num_key_value_heads = 8L,
         head_dim = 256L,
         max_position_embeddings = 131072L,
         rms_norm_eps = 1e-6,
         rope_theta = 1000000.0, # Gemma3 uses 1M, not 10K
         rope_scaling = list(factor = 8.0, type = "linear"),
         sliding_window = 1024L,
         sliding_window_pattern = 6L,
         attn_logit_softcapping = NULL, # Gemma3 doesn't use softcapping by default
         query_pre_attn_scalar = 256L
    )
}

# -----------------------------------------------------------------------------
# Weight Loading
# -----------------------------------------------------------------------------

# Model config with HF defaults, shared by the checkpoint and NF4
# artifact loaders (the raw text_config travels in the NF4 manifest)
.gemma3_build_config <- function(config_raw) {
    list(
         vocab_size = config_raw$vocab_size %||% 262208L,
         hidden_size = config_raw$hidden_size %||% 3840L,
         intermediate_size = config_raw$intermediate_size %||% 15360L,
         num_hidden_layers = config_raw$num_hidden_layers %||% 48L,
         num_attention_heads = config_raw$num_attention_heads %||% 16L,
         num_key_value_heads = config_raw$num_key_value_heads %||% 8L,
         head_dim = config_raw$head_dim %||% 256L,
         max_position_embeddings = config_raw$max_position_embeddings %||% 131072L,
         rms_norm_eps = config_raw$rms_norm_eps %||% 1e-6,
         rope_theta = config_raw$rope_theta %||% 1000000.0, # Gemma3 uses 1M
         rope_scaling = list(factor = config_raw$rope_scaling$factor %||% 8.0),
         sliding_window = config_raw$sliding_window %||% 1024L,
         sliding_window_pattern = config_raw$sliding_window_pattern %||% 6L,
         attn_logit_softcapping = config_raw$attn_logit_softcapping, # NULL = no softcapping (HF default)
         query_pre_attn_scalar = config_raw$query_pre_attn_scalar %||% 256L
    )
}

#' Load Gemma3 Text Model from safetensors
#'
#' Loads pre-trained Gemma3 weights from HuggingFace safetensors files.
#' An NF4 artifact directory (from \code{\link{gemma3_quantize_nf4}})
#' dispatches to \code{\link{load_gemma3_nf4}}.
#'
#' @param model_path Character. Path to directory containing model files.
#' @param device Character. Device to load model to.
#' @param dtype Character. Data type ("float32", "float16", "bfloat16").
#' @param pin Logical. When loading to the CPU, page-lock the weights
#'   so \code{\link{encode_with_gemma3}} can swap the model to the GPU
#'   at DMA speed per encode (~0.3 s on, free off) instead of holding
#'   VRAM or reloading. Default follows
#'   \code{options(diffuseR.pin_staging)}.
#' @param verbose Logical. Print loading progress.
#' @return Initialized gemma3_text_model with loaded weights.
#' @export
load_gemma3_text_encoder <- function(model_path, device = "cpu",
                                     dtype = "float16",
                                     pin = getOption("diffuseR.pin_staging", TRUE),
                                     verbose = TRUE) {
    # An NF4 artifact directory (gemma3_quantize_nf4) loads through the
    # quantized path; bf16 compute unless the caller says otherwise
    if (file.exists(file.path(model_path, "manifest.json"))) {
        if (identical(dtype, "float16")) {
            dtype <- "bfloat16"
        }
        return(load_gemma3_nf4(model_path, device = device, dtype = dtype,
                               pin = pin, verbose = verbose))
    }
    # Load config
    config_path <- file.path(model_path, "config.json")
    if (!file.exists(config_path)) {
        stop("Config file not found: ", config_path)
    }

    config_raw <- jsonlite::fromJSON(config_path)

    # Extract text config if this is a multimodal model
    if (!is.null(config_raw$text_config)) {
        config_raw <- config_raw$text_config
    }

    config <- .gemma3_build_config(config_raw)

    if (verbose) {
        message(sprintf("Creating Gemma3 model: %d layers, hidden_size=%d",
                        config$num_hidden_layers, config$hidden_size))
    }

    torch_dtype <- switch(dtype,
                          "float32" = torch::torch_float32(),
                          "float16" = torch::torch_float16(),
                          "bfloat16" = torch::torch_bfloat16(),
                          torch::torch_float32()
    )

    # Skeleton at the target dtype: weights fill from the checkpoint,
    # so the 12B fp32 init + cast pass (~72 GB of writes) is skipped
    model <- .construct_skeleton(gemma3_text_model, config, dtype = torch_dtype)

    # Find safetensor files
    safetensor_files <- list.files(model_path, pattern = "\\.safetensors$", full.names = TRUE)
    # Prefer model-* files over diffusion_pytorch_model-*
    model_files <- grep("^model-", basename(safetensor_files), value = TRUE)
    if (length(model_files) == 0) {
        model_files <- grep("diffusion_pytorch_model", safetensor_files, value = TRUE)
    } else {
        model_files <- file.path(model_path, model_files)
        safetensor_files <- model_files
    }

    if (length(safetensor_files) == 0) {
        stop("No safetensor files found in: ", model_path)
    }

    if (verbose) {
        message(sprintf("Loading weights from %d safetensor files...", length(safetensor_files)))
    }

    # Load and apply weights
    total_loaded <- 0L
    total_skipped <- 0L

    all_filled <- character(0)
    for (sf_path in safetensor_files) {
        if (verbose) {
            message("  Loading: ", basename(sf_path))
        }
        weights <- safetensors::safe_load_file(sf_path, framework = "torch")

        result <- load_gemma3_weights(model, weights, verbose = FALSE)
        total_loaded <- total_loaded + result$loaded
        total_skipped <- total_skipped + result$skipped
        all_filled <- c(all_filled, result$filled)
        rm(weights)
        gc(verbose = FALSE)
    }

    # The model is a skeleton: any parameter the checkpoint didn't fill
    # is uninitialized memory, so unfilled params are a hard error
    unfilled <- setdiff(names(model$parameters), all_filled)
    if (length(unfilled)) {
        stop("Gemma3 load: ", length(unfilled),
             " parameters not filled from checkpoint, e.g. ",
             paste(utils::head(unfilled, 3), collapse = ", "))
    }

    if (verbose) {
        message(sprintf("Loaded %d parameters, skipped %d", total_loaded, total_skipped))
    }

    model$to(device = device)

    if (pin && device == "cpu") {
        st <- .pin_component(model)
        if (!is.null(st)) {
            attr(model, "staging") <- st
            if (verbose) {
                message("Gemma3 weights pinned for staged transfer")
            }
        }
    }

    if (verbose) {
        message("Gemma3 text encoder loaded on device: ", device)
    }

    model
}

#' Load weights into Gemma3 model
#' @keywords internal
load_gemma3_weights <- function(model, weights, verbose = TRUE) {
    native_params <- names(model$parameters)

    # Remap HuggingFace keys to our module structure
    remap_gemma3_key <- function(key) {
        # Remove 'language_model.' prefix if present (for multimodal models)
        key <- sub("^language_model\\.", "", key)

        # Map HuggingFace structure to our structure:
        # model.embed_tokens -> embed_tokens
        # model.layers.0.self_attn.q_proj -> layers.1.self_attn.q_proj (1-indexed)
        # model.norm -> norm

        key <- sub("^model\\.", "", key)

        # Note: R torch model$parameters uses 0-indexed layer names even though
        # R list indexing is 1-based. So we keep the layer numbers as-is.

        key
    }

    loaded <- 0L
    skipped <- 0L
    filled <- character(0)

    torch::with_no_grad({
        for (hf_name in names(weights)) {
            native_name <- remap_gemma3_key(hf_name)

            if (native_name %in% native_params) {
                hf_tensor <- weights[[hf_name]]
                native_tensor <- model$parameters[[native_name]]

                if (all(as.integer(hf_tensor$shape) == as.integer(native_tensor$shape))) {
                    native_tensor$copy_(hf_tensor)
                    loaded <- loaded + 1L
                    filled <- c(filled, native_name)
                } else {
                    if (verbose) {
                        message("Shape mismatch: ", native_name,
                                " (HF: ", paste(as.integer(hf_tensor$shape),
                                collapse = "x"),
                                " vs R: ", paste(as.integer(native_tensor$shape), collapse = "x"), ")")
                    }
                    skipped <- skipped + 1L
                }
            } else {
                skipped <- skipped + 1L
            }
        }
    })

    if (verbose) {
        message(sprintf("Gemma3 weights: %d loaded, %d skipped", loaded, skipped))
    }

    invisible(list(loaded = loaded, skipped = skipped, filled = filled))
}

# -----------------------------------------------------------------------------
# Tokenizer Wrapper
# -----------------------------------------------------------------------------

#' Gemma3 Tokenizer
#'
#' Native R tokenizer for Gemma3 using BPE.
#' Loads from HuggingFace tokenizer.json format.
#'
#' @param tokenizer_path Character. Path to tokenizer directory or tokenizer.json file.
#' @return A gemma3_tokenizer object (extends bpe_tokenizer).
#' @export
gemma3_tokenizer <- function(tokenizer_path) {
    # Load base BPE tokenizer
    tokenizer <- bpe_tokenizer(tokenizer_path)

    # Add Gemma-specific configuration
    tokenizer$padding_side <- "left" # Gemma uses left padding

    # Update class

    class(tokenizer) <- c("gemma3_tokenizer", class(tokenizer))

    tokenizer
}

#' Tokenize text for Gemma3
#'
#' @param tokenizer Gemma3 tokenizer object.
#' @param text Character vector of prompts.
#' @param max_length Integer. Maximum sequence length.
#' @param padding Character. Padding strategy ("left", "right", "max_length", "none").
#' @param return_tensors Character. Return type ("list" or "pt" for torch tensors).
#' @return List with input_ids and attention_mask.
#' @export
tokenize_gemma3 <- function(tokenizer, text, max_length = 1024L,
                            padding = "max_length", return_tensors = "pt") {
    if (!inherits(tokenizer, "gemma3_tokenizer") &&
        !inherits(tokenizer, "bpe_tokenizer")) {
        stop("tokenizer must be a gemma3_tokenizer or bpe_tokenizer object")
    }

    # Use native BPE encoding
    result <- encode_bpe(tokenizer = tokenizer, text = text,
                         add_special_tokens = TRUE, max_length = max_length,
                         padding = padding, truncation = TRUE,
                         return_tensors = return_tensors)

    result
}

# -----------------------------------------------------------------------------
# Full Text Encoding Pipeline
# -----------------------------------------------------------------------------

#' Encode text with Gemma3 for LTX-2
#'
#' Full pipeline for encoding text prompts using Gemma3 text encoder.
#' Returns the raw stacked per-layer hidden states (embedding layer plus all
#' transformer layers) for downstream connector modules, which handle
#' normalization and projection themselves.
#'
#' @param prompts Character vector of prompts.
#' @param model Gemma3 text model (or path to load from).
#' @param tokenizer Gemma3 tokenizer (or path to load from).
#' @param max_sequence_length Integer. Maximum sequence length.
#' @param device Character. Device for computation.
#' @param dtype Character. Data type.
#' @param verbose Logical. Print progress.
#' @return List with prompt_embeds (raw stacked hidden states,
#'   shape \code{[batch, seq_len, hidden_size, num_layers + 1]}) and
#'   prompt_attention_mask.
#' @export
encode_with_gemma3 <- function(prompts, model = NULL, tokenizer = NULL,
                               max_sequence_length = 1024L, device = "cuda",
                               dtype = "float16", verbose = TRUE) {
    # Load model if path provided
    if (is.character(model)) {
        model <- load_gemma3_text_encoder(model, device = device,
            dtype = dtype, verbose = verbose)
    }

    # Load tokenizer if path provided
    if (is.character(tokenizer)) {
        tokenizer <- gemma3_tokenizer(tokenizer)
    }

    if (is.null(model) || is.null(tokenizer)) {
        stop("Both model and tokenizer are required")
    }

    # A pinned CPU-resident model (see the loaders' pin argument) swaps
    # to the compute device for the encode and back for free afterwards
    staging <- attr(model, "staging")
    if (!is.null(staging) && device != "cpu") {
        cur <- tryCatch(staging[[1]]$live$device$type, error = function(e) NULL)
        if (!identical(cur, device)) {
            .staged_onload(staging, device)
            on.exit(.staged_offload(staging), add = TRUE)
        }
    }

    # Ensure prompts is a list
    if (is.character(prompts)) {
        prompts <- as.list(prompts)
    }

    # Tokenize (left-padding to max_length for Gemma3)
    if (verbose) {
        message("Tokenizing prompts...")
    }
    tokens <- tokenize_gemma3(tokenizer, unlist(prompts),
                              max_length = max_sequence_length,
                              padding = "max_length")

    input_ids <- tokens$input_ids$to(device = device)
    attention_mask <- tokens$attention_mask$to(device = device)

    # Run through model
    if (verbose) {
        message("Encoding with Gemma3...")
    }
    torch::with_no_grad({
        output <- model(input_ids, attention_mask = attention_mask, output_hidden_states = TRUE)
    })

    # Stack ALL hidden states: embedding layer + every transformer layer.
    # The LTX connectors consume the full stack (their projection expects
    # hidden_size * (num_layers + 1) inputs) and normalize it themselves.
    hidden_states_list <- output$hidden_states
    # Stack: [batch, seq_len, hidden_size, num_layers + 1]
    hidden_states_stacked <- torch::torch_stack(hidden_states_list, dim = -1L)

    list(
         prompt_embeds = hidden_states_stacked,
         prompt_attention_mask = attention_mask
    )
}

# Null-coalescing operator
if (!exists("%||%", mode = "function")) {
    `%||%` <- function(x, y) if (is.null(x)) {
        y
    } else {
        x
    }
}

#' Batch-encode prompts with Gemma3, cached to disk
#'
#' Encodes a vector of prompts in sub-batches (bounding activation
#' VRAM) and optionally caches each prompt's result under
#' \code{cache_dir}, keyed by the prompt text and sequence length.
#' Already-cached prompts are skipped, so an interrupted batch resumes
#' where it stopped. Embeddings land on the CPU either way; the
#' renderer moves them per phase.
#'
#' Budget note: each prompt's result is the full hidden-state stack the
#' LTX connectors consume - roughly 0.4 GB at 1024 tokens - so caching
#' N prompts needs ~0.4N GB of disk.
#'
#' @param prompts Character vector.
#' @param model,tokenizer As in \code{\link{encode_with_gemma3}}.
#' @param batch_size Integer. Prompts per forward pass (default 4;
#'   raise on cards with headroom, lower if the encode OOMs).
#' @param cache_dir Optional directory. When given, each prompt is
#'   written to \code{gemma3-<md5>.pt} and the return value is the
#'   character vector of those paths (load one with
#'   \code{torch::torch_load}); when NULL, the return value is a list
#'   of \code{list(prompt_embeds, prompt_attention_mask)} in prompt
#'   order.
#' @param max_sequence_length,device,verbose As in
#'   \code{\link{encode_with_gemma3}}.
#'
#' @return Character vector of cache paths (with \code{cache_dir}) or
#'   a list of per-prompt embedding results (without).
#'
#' @export
gemma3_encode_batch <- function(prompts, model = NULL, tokenizer = NULL,
                                batch_size = 4L, cache_dir = NULL,
                                max_sequence_length = 1024L, device = "cuda",
                                verbose = TRUE) {
    prompts <- as.character(prompts)
    n <- length(prompts)
    if (!n) {
        stop("No prompts given")
    }

    cache_paths <- NULL
    if (!is.null(cache_dir)) {
        dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
        cache_paths <- vapply(prompts, function(p) {
            tf <- tempfile()
            writeLines(c(p, as.character(max_sequence_length)), tf)
            key <- unname(tools::md5sum(tf))
            unlink(tf)
            file.path(cache_dir, paste0("gemma3-", key, ".pt"))
        }, character(1), USE.NAMES = FALSE)
        todo <- which(!file.exists(cache_paths))
        if (verbose && length(todo) < n) {
            message(sprintf("%d/%d prompts already cached", n - length(todo),
                            n))
        }
    } else {
        todo <- seq_len(n)
    }

    if (is.null(cache_dir)) {
        results <- vector("list", n)
    } else {
        results <- NULL
    }
    done <- 0L
    for (chunk in split(todo, ceiling(seq_along(todo) / batch_size))) {
        enc <- encode_with_gemma3(prompts[chunk], model = model,
                                  tokenizer = tokenizer,
                                  max_sequence_length = max_sequence_length,
                                  device = device, verbose = FALSE)
        embeds <- enc$prompt_embeds$cpu()
        mask <- enc$prompt_attention_mask$cpu()
        for (j in seq_along(chunk)) {
            # clone(), not contiguous(): a narrow of a contiguous
            # tensor is already contiguous, so contiguous() keeps the
            # storage-offset view - and torch_save serializes offset
            # views from the base storage (every row would save as
            # row 1). clone() forces fresh storage.
            one <- list(
                        prompt_embeds = embeds$narrow(1L, j, 1L)$clone(),
                        prompt_attention_mask = mask$narrow(1L, j, 1L)$clone()
            )
            if (is.null(cache_dir)) {
                results[[chunk[j]]] <- one
            } else {
                torch::torch_save(one, cache_paths[chunk[j]])
            }
        }
        rm(enc, embeds, mask)
        gc(verbose = FALSE)
        done <- done + length(chunk)
        if (verbose) {
            message(sprintf("  encoded %d/%d prompts", done, length(todo)))
        }
    }

    if (is.null(cache_dir)) {
        results
    } else {
        cache_paths
    }
}

# load_gemma3_text_encoder(): skeleton construction at the target dtype,
# checkpoint fill, and the hard error when a parameter is not covered by
# the checkpoint (skeleton weights are uninitialized memory, so a silent
# partial load must be impossible).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

# Tiny model dir: config.json + one model-*.safetensors shard whose keys
# carry the HF "model." prefix, generated from a bare (initialized) model
config <- list(
  vocab_size = 64L, hidden_size = 16L, intermediate_size = 32L,
  num_hidden_layers = 2L, num_attention_heads = 2L,
  num_key_value_heads = 1L, head_dim = 8L, query_pre_attn_scalar = 8L,
  sliding_window = 4L, sliding_window_pattern = 2L
)

ref_config <- gemma3_config_ltx2()
for (nm in names(config)) ref_config[[nm]] <- config[[nm]]
torch::torch_manual_seed(11)
ref <- gemma3_text_model(ref_config)
ref$eval()

dir <- file.path(tempdir(), "gemma3-loader-test")
dir.create(dir, showWarnings = FALSE)
jsonlite::write_json(config, file.path(dir, "config.json"), auto_unbox = TRUE)

params <- ref$parameters
weights <- stats::setNames(
  lapply(params, function(p) p$detach()$contiguous()),
  paste0("model.", names(params))
)
safetensors::safe_save_file(weights, file.path(dir, "model-00001-of-00001.safetensors"))

# --- full checkpoint loads; weights and forward match the reference ----------------

enc <- load_gemma3_text_encoder(dir, device = "cpu", dtype = "float32",
                                verbose = FALSE)
expect_true(all(names(enc$parameters) %in% names(params)))
expect_true(as.numeric((enc$parameters[["embed_tokens.weight"]] -
                        params[["embed_tokens.weight"]])$abs()$max()) == 0)

ids <- torch::torch_randint(1L, 64L, size = c(1L, 5L), dtype = torch::torch_long())
mask <- torch::torch_ones(1L, 5L)
torch::with_no_grad({
  out_ref <- ref(ids, attention_mask = mask, output_hidden_states = TRUE)
  out_enc <- enc(ids, attention_mask = mask, output_hidden_states = TRUE)
})
expect_true(as.numeric((out_enc$last_hidden_state -
                        out_ref$last_hidden_state)$abs()$max()) < 1e-6)

# --- skeleton dtype follows the dtype argument -------------------------------------

enc16 <- load_gemma3_text_encoder(dir, device = "cpu", dtype = "float16",
                                  verbose = FALSE)
expect_true(enc16$parameters[[1]]$dtype == torch::torch_float16())

# --- a checkpoint that misses a parameter is a hard error --------------------------

partial <- weights[setdiff(names(weights), "model.norm.weight")]
unlink(file.path(dir, "model-00001-of-00001.safetensors"))
safetensors::safe_save_file(partial, file.path(dir, "model-00001-of-00001.safetensors"))
expect_error(
  load_gemma3_text_encoder(dir, device = "cpu", dtype = "float32",
                           verbose = FALSE),
  "not filled"
)

unlink(dir, recursive = TRUE)

# --- pinned staging surface ----------------------------------------------------------

# Both loaders take pin (default: the pin_staging option). Actual
# page-locking needs CUDA; validated on GPU outside R CMD check.
expect_true("pin" %in% names(formals(load_gemma3_text_encoder)))
expect_true("pin" %in% names(formals(load_gemma3_nf4)))

# Hidden-state collection semantics of the Gemma3 text encoder.
#
# HF transformers returns num_layers + 1 hidden states: the state BEFORE
# each decoder layer (the first being the embedding output) plus the
# post-final-norm output. The un-normed last-layer output is never
# included. The LTX connectors' input projection expects exactly
# hidden_size * (num_layers + 1) features, so an extra or missing state
# fails at the first matmul with real weights.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)

config <- gemma3_config_ltx2()
config$vocab_size <- 64L
config$hidden_size <- 16L
config$intermediate_size <- 32L
config$num_hidden_layers <- 3L
config$num_attention_heads <- 2L
config$num_key_value_heads <- 1L
config$head_dim <- 8L
config$query_pre_attn_scalar <- 8L
config$sliding_window <- 4L

model <- gemma3_text_model(config)
model$eval()

ids <- torch::torch_randint(1L, 64L, size = c(1L, 7L), dtype = torch::torch_long())
mask <- torch::torch_ones(1L, 7L)
torch::with_no_grad({
  out <- model(ids, attention_mask = mask, output_hidden_states = TRUE)
})

# num_layers + 1 states
expect_equal(length(out$hidden_states), config$num_hidden_layers + 1L)

# Last collected state is the post-norm output (identical to last_hidden_state)
last_collected <- out$hidden_states[[length(out$hidden_states)]]
expect_true(as.numeric((last_collected - out$last_hidden_state)$abs()$max()) == 0)

# The un-normed last-layer output must NOT be in the stack: the
# second-to-last entry is the input to the final layer, so running the
# final layer + norm over it must reproduce last_hidden_state.
pre_last <- out$hidden_states[[config$num_hidden_layers]]
expect_false(as.numeric((pre_last - out$last_hidden_state)$abs()$max()) == 0)

# Without the flag, no states are collected
torch::with_no_grad({
  out2 <- model(ids, attention_mask = mask, output_hidden_states = FALSE)
})
expect_null(out2$hidden_states)

# gemma3_encode_batch: sub-batched encoding matches solo encodes, disk
# caching returns paths, resumes without re-encoding, and round-trips
# through torch_load.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}

library(diffuseR)

config <- gemma3_config_ltx2()
for (nm in names(cfg <- list(
  vocab_size = 64L, hidden_size = 16L, intermediate_size = 32L,
  num_hidden_layers = 2L, num_attention_heads = 2L,
  num_key_value_heads = 1L, head_dim = 8L, query_pre_attn_scalar = 8L,
  sliding_window = 4L, sliding_window_pattern = 2L
))) config[[nm]] <- cfg[[nm]]
torch::torch_manual_seed(7)
model <- gemma3_text_model(config)
model$eval()

# A stub tokenizer: encode_with_gemma3 only needs tokenize_gemma3 to
# work, which needs ids within vocab; use the real path via a fake
# minimal bpe? Simpler: call encode_with_gemma3 through its tokens by
# monkeying is overkill - instead drive the batch helper with a real
# tokenizer-free path: fake tokenizer via the exported tokenize seam is
# not available, so tokenize through gemma3_tokenizer would need vocab
# files. Drive the model directly instead for parity, and the batch
# helper end-to-end with a trivial tokenizer object.

# Minimal tokenizer standing in for gemma3_tokenizer: the helper only
# passes it through to encode_with_gemma3 -> tokenize_gemma3, which
# calls encode_bpe; emulate the contract with a closure-based double.
fake_tok <- structure(list(), class = "fake_tok")
tokenize_real <- diffuseR:::tokenize_gemma3

# Patch tokenize_gemma3 inside the namespace for this test: map each
# prompt deterministically to ids from its characters (padded left)
fake_tokenize <- function(tokenizer, prompts, max_length = 16L,
                          padding = "max_length") {
  n <- length(prompts)
  ids <- torch::torch_zeros(n, max_length, dtype = torch::torch_long())
  mask <- torch::torch_zeros(n, max_length)
  for (i in seq_len(n)) {
    v <- utf8ToInt(prompts[[i]]) %% 60L + 1L
    v <- utils::tail(v, max_length)
    ids[i, (max_length - length(v) + 1L):max_length] <-
      torch::torch_tensor(v, dtype = torch::torch_long())
    mask[i, (max_length - length(v) + 1L):max_length] <- 1
  }
  list(input_ids = ids, attention_mask = mask)
}
# No on.exit here: at test-file top level tinytest evaluates each
# expression in its own frame, so on.exit would fire immediately and
# restore the real tokenizer before the calls below. Restored at the
# end of the file instead.
assignInNamespace("tokenize_gemma3", fake_tokenize, ns = "diffuseR")

prompts <- c("a cat on a mat", "neon fog rolling", "a robot dancing")

# --- no cache: list results match solo encodes -------------------------------------

batch <- gemma3_encode_batch(prompts, model = model, tokenizer = fake_tok,
                             batch_size = 2L, max_sequence_length = 16L,
                             device = "cpu", verbose = FALSE)
expect_equal(length(batch), 3L)
solo <- encode_with_gemma3(prompts[2], model = model, tokenizer = fake_tok,
                           max_sequence_length = 16L, device = "cpu",
                           verbose = FALSE)
expect_equal(as.integer(batch[[2]]$prompt_embeds$shape),
             as.integer(solo$prompt_embeds$shape))
expect_true(as.numeric((batch[[2]]$prompt_embeds -
                        solo$prompt_embeds)$abs()$max()) < 1e-5)
expect_true(as.numeric((batch[[2]]$prompt_attention_mask -
                        solo$prompt_attention_mask)$abs()$max()) == 0)

# --- cache: paths returned, resume skips, round-trip matches -----------------------

cache <- file.path(tempdir(), "gemma3-batch-cache")
unlink(cache, recursive = TRUE)
paths <- gemma3_encode_batch(prompts, model = model, tokenizer = fake_tok,
                             batch_size = 2L, cache_dir = cache,
                             max_sequence_length = 16L, device = "cpu",
                             verbose = FALSE)
expect_equal(length(paths), 3L)
expect_true(all(file.exists(paths)))

loaded <- torch::torch_load(paths[2])
expect_true(as.numeric((loaded$prompt_embeds -
                        batch[[2]]$prompt_embeds)$abs()$max()) < 1e-6)

# Resume: files untouched on the second call (no re-encode)
mt <- file.mtime(paths)
paths2 <- gemma3_encode_batch(prompts, model = model, tokenizer = fake_tok,
                              batch_size = 2L, cache_dir = cache,
                              max_sequence_length = 16L, device = "cpu",
                              verbose = FALSE)
expect_equal(paths2, paths)
expect_equal(file.mtime(paths), mt)

# A new prompt encodes into the same cache without touching the others
paths3 <- gemma3_encode_batch(c(prompts, "one more scene"), model = model,
                              tokenizer = fake_tok, batch_size = 2L,
                              cache_dir = cache,
                              max_sequence_length = 16L, device = "cpu",
                              verbose = FALSE)
expect_equal(length(paths3), 4L)
expect_equal(paths3[1:3], paths)
expect_true(file.exists(paths3[4]))

unlink(cache, recursive = TRUE)
assignInNamespace("tokenize_gemma3", tokenize_real, ns = "diffuseR")

#' FLUX Checkpoint Readers
#'
#' FLUX transformers ship in the diffusers layout: a directory with
#' \code{config.json}, one or more \code{diffusion_pytorch_model*.safetensors}
#' shards, and (when sharded) a
#' \code{diffusion_pytorch_model.safetensors.index.json} weight map.
#' These helpers open that layout behind the same checkpoint interface as
#' \code{\link{ltx23_open_checkpoint}}, so the LTX group loaders and
#' quantization machinery work unchanged. FLUX module names mirror the
#' checkpoint keys 1:1 - no key mapping is needed.
#'
#' @name checkpoint_flux
NULL

# Quantization cast set: every large linear weight in the FLUX blocks,
# including the adaLN modulation linears (norm*.linear are 3.2B params
# across the model; leaving them bf16 would not fit resident on 16 GB).
# Everything else (embedders, q/k norms, final norm, biases) stays bf16.
# Full-size census: 19 double blocks x 14 + 38 single blocks x 6 = 494
# cast weights, ~11.8B of the 12B parameters.
.flux_quant_cast_pattern <- paste0(
                                   "^(",
                                   "transformer_blocks\\.[0-9]+\\.(",
                                   "attn\\.(to_q|to_k|to_v|add_q_proj|add_k_proj|add_v_proj|to_out\\.0|to_add_out)",
                                   "|ff\\.net\\.(0\\.proj|2)|ff_context\\.net\\.(0\\.proj|2)",
                                   "|norm1\\.linear|norm1_context\\.linear",
                                   ")",
                                   "|single_transformer_blocks\\.[0-9]+\\.(",
                                   "attn\\.(to_q|to_k|to_v)|proj_mlp|proj_out|norm\\.linear",
                                   ")",
                                   ")\\.weight$"
)

#' Test whether a FLUX key is in the quantization cast set
#'
#' @param key Character vector of parameter names (diffusers-style).
#'
#' @return Logical vector.
#'
#' @export
flux_is_quant_key <- function(key) {
    grepl(.flux_quant_cast_pattern, key)
}

# FLUX.2 cast set: block linears, the three shared modulation
# projections, and the context embedder (7680 x 3072). Everything else
# (x_embedder, timestep MLP, norm_out, q/k norms) stays in the resident
# dtype. Full klein-4B census: 5 double x 12 + 20 single x 2 + 3
# modulations + context_embedder = 104 cast weights, ~3.9B of 4B params.
.flux2_quant_cast_pattern <- paste0(
                                    "^(",
                                    "transformer_blocks\\.[0-9]+\\.(",
                                    "attn\\.(to_q|to_k|to_v|add_q_proj|add_k_proj|add_v_proj|to_out\\.0|to_add_out)",
                                    "|ff\\.(linear_in|linear_out)|ff_context\\.(linear_in|linear_out)",
                                    ")",
                                    "|single_transformer_blocks\\.[0-9]+\\.attn\\.(to_qkv_mlp_proj|to_out)",
                                    "|double_stream_modulation_img\\.linear",
                                    "|double_stream_modulation_txt\\.linear",
                                    "|single_stream_modulation\\.linear",
                                    "|context_embedder",
                                    ")\\.weight$"
)

# Z-Image cast set: the seven core linears (attention q/k/v/out and the
# three SwiGLU weights) in all three block stacks. The per-block adaLN
# modulation linears, embedders, pad tokens and norms stay in the
# resident dtype. Full Turbo census: (30 layers + 2 noise_refiner + 2
# context_refiner) x 7 = 238 cast weights, ~6.0B of the 6B parameters.
.zimage_quant_cast_pattern <- paste0(
                                     "^(noise_refiner|context_refiner|layers)\\.[0-9]+\\.(",
                                     "attention\\.(to_q|to_k|to_v|to_out\\.0)",
                                     "|feed_forward\\.(w1|w2|w3)",
                                     ")\\.weight$"
)

#' Test whether a Z-Image key is in the quantization cast set
#'
#' @param key Character vector of parameter names (diffusers-style).
#'
#' @return Logical vector.
#'
#' @export
zimage_is_quant_key <- function(key) {
    grepl(.zimage_quant_cast_pattern, key)
}

#' Test whether a FLUX.2 key is in the quantization cast set
#'
#' @param key Character vector of parameter names (diffusers-style).
#'
#' @return Logical vector.
#'
#' @export
flux2_is_quant_key <- function(key) {
    grepl(.flux2_quant_cast_pattern, key)
}

# Model family from a diffusers transformer config
.flux_family <- function(config) {
    cls <- config$`_class_name` %||% "FluxTransformer2DModel"
    if (identical(cls, "Flux2Transformer2DModel")) {
        "flux2"
    } else if (identical(cls, "ZImageTransformer2DModel")) {
        "zimage"
    } else {
        "flux1"
    }
}

#' Open a FLUX transformer checkpoint directory
#'
#' Opens a diffusers-layout transformer directory lazily (headers only).
#' Sharded checkpoints are resolved through the index.json weight map;
#' single-file checkpoints are opened directly. The transformer
#' \code{config.json} is attached as \code{$config}.
#'
#' @param transformer_dir Directory containing \code{config.json} and the
#'   \code{diffusion_pytorch_model*.safetensors} file(s).
#'
#' @return An object of class \code{ltx23_checkpoint} (shared checkpoint
#'   interface): list with \code{handle$get_tensor}, \code{keys},
#'   \code{config}, and \code{path}.
#'
#' @export
flux_open_checkpoint <- function(transformer_dir) {
    if (!requireNamespace("safetensors", quietly = TRUE)) {
        stop("The safetensors package is required to read FLUX checkpoints.")
    }
    transformer_dir <- path.expand(transformer_dir)
    if (!dir.exists(transformer_dir)) {
        stop("Checkpoint directory not found: ", transformer_dir)
    }

    config <- NULL
    config_path <- file.path(transformer_dir, "config.json")
    if (file.exists(config_path)) {
        config <- jsonlite::fromJSON(config_path, simplifyVector = TRUE)
    }

    opened <- .flux_open_sharded_dir(transformer_dir, "diffusion_pytorch_model")

    structure(
              list(handle = opened$handle, keys = opened$keys, version = NULL,
                   config = config, path = transformer_dir),
              class = "ltx23_checkpoint"
    )
}

# Open a HF-layout safetensors directory (sharded via <base>.safetensors
# .index.json, or a single <base>.safetensors) lazily; returns the
# handle/keys pair shared by all checkpoint objects.
.flux_open_sharded_dir <- function(dir, base) {
    index_path <- file.path(dir, paste0(base, ".safetensors.index.json"))
    if (file.exists(index_path)) {
        index <- jsonlite::fromJSON(index_path, simplifyVector = TRUE)
        weight_map <- unlist(index$weight_map)
        shard_files <- unique(weight_map)
        missing <- shard_files[!file.exists(file.path(dir, shard_files))]
        if (length(missing)) {
            stop("Missing checkpoint shards: ", paste(missing, collapse = ", "))
        }
        handles <- lapply(file.path(dir, shard_files), function(p) {
            safetensors::safetensors$new(p, framework = "torch")
        })
        names(handles) <- shard_files
        keys <- names(weight_map)
        handle <- list(
                       get_tensor = function(key) {
            shard <- weight_map[[key]]
            if (is.null(shard) || is.na(shard)) {
                stop("Key not found in checkpoint index: ", key)
            }
            .st_read_or_breadcrumb(function() handles[[shard]]$get_tensor(key),
                                   file.path(dir, shard))
        }
        )
    } else {
        single_path <- file.path(dir, paste0(base, ".safetensors"))
        if (!file.exists(single_path)) {
            stop("No ", base, " safetensors (or index) in ", dir)
        }
        h <- safetensors::safetensors$new(single_path, framework = "torch")
        keys <- setdiff(h$keys(), "__metadata__")
        handle <- list(get_tensor = function(key) {
            .st_read_or_breadcrumb(function() h$get_tensor(key), single_path)
        })
    }
    list(handle = handle, keys = keys)
}

#' Open a quantized FLUX artifact directory
#'
#' Opens the sharded NF4/fp8 artifact written by
#' \code{\link{flux_quantize}} through the shared checkpoint interface.
#' The manifest's embedded transformer config and \code{format} ride
#' along, so \code{\link{flux_load_transformer}} needs nothing else.
#'
#' @param dir The quantized artifact directory (with manifest.json).
#'
#' @return An \code{ltx23_checkpoint} with \code{$format} set.
#'
#' @export
flux_open_quantized <- function(dir) {
    manifest_path <- file.path(dir, "manifest.json")
    if (!file.exists(manifest_path)) {
        stop("No manifest.json in ", dir, "; run flux_quantize() first.")
    }
    ltx23_open_fp8_checkpoint(dir)
}

# reshard_safetensors: split a single safetensors into sub-2 GB shards
# with a diffusers index, and read it back through the sharded reader.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}
library(diffuseR)

# three 100x100 float32 tensors (~40 KB each)
set.seed(1)
tensors <- list(a = torch::torch_randn(100L, 100L),
                b = torch::torch_randn(100L, 100L),
                c = torch::torch_randn(100L, 100L))
src <- tempfile(fileext = ".safetensors")
safetensors::safe_save_file(tensors, src)

out <- tempfile()
# ~60 KB target forces one tensor per shard -> three shards
reshard_safetensors(src, out, base = "model", shard_bytes = 60000,
  verbose = FALSE)

idx <- file.path(out, "model.safetensors.index.json")
expect_true(file.exists(idx))
j <- jsonlite::fromJSON(idx, simplifyVector = FALSE)
expect_equal(length(j$weight_map), 3L)                 # every key mapped
expect_true(length(unique(unlist(j$weight_map))) >= 2) # actually sharded
# diffusers -of- naming
expect_true(all(grepl("model-\\d{5}-of-\\d{5}\\.safetensors",
  unlist(j$weight_map))))

# read back through the sharded reader: same keys, same values
opened <- diffuseR:::.flux_open_sharded_dir(out, "model")
expect_true(setequal(opened$keys, c("a", "b", "c")))
for (k in c("a", "b", "c")) {
  expect_true(as.logical(torch::torch_allclose(
    opened$handle$get_tensor(k), tensors[[k]])))
}

unlink(c(src, out), recursive = TRUE)

# FLUX checkpoint opening + quantization round trip on the tiny sharded
# diffusers checkpoint (fixture generated by tools/gen_fixtures_flux_model.py,
# checked in). Everything runs on CPU.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

ckpt_dir <- system.file("tinytest", "fixtures", "flux_tiny_ckpt",
  package = "diffuseR")
if (ckpt_dir == "") ckpt_dir <- "fixtures/flux_tiny_ckpt"
if (!dir.exists(ckpt_dir)) exit_file("flux tiny checkpoint missing")

fixture_path <- system.file("tinytest", "fixtures", "flux_model.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/flux_model.safetensors"
if (!file.exists(fixture_path)) exit_file("flux model fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}
cosine_sim <- function(a, b) {
  a <- a$to(dtype = torch::torch_float32())$flatten()
  b <- b$to(dtype = torch::torch_float32())$flatten()
  as.numeric(torch::torch_dot(a, b) / (a$norm() * b$norm()))
}

# --- cast-set census ------------------------------------------------------------
# Tiny config: 1 double block x 14 + 1 single block x 6 = 20 cast weights.
# (Full schnell: 19 x 14 + 38 x 6 = 494.)

ckpt <- flux_open_checkpoint(ckpt_dir)
expect_true(inherits(ckpt, "ltx23_checkpoint"))
expect_equal(sum(flux_is_quant_key(ckpt$keys)), 20L)
expect_equal(ckpt$config$num_layers, 1L)

# The sharded index resolves every key to a readable tensor
t0 <- ckpt$handle$get_tensor("x_embedder.weight")
expect_equal(as.integer(t0$shape), c(16L, 4L))

# --- full-precision load through the sharded index --------------------------------

model <- flux_load_transformer(ckpt, device = "cpu", dtype = "float32",
  verbose = FALSE)
ids <- torch::torch_cat(list(fx$txt_ids, fx$img_ids), dim = 1L)
rope <- flux_pos_embed(ids, axes_dim = c(2L, 2L, 4L))
out_full <- torch::with_no_grad(model(
  hidden_states = fx$hidden,
  encoder_hidden_states = fx$encoder,
  pooled_projections = fx$pooled,
  timestep = fx$timestep,
  image_rotary_emb = rope
))
expect_true(max_abs_diff(out_full, fx$out) < 1e-4)

# --- NF4 round trip -----------------------------------------------------------------

nf4_dir <- file.path(tempdir(), "flux-tiny-nf4")
unlink(nf4_dir, recursive = TRUE)
manifest <- flux_quantize(ckpt_dir, output_dir = nf4_dir, format = "nf4",
  verbose = FALSE)
expect_equal(manifest$cast, 20L)
expect_true(file.exists(file.path(nf4_dir, "manifest.json")))

nf4_ckpt <- flux_open_quantized(nf4_dir)
expect_equal(nf4_ckpt$format, "nf4")
model_nf4 <- flux_load_transformer(nf4_ckpt, device = "cpu", verbose = FALSE)
out_nf4 <- torch::with_no_grad(model_nf4(
  hidden_states = fx$hidden$to(dtype = torch::torch_bfloat16()),
  encoder_hidden_states = fx$encoder$to(dtype = torch::torch_bfloat16()),
  pooled_projections = fx$pooled$to(dtype = torch::torch_bfloat16()),
  timestep = fx$timestep,
  image_rotary_emb = rope
))
expect_true(all(is.finite(as.numeric(out_nf4$to(dtype = torch::torch_float32())))))
expect_true(cosine_sim(out_nf4, out_full) > 0.98)
ltx23_release_dequant_buffers()

# --- fp8 round trip (needs an F8-capable safetensors build) ---------------------------

# Ask the same gate flux_quantize() asks, rather than re-probing here: a
# private probe can disagree with .st_can_write() whenever the
# diffuseR.st_caps override is set, and then this block runs only to hit
# the hard error the override was meant to simulate.
f8_ok <- diffuseR:::.st_can_write("float8_e4m3fn")

if (f8_ok) {
  fp8_dir <- file.path(tempdir(), "flux-tiny-fp8")
  unlink(fp8_dir, recursive = TRUE)
  manifest8 <- flux_quantize(ckpt_dir, output_dir = fp8_dir, format = "fp8",
    verbose = FALSE)
  expect_equal(manifest8$cast, 20L)

  fp8_ckpt <- flux_open_quantized(fp8_dir)
  model_fp8 <- flux_load_transformer(fp8_ckpt, device = "cpu", pin = FALSE,
    verbose = FALSE)
  out_fp8 <- torch::with_no_grad(model_fp8(
    hidden_states = fx$hidden$to(dtype = torch::torch_bfloat16()),
    encoder_hidden_states = fx$encoder$to(dtype = torch::torch_bfloat16()),
    pooled_projections = fx$pooled$to(dtype = torch::torch_bfloat16()),
    timestep = fx$timestep,
    image_rotary_emb = rope
  ))
  expect_true(cosine_sim(out_fp8, out_full) > 0.99)
}

options(diffuseR.block_gc = NULL)

# FLUX.2 quantization round trip on the tiny sharded checkpoint:
# family auto-detection via config _class_name, cast census, NF4 and
# fp8 (streamed + resident) loads. Everything runs on CPU.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

ckpt_dir <- system.file("tinytest", "fixtures", "flux2_tiny_ckpt",
  package = "diffuseR")
if (ckpt_dir == "") ckpt_dir <- "fixtures/flux2_tiny_ckpt"
if (!dir.exists(ckpt_dir)) exit_file("flux2 tiny checkpoint missing")

fixture_path <- system.file("tinytest", "fixtures", "flux2_model.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/flux2_model.safetensors"
if (!file.exists(fixture_path)) exit_file("flux2 model fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}
cosine_sim <- function(a, b) {
  a <- a$to(dtype = torch::torch_float32())$flatten()
  b <- b$to(dtype = torch::torch_float32())$flatten()
  as.numeric(torch::torch_dot(a, b) / (a$norm() * b$norm()))
}

# --- family detection + cast census ------------------------------------------------
# Tiny config: 1 double x 12 + 1 single x 2 + 3 modulations +
# context_embedder = 18 cast weights. (Full klein-4B: 104.)

ckpt <- flux_open_checkpoint(ckpt_dir)
expect_equal(diffuseR:::.flux_family(ckpt$config), "flux2")
expect_equal(sum(flux2_is_quant_key(ckpt$keys)), 18L)

# --- full-precision load through family dispatch ------------------------------------

model <- flux_load_transformer(ckpt, device = "cpu", dtype = "float32",
  verbose = FALSE)
ids <- torch::torch_cat(list(fx$txt_ids, fx$img_ids), dim = 1L)
rope <- flux_pos_embed(ids, axes_dim = c(2L, 2L, 2L, 2L), theta = 2000)
out_full <- torch::with_no_grad(model(
  hidden_states = fx$hidden,
  encoder_hidden_states = fx$encoder,
  timestep = fx$timestep,
  image_rotary_emb = rope
))
expect_true(max_abs_diff(out_full, fx$out) < 1e-4)

# --- NF4 round trip ------------------------------------------------------------------

nf4_dir <- file.path(tempdir(), "flux2-tiny-nf4")
unlink(nf4_dir, recursive = TRUE)
manifest <- flux_quantize(ckpt_dir, output_dir = nf4_dir, format = "nf4",
  verbose = FALSE)
expect_equal(manifest$cast, 18L)
expect_true(grepl("^flux2-klein-nf4", manifest$shards[[1]]))

model_nf4 <- flux_load_transformer(flux_open_quantized(nf4_dir),
  device = "cpu", verbose = FALSE)
out_nf4 <- torch::with_no_grad(model_nf4(
  hidden_states = fx$hidden$to(dtype = torch::torch_bfloat16()),
  encoder_hidden_states = fx$encoder$to(dtype = torch::torch_bfloat16()),
  timestep = fx$timestep,
  image_rotary_emb = rope
))
expect_true(cosine_sim(out_nf4, out_full) > 0.98)
ltx23_release_dequant_buffers()

# --- fp8 round trips (streamed and resident) ------------------------------------------

# Same gate the quantizer uses, so the diffuseR.st_caps override
# reaches this guard too; a private probe would disagree with it.
f8_ok <- diffuseR:::.st_can_write("float8_e4m3fn")

if (f8_ok) {
  fp8_dir <- file.path(tempdir(), "flux2-tiny-fp8")
  unlink(fp8_dir, recursive = TRUE)
  manifest8 <- flux_quantize(ckpt_dir, output_dir = fp8_dir, format = "fp8",
    verbose = FALSE)
  expect_equal(manifest8$cast, 18L)

  model_fp8 <- flux_load_transformer(flux_open_quantized(fp8_dir),
    device = "cpu", pin = FALSE, verbose = FALSE)
  out_fp8 <- torch::with_no_grad(model_fp8(
    hidden_states = fx$hidden$to(dtype = torch::torch_bfloat16()),
    encoder_hidden_states = fx$encoder$to(dtype = torch::torch_bfloat16()),
    timestep = fx$timestep,
    image_rotary_emb = rope
  ))
  expect_true(cosine_sim(out_fp8, out_full) > 0.99)

  # Resident variant: weights moved to the compute device (CPU here just
  # exercises the walker), forward unchanged
  model_res <- flux_load_transformer(flux_open_quantized(fp8_dir),
    device = "cpu", pin = FALSE, fp8_resident = TRUE, verbose = FALSE)
  out_res <- torch::with_no_grad(model_res(
    hidden_states = fx$hidden$to(dtype = torch::torch_bfloat16()),
    encoder_hidden_states = fx$encoder$to(dtype = torch::torch_bfloat16()),
    timestep = fx$timestep,
    image_rotary_emb = rope
  ))
  expect_true(max_abs_diff(out_res, out_fp8) == 0)
}

options(diffuseR.block_gc = NULL)

#' NF4 Weight Storage for the LTX-2.3 Transformer
#'
#' 4-bit NormalFloat quantization (the QLoRA scheme: per-block absmax
#' normalization against a 16-level quantile code, two indices packed
#' per byte). At ~4.5 bits/parameter the 22B transformer fits in about
#' 12.5 GB, small enough to stay resident on a 16 GB GPU: no per-step
#' PCIe weight streaming, at a small quality cost relative to fp8.
#' Quantization and dequantization are pure torch ops (bucketize,
#' index_select) - no custom kernels.
#'
#' @name nf4_ltx23
NULL

# The 16 NF4 quantile levels (QLoRA, Dettmers et al. 2023)
.ltx23_nf4_table <- c(-1.0, -0.6961928009986877, -0.5250730514526367,
                      -0.39491748809814453, -0.28444138169288635,
                      -0.18477343022823334, -0.09105003625154495, 0.0,
                      0.07958029955625534, 0.16093020141124725,
                      0.24611230194568634, 0.33791524171829224,
                      0.44070982933044434, 0.5626170039176941,
                      0.7229568362236023, 1.0)

.ltx23_nf4_block_size <- 64L

#' Quantize a tensor to NF4
#'
#' @param x Float tensor (any shape; total elements must be a multiple
#'   of 128, i.e. two 64-element blocks - always true for the LTX
#'   linears).
#'
#' @return List with \code{packed} (uint8, two indices per byte) and
#'   \code{absmax} (float32, one per 64-element block).
#'
#' @export
ltx23_nf4_quantize <- function(x) {
    n <- prod(x$shape)
    block <- .ltx23_nf4_block_size
    if (n %% block != 0L) {
        stop("Tensor length must be a multiple of ", block)
    }

    table <- torch::torch_tensor(.ltx23_nf4_table,
                                 dtype = torch::torch_float32(),
                                 device = x$device)
    # Decision boundaries at the midpoints between adjacent levels
    midpoints <- (table$narrow(1L, 1L, 15L) + table$narrow(1L, 2L, 15L)) / 2

    blocks <- x$detach()$to(dtype = torch::torch_float32())$reshape(c(-1L, block))
    absmax <- blocks$abs()$amax(dim = 2L)$clamp(min = 1e-12)
    normalized <- blocks / absmax$unsqueeze(2L)

    idx <- torch::torch_bucketize(normalized$flatten(), midpoints) # 0..15
    idx <- idx$to(dtype = torch::torch_uint8())

    # Pack pairs: first index in the high nibble
    pairs <- idx$reshape(c(-1L, 2L))
    packed <- pairs$narrow(2L, 1L, 1L)$squeeze(2L) * 16L +
    pairs$narrow(2L, 2L, 1L)$squeeze(2L)

    list(packed = packed, absmax = absmax)
}

#' Dequantize NF4 data to a float tensor
#'
#' @param packed uint8 tensor of packed index pairs.
#' @param absmax float32 tensor of per-block scales.
#' @param shape Integer vector. Original tensor shape.
#' @param dtype Target torch dtype.
#' @param chunk_elements Integer. Elements dequantized per slice (bounds
#'   the int64 index temporary).
#' @param out Optional preallocated tensor of \code{shape} to write into
#'   (avoids allocating a fresh weight tensor per call).
#'
#' @return Tensor of \code{shape} in \code{dtype} on the input's device.
#'
#' @export
ltx23_nf4_dequantize <- function(packed, absmax, shape,
                                 dtype = torch::torch_bfloat16(),
                                 chunk_elements = 8388608L, out = NULL) {
    block <- .ltx23_nf4_block_size

    if (is.null(out)) {
        out <- torch::torch_empty(shape, dtype = dtype, device = packed$device)
    }
    out_pairs <- out$view(c(-1L, 2L))
    out_blocks <- out$view(c(-1L, block))

    n_bytes <- packed$shape[1]
    bytes_per_chunk <- max(chunk_elements %/% 2L, block)
    scratch <- .ltx23_get_dequant_scratch(min(bytes_per_chunk, n_bytes),
        packed$device)
    # Byte LUT [256, 2] in the target dtype (shared with the jit block
    # stack): one 1-based index per PACKED byte and a single gather of
    # (hi, lo) pairs written straight into out - no nibble unpack, no
    # per-value int64 index, no float32 staging copy
    tbl <- .ltx23_jit_table(packed$device, dtype)
    am <- absmax$to(dtype = dtype)

    start <- 1L
    torch::with_no_grad({
        while (start <= n_bytes) {
            len <- min(bytes_per_chunk, n_bytes - start + 1L)
            idx <- scratch$idx$narrow(1L, 1L, len)
            idx$copy_(packed$narrow(1L, start, len))$add_(1L)
            .ltx23_index_select_into(out_pairs$narrow(1L, start, len), tbl, idx)

            block_start <- ((start - 1L) * 2L) %/% block + 1L
            n_blocks <- (len * 2L) %/% block
            out_blocks$narrow(1L, block_start, n_blocks)$
            mul_(am$narrow(1L, block_start, n_blocks)$unsqueeze(2L))
            start <- start + len
        }
    })
    out
}

# torch_index_select_out is not exported from torch; fall back to an
# allocating index_select if it ever disappears
.ltx23_index_select_fn <- local({
    fn <- NULL
    function() {
        if (is.null(fn)) {
            fn <<- tryCatch(
                            get("torch_index_select_out", envir = asNamespace("torch")),
                            error = function(e) FALSE
            )
        }
        fn
    }
})

.ltx23_index_select_into <- function(out, table, idx) {
    fn <- .ltx23_index_select_fn()
    if (isFALSE(fn)) {
        out$copy_(torch::torch_index_select(table, 1L, idx))
    } else {
        fn(out, table, 1L, idx)
    }
    invisible(out)
}

# Persistent per-device dequantization scratch (nibbles, indices, values,
# and the level table), sized to the chunk length
.ltx23_dequant_scratch <- new.env(parent = emptyenv())

.ltx23_get_dequant_scratch <- function(n_bytes, device) {
    key <- paste(device$type, device$index %||% 0L, sep = "|")
    scratch <- .ltx23_dequant_scratch[[key]]
    if (is.null(scratch) || scratch$n_bytes < n_bytes) {
        scratch <- list(
                        n_bytes = n_bytes,
                        idx = torch::torch_empty(n_bytes, dtype = torch::torch_long(),
                device = device)
        )
        .ltx23_dequant_scratch[[key]] <- scratch
    }
    scratch
}

# Reusable dequantization buffers, keyed by shape/dtype/device: each
# distinct weight shape gets one long-lived buffer, so per-step
# dequantization allocates nothing (the buffer is overwritten in place
# by the next linear of the same shape)
.ltx23_dequant_buffers <- new.env(parent = emptyenv())

.ltx23_get_dequant_buffer <- function(shape, dtype, device) {
    key <- paste(paste(shape, collapse = "x"), dtype$.type(),
                 paste(device$type, device$index %||% 0L), sep = "|")
    buf <- .ltx23_dequant_buffers[[key]]
    if (is.null(buf)) {
        buf <- torch::torch_empty(shape, dtype = dtype, device = device)
        .ltx23_dequant_buffers[[key]] <- buf
    }
    buf
}

#' Release the NF4 dequantization buffers
#'
#' Frees the cached per-shape weight buffers (e.g. before decoding at
#' high resolution).
#'
#' @return Invisibly, NULL.
#'
#' @export
ltx23_release_dequant_buffers <- function() {
    rm(list = ls(.ltx23_dequant_buffers), envir = .ltx23_dequant_buffers)
    rm(list = ls(.ltx23_dequant_scratch), envir = .ltx23_dequant_scratch)
    .ltx23_release_attn_buffers()
    gc(verbose = FALSE)
    if (torch::cuda_is_available()) {
        tryCatch(torch::cuda_empty_cache(), error = function(e) NULL)
    }
    invisible(NULL)
}

#' NF4 linear layer
#'
#' Packed weights and per-block scales are registered as buffers, so
#' they move with the module (uint8 packs are untouched by dtype
#' conversions). The forward pass dequantizes on the weight's device.
#'
#' @param out_features,in_features Integers.
#' @param bias Logical.
#'
#' @return Module whose forward(x) returns the linear projection of
#'   \code{x}, dequantizing the NF4 weight into a reusable buffer first.
#'   Same result as an \code{nn_linear} of the same shape, at roughly an
#'   eighth of the resident weight bytes.
#'
#' @export
ltx23_nf4_linear <- torch::nn_module(
                                     "ltx23_nf4_linear",
                                     initialize = function(out_features, in_features, bias = TRUE) {
    self$out_features <- as.integer(out_features)
    self$in_features <- as.integer(in_features)
    n <- self$out_features * self$in_features
    self$weight_nf4 <- torch::nn_buffer(
                                        torch::torch_zeros(n %/% 2L, dtype = torch::torch_uint8())
    )
    self$weight_absmax <- torch::nn_buffer(
        torch::torch_ones(n %/% .ltx23_nf4_block_size,
                          dtype = torch::torch_float32())
    )
    if (bias) {
        self$bias <- torch::nn_parameter(torch::torch_zeros(out_features))
    }
},
                                     set_nf4_weight = function(packed, absmax) {
    torch::with_no_grad({
        self$weight_nf4$copy_(packed)
        self$weight_absmax$copy_(absmax)
    })
    invisible(self)
},
                                     forward = function(x) {
    w <- .ltx23_get_dequant_buffer(
                                   c(self$out_features, self$in_features), x$dtype,
                                   self$weight_nf4$device
    )
    ltx23_nf4_dequantize(
                         self$weight_nf4, self$weight_absmax,
                         c(self$out_features, self$in_features),
                         dtype = x$dtype, out = w
    )
    torch::nnf_linear(x, w, self$bias)
}
)

#' Quantize an LTX-2.3 checkpoint to NF4 shards
#'
#' Same streaming layout and cast policy as
#' \code{\link{ltx23_quantize_fp8}}, but cast-set weights are stored as
#' NF4 (\code{<key>} packed uint8 + \code{<key>_absmax} float32 blocks +
#' the original shape recovered from the model config at load time).
#' Non-cast tensors are copied through unchanged. The manifest carries
#' \code{format = "nf4"}.
#'
#' @param checkpoint_path Source .safetensors (bf16 single file).
#' @param output_dir Output directory for shards + manifest; NULL (the
#'   default) resolves under tools::R_user_dir("diffuseR", "data").
#' @param shard_bytes Numeric. Target shard size in bytes. The default
#'   1.9e9 keeps every shard under the 2^31-byte (~2.15 GB) ceiling that
#'   stock CRAN safetensors can read. Pass a larger value (e.g. 4e9) only
#'   for local builds you will read back with a fork-patched safetensors.
#' @param force Logical. Re-quantize even if a valid manifest exists.
#' @param verbose Logical.
#'
#' @return Invisibly, the manifest list.
#'
#' @export
ltx23_quantize_nf4 <- function(checkpoint_path, output_dir = NULL,
                               shard_bytes = 1.9e9, force = FALSE,
                               verbose = TRUE) {
    if (is.null(output_dir)) {
        output_dir <- file.path(tools::R_user_dir("diffuseR", "data"),
                                "ltx2.3-nf4")
    }
    manifest_path <- file.path(output_dir, "manifest.json")
    if (!force && file.exists(manifest_path)) {
        manifest <- jsonlite::fromJSON(manifest_path)
        if (all(file.exists(file.path(output_dir, manifest$shards)))) {
            if (verbose) {
                message("NF4 artifact already present: ", output_dir)
            }
            return(invisible(manifest))
        }
    }
    dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

    ckpt <- ltx23_open_checkpoint(checkpoint_path)

    shard <- list()
    shard_size <- 0
    shard_files <- character(0)
    n_cast <- 0L

    flush_shard <- function() {
        if (!length(shard)) {
            return()
        }
        fname <- sprintf("ltx2.3-nf4-%05d.safetensors",
                         length(shard_files) + 1L)
        safetensors::safe_save_file(shard, file.path(output_dir, fname))
        shard_files[[length(shard_files) + 1L]] <<- fname
        if (verbose) {
            message(sprintf("  wrote %s (%.2f GB, %d tensors)", fname,
                            shard_size / 1e9, length(shard)))
        }
        shard <<- list()
        shard_size <<- 0
        gc(verbose = FALSE)
    }

    keys <- ckpt$keys
    for (i in seq_along(keys)) {
        key <- keys[[i]]
        tensor <- ckpt$handle$get_tensor(key)

        mapped <- ltx23_map_dit_key(key)
        if (startsWith(key, "model.diffusion_model.") &&
            ltx23_is_fp8_cast_key(mapped)) {
            torch::with_no_grad({
                q <- ltx23_nf4_quantize(tensor)
                shard[[key]] <- q$packed
                shard[[paste0(key, "_absmax")]] <- q$absmax
            })
            shard_size <- shard_size + prod(tensor$shape) * 0.5625
            n_cast <- n_cast + 1L
        } else {
            shard[[key]] <- tensor
            shard_size <- shard_size + prod(tensor$shape) * 2
        }
        rm(tensor)

        if (shard_size >= shard_bytes) {
            flush_shard()
        }
        if (i %% 200L == 0L) {
            gc(verbose = FALSE)
            if (verbose) {
                message(sprintf("  quantizing %d/%d tensors", i, length(keys)))
            }
        }
    }
    flush_shard()

    manifest <- list(
                     source = basename(checkpoint_path),
                     model_version = ckpt$version,
                     format = "nf4",
                     shards = shard_files,
                     tensors = length(keys),
                     nf4_cast = n_cast,
                     config = ckpt$config
    )
    jsonlite::write_json(manifest, manifest_path, auto_unbox = TRUE, pretty = TRUE)
    if (verbose) {
        message(sprintf("Quantized %d/%d tensors to nf4 across %d shards: %s",
                        n_cast, length(keys), length(shard_files), output_dir))
    }
    invisible(manifest)
}

#' Load the LTX-2.3 transformer with resident NF4 weights
#'
#' Builds the transformer, swaps the cast-set linears for
#' \code{\link{ltx23_nf4_linear}}, and loads everything onto
#' \code{device}: at ~4.5 bits/parameter the whole 22B transformer stays
#' GPU-resident, avoiding per-step weight transfers.
#'
#' @param ckpt An NF4 \code{ltx23_checkpoint}
#'   (\code{\link{ltx23_open_fp8_checkpoint}} reads any shard artifact).
#' @param device Character.
#' @param verbose Logical.
#' @param ... Passed to \code{\link{ltx23_transformer}} (tiny test configs).
#'
#' @return The loaded \code{ltx23_transformer}.
#'
#' @export
ltx23_load_transformer_nf4 <- function(ckpt, device = "cuda", verbose = TRUE,
                                       ...) {
    stopifnot(inherits(ckpt, "ltx23_checkpoint"))
    model <- .construct_skeleton(ltx23_transformer, ...)

    groups <- ltx23_split_keys(ckpt$keys)
    dit_keys <- groups$dit
    absmax_keys <- dit_keys[endsWith(dit_keys, ".weight_absmax")]
    main_keys <- setdiff(dit_keys, absmax_keys)

    dests <- c(model$named_parameters(), model$named_buffers())
    filled <- character(0)
    unmapped <- character(0)

    torch::with_no_grad({
        for (i in seq_along(main_keys)) {
            key <- main_keys[[i]]
            mapped <- ltx23_map_dit_key(key)

            if (ltx23_is_fp8_cast_key(mapped) &&
                        paste0(key, "_absmax") %in% absmax_keys) {
                segments <- strsplit(mapped, ".", fixed = TRUE)[[1]]
                parent <- .ltx23_walk_module(model, utils::head(segments, -2L))
                leaf <- segments[length(segments) - 1L]
                old <- .ltx23_walk_module(parent, leaf)
                if (is.null(old)) {
                    unmapped <- c(unmapped, key)
                    next
                }
                w_shape <- old$weight$shape
                nf4_mod <- ltx23_nf4_linear(w_shape[1], w_shape[2],
                    bias = !is.null(old$bias))
                if (!is.null(old$bias)) {
                    # Adopt the original bias parameter; its checkpoint key
                    # loads through the pre-swap destination map
                    nf4_mod$bias <- old$bias
                }
                nf4_mod$set_nf4_weight(
                                       ckpt$handle$get_tensor(key),
                                       ckpt$handle$get_tensor(paste0(key, "_absmax"))
                )
                do.call(`$<-`, list(parent, leaf, nf4_mod))
                filled <- c(filled, mapped)
            } else {
                dest <- dests[[mapped]]
                if (is.null(dest)) {
                    unmapped <- c(unmapped, key)
                    next
                }
                dest$copy_(ckpt$handle$get_tensor(key))
                filled <- c(filled, mapped)
            }

            if (i %% 100L == 0L) {
                gc(verbose = FALSE)
                if (verbose && i %% 500L == 0L) {
                    message(sprintf("  loaded %d/%d transformer tensors", i,
                                    length(main_keys)))
                }
            }
        }
    })
    gc(verbose = FALSE)

    if (length(unmapped)) {
        stop("NF4 transformer load: ", length(unmapped), " unmapped keys, e.g. ",
             paste(utils::head(unmapped, 3), collapse = ", "))
    }
    expected_missing <- grepl(.ltx23_fp8_cast_pattern, names(dests))
    unfilled <- setdiff(names(dests)[!expected_missing], filled)
    if (length(unfilled)) {
        stop("NF4 transformer load: ", length(unfilled), " unfilled params, e.g. ",
             paste(utils::head(unfilled, 3), collapse = ", "))
    }

    # Everything (packed weights included, as buffers) onto the GPU
    model$to(device = device)
    model$eval()
    # Block intermediates (norms, projections, FF activations) are still
    # ~1.5GB per block at high resolution; per-block gc keeps them bounded
    options(diffuseR.block_gc = TRUE)
    if (verbose) {
        message("Transformer ready: NF4 weights resident on ", device)
    }
    model
}

# NF4 quantization: round trip quality, linear parity, and resident
# transformer loading from a tiny official-named checkpoint.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

torch::torch_manual_seed(321)

# --- Quantize/dequantize round trip ------------------------------------------

w <- torch::torch_randn(64L, 128L) * 0.02
q <- ltx23_nf4_quantize(w)
expect_equal(as.integer(q$packed$shape), 64L * 128L / 2L)
expect_equal(as.integer(q$absmax$shape), 64L * 128L / 64L)
expect_equal(q$packed$dtype$.type(), "Byte")

w_rt <- ltx23_nf4_dequantize(q$packed, q$absmax, c(64L, 128L),
  dtype = torch::torch_float32())
expect_equal(as.integer(w_rt$shape), c(64L, 128L))
# NF4 round-trip error on Gaussian data: a few percent relative
rel <- as.numeric((w_rt - w)$abs()$mean() / w$abs()$mean())
expect_true(rel < 0.1)

# Values land exactly on table levels x absmax
blocks <- w_rt$reshape(c(-1L, 64L)) / q$absmax$unsqueeze(2L)
dists <- torch::torch_min(
  (blocks$flatten()$unsqueeze(2L) -
    torch::torch_tensor(diffuseR:::.ltx23_nf4_table)$unsqueeze(1L))$abs(),
  dim = 2L
)[[1]]
expect_true(as.numeric(dists$max()) < 1e-5)

# Chunked dequant is identical to single-shot
w_chunked <- ltx23_nf4_dequantize(q$packed, q$absmax, c(64L, 128L),
  dtype = torch::torch_float32(), chunk_elements = 512L)
expect_true(as.numeric((w_chunked - w_rt)$abs()$max()) == 0)

# --- nf4_linear parity -----------------------------------------------------------

lin <- torch::nn_linear(128L, 64L)
qw <- ltx23_nf4_quantize(lin$weight)
nf4 <- ltx23_nf4_linear(64L, 128L, bias = TRUE)
nf4$set_nf4_weight(qw$packed, qw$absmax)
torch::with_no_grad(nf4$bias$copy_(lin$bias))

x <- torch::torch_randn(5L, 128L)
torch::with_no_grad({
  y_ref <- lin(x)
  y_nf4 <- nf4(x)
})
rel_y <- as.numeric((y_nf4 - y_ref)$abs()$mean() / y_ref$abs()$mean())
expect_true(rel_y < 0.15)

# Buffers move with the module and survive dtype casts
nf4$to(dtype = torch::torch_float64())
expect_equal(nf4$weight_nf4$dtype$.type(), "Byte")

# --- Tiny transformer through the NF4 artifact ------------------------------------

tiny_cfg <- list(
  in_channels = 4L, out_channels = 4L,
  num_attention_heads = 2L, attention_head_dim = 8L,
  cross_attention_dim = 16L,
  audio_in_channels = 4L, audio_out_channels = 4L,
  audio_num_attention_heads = 2L, audio_attention_head_dim = 4L,
  audio_cross_attention_dim = 8L,
  num_layers = 1L
)
ref_model <- do.call(ltx23_transformer, tiny_cfg)
ref_model$eval()

to_official <- function(name) {
  name <- gsub("av_cross_attn_video_scale_shift", "av_ca_video_scale_shift_adaln_single", name, fixed = TRUE)
  name <- gsub("av_cross_attn_video_a2v_gate", "av_ca_a2v_gate_adaln_single", name, fixed = TRUE)
  name <- gsub("av_cross_attn_audio_scale_shift", "av_ca_audio_scale_shift_adaln_single", name, fixed = TRUE)
  name <- gsub("av_cross_attn_audio_v2a_gate", "av_ca_v2a_gate_adaln_single", name, fixed = TRUE)
  name <- gsub("video_a2v_cross_attn_scale_shift_table", "scale_shift_table_a2v_ca_video", name, fixed = TRUE)
  name <- gsub("audio_a2v_cross_attn_scale_shift_table", "scale_shift_table_a2v_ca_audio", name, fixed = TRUE)
  name <- gsub("prompt_adaln", "prompt_adaln_single", name, fixed = TRUE)
  name <- sub("^audio_time_embed\\.", "audio_adaln_single.", name)
  name <- sub("^time_embed\\.", "adaln_single.", name)
  name <- gsub("proj_in", "patchify_proj", name, fixed = TRUE)
  name <- gsub("norm_q", "q_norm", name, fixed = TRUE)
  name <- gsub("norm_k", "k_norm", name, fixed = TRUE)
  paste0("model.diffusion_model.", name)
}

params <- ref_model$named_parameters()
tensors <- list()
for (name in names(params)) {
  tensors[[to_official(name)]] <- params[[name]]$detach()
}
src <- tempfile(fileext = ".safetensors")
on.exit(unlink(src), add = TRUE)
safetensors::safe_save_file(tensors, src, metadata = list(model_version = "2.3.0"))

nf4_dir <- tempfile("nf4_")
on.exit(unlink(nf4_dir, recursive = TRUE), add = TRUE)
manifest <- ltx23_quantize_nf4(src, nf4_dir, verbose = FALSE)
expect_equal(manifest$format, "nf4")
expect_equal(manifest$nf4_cast, sum(ltx23_is_fp8_cast_key(names(params))))

ckpt <- ltx23_open_fp8_checkpoint(nf4_dir)
expect_equal(ckpt$format, "nf4")

nf4_model <- do.call(ltx23_load_transformer_nf4, c(
  list(ckpt = ckpt, device = "cpu", verbose = FALSE),
  tiny_cfg
))
expect_inherits(nf4_model$transformer_blocks[[1]]$attn1$to_q, "ltx23_nf4_linear")

# Forward runs in bf16 and stays finite
torch::torch_manual_seed(5)
bf <- torch::torch_bfloat16()
torch::with_no_grad({
  out <- nf4_model(
    hidden_states = torch::torch_randn(1L, 24L, 4L, dtype = bf),
    audio_hidden_states = torch::torch_randn(1L, 5L, 4L, dtype = bf),
    encoder_hidden_states = torch::torch_randn(1L, 7L, 16L, dtype = bf),
    audio_encoder_hidden_states = torch::torch_randn(1L, 7L, 8L, dtype = bf),
    timestep = torch::torch_tensor(700, dtype = torch::torch_float32()),
    sigma = torch::torch_tensor(700, dtype = torch::torch_float32()),
    num_frames = 2L, height = 3L, width = 4L, audio_num_frames = 5L,
    use_cross_timestep = TRUE
  )
})
expect_true(as.logical(out$sample$isfinite()$all()$item()))
expect_true(as.logical(out$audio_sample$isfinite()$all()$item()))

# treesitR coverage report: diffusers LTX-2 reference vs the diffuseR port.
#
# Parses the Python reference files and the R port with tree-sitter,
# extracts class/function inventories, and prints a coverage table using
# a curated mapping (which also documents deliberate omissions).
#
# Usage: r tools/ltx23_compare.R

library(treesitR)

`%||%` <- function(x, y) if (is.null(x)) y else x

py_entities <- function(path) {
  src <- paste(readLines(path, warn = FALSE), collapse = "\n")
  parser <- ts_parser_new()
  ts_parser_set_language(parser, ts_language_python())
  tree <- ts_parse(parser, src)
  root <- ts_tree_root_node(tree)

  out <- character(0)
  # Top-level classes and functions only (methods are per-class detail)
  for (i in seq_len(ts_node_named_child_count(root))) {
    node <- ts_node_named_child(root, i - 1L)
    if (ts_node_type(node) == "decorated_definition") {
      node <- ts_node_child_by_field(node, "definition")
    }
    type <- ts_node_type(node)
    if (type %in% c("class_definition", "function_definition")) {
      name <- ts_node_text(ts_node_child_by_field(node, "name"))
      prefix <- if (type == "class_definition") "class" else "def"
      out[[length(out) + 1L]] <- paste(prefix, name)
    }
  }
  unique(out)
}

r_entities <- function(path) {
  src <- paste(readLines(path, warn = FALSE), collapse = "\n")
  parser <- ts_parser_new()
  ts_parser_set_language(parser, ts_language_r())
  tree <- ts_parse(parser, src)
  root <- ts_tree_root_node(tree)

  out <- character(0)
  for (i in seq_len(ts_node_named_child_count(root))) {
    node <- ts_node_named_child(root, i - 1L)
    if (ts_node_type(node) == "binary_operator") {
      lhs <- ts_node_child_by_field(node, "lhs")
      if (!is.null(lhs) && ts_node_type(lhs) == "identifier") {
        out[[length(out) + 1L]] <- ts_node_text(lhs)
      }
    }
  }
  unique(out)
}

ref_dir <- "ref/upstream/diffusers/src/diffusers"
py_files <- c(
  transformer = file.path(ref_dir, "models/transformers/transformer_ltx2.py"),
  video_vae = file.path(ref_dir, "models/autoencoders/autoencoder_kl_ltx2.py"),
  audio_vae = file.path(ref_dir, "models/autoencoders/autoencoder_kl_ltx2_audio.py"),
  connectors = file.path(ref_dir, "pipelines/ltx2/connectors.py"),
  vocoder = file.path(ref_dir, "pipelines/ltx2/vocoder.py"),
  upsampler = file.path(ref_dir, "pipelines/ltx2/latent_upsampler.py")
)

# Curated mapping: reference entity -> R counterpart or a documented skip
MAPPING <- c(
  # transformer_ltx2.py
  "def apply_interleaved_rotary_emb" = "ltx23_apply_interleaved_rotary_emb",
  "def apply_split_rotary_emb" = "ltx23_apply_split_rotary_emb",
  "class LTX2AdaLayerNormSingle" = "ltx23_ada_layer_norm_single",
  "class LTX2AudioVideoAttnProcessor" = "ltx23_attention (inlined)",
  "class LTX2PerturbedAttnProcessor" = "ltx23_attention (inlined)",
  "class LTX2Attention" = "ltx23_attention",
  "class LTX2VideoTransformerBlock" = "ltx23_transformer_block",
  "class LTX2AudioVideoRotaryPosEmbed" = "ltx23_rotary_pos_embed",
  "class LTX2VideoTransformer3DModel" = "ltx23_transformer",
  "class AudioVisualModelOutput" = "SKIP: plain list return",
  # autoencoder_kl_ltx2.py
  "class PerChannelRMSNorm" = "ltx23_per_channel_rms_norm",
  "class LTX2VideoCausalConv3d" = "ltx23_causal_conv3d",
  "class LTX2VideoResnetBlock3d" = "ltx23_video_resnet_block3d",
  "class LTX2VideoDownsampler3d" = "ltx23_video_downsampler3d",
  "class LTX2VideoUpsampler3d" = "ltx23_video_upsampler3d",
  "class LTX2VideoDownBlock3D" = "ltx23_video_down_block3d",
  "class LTX2VideoMidBlock3d" = "ltx23_video_mid_block3d",
  "class LTX2VideoUpBlock3d" = "ltx23_video_up_block3d",
  "class LTX2VideoEncoder3d" = "ltx23_video_encoder3d",
  "class LTX2VideoDecoder3d" = "ltx23_video_decoder3d",
  "class AutoencoderKLLTX2Video" = "ltx23_video_vae",
  # autoencoder_kl_ltx2_audio.py
  "class LTX2AudioCausalConv2d" = "ltx23_audio_causal_conv2d",
  "class LTX2AudioPixelNorm" = "ltx23_per_channel_rms_norm (shared)",
  "class LTX2AudioAttnBlock" = "SKIP: mid_block_add_attention FALSE in 2.3",
  "class LTX2AudioResnetBlock" = "ltx23_audio_resnet_block",
  "class LTX2AudioDownsample" = "SKIP: encoder-only (t2v never encodes audio)",
  "class LTX2AudioUpsample" = "ltx23_audio_upsample",
  "class LTX2AudioAudioPatchifier" = "SKIP: pipeline packs latents directly",
  "class LTX2AudioEncoder" = "SKIP: encoder-only (t2v never encodes audio)",
  "class LTX2AudioDecoder" = "ltx23_audio_decoder",
  "class AutoencoderKLLTX2Audio" = "ltx23_audio_vae",
  # connectors.py
  "def per_layer_masked_mean_norm" = "SKIP: LTX-2.0 path (per_modality_projections)",
  "def per_token_rms_norm" = "ltx23_per_token_rms_norm",
  "class LTX2RotaryPosEmbed1d" = "ltx23_rotary_pos_embed_1d",
  "class LTX2TransformerBlock1d" = "ltx23_transformer_block_1d",
  "class LTX2ConnectorTransformer1d" = "ltx23_connector_transformer_1d",
  "class LTX2TextConnectors" = "ltx23_text_connectors",
  # vocoder.py
  "def kaiser_sinc_filter1d" = "ltx23_kaiser_sinc_filter1d",
  "class DownSample1d" = "ltx23_downsample1d",
  "class UpSample1d" = "ltx23_upsample1d",
  "class AntiAliasAct1d" = "ltx23_antialias_act1d",
  "class SnakeBeta" = "ltx23_snake_beta",
  "class ResBlock" = "ltx23_vocoder_resblock / ltx23_upsampler_res_block",
  "class LTX2Vocoder" = "ltx23_vocoder",
  "class CausalSTFT" = "ltx23_causal_stft",
  "class MelSTFT" = "ltx23_mel_stft",
  "class LTX2VocoderWithBWE" = "ltx23_vocoder_with_bwe",
  # latent_upsampler.py
  "class PixelShuffleND" = "nnf_pixel_shuffle (2D case only in 2.3)",
  "class BlurDownsample" = "SKIP: rational resampler off in 2.3",
  "class SpatialRationalResampler" = "SKIP: rational resampler off in 2.3",
  "class LTX2LatentUpsamplerModel" = "ltx23_latent_upsampler"
)

r_files <- list.files("R", pattern = "ltx23|txt2vid_ltx23", full.names = TRUE)
r_defined <- unlist(lapply(r_files, r_entities))

missing <- character(0)
cat(sprintf("%-45s %-50s %s\n", "REFERENCE", "R PORT", "STATUS"))
cat(strrep("-", 110), "\n")
for (section in names(py_files)) {
  cat("##", section, "\n")
  for (entity in py_entities(py_files[[section]])) {
    mapped <- if (entity %in% names(MAPPING)) MAPPING[[entity]] else NULL
    status <- if (is.null(mapped)) {
      missing <- c(missing, paste0(section, ": ", entity))
      "MISSING FROM MAPPING"
    } else if (startsWith(mapped, "SKIP")) {
      "skipped (documented)"
    } else {
      r_name <- strsplit(mapped, " ")[[1]][1]
      if (r_name %in% r_defined || grepl("nnf_", r_name)) "ported" else "MAPPED BUT NOT DEFINED"
    }
    cat(sprintf("%-45s %-50s %s\n", entity, mapped %||% "-", status))
  }
}

if (length(missing)) {
  cat("\nUNMAPPED reference entities:\n")
  for (m in missing) cat("  -", m, "\n")
  quit(status = 1)
}
cat("\nAll reference entities accounted for.\n")

#' Native UNet for Stable Diffusion
#'
#' Native R torch implementation of UNet2DConditionModel.
#' Replaces TorchScript for better GPU compatibility.
#'
#' @param in_channels Input channels (default 4 for latent space)
#' @param out_channels Output channels (default 4)
#' @param block_out_channels Channel multipliers per block
#' @param layers_per_block Number of ResBlocks per down/up block
#' @param cross_attention_dim Context dimension from text encoder
#' @param attention_head_dim Dimension per attention head
#'
#' @return An nn_module representing the UNet
#' @export
unet_native <- torch::nn_module(
                                "UNetNative",

                                initialize = function(
        in_channels = 4L,
        out_channels = 4L,
        block_out_channels = c(320L, 640L, 1280L, 1280L),
        layers_per_block = 2L,
        cross_attention_dim = 1024L,
        attention_head_dim = 64L
    ) {
    self$in_channels <- in_channels
    self$out_channels <- out_channels
    self$block_out_channels <- block_out_channels
    self$layers_per_block <- layers_per_block

    # Time embedding dimension
    time_embed_dim <- block_out_channels[1] * 4L # 320 * 4 = 1280

    # Input convolution
    self$conv_in <- torch::nn_conv2d(in_channels, block_out_channels[1], 3L,
                                     padding = 1L)

    # Time embedding MLP
    self$time_embedding_linear_1 <- torch::nn_linear(block_out_channels[1], time_embed_dim)
    self$time_embedding_linear_2 <- torch::nn_linear(time_embed_dim, time_embed_dim)

    # Down blocks
    # SD21: blocks 0-2 have attention, block 3 does not
    self$down_blocks <- torch::nn_module_list()
    output_channel <- block_out_channels[1]

    for (i in seq_along(block_out_channels)) {
        input_channel <- output_channel
        output_channel <- block_out_channels[i]
        is_final_block <- (i == length(block_out_channels))
        # Last block has no attention
        has_attention <- !is_final_block

        down_block <- self$create_down_block(
            input_channel, output_channel, time_embed_dim,
            cross_attention_dim, attention_head_dim,
            layers_per_block,
            add_downsample = !is_final_block,
            add_attention = has_attention
        )
        self$down_blocks$append(down_block)
    }

    # Mid block
    mid_channels <- block_out_channels[length(block_out_channels)]
    self$mid_block <- self$create_mid_block(
        mid_channels, time_embed_dim, cross_attention_dim, attention_head_dim
    )

    # Up blocks (reverse order)
    # SD21: first up_block (index 0) has no attention, rest do
    # Skip channels vary per resnet - we store them for forward pass
    self$up_blocks <- torch::nn_module_list()
    reversed_channels <- rev(block_out_channels)

    # Calculate skip channels for each resnet
    # The skip stack order (bottom to top):
    # down0: [in, r0, r1, ds], down1: [r0, r1, ds], down2: [r0, r1, ds], down3: [r0, r1]
    # This gives us: [320, 320, 320, 320, 640, 640, 640, 1280, 1280, 1280, 1280, 1280]
    skip_channels <- c(
                       block_out_channels[1], # conv_in
                       rep(block_out_channels[1], layers_per_block), # down0 resnets
                       block_out_channels[1], # down0 downsample
                       rep(block_out_channels[2], layers_per_block), # down1 resnets
                       block_out_channels[2], # down1 downsample
                       rep(block_out_channels[3], layers_per_block), # down2 resnets
                       block_out_channels[3], # down2 downsample
                       rep(block_out_channels[4], layers_per_block) # down3 resnets (no downsample)
    )
    self$skip_channels <- rev(skip_channels) # Reverse for popping order

    for (i in seq_along(reversed_channels)) {
        output_channel <- reversed_channels[i]
        if (i == 1) {
            prev_output <- mid_channels
        } else {
            prev_output <- reversed_channels[i - 1]
        }

        is_final_block <- (i == length(reversed_channels))
        is_first_block <- (i == 1)
        has_attention <- !is_first_block

        # Calculate per-resnet skip channels for this up_block
        num_resnets <- layers_per_block + 1L
        resnet_skip_channels <- integer(num_resnets)
        for (j in seq_len(num_resnets)) {
            skip_idx <- (i - 1) * num_resnets + j
            resnet_skip_channels[j] <- self$skip_channels[skip_idx]
        }

        up_block <- self$create_up_block_v2(
            prev_output, output_channel, resnet_skip_channels, time_embed_dim,
            cross_attention_dim, attention_head_dim,
            add_upsample = !is_final_block,
            add_attention = has_attention
        )
        self$up_blocks$append(up_block)
    }

    # Output
    self$conv_norm_out <- group_norm_32(block_out_channels[1])
    self$conv_out <- torch::nn_conv2d(block_out_channels[1], out_channels, 3L, padding = 1L)
},

                                create_down_block = function(
        in_channels,
        out_channels,
        time_embed_dim,
        cross_attention_dim,
        attention_head_dim,
        num_layers,
        add_downsample = TRUE,
        add_attention = TRUE
    ) {
    # Create block as a simple nn_module
    DownBlock <- torch::nn_module(
                                  "DownBlock",
                                  initialize = function() {
        # ResNets
        self$resnets <- torch::nn_module_list()
        for (i in seq_len(num_layers)) {
            if (i == 1) {
                in_ch <- in_channels
            } else {
                in_ch <- out_channels
            }
            self$resnets$append(UNetResBlock(in_ch, out_channels, time_embed_dim))
        }

        # Attention blocks (optional)
        self$has_attentions <- add_attention
        if (add_attention) {
            self$attentions <- torch::nn_module_list()
            n_heads <- out_channels %/% attention_head_dim
            for (i in seq_len(num_layers)) {
                self$attentions$append(
                                       SpatialTransformer(out_channels, n_heads, attention_head_dim,
                        depth = 1L, context_dim = cross_attention_dim)
                )
            }
        }

        # Downsample
        self$has_downsamplers <- add_downsample
        if (add_downsample) {
            self$downsamplers <- torch::nn_module_list()
            self$downsamplers$append(Downsample2D(out_channels))
        }
    }
    )
    DownBlock()
},

                                create_mid_block = function(
        channels,
        time_embed_dim,
        cross_attention_dim,
        attention_head_dim
    ) {
    MidBlock <- torch::nn_module(
                                 "MidBlock",
                                 initialize = function() {
        # Two resnets with attention in between
        self$resnets <- torch::nn_module_list()
        self$resnets$append(UNetResBlock(channels, channels, time_embed_dim))
        self$resnets$append(UNetResBlock(channels, channels, time_embed_dim))

        # Attention
        n_heads <- channels %/% attention_head_dim
        self$attentions <- torch::nn_module_list()
        self$attentions$append(
                               SpatialTransformer(channels, n_heads, attention_head_dim,
                depth = 1L, context_dim = cross_attention_dim)
        )
    }
    )
    MidBlock()
},

                                create_up_block_v2 = function(
        in_channels,
        out_channels,
        resnet_skip_channels,
        time_embed_dim,
        cross_attention_dim,
        attention_head_dim,
        add_upsample = TRUE,
        add_attention = TRUE
    ) {
    # resnet_skip_channels is a vector with one entry per resnet
    num_resnets <- length(resnet_skip_channels)

    UpBlock <- torch::nn_module(
                                "UpBlock",
                                initialize = function() {
        # ResNets (handle skip connections with variable channels)
        self$resnets <- torch::nn_module_list()
        for (i in seq_len(num_resnets)) {
            # First resnet takes prev_output, rest take out_channels
            if (i == 1) {
                in_ch <- in_channels + resnet_skip_channels[i]
            } else {
                in_ch <- out_channels + resnet_skip_channels[i]
            }
            self$resnets$append(UNetResBlock(in_ch, out_channels, time_embed_dim))
        }

        # Attention blocks (optional)
        self$has_attentions <- add_attention
        if (add_attention) {
            self$attentions <- torch::nn_module_list()
            n_heads <- out_channels %/% attention_head_dim
            for (i in seq_len(num_resnets)) {
                self$attentions$append(
                                       SpatialTransformer(out_channels, n_heads, attention_head_dim,
                        depth = 1L, context_dim = cross_attention_dim)
                )
            }
        }

        # Upsample
        self$has_upsamplers <- add_upsample
        if (add_upsample) {
            self$upsamplers <- torch::nn_module_list()
            self$upsamplers$append(Upsample2D(out_channels))
        }
    }
    )
    UpBlock()
},

                                forward = function(
        sample,
        timestep,
        encoder_hidden_states
    ) {
    # Get model dtype from weights
    model_dtype <- self$time_embedding_linear_1$weight$dtype

    # Time embedding (computed in float32, then cast to model dtype).
    # SD 2.1 uses the standard diffusers timestep embedding
    # (flip_sin_to_cos=TRUE, downscale_freq_shift=0), same as SDXL. The
    # old FALSE/1 scrambled the sin/cos ordering the trained
    # time_embedding weights expect - masked by constant inputs (so the
    # parity test missed it) but compounding through the spatial path
    # into tiled output. Verified cos 1.0 vs the TorchScript reference.
    t_emb <- timestep_embedding(timestep, self$block_out_channels[1],
                                flip_sin_to_cos = TRUE, downscale_freq_shift = 0L)
    t_emb <- t_emb$to(dtype = model_dtype)
    t_emb <- self$time_embedding_linear_1(t_emb)
    t_emb <- torch::nnf_silu(t_emb)
    t_emb <- self$time_embedding_linear_2(t_emb)

    # Input conv
    sample <- self$conv_in(sample)

    # Store skip connections
    down_block_res_samples <- list(sample)

    # Down blocks
    for (i in seq_along(self$down_blocks)) {
        block <- self$down_blocks[[i]]

        for (j in seq_along(block$resnets)) {
            sample <- block$resnets[[j]](sample, t_emb)
            if (block$has_attentions) {
                sample <- block$attentions[[j]](sample, encoder_hidden_states)
            }
            down_block_res_samples <- c(down_block_res_samples, list(sample))
        }

        if (block$has_downsamplers) {
            sample <- block$downsamplers[[1]](sample)
            down_block_res_samples <- c(down_block_res_samples, list(sample))
        }
    }

    # Mid block
    sample <- self$mid_block$resnets[[1]](sample, t_emb)
    sample <- self$mid_block$attentions[[1]](sample, encoder_hidden_states)
    sample <- self$mid_block$resnets[[2]](sample, t_emb)

    # Up blocks
    for (i in seq_along(self$up_blocks)) {
        block <- self$up_blocks[[i]]

        for (j in seq_along(block$resnets)) {
            # Pop skip connection
            res_sample <- down_block_res_samples[[length(down_block_res_samples)]]
            down_block_res_samples <- down_block_res_samples[-length(down_block_res_samples)]

            # Concatenate
            sample <- torch::torch_cat(list(sample, res_sample), dim = 2L)

            sample <- block$resnets[[j]](sample, t_emb)
            if (block$has_attentions) {
                sample <- block$attentions[[j]](sample, encoder_hidden_states)
            }
        }

        if (block$has_upsamplers) {
            sample <- block$upsamplers[[1]](sample)
        }
    }

    # Output
    sample <- self$conv_norm_out(sample)
    sample <- torch::nnf_silu(sample)
    sample <- self$conv_out(sample)

    sample
}
)

#' Detect UNet architecture from TorchScript file
#'
#' @param torchscript_path Path to TorchScript UNet .pt file
#' @return List with architecture parameters
#' @keywords internal
detect_unet_architecture <- function(torchscript_path) {
    ts_unet <- torch::jit_load(torchscript_path)
    ts_params <- ts_unet$parameters
    param_names <- names(ts_params)

    # Get block_out_channels from down_blocks resnets
    # Use norm2 (not norm1) to get output channels
    block_channels <- integer(0)
    for (i in 0:3) {
        key <- sprintf("unet.down_blocks.%d.resnets.0.norm2.weight", i)
        if (key %in% param_names) {
            block_channels <- c(block_channels,
                                as.integer(ts_params[[key]]$shape[1]))
        }
    }

    # Get cross_attention_dim from attn2.to_k
    attn_key <- "unet.down_blocks.0.attentions.0.transformer_blocks.0.attn2.to_k.weight"
    cross_attention_dim <- as.integer(ts_params[[attn_key]]$shape[2])

    # attention_head_dim is a fixed architectural choice for SD models
    # SD21 and SDXL use 64
    attention_head_dim <- 64L

    list(
         in_channels = 4L,
         out_channels = 4L,
         block_out_channels = block_channels,
         layers_per_block = 2L,
         cross_attention_dim = cross_attention_dim,
         attention_head_dim = attention_head_dim
    )
}

#' Create native UNet from TorchScript
#'
#' Detects architecture and loads weights from a TorchScript UNet file.
#'
#' @param torchscript_path Path to TorchScript UNet .pt file
#' @param verbose Print loading progress
#'
#' @return A native UNet module with loaded weights
#' @export
unet_native_from_torchscript <- function(torchscript_path, verbose = TRUE) {
    # Detect architecture
    arch <- detect_unet_architecture(torchscript_path)

    if (verbose) {
        message("Detected UNet architecture:")
        message("  block_out_channels: ",
                paste(arch$block_out_channels, collapse = ", "))
        message("  cross_attention_dim: ", arch$cross_attention_dim)
        message("  attention_head_dim: ", arch$attention_head_dim)
    }

    # Create native UNet
    unet <- unet_native(
                        in_channels = arch$in_channels,
                        out_channels = arch$out_channels,
                        block_out_channels = arch$block_out_channels,
                        layers_per_block = arch$layers_per_block,
                        cross_attention_dim = arch$cross_attention_dim,
                        attention_head_dim = arch$attention_head_dim
    )

    # Load weights
    load_unet_weights(unet, torchscript_path, verbose = verbose)

    unet
}

#' Load weights from TorchScript UNet into native UNet
#'
#' @param native_unet Native UNet module
#' @param torchscript_path Path to TorchScript UNet .pt file
#' @param verbose Print loading progress
#'
#' @return The native UNet with loaded weights (invisibly)
#' @export
load_unet_weights <- function(native_unet, torchscript_path, verbose = TRUE) {
    ts_unet <- torch::jit_load(torchscript_path)
    ts_params <- ts_unet$parameters

    # Build weight mapping
    remap_key <- function(key) {
        # Strip unet. prefix
        key <- sub("^unet\\.", "", key)

        # Time embedding: time_embedding.linear_1 -> time_embedding_linear_1
        key <- sub("^time_embedding\\.linear_1", "time_embedding_linear_1", key)
        key <- sub("^time_embedding\\.linear_2", "time_embedding_linear_2", key)

        # Note: nn_sequential children use 0-indexing (matching Python/TorchScript)
        # so no conversion needed for to_out.0, net.0, net.2

        key
    }

    loaded <- 0
    skipped <- 0
    unmapped <- character(0)

    torch::with_no_grad({
        for (ts_name in names(ts_params)) {
            native_name <- remap_key(ts_name)

            if (native_name %in% names(native_unet$parameters)) {
                ts_tensor <- ts_params[[ts_name]]
                native_tensor <- native_unet$parameters[[native_name]]

                if (all(ts_tensor$shape == native_tensor$shape)) {
                    native_tensor$copy_(ts_tensor)
                    loaded <- loaded + 1
                } else if (verbose) {
                    message("Shape mismatch: ", native_name,
                            " (", paste(as.integer(ts_tensor$shape),
                                        collapse = "x"), " vs ",
                            paste(as.integer(native_tensor$shape), collapse = "x"), ")")
                    skipped <- skipped + 1
                }
            } else {
                skipped <- skipped + 1
                unmapped <- c(unmapped, paste0(ts_name, " -> ", native_name))
            }
        }
    })

    if (verbose) {
        if (length(unmapped) > 0 && length(unmapped) <= 10) {
            message("Unmapped parameters:")
            for (u in unmapped) {
                message("  ", u)
            }
        } else if (length(unmapped) > 10) {
            message("Unmapped: ", length(unmapped), " parameters (showing first 10)")
            for (u in head(unmapped, 10)) {
                message("  ", u)
            }
        }
        message("Loaded ", loaded, "/", loaded + skipped, " parameters")
    }

    invisible(native_unet)
}

#' Native SDXL UNet
#'
#' Native R torch implementation of SDXL UNet2DConditionModel.
#' SDXL has a different architecture from SD21:
#' - 3 down/up blocks (not 4)
#' - Variable transformer depth per block
#' - Additional conditioning via add_embedding
#'
#' @param in_channels Input channels (default 4 for latent space)
#' @param out_channels Output channels (default 4)
#' @param block_out_channels Channel multipliers per block
#' @param layers_per_block Number of ResBlocks per down/up block
#' @param transformer_layers_per_block Transformer depth per block
#' @param cross_attention_dim Context dimension from text encoder
#' @param attention_head_dim Dimension per attention head
#' @param addition_embed_dim Dimension for additional embeddings
#' @param addition_time_embed_dim Dimension for time embedding projection
#'
#' @return An nn_module representing the SDXL UNet
#' @export
unet_sdxl_native <- torch::nn_module(
                                     "UNetSDXLNative",

                                     initialize = function(
        in_channels = 4L,
        out_channels = 4L,
        block_out_channels = c(320L, 640L, 1280L),
        layers_per_block = 2L,
        transformer_layers_per_block = c(0L, 2L, 10L),
        cross_attention_dim = 2048L,
        attention_head_dim = 64L,
        addition_embed_dim = 1280L,
        addition_time_embed_dim = 256L
    ) {
    self$in_channels <- in_channels
    self$out_channels <- out_channels
    self$block_out_channels <- block_out_channels
    self$layers_per_block <- layers_per_block
    self$transformer_layers_per_block <- transformer_layers_per_block

    # Time embedding dimension (same as SD21)
    time_embed_dim <- block_out_channels[1] * 4L # 320 * 4 = 1280

    # Input convolution
    self$conv_in <- torch::nn_conv2d(in_channels, block_out_channels[1], 3L,
                                     padding = 1L)

    # Time embedding MLP
    self$time_embedding_linear_1 <- torch::nn_linear(block_out_channels[1], time_embed_dim)
    self$time_embedding_linear_2 <- torch::nn_linear(time_embed_dim, time_embed_dim)

    # Additional embedding (SDXL-specific): projects text_embeds + time_ids
    # time_ids has 6 values, projected through 256-dim Fourier features = 1536
    # text_embeds is 1280-dim, so total input is 2816
    add_embed_input_dim <- addition_embed_dim + 6L * addition_time_embed_dim
    self$add_embedding_linear_1 <- torch::nn_linear(add_embed_input_dim, time_embed_dim)
    self$add_embedding_linear_2 <- torch::nn_linear(time_embed_dim, time_embed_dim)

    # Time projection for time_ids (SDXL uses Fourier features)
    self$add_time_proj_dim <- addition_time_embed_dim

    # Down blocks (SDXL: 3 blocks with variable attention)
    self$down_blocks <- torch::nn_module_list()
    output_channel <- block_out_channels[1]

    for (i in seq_along(block_out_channels)) {
        input_channel <- output_channel
        output_channel <- block_out_channels[i]
        is_final_block <- (i == length(block_out_channels))
        transformer_depth <- transformer_layers_per_block[i]

        down_block <- self$create_down_block_sdxl(
            input_channel, output_channel, time_embed_dim,
            cross_attention_dim, attention_head_dim,
            layers_per_block, transformer_depth,
            add_downsample = !is_final_block
        )
        self$down_blocks$append(down_block)
    }

    # Mid block (SDXL: 10 transformer blocks)
    mid_channels <- block_out_channels[length(block_out_channels)]
    mid_transformer_depth <- transformer_layers_per_block[length(transformer_layers_per_block)]
    self$mid_block <- self$create_mid_block_sdxl(
        mid_channels, time_embed_dim, cross_attention_dim, attention_head_dim,
        mid_transformer_depth
    )

    # Up blocks (SDXL: reverse order, 3 blocks)
    self$up_blocks <- torch::nn_module_list()
    reversed_channels <- rev(block_out_channels)
    reversed_transformer_depths <- rev(transformer_layers_per_block)

    # Calculate skip channels for SDXL (3 blocks)
    # down0: [in, r0, r1, ds], down1: [r0, r1, ds], down2: [r0, r1]
    skip_channels <- c(
                       block_out_channels[1], # conv_in
                       rep(block_out_channels[1], layers_per_block), # down0 resnets
                       block_out_channels[1], # down0 downsample
                       rep(block_out_channels[2], layers_per_block), # down1 resnets
                       block_out_channels[2], # down1 downsample
                       rep(block_out_channels[3], layers_per_block) # down2 resnets (no downsample)
    )
    self$skip_channels <- rev(skip_channels)

    for (i in seq_along(reversed_channels)) {
        output_channel <- reversed_channels[i]
        if (i == 1) {
            prev_output <- mid_channels
        } else {
            prev_output <- reversed_channels[i - 1]
        }

        is_final_block <- (i == length(reversed_channels))
        transformer_depth <- reversed_transformer_depths[i]

        # Calculate per-resnet skip channels
        num_resnets <- layers_per_block + 1L
        resnet_skip_channels <- integer(num_resnets)
        for (j in seq_len(num_resnets)) {
            skip_idx <- (i - 1) * num_resnets + j
            resnet_skip_channels[j] <- self$skip_channels[skip_idx]
        }

        up_block <- self$create_up_block_sdxl(
            prev_output, output_channel, resnet_skip_channels, time_embed_dim,
            cross_attention_dim, attention_head_dim, transformer_depth,
            add_upsample = !is_final_block
        )
        self$up_blocks$append(up_block)
    }

    # Output
    self$conv_norm_out <- group_norm_32(block_out_channels[1])
    self$conv_out <- torch::nn_conv2d(block_out_channels[1], out_channels, 3L, padding = 1L)
},

                                     create_down_block_sdxl = function(
        in_channels,
        out_channels,
        time_embed_dim,
        cross_attention_dim,
        attention_head_dim,
        num_layers,
        transformer_depth,
        add_downsample = TRUE
    ) {
    DownBlock <- torch::nn_module(
                                  "DownBlockSDXL",
                                  initialize = function() {
        # ResNets
        self$resnets <- torch::nn_module_list()
        for (i in seq_len(num_layers)) {
            if (i == 1) {
                in_ch <- in_channels
            } else {
                in_ch <- out_channels
            }
            self$resnets$append(UNetResBlock(in_ch, out_channels, time_embed_dim))
        }

        # Attention blocks (only if transformer_depth > 0)
        self$has_attentions <- transformer_depth > 0
        if (self$has_attentions) {
            self$attentions <- torch::nn_module_list()
            n_heads <- out_channels %/% attention_head_dim
            for (i in seq_len(num_layers)) {
                self$attentions$append(
                                       SpatialTransformer(out_channels, n_heads, attention_head_dim,
                        depth = transformer_depth, context_dim = cross_attention_dim)
                )
            }
        }

        # Downsample
        self$has_downsamplers <- add_downsample
        if (add_downsample) {
            self$downsamplers <- torch::nn_module_list()
            self$downsamplers$append(Downsample2D(out_channels))
        }
    }
    )
    DownBlock()
},

                                     create_mid_block_sdxl = function(
        channels,
        time_embed_dim,
        cross_attention_dim,
        attention_head_dim,
        transformer_depth
    ) {
    MidBlock <- torch::nn_module(
                                 "MidBlockSDXL",
                                 initialize = function() {
        self$resnets <- torch::nn_module_list()
        self$resnets$append(UNetResBlock(channels, channels, time_embed_dim))
        self$resnets$append(UNetResBlock(channels, channels, time_embed_dim))

        n_heads <- channels %/% attention_head_dim
        self$attentions <- torch::nn_module_list()
        self$attentions$append(
                               SpatialTransformer(channels, n_heads, attention_head_dim,
                depth = transformer_depth, context_dim = cross_attention_dim)
        )
    }
    )
    MidBlock()
},

                                     create_up_block_sdxl = function(
        in_channels,
        out_channels,
        resnet_skip_channels,
        time_embed_dim,
        cross_attention_dim,
        attention_head_dim,
        transformer_depth,
        add_upsample = TRUE
    ) {
    num_resnets <- length(resnet_skip_channels)

    UpBlock <- torch::nn_module(
                                "UpBlockSDXL",
                                initialize = function() {
        self$resnets <- torch::nn_module_list()
        for (i in seq_len(num_resnets)) {
            if (i == 1) {
                in_ch <- in_channels + resnet_skip_channels[i]
            } else {
                in_ch <- out_channels + resnet_skip_channels[i]
            }
            self$resnets$append(UNetResBlock(in_ch, out_channels, time_embed_dim))
        }

        # Attention blocks (only if transformer_depth > 0)
        self$has_attentions <- transformer_depth > 0
        if (self$has_attentions) {
            self$attentions <- torch::nn_module_list()
            n_heads <- out_channels %/% attention_head_dim
            for (i in seq_len(num_resnets)) {
                self$attentions$append(
                                       SpatialTransformer(out_channels, n_heads, attention_head_dim,
                        depth = transformer_depth, context_dim = cross_attention_dim)
                )
            }
        }

        # Upsample
        self$has_upsamplers <- add_upsample
        if (add_upsample) {
            self$upsamplers <- torch::nn_module_list()
            self$upsamplers$append(Upsample2D(out_channels))
        }
    }
    )
    UpBlock()
},

                                     forward = function(
        sample,
        timestep,
        encoder_hidden_states,
        text_embeds,
        time_ids
    ) {
    # Accepts same signature as TorchScript:
    # unet(sample, timestep, encoder_hidden_states, text_embeds, time_ids)

    # Get model dtype from weights
    model_dtype <- self$time_embedding_linear_1$weight$dtype

    # Time embedding (computed in float32, then cast to model dtype)
    # SDXL uses flip_sin_to_cos=TRUE (default), downscale_freq_shift=0 (default)
    t_emb <- timestep_embedding(timestep, self$block_out_channels[1])
    t_emb <- t_emb$to(dtype = model_dtype)
    t_emb <- self$time_embedding_linear_1(t_emb)
    t_emb <- torch::nnf_silu(t_emb)
    t_emb <- self$time_embedding_linear_2(t_emb)

    # Additional embedding (Fourier projection of time_ids + text_embeds)
    # Uses same defaults as main time embedding
    time_ids_emb <- timestep_embedding(time_ids$flatten(), self$add_time_proj_dim)
    time_ids_emb <- time_ids_emb$to(dtype = model_dtype)
    time_ids_emb <- time_ids_emb$reshape(c(text_embeds$shape[1], -1L))
    add_embeds <- torch::torch_cat(list(text_embeds, time_ids_emb), dim = 2L)
    add_emb <- self$add_embedding_linear_1(add_embeds)
    add_emb <- torch::nnf_silu(add_emb)
    add_emb <- self$add_embedding_linear_2(add_emb)

    # Combine time and additional embeddings
    emb <- t_emb + add_emb

    # Input conv
    sample <- self$conv_in(sample)

    # Store skip connections
    down_block_res_samples <- list(sample)

    # Down blocks
    for (i in seq_along(self$down_blocks)) {
        block <- self$down_blocks[[i]]

        for (j in seq_along(block$resnets)) {
            sample <- block$resnets[[j]](sample, emb)
            if (block$has_attentions) {
                sample <- block$attentions[[j]](sample, encoder_hidden_states)
            }
            down_block_res_samples <- c(down_block_res_samples, list(sample))
        }

        if (block$has_downsamplers) {
            sample <- block$downsamplers[[1]](sample)
            down_block_res_samples <- c(down_block_res_samples, list(sample))
        }
    }

    # Mid block
    sample <- self$mid_block$resnets[[1]](sample, emb)
    sample <- self$mid_block$attentions[[1]](sample, encoder_hidden_states)
    sample <- self$mid_block$resnets[[2]](sample, emb)

    # Up blocks
    for (i in seq_along(self$up_blocks)) {
        block <- self$up_blocks[[i]]

        for (j in seq_along(block$resnets)) {
            # Pop skip connection
            res_sample <- down_block_res_samples[[length(down_block_res_samples)]]
            down_block_res_samples <- down_block_res_samples[-length(down_block_res_samples)]

            # Concatenate
            sample <- torch::torch_cat(list(sample, res_sample), dim = 2L)

            sample <- block$resnets[[j]](sample, emb)
            if (block$has_attentions) {
                sample <- block$attentions[[j]](sample, encoder_hidden_states)
            }
        }

        if (block$has_upsamplers) {
            sample <- block$upsamplers[[1]](sample)
        }
    }

    # Output
    sample <- self$conv_norm_out(sample)
    sample <- torch::nnf_silu(sample)
    sample <- self$conv_out(sample)

    sample
}
)

#' Detect SDXL UNet architecture from TorchScript file
#'
#' @param torchscript_path Path to TorchScript SDXL UNet .pt file
#' @return List with architecture parameters
#' @keywords internal
detect_unet_sdxl_architecture <- function(torchscript_path) {
    ts_unet <- torch::jit_load(torchscript_path, device = "cpu")
    ts_params <- ts_unet$parameters
    param_names <- names(ts_params)

    # Get block_out_channels from down_blocks resnets
    block_channels <- integer(0)
    for (i in 0:2) { # SDXL has 3 blocks
        key <- sprintf("unet.down_blocks.%d.resnets.0.norm2.weight", i)
        if (key %in% param_names) {
            block_channels <- c(block_channels,
                                as.integer(ts_params[[key]]$shape[1]))
        }
    }

    # Get transformer depths per block
    transformer_depths <- integer(length(block_channels))
    for (i in seq_along(block_channels)) {
        depth <- 0L
        for (j in 0:15) {
            key <- sprintf("unet.down_blocks.%d.attentions.0.transformer_blocks.%d.attn1.to_q.weight", i - 1, j)
            if (key %in% param_names) {
                depth <- depth + 1L
            }
        }
        transformer_depths[i] <- depth
    }

    # Get cross_attention_dim from attn2.to_k (use block with attention)
    cross_attention_dim <- NULL
    for (i in 0:2) {
        attn_key <- sprintf("unet.down_blocks.%d.attentions.0.transformer_blocks.0.attn2.to_k.weight", i)
        if (attn_key %in% param_names) {
            cross_attention_dim <- as.integer(ts_params[[attn_key]]$shape[2])
            break
        }
    }

    list(
         in_channels = 4L,
         out_channels = 4L,
         block_out_channels = block_channels,
         layers_per_block = 2L,
         transformer_layers_per_block = transformer_depths,
         cross_attention_dim = cross_attention_dim,
         attention_head_dim = 64L,
         addition_embed_dim = 1280L,
         addition_time_embed_dim = 256L
    )
}

#' Create native SDXL UNet from TorchScript
#'
#' @param torchscript_path Path to TorchScript SDXL UNet .pt file
#' @param verbose Print loading progress
#'
#' @return A native SDXL UNet module with loaded weights
#' @export
unet_sdxl_native_from_torchscript <- function(torchscript_path,
    verbose = TRUE) {
    arch <- detect_unet_sdxl_architecture(torchscript_path)

    if (verbose) {
        message("Detected SDXL UNet architecture:")
        message("  block_out_channels: ",
                paste(arch$block_out_channels, collapse = ", "))
        message("  transformer_layers_per_block: ", paste(arch$transformer_layers_per_block, collapse = ", "))
        message("  cross_attention_dim: ", arch$cross_attention_dim)
    }

    unet <- unet_sdxl_native(
                             in_channels = arch$in_channels,
                             out_channels = arch$out_channels,
                             block_out_channels = arch$block_out_channels,
                             layers_per_block = arch$layers_per_block,
                             transformer_layers_per_block = arch$transformer_layers_per_block,
                             cross_attention_dim = arch$cross_attention_dim,
                             attention_head_dim = arch$attention_head_dim,
                             addition_embed_dim = arch$addition_embed_dim,
                             addition_time_embed_dim = arch$addition_time_embed_dim
    )

    load_unet_sdxl_weights(unet, torchscript_path, verbose = verbose)

    unet
}

#' Load weights from TorchScript SDXL UNet into native SDXL UNet
#'
#' @param native_unet Native SDXL UNet module
#' @param torchscript_path Path to TorchScript SDXL UNet .pt file
#' @param verbose Print loading progress
#'
#' @return The native UNet with loaded weights (invisibly)
#' @export
load_unet_sdxl_weights <- function(native_unet, torchscript_path,
                                   verbose = TRUE) {
    ts_unet <- torch::jit_load(torchscript_path, device = "cpu")
    ts_params <- ts_unet$parameters

    remap_key <- function(key) {
        # Strip unet. prefix
        key <- sub("^unet\\.", "", key)

        # Time embedding: time_embedding.linear_1 -> time_embedding_linear_1
        key <- sub("^time_embedding\\.linear_1", "time_embedding_linear_1", key)
        key <- sub("^time_embedding\\.linear_2", "time_embedding_linear_2", key)

        # Add embedding: add_embedding.linear_1 -> add_embedding_linear_1
        key <- sub("^add_embedding\\.linear_1", "add_embedding_linear_1", key)
        key <- sub("^add_embedding\\.linear_2", "add_embedding_linear_2", key)

        key
    }

    loaded <- 0
    skipped <- 0
    unmapped <- character(0)

    torch::with_no_grad({
        for (ts_name in names(ts_params)) {
            native_name <- remap_key(ts_name)

            if (native_name %in% names(native_unet$parameters)) {
                ts_tensor <- ts_params[[ts_name]]
                native_tensor <- native_unet$parameters[[native_name]]

                if (all(ts_tensor$shape == native_tensor$shape)) {
                    native_tensor$copy_(ts_tensor)
                    loaded <- loaded + 1
                } else if (verbose) {
                    message("Shape mismatch: ", native_name,
                            " (", paste(as.integer(ts_tensor$shape),
                                        collapse = "x"), " vs ",
                            paste(as.integer(native_tensor$shape), collapse = "x"), ")")
                    skipped <- skipped + 1
                }
            } else {
                skipped <- skipped + 1
                unmapped <- c(unmapped, paste0(ts_name, " -> ", native_name))
            }
        }
    })

    if (verbose) {
        if (length(unmapped) > 0 && length(unmapped) <= 10) {
            message("Unmapped parameters:")
            for (u in unmapped) {
                message("  ", u)
            }
        } else if (length(unmapped) > 10) {
            message("Unmapped: ", length(unmapped), " parameters (showing first 10)")
            for (u in head(unmapped, 10)) {
                message("  ", u)
            }
        }
        message("Loaded ", loaded, "/", loaded + skipped, " parameters")
    }

    invisible(native_unet)
}

#' LTX-2.3 Audio VAE
#'
#' Fresh R port of the LTX-2 audio autoencoder from the diffusers
#' reference (Apache-2.0, autoencoder_kl_ltx2_audio.py), configured per
#' the checkpoint: pixel norm, height-axis causality, base 128 channels
#' with multipliers (1, 2, 4), 8 latent channels, 64 mel bins, no
#' attention. The decoder produces mel for the vocoder; the encoder
#' turns user audio into conditioning latents (lip sync).
#'
#' @name audio_vae_ltx23
NULL

# Audio latents are 4x downsampled in time relative to mel frames
.ltx23_audio_latent_downsample_factor <- 4L

#' Causal 2D convolution for audio spectrograms
#'
#' Pads asymmetrically along the causal axis ("height" = time frames for
#' LTX audio) before an unpadded Conv2d.
#'
#' @param in_channels,out_channels Integers.
#' @param kernel_size Integer or length-2 vector.
#' @param stride Integer.
#' @param causality_axis "height", "width", "width-compatibility", or "none".
#'
#' @return Module whose forward(x) returns the convolved tensor, padded
#'   so that each output frame depends only on current and earlier
#'   input frames.
#'
#' @export
ltx23_audio_causal_conv2d <- torch::nn_module(
    "ltx23_audio_causal_conv2d",
    initialize = function(
                          in_channels,
                          out_channels,
                          kernel_size = 3L,
                          stride = 1L,
                          causality_axis = "height"
    ) {
    if (length(kernel_size) == 1L) kernel_size <- rep(kernel_size, 2L)
    pad_h <- kernel_size[1] - 1L
    pad_w <- kernel_size[2] - 1L

    self$padding <- switch(causality_axis,
                           none = c(pad_w %/% 2L, pad_w - pad_w %/% 2L, pad_h %/% 2L,
                                    pad_h - pad_h %/% 2L),
                           width = c(pad_w, 0L, pad_h %/% 2L, pad_h - pad_h %/% 2L),
                           `width-compatibility` = c(pad_w, 0L, pad_h %/% 2L, pad_h - pad_h %/% 2L),
                           height = c(pad_w %/% 2L, pad_w - pad_w %/% 2L, pad_h, 0L),
                           stop("Invalid causality_axis: ", causality_axis)
    )
    self$conv <- torch::nn_conv2d(
                                  in_channels, out_channels, kernel_size,
                                  stride = stride, padding = 0L
    )
},
    forward = function(x) {
    self$conv(torch::nnf_pad(x, self$padding))
}
)

#' LTX audio ResNet block
#'
#' PixelNorm -> SiLU -> causal conv, twice, with a 1x1 causal conv
#' shortcut (\code{nin_shortcut}) on channel change.
#'
#' @param in_channels,out_channels Integers.
#' @param causality_axis Character.
#'
#' @return Module whose forward(x) returns \code{x} plus the residual
#'   branch, a tensor of the same shape as \code{x}.
#'
#' @export
ltx23_audio_resnet_block <- torch::nn_module(
    "ltx23_audio_resnet_block",
    initialize = function(
                          in_channels,
                          out_channels = NULL,
                          causality_axis = "height"
    ) {
    out_channels <- out_channels %||% in_channels
    self$changes_channels <- in_channels != out_channels

    self$norm1 <- ltx23_per_channel_rms_norm(eps = 1e-6)
    self$conv1 <- ltx23_audio_causal_conv2d(in_channels, out_channels,
        kernel_size = 3L, causality_axis = causality_axis)
    self$norm2 <- ltx23_per_channel_rms_norm(eps = 1e-6)
    self$conv2 <- ltx23_audio_causal_conv2d(
        out_channels, out_channels, kernel_size = 3L, causality_axis = causality_axis
    )
    if (self$changes_channels) {
        self$nin_shortcut <- ltx23_audio_causal_conv2d(
            in_channels, out_channels, kernel_size = 1L, causality_axis = causality_axis
        )
    }
},
    forward = function(x) {
    h <- self$norm1(x)
    h <- torch::nnf_silu(h)
    h <- self$conv1(h)
    h <- self$norm2(h)
    h <- torch::nnf_silu(h)
    h <- self$conv2(h)
    if (self$changes_channels) {
        x <- self$nin_shortcut(x)
    }
    x + h
}
)

#' LTX audio upsampler
#'
#' Nearest 2x interpolation, causal conv, then a crop of the first
#' element along the causal axis.
#'
#' @param in_channels Integer.
#' @param causality_axis Character.
#'
#' @return Module whose forward(x) returns the tensor upsampled 2x by
#'   nearest-neighbour interpolation and convolved.
#'
#' @export
ltx23_audio_upsample <- torch::nn_module(
    "ltx23_audio_upsample",
    initialize = function(in_channels, causality_axis = "height") {
    self$causality_axis <- causality_axis
    self$conv <- ltx23_audio_causal_conv2d(in_channels, in_channels,
        kernel_size = 3L, causality_axis = causality_axis)
},
    forward = function(x) {
    x <- torch::nnf_interpolate(x, scale_factor = 2, mode = "nearest")
    x <- self$conv(x)
    if (self$causality_axis == "height") {
        # Drop the first (causally padded) frame; Python [:, :, 1:, :]
        x <- x$narrow(3L, 2L, x$shape[3] - 1L)
    } else if (self$causality_axis == "width") {
        x <- x$narrow(4L, 2L, x$shape[4] - 1L)
    }
    x
}
)

#' LTX audio downsampler
#'
#' Causal zero-pad followed by a plain stride-2 conv (reference
#' LTX2AudioDownsample; note the conv is unwrapped, so its checkpoint
#' key is \code{downsample.conv.*}).
#'
#' @param in_channels Integer.
#' @param causality_axis Character.
#'
#' @return Module whose forward(x) returns the strided convolution of
#'   \code{x}, halving the downsampled axes.
#'
#' @export
ltx23_audio_downsample <- torch::nn_module(
    "ltx23_audio_downsample",
    initialize = function(in_channels, causality_axis = "height") {
    # Padding order: (left, right, top, bottom)
    self$padding <- switch(causality_axis, none = c(0L, 1L, 0L, 1L),
                           width = c(2L, 0L, 0L, 1L),
                           height = c(0L, 1L, 2L, 0L),
                           `width-compatibility` = c(1L, 0L, 0L, 1L),
                           stop("Invalid causality_axis: ", causality_axis))
    self$conv <- torch::nn_conv2d(in_channels, in_channels, kernel_size = 3L,
                                  stride = 2L, padding = 0L)
},
    forward = function(x) {
    self$conv(torch::nnf_pad(x, self$padding))
}
)

#' LTX-2.3 audio VAE encoder
#'
#' Mel spectrogram [B, 2, T, 64] -> latent distribution moments
#' [B, 2 * latent_channels, ceil(T/4), 16]. Structure mirrors the
#' decoder: causal convs, parameterless pixel norms, ResNet stages with
#' stride-2 downsampling between levels (reference LTX2AudioEncoder).
#'
#' @param base_channels,num_res_blocks,latent_channels,ch_mult,causality_axis
#'   See \code{\link{ltx23_audio_decoder}}.
#' @param in_channels Integer. Mel channels (2 = stereo).
#'
#' @return Module whose forward(x) returns the encoded audio latent, a
#'   tensor downsampled along time and mel axes with the configured
#'   latent channel count.
#'
#' @export
ltx23_audio_encoder <- torch::nn_module(
                                        "ltx23_audio_encoder",
                                        initialize = function(
        base_channels = 128L,
        in_channels = 2L,
        num_res_blocks = 2L,
        latent_channels = 8L,
        ch_mult = c(1L, 2L, 4L),
        causality_axis = "height"
    ) {
    num_levels <- length(ch_mult)
    self$conv_in <- ltx23_audio_causal_conv2d(in_channels, base_channels,
        kernel_size = 3L, causality_axis = causality_axis)

    block_in <- base_channels
    stages <- list()
    for (level in seq_len(num_levels)) {
        block_out <- base_channels * ch_mult[level]
        blocks <- list()
        for (j in seq_len(num_res_blocks)) {
            blocks[[j]] <- ltx23_audio_resnet_block(
                if (j == 1L) block_in else block_out, block_out,
                causality_axis = causality_axis
            )
        }
        block_in <- block_out
        stage <- torch::nn_module(
                                  "ltx23_audio_down_stage",
                                  initialize = function(blocks, downsample) {
            self$block <- torch::nn_module_list(blocks)
            if (!is.null(downsample)) {
                self$downsample <- downsample
            }
        },
                                  forward = function(x) {
            for (i in seq_along(self$block)) {
                x <- self$block[[i]](x)
            }
            if (!is.null(self$downsample)) {
                x <- self$downsample(x)
            }
            x
        }
        )
        stages[[level]] <- stage(
                                 blocks,
            if (level != num_levels) {
                ltx23_audio_downsample(block_in, causality_axis = causality_axis)
            } else {
                NULL
            }
        )
    }
    self$down <- torch::nn_module_list(stages)

    self$mid <- torch::nn_module(
                                 "ltx23_audio_mid",
                                 initialize = function(channels, causality_axis) {
        self$block_1 <- ltx23_audio_resnet_block(channels,
            causality_axis = causality_axis)
        self$block_2 <- ltx23_audio_resnet_block(channels,
            causality_axis = causality_axis)
    },
                                 forward = function(x) {
        self$block_2(self$block_1(x))
    }
    )(block_in, causality_axis)

    self$norm_out <- ltx23_per_channel_rms_norm(eps = 1e-6)
    self$conv_out <- ltx23_audio_causal_conv2d(
        block_in, 2L * latent_channels, kernel_size = 3L,
        causality_axis = causality_axis
    )
    self$latent_channels <- as.integer(latent_channels)
},
                                        forward = function(x) {
    h <- self$conv_in(x)
    for (level in seq_along(self$down)) {
        h <- self$down[[level]](h)
    }
    h <- self$mid(h)
    h <- self$norm_out(h)
    h <- torch::nnf_silu(h)
    self$conv_out(h)
}
)

#' LTX-2.3 audio VAE decoder
#'
#' Latents [B, 8, L, 16] -> mel spectrogram [B, 2, 4L - 3, 64].
#'
#' @param base_channels Integer.
#' @param output_channels Integer. Audio channels (2 = stereo).
#' @param num_res_blocks Integer. Per-level ResNet count (a stage runs
#'   \code{num_res_blocks + 1} blocks).
#' @param latent_channels Integer.
#' @param ch_mult Integer vector. Channel multipliers per level.
#' @param causality_axis Character.
#' @param mel_bins Integer. Output mel bins (crop/pad target).
#'
#' @return Module whose forward(x) returns the decoded mel
#'   spectrogram reconstructed from an audio latent.
#'
#' @export
ltx23_audio_decoder <- torch::nn_module(
                                        "ltx23_audio_decoder",
                                        initialize = function(
        base_channels = 128L,
        output_channels = 2L,
        num_res_blocks = 2L,
        latent_channels = 8L,
        ch_mult = c(1L, 2L, 4L),
        causality_axis = "height",
        mel_bins = 64L
    ) {
    self$num_resolutions <- length(ch_mult)
    self$num_res_blocks <- num_res_blocks
    self$out_ch <- as.integer(output_channels)
    self$mel_bins <- as.integer(mel_bins)
    self$causal <- causality_axis != "none"

    base_block_channels <- base_channels * ch_mult[length(ch_mult)]

    self$conv_in <- ltx23_audio_causal_conv2d(latent_channels,
        base_block_channels, kernel_size = 3L,
        causality_axis = causality_axis)

    mid_container <- torch::nn_module(
                                      "ltx23_audio_mid",
                                      initialize = function(channels, causality_axis) {
        self$block_1 <- ltx23_audio_resnet_block(channels, channels,
            causality_axis = causality_axis)
        self$block_2 <- ltx23_audio_resnet_block(channels, channels,
            causality_axis = causality_axis)
    },
                                      forward = function(x) {
        self$block_2(self$block_1(x))
    }
    )
    self$mid <- mid_container(base_block_channels, causality_axis)

    # Stages are stored at their level index; built from the top level
    # down, tracking the running channel count like the reference
    stages <- vector("list", self$num_resolutions)
    block_in <- base_block_channels
    for (level in rev(seq_len(self$num_resolutions))) {
        block_out <- base_channels * ch_mult[level]
        stage_blocks <- list()
        for (j in seq_len(num_res_blocks + 1L)) {
            stage_blocks[[j]] <- ltx23_audio_resnet_block(
                block_in, block_out, causality_axis = causality_axis
            )
            block_in <- block_out
        }
        stage <- torch::nn_module(
                                  "ltx23_audio_up_stage",
                                  initialize = function(blocks, upsample) {
            self$block <- torch::nn_module_list(blocks)
            if (!is.null(upsample)) self$upsample <- upsample
        },
                                  forward = function(x) {
            for (i in seq_along(self$block)) {
                x <- self$block[[i]](x)
            }
            if (!is.null(self$upsample)) x <- self$upsample(x)
            x
        }
        )
        upsample <- if (level != 1L) {
            ltx23_audio_upsample(block_in, causality_axis = causality_axis)
        } else {
            NULL
        }
        stages[[level]] <- stage(stage_blocks, upsample)
    }
    self$up <- torch::nn_module_list(stages)

    self$norm_out <- ltx23_per_channel_rms_norm(eps = 1e-6)
    self$conv_out <- ltx23_audio_causal_conv2d(
        block_in, output_channels, kernel_size = 3L, causality_axis = causality_axis
    )
},
                                        forward = function(sample) {
    frames <- sample$shape[3]
    target_frames <- frames * .ltx23_audio_latent_downsample_factor
    if (self$causal) {
        target_frames <- max(target_frames - (.ltx23_audio_latent_downsample_factor - 1L), 1L)
    }

    h <- self$conv_in(sample)
    h <- self$mid(h)
    for (level in rev(seq_len(self$num_resolutions))) {
        h <- self$up[[level]](h)
    }

    h <- self$norm_out(h)
    h <- torch::nnf_silu(h)
    h <- self$conv_out(h)

    # Crop/pad to the target (channels, time, mel bins)
    cur_time <- h$shape[3]
    cur_freq <- h$shape[4]
    h <- h$narrow(2L, 1L, self$out_ch)
    h <- h$narrow(3L, 1L, min(cur_time, target_frames))
    h <- h$narrow(4L, 1L, min(cur_freq, self$mel_bins))

    time_pad <- target_frames - h$shape[3]
    freq_pad <- self$mel_bins - h$shape[4]
    if (time_pad > 0L || freq_pad > 0L) {
        h <- torch::nnf_pad(h, c(0L, max(freq_pad, 0L), 0L, max(time_pad, 0L)))
    }
    h
}
)

#' LTX-2.3 audio VAE
#'
#' Encoder + decoder plus the per-channel latent statistics loaded from
#' the checkpoint. Encoding is used for audio-conditioned generation
#' (lip sync); decoding for generated audio.
#'
#' @param base_channels,output_channels,num_res_blocks,latent_channels,ch_mult,causality_axis,mel_bins
#'   See \code{\link{ltx23_audio_decoder}}.
#' @param in_channels Integer. Mel input channels (2 = stereo).
#'
#' @return Module bundling the audio encoder and decoder. Its
#'   forward(z) is \code{decode(z)}, returning the mel spectrogram for a
#'   latent; \code{$encode()} and \code{$decode()} are callable
#'   separately.
#'
#' @export
ltx23_audio_vae <- torch::nn_module(
                                    "ltx23_audio_vae",
                                    initialize = function(
        base_channels = 128L,
        output_channels = 2L,
        num_res_blocks = 2L,
        latent_channels = 8L,
        ch_mult = c(1L, 2L, 4L),
        causality_axis = "height",
        mel_bins = 64L,
        in_channels = 2L
    ) {
    self$encoder <- ltx23_audio_encoder(base_channels = base_channels,
                                        in_channels = in_channels,
                                        num_res_blocks = num_res_blocks,
                                        latent_channels = latent_channels,
                                        ch_mult = ch_mult,
                                        causality_axis = causality_axis)
    self$decoder <- ltx23_audio_decoder(base_channels = base_channels,
                                        output_channels = output_channels,
                                        num_res_blocks = num_res_blocks,
                                        latent_channels = latent_channels,
                                        ch_mult = ch_mult,
                                        causality_axis = causality_axis,
                                        mel_bins = mel_bins)
    # Statistics are stored at base_channels size in the checkpoint;
    # only the first latent_channels entries are meaningful
    self$latents_mean <- torch::nn_buffer(torch::torch_zeros(base_channels))
    self$latents_std <- torch::nn_buffer(torch::torch_ones(base_channels))
    self$latent_channels <- as.integer(latent_channels)
},
                                    encode = function(mel) {
    moments <- self$encoder(mel)
    n <- self$latent_channels
    list(
         mean = moments$narrow(2L, 1L, n),
         logvar = moments$narrow(2L, n + 1L, n)
    )
},
                                    decode = function(z) {
    .ltx23_traced_call(self$decoder, z)
},
                                    forward = function(z) {
    self$decode(z)
}
)

#' Map an official audio VAE checkpoint key to the R module name
#'
#' @param key Character. Checkpoint key.
#'
#' @return Character.
#'
#' @export
ltx23_map_audio_vae_key <- function(key) {
    key <- sub("^audio_vae\\.", "", key)
    key <- sub("^per_channel_statistics\\.mean-of-means$", "latents_mean", key)
    key <- sub("^per_channel_statistics\\.std-of-means$", "latents_std", key)
    key
}

# Penultimate hidden-state support for SDXL text encoders. SDXL feeds the
# UNet hidden_states[-2] (the second-to-last block output, pre-final-LN)
# from both encoders, and te2's pooled text_embeds from the full stack.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
    exit_file("torch not fully installed")
}
library(diffuseR)

set.seed(1)
torch::torch_manual_seed(1)

# tiny encoders (embed_dim divisible by num_heads; a pseudo-EOS at the
# highest id so argmax locates the pooled position)
mk1 <- function(pen) {
    m <- text_encoder_native(vocab_size = 50, context_length = 8,
        embed_dim = 32, num_layers = 4, num_heads = 2, mlp_dim = 64,
        apply_final_ln = FALSE, return_penultimate = pen, gelu_type = "quick")
    m$eval()
    m
}
ids <- torch::torch_tensor(matrix(c(40L, 5L, 12L, 49L, 0L, 0L, 0L, 0L),
    nrow = 1), dtype = torch::torch_long())

# --- text_encoder_native: penultimate differs from last, same shape ---
e_last <- mk1(FALSE)
e_pen <- mk1(TRUE)
# copy weights so the two modules are identical apart from the flag
e_pen$load_state_dict(e_last$state_dict())
h_last <- torch::with_no_grad(e_last(ids))
h_pen <- torch::with_no_grad(e_pen(ids))
expect_equal(dim(h_pen), dim(h_last))                       # [1, 8, 32]
expect_true(as.numeric((h_pen - h_last)$abs()$max()) > 1e-4) # genuinely earlier layer

# --- text_encoder2_native: pooled is stack-invariant, hidden is not ---
mk2 <- function(pen) {
    m <- text_encoder2_native(vocab_size = 50, context_length = 8,
        embed_dim = 32, num_layers = 4, num_heads = 2, mlp_dim = 64,
        return_penultimate = pen)
    m$eval()
    m
}
e2_last <- mk2(FALSE)
e2_pen <- mk2(TRUE)
e2_pen$load_state_dict(e2_last$state_dict())
o_last <- torch::with_no_grad(e2_last(ids))
o_pen <- torch::with_no_grad(e2_pen(ids))

# hidden output: penultimate vs last differ
expect_equal(dim(o_pen[[1]]), dim(o_last[[1]]))             # [1, 8, 32]
expect_true(as.numeric((o_pen[[1]] - o_last[[1]])$abs()$max()) > 1e-4)

# pooled output: computed from the full stack, so identical either way
expect_equal(dim(o_pen[[2]]), c(1L, 32L))
expect_true(as.numeric((o_pen[[2]] - o_last[[2]])$abs()$max()) < 1e-5)

# Parity tests for the Z-Image transformer blocks against diffusers
# reference fixtures (generated by tools/gen_fixtures_zimage_dit.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "dit_zimage.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/dit_zimage.safetensors"
if (!file.exists(fixture_path)) exit_file("zimage dit fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

load_named_weights <- function(module, weights) {
  dests <- c(module$named_parameters(), module$named_buffers())
  missing_dest <- setdiff(names(weights), names(dests))
  if (length(missing_dest)) {
    stop("No destination for: ", paste(utils::head(missing_dest, 5), collapse = ", "))
  }
  unfilled <- setdiff(names(dests), names(weights))
  if (length(unfilled)) {
    stop("Unfilled params: ", paste(utils::head(unfilled, 5), collapse = ", "))
  }
  torch::with_no_grad({
    for (name in names(weights)) dests[[name]]$copy_(weights[[name]])
  })
  invisible(module)
}

subset_prefix <- function(fx, prefix) {
  keys <- grep(paste0("^", prefix, "\\."), names(fx), value = TRUE)
  out <- fx[keys]
  names(out) <- sub(paste0("^", prefix, "\\."), "", keys)
  out
}

freqs <- list(fx$freqs_cos, fx$freqs_sin)

# --- modulated block (noise refiner / main trunk) --------------------------------

blk <- zimage_block(dim = 32L, n_heads = 2L, modulation = TRUE)
load_named_weights(blk, subset_prefix(fx, "mod"))
blk$eval()
out <- torch::with_no_grad(blk(fx$x, freqs, adaln_input = fx$adaln))
expect_equal(out$shape, c(1L, 24L, 32L))
expect_true(max_abs_diff(out, fx$mod_out) < 1e-5)

# --- unmodulated block (context refiner) ------------------------------------------

ublk <- zimage_block(dim = 32L, n_heads = 2L, modulation = FALSE)
load_named_weights(ublk, subset_prefix(fx, "unmod"))
ublk$eval()
uout <- torch::with_no_grad(ublk(fx$x, freqs))
expect_true(max_abs_diff(uout, fx$unmod_out) < 1e-5)

# The unmodulated block must carry no adaLN weights
expect_false(any(grepl("adaLN", names(ublk$named_parameters()))))

# --- chunked attention matches unchunked -------------------------------------------

out_chunked <- torch::with_no_grad(
  blk(fx$x, freqs, adaln_input = fx$adaln, chunk_size = 7L))
expect_true(max_abs_diff(out_chunked, out) < 1e-6)

# --- final layer -------------------------------------------------------------------

fin <- zimage_final_layer(hidden_size = 32L, out_channels = 64L)
load_named_weights(fin, subset_prefix(fx, "final"))
fin$eval()
fout <- torch::with_no_grad(fin(fx$x, fx$adaln))
expect_equal(fout$shape, c(1L, 24L, 64L))
expect_true(max_abs_diff(fout, fx$final_out) < 1e-5)

# --- timestep embedder --------------------------------------------------------------

temb <- zimage_t_embedder(out_size = 32L, mid_size = 48L)
load_named_weights(temb, subset_prefix(fx, "temb"))
temb$eval()
tout <- torch::with_no_grad(temb(fx$temb_in))
expect_true(max_abs_diff(tout, fx$temb_out) < 1e-5)

# --- cap embedder (Sequential RMSNorm + Linear, keys 0/1) ----------------------------

cap <- torch::nn_sequential(
  diffuseR:::ltx23_rms_norm(24L, eps = 1e-5),
  torch::nn_linear(24L, 32L, bias = TRUE)
)
load_named_weights(cap, subset_prefix(fx, "cap"))
cap$eval()
cout <- torch::with_no_grad(cap(fx$cap_in))
expect_true(max_abs_diff(cout, fx$cap_out) < 1e-6)

# Parity tests for the FLUX.2 transformer blocks against diffusers
# reference fixtures (generated by tools/gen_fixtures_flux2_dit.py).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "dit_flux2.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/dit_flux2.safetensors"
if (!file.exists(fixture_path)) exit_file("flux2 dit fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

load_named_weights <- function(module, weights) {
  dests <- c(module$named_parameters(), module$named_buffers())
  missing_dest <- setdiff(names(weights), names(dests))
  if (length(missing_dest)) {
    stop("No destination for: ", paste(utils::head(missing_dest, 5), collapse = ", "))
  }
  unfilled <- setdiff(names(dests), names(weights))
  if (length(unfilled)) {
    stop("Unfilled params: ", paste(utils::head(unfilled, 5), collapse = ", "))
  }
  torch::with_no_grad({
    for (name in names(weights)) dests[[name]]$copy_(weights[[name]])
  })
  invisible(module)
}

fixture_group <- function(prefix) {
  keys <- grep(paste0("^", prefix, "\\."), names(fx), value = TRUE)
  w <- fx[keys]
  names(w) <- sub(paste0("^", prefix, "\\."), "", keys)
  w
}

rope <- list(fx$rope_cos, fx$rope_sin)
DIM <- 16L
HEADS <- 2L
HEAD_DIM <- 8L

# --- flux2_modulation --------------------------------------------------------------

mod2 <- flux2_modulation(DIM, mod_param_sets = 2L)
mod2$eval()
load_named_weights(mod2, fixture_group("mod2"))
expect_true(max_abs_diff(torch::with_no_grad(mod2(fx$temb)), fx$mod2_out) < 1e-5)

mod1 <- flux2_modulation(DIM, mod_param_sets = 1L)
mod1$eval()
load_named_weights(mod1, fixture_group("mod1"))
expect_true(max_abs_diff(torch::with_no_grad(mod1(fx$temb)), fx$mod1_out) < 1e-5)

# --- flux2_feed_forward (SwiGLU) ------------------------------------------------------

ff <- flux2_feed_forward(DIM, DIM, mult = 3.0)
ff$eval()
load_named_weights(ff, fixture_group("ff"))
expect_true(max_abs_diff(torch::with_no_grad(ff(fx$joint_x)), fx$ff_out) < 1e-5)

# --- flux2_double_block ----------------------------------------------------------------

dbl <- flux2_double_block(DIM, HEADS, HEAD_DIM, mlp_ratio = 3.0)
dbl$eval()
load_named_weights(dbl, fixture_group("dbl"))
db <- torch::with_no_grad(dbl(
  hidden_states = fx$img_x,
  encoder_hidden_states = fx$txt_x,
  temb_mod_img = fx$dbl_mod_img,
  temb_mod_txt = fx$dbl_mod_txt,
  image_rotary_emb = rope
))
expect_true(max_abs_diff(db[[1]], fx$dbl_enc_out) < 1e-5)
expect_true(max_abs_diff(db[[2]], fx$dbl_hid_out) < 1e-5)

# --- flux2_single_block ------------------------------------------------------------------

sgl <- flux2_single_block(DIM, HEADS, HEAD_DIM, mlp_ratio = 3.0)
sgl$eval()
load_named_weights(sgl, fixture_group("sgl"))
sg <- torch::with_no_grad(sgl(
  hidden_states = fx$joint_x,
  temb_mod = fx$sgl_mod,
  image_rotary_emb = rope
))
expect_true(max_abs_diff(sg, fx$sgl_out) < 1e-5)

# Parity tests for the FLUX transformer blocks against diffusers
# reference fixtures (generated by tools/gen_fixtures_flux_dit.py,
# checked in).

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "dit_flux.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/dit_flux.safetensors"
if (!file.exists(fixture_path)) exit_file("flux dit fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

# Copy fixture weights (diffusers state-dict names) into an R module.
# Errors if any name has no destination or any destination stays unfilled.
load_named_weights <- function(module, weights) {
  dests <- c(module$named_parameters(), module$named_buffers())
  missing_dest <- setdiff(names(weights), names(dests))
  if (length(missing_dest)) {
    stop("No destination for: ", paste(utils::head(missing_dest, 5), collapse = ", "))
  }
  unfilled <- setdiff(names(dests), names(weights))
  if (length(unfilled)) {
    stop("Unfilled params: ", paste(utils::head(unfilled, 5), collapse = ", "))
  }
  torch::with_no_grad({
    for (name in names(weights)) dests[[name]]$copy_(weights[[name]])
  })
  invisible(module)
}

fixture_group <- function(prefix) {
  keys <- grep(paste0("^", prefix, "\\."), names(fx), value = TRUE)
  w <- fx[keys]
  names(w) <- sub(paste0("^", prefix, "\\."), "", keys)
  w
}

rope <- list(fx$rope_cos, fx$rope_sin)
DIM <- 16L
HEADS <- 2L
HEAD_DIM <- 8L
S_TXT <- 7L

# --- flux_ada_layer_norm_zero ---------------------------------------------------

adazero <- flux_ada_layer_norm_zero(DIM)
adazero$eval()
load_named_weights(adazero, fixture_group("adazero"))
az <- torch::with_no_grad(adazero(fx$joint_x, emb = fx$temb))
expect_true(max_abs_diff(az[[1]], fx$adazero_x_norm) < 1e-5)
expect_true(max_abs_diff(az[[2]], fx$adazero_gate_msa) < 1e-5)
expect_true(max_abs_diff(az[[3]], fx$adazero_shift_mlp) < 1e-5)
expect_true(max_abs_diff(az[[4]], fx$adazero_scale_mlp) < 1e-5)
expect_true(max_abs_diff(az[[5]], fx$adazero_gate_mlp) < 1e-5)

# --- flux_ada_layer_norm_zero_single ----------------------------------------------

adasingle <- flux_ada_layer_norm_zero_single(DIM)
adasingle$eval()
load_named_weights(adasingle, fixture_group("adasingle"))
as_ <- torch::with_no_grad(adasingle(fx$joint_x, emb = fx$temb))
expect_true(max_abs_diff(as_[[1]], fx$adasingle_x_norm) < 1e-5)
expect_true(max_abs_diff(as_[[2]], fx$adasingle_gate) < 1e-5)

# --- flux_ada_layer_norm_continuous (scale-first chunk order) ----------------------

adacont <- flux_ada_layer_norm_continuous(DIM, DIM)
adacont$eval()
load_named_weights(adacont, fixture_group("adacont"))
ac <- torch::with_no_grad(adacont(fx$joint_x, fx$temb))
expect_true(max_abs_diff(ac, fx$adacont_out) < 1e-5)

# --- flux_attention, double-stream variant ------------------------------------------

attn_d <- flux_attention(DIM, HEADS, HEAD_DIM, added_kv = TRUE)
attn_d$eval()
load_named_weights(attn_d, fixture_group("attnd"))
ad <- torch::with_no_grad(attn_d(
  hidden_states = fx$img_x,
  encoder_hidden_states = fx$txt_x,
  image_rotary_emb = rope
))
expect_equal(as.integer(ad[[1]]$shape), as.integer(fx$attnd_out$shape))
expect_true(max_abs_diff(ad[[1]], fx$attnd_out) < 1e-5)
expect_true(max_abs_diff(ad[[2]], fx$attnd_ctx_out) < 1e-5)

# --- flux_attention, pre_only variant -------------------------------------------------

attn_s <- flux_attention(DIM, HEADS, HEAD_DIM, pre_only = TRUE)
attn_s$eval()
load_named_weights(attn_s, fixture_group("attns"))
asn <- torch::with_no_grad(attn_s(
  hidden_states = fx$joint_x,
  image_rotary_emb = rope
))
expect_true(max_abs_diff(asn, fx$attns_out) < 1e-5)

# --- flux_double_block -----------------------------------------------------------------

dbl <- flux_double_block(DIM, HEADS, HEAD_DIM)
dbl$eval()
load_named_weights(dbl, fixture_group("dbl"))
db <- torch::with_no_grad(dbl(
  hidden_states = fx$img_x,
  encoder_hidden_states = fx$txt_x,
  temb = fx$temb,
  image_rotary_emb = rope
))
expect_true(max_abs_diff(db[[1]], fx$dbl_enc_out) < 1e-5)
expect_true(max_abs_diff(db[[2]], fx$dbl_hid_out) < 1e-5)

# --- flux_single_block -------------------------------------------------------------------
# The R block takes the pre-concatenated [txt; img] sequence and returns it
# joint; the reference returns the split streams.

sgl <- flux_single_block(DIM, HEADS, HEAD_DIM)
sgl$eval()
load_named_weights(sgl, fixture_group("sgl"))
joint_in <- torch::torch_cat(list(fx$txt_x, fx$img_x), dim = 2L)
sg <- torch::with_no_grad(sgl(
  hidden_states = joint_in,
  temb = fx$temb,
  image_rotary_emb = rope
))
expect_true(max_abs_diff(sg$narrow(2L, 1L, S_TXT), fx$sgl_enc_out) < 1e-5)
expect_true(max_abs_diff(
  sg$narrow(2L, S_TXT + 1L, sg$shape[2] - S_TXT),
  fx$sgl_hid_out
) < 1e-5)

# Full-forward parity test for the Z-Image transformer against a tiny
# random-init diffusers reference model (generated by
# tools/gen_fixtures_zimage_model.py). Includes a strict bidirectional
# key census.

if (!requireNamespace("torch", quietly = TRUE) || !torch::torch_is_installed()) {
  exit_file("torch not fully installed")
}
if (!requireNamespace("safetensors", quietly = TRUE)) {
  exit_file("safetensors not installed")
}

library(diffuseR)

fixture_path <- system.file("tinytest", "fixtures", "zimage_model.safetensors",
  package = "diffuseR")
if (fixture_path == "") fixture_path <- "fixtures/zimage_model.safetensors"
if (!file.exists(fixture_path)) exit_file("zimage model fixtures missing")

fx <- safetensors::safe_load_file(fixture_path, framework = "torch")

max_abs_diff <- function(a, b) {
  as.numeric(torch::torch_max(torch::torch_abs(
    a$to(dtype = torch::torch_float32()) - b$to(dtype = torch::torch_float32())
  )))
}

sd <- fx[grep("^sd\\.", names(fx))]
names(sd) <- sub("^sd\\.", "", names(sd))

model <- zimage_transformer(
  in_channels = 4L, dim = 48L, n_layers = 2L, n_refiner_layers = 1L,
  n_heads = 2L, cap_feat_dim = 24L, axes_dims = c(8L, 8L, 8L)
)

# Strict bidirectional key census: every checkpoint key lands, every
# module parameter fills, and the context refiner carries no adaLN
dests <- c(model$named_parameters(), model$named_buffers())
expect_equal(sort(names(dests)), sort(names(sd)))
expect_false(any(grepl("context_refiner.*adaLN", names(dests))))
expect_true(any(grepl("noise_refiner.*adaLN", names(dests))))

torch::with_no_grad({
  for (name in names(sd)) dests[[name]]$copy_(sd[[name]])
})
model$eval()

out <- torch::with_no_grad(model(fx$x, fx$t, fx$cap))
expect_equal(out$shape, c(4L, 1L, 12L, 20L))
expect_true(max_abs_diff(out, fx$out) < 1e-4)

# Chunked attention path matches
out_chunked <- torch::with_no_grad(model(fx$x, fx$t, fx$cap, chunk_size = 16L))
expect_true(max_abs_diff(out_chunked, out) < 1e-5)

#!/usr/bin/env bash
# Run LTX-2.3 fixture generators against the diffusers reference checkout.
# Usage: tools/gen_fixtures.sh [script.py ...]   (default: all gen_fixtures_*.py)
#
# Uses uv with a CPU-only torch index; never touches the system Python.
set -euo pipefail
cd "$(dirname "$0")/.."

scripts=("$@")
if [ ${#scripts[@]} -eq 0 ]; then
  scripts=(tools/gen_fixtures_*.py)
fi

for s in "${scripts[@]}"; do
  echo "== $s"
  uv run --no-project \
    --index https://download.pytorch.org/whl/cpu \
    --index-strategy unsafe-best-match \
    --with torch \
    --with numpy \
    --with safetensors \
    --with transformers \
    --with huggingface_hub \
    --with packaging \
    --with filelock \
    --with regex \
    --with requests \
    --with tqdm \
    --with pillow \
    python "$s"
done

# Implementation references (LTX-2.3, FLUX.1-schnell)

Every technique in diffuseR's LTX-2.3 and FLUX.1 support traces to a
public, permissively licensed source or to our own measured engineering.
None of it derives from Wan2GP (WanGP Community License) or its `mmgp`
module; this file documents the actual lineage, idea by idea.

## Model architecture and pipeline

| What | Source |
|---|---|
| DiT, VAEs, connectors, vocoder, upsampler module code | Ported from HuggingFace **diffusers** (Apache-2.0): `models/transformers/transformer_ltx2.py`, `models/autoencoders/autoencoder_kl_ltx2.py`, `autoencoder_kl_ltx2_audio.py`, `pipelines/ltx2/{connectors,vocoder,latent_upsampler}.py` |
| Pipeline flow, Euler velocity steps, latent packing, AdaIN filter, tone mapping | diffusers `pipelines/ltx2/pipeline_ltx2.py`, `pipeline_ltx2_latent_upsample.py` (Apache-2.0) |
| Distilled sigma schedules, default negative prompt | diffusers `pipelines/ltx2/utils.py` (Apache-2.0); values originate from Lightricks' LTX-2 release (numeric facts) |
| Single-file checkpoint key layout, embedded config, `model_version` metadata | Format facts of the official Lightricks checkpoints, cross-checked against diffusers `scripts/convert_ltx2_to_diffusers.py` (Apache-2.0) |
| Gemma3 text encoder | Ported from HuggingFace **transformers** (Apache-2.0) |

## FLUX.1-schnell

| What | Source |
|---|---|
| MMDiT transformer (double/single blocks, joint attention, adaLN-Zero variants, RoPE position ids, timestep + pooled-text conditioning) | Ported from HuggingFace **diffusers** (Apache-2.0): `models/transformers/transformer_flux.py`, `models/normalization.py`, `models/embeddings.py` |
| Pipeline flow: prompt encoding contract, sigma schedule (`linspace(1, 1/N, N)`, static shift), 2x2 latent pack/unpack, latent image ids, VAE scale/shift decode | diffusers `pipelines/flux/pipeline_flux.py` (Apache-2.0) |
| FlowMatch Euler scheduler | diffusers `schedulers/scheduling_flow_match_euler_discrete.py` (Apache-2.0); shared with the LTX port |
| 16-channel AutoencoderKL decoder config (no quant convs, scaling 0.3611 / shift 0.1159) | diffusers `scripts/convert_flux_to_diffusers.py` + `convert_sd3_to_diffusers.py` (Apache-2.0) |
| T5-v1.1 encoder (RMS norms, unscaled unbiased attention, shared relative position bias, gated-GELU FFN) | Ported from HuggingFace **transformers** (Apache-2.0): `models/t5/modeling_t5.py` |
| CLIP ViT-L text encoder with quick-GELU and argmax-EOS pooling | HuggingFace **transformers** `models/clip/modeling_clip.py` (Apache-2.0); native module shared with the SD/SDXL port |
| SentencePiece Unigram tokenization (Viterbi best-path over piece log-probs) | Kudo (2018), "Subword Regularization", arXiv:1804.10959; SentencePiece (Apache-2.0); tokenizer.json format facts from HuggingFace tokenizers documentation |
| NF4/fp8 transformer quantization, cast-set policy, phase offloading, allocator tuning | Same sources as the LTX sections below, applied to the FLUX cast set |
| Weights | black-forest-labs/FLUX.1-schnell (Apache-2.0; gated HuggingFace repo, downloaded by the user, never redistributed) |

## FLUX.2 klein

| What | Source |
|---|---|
| MMDiT transformer (shared modulation, SwiGLU feed-forwards, ViT-22B parallel single blocks, 4-axis RoPE) | Ported from HuggingFace **diffusers** (Apache-2.0): `models/transformers/transformer_flux2.py` |
| Klein pipeline flow: Qwen3 prompt encoding contract (chat template, mid-stack hidden states 9/18/27), 4-channel position ids, latent patchify/pack chain, empirical dynamic-shift mu, BatchNorm latent statistics | diffusers `pipelines/flux2/pipeline_flux2_klein.py` (Apache-2.0); the empirical mu formula originates from BFL `sampling.py` (numeric facts) |
| 32-channel `AutoencoderKLFlux2` decode path (post_quant_conv + standard decoder + BN running stats) | diffusers `models/autoencoders/autoencoder_kl_flux2.py` (Apache-2.0); decoder body shared with the SD/FLUX.1 port |
| Qwen3 encoder (GQA, per-head q/k RMS norms, SwiGLU, RoPE theta 1e6, causal + padding mask) | Ported from HuggingFace **transformers** (Apache-2.0): `models/qwen3/` |
| Qwen2 byte-level BPE tokenization (GPT-2 byte table, GPT-4-style split regex, rank-based merges) | GPT-2 BPE (Radford et al. 2019; openai/gpt-2 `encoder.py`, MIT); format facts from HuggingFace tokenizers documentation |
| Weights | black-forest-labs/FLUX.2-klein-4B (Apache-2.0, ungated; downloaded by the user, never redistributed) |

## Z-Image Turbo

| What | Source |
|---|---|
| Single-stream DiT (sandwich RMSNorms, scale/gate-only tanh modulation, noise/context refiner stacks, SwiGLU with separate gate, 3-axis RoPE theta 256, SEQ_MULTI_OF padding with learned pad tokens) | Ported from HuggingFace **diffusers** (Apache-2.0): `models/transformers/transformer_z_image.py` |
| Pipeline flow: Qwen3 prompt encoding contract (thinking-enabled chat template, penultimate hidden state, mask-sliced captions), linspace sigma schedule, reversed-timestep + negated-output convention, static shift | diffusers `pipelines/z_image/pipeline_z_image.py` (Apache-2.0) |
| VAE decode | The FLUX.1 16-channel AutoencoderKL port, unchanged (the checkpoint's VAE config is the flux-dev autoencoder) |
| Qwen3-4B encoder + Qwen2 byte-level BPE tokenizer | Shared with the FLUX.2 klein port (see above); the tokenizer files are byte-identical between the two checkpoints |
| Weights | Tongyi-MAI/Z-Image-Turbo (Apache-2.0, ungated; downloaded by the user, never redistributed) |

## Quantization

| What | Source |
|---|---|
| FP8 e4m3fn weights with per-tensor scales, upcast-in-forward; the exact linear cast set (attention + FFN projections in transformer blocks) | Official Lightricks LTX-2 quantization policy (design facts from `ltx-core/quantization/fp8_cast.py`, LTX-2 Community License — policy observed, no code taken); fp8 storage-with-upcast is also diffusers' documented "layerwise casting" (Apache-2.0) |
| NF4 4-bit format: 16-level NormalFloat quantile code, per-64-block absmax, packed nibbles | **QLoRA**: Dettmers, Pagnoni, Holtzman, Zettlemoyer (2023), "QLoRA: Efficient Finetuning of Quantized LLMs", arXiv:2305.14314; format details as implemented in bitsandbytes (MIT) — reimplemented here in pure torch ops |
| 4-bit weights for consumer-GPU video DiTs as a practice | Community standard: GGUF Q4 checkpoints for LTX/ComfyUI (e.g. city96/ComfyUI-GGUF ecosystem), diffusers bitsandbytes integration docs |

## Conditioning (image-to-video, continuation, audio-driven)

| What | Source |
|---|---|
| Prefix frame conditioning: VAE-encoded pixels as frozen latent tokens, per-token timestep `t * (1 - mask)`, frozen Euler updates, partial-noise option | Ported from diffusers `pipelines/ltx2/pipeline_ltx2_image2video.py` and `pipeline_ltx2_condition.py` (Apache-2.0), restricted to index-0 / strength-1 |
| Audio-driven generation (lip sync): user audio encoded to clean, frozen audio latents; audio stream sees timestep/sigma zero while video denoises attending to it | Our own design — the audio-side mirror of the Apache i2v conditioning pattern above (diffusers has no audio-input conditioning) |
| Audio VAE encoder module | Ported from diffusers `autoencoder_kl_ltx2_audio.py` `LTX2AudioEncoder` (Apache-2.0) |
| 16 kHz log-mel frontend (causal STFT 1024/160, 64 slaney-normed mel bins to 8 kHz) | Parameters are the checkpoint's embedded preprocessing config (numeric facts); STFT/mel constructions are textbook DSP, verified against the checkpoint's stored vocoder bases |

## Memory management

| What | Source |
|---|---|
| Phase-sequential component offloading (each component on the GPU only for its phase) | diffusers memory docs: `enable_model_cpu_offload`, group offloading (`docs/source/en/optimization/memory.md`, Apache-2.0) |
| CPU-resident weights streamed per layer | diffusers leaf-level/group offloading with pinned memory (same doc); long-standing technique (e.g. DeepSpeed ZeRO-Offload, arXiv:2101.06840) |
| Tiled VAE decode (overlapping tiles, crossfaded seams; spatial + temporal) | Direct port of diffusers `AutoencoderKLLTX2Video.tiled_decode` / `_temporal_tiled_decode` (Apache-2.0); tiled VAE decoding dates to the original Stable Diffusion AutoencoderKL (`enable_tiling`) |
| Attention query chunking under a memory budget | diffusers attention slicing (`enable_attention_slicing`, Apache-2.0); memory-efficient attention lineage: Rabe & Staats (2021), arXiv:2112.05682 |
| Pinned host memory + non-blocking transfers | PyTorch CUDA semantics documentation |
| `PYTORCH_CUDA_ALLOC_CONF=expandable_segments` | PyTorch CUDA caching-allocator documentation |
| R-level allocator GC tuning (`torch.cuda_allocator_reserved_rate`, `torch.threshold_call_gc`) | mlverse torch memory-management article (`vignettes/articles/memory-management.Rmd`); A/B methodology and measurements from our own whisper/chatterbox work (both public cornball-ai packages) |
| Persistent scratch buffers for dequantization and attention temporaries | Our own engineering, driven by R's lazy garbage collection of tensor handles (documented in the mlverse article above) |
| Per-block `gc()` for streamed-weight temporaries | Our own engineering (see whisper/chatterbox inference loops) |
| TorchScript-compiled block stack (`torch::jit_compile`, weights as `List[Tensor]`, fused `scaled_dot_product_attention`) | mlverse torch JIT API; the JIT-decode pattern from our own whisper (`decode_jit.R`) and chatterbox (`t3_jit.R`) work, both public cornball-ai packages |

## Measured on our hardware (RTX 5060 Ti 16GB)

Resolution caps, VRAM peaks, profile thresholds, and the GC A/B results
in `tasks/todo.md` and commit messages are our own measurements of the
above techniques, not third-party numbers.

<!--
%\VignetteIndexEntry{Performance Levers}
%\VignetteEngine{simplermarkdown::mdweave_to_html}
%\VignetteEncoding{UTF-8}
-->
---
title: "Performance Levers"
---

# Performance Levers

diffuseR runs 20B-parameter models on 16 GB consumer GPUs, and most
models on no GPU at all. It does that with three independent levers:
weight precision, per-component device placement, and memory residency.
This vignette is the map. The machine-readable version of the same
policy is `recommend()`, which inspects your VRAM, host RAM, and
installed `safetensors` capabilities and returns a configuration.
Today `flux_memory_profile()` delegates to it and `serve()` consults
it to pick between built LTX artifacts; the FLUX-family loaders resolve
their `pin` (and FLUX.1 its `text_device`) through it. For the rest of
the fields it is advisory — call it and pass them to the loaders
yourself.

## Lever 1: weight precision

The ladder that exists in code: fp32 → bf16/fp16 → fp8 (e4m3fn) → nf4.
Quantization applies to the big diffusion transformers only — VAEs,
vocoders, connectors, and modulation/embedding layers always stay at
16/32 bits (they are small and precision-sensitive; each quantizer
carries an exact census of which weights it may touch).

| model | DiT / UNet | text encoder(s) | VAE |
|---|---|---|---|
| FLUX.1 (12B) | nf4, fp8 (streamed), bf16, fp32 | T5: bf16 (GPU, 14 GB+) or fp32 (CPU) · CLIP-L: fp16/fp32 | 16/32 |
| FLUX.2 klein (4B) | nf4, fp8 (resident), bf16, fp32 | Qwen3-4B: bf16/fp32 | 16/32 |
| Z-Image (6B) | nf4, fp8 (resident), bf16, fp32 | Qwen3-4B: bf16/fp32 | 16/32 |
| LTX-2.3 (22B video) | nf4, fp8 (streamed), bf16, fp32 | Gemma3-12B: nf4, bf16, fp32 | 16/32 |
| SD 2.1 / SDXL | fp16, fp32 | CLIP: fp16/fp32 | 16/32 |

Two readability rules govern the ladder:

- **nf4 always loads.** Its artifacts are packed uint8 plus float32
  scale blocks in sub-2 GB shards — every `safetensors` build reads
  them. It is the default tier for the quantized families.
- **fp8 needs a capable `safetensors`.** The float8 dtypes are not yet
  readable by the CRAN `safetensors`; `recommend()` probes the
  installed build and, when a card could run fp8 but the reader
  cannot, recommends nf4 and surfaces the suggestion in `$note`
  (never an error).

The SD-family models ship no quantized weights: their floor is fp16,
and what varies across VRAM is placement, not precision.

## Lever 2: device placement

Every component takes its own device. The SD family uses explicit
device maps (`auto_devices()` strategies: `full_gpu`, `unet_gpu`,
`cpu_only`); the flux family and LTX use phase offloading, where each
component holds the GPU only for its own phase — text encoding,
denoising, decoding — and the denoiser is the sole GPU tenant during
the loop. Text encoders earn special placement: FLUX.1's T5
phase-onloads in bf16 on 14 GB+ cards (its ~9.8 GB encode phase fits)
and runs fp32 on the CPU below that; the Qwen3 encoders phase-onload in
bf16; the Gemma3 encoder can be GPU-resident or CPU-resident with a
staged swap (see below).

## Lever 3: residency

From most to least VRAM:

1. **Fully GPU-resident** — everything fits, nothing moves (large
   cards only).
2. **Resident quantized DiT, phase-swapped everything else** — the
   16 GB sweet spot: LTX nf4 and the fp8 FLUX.2/Z-Image
   configurations keep the DiT on the card while text encoders swap
   per phase.
3. **Pinned phase swap** — component weights live in page-locked
   ("pinned") host RAM and move to the GPU for their phase at DMA
   rate. Measured on a PCIe 5.0 x8 card: 25 GB/s host-to-device
   (11 GB of transformer re-onloads in 0.5 s), and the offload is a
   pointer swap back to the still-valid pinned copy — zero bytes
   moved, because inference never mutates weights. Page-locking costs
   ~0.6 s/GB once at load. Built for the LTX pipeline components, the
   Gemma3 encoder, and the FLUX-family transformer, VAE decoder, and
   text encoder(s).
4. **Pageable phase swap** — the same movement through ordinary
   memory, at roughly 2-16 GB/s depending on tensor layout. The
   fallback when pinning is opted out or page-locking fails.
5. **Streamed weights** — the bigger-than-VRAM tier: weights stay in
   (pinned) host RAM permanently and stream across PCIe during each
   forward pass, about one byte per parameter per step. LTX fp8 and
   FLUX.1 fp8 run this way.
6. **CPU-only** — every model runs without a GPU, at its floor
   precision.

There is no disk tier at inference time: weights load from disk once,
and the closest thing to "swap to disk" is *unpinned* host memory
being paged out by the OS — which is exactly the trade `recommend()`
weighs.

## Pinning: when and when not

Pinned pages are unswappable — they subtract from what the OS can page
out, so on small-RAM machines they convert memory pressure into
process kills rather than slowdowns. `recommend()` therefore returns
`pin = TRUE` only when available host RAM covers the model's estimated
pinned set twice over, `FALSE` on the CPU tier (nothing stages), and
`TRUE` when RAM cannot be detected, because page-locking already fails
soft per component. The global switch is
`options(diffuseR.pin_staging = FALSE)` — reach for it under host
memory pressure, in containers with hard memory caps, or for
single-generation sessions where the one-time page-lock never pays
itself back. The LTX pipeline, the Gemma3 loaders, and the FLUX-family
image loaders consume the decision (their `pin` argument defaults to
it); the SD-family loaders place components statically, so pinning is
inert for them.

## Putting it together

```r
r <- recommend("ltx")     # or "flux1", "flux2", "zimage", "sdxl", "sd21"
r$precision               # tier the card + safetensors support
r$devices                 # per-component placement
r$pin                     # page-lock the phase-swapped host copies?
r$note                    # fork suggestion when fp8 wanted but unreadable
```

Treat the result as the machine's advice: pass its fields to the
loaders and generators. The FLUX-family loaders' `pin` (and FLUX.1's
`text_device`), `flux_memory_profile()`, and `serve()`'s LTX artifact
selection consume it automatically.

% tinyrox says don't edit this manually, but it can't stop you!
\name{txt2img_zimage}
\alias{txt2img_zimage}
\title{Generate an image with Z-Image-Turbo}
\usage{
txt2img_zimage(prompt, pipeline = NULL, width = 1024L, height = 1024L,
               num_inference_steps = 8L, max_sequence_length = 512L,
               seed = NULL, prompt_embeds = NULL, save_file = TRUE,
               filename = NULL, verbose = TRUE, ...)
}
\arguments{
\item{prompt}{Character. The prompt.}

\item{pipeline}{A \code{zimage_pipeline} from
\code{\link{zimage_load_pipeline}}; NULL loads one (passing
\code{...} through).}

\item{num_inference_steps}{Integer. Denoising steps (Turbo: 8).}

\item{max_sequence_length}{Integer. Qwen3 token length (512).}

\item{seed}{Integer or NULL. Latents are drawn on the CPU, so a seed
matches a Python diffusers run with a CPU generator.}

\item{prompt_embeds}{Optional precomputed [L, 2560] caption
embeddings (valid tokens only).}

\item{save_file}{Logical. Write a PNG.}

\item{filename}{Output path (default derived from the prompt).}

\item{verbose}{Logical, or one of "silent", "progress", "steps".
TRUE = "steps" (full per-phase chatter), FALSE = "silent".
"progress" prints a one-line generation summary plus a denoise
progress bar (interactive) or periodic step ticks (captured logs).}

\item{...}{Passed to \code{\link{zimage_load_pipeline}} when
\code{pipeline} is NULL.}

\item{width,height}{Integers, divisible by 16.}
}
\value{
Invisibly, \code{list(image, metadata)} where \code{image} is
  an [H, W, 3] array in [0, 1].
}
\description{
Guidance-distilled text-to-image (8 steps, no CFG): Qwen3-4B prompt
encoding (thinking-enabled chat template, penultimate hidden state),
FlowMatch denoising with the reversed-timestep convention, and
16-channel VAE decode. Strong at legible text rendering, English and
Chinese both.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_text_encoder_safetensors}
\alias{load_text_encoder_safetensors}
\title{Load HF safetensors weights into the native CLIP text encoder}
\usage{
load_text_encoder_safetensors(native_encoder, path, verbose = TRUE)
}
\arguments{
\item{native_encoder}{Native text encoder module}

\item{path}{Path to model.safetensors (or a directory containing it)}

\item{verbose}{Print loading progress}
}
\value{
The native encoder with loaded weights (invisibly)
}
\description{
Loads a HuggingFace CLIPTextModel \code{model.safetensors} (e.g.
FLUX.1-schnell's \code{text_encoder} or SDXL's \code{text_encoder})
into \code{\link{text_encoder_native}}, reusing the TorchScript key
remaps minus the export prefixes.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vram}
\alias{vram}
\title{VRAM Detection and Management Utilities}
\description{
Device detection, VRAM reporting, and module offloading helpers shared
by the image and video pipelines.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_pad_len}
\alias{zimage_pad_len}
\title{Padding length to the next multiple of 32}
\usage{
zimage_pad_len(n)
}
\arguments{
\item{n}{Integer token count.}
}
\value{
Integer pad length in [0, 31].
}
\description{
Padding length to the next multiple of 32
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.staged_onload}
\alias{.staged_onload}
\title{Move a pinned component onto the compute device}
\usage{
.staged_onload(staging, device)
}
\description{
Non-blocking copies from pinned memory share the default stream,
so later kernels are ordered after them; no explicit sync needed.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_is_fp8_cast_key}
\alias{ltx23_is_fp8_cast_key}
\title{Test whether a mapped DiT key is in the official fp8 cast set}
\usage{
ltx23_is_fp8_cast_key(mapped_key)
}
\arguments{
\item{mapped_key}{Character vector of mapped (diffusers-style)
parameter names.}
}
\value{
Logical vector.
}
\description{
Test whether a mapped DiT key is in the official fp8 cast set
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{save_video_ffmpeg}
\alias{save_video_ffmpeg}
\title{Save Video using FFmpeg}
\usage{
save_video_ffmpeg(video, file, fps = 24, format = "mp4", quality = 85,
                  verbose = TRUE)
}
\arguments{
\item{video}{Array of video frames [T, H, W, C].}

\item{file}{Output file path.}

\item{fps}{Frames per second.}

\item{format}{Output format.}

\item{quality}{Quality level 1-100.}

\item{verbose}{Logical.}
}
\description{
Save Video using FFmpeg
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_prepare_text_ids}
\alias{flux2_prepare_text_ids}
\title{Build FLUX.2 text position ids}
\usage{
flux2_prepare_text_ids(len, device = "cpu")
}
\arguments{
\item{len}{Integer. Text sequence length.}

\item{device}{Device for the resulting tensor.}
}
\value{
Float tensor [len, 4].
}
\description{
Columns (T, H, W, L) with only L varying: 0..len-1. Reference:
Flux2KleinPipeline._prepare_text_ids.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{text_encoder2_native_from_safetensors}
\alias{text_encoder2_native_from_safetensors}
\title{Build a native SDXL text encoder 2 from a diffusers safetensors directory}
\usage{
text_encoder2_native_from_safetensors(path, return_penultimate = TRUE,
                                      verbose = TRUE, ...)
}
\arguments{
\item{path}{diffusers text_encoder_2 directory (config.json +
model.safetensors) or the config.json path.}

\item{return_penultimate}{Return the penultimate hidden state for the
SDXL cross-attention embeds (default TRUE); the pooled output is
always computed from the full stack.}

\item{verbose}{Print how many parameters were loaded.}

\item{...}{Overrides for \code{\link{text_encoder2_native}} args.}
}
\value{
The native text encoder 2 in eval mode.
}
\description{
Reads the CLIPTextConfig from \code{<dir>/config.json}, constructs
\code{\link{text_encoder2_native}} to match, and loads
\code{model.safetensors} - the safetensors counterpart to the
TorchScript \code{text_encoder_2} path (no TorchScript, Blackwell-safe).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{txt2vid_ltx2}
\alias{txt2vid_ltx2}
\title{Generate video (and audio) with LTX-2.3}
\usage{
txt2vid_ltx2(prompt, pipeline, text_encoder = NULL, tokenizer = NULL,
             prompt_embeds = NULL, connector_embeds = NULL, width = 768L,
             height = 512L, num_frames = 121L, frame_rate = 24,
             sigmas = ltx23_distilled_sigmas(), guidance_scale = 1,
             seed = NULL, device = "cuda", dtype = "bfloat16", filename = NULL,
             max_sequence_length = 1024L, decode_video = TRUE,
             decode_audio = TRUE, two_stage = FALSE, upsampler = NULL,
             adain_factor = 1, tone_map_compression = 0, phase_offload = TRUE,
             image = NULL, condition_video = NULL, conditioning_frames = 9L,
             cond_noise_scale = 0, condition_latents = NULL,
             resident = character(), trim_frames = 0L, audio = NULL,
             verbose = TRUE)
}
\arguments{
\item{prompt}{Character. The prompt.}

\item{pipeline}{An \code{ltx23_pipeline} from
\code{\link{ltx23_load_pipeline}}.}

\item{prompt_embeds}{Optional precomputed list with
\code{prompt_embeds} (raw stacked Gemma3 states) and
\code{prompt_attention_mask}; bypasses the text encoder.}

\item{connector_embeds}{Optional precomputed text-connector outputs:
a list with \code{video_text_embedding},
\code{audio_text_embedding}, and \code{attention_mask} (the result
of \code{pipeline$connectors} on the Gemma3 states). The prompt is
constant across a chained track, so compute this once and pass it
to every chunk: it skips the per-call connectors phase, whose
GPU-side handling of the raw hidden-state stack does not fit next
to a resident transformer.}

\item{num_frames}{Integer. 8k + 1 frames (e.g. 121).}

\item{frame_rate}{Numeric. Frames per second.}

\item{sigmas}{Numeric vector. Denoising schedule (default: official
distilled schedule; must end in 0).}

\item{guidance_scale}{Numeric. Only 1 (no CFG) is supported; the
distilled checkpoints are trained for CFG-free sampling.}

\item{seed}{Integer or NULL.}

\item{device}{Character. Compute device for the denoising loop.}

\item{dtype}{Character. Model compute dtype ("bfloat16" or "float32").}

\item{filename}{Character or NULL. Output video path (.mp4). Audio is
muxed in when the av package is available.}

\item{max_sequence_length}{Integer. Text token length (multiple of 128).}

\item{two_stage}{Logical. Generate at half resolution, upsample the
latents 2x spatially, and refine over the stage-2 schedule
(resolution must then be a multiple of 64; requires
\code{upsampler}).}

\item{upsampler}{An \code{\link{ltx23_latent_upsampler}} (see
\code{\link{ltx23_load_upsampler}}).}

\item{adain_factor}{Numeric. AdaIN blend of the upsampled latents
toward the stage-1 statistics (0 disables).}

\item{tone_map_compression}{Numeric in [0, 1]. Optional latent tone
mapping before stage 2.}

\item{phase_offload}{Logical. Move each small component to the compute
device only for its phase (text encoding, upsampling, decoding) and
back to the CPU afterwards, keeping the denoise phase as the sole
GPU tenant.}

\item{image}{Optional start image for image-to-video: a PNG/JPEG path
or an [H, W, 3] array in [0, 1]. The image conditions the first
frame; the rest of the video is generated (reference i2v).}

\item{condition_video}{Optional continuation source: a video path
(its trailing \code{conditioning_frames} frames are used) or an
[F, H, W, 3] array. The clip's tail becomes the frozen prefix of
the new video, so the output's first \code{conditioning_frames}
frames overlap the source (trim or crossfade when concatenating).}

\item{conditioning_frames}{Integer. Trailing pixel frames taken from
\code{condition_video} (8k + 1, default 9 = 2 latent frames).}

\item{cond_noise_scale}{Numeric in [0, 1]. Optional partial noising
of the conditioned tokens (0 = keep them exactly).}

\item{condition_latents}{Optional continuation source already in
latent space: normalized video latents [1, 128, k, height/32,
width/32] (e.g. \code{\link{ltx23_tail_latents}} on a previous
result), used directly as the frozen prefix with no VAE encode.
Mutually exclusive with \code{image} and \code{condition_video}.}

\item{resident}{Character vector of pipeline component names
("transformer", "vae", "audio_vae", "connectors", "vocoder") to
keep on the compute device after their phase instead of
offloading, for callers running several generations back to back
(chained chunks). Components already on the device are not
re-copied on later calls.}

\item{trim_frames}{Integer. Drop this many leading pixel frames
from the decoded video (and the saved file), e.g. the
conditioning-head overlap of a continuation. The returned
\code{latents} keep the full sequence (tail slicing for chaining
needs it). Audio is muxed exactly as supplied, so drop any head
padding from the conditioning audio when trimming.}

\item{audio}{Optional conditioning audio for audio-driven generation
(lip sync): a file path (decoded via \code{av}) or a matrix
[2, samples] in [-1, 1] at 16 kHz. The audio is encoded into
clean, frozen audio latents that the video attends to while
denoising, and the original samples are muxed into the output
(audio decoding is skipped).}

\item{verbose}{Logical, or one of "silent", "progress", "steps".
TRUE = "steps" (full per-phase chatter, per-step sigma/timing
lines), FALSE = "silent". "progress" prints a one-line generation
summary plus a denoise progress bar (interactive) or periodic
step ticks (captured logs).}

\item{text_encoder,tokenizer}{Gemma3 model and tokenizer (or paths;
see \code{\link{load_gemma3_text_encoder}} and
\code{\link{gemma3_tokenizer}}). Ignored when \code{prompt_embeds}
is supplied.}

\item{width,height}{Integers. Output resolution (multiples of 32).}

\item{decode_video,decode_audio}{Logicals. Decode the respective
latents (disable for latent-space work).}
}
\value{
Invisibly, a list with \code{video} (array
  [frames, height, width, 3] in [0, 1]), \code{audio} (matrix
  [2, samples] in [-1, 1]), \code{sample_rate}, the raw
  \code{latents} and \code{audio_latents}, and \code{latent_shape}
  (c(frames, height, width) of the latent geometry, for
  \code{\link{ltx23_tail_latents}}).
}
\description{
Distilled text-to-video generation: encodes the prompt with Gemma3 +
connectors, denoises joint audio/video latents over the official
8-step distilled schedule (no classifier-free guidance), decodes the
video with the causal VAE and the audio with the audio VAE + BWE
vocoder, and optionally muxes both into an MP4.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{rotate_half}
\alias{rotate_half}
\title{Rotate half of the hidden dims}
\usage{
rotate_half(x)
}
\description{
Rotate half of the hidden dims
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.ltx23_release_vae_traces}
\alias{.ltx23_release_vae_traces}
\title{Release all cached decode traces}
\usage{
.ltx23_release_vae_traces()
}
\value{
Invisibly, NULL.
}
\description{
Traces hold references to the weight tensors they captured; drop
them when a component leaves the GPU so its memory actually frees.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_nf4_dequantize}
\alias{ltx23_nf4_dequantize}
\title{Dequantize NF4 data to a float tensor}
\usage{
ltx23_nf4_dequantize(packed, absmax, shape, dtype = torch::torch_bfloat16(),
                     chunk_elements = 8388608L, out = NULL)
}
\arguments{
\item{packed}{uint8 tensor of packed index pairs.}

\item{absmax}{float32 tensor of per-block scales.}

\item{shape}{Integer vector. Original tensor shape.}

\item{dtype}{Target torch dtype.}

\item{chunk_elements}{Integer. Elements dequantized per slice (bounds
the int64 index temporary).}

\item{out}{Optional preallocated tensor of \code{shape} to write into
(avoids allocating a fresh weight tensor per call).}
}
\value{
Tensor of \code{shape} in \code{dtype} on the input's device.
}
\description{
Dequantize NF4 data to a float tensor
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_audio_vae}
\alias{ltx23_audio_vae}
\title{LTX-2.3 audio VAE}
\usage{
ltx23_audio_vae(base_channels = 128L, output_channels = 2L,
                num_res_blocks = 2L, latent_channels = 8L,
                ch_mult = c(1L, 2L, 4L), causality_axis = "height",
                mel_bins = 64L, in_channels = 2L)
}
\arguments{
\item{in_channels}{Integer. Mel input channels (2 = stereo).}

\item{base_channels,output_channels,num_res_blocks,latent_channels,ch_mult,causality_axis,mel_bins}{See \code{\link{ltx23_audio_decoder}}.}
}
\value{
Module bundling the audio encoder and decoder. Its
  forward(z) is \code{decode(z)}, returning the mel spectrogram for a
  latent; \code{$encode()} and \code{$decode()} are callable
  separately.
}
\description{
Encoder + decoder plus the per-channel latent statistics loaded from
the checkpoint. Encoding is used for audio-conditioned generation
(lip sync); decoding for generated audio.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{detect_unet_architecture}
\alias{detect_unet_architecture}
\title{Detect UNet architecture from TorchScript file}
\usage{
detect_unet_architecture(torchscript_path)
}
\arguments{
\item{torchscript_path}{Path to TorchScript UNet .pt file}
}
\value{
List with architecture parameters
}
\description{
Detect UNet architecture from TorchScript file
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{sdxl_pipeline_safetensors}
\alias{sdxl_pipeline_safetensors}
\title{Native SDXL pipeline from diffusers safetensors}
\description{
Assemble and run the native SDXL pipeline directly from a HuggingFace
diffusers directory (\code{unet/}, \code{vae/}, \code{text_encoder/},
\code{text_encoder_2/}), with no TorchScript \code{.pt} step - so it
works on Blackwell and loads the same weights everyone else uses. The
counterpart to \code{\link{sd_pipeline_from_safetensors}}, adding the
second text encoder (OpenCLIP ViT-bigG) and the added-conditioning
embeddings (pooled \code{text_embeds} + \code{time_ids}) SDXL needs.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_rotary_pos_embed_1d}
\alias{ltx23_rotary_pos_embed_1d}
\title{1D rotary embeddings for the text connectors}
\usage{
ltx23_rotary_pos_embed_1d(dim, base_seq_len = 4096L, theta = 10000,
                          double_precision = TRUE, rope_type = "split",
                          num_attention_heads = 32L)
}
\arguments{
\item{dim}{Integer. Rotary dimension (connector inner dim).}

\item{base_seq_len}{Integer. Base sequence length for normalization.}

\item{theta}{Numeric. RoPE theta.}

\item{double_precision}{Logical. Compute base frequencies in float64.}

\item{rope_type}{"split" (LTX-2.3) or "interleaved".}

\item{num_attention_heads}{Integer. For the split per-head layout.}
}
\value{
Module whose forward(batch_size, pos, device) returns
  \code{list(cos_freqs, sin_freqs)}, the 1-D rotary tables for a
  sequence of length \code{pos}.
}
\description{
1D rotary embeddings for the text connectors
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{dit_flux2_modules}
\alias{dit_flux2_modules}
\title{FLUX.2 Transformer Building Blocks}
\description{
Fresh R port of the FLUX.2 MMDiT blocks from the diffusers reference
implementation (Apache-2.0,
src/diffusers/models/transformers/transformer_flux2.py). Key
differences from FLUX.1: modulation is computed ONCE at model level
by shared \code{flux2_modulation} projections and passed into the
blocks (block norms are parameterless), feed-forwards use SwiGLU with
the gate fused into \code{linear_in}, the single-stream block is a
ViT-22B-style parallel block with fully fused projections, and every
linear is bias-free. Module field names mirror the diffusers
state-dict keys 1:1. Reuses \code{flux_attention} (bias = FALSE),
\code{ltx23_rms_norm}, \code{.ltx23_sdpa}, and
\code{flux_apply_rotary_emb}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_vocoder}
\alias{ltx23_vocoder}
\title{LTX-2.3 vocoder stage}
\usage{
ltx23_vocoder(in_channels = 128L, hidden_channels = 1536L, out_channels = 2L,
              upsample_kernel_sizes = c(11L, 4L, 4L, 4L, 4L, 4L),
              upsample_factors = c(5L, 2L, 2L, 2L, 2L, 2L),
              resnet_kernel_sizes = c(3L, 7L, 11L),
              resnet_dilations = list(c(1L, 3L, 5L), c(1L, 3L, 5L), c(1L, 3L, 5L)),
              antialias_ratio = 2L, antialias_kernel_size = 12L,
              final_bias = FALSE)
}
\arguments{
\item{in_channels}{Integer. Flattened input channels (C * mel bins / 1).}

\item{hidden_channels}{Integer.}

\item{out_channels}{Integer.}

\item{resnet_kernel_sizes}{Integer vector.}

\item{resnet_dilations}{List of integer vectors.}

\item{final_bias}{Logical.}

\item{upsample_kernel_sizes,upsample_factors}{Integer vectors.}

\item{antialias_ratio,antialias_kernel_size}{Integers.}
}
\value{
Module whose forward(hidden_states, time_last) returns the
  synthesized waveform [B, 1, samples] for a mel spectrogram.
}
\description{
Mel spectrogram [B, C, T, M] -> waveform [B, out_channels, samples].
Channel and mel dims are flattened into conv channels; each upsample
stage halves the channel count and averages three parallel ResNet
branches.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{checkpoint_ltx23}
\alias{checkpoint_ltx23}
\title{LTX-2.3 Single-File Checkpoint Reader}
\description{
LTX 2.3 checkpoints ship as one safetensors file containing every
component (transformer, connectors, video VAE, audio VAE, vocoder),
with the model version and full component configuration embedded in
the safetensors metadata. These helpers open the file, validate the
version, split the key space by component, and stream tensors into
R torch modules one at a time so the 46 GB file is never fully
materialized in memory.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_vocoder_resblock}
\alias{ltx23_vocoder_resblock}
\title{Vocoder ResNet block (AMP)}
\usage{
ltx23_vocoder_resblock(channels, kernel_size = 3L, dilations = c(1L, 3L, 5L),
                       antialias_ratio = 2L, antialias_kernel_size = 12L)
}
\arguments{
\item{channels}{Integer.}

\item{kernel_size}{Integer.}

\item{dilations}{Integer vector.}

\item{antialias_ratio,antialias_kernel_size}{Integers.}
}
\value{
Module whose forward(x) returns \code{x} after the dilated
  convolution pairs have been added back as residuals, a tensor of
  the same shape.
}
\description{
Dilated conv pairs, each preceded by an anti-aliased SnakeBeta
activation, with residual connections.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{checkpoint_flux}
\alias{checkpoint_flux}
\title{FLUX Checkpoint Readers}
\description{
FLUX transformers ship in the diffusers layout: a directory with
\code{config.json}, one or more \code{diffusion_pytorch_model*.safetensors}
shards, and (when sharded) a
\code{diffusion_pytorch_model.safetensors.index.json} weight map.
These helpers open that layout behind the same checkpoint interface as
\code{\link{ltx23_open_checkpoint}}, so the LTX group loaders and
quantization machinery work unchanged. FLUX module names mirror the
checkpoint keys 1:1 - no key mapping is needed.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{resident_status}
\alias{resident_status}
\title{Status of a resident handle}
\usage{
resident_status(res)
}
\arguments{
\item{res}{A \code{diffuseR_resident} handle.}
}
\value{
A list with \code{model}, \code{state}, \code{device},
  \code{components} (character vector of pinned component names),
  \code{pinned_bytes} (page-locked host bytes held),
  \code{gpu_allocated} and \code{gpu_reserved} (bytes the CUDA
  caching allocator reports live and held for this process, NA
  without CUDA), \code{components_on_gpu} (how many components are
  *actually* resident right now), \code{loaded_at}, and
  \code{last_error} (NULL unless a transition failed).

  \code{state} is the handle's claim on the card;
  \code{components_on_gpu} is the measurement. They disagree by
  design after a render on a \code{phase_offload = TRUE} pipeline,
  which returns each component to pinned host memory as its phase
  finishes: the handle stays "active" (it still owns the card's
  budget and can render again without touching disk) while
  \code{components_on_gpu} is 0. Schedule on the measurement.
}
\description{
Status of a resident handle
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.resident_all_on}
\alias{.resident_all_on}
\title{TRUE when every staged tensor sits on the expected device type}
\usage{
.resident_all_on(staging, type)
}
\arguments{
\item{staging}{A named list of staging sets.}

\item{type}{"cpu" or "cuda".}
}
\value{
Logical.
}
\description{
TRUE when every staged tensor sits on the expected device type
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{UNetCrossAttention}
\alias{UNetCrossAttention}
\title{Cross-Attention for UNet}
\usage{
UNetCrossAttention(query_dim, context_dim = NULL, heads = 8L, dim_head = 64L)
}
\description{
Cross-Attention for UNet
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{timestep_embedding}
\alias{timestep_embedding}
\title{Sinusoidal Timestep Embedding}
\usage{
timestep_embedding(timesteps, dim, flip_sin_to_cos = TRUE,
                   downscale_freq_shift = 0L)
}
\arguments{
\item{timesteps}{Tensor of timesteps (batch_size,)}

\item{dim}{Embedding dimension}

\item{flip_sin_to_cos}{If TRUE, output [cos, sin] instead of [sin, cos].
SDXL uses TRUE (default), SD21 uses FALSE.}

\item{downscale_freq_shift}{Frequency shift parameter. SDXL uses 0 (default),
SD21 uses 1. With 0: exponent = log(10000) / half_dim. With 1: exponent = log(10000) / (half_dim - 1).}
}
\value{
Tensor (batch_size, dim)
}
\description{
Sinusoidal Timestep Embedding
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_prepare_conditioned_latents}
\alias{ltx23_prepare_conditioned_latents}
\title{Build conditioned initial latents and the conditioning mask}
\usage{
ltx23_prepare_conditioned_latents(cond_latents, latent_frames, latent_height,
                                  latent_width, noise, cond_noise_scale = 0)
}
\arguments{
\item{cond_latents}{Normalized condition latents
[1, 128, k, H', W'] from \code{\link{ltx23_encode_video_frames}}.}

\item{noise}{Tensor [1, 128, F', H', W'] of standard noise (caller
provides so seeding stays in one place).}

\item{cond_noise_scale}{Numeric. Optional partial noising of the
conditioned tokens (diffusers \code{noise_scale}, default 0).}

\item{latent_frames,latent_height,latent_width}{Integers. Full
latent geometry of the generation.}
}
\value{
list(latents [1, S, 128] float32 packed,
  conditioning_mask [1, S] float32 packed).
}
\description{
i2v (\code{cond_latents} has one latent frame): the encoded frame is
repeated across all latent frames and only latent frame 0 is marked
conditioned. Continuation (k latent frames): the prefix tokens are
replaced and marked. Unconditioned positions start as pure noise.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{VAEAttentionBlock}
\alias{VAEAttentionBlock}
\title{VAE Attention Block}
\usage{
VAEAttentionBlock(channels, norm_groups = 32)
}
\arguments{
\item{channels}{Number of channels}
}
\description{
Self-attention for VAE mid block
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{VAEResnetBlock}
\alias{VAEResnetBlock}
\title{VAE ResNet Block}
\usage{
VAEResnetBlock(in_channels, out_channels, norm_groups = 32)
}
\arguments{
\item{in_channels}{Input channels}

\item{out_channels}{Output channels}
}
\description{
VAE ResNet Block
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unet_sdxl_native_from_torchscript}
\alias{unet_sdxl_native_from_torchscript}
\title{Create native SDXL UNet from TorchScript}
\usage{
unet_sdxl_native_from_torchscript(torchscript_path, verbose = TRUE)
}
\arguments{
\item{torchscript_path}{Path to TorchScript SDXL UNet .pt file}

\item{verbose}{Print loading progress}
}
\value{
A native SDXL UNet module with loaded weights
}
\description{
Create native SDXL UNet from TorchScript
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{rope_flux2}
\alias{rope_flux2}
\title{FLUX.2 Position Ids and Empirical Shift}
\description{
Fresh R port of the FLUX.2 position-id builders and the empirical
timestep-shift formula from the diffusers reference (Apache-2.0,
src/diffusers/pipelines/flux2/pipeline_flux2_klein.py). FLUX.2 uses
4-axis rotary position ids (T, H, W, L): text tokens carry only the
L axis (sequence position), image latents carry H and W, and the T
axis distinguishes reference images (unused for txt2img). Frequencies
come from \code{\link{flux_pos_embed}} with
\code{axes_dim = c(32, 32, 32, 32)} and \code{theta = 2000}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_sd21}
\alias{download_sd21}
\title{Download the Stable Diffusion 2.1 diffusers weights}
\usage{
download_sd21(verbose = TRUE)
}
\arguments{
\item{verbose}{Logical.}
}
\value{
Invisibly, the diffusers directory (the parent of
  \code{unet/}, \code{vae/}, \code{text_encoder/}).
}
\description{
Fetches the UNet, VAE, and CLIP text encoder from the
\code{cornball-ai/sd21-R} HuggingFace dataset (fp16 diffusers
safetensors, converted from the original OpenRAIL weights; the
upstream \code{stabilityai} repo was deprecated). About 2.5 GB,
one-time. The native tokenizer and DDIM scheduler need no downloads.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_encode_batch}
\alias{gemma3_encode_batch}
\title{Batch-encode prompts with Gemma3, cached to disk}
\usage{
gemma3_encode_batch(prompts, model = NULL, tokenizer = NULL, batch_size = 4L,
                    cache_dir = NULL, max_sequence_length = 1024L,
                    device = "cuda", verbose = TRUE)
}
\arguments{
\item{prompts}{Character vector.}

\item{batch_size}{Integer. Prompts per forward pass (default 4;
raise on cards with headroom, lower if the encode OOMs).}

\item{cache_dir}{Optional directory. When given, each prompt is
written to \code{gemma3-<md5>.pt} and the return value is the
character vector of those paths (load one with
\code{torch::torch_load}); when NULL, the return value is a list
of \code{list(prompt_embeds, prompt_attention_mask)} in prompt
order.}

\item{model,tokenizer}{As in \code{\link{encode_with_gemma3}}.}

\item{max_sequence_length,device,verbose}{As in
\code{\link{encode_with_gemma3}}.}
}
\value{
Character vector of cache paths (with \code{cache_dir}) or
  a list of per-prompt embedding results (without).
}
\description{
Encodes a vector of prompts in sub-batches (bounding activation
VRAM) and optionally caches each prompt's result under
\code{cache_dir}, keyed by the prompt text and sequence length.
Already-cached prompts are skipped, so an interrupted batch resumes
where it stopped. Embeddings land on the CPU either way; the
renderer moves them per phase.
}
\details{
Budget note: each prompt's result is the full hidden-state stack the
LTX connectors consume - roughly 0.4 GB at 1024 tokens - so caching
N prompts needs ~0.4N GB of disk.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_flux}
\alias{download_flux}
\title{Download and Prepare FLUX.1-schnell Weights}
\description{
Downloads FLUX.1-schnell from HuggingFace (weights Apache-2.0, but
the repo is gated behind a license click-through) and quantizes the
12B transformer to a local NF4 (~7 GB) or fp8 (~12 GB) artifact.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{SpatialTransformer}
\alias{SpatialTransformer}
\title{Spatial Transformer (Attention Block)}
\usage{
SpatialTransformer(in_channels, n_heads, d_head, depth = 1L, context_dim = NULL)
}
\description{
Spatial Transformer (Attention Block)
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{quantize_flux}
\alias{quantize_flux}
\title{FLUX Transformer Quantization and Loading}
\description{
Quantize the 12B FLUX transformer to NF4 (~7 GB, GPU-resident on
16 GB cards) or fp8 (~12 GB, CPU-resident and streamed per forward),
and load any format back into \code{\link{flux_transformer}}. Reuses
the LTX-2.3 quantization machinery (\code{ltx23_nf4_quantize},
\code{ltx23_nf4_linear}, \code{ltx23_fp8_linear}); only the cast set
and the diffusers directory layout are FLUX-specific.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{preprocess_image}
\alias{preprocess_image}
\title{Preprocess image for Stable Diffusion}
\usage{
preprocess_image(input, device = "cpu", width = 512, height = 512)
}
\arguments{
\item{input}{File path to .jpg or .png, or a 3D array}

\item{device}{Target device for torch ("cpu" or "cuda")}

\item{width}{Desired width of the output image}

\item{height}{Desired height of the output image}
}
\value{
Torch tensor of shape c(1, 3, 512, 512), scaled to c(-1, 1)
}
\description{
Preprocess image for Stable Diffusion
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_block}
\alias{zimage_block}
\title{Z-Image transformer block}
\usage{
zimage_block(dim, n_heads, norm_eps = 1e-05, modulation = TRUE)
}
\arguments{
\item{dim}{Integer. Model width.}

\item{n_heads}{Integer. Attention heads; head dim is dim / n_heads.}

\item{norm_eps}{Numeric. RMSNorm epsilon. Default 1e-5.}

\item{modulation}{Logical. Whether the block is timestep-modulated.}
}
\value{
Module whose forward(x, freqs, adaln_input, chunk_size)
  returns the residual block output, a tensor of the same shape as
  \code{x}. \code{adaln_input} is used only when the block was built
  with \code{modulation = TRUE}.
}
\description{
Sandwich-norm residual block shared by the noise refiner, the context
refiner and the main trunk. With \code{modulation = TRUE} the block
carries an adaLN linear producing (scale_msa, gate_msa, scale_mlp,
gate_mlp); the context refiner uses \code{modulation = FALSE} and has
no adaLN weights at all.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_unpatchify_latents}
\alias{flux2_unpatchify_latents}
\title{Unpatchify FLUX.2 latents (channels -> 2x2)}
\usage{
flux2_unpatchify_latents(latents)
}
\arguments{
\item{latents}{Tensor [B, 4C, H, W].}
}
\value{
Tensor [B, C, H * 2, W * 2].
}
\description{
Inverse of \code{\link{flux2_patchify_latents}}. Reference:
Flux2KleinPipeline._unpatchify_latents.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{decode_bpe}
\alias{decode_bpe}
\title{Decode token IDs to text}
\usage{
decode_bpe(tokenizer, ids, skip_special_tokens = TRUE)
}
\arguments{
\item{tokenizer}{A bpe_tokenizer object.}

\item{ids}{Integer vector or matrix of token IDs.}

\item{skip_special_tokens}{Logical. Skip special tokens in output.}
}
\value{
Character string or vector.
}
\description{
Decode token IDs to text
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{print.bpe_tokenizer}
\alias{print.bpe_tokenizer}
\title{Print BPE Tokenizer}
\usage{
\method{print}{bpe_tokenizer}(x, ...)
}
\arguments{
\item{x}{A bpe_tokenizer object.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns \code{x}. Called for the side effect of
  printing a summary of the tokenizer to the console.
}
\description{
Print BPE Tokenizer
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_flux1}
\alias{download_flux1}
\title{Download FLUX.1-schnell and build the quantized artifact}
\usage{
download_flux1(quantize = TRUE, precision = c("nf4", "fp8"), output_dir = NULL,
               text_encoders = TRUE, verbose = TRUE)
}
\arguments{
\item{quantize}{Logical. Build the quantized artifact after
downloading.}

\item{precision}{"nf4" (~7 GB, GPU-resident on 16 GB cards) or
"fp8" (~12 GB, CPU-resident, streamed; near-bf16 quality).}

\item{output_dir}{Directory for the quantized artifact.}

\item{text_encoders}{Logical. Also fetch the CLIP + T5 text encoders,
tokenizer, VAE, and scheduler config (~10 GB).}

\item{verbose}{Logical.}
}
\value{
Invisibly, a list with \code{transformer_dir},
  \code{artifact_dir}, and \code{support} (named file paths).
}
\description{
Skips work already done: a valid quantized manifest short-circuits
the transformer download; cached files are not re-fetched. Needs
\code{HF_TOKEN} set for the gated repo (see the error message it
raises without one). The bf16 transformer source (~24 GB in the
HuggingFace cache) may be deleted after quantization.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_ltx23}
\alias{download_ltx23}
\title{Download and Prepare LTX-2.3 Model Weights}
\description{
Downloads the LTX-2.3 distilled checkpoint (46 GB, LTX-2 Community
License) and the Gemma3 text encoder from HuggingFace with an explicit
consent prompt, then quantizes the transformer to the local fp8
artifact (~26 GB) used by the GPU-poor pipeline.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_snake_beta}
\alias{ltx23_snake_beta}
\title{SnakeBeta activation}
\usage{
ltx23_snake_beta(channels, eps = 1e-09)
}
\arguments{
\item{channels}{Integer.}

\item{eps}{Numeric.}
}
\value{
Module whose forward(hidden_states) returns the Snake
  activation \code{x + sin(alpha * x)^2 / beta}, a tensor of the same
  shape as the input.
}
\description{
\code{x + (1 / (beta + eps)) * sin(x * alpha)^2} with per-channel
log-scale alpha/beta parameters.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_attention}
\alias{flux_attention}
\title{FLUX joint attention}
\usage{
flux_attention(query_dim, heads, dim_head, added_kv = FALSE, pre_only = FALSE,
               eps = 1e-06, bias = TRUE)
}
\arguments{
\item{query_dim}{Integer. Model dimension.}

\item{heads}{Integer. Number of attention heads.}

\item{dim_head}{Integer. Per-head dimension.}

\item{added_kv}{Logical. Add text-stream projections (double blocks).}

\item{pre_only}{Logical. Skip the output projection (single blocks).}

\item{eps}{Numeric. RMS norm epsilon.}

\item{bias}{Logical. Bias on the linear projections (TRUE for FLUX.1,
FALSE for FLUX.2).}
}
\value{
Module whose forward(hidden_states, encoder_hidden_states,
  image_rotary_emb, chunk_size) returns the attended image stream
  [B, S, query_dim]. When \code{encoder_hidden_states} is supplied
  (double-stream blocks) it returns \code{list(image, text)} instead,
  each projected by its own output layer.
}
\description{
Multi-head attention with per-head RMS q/k norms and rotary position
embeddings. With \code{added_kv = TRUE} (double-stream blocks) the
text stream gets its own q/k/v projections and both streams attend
jointly (text tokens first); the outputs are split back and projected
per stream. With \code{pre_only = TRUE} (single-stream blocks) there
is no output projection. Reference: diffusers FluxAttention +
FluxAttnProcessor.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{connectors_ltx23}
\alias{connectors_ltx23}
\title{LTX-2.3 Text Embedding Connectors}
\description{
Fresh R port of the LTX text connectors from the diffusers reference
(Apache-2.0, src/diffusers/pipelines/ltx2/connectors.py). The
connectors take raw stacked per-layer Gemma3 hidden states
[batch, seq, caption_channels, num_layers + 1], normalize and project
them per modality, replace padding with learnable registers, and run a
small 1D transformer per modality to produce the DiT text embeddings.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flowmatch_set_timesteps}
\alias{flowmatch_set_timesteps}
\title{Set timesteps for inference}
\usage{
flowmatch_set_timesteps(schedule, num_inference_steps = 50L, device = "cpu",
                        mu = NULL, sigmas = NULL, timesteps = NULL)
}
\arguments{
\item{schedule}{List. The FlowMatch scheduler object.}

\item{num_inference_steps}{Integer. Number of denoising steps. Default: 50}

\item{device}{Character or torch device. Device for tensors. Default: "cpu"}

\item{mu}{Numeric or NULL. Shift parameter for dynamic shifting. Required
if use_dynamic_shifting is TRUE. Default: NULL}

\item{sigmas}{Numeric vector or NULL. Custom sigma values. Default: NULL}

\item{timesteps}{Numeric vector or NULL}
}
\value{
Updated scheduler with configured timesteps and sigmas.
}
\description{
Configures the scheduler timesteps for a specific number of inference steps.
This must be called before using the scheduler for denoising.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_get_timestep_embedding}
\alias{ltx23_get_timestep_embedding}
\title{Sinusoidal timestep embedding}
\usage{
ltx23_get_timestep_embedding(timesteps, embedding_dim, flip_sin_to_cos = TRUE,
                             downscale_freq_shift = 0, max_period = 10000)
}
\arguments{
\item{timesteps}{1D tensor of timestep values.}

\item{embedding_dim}{Integer. Output embedding size.}

\item{flip_sin_to_cos}{Logical. Put cos before sin.}

\item{downscale_freq_shift}{Numeric. Frequency delta control.}

\item{max_period}{Numeric. Maximum embedding frequency period.}
}
\value{
Tensor [N, embedding_dim].
}
\description{
DDPM-style sinusoidal embedding. LTX uses \code{flip_sin_to_cos=TRUE}
(cos first) and \code{downscale_freq_shift=0}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{print.diffuseR_resident}
\alias{print.diffuseR_resident}
\title{Print a resident handle}
\usage{
\method{print}{diffuseR_resident}(x, ...)
}
\arguments{
\item{x}{A \code{diffuseR_resident} handle.}

\item{...}{Ignored.}
}
\value{
Invisibly \code{x}. Called for the side effect of printing a
  one-block summary to the console.
}
\description{
Print a resident handle
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{sd_pipeline_safetensors}
\alias{sd_pipeline_safetensors}
\title{Native Stable Diffusion pipelines from diffusers safetensors}
\description{
Assemble and run the native SD pipeline directly from a HuggingFace
diffusers directory (\code{unet/}, \code{vae/}, \code{text_encoder/}),
with no TorchScript \code{.pt} step - so it works on Blackwell and
loads the same weights everyone else uses. SD21 is wired end to end
here; SDXL still needs its second text encoder and added-conditioning
embeddings (tracked in tasks/todo.md).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_parallel_self_attention}
\alias{flux2_parallel_self_attention}
\title{FLUX.2 parallel self-attention (single-stream)}
\usage{
flux2_parallel_self_attention(query_dim, heads, dim_head, mlp_ratio = 3,
                              eps = 1e-06, bias = FALSE)
}
\arguments{
\item{query_dim}{Integer. Model dimension.}

\item{heads}{Integer. Attention heads.}

\item{dim_head}{Integer. Per-head dimension.}

\item{mlp_ratio}{Numeric. MLP hidden multiplier (FLUX.2: 3.0).}

\item{eps}{Numeric. RMS norm epsilon.}

\item{bias}{Logical.}
}
\value{
Module whose forward(hidden_states, image_rotary_emb,
  chunk_size) returns the block output [B, S, query_dim]: attention
  and MLP branches computed in parallel from one fused projection,
  concatenated, and projected back by a second fused layer.
}
\description{
ViT-22B-style parallel block internals: one fused projection produces
QKV and the SwiGLU MLP input; one fused projection consumes
cat(attention output, MLP output). Reference:
Flux2ParallelSelfAttention + Flux2ParallelSelfAttnProcessor.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.resident_components}
\alias{.resident_components}
\title{Every nn_module field of a pipeline, by name}
\usage{
.resident_components(pipeline)
}
\arguments{
\item{pipeline}{A loaded diffuseR pipeline.}
}
\value{
A named list of the pipeline's \code{nn_module} fields,
  possibly empty.
}
\description{
Discovery beats a hard-coded list: the families disagree on which
components exist (FLUX.1 has two text encoders, LTX adds a video VAE,
an audio VAE and a vocoder), and a field added later is picked up
without touching this file.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_set_attn_chunk}
\alias{ltx23_set_attn_chunk}
\title{Set the attention query-chunk size across a transformer}
\usage{
ltx23_set_attn_chunk(transformer, chunk)
}
\arguments{
\item{transformer}{An \code{ltx23_transformer}.}

\item{chunk}{Integer or NULL.}
}
\value{
Invisibly, the transformer.
}
\description{
R torch has no fused attention, so the [B, H, S, S] matrix
materializes; chunking queries bounds the peak. NULL disables
chunking.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{Upsample2D}
\alias{Upsample2D}
\title{Upsample Block}
\usage{
Upsample2D(channels)
}
\description{
Upsample Block
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_encoder3d}
\alias{ltx23_video_encoder3d}
\title{LTX-2.3 video encoder}
\usage{
ltx23_video_encoder3d(in_channels = 3L, out_channels = 128L,
                      block_out_channels = c(256L, 512L, 1024L, 1024L),
                      spatio_temporal_scaling = c(TRUE, TRUE, TRUE, TRUE),
                      layers_per_block = c(4L, 6L, 4L, 2L, 2L),
                      downsample_type = NULL, patch_size = 4L,
                      patch_size_t = 1L, resnet_norm_eps = 1e-06,
                      is_causal = TRUE, spatial_padding_mode = "zeros")
}
\arguments{
\item{block_out_channels}{Integer vector. Per-block output channels.}

\item{spatio_temporal_scaling}{Logical vector per block.}

\item{layers_per_block}{Integer vector (blocks then mid).}

\item{downsample_type}{Character vector per block.}

\item{resnet_norm_eps}{Numeric.}

\item{is_causal}{Logical.}

\item{spatial_padding_mode}{Character.}

\item{in_channels,out_channels}{Integers. Pixel and latent channels.}

\item{patch_size,patch_size_t}{Integers. Pixel patchification.}
}
\value{
Module whose forward(hidden_states, causal) returns the
  encoded video latent [B, C, F, H, W], with the last channel
  repeated to carry the per-channel scale expected downstream.
}
\description{
Pixel video [B, 3, F, H, W] -> latent statistics
[B, 2 * latent_channels, F/8, H/32, W/32] (mean and a uniform log-var
channel broadcast across the latent channels).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{dit_flux_modules}
\alias{dit_flux_modules}
\title{FLUX Transformer Building Blocks}
\description{
Fresh R port of the FLUX MMDiT blocks from the diffusers reference
implementation (Apache-2.0,
src/diffusers/models/transformers/transformer_flux.py and
src/diffusers/models/normalization.py). Module field names mirror the
diffusers state-dict keys 1:1 so checkpoints load without remapping.
Reuses the LTX primitives \code{ltx23_rms_norm}, \code{.ltx23_sdpa}
and \code{ltx23_feed_forward}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_ada_layer_norm_continuous}
\alias{flux_ada_layer_norm_continuous}
\title{FLUX continuous adaLN (final norm)}
\usage{
flux_ada_layer_norm_continuous(dim, cond_dim = dim, bias = TRUE)
}
\arguments{
\item{dim}{Integer. Model dimension.}

\item{cond_dim}{Integer. Conditioning embedding dimension.}

\item{bias}{Logical. Bias on the projection (TRUE for FLUX.1, FALSE
for FLUX.2).}
}
\value{
Module whose forward(x, cond) returns \code{x} normalized and
  then scaled and shifted by the conditioning embedding, a tensor of
  the same shape as \code{x}.
}
\description{
Scale/shift conditioning of the final norm. Note the chunk order:
scale first, then shift (the reverse of adaLN-Zero). Reference:
diffusers AdaLayerNormContinuous as used by FLUX norm_out
(elementwise_affine = FALSE, eps = 1e-6).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unet_native}
\alias{unet_native}
\title{Native UNet for Stable Diffusion}
\usage{
unet_native(in_channels = 4L, out_channels = 4L,
            block_out_channels = c(320L, 640L, 1280L, 1280L),
            layers_per_block = 2L, cross_attention_dim = 1024L,
            attention_head_dim = 64L)
}
\arguments{
\item{in_channels}{Input channels (default 4 for latent space)}

\item{out_channels}{Output channels (default 4)}

\item{block_out_channels}{Channel multipliers per block}

\item{layers_per_block}{Number of ResBlocks per down/up block}

\item{cross_attention_dim}{Context dimension from text encoder}

\item{attention_head_dim}{Dimension per attention head}
}
\value{
An nn_module representing the UNet
}
\description{
Native R torch implementation of UNet2DConditionModel.
Replaces TorchScript for better GPU compatibility.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_patchify}
\alias{zimage_patchify}
\title{Patchify a latent image to Z-Image tokens}
\usage{
zimage_patchify(image, patch_size = 2L, f_patch_size = 1L)
}
\arguments{
\item{image}{Tensor of shape [C, F, H, W].}

\item{patch_size}{Integer spatial patch size. Default 2.}

\item{f_patch_size}{Integer temporal patch size. Default 1.}
}
\value{
Tensor of shape [num_tokens, patch_dim].
}
\description{
(C, F, H, W) -> [F/pF * H/p * W/p, pF * p * p * C], matching
_patchify_image. No padding is applied here.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_single_block}
\alias{flux_single_block}
\title{FLUX single-stream transformer block}
\usage{
flux_single_block(dim, num_attention_heads, attention_head_dim, mlp_ratio = 4)
}
\arguments{
\item{dim}{Integer. Model dimension.}

\item{num_attention_heads}{Integer. Attention heads.}

\item{attention_head_dim}{Integer. Per-head dimension.}

\item{mlp_ratio}{Numeric. MLP hidden dim multiplier.}
}
\value{
Module whose forward(hidden_states, temb, image_rotary_emb)
  returns the joint hidden states.
}
\description{
Parallel attention + MLP over the joint [text; image] sequence with a
shared gate: \code{x + gate * proj_out(cat(attn, gelu(mlp)))}. The
reference concatenates the streams inside every block and splits after;
here the caller concatenates once before the single-block stack, which
is numerically identical. Reference: diffusers
FluxSingleTransformerBlock.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vae_ltx23}
\alias{vae_ltx23}
\title{LTX-2.3 Causal Video VAE}
\description{
Fresh R port of the LTX-2 video autoencoder from the diffusers
reference (Apache-2.0, autoencoder_kl_ltx2.py), with LTX 2.3 defaults:
encoder blocks (256, 512, 1024, 1024), a 4-up-block decoder with mixed
(spatiotemporal, spatiotemporal, temporal, spatial) upsampling, no
upsample residuals, and zeros spatial padding throughout. The encoder
is causal; the decoder is not.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{scheduler_add_noise}
\alias{scheduler_add_noise}
\title{Add noise to latents using DDIM scheduler}
\usage{
scheduler_add_noise(original_latents, noise, timestep, scheduler_obj)
}
\arguments{
\item{original_latents}{A torch tensor representing the original latents.}

\item{noise}{A torch tensor representing the noise to be added.}

\item{timestep}{An integer representing the current timestep in the diffusion
process.}

\item{scheduler_obj}{A list containing the DDIM scheduler parameters, including
alphas_cumprod and timesteps. The alphas_cumprod represents how much of the
original signal remains at each timestep of the diffusion process.}
}
\value{
A torch tensor containing the noised latents, which represents the
original latents with the appropriate amount of noise added for the given
timestep.
}
\description{
This function adds noise to the original latents according to the DDIM
scheduler's diffusion process. It computes the noisy latents based on the
original latents, noise, and the current timestep.
}
\details{
The noise is added according to the standard diffusion forward process formula:
noised_latents = sqrt(alpha_cumprod) * original_latents + sqrt(1-alpha_cumprod) * noise

Where alpha_cumprod is the cumulative product of (1-beta) values up to the
specified timestep, with beta being the noise schedule.
}
\examples{
if (torch::torch_is_installed()) {
  scheduler <- ddim_scheduler_create(num_inference_steps = 5)
  latents <- torch::torch_randn(c(1, 4, 8, 8))
  noised_latents <- scheduler_add_noise(
    original_latents = latents,
    noise = torch::torch_randn_like(latents),
    timestep = scheduler$timesteps[1],
    scheduler_obj = scheduler
  )
  noised_latents$shape
}

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{bpe_tokenizer}
\alias{bpe_tokenizer}
\title{BPE Tokenizer}
\usage{
bpe_tokenizer(tokenizer_path)
}
\arguments{
\item{tokenizer_path}{Path to tokenizer.json or directory containing it.}
}
\value{
A bpe_tokenizer object.
}
\description{
Native R implementation of Byte-Pair Encoding tokenizer.
Loads from HuggingFace tokenizer.json format.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_unpack_latents_with_ids}
\alias{flux2_unpack_latents_with_ids}
\title{Unpack FLUX.2 tokens back to a latent grid via position ids}
\usage{
flux2_unpack_latents_with_ids(x, ids, height, width)
}
\arguments{
\item{x}{Tensor [B, S, C] of tokens.}

\item{ids}{Tensor [S, 4] (or [B, S, 4]) of position ids.}

\item{height,width}{Integers. Packed grid dimensions.}
}
\value{
Tensor [B, C, height, width].
}
\description{
Scatters tokens to (H, W) positions taken from the id columns
(H = column 2, W = column 3, 0-based values). Reference:
Flux2KleinPipeline._unpack_latents_with_ids.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.ltx23_release_attn_buffers}
\alias{.ltx23_release_attn_buffers}
\title{Release the attention scratch buffers}
\usage{
.ltx23_release_attn_buffers()
}
\value{
Invisibly, NULL.
}
\description{
Release the attention scratch buffers
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{txt2img}
\alias{txt2img}
\title{Generate an image from a text prompt using a diffusion pipeline}
\usage{
txt2img(prompt, model_name = c("sd21", "sdxl", "flux1", "flux2", "zimage"), ...)
}
\arguments{
\item{prompt}{A character string prompt describing the image to generate.}

\item{model_name}{Name of the model to use (e.g., `"sd21"`).}

\item{...}{Additional parameters passed to the diffusion process.}
}
\value{
A tensor or image object, depending on implementation.
}
\description{
Generate an image from a text prompt using a diffusion pipeline
}
\examples{
\dontrun{
img <- txt2img("a cat wearing sunglasses in space", device = "cuda")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{clip_pooled_output}
\alias{clip_pooled_output}
\title{Pooled CLIP output at the EOS position}
\usage{
clip_pooled_output(hidden_states, input_ids)
}
\arguments{
\item{hidden_states}{Final-LN hidden states [B, S, D] from
\code{\link{text_encoder_native}} (with \code{apply_final_ln = TRUE}).}

\item{input_ids}{Token ids [B, S] (0-based, as fed to the encoder).}
}
\value{
Tensor [B, D].
}
\description{
The HF CLIPTextModel pooler_output: the final-layer-norm hidden state
at the EOS token position, located by argmax over the token ids (EOS
is the highest id in the CLIP vocab, and causal attention makes any
padding after it irrelevant). No text projection is applied - this is
what FLUX uses as pooled_projections.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{models2devices}
\alias{models2devices}
\title{models2devices}
\usage{
models2devices(model_name, devices = "cpu", unet_dtype_str = NULL,
               download_models = FALSE)
}
\arguments{
\item{model_name}{A character string representing the name of the model to be used.}

\item{devices}{A character string or a named list specifying the devices for different components of the model.}

\item{unet_dtype_str}{A character string specifying the data type for the UNet model.}

\item{download_models}{Logical indicating whether to download models if they are not found.}
}
\value{
A list containing the device configuration, UNet data type, and CPU/CUDA devices.
}
\description{
This function sets up the model directory, device configuration, and data types for diffusion models.
It checks the validity of the model name and devices, detects model type, and downloads the model if necessary.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_prepare_latent_ids}
\alias{flux2_prepare_latent_ids}
\title{Build FLUX.2 latent position ids}
\usage{
flux2_prepare_latent_ids(height, width, device = "cpu")
}
\arguments{
\item{height}{Integer. Packed grid height (pixel height / 16).}

\item{width}{Integer. Packed grid width (pixel width / 16).}

\item{device}{Device for the resulting tensor.}
}
\value{
Float tensor [height * width, 4].
}
\description{
Columns (T, H, W, L) with H and W carrying the packed-grid position
(row-major: H varies slowest), T = L = 0. Reference:
Flux2KleinPipeline._prepare_latent_ids.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_model_component}
\alias{load_model_component}
\title{Load a specific component of a diffusion model}
\usage{
load_model_component(component, model_name = "sd21", device = "cpu",
                     unet_dtype_str = NULL, download = TRUE, use_native = FALSE)
}
\arguments{
\item{component}{Character string, the component to load: "unet", "decoder", or "text_encoder".}

\item{model_name}{Character string, the name of the model to use.}

\item{device}{Character string, the torch device to load the model onto ("cpu" or "cuda").}

\item{unet_dtype_str}{Optional; the data type for the UNet model. If `NULL`, defaults to `float32` for CPU and `float16` for CUDA.}

\item{download}{Logical; if `TRUE` (default), downloads the model if it doesn't exist locally.}

\item{use_native}{Logical; if `TRUE`, uses native R torch modules instead of TorchScript.
Supported for unet, decoder, text_encoder, and text_encoder2. Native modules have
better GPU compatibility (especially on Blackwell/RTX 50xx).}
}
\value{
A torch model object.
}
\description{
Loads a TorchScript model component (UNet, decoder, or text encoder) from the
hfhub cache or legacy model directory, downloading it first if necessary.
}
\examples{
\dontrun{
unet <- load_model_component("unet", "sd21", "cpu")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{clear_vram}
\alias{clear_vram}
\title{Clear VRAM Cache}
\usage{
clear_vram(verbose = FALSE)
}
\arguments{
\item{verbose}{Logical. Print memory status before/after.}
}
\value{
Invisibly returns NULL.
}
\description{
Forces garbage collection and clears CUDA memory cache.
}
\examples{
if (torch::torch_is_installed()) {
  clear_vram()
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.st_can_read}
\alias{.st_can_read}
\title{Probe whether the installed safetensors can READ a dtype}
\usage{
.st_can_read(dtype = c("bfloat16", "float8_e4m3fn"))
}
\arguments{
\item{dtype}{"bfloat16" or "float8_e4m3fn".}
}
\value{
Logical.
}
\description{
Hand-builds a tiny safetensors file of the dtype (via
\code{.st_write_min}, no safetensors writer involved) and tries to
load it back. Cached per session;
\code{options(diffuseR.st_read_caps = list(bfloat16 = TRUE, ...))}
overrides the probe for tests and for forcing a tier.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_load_pipeline}
\alias{flux_load_pipeline}
\title{Load the FLUX.1-schnell pipeline}
\usage{
flux_load_pipeline(model_dir = NULL, device = "cuda", precision = NULL,
                   text_device = NULL, attn_chunk = NULL, phase_offload = TRUE,
                   pin = NULL, verbose = TRUE)
}
\arguments{
\item{model_dir}{Quantized artifact directory (default: the
\code{download_flux1} location for \code{precision}), or a raw
diffusers transformer directory for full-precision loading.}

\item{device}{Character. Compute device.}

\item{precision}{"nf4" or "fp8"; NULL picks the
\code{\link{flux_memory_profile}} recommendation.}

\item{text_device}{Where the text encoders compute. NULL (default)
takes the \code{\link{flux_memory_profile}} recommendation: "cuda"
on 14 GB+ cards (T5-XXL onloads in bfloat16 for its encode) and
"cpu" below that (T5-XXL runs float32 in place, since its ~9.8 GB
GPU encode phase would not fit).}

\item{attn_chunk}{Integer or NULL. Attention query-chunk override.}

\item{phase_offload}{Logical. One GPU tenant per phase.}

\item{pin}{Logical or NULL. Page-lock the phase-swapped weights for
DMA-rate transfer (see \code{\link{staging}}). NULL (default)
resolves via \code{options(diffuseR.pin_staging)} then the
host-RAM-aware \code{\link{recommend}} decision.}

\item{verbose}{Logical.}
}
\value{
A \code{flux_pipeline} list.
}
\description{
Loads the quantized transformer artifact plus the VAE decoder, CLIP
and T5 text encoders, tokenizer, and scheduler config (from the
HuggingFace cache populated by \code{\link{download_flux1}}).
Components load to the CPU when \code{phase_offload} is on and move
to the GPU only for their phase of the generation.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{recommend}
\alias{recommend}
\title{Recommend a precision and device configuration for a model}
\usage{
recommend(model = c("sd21", "sdxl", "flux1", "flux2", "zimage", "ltx"),
          vram_gb = NULL, st_caps = NULL, host_ram_gb = NULL)
}
\arguments{
\item{model}{"sd21", "sdxl", "flux1", "flux2", "zimage", or "ltx".}

\item{vram_gb}{Numeric or NULL. Free VRAM in GB; auto-detected via
nvidia-smi when NULL.}

\item{st_caps}{NULL or a named logical list with \code{bfloat16}
and/or \code{float8_e4m3fn} - the safetensors READ capabilities.
NULL probes the installed safetensors.}

\item{host_ram_gb}{Numeric or NULL. Available host RAM in GB;
auto-detected (Linux \code{MemAvailable}) when NULL, NA where
undetectable.}
}
\value{
A list with \code{model}, \code{precision}, \code{devices}
  (named component -> device map), \code{offload} (phase-offloading
  logical), \code{max_pixels}, \code{text_device}, \code{attn_chunk},
  \code{vram_gb}, \code{pin} (page-lock the phase-swapped host
  copies), \code{pinned_set_gb} (estimated pinned bytes),
  \code{host_ram_gb}, \code{fork_suggested} (logical), and
  \code{note} (the fork suggestion string, or NULL).
}
\description{
One VRAM-and-capability-aware recommendation for every diffuseR
model. The policy:
}
\details{
\itemize{
\item nf4 is the default tier for the quantized families (flux1,
flux2, zimage, ltx). Its weights are packed uint8 plus float32
blocks in sub-2 GB shards, which every safetensors reads, so it
always loads. The SD models ship no quantized weights; their
floor is fp16 with placement varying by VRAM.
\item When the card has room for a higher-quality tier (fp8 or
bf16) AND the installed safetensors can \emph{read} that dtype
(\code{\link{.st_can_read}}), that tier is recommended instead.
\item When the card has room but safetensors cannot read the tier,
nf4 is recommended and the fork suggestion is surfaced in
\code{note} (never an error).
}

This is the policy engine; it does no disk I/O and does not know which
artifacts are built. Loaders reconcile the recommendation with what is
on disk (see \code{\link{flux_load_pipeline}}). Thresholds are
validated on an RTX 5060 Ti (16 GB) and are deliberately conservative
elsewhere. Video sizing for \code{"ltx"} is coarse here; the LTX
pipeline uses \code{\link{ltx23_memory_profile}} for frame-aware
placement.

The pinning decision: phase-swapped weights are page-locked host
copies (see \code{\link{staging}}) that transfer at DMA rate -
but pinned pages are unswappable, so on small-RAM machines they turn
memory pressure into OOM kills. \code{pin} is TRUE when available
host RAM covers the model's pinned set twice over, FALSE below that,
FALSE on the cpu tier (nothing stages), and TRUE when RAM cannot be
detected (page-locking already fails soft per component). The LTX
pipeline, the Gemma3 encoder, and the FLUX-family image loaders
(flux1, flux2, zimage) take \code{pin} arguments and stage pinned
weights (see \code{\link{staging}}); the SD-family loaders place
components statically and do not phase-swap, so \code{pin} is inert
for them. \code{options(diffuseR.pin_staging)} is the global switch.

}
\examples{
# Stating vram_gb and st_caps makes the policy deterministic: no GPU
# and no installed safetensors needed.
r <- recommend("flux1", vram_gb = 16,
               st_caps = list(bfloat16 = TRUE, float8_e4m3fn = FALSE))
r$precision       # "nf4": fp8 fits the card, but cannot be read
r$fork_suggested  # TRUE
cat(r$note)       # the fork-or-nf4 message

# Same card, once safetensors can read float8
recommend("flux1", vram_gb = 16,
          st_caps = list(bfloat16 = TRUE, float8_e4m3fn = TRUE))$precision

# Auto-detect VRAM and probe the installed safetensors
str(recommend("flux2"))
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_pos_embed}
\alias{zimage_pos_embed}
\title{Compute Z-Image rotary frequencies from position ids}
\usage{
zimage_pos_embed(ids, axes_dim = c(32L, 48L, 48L), theta = 256)
}
\arguments{
\item{ids}{Tensor of shape [S, 3] from \code{zimage_cap_pos_ids} /
\code{zimage_img_pos_ids}.}

\item{axes_dim}{Integer vector of per-axis rotary dims; must sum to
the attention head dim. Z-Image uses c(32, 48, 48).}

\item{theta}{Numeric. RoPE base frequency. Z-Image uses 256.}
}
\value{
List of two tensors (cos, sin), each [S, sum(axes_dim)],
  float32, on the device of \code{ids}.
}
\description{
Per-axis 1D rotary frequencies in the interleaved-real convention.
Frequencies and angles are built in float64, then the angles are cast
to float32 before cos/sin — matching the reference torch.polar call on
a .float() tensor. Output format matches \code{flux_pos_embed} so
\code{flux_apply_rotary_emb} applies unchanged.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.verbosity}
\alias{.verbosity}
\title{Normalize a verbosity flag}
\usage{
.verbosity(verbose)
}
\arguments{
\item{verbose}{Logical, or one of "silent", "progress", "steps".
TRUE maps to "steps" and FALSE to "silent".}
}
\value{
One of "silent", "progress", "steps".
}
\description{
Normalize a verbosity flag
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_unet_sdxl_safetensors}
\alias{load_unet_sdxl_safetensors}
\title{Load HF safetensors weights into the native SDXL UNet}
\usage{
load_unet_sdxl_safetensors(native_unet, path, verbose = TRUE)
}
\arguments{
\item{native_unet}{A \code{\link{unet_sdxl_native}} module.}

\item{path}{Path to the UNet directory (containing
\code{diffusion_pytorch_model.safetensors} or its shard index) or
directly to the single-file checkpoint.}

\item{verbose}{Print how many parameters were loaded.}
}
\value{
The native UNet with weights loaded (invisibly).
}
\description{
Load HF safetensors weights into the native SDXL UNet
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_downsample1d}
\alias{ltx23_downsample1d}
\title{Anti-aliasing 1D downsampler (low-pass then stride)}
\usage{
ltx23_downsample1d(ratio = 2L, kernel_size = NULL)
}
\arguments{
\item{ratio}{Integer. Downsampling ratio.}

\item{kernel_size}{Integer or NULL (default 6*ratio rounded even).}
}
\value{
Module whose forward(x) returns \code{x} low-pass filtered
  and decimated by \code{ratio} along the time axis.
}
\description{
Anti-aliasing 1D downsampler (low-pass then stride)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_unpack_latents}
\alias{flux_unpack_latents}
\title{Unpack a FLUX patch sequence back into latents}
\usage{
flux_unpack_latents(latents, height, width, vae_scale_factor = 8L)
}
\arguments{
\item{latents}{Tensor of shape [B, S, C_packed].}

\item{vae_scale_factor}{Integer. Spatial downsampling of the VAE (8).}

\item{height,width}{Integers. Image height/width in pixels.}
}
\value{
Tensor of shape [B, C_packed / 4, height / 8, width / 8].
}
\description{
Inverse of \code{flux_pack_latents}. Height and width are the target
image dimensions in pixels; the latent grid is derived via the VAE
scale factor and the 2x2 patch size. Reference:
FluxPipeline._unpack_latents.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{apply_bpe_merges}
\alias{apply_bpe_merges}
\title{Apply BPE merge rules}
\usage{
apply_bpe_merges(tokens, merge_priority, vocab)
}
\description{
Apply BPE merge rules
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.resident_pin}
\alias{.resident_pin}
\title{Pin every component of a pipeline for fast transfer}
\usage{
.resident_pin(pipeline, verbose = TRUE)
}
\arguments{
\item{pipeline}{A loaded diffuseR pipeline.}

\item{verbose}{Print progress.}
}
\value{
A named list of staging sets, one per component that could be
  pinned. Components that fail to page-lock are absent, and fall back
  to the pageable \code{$to()} path.
}
\description{
Re-uses any staging the loader already built (the phase-offload path
pins as part of loading), and pins the rest. Pinning a component that
is currently on the GPU also evicts it, since \code{.pin_component}
copies into page-locked host memory and re-points the live tensors at
it, so this doubles as the initial offload.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{qwen3_encoder}
\alias{qwen3_encoder}
\title{Qwen3 encoder stack}
\usage{
qwen3_encoder(vocab_size = 151936L, hidden_size = 2560L,
              intermediate_size = 9728L, num_hidden_layers = 36L,
              num_attention_heads = 32L, num_key_value_heads = 8L,
              head_dim = 128L, rope_theta = 1e+06, rms_norm_eps = 1e-06)
}
\arguments{
\item{rope_theta}{Numeric.}

\item{rms_norm_eps}{Numeric.}

\item{vocab_size,hidden_size,intermediate_size,num_hidden_layers}{Integers.}

\item{num_attention_heads,num_key_value_heads,head_dim}{Integers.}
}
\value{
Module whose forward(input_ids, attention_mask = NULL,
  out_layers) returns a list of hidden-state tensors [B, S, hidden],
  one per requested layer (a value of k means the state after k
  layers, matching HF \code{output.hidden_states[k]}). Runs only to
  \code{max(out_layers)}. \code{input_ids} are 1-based.
}
\description{
Defaults are the Qwen3-4B configuration used by FLUX.2 klein. The
module tree mirrors the checkpoint keys (\code{model.embed_tokens},
\code{model.layers.*}, \code{model.norm}); the tied LM head carries
no weights of its own and is not implemented.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{quant_conv}
\alias{quant_conv}
\title{Quant Conv}
\usage{
quant_conv(x, dtype, device)
}
\arguments{
\item{x}{Input tensor to be processed.}

\item{dtype}{Data type for the tensor (e.g., "torch_float16" or "torch_float32").}

\item{device}{Device on which the tensor is located (e.g., "cpu" or "cuda").}
}
\value{
Processed tensor after applying the quantized convolution.
}
\description{
This function applies a quantized convolution operation to an input tensor.
It is typically used in the context of image processing, particularly in
generative models like Stable Diffusion.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{UNetResBlock}
\alias{UNetResBlock}
\title{ResNet Block for UNet}
\usage{
UNetResBlock(in_channels, out_channels, time_embed_dim)
}
\description{
ResNet Block for UNet
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_quantize_nf4}
\alias{ltx23_quantize_nf4}
\title{Quantize an LTX-2.3 checkpoint to NF4 shards}
\usage{
ltx23_quantize_nf4(checkpoint_path, output_dir = NULL, shard_bytes = 1.9e+09,
                   force = FALSE, verbose = TRUE)
}
\arguments{
\item{checkpoint_path}{Source .safetensors (bf16 single file).}

\item{output_dir}{Output directory for shards + manifest; NULL (the
default) resolves under tools::R_user_dir("diffuseR", "data").}

\item{shard_bytes}{Numeric. Target shard size in bytes. The default
1.9e9 keeps every shard under the 2^31-byte (~2.15 GB) ceiling that
stock CRAN safetensors can read. Pass a larger value (e.g. 4e9) only
for local builds you will read back with a fork-patched safetensors.}

\item{force}{Logical. Re-quantize even if a valid manifest exists.}

\item{verbose}{Logical.}
}
\value{
Invisibly, the manifest list.
}
\description{
Same streaming layout and cast policy as
\code{\link{ltx23_quantize_fp8}}, but cast-set weights are stored as
NF4 (\code{<key>} packed uint8 + \code{<key>_absmax} float32 blocks +
the original shape recovered from the model config at load time).
Non-cast tensors are copied through unchanged. The manifest carries
\code{format = "nf4"}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_quantize_nf4}
\alias{gemma3_quantize_nf4}
\title{Quantize a Gemma3 text encoder to NF4 shards}
\usage{
gemma3_quantize_nf4(model_path, output_dir = NULL, shard_bytes = 1.9e+09,
                    force = FALSE, verbose = TRUE)
}
\arguments{
\item{model_path}{HuggingFace snapshot directory (config.json +
model-*.safetensors).}

\item{output_dir}{Output directory for shards + manifest (default:
\code{gemma3-nf4} under \code{tools::R_user_dir}).}

\item{shard_bytes}{Numeric. Target shard size in bytes; the 1.9e9
default keeps shards readable by stock CRAN safetensors.}

\item{force}{Logical. Re-quantize even if a valid manifest exists.}

\item{verbose}{Logical.}
}
\value{
Invisibly, the manifest list.
}
\description{
Streams the HuggingFace Gemma3 checkpoint tensor by tensor. The
language model's projection weights (q/k/v/o and gate/up/down, ~11B
of the 12B parameters) are stored as NF4 (packed uint8 +
\code{<key>_absmax} float32 blocks); embeddings and norms are copied
at the resident dtype. Vision-tower and projector weights are
dropped - the text encoder never uses them. The result is a ~8 GB
artifact that fits a 16 GB card during the encode phase (vs 45 GB
of host RAM for the fp32 CPU path).
}
\details{
Keys are stored normalized (\code{language_model.} / \code{model.}
prefixes stripped), matching the module tree of
\code{\link{gemma3_text_model}}.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_component}
\alias{download_component}
\title{Download a single TorchScript model component}
\usage{
download_component(model_name = "sd21", component, device = "cpu",
                   overwrite = FALSE, show_progress = TRUE)
}
\arguments{
\item{model_name}{Character string, the name of the model (e.g., \code{"sd21"}).}

\item{component}{Character string, the component to download (e.g., \code{"unet"}, \code{"decoder"}).}

\item{device}{Character string, the device type (e.g., \code{"cpu"} or \code{"cuda"}).}

\item{overwrite}{Logical; if \code{TRUE}, force re-download even if cached.}

\item{show_progress}{Logical; if \code{TRUE} (default), displays progress during download.}
}
\value{
The local file path to the downloaded component (character string).
}
\description{
Downloads a specific model component file (e.g., UNet, decoder, text encoder)
using \code{hfhub::hub_download()} from the cornball-ai dataset repos.
}
\examples{
\dontrun{
path <- download_component("sd21", "text_encoder", "cpu")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_modulation}
\alias{flux2_modulation}
\title{FLUX.2 shared modulation projection}
\usage{
flux2_modulation(dim, mod_param_sets = 2L, bias = FALSE)
}
\arguments{
\item{dim}{Integer. Model dimension.}

\item{mod_param_sets}{Integer. Number of (shift, scale, gate) triples.}

\item{bias}{Logical.}
}
\value{
Module whose forward(temb) returns the modulation tensor
  \code{linear(silu(temb))}, holding \code{mod_param_sets} triples of
  (shift, scale, gate) along the last axis.
}
\description{
\code{linear(silu(temb))} producing \code{mod_param_sets} triples of
(shift, scale, gate). Computed once per forward at model level and
broadcast to every block. Reference: Flux2Modulation.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{dit_ltx23}
\alias{dit_ltx23}
\title{LTX-2.3 Audio-Video Diffusion Transformer}
\description{
Fresh R port of the LTX-2 transformer from the diffusers reference
(Apache-2.0, src/diffusers/models/transformers/transformer_ltx2.py),
configured for LTX 2.3: gated attention, cross-attention modulation,
prompt AdaLN, split RoPE, and connector-projected text embeddings
(no in-model caption projection).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_text_model}
\alias{gemma3_text_model}
\title{Gemma3 Text Model}
\usage{
gemma3_text_model(config)
}
\arguments{
\item{config}{Model configuration list.}
}
\value{
Module whose forward(input_ids, ...) returns
  \code{list(last_hidden_state, hidden_states)}: the final hidden
  state [B, S, hidden_size] and the list of per-layer hidden states.
}
\description{
Full Gemma3 text encoder model.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{encode_with_gemma3}
\alias{encode_with_gemma3}
\title{Encode text with Gemma3 for LTX-2}
\usage{
encode_with_gemma3(prompts, model = NULL, tokenizer = NULL,
                   max_sequence_length = 1024L, device = "cuda",
                   dtype = "float16", verbose = TRUE)
}
\arguments{
\item{prompts}{Character vector of prompts.}

\item{model}{Gemma3 text model (or path to load from).}

\item{tokenizer}{Gemma3 tokenizer (or path to load from).}

\item{max_sequence_length}{Integer. Maximum sequence length.}

\item{device}{Character. Device for computation.}

\item{dtype}{Character. Data type.}

\item{verbose}{Logical. Print progress.}
}
\value{
List with prompt_embeds (raw stacked hidden states,
  shape \code{[batch, seq_len, hidden_size, num_layers + 1]}) and
  prompt_attention_mask.
}
\description{
Full pipeline for encoding text prompts using Gemma3 text encoder.
Returns the raw stacked per-layer hidden states (embedding layer plus all
transformer layers) for downstream connector modules, which handle
normalization and projection themselves.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{jit_vae_ltx23}
\alias{jit_vae_ltx23}
\title{JIT-Traced Decode for the LTX-2.3 VAEs and Vocoder}
\description{
The video/audio decoders and the vocoder are static feed-forward
graphs, so \code{torch::jit_trace} converts them wholesale: one
R-to-libtorch crossing per forward, intermediates freed eagerly by
libtorch instead of accumulating as R handles until gc. Traces are
shape-specialized (runtime sizes bake into the graph as constants;
a mismatched input errors), so they are cached per instance, input
shape, dtype, device, and call tag, and re-traced on a miss.
}
\details{
A trace captures the module's weight tensors, which would pin them
on the GPU across phase offloads; the pipeline releases all traces
whenever a component offloads (\code{.ltx23_release_vae_traces}).

Tracing hazard on this lantern build: if the allocator callback
runs R's gc \emph{during} trace recording (memory pressure), the
recorded graph can capture garbage argument values (observed as
corrupted \code{narrow} starts on the full-size decoder; verified
5/5 clean once gc cannot fire mid-trace). Defenses, in order: a
gc + cache flush right before each trace so pressure starts near
zero, \code{tryCatch} around trace and replay, and a one-time
validation of every fresh trace against the eager output — any
mismatch permanently blacklists that shape and runs eager. With
those in place the traced path cannot corrupt output — but per
render it measured slower than eager (traces are released on phase
offload, so every render re-pays trace + validation), so it stays
opt-in: \code{options(diffuseR.jit_vae = TRUE)}.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{condition_ltx23}
\alias{condition_ltx23}
\title{LTX-2.3 Prefix Conditioning (image-to-video, video continuation)}
\description{
Fresh R port of the frame-conditioning mechanics from the diffusers
reference (Apache-2.0, pipelines/ltx2/pipeline_ltx2_image2video.py
and pipeline_ltx2_condition.py), restricted to prefix conditioning
at latent index 0 with strength 1: a single start image (i2v) or the
leading pixel frames of a previous clip (continuation). Conditioned
latent tokens are initialized from the VAE-encoded pixels, see a
per-token timestep of zero, and are frozen through the Euler loop.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{img2img}
\alias{img2img}
\title{Image-to-Image Generation with Stable Diffusion}
\usage{
img2img(input_image, prompt, negative_prompt = NULL, img_dim = 512,
        model_name = c("sd21", "sdxl"), pipeline = NULL, devices = "auto",
        unet_dtype_str = "float16", download_models = FALSE,
        scheduler = "ddim", num_inference_steps = 50, strength = 0.8,
        guidance_scale = 7.5, seed = NULL, save_file = TRUE, filename = NULL,
        metadata_path = NULL, use_native_decoder = FALSE,
        use_native_text_encoder = FALSE, use_native_unet = FALSE, ...)
}
\arguments{
\item{input_image}{Path to the input image or a tensor representing the image.}

\item{prompt}{Text prompt to guide the image generation.}

\item{negative_prompt}{Optional negative prompt to guide the image generation.`}

\item{img_dim}{Dimension of the output image (default: 512).}

\item{model_name}{Name of the Stable Diffusion model to use (default: "sd21").}

\item{pipeline}{Optional pre-loaded pipeline. If `NULL`, it will be loaded based on `model_name`.}

\item{devices}{A named list of devices for each model component (e.g., `list(unet = "cuda", decoder = "cpu", text_encoder = "cpu", encoder = "cpu")`).}

\item{unet_dtype_str}{Optional A character for dtype of the unet component (typically "torch_float16" for cuda and "torch_float32" for cpu).}

\item{download_models}{Logical indicating whether to download models if not found (default: FALSE).}

\item{scheduler}{Scheduler to use for the diffusion process (default: "ddim").}

\item{num_inference_steps}{Number of diffusion steps (default: 50).}

\item{strength}{Strength of the image-to-image transformation (default: 0.8).}

\item{guidance_scale}{Scale for classifier-free guidance (default: 7.5).}

\item{seed}{Random seed for reproducibility (default: NULL).}

\item{save_file}{Logical indicating whether to save the generated image.}

\item{filename}{Optional filename for saving the image. If `NULL`, a default name is generated.}

\item{metadata_path}{Path to save metadata (default: NULL).}

\item{use_native_decoder}{Logical; if TRUE, uses native R torch decoder instead of TorchScript.
Native decoder has better GPU compatibility (especially Blackwell).}

\item{use_native_text_encoder}{Logical; if TRUE, uses native R torch text encoder instead of TorchScript.
Native text encoder has better GPU compatibility (especially Blackwell).}

\item{use_native_unet}{Logical; if TRUE, uses native R torch UNet instead of TorchScript.
Native UNet has better GPU compatibility (especially Blackwell).}

\item{...}{Additional arguments for future use.}
}
\value{
An image array and metadata
}
\description{
This function generates an image based on an input image and a text prompt using the Stable Diffusion model.
It allows for various configurations such as model name, device, scheduler, and more.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_img_pos_ids}
\alias{zimage_img_pos_ids}
\title{Build Z-Image latent image position ids}
\usage{
zimage_img_pos_ids(h_tokens, w_tokens, start0, f_tokens = 1L, device = "cpu")
}
\arguments{
\item{h_tokens}{Integer. Token grid height (latent height / patch).}

\item{w_tokens}{Integer. Token grid width (latent width / patch).}

\item{start0}{Integer. First-axis start, cap_padded_len + 1.}

\item{f_tokens}{Integer. Token grid frames; 1 for txt2img.}

\item{device}{Device for the resulting tensor.}
}
\value{
Float tensor of shape [padded token count, 3].
}
\description{
Image tokens use axis 1 for the frame index offset past the caption
(start0 = cap_padded_len + 1), axis 2 for the token row and axis 3 for
the token column. Trailing pad tokens (token count not a multiple of
32) sit at (0, 0, 0). Reference: patchify_and_embed / _pad_with_ids.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{t5_text_encoder}
\alias{t5_text_encoder}
\title{T5 Text Encoder (T5-v1.1)}
\description{
Fresh R port of the T5 encoder stack from HuggingFace transformers
(Apache-2.0, src/transformers/models/t5/modeling_t5.py), as used by
FLUX's second text encoder (T5-v1.1-XXL: 24 layers, d_model 4096,
64 heads x d_kv 64, gated-GELU FFN). Distinctives faithfully carried
over: RMS layer norms (no mean subtraction), no biases anywhere, no
1/sqrt(d) attention scaling (folded into the weights), and a shared
relative position bias computed once from block 1's embedding and
added to every layer's attention logits. Module field names mirror
the checkpoint keys (minus the \code{encoder.} prefix).
}
\details{
FLUX passes no attention mask - padding tokens attend and are
attended to - so none is implemented.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{serve}
\alias{serve}
\title{Serve diffuseR over HTTP}
\usage{
serve(port = 7812L, model = c("flux2", "zimage", "flux1", "ltx"),
      device = "cuda", token = NULL, max_pixels = 1024L^2, max_frames = 161L,
      max_steps = 50L, max_pixel_frames = NULL, max_prompts = 32L,
      timeout = 300L, max_body = 1024L^2, warmup = TRUE)
}
\arguments{
\item{port}{Integer. TCP port. Default 7812 (cornball serve range: whisper 7809, chatterbox 7810, qwen3 TTS 7811).}

\item{model}{One of "flux2", "zimage", "flux1" (images) or "ltx"
(video). SD 2.1/SDXL are not served yet.}

\item{device}{Character. "cuda" or "cpu".}

\item{token}{Character or NULL. Shared secret; when set, requests
must send \code{Authorization: Bearer <token>}.}

\item{max_pixels}{Integer. Maximum width x height accepted (images
and video frames). Default 1024^2.}

\item{max_frames}{Integer. Maximum video frame count. Default 161.}

\item{max_steps}{Integer. Maximum image inference steps. Default 50.}

\item{max_pixel_frames}{Numeric. Joint video budget: width x height
x frames must stay under it (NULL = max_pixels x 121, so full-
resolution clips top out at 121 frames and longer clips must
shrink spatially).}

\item{max_prompts}{Integer. Bound on the per-prompt connector-embed
cache for "ltx" (~9 MB per entry, LRU-evicted). Default 32.}

\item{timeout}{Integer. Per-connection I/O timeout in seconds.}

\item{max_body}{Integer. Maximum request body bytes. Default 1 MB
(bodies are JSON).}

\item{warmup}{Logical. Image models: run one small generation at
startup so the first request doesn't pay tracing and allocator
growth. Ignored for "ltx".}
}
\value{
Does not return normally; runs until interrupted.
}
\description{
Starts a blocking HTTP server that loads one model and answers
OpenAI-style generation requests. Never downloads weights: if the
model's artifacts are missing, startup stops with the loader's
pointer to the explicit \code{download_*()} function.
}
\details{
Endpoints:
\itemize{
\item \code{GET /health} - liveness probe, returns
\code{{"status":"ok","model":...}}. The server is
single-threaded, so health only answers between requests.
\item \code{POST /v1/images/generations} - image models. JSON body
\code{{prompt, size, seed, steps}} (\code{size} like
\code{"1024x1024"}; \code{n > 1} is not supported). Returns
\code{{created, data: [{b64_json}]}} with a base64 PNG.
\item \code{POST /v1/videos/generations} - \code{model = "ltx"}
only. JSON body \code{{prompt, width, height, num_frames,
frame_rate, seed}}. Returns raw \code{video/mp4} bytes. Video
generation takes minutes: give your client a matching timeout.
}

The server is single-threaded and runs until interrupted. Run it
under a process supervisor (systemd, tmux); an example unit ships
with the package: \code{system.file("diffuser.service",
package = "diffuseR")}.

Security: base R's \code{serverSocket} binds all interfaces, so the
server is reachable by anything that can reach the machine. Keep it
behind a firewall or reverse proxy, and/or set \code{token}: when
set, every request must carry \code{Authorization: Bearer <token>}
or it is refused with 401. Generation size is capped by
\code{max_pixels}/\code{max_frames}; oversized requests get 400. A
CUDA out-of-memory during a request answers 500 and then exits the
process (status 70) so a supervisor restarts it with clean GPU
state rather than serving on with stranded components.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_pos_embed}
\alias{flux_pos_embed}
\title{Compute FLUX rotary frequencies from position ids}
\usage{
flux_pos_embed(ids, axes_dim = c(16L, 56L, 56L), theta = 10000)
}
\arguments{
\item{ids}{Tensor of shape [S, 3]: concatenated text ids (all zero)
and image ids from \code{flux_prepare_latent_image_ids}.}

\item{axes_dim}{Integer vector of per-axis rotary dims; must sum to
the attention head dim. FLUX uses c(16, 56, 56).}

\item{theta}{Numeric. RoPE base frequency.}
}
\value{
List of two tensors (cos, sin), each [S, sum(axes_dim)],
  float32, on the device of \code{ids}.
}
\description{
Per-axis 1D rotary frequencies (interleaved-real convention), computed
in float64 on CPU and concatenated over the axes. Reference:
FluxPosEmbed with get_1d_rotary_pos_embed(repeat_interleave_real=TRUE,
use_real=TRUE, freqs_dtype=float64).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_upsampler3d}
\alias{ltx23_video_upsampler3d}
\title{Pixel-shuffle 3D upsampler}
\usage{
ltx23_video_upsampler3d(in_channels, stride = c(1L, 1L, 1L), residual = FALSE,
                        upscale_factor = 1L, spatial_padding_mode = "zeros")
}
\arguments{
\item{in_channels}{Integer.}

\item{stride}{Length-3 integer vector (t, h, w).}

\item{residual}{Logical. Add the rearranged input as a residual.}

\item{upscale_factor}{Integer.}

\item{spatial_padding_mode}{Character.}
}
\value{
Module whose forward(hidden_states, causal) returns the
  depth-to-space upsampled tensor: spatial and temporal extents grow
  by \code{stride}, channels shrink to match.
}
\description{
Conv followed by channel-to-space/time rearrangement, with an optional
channel-repeat residual and an upscale factor that divides the conv
output channels.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{is_blackwell_gpu}
\alias{is_blackwell_gpu}
\title{Check if GPU is Blackwell Architecture}
\usage{
is_blackwell_gpu()
}
\value{
Logical. TRUE if Blackwell GPU detected.
}
\description{
Blackwell GPUs (RTX 50xx) may need special handling.
}
\examples{
# Soft probe: FALSE on any machine without a Blackwell card, and on
# machines where torch has no lantern binaries.
is_blackwell_gpu()
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{sdxl_pipeline_from_safetensors}
\alias{sdxl_pipeline_from_safetensors}
\title{Assemble a native SDXL pipeline from a diffusers safetensors directory}
\usage{
sdxl_pipeline_from_safetensors(diffusers_dir, devices = NULL,
                               unet_dtype = NULL, verbose = TRUE)
}
\arguments{
\item{diffusers_dir}{Directory with \code{unet/}, \code{vae/},
\code{text_encoder/}, \code{text_encoder_2/} subdirectories.}

\item{devices}{Named list of component devices (\code{unet},
\code{decoder}, \code{text_encoder}, \code{text_encoder2});
defaults to all-CPU. \code{text_encoder2} defaults to the
\code{text_encoder} device when unset.}

\item{unet_dtype}{A torch dtype for the UNet (default float16 on CUDA,
float32 on CPU).}

\item{verbose}{Logical.}
}
\value{
A list with \code{unet}, \code{decoder}, \code{text_encoder},
  \code{text_encoder2}, \code{vae_scaling}, and \code{native_decode}
  (\code{TRUE}; the decoder already applies \code{post_quant_conv}).
}
\description{
Builds the two native CLIP text encoders, the native SDXL UNet, and the
VAE decode module from a diffusers directory using the
\code{*_from_safetensors} constructors, places each on its component
device, and returns the list the \code{\link{txt2img_sdxl}} denoise
loop expects.
}
\details{
Both encoders return their penultimate hidden state (SDXL feeds the
UNet the concatenated \code{[text_encoder (768) | text_encoder_2
(1280)]} = 2048-dim penultimate embeds); \code{text_encoder} uses
\code{quick_gelu} (OpenAI CLIP ViT-L) and \code{text_encoder_2} uses
exact GELU (OpenCLIP bigG). The pooled \code{text_embeds} come from
\code{text_encoder_2}'s full stack. The VAE decodes in float32 (the
SDXL fp16 VAE overflows in fp16) and its \code{scaling_factor} is read
from \code{vae/config.json}.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_ada_layer_norm_zero}
\alias{flux_ada_layer_norm_zero}
\title{FLUX adaLN-Zero modulation (double-stream)}
\usage{
flux_ada_layer_norm_zero(dim)
}
\arguments{
\item{dim}{Integer. Model dimension.}
}
\value{
Module whose forward(x, emb) returns
  \code{list(x_norm, gate_msa, shift_mlp, scale_mlp, gate_mlp)}.
}
\description{
Projects the conditioning embedding to six modulation vectors and
returns the msa-modulated input plus the remaining parameters.
Reference: diffusers AdaLayerNormZero.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{text_encoder2_native}
\alias{text_encoder2_native}
\title{Native CLIP Text Encoder 2 (OpenCLIP ViT-bigG for SDXL)}
\usage{
text_encoder2_native(vocab_size = 49408, context_length = 77, embed_dim = 1280,
                     num_layers = 32, num_heads = 20, mlp_dim = 5120,
                     return_penultimate = FALSE)
}
\arguments{
\item{vocab_size}{Vocabulary size (default 49408)}

\item{context_length}{Maximum sequence length (default 77)}

\item{embed_dim}{Embedding dimension (default 1280)}

\item{num_layers}{Number of transformer layers (default 32)}

\item{num_heads}{Number of attention heads (default 20)}

\item{mlp_dim}{MLP hidden dimension (default 5120)}

\item{return_penultimate}{Return the penultimate hidden state
alongside the pooled output (SDXL's \code{hidden_states[-2]}
prompt embeds; the pooled output still comes from the full stack)}
}
\value{
An nn_module representing the text encoder
}
\description{
Native R torch implementation of OpenCLIP text encoder used in SDXL.
Returns both hidden states and pooled output.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_per_channel_rms_norm}
\alias{ltx23_per_channel_rms_norm}
\title{Per-channel RMS normalization}
\usage{
ltx23_per_channel_rms_norm(eps = 1e-08)
}
\arguments{
\item{eps}{Numeric. Stability epsilon.}
}
\value{
Module whose forward(x) returns \code{x} divided by its
  per-channel root mean square, a tensor of the same shape.
}
\description{
Normalizes by the root-mean-square across the channel dimension
(dim 2 of [B, C, F, H, W]); no learned parameters.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_up_block3d}
\alias{ltx23_video_up_block3d}
\title{LTX video up block}
\usage{
ltx23_video_up_block3d(in_channels, out_channels = NULL, num_layers = 1L,
                       resnet_eps = 1e-06, spatio_temporal_scale = TRUE,
                       upsample_type = "spatiotemporal",
                       upsample_residual = FALSE, upscale_factor = 1L,
                       spatial_padding_mode = "zeros")
}
\arguments{
\item{num_layers}{Integer.}

\item{resnet_eps}{Numeric.}

\item{spatio_temporal_scale}{Logical.}

\item{upsample_type}{"spatial", "temporal", or "spatiotemporal".}

\item{upsample_residual}{Logical.}

\item{upscale_factor}{Integer.}

\item{spatial_padding_mode}{Character.}

\item{in_channels,out_channels}{Integers.}
}
\value{
Module whose forward(hidden_states, causal) returns the
  stage output: the optional input projection and upsampler, then the
  resnet stack applied in sequence.
}
\description{
Optional channel-changing conv-in ResNet, pixel-shuffle upsampler,
then a ResNet stack at the output channel count.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_down_block3d}
\alias{ltx23_video_down_block3d}
\title{LTX video down block}
\usage{
ltx23_video_down_block3d(in_channels, out_channels = NULL, num_layers = 1L,
                         resnet_eps = 1e-06, spatio_temporal_scale = TRUE,
                         downsample_type = "spatiotemporal",
                         spatial_padding_mode = "zeros")
}
\arguments{
\item{num_layers}{Integer. ResNet count.}

\item{resnet_eps}{Numeric.}

\item{spatio_temporal_scale}{Logical. Whether to downsample at all.}

\item{downsample_type}{"spatial", "temporal", or "spatiotemporal".}

\item{spatial_padding_mode}{Character.}

\item{in_channels,out_channels}{Integers.}
}
\value{
Module whose forward(hidden_states, causal) returns the
  stage output: the resnet stack applied in sequence, then the
  optional downsampler.
}
\description{
ResNet stack (at the input channel count) followed by a
pixel-unshuffle downsampler that also changes the channel count.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{CLIPTokenizer}
\alias{CLIPTokenizer}
\title{Tokenize a prompt}
\usage{
CLIPTokenizer(prompt,
              merges = system.file("tokenizer/merges.txt", package = "diffuseR"),
              vocab_file = system.file("tokenizer/vocab.json", package = "diffuseR"),
              pad_token = 0L)
}
\arguments{
\item{prompt}{A character string prompt describing the image to generate.}

\item{merges}{Path to the merges file (BPE merges).}

\item{vocab_file}{Path to the vocabulary file (token->id mapping).}

\item{pad_token}{The token ID used for padding (default is 0).}
}
\value{
A 2D torch tensor of shape c(1, 77) containing the token IDs.
}
\description{
Tokenize a prompt
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unet_sdxl_native_from_safetensors}
\alias{unet_sdxl_native_from_safetensors}
\title{Build a native SDXL UNet from a diffusers safetensors directory}
\usage{
unet_sdxl_native_from_safetensors(path, verbose = TRUE, ...)
}
\arguments{
\item{path}{Path to the UNet directory or its single-file checkpoint.}

\item{verbose}{Print how many parameters were loaded.}

\item{...}{Overrides for \code{\link{unet_sdxl_native}} constructor args.}
}
\value{
The native SDXL UNet in eval mode.
}
\description{
The safetensors counterpart to
\code{\link{unet_sdxl_native_from_torchscript}}: constructs
\code{\link{unet_sdxl_native}} and loads its weights from
\code{unet/diffusion_pytorch_model.safetensors}. Validated against the
cached \code{stabilityai/stable-diffusion-xl-base-1.0} UNet (all 1680
parameters map with matching shapes). Pass constructor overrides
through \code{...} for a variant checkpoint.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_zimage_turbo}
\alias{download_zimage_turbo}
\title{Download Z-Image-Turbo and build the quantized artifact}
\usage{
download_zimage_turbo(quantize = TRUE, precision = c("auto", "fp8", "nf4"),
                      output_dir = NULL, text_encoders = TRUE, verbose = TRUE)
}
\arguments{
\item{quantize}{Logical. Build the quantized artifact.}

\item{precision}{"auto" (default: fp8 when safetensors supports
float8, else nf4), "fp8" (~6.3 GB, GPU-resident; near-bf16
quality), or "nf4" (~3.6 GB).}

\item{output_dir}{Directory for the quantized artifact.}

\item{text_encoders}{Logical. Also fetch the Qwen3-4B text encoder,
tokenizer, VAE, and scheduler config (~8.2 GB).}

\item{verbose}{Logical.}
}
\value{
Invisibly, a list with \code{transformer_dir},
  \code{artifact_dir}, and \code{support} (named file paths).
}
\description{
Skips work already done: a valid quantized manifest short-circuits
the transformer download; cached files are not re-fetched. No token
is needed (the repo is ungated). The float32 transformer source
(~24.6 GB in the HuggingFace cache) may be deleted after
quantization.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_load_group}
\alias{ltx23_load_group}
\title{Stream a checkpoint key group into a module}
\usage{
ltx23_load_group(ckpt, keys, module, map_key = identity, verbose = TRUE,
                 gc_every = 50L)
}
\arguments{
\item{ckpt}{An \code{ltx23_checkpoint}.}

\item{keys}{Character vector of checkpoint keys to load (one group
from \code{\link{ltx23_split_keys}}).}

\item{module}{A torch nn_module to populate.}

\item{map_key}{Function mapping a checkpoint key to the module's
parameter/buffer name, or NA to skip the key deliberately.}

\item{verbose}{Logical. Report progress and coverage.}

\item{gc_every}{Integer. Run \code{gc()} after this many tensors.}
}
\value{
Invisibly, a list with \code{unmapped} (checkpoint keys that
  found no destination), \code{skipped} (keys the mapper declined),
  and \code{unfilled} (module parameters/buffers never written).
  A perfectly loaded group has zero \code{unmapped} and zero
  \code{unfilled}.
}
\description{
Reads tensors one at a time from an open checkpoint and copies them
into the matching parameters/buffers of \code{module}. Destination
names are derived by \code{map_key}; \code{$copy_()} handles any
dtype/device conversion, so the module may already live on its target
device in its target dtype.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_transformer}
\alias{flux2_transformer}
\title{FLUX.2 transformer model}
\usage{
flux2_transformer(in_channels = 128L, num_layers = 5L, num_single_layers = 20L,
                  attention_head_dim = 128L, num_attention_heads = 24L,
                  joint_attention_dim = 7680L, mlp_ratio = 3,
                  timestep_guidance_channels = 256L,
                  axes_dims_rope = c(32L, 32L, 32L, 32L), rope_theta = 2000,
                  eps = 1e-06, out_channels = NULL)
}
\arguments{
\item{in_channels}{Integer. Packed latent channels (128).}

\item{num_layers}{Integer. Double-stream block count (klein-4B: 5).}

\item{num_single_layers}{Integer. Single-stream block count (20).}

\item{attention_head_dim}{Integer. Per-head dimension.}

\item{num_attention_heads}{Integer. Attention heads.}

\item{joint_attention_dim}{Integer. Text embedding dim (7680).}

\item{mlp_ratio}{Numeric. Feed-forward multiplier (3.0).}

\item{timestep_guidance_channels}{Integer. Sinusoid width (256).}

\item{axes_dims_rope}{Integer vector. Per-axis rotary dims.}

\item{rope_theta}{Numeric. Rotary base frequency (2000).}

\item{eps}{Numeric. Norm epsilon.}

\item{out_channels}{Integer or NULL. Defaults to \code{in_channels}.}
}
\value{
Module whose forward(hidden_states, encoder_hidden_states,
  timestep, image_rotary_emb) returns the predicted velocity for the
  image tokens [B, S_img, out_channels]. \code{timestep} is in sigma
  space (0-1); it is scaled by 1000 internally.
}
\description{
Shared modulation computed once per forward; double blocks over
separate text/image streams, then single (parallel) blocks over the
concatenated [text; image] sequence. Rotary embeddings are
precomputed by the caller with \code{\link{flux_pos_embed}}
(\code{axes_dim = c(32, 32, 32, 32)}, \code{theta = 2000}) over the
concatenated [text; image] 4-axis position ids.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{qwen3_text_encoder}
\alias{qwen3_text_encoder}
\title{Qwen3 Text Encoder}
\description{
Fresh R port of the Qwen3 decoder stack from HuggingFace transformers
(Apache-2.0, src/transformers/models/qwen3/), used by FLUX.2 klein as
its text encoder (Qwen3-4B: 36 layers, hidden 2560, 32 query / 8 KV
heads, head_dim 128, SwiGLU 9728, RoPE theta 1e6). The pipeline
consumes mid-stack hidden states (layers 9, 18, 27 for klein-4B)
concatenated per token, so the forward runs only as deep as the last
requested layer; the LM head is never needed (embeddings are tied).
Causal attention with the tokenizer's padding mask, matching the
reference exactly.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.resident_on_gpu_count}
\alias{.resident_on_gpu_count}
\title{How many components actually have their tensors on the GPU}
\usage{
.resident_on_gpu_count(staging)
}
\arguments{
\item{staging}{A named list of staging sets.}
}
\value{
Integer. Number of components whose live tensors are on CUDA.
}
\description{
Ground truth, as opposed to the handle's declared state. The two can
disagree: a pipeline built with \code{phase_offload = TRUE} swaps each
component back to pinned host memory as its phase finishes, so after a
render the handle is still "active" while the card holds nothing. A
broker deciding who to evict needs the measurement, not the claim.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flowmatch_scheduler_step}
\alias{flowmatch_scheduler_step}
\title{Perform a FlowMatch scheduler step}
\usage{
flowmatch_scheduler_step(model_output, timestep, sample, schedule,
                         generator = NULL)
}
\arguments{
\item{model_output}{torch tensor. The output from the diffusion model
(velocity prediction).}

\item{timestep}{Numeric. The current timestep.}

\item{sample}{torch tensor. The current noisy sample.}

\item{schedule}{List. The FlowMatch scheduler object.}

\item{generator}{torch generator or NULL. Random generator for reproducibility.}
}
\value{
A list containing:
  \describe{
    \item{prev_sample}{The denoised sample at the previous timestep}
    \item{schedule}{The updated scheduler with incremented step_index}
  }
}
\description{
Performs a single denoising step using Euler integration. This is the
core sampling function for FlowMatch models.
}
\details{
The FlowMatch Euler step is remarkably simple:
\code{prev_sample = sample + dt * model_output}
where \code{dt = sigma_next - sigma_current}.

This implements the Euler method for solving the probability flow ODE
in continuous normalizing flows.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unet_native_from_safetensors}
\alias{unet_native_from_safetensors}
\title{Build a native SD21 UNet from a diffusers safetensors directory}
\usage{
unet_native_from_safetensors(path, verbose = TRUE, ...)
}
\arguments{
\item{path}{Path to the UNet directory or its single-file checkpoint.}

\item{verbose}{Print how many parameters were loaded.}

\item{...}{Overrides for \code{\link{unet_native}} constructor args.}
}
\value{
The native SD21 UNet in eval mode.
}
\description{
The safetensors counterpart to
\code{\link{unet_native_from_torchscript}}: constructs
\code{\link{unet_native}} and loads its weights from
\code{unet/diffusion_pytorch_model.safetensors} (no TorchScript, so it
works on Blackwell). The default construction matches the canonical
Stable Diffusion 2.1 UNet; pass constructor overrides through
\code{...} for a variant checkpoint (the loader fails loudly on any
shape mismatch, so a wrong architecture surfaces immediately rather
than loading silently wrong weights).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_causal_conv3d}
\alias{ltx23_causal_conv3d}
\title{Causal 3D convolution}
\usage{
ltx23_causal_conv3d(in_channels, out_channels, kernel_size = 3L, stride = 1L,
                    spatial_padding_mode = "zeros")
}
\arguments{
\item{kernel_size}{Integer or length-3 vector (t, h, w).}

\item{stride}{Integer or length-3 vector.}

\item{spatial_padding_mode}{Character. Conv padding mode.}

\item{in_channels,out_channels}{Integers.}
}
\value{
Module whose forward(hidden_states, causal) returns the 3-D
  convolution of the input. With \code{causal = TRUE} the temporal
  axis is left-padded by replicating the first frame, so no output
  frame sees a later input frame.
}
\description{
Spatial padding is handled by the convolution; temporal padding
replicates the first frame (causal) or both edge frames (non-causal),
chosen at call time.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.ltx23_jit_pack_block}
\alias{.ltx23_jit_pack_block}
\title{Pack a transformer block's weights for the JIT stack}
\usage{
.ltx23_jit_pack_block(block)
}
\arguments{
\item{block}{An NF4-quantized \code{ltx23_transformer_block}.}
}
\value{
List of 114 tensors.
}
\description{
Returns the block's tensors in the fixed 114-slot layout consumed by
the compiled \code{stack_nf4}/\code{block_nf4} TorchScript functions.
Tensor handles are borrowed, not copied.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_rms_norm}
\alias{ltx23_rms_norm}
\title{RMS normalization}
\usage{
ltx23_rms_norm(dim, eps = 1e-06, elementwise_affine = TRUE)
}
\arguments{
\item{dim}{Integer. Normalized dimension size.}

\item{eps}{Numeric. Stability epsilon.}

\item{elementwise_affine}{Logical. Learn a scale weight.}
}
\value{
Module whose forward(x) returns \code{x} RMS-normalized over
  the last axis and cast back to the input dtype, a tensor of the
  same shape.
}
\description{
Variance is computed in float32; the result is cast back to the input
dtype (or the weight dtype when elementwise affine).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{encode_with_t5}
\alias{encode_with_t5}
\title{Encode prompts with the T5 encoder}
\usage{
encode_with_t5(prompts, model, tokenizer, max_sequence_length = 256L,
               device = NULL)
}
\arguments{
\item{prompts}{Character vector.}

\item{model}{A \code{\link{t5_encoder}}.}

\item{tokenizer}{A \code{\link{unigram_tokenizer}}.}

\item{max_sequence_length}{Integer. Fixed token length (schnell: 256).}

\item{device}{Device for the input ids (defaults to the model's).}
}
\value{
Tensor [length(prompts), max_sequence_length, d_model].
}
\description{
Tokenizes with \code{\link{encode_unigram}} (right padding to
\code{max_sequence_length}) and runs the encoder. Matching the FLUX
reference pipeline, no attention mask is used.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_attention}
\alias{ltx23_attention}
\title{LTX-2 attention layer}
\usage{
ltx23_attention(query_dim, heads = 8L, kv_heads = NULL, dim_head = 64L,
                bias = TRUE, cross_attention_dim = NULL, out_bias = TRUE,
                norm_eps = 1e-06, norm_elementwise_affine = TRUE,
                rope_type = "split", apply_gated_attention = FALSE)
}
\arguments{
\item{query_dim}{Integer. Query feature dimension.}

\item{dim_head}{Integer. Per-head dimension.}

\item{cross_attention_dim}{Integer or NULL. Key/value input dimension
(NULL for self-attention).}

\item{norm_eps}{Numeric. RMS norm epsilon.}

\item{norm_elementwise_affine}{Logical. RMS norms carry weights.}

\item{rope_type}{"split" or "interleaved".}

\item{apply_gated_attention}{Logical. Add per-head sigmoid output gates.}

\item{heads,kv_heads}{Integers. Attention head counts.}

\item{bias,out_bias}{Logicals. Projection biases.}
}
\value{
Module whose forward(hidden_states, ...) returns the attended
  states [B, S, query_dim] after the output projection.
}
\description{
Attention with RMS q/k norms across heads, optional per-head output
gating (LTX-2.3), separate query/key RoPE (for a2v/v2a cross
attention), and optional STG perturbation (skip attention, use the
value projection).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_vae_decoder}
\alias{flux2_vae_decoder}
\title{FLUX.2 VAE decoder}
\usage{
flux2_vae_decoder(latent_channels = 32L,
                  block_channels = c(512L, 512L, 256L, 128L), norm_groups = 32L)
}
\arguments{
\item{latent_channels}{Integer (32 for FLUX.2).}

\item{block_channels}{Decoder block channels (reversed encoder
block_out_channels).}

\item{norm_groups}{Integer. Group norm groups.}
}
\value{
Module whose forward(z) decodes [B, 32, H, W] latents to
  [B, 3, 8H, 8W] images; \code{$bn$running_mean} /
  \code{$bn$running_var} carry the normalization statistics.
}
\description{
The AutoencoderKLFlux2 decode path: post_quant_conv (1x1, 32
channels) followed by the standard AutoencoderKL decoder body
(reused from \code{\link{vae_decoder_native}}), plus the BatchNorm
running statistics used for latent (de)normalization. Reference:
src/diffusers/models/autoencoders/autoencoder_kl_flux2.py.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.resident_pinned_bytes}
\alias{.resident_pinned_bytes}
\title{Total pinned host bytes across a staging set}
\usage{
.resident_pinned_bytes(staging)
}
\arguments{
\item{staging}{A named list of staging sets.}
}
\value{
Numeric. Bytes of page-locked host memory held.
}
\description{
Total pinned host bytes across a staging set
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_audio_causal_conv2d}
\alias{ltx23_audio_causal_conv2d}
\title{Causal 2D convolution for audio spectrograms}
\usage{
ltx23_audio_causal_conv2d(in_channels, out_channels, kernel_size = 3L,
                          stride = 1L, causality_axis = "height")
}
\arguments{
\item{kernel_size}{Integer or length-2 vector.}

\item{stride}{Integer.}

\item{causality_axis}{"height", "width", "width-compatibility", or "none".}

\item{in_channels,out_channels}{Integers.}
}
\value{
Module whose forward(x) returns the convolved tensor, padded
  so that each output frame depends only on current and earlier
  input frames.
}
\description{
Pads asymmetrically along the causal axis ("height" = time frames for
LTX audio) before an unpadded Conv2d.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_fp8_linear}
\alias{ltx23_fp8_linear}
\title{FP8 linear layer}
\usage{
ltx23_fp8_linear(out_features, in_features, bias = TRUE)
}
\arguments{
\item{bias}{Logical.}

\item{out_features,in_features}{Integers.}
}
\value{
Module whose forward(x) returns the linear projection of
  \code{x}, with the fp8 weight bytes transferred and cast up to the
  compute dtype for the matmul. Same result as an
  \code{nn_linear} of the same shape, at a quarter of the resident
  weight bytes.
}
\description{
Weight lives as float8_e4m3fn plus a float32 scale in plain module
fields (so \code{$to(device)} moves only the bias); the forward pass
ships 1 byte/param to the input's device, upcasts, rescales, and runs
\code{nnf_linear}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_unet_weights}
\alias{load_unet_weights}
\title{Load weights from TorchScript UNet into native UNet}
\usage{
load_unet_weights(native_unet, torchscript_path, verbose = TRUE)
}
\arguments{
\item{native_unet}{Native UNet module}

\item{torchscript_path}{Path to TorchScript UNet .pt file}

\item{verbose}{Print loading progress}
}
\value{
The native UNet with loaded weights (invisibly)
}
\description{
Load weights from TorchScript UNet into native UNet
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{auto_devices}
\alias{auto_devices}
\title{Auto-Configure Device Assignment}
\usage{
auto_devices(model = "sdxl", strategy = "auto")
}
\arguments{
\item{model}{Character. Model type: "sd21" or "sdxl".}

\item{strategy}{Character. Memory strategy: "auto" (default), "full_gpu",
"unet_gpu", or "cpu_only". See Details.}
}
\value{
A named list of device assignments suitable for `models2devices()`.
}
\description{
Automatically determines optimal device configuration for diffusion model
components based on available VRAM (via nvidia-smi) and GPU architecture.
}
\details{
Strategies:
\describe{
  \item{"auto"}{Detect free VRAM and choose the best strategy}
  \item{"full_gpu"}{All components on CUDA (10GB+ free VRAM for SDXL)}
  \item{"unet_gpu"}{Only unet on CUDA, rest on CPU (6GB+ for SDXL)}
  \item{"cpu_only"}{All components on CPU}
}

On Blackwell GPUs (RTX 50xx), "unet_gpu" is forced due to TorchScript
compatibility issues, regardless of available VRAM. The native modules
(`use_native_unet` and friends) do not have this restriction.
}
\examples{
# Force a strategy: no GPU or nvidia-smi needed.
str(auto_devices("sdxl", strategy = "cpu_only"))

str(auto_devices("sd21", strategy = "unet_gpu"))

# Auto-detect free VRAM and pick a strategy for this machine.
str(auto_devices("sdxl"))
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{CLIPMLP}
\alias{CLIPMLP}
\title{CLIP MLP Block}
\usage{
CLIPMLP(in_dim, hidden_dim, gelu_type = "tanh")
}
\arguments{
\item{in_dim}{Input dimension}

\item{hidden_dim}{Hidden dimension}

\item{gelu_type}{GELU variant: "tanh" (tanh approximation), "quick" (QuickGELU), "exact" (standard GELU)}
}
\description{
Feed-forward network with configurable activation
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{resident_deactivate}
\alias{resident_deactivate}
\title{Release a resident pipeline's VRAM}
\usage{
resident_deactivate(res, release = TRUE)
}
\arguments{
\item{res}{A \code{diffuseR_resident} handle.}

\item{release}{Empty the CUDA caching allocator afterwards. Leave TRUE
unless another handle on the same device is about to reuse the pool.}
}
\value{
Invisibly the handle, with state "inactive".
}
\description{
Re-points every component at its pinned host copy and drops the GPU
storage. Weights are immutable during inference, so the pinned copies
are still current and this moves no bytes: it is a pointer swap plus a
cache release. The handle stays loaded and can be reactivated without
touching the disk.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_upsample1d}
\alias{ltx23_upsample1d}
\title{Anti-aliasing 1D upsampler (transposed low-pass)}
\usage{
ltx23_upsample1d(ratio = 2L, kernel_size = NULL, window_type = "kaiser",
                 persistent = TRUE)
}
\arguments{
\item{ratio}{Integer. Upsampling ratio.}

\item{kernel_size}{Integer or NULL.}

\item{window_type}{"kaiser" (BigVGAN default) or "hann" (final resampler).}

\item{persistent}{Logical. Register the filter as a buffer (present in
checkpoints); FALSE stores the computed filter as a plain field.}
}
\value{
Module whose forward(x) returns \code{x} interpolated up by
  \code{ratio} along the time axis, with the filter padding trimmed
  off.
}
\description{
Anti-aliasing 1D upsampler (transposed low-pass)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_vocoder_with_bwe}
\alias{ltx23_vocoder_with_bwe}
\title{LTX-2.3 vocoder with bandwidth extension}
\usage{
ltx23_vocoder_with_bwe(in_channels = 128L, hidden_channels = 1536L,
                       out_channels = 2L,
                       upsample_kernel_sizes = c(11L, 4L, 4L, 4L, 4L, 4L),
                       upsample_factors = c(5L, 2L, 2L, 2L, 2L, 2L),
                       resnet_kernel_sizes = c(3L, 7L, 11L),
                       resnet_dilations = NULL, bwe_in_channels = 128L,
                       bwe_hidden_channels = 512L,
                       bwe_upsample_kernel_sizes = c(12L, 11L, 4L, 4L, 4L),
                       bwe_upsample_factors = c(6L, 5L, 2L, 2L, 2L),
                       bwe_resnet_kernel_sizes = c(3L, 7L, 11L),
                       bwe_resnet_dilations = NULL, filter_length = 512L,
                       hop_length = 80L, window_length = 512L,
                       num_mel_channels = 64L, input_sampling_rate = 16000L,
                       output_sampling_rate = 48000L)
}
\arguments{
\item{out_channels}{Integer. Audio channels.}

\item{hop_length}{Integer. Mel analysis hop.}

\item{in_channels,bwe_in_channels}{Integers. Flattened mel input channels.}

\item{hidden_channels,bwe_hidden_channels}{Integers.}

\item{upsample_kernel_sizes,upsample_factors,bwe_upsample_kernel_sizes,bwe_upsample_factors}{Integer vectors. Per-stage transposed-conv configs.}

\item{resnet_kernel_sizes,bwe_resnet_kernel_sizes}{Integer vectors.}

\item{resnet_dilations,bwe_resnet_dilations}{Lists of integer vectors.}

\item{filter_length,window_length,num_mel_channels}{Integers. Mel
re-analysis configuration.}

\item{input_sampling_rate,output_sampling_rate}{Integers.}
}
\value{
Module whose forward(mel_spec) returns the
  bandwidth-extended waveform [B, 1, samples], trimmed to the sample
  count implied by the input frames and the rate ratio.
}
\description{
Full mel [B, 2, T, 64] -> 48 kHz stereo waveform pipeline: 16 kHz
vocoder, causal mel re-analysis, BWE vocoder residual added to a
Hann-resampled skip path, clamped to [-1, 1].
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_text_encoder2_safetensors}
\alias{load_text_encoder2_safetensors}
\title{Load HF safetensors weights into the native SDXL text encoder 2}
\usage{
load_text_encoder2_safetensors(native_encoder, path, verbose = TRUE)
}
\arguments{
\item{native_encoder}{Native text encoder 2 module}

\item{path}{Path to model.safetensors (or a directory containing it)}

\item{verbose}{Print loading progress}
}
\value{
The native encoder with loaded weights (invisibly)
}
\description{
Loads a HuggingFace CLIPTextModelWithProjection \code{model.safetensors}
(SDXL's OpenCLIP ViT-bigG \code{text_encoder_2}) into
\code{\link{text_encoder2_native}}. Same key layout as encoder 1, plus
the top-level \code{text_projection.weight} that produces the pooled
\code{text_embeds}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_census}
\alias{ltx23_census}
\title{Summarize checkpoint key coverage}
\usage{
ltx23_census(ckpt)
}
\arguments{
\item{ckpt}{An \code{ltx23_checkpoint}.}
}
\value{
A data.frame with one row per component group and its key count.
}
\description{
Summarize checkpoint key coverage
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_encode_video_frames}
\alias{ltx23_encode_video_frames}
\title{Encode pixel frames to normalized video latents}
\usage{
ltx23_encode_video_frames(vae, frames)
}
\arguments{
\item{vae}{An \code{ltx23_video_vae}.}

\item{frames}{Tensor [1, 3, F, H, W] in [-1, 1] (see
\code{\link{ltx23_preprocess_frames}}).}
}
\value{
Normalized latents [1, 128, F', H/32, W/32] (float32).
}
\description{
VAE encode in "argmax" mode (the distribution mean), then normalize
with the checkpoint's per-channel statistics — the exact inverse of
the decode path.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{GEGLU}
\alias{GEGLU}
\title{GEGLU Feedforward}
\usage{
GEGLU(dim_in, dim_out)
}
\description{
GEGLU Feedforward
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{dit_ltx23_modules}
\alias{dit_ltx23_modules}
\title{LTX-2.3 Transformer Building Blocks}
\description{
Fresh R port of the LTX-2 transformer components from the diffusers
reference (Apache-2.0, src/diffusers/models/transformers/
transformer_ltx2.py and shared modules). Field names mirror the
diffusers module tree so checkpoint keys map 1:1.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{rope_flux}
\alias{rope_flux}
\title{FLUX Rotary Positional Embeddings}
\description{
Fresh R port of the FLUX rotary positional embedding scheme from the
diffusers reference implementation (Apache-2.0,
src/diffusers/models/transformers/transformer_flux.py FluxPosEmbed and
src/diffusers/models/embeddings.py get_1d_rotary_pos_embed /
apply_rotary_emb). FLUX uses the interleaved adjacent-pair convention
(use_real_unbind_dim = -1) with per-axis frequencies computed in
float64 and applied in float32. Text tokens carry all-zero ids, so
they receive the identity rotation.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{text_encoder_native}
\alias{text_encoder_native}
\title{Native CLIP Text Encoder}
\usage{
text_encoder_native(vocab_size = 49408, context_length = 77, embed_dim = 768,
                    num_layers = 12, num_heads = 12, mlp_dim = 3072,
                    apply_final_ln = TRUE, return_penultimate = FALSE,
                    gelu_type = "tanh")
}
\arguments{
\item{vocab_size}{Vocabulary size (default 49408)}

\item{context_length}{Maximum sequence length (default 77)}

\item{embed_dim}{Embedding dimension}

\item{num_layers}{Number of transformer layers}

\item{num_heads}{Number of attention heads}

\item{mlp_dim}{MLP hidden dimension}

\item{apply_final_ln}{Whether to apply final layer norm (default TRUE).
Set to FALSE to match TorchScript exports that don't include final LN.}

\item{return_penultimate}{Return the second-to-last transformer block's
output (\code{hidden_states[-2]}, pre-final-LN) instead of the last
layer. This is what SDXL feeds the UNet cross-attention; final LN is
never applied to the penultimate output (default FALSE).}

\item{gelu_type}{GELU variant: "tanh" (matches the TorchScript exports),
"quick" (HF CLIP ViT-L, used by SDXL text_encoder and FLUX), or "exact"}
}
\value{
An nn_module representing the text encoder
}
\description{
Native R torch implementation of CLIP text encoder.
Replaces TorchScript for better GPU compatibility.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_attention}
\alias{gemma3_attention}
\title{Gemma3 Attention}
\usage{
gemma3_attention(config, layer_idx = 0L)
}
\arguments{
\item{config}{Model configuration.}

\item{layer_idx}{Integer. Layer index for layer-specific settings.}
}
\description{
Multi-head attention with Grouped Query Attention (GQA) and optional
sliding window attention.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{audio_vae_ltx23}
\alias{audio_vae_ltx23}
\title{LTX-2.3 Audio VAE}
\description{
Fresh R port of the LTX-2 audio autoencoder from the diffusers
reference (Apache-2.0, autoencoder_kl_ltx2_audio.py), configured per
the checkpoint: pixel norm, height-axis causality, base 128 channels
with multipliers (1, 2, 4), 8 latent channels, 64 mel bins, no
attention. The decoder produces mel for the vocoder; the encoder
turns user audio into conditioning latents (lip sync).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{CLIPAttention}
\alias{CLIPAttention}
\title{CLIP Attention Block}
\usage{
CLIPAttention(embed_dim, num_heads)
}
\arguments{
\item{embed_dim}{Embedding dimension}

\item{num_heads}{Number of attention heads}
}
\description{
Multi-head self-attention with separate Q/K/V projections (HuggingFace style)
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_adain_filter_latent}
\alias{ltx23_adain_filter_latent}
\title{Adaptive instance normalization between latent tensors}
\usage{
ltx23_adain_filter_latent(latents, reference_latents, factor = 1)
}
\arguments{
\item{latents}{Tensor [B, C, F, H, W].}

\item{reference_latents}{Tensor with the target statistics.}

\item{factor}{Numeric blend in [-10, 10]; 0 is identity.}
}
\value{
Filtered latents.
}
\description{
Matches each (batch, channel) slice's mean/std to the reference
latents, blended by \code{factor} (cf. diffusers
\code{LTX2LatentUpsamplePipeline.adain_filter_latent}).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unigram_tokenizer}
\alias{unigram_tokenizer}
\title{Load a Unigram tokenizer from tokenizer.json}
\usage{
unigram_tokenizer(tokenizer_path)
}
\arguments{
\item{tokenizer_path}{Path to a HuggingFace tokenizer.json with a
Unigram model, or a directory containing one.}
}
\value{
A \code{unigram_tokenizer} object.
}
\description{
Load a Unigram tokenizer from tokenizer.json
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_tokenizer}
\alias{gemma3_tokenizer}
\title{Gemma3 Tokenizer}
\usage{
gemma3_tokenizer(tokenizer_path)
}
\arguments{
\item{tokenizer_path}{Character. Path to tokenizer directory or tokenizer.json file.}
}
\value{
A gemma3_tokenizer object (extends bpe_tokenizer).
}
\description{
Native R tokenizer for Gemma3 using BPE.
Loads from HuggingFace tokenizer.json format.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_pack_latents}
\alias{flux2_pack_latents}
\title{Pack patchified FLUX.2 latents into tokens}
\usage{
flux2_pack_latents(latents)
}
\arguments{
\item{latents}{Tensor [B, C, H, W].}
}
\value{
Tensor [B, H * W, C].
}
\description{
[B, C, H, W] -> [B, H * W, C] (row-major spatial flatten,
channels-last). Reference: Flux2KleinPipeline._pack_latents.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{st_caps}
\alias{st_caps}
\title{safetensors read-capability probes and fork messaging}
\description{
The CRAN build of safetensors 0.2.1 reads bfloat16 but cannot write
it, and has no float8 support at all. Both fixes merged upstream on
2026-07-31 (mlverse/safetensors#11 for bfloat16 write, #13 for
float8) without a version bump, so the installed version number
cannot tell you which build you have. That is why every gate here is
a runtime probe: write a tiny tensor, read it back, cache the answer.
Two capabilities matter and they differ:
}
\details{
\itemize{
\item \emph{write} (\code{\link{flux_quantize}}'s internal
\code{.st_can_write}, in quantize_flux.R): needed to BUILD a
quantized artifact in that dtype.
\item \emph{read} (\code{.st_can_read}, here): needed to LOAD a
hosted artifact in that dtype. This is the capability that gates
user-facing recommendations. It is strictly weaker than write:
CRAN safetensors reads bfloat16 it cannot write, so the write
probe is the wrong signal for whether a hosted bf16 artifact will
load.
}

Both are capability-probed, never version-pinned, so the fork
requirement self-heals the day the fixes reach CRAN.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{resident_activate}
\alias{resident_activate}
\title{Bring a resident pipeline onto the GPU}
\usage{
resident_activate(res)
}
\arguments{
\item{res}{A \code{diffuseR_resident} handle.}
}
\value{
Invisibly the handle, with state "active".
}
\description{
Copies every pinned component to the handle's bound device by DMA and
verifies the result tensor-by-tensor. A failure rolls back to the
pinned host state; a rollback that cannot itself be verified leaves
the handle broken.
}
\details{
What activation does depends on how the pipeline was loaded:

\itemize{
\item \code{phase_offload = TRUE} (the default, and what the
\code{txt2img_*} functions expect): no bulk transfer. The render
moves each component on as its phase begins and back off as it
ends, from these same pinned copies, so pre-loading them would be
undone within one phase. Activation is the ownership claim.
\item \code{phase_offload = FALSE}: every component is copied to the
card up front and stays there across renders. This is the fast
path, and it is checked against free VRAM first.
}

The distinction is not cosmetic. FLUX.1's pinned set is 15.73 GB,
which does not fit a 15.47 GiB card -- bulk-onloading it OOMs even
though the phased render fits comfortably. So \code{state} is a claim
about who owns the card, not a measurement of what is on it; read
\code{components_on_gpu} from \code{\link{resident_status}} for the
measurement.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{text_encoder_native_from_safetensors}
\alias{text_encoder_native_from_safetensors}
\title{Build a native CLIP text encoder from a diffusers safetensors directory}
\usage{
text_encoder_native_from_safetensors(path, apply_final_ln = TRUE,
                                     verbose = TRUE, ...)
}
\arguments{
\item{path}{diffusers text_encoder directory (config.json +
model.safetensors) or the config.json path.}

\item{apply_final_ln}{Apply the final layer norm in forward (default TRUE).}

\item{verbose}{Print how many parameters were loaded.}

\item{...}{Overrides for \code{\link{text_encoder_native}} args (e.g.
\code{gelu_type}).}
}
\value{
The native text encoder in eval mode.
}
\description{
Reads the CLIPTextConfig from \code{<dir>/config.json}, constructs
\code{\link{text_encoder_native}} to match, and loads
\code{model.safetensors} - the safetensors counterpart to the
TorchScript text-encoder path (no TorchScript, Blackwell-safe).
Handles SD21's OpenCLIP ViT-H and SDXL's CLIP ViT-L (which is the same
checkpoint as FLUX's \code{text_encoder}). \code{apply_final_ln}
governs only the forward output; the \code{final_layer_norm} weights
load either way. Use \code{TRUE} for SD21 and pooled CLIP outputs,
\code{FALSE} for the SDXL penultimate-layer prompt embeds.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{CLIPTransformerBlock}
\alias{CLIPTransformerBlock}
\title{CLIP Transformer Block}
\usage{
CLIPTransformerBlock(embed_dim, num_heads, mlp_dim, gelu_type = "tanh")
}
\arguments{
\item{embed_dim}{Embedding dimension}

\item{num_heads}{Number of attention heads}

\item{mlp_dim}{MLP hidden dimension}

\item{gelu_type}{GELU variant: "tanh", "quick", or "exact"}
}
\description{
Pre-norm transformer block with attention and MLP (HuggingFace style)
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_prepare_latent_image_ids}
\alias{flux_prepare_latent_image_ids}
\title{Build FLUX latent image position ids}
\usage{
flux_prepare_latent_image_ids(height, width, device = "cpu")
}
\arguments{
\item{height}{Integer. Packed grid height (latent height / 2).}

\item{width}{Integer. Packed grid width (latent width / 2).}

\item{device}{Device for the resulting tensor.}
}
\value{
Float tensor of shape [height * width, 3].
}
\description{
Position ids over the packed latent grid (latent height/2 x width/2).
Channel 1 is always zero, channel 2 holds the row index, channel 3 the
column index. Reference: FluxPipeline._prepare_latent_image_ids.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{apply_rotary_pos_emb}
\alias{apply_rotary_pos_emb}
\title{Apply rotary position embeddings}
\usage{
apply_rotary_pos_emb(q, k, cos, sin)
}
\arguments{
\item{q}{Query tensor [batch, heads, seq, head_dim]}

\item{k}{Key tensor [batch, heads, seq, head_dim]}

\item{cos}{Cosine embeddings [batch, seq, head_dim]}

\item{sin}{Sine embeddings [batch, seq, head_dim]}
}
\description{
Apply rotary position embeddings
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_config_ltx2}
\alias{gemma3_config_ltx2}
\title{Create Gemma3 configuration for LTX-2}
\usage{
gemma3_config_ltx2()
}
\value{
List with model configuration parameters.
}
\description{
Returns the default configuration used by LTX-2's text encoder.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_rotary_embedding}
\alias{gemma3_rotary_embedding}
\title{Gemma3 Rotary Position Embeddings}
\usage{
gemma3_rotary_embedding(dim, max_position_embeddings = 8192L, base = 10000,
                        scaling_factor = 1)
}
\arguments{
\item{dim}{Integer. Head dimension.}

\item{max_position_embeddings}{Integer. Maximum sequence length.}

\item{base}{Numeric. RoPE base frequency.}

\item{scaling_factor}{Numeric. Optional scaling factor for extended context.}
}
\description{
Standard RoPE with optional scaling factor for extended context.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{txt2img_sdxl}
\alias{txt2img_sdxl}
\title{Generate an image from a text prompt using SDXL}
\usage{
txt2img_sdxl(prompt, negative_prompt = NULL, img_dim = 1024, pipeline = NULL,
             devices = "auto", memory_profile = NULL, unet_dtype_str = NULL,
             download_models = FALSE, scheduler = "ddim", timesteps = NULL,
             initial_latents = NULL, num_inference_steps = 30,
             guidance_scale = 7.5, seed = NULL, save_file = TRUE,
             filename = NULL, metadata_path = NULL, use_native_decoder = FALSE,
             use_native_text_encoder = FALSE, use_native_unet = FALSE,
             diffusers_dir = NULL, verbose = TRUE, ...)
}
\arguments{
\item{prompt}{A character string prompt describing the image to generate.}

\item{negative_prompt}{Optional negative prompt to guide the generation.}

\item{img_dim}{Dimension of the output image (e.g., 512 for 512x512).}

\item{pipeline}{Optional A pre-loaded diffusion pipeline. If `NULL`, it will be loaded based on the model name and devices.}

\item{devices}{A named list of devices for each model component (e.g., `list(unet = "cuda", decoder = "cpu", text_encoder = "cpu")`),
or "auto" to use `auto_devices()`, or NULL to use memory_profile devices.}

\item{memory_profile}{Character or list. Memory profile for GPU-poor optimization:
"auto" for auto-detection, or a profile name ("full_gpu", "balanced", "unet_gpu", "cpu_only"),
or a list from `sdxl_memory_profile()`. When specified, overrides devices parameter.}

\item{unet_dtype_str}{Optional A character for dtype of the unet component (typically "float16" for cuda and "float32" for cpu; float32 is available for cuda).}

\item{download_models}{Logical indicating whether to download the model files if they are not found.}

\item{scheduler}{Scheduler to use (e.g., `"ddim"`, `"euler"`).}

\item{timesteps}{Optional A vector of timesteps to use.}

\item{initial_latents}{Optional initial latents for the diffusion process.}

\item{num_inference_steps}{Number of inference steps to run.}

\item{guidance_scale}{Scale for classifier-free guidance (typically 7.5).}

\item{seed}{Optional seed for reproducibility.}

\item{save_file}{Logical indicating whether to save the generated image.}

\item{filename}{Optional filename for saving the image. If `NULL`, a default name is generated.}

\item{metadata_path}{Optional file path to save metadata.}

\item{use_native_decoder}{Logical; if TRUE, uses native R torch decoder instead of TorchScript.
Native decoder has better GPU compatibility (especially Blackwell).}

\item{use_native_text_encoder}{Logical; if TRUE, uses native R torch text encoder instead of TorchScript.
Native text encoder has better GPU compatibility (especially Blackwell).}

\item{use_native_unet}{Logical; if TRUE, uses native R torch UNet instead of TorchScript.
Native UNet has better GPU compatibility (especially Blackwell).}

\item{diffusers_dir}{Optional path to a diffusers safetensors directory
(\code{unet/}, \code{vae/}, \code{text_encoder/}, \code{text_encoder_2/}).
When supplied, the pipeline is built end-to-end from safetensors via
\code{\link{sdxl_pipeline_from_safetensors}} (native, Blackwell-safe,
no TorchScript) and the other \code{use_native_*} flags are ignored.}

\item{verbose}{Logical. Print progress and memory status messages.}

\item{...}{Additional parameters passed to the diffusion process.}
}
\value{
An image array and metadata
}
\description{
Generate an image from a text prompt using SDXL
}
\examples{
\dontrun{
# Basic usage with auto-detection
img <- txt2img_sdxl("a cat wearing sunglasses in space")

# GPU-poor mode (8GB VRAM)
img <- txt2img_sdxl("a sunset over mountains", memory_profile = "unet_gpu")

# Explicit memory profile
profile <- sdxl_memory_profile(vram_gb = 8)
img <- txt2img_sdxl("a forest path", memory_profile = profile)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_open_quantized}
\alias{flux_open_quantized}
\title{Open a quantized FLUX artifact directory}
\usage{
flux_open_quantized(dir)
}
\arguments{
\item{dir}{The quantized artifact directory (with manifest.json).}
}
\value{
An \code{ltx23_checkpoint} with \code{$format} set.
}
\description{
Opens the sharded NF4/fp8 artifact written by
\code{\link{flux_quantize}} through the shared checkpoint interface.
The manifest's embedded transformer config and \code{format} ride
along, so \code{\link{flux_load_transformer}} needs nothing else.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_bn_normalize}
\alias{flux2_bn_normalize}
\title{Normalize patchified latents with the VAE BatchNorm statistics}
\usage{
flux2_bn_normalize(latents, bn_mean, bn_var, eps = 1e-04, inverse = FALSE)
}
\arguments{
\item{latents}{Tensor [B, 128, H, W] (patchified).}

\item{eps}{Numeric. BatchNorm epsilon.}

\item{inverse}{Logical. De-normalize (decode path) instead.}

\item{bn_mean,bn_var}{Float tensors [128].}
}
\value{
Tensor like \code{latents}.
}
\description{
FLUX.2 has no scalar scaling/shift factor; latents are standardized
per packed channel with the VAE's \code{bn.running_mean} /
\code{bn.running_var} (eps 1e-4). Reference: encode/decode paths of
Flux2KleinPipeline.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_transformer_block}
\alias{ltx23_transformer_block}
\title{LTX-2 transformer block}
\usage{
ltx23_transformer_block(dim, num_attention_heads, attention_head_dim,
                        cross_attention_dim, audio_dim,
                        audio_num_attention_heads, audio_attention_head_dim,
                        audio_cross_attention_dim, video_gated_attn = TRUE,
                        video_cross_attn_adaln = TRUE, audio_gated_attn = TRUE,
                        audio_cross_attn_adaln = TRUE, eps = 1e-06,
                        elementwise_affine = FALSE, rope_type = "split",
                        perturbed_attn = TRUE)
}
\arguments{
\item{cross_attention_dim}{Integer. Text embedding dim for video.}

\item{audio_cross_attention_dim}{Integer. Text embedding dim for audio.}

\item{eps}{Numeric. Norm epsilon.}

\item{elementwise_affine}{Logical. Block norms carry weights (FALSE for LTX).}

\item{rope_type}{"split" or "interleaved".}

\item{perturbed_attn}{Logical. Enable the STG perturbation arguments.}

\item{dim,audio_dim}{Integers. Video/audio stream dimensions.}

\item{num_attention_heads,attention_head_dim}{Video attention shape.}

\item{audio_num_attention_heads,audio_attention_head_dim}{Audio attention shape.}

\item{video_gated_attn,audio_gated_attn}{Logicals. Per-head output gates.}

\item{video_cross_attn_adaln,audio_cross_attn_adaln}{Logicals. LTX-2.3
text cross-attention modulation (9 mod params instead of 6).}
}
\value{
Module whose forward(hidden_states, ...) returns
  \code{list(hidden_states, audio_hidden_states)}, the video and audio
  streams after self-attention, cross-attention and the feed-forward,
  each the same shape as its input.
}
\description{
Dual-stream (video + audio) block: modulated self-attention per
modality, text cross-attention per modality (with LTX-2.3 query and
key/value modulation), bidirectional audio-video cross-attention with
global+per-block modulation, and modulated feed-forward.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_load_transformer}
\alias{flux_load_transformer}
\title{Load a FLUX transformer from any checkpoint format}
\usage{
flux_load_transformer(ckpt, device = "cuda", dtype = "bfloat16", pin = TRUE,
                      fp8_resident = FALSE, verbose = TRUE, ...)
}
\arguments{
\item{ckpt}{A checkpoint from \code{\link{flux_open_checkpoint}} or
\code{\link{flux_open_quantized}}.}

\item{device}{Character. Compute device.}

\item{dtype}{Character. Model dtype ("bfloat16" or "float32"). For
quantized formats this sets the resident (non-quantized) tensors
and must match the compute dtype: bfloat16 for GPU compute,
float32 for CPU compute.}

\item{pin}{Logical. Pin fp8 host memory for faster transfers
(streamed fp8 only).}

\item{fp8_resident}{Logical. Keep the fp8 weights on \code{device}
instead of streaming from the CPU - right for models whose whole
quantized footprint fits in VRAM (FLUX.2 klein-4B: ~4 GB).}

\item{verbose}{Logical.}

\item{...}{Overrides for \code{\link{flux_transformer}} arguments
(tiny test configs).}
}
\value{
The loaded \code{flux_transformer} in eval mode.
}
\description{
Builds \code{\link{flux_transformer}} from the checkpoint's embedded
config and loads the weights. Dispatches on the checkpoint format:
}
\details{
\itemize{
\item full precision (\code{\link{flux_open_checkpoint}}): weights
stream into the model in \code{dtype} on \code{device}.
\item \code{"nf4"} (\code{\link{flux_open_quantized}}): cast-set
linears become \code{ltx23_nf4_linear}; the whole model (packed
weights included) moves to \code{device} and stays resident.
\item \code{"fp8"}: cast-set linears become
\code{ltx23_fp8_linear}; fp8 weights stay CPU-resident (optionally
pinned) and stream to \code{device} inside each forward.
}

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_feed_forward}
\alias{zimage_feed_forward}
\title{Z-Image feed-forward (SwiGLU with separate gate weights)}
\usage{
zimage_feed_forward(dim, hidden_dim)
}
\arguments{
\item{dim}{Integer. Model width.}

\item{hidden_dim}{Integer. Hidden width.}
}
\value{
Module whose forward(x) returns \code{w2(silu(w1(x)) * w3(x))},
  a tensor of the same shape as \code{x}.
}
\description{
w2(silu(w1(x)) * w3(x)) with all three linears bias-free. The hidden
width is int(dim / 3 * 8).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_tune_gc}
\alias{ltx23_tune_gc}
\title{Tune the torch CUDA allocator for large-resident inference}
\usage{
ltx23_tune_gc(footprint_gb = 12, total_gb = NULL)
}
\arguments{
\item{footprint_gb}{Numeric. Expected resident GPU footprint in GB
(NF4 transformer: ~12).}

\item{total_gb}{Numeric or NULL (auto-detect total VRAM).}
}
\value{
Invisibly, the applied reserved rate (NULL if skipped).
}
\description{
Stops the allocator GC storm (cf. ~/skills/torch
torch-jit-gc-performance.md): lantern proactively calls R's gc()
whenever reserved memory exceeds \code{torch.cuda_allocator_reserved_rate}
(default 0.20) of the card. With ~75\\% of VRAM occupied by resident
weights that fires on nearly every allocation. Raising the rate to the
actual footprint is safe here because the LTX hot loops compute into
persistent scratch buffers (near-zero per-step garbage). Also raises
the host-allocation GC threshold and defaults
\code{PYTORCH_CUDA_ALLOC_CONF} to expandable segments. User-set
options win.
}
\details{
\code{start_torch()} reads the gate options exactly once, so setting
them after torch has started is inert on its own. The three CUDA
gates are therefore also pushed into the live allocator here (the
\code{.flux_gc_gates} pattern), which makes this function effective
whenever it runs. The host-side \code{torch.threshold_call_gc} has
no live setter; the package defaults it in \code{.onLoad} so torch
reads it at init in any session that loads diffuseR before running
torch ops.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_audio_upsample}
\alias{ltx23_audio_upsample}
\title{LTX audio upsampler}
\usage{
ltx23_audio_upsample(in_channels, causality_axis = "height")
}
\arguments{
\item{in_channels}{Integer.}

\item{causality_axis}{Character.}
}
\value{
Module whose forward(x) returns the tensor upsampled 2x by
  nearest-neighbour interpolation and convolved.
}
\description{
Nearest 2x interpolation, causal conv, then a crop of the first
element along the causal axis.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_mid_block3d}
\alias{ltx23_video_mid_block3d}
\title{LTX video mid block}
\usage{
ltx23_video_mid_block3d(in_channels, num_layers = 1L, resnet_eps = 1e-06,
                        spatial_padding_mode = "zeros")
}
\arguments{
\item{in_channels}{Integer.}

\item{num_layers}{Integer.}

\item{resnet_eps}{Numeric.}

\item{spatial_padding_mode}{Character.}
}
\value{
Module whose forward(hidden_states, causal) returns the
  bottleneck output, a tensor of the same shape as the input.
}
\description{
A plain ResNet stack at a fixed channel count.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_pipeline}
\alias{load_pipeline}
\title{Load a diffusion model pipeline}
\usage{
load_pipeline(model_name, m2d, i2i = FALSE, unet_dtype_str,
              use_native_decoder = FALSE, use_native_text_encoder = FALSE,
              use_native_unet = FALSE, ...)
}
\arguments{
\item{model_name}{The name of the model to load.}

\item{m2d}{A list containing model-to-device mappings and configurations.}

\item{i2i}{Logical indicating whether to load the encoder for img2img().}

\item{unet_dtype_str}{A string representing the data type for the UNet model (e.g., "float32", "float16").}

\item{use_native_decoder}{Logical; if TRUE, uses native R torch decoder instead of TorchScript.
Native decoder has better GPU compatibility (especially Blackwell).}

\item{use_native_text_encoder}{Logical; if TRUE, uses native R torch text encoder instead of TorchScript.
Native text encoder has better GPU compatibility (especially Blackwell).}

\item{use_native_unet}{Logical; if TRUE, uses native R torch UNet instead of TorchScript.
Native UNet has better GPU compatibility (especially Blackwell).}

\item{...}{Additional arguments passed to the model loading functions.}
}
\value{
An environment containing the loaded models and configuration.
}
\description{
This function loads a diffusion model pipeline consisting of a UNet, VAE decoder, and text encoder.
It initializes the models and sets up the environment for inference.
}
\examples{
\dontrun{
pipeline <- load_pipeline("my_model", device = "cuda")
}

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.detect_vram}
\alias{.detect_vram}
\title{Detect Available VRAM}
\usage{
.detect_vram(use_free = FALSE)
}
\arguments{
\item{use_free}{Logical. If TRUE, return free VRAM. If FALSE, return total.}
}
\value{
Numeric. VRAM in GB, or 0 if no GPU detected.
}
\description{
Asks nvidia-smi.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.ffmpeg_available}
\alias{.ffmpeg_available}
\title{Check if FFmpeg is Available}
\usage{
.ffmpeg_available()
}
\value{
Logical. TRUE if ffmpeg is in PATH.
}
\description{
Check if FFmpeg is Available
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.ltx23_traced_call}
\alias{.ltx23_traced_call}
\title{Run a module forward through a shape-specialized trace}
\usage{
.ltx23_traced_call(module, x, forward = NULL, tag = "")
}
\arguments{
\item{module}{The nn_module (identity for the cache key; also the
default callable).}

\item{x}{Input tensor.}

\item{forward}{Optional closure wrapping the call (for extra fixed
arguments like \code{causal}); must be pure in \code{x}.}

\item{tag}{Character. Distinguishes call variants of one module.}
}
\value{
The forward result.
}
\description{
Run a module forward through a shape-specialized trace
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_feed_forward}
\alias{flux2_feed_forward}
\title{FLUX.2 feed-forward (fused SwiGLU)}
\usage{
flux2_feed_forward(dim, dim_out = NULL, mult = 3, bias = FALSE)
}
\arguments{
\item{dim}{Integer. Input dimension.}

\item{dim_out}{Integer. Output dimension (defaults to \code{dim}).}

\item{mult}{Numeric. Inner dim multiplier (FLUX.2: 3.0).}

\item{bias}{Logical.}
}
\value{
Module whose forward(x) returns the SwiGLU-gated projection
  of \code{x}, a tensor with the last axis of width \code{dim_out}.
}
\description{
\code{linear_in} projects to twice the inner dim; SwiGLU gates the
first half with SiLU and multiplies by the second half;
\code{linear_out} projects back. Reference: Flux2FeedForward +
Flux2SwiGLU.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_decoder_weights}
\alias{load_decoder_weights}
\title{Load weights from TorchScript decoder into native decoder}
\usage{
load_decoder_weights(native_decoder, torchscript_path, verbose = TRUE)
}
\arguments{
\item{native_decoder}{Native VAE decoder module}

\item{torchscript_path}{Path to TorchScript decoder .pt file}

\item{verbose}{Print loading progress}
}
\value{
The native decoder with loaded weights (invisibly)
}
\description{
Load weights from TorchScript decoder into native decoder
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_quantize}
\alias{flux_quantize}
\title{Quantize a FLUX transformer to NF4 or fp8 shards}
\usage{
flux_quantize(transformer_dir, output_dir = NULL, format = c("nf4", "fp8"),
              shard_bytes = 1.9e+09, force = FALSE, verbose = TRUE)
}
\arguments{
\item{transformer_dir}{Source diffusers transformer directory.}

\item{output_dir}{Output directory for shards + manifest (default:
the per-format location under \code{tools::R_user_dir}).}

\item{format}{"nf4" or "fp8".}

\item{shard_bytes}{Numeric. Target shard size in bytes. The default
1.9e9 keeps every shard under the 2^31-byte (~2.15 GB) ceiling that
stock CRAN safetensors can read, so the artifact loads fork-free.
Pass a larger value (e.g. 4e9) only for local builds you will read
back with a fork-patched safetensors.}

\item{force}{Logical. Re-quantize even if a valid manifest exists.}

\item{verbose}{Logical.}
}
\value{
Invisibly, the manifest list.
}
\description{
Streams the bf16 diffusers checkpoint tensor by tensor. Cast-set
weights (see \code{\link{flux_is_quant_key}}) are stored as NF4
(packed uint8 + \code{<key>_absmax} float32 blocks) or as
float8_e4m3fn with an absmax/448 per-tensor \code{<key>_scale};
everything else is copied through unchanged. The manifest embeds the
transformer config, so the source checkpoint is not needed again
after quantization.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.resident_guard}
\alias{.resident_guard}
\title{Refuse operations that the current state cannot serve}
\usage{
.resident_guard(res, verb)
}
\arguments{
\item{res}{A resident handle.}

\item{verb}{What the caller is attempting, for the message.}
}
\value{
Invisibly TRUE, or an error.
}
\description{
Refuse operations that the current state cannot serve
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{save_video_av}
\alias{save_video_av}
\title{Save Video using av Package}
\usage{
save_video_av(video, file, fps = 24, verbose = TRUE)
}
\arguments{
\item{video}{Array of video frames [T, H, W, C].}

\item{file}{Output file path.}

\item{fps}{Frames per second.}

\item{verbose}{Logical.}
}
\description{
Save Video using av Package
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_transformer}
\alias{flux_transformer}
\title{FLUX transformer model}
\usage{
flux_transformer(in_channels = 64L, num_layers = 19L, num_single_layers = 38L,
                 attention_head_dim = 128L, num_attention_heads = 24L,
                 joint_attention_dim = 4096L, pooled_projection_dim = 768L,
                 axes_dims_rope = c(16L, 56L, 56L), out_channels = NULL)
}
\arguments{
\item{in_channels}{Integer. Packed latent channels (64).}

\item{num_layers}{Integer. Double-stream block count.}

\item{num_single_layers}{Integer. Single-stream block count.}

\item{attention_head_dim}{Integer. Per-head dimension.}

\item{num_attention_heads}{Integer. Attention heads.}

\item{joint_attention_dim}{Integer. T5 embedding dim (4096).}

\item{pooled_projection_dim}{Integer. CLIP pooled dim (768).}

\item{axes_dims_rope}{Integer vector. Per-axis rotary dims.}

\item{out_channels}{Integer or NULL. Output channels (defaults to
\code{in_channels}).}
}
\value{
Module whose forward(hidden_states, encoder_hidden_states,
  pooled_projections, timestep, image_rotary_emb) returns the
  predicted velocity for the image tokens [B, S_img, out_channels].
  \code{timestep} is in sigma space (0-1); it is scaled by 1000
  internally, matching the reference.
}
\description{
19 double-stream (MMDiT) blocks followed by 38 single-stream blocks
over the joint [text; image] sequence, with adaLN-Zero conditioning on
timestep + pooled CLIP text. Rotary embeddings are precomputed by the
caller with \code{flux_pos_embed} (they are static across denoise
steps). Defaults are the FLUX.1-schnell configuration.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{save_image}
\alias{save_image}
\title{Save and Display an Image from a Torch Tensor}
\usage{
save_image(img, save_to = "output.png", normalize = TRUE)
}
\arguments{
\item{img}{A numeric with shape `[3, H, W]`.}

\item{save_to}{File path for the PNG image (default is `"output.png"`).}

\item{normalize}{Logical; whether to normalize pixel values to `[0, 1]`. Default is `TRUE`.}
}
\value{
Invisibly returns the saved file path.
}
\description{
Converts a Torch tensor to a normalized RGB image array, saves it as a PNG file,
and optionally displays it in the RStudio Viewer pane using `grid::grid.raster()`.
}
\examples{
img <- array(runif(32 * 32 * 3), dim = c(32, 32, 3))
out <- file.path(tempdir(), "sample.png")
save_image(img, out)
file.exists(out)
unlink(out)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_kaiser_sinc_filter1d}
\alias{ltx23_kaiser_sinc_filter1d}
\title{Kaiser sinc low-pass filter kernel}
\usage{
ltx23_kaiser_sinc_filter1d(cutoff, half_width, kernel_size)
}
\arguments{
\item{cutoff}{Numeric. Normalized cutoff in (0, 0.5].}

\item{half_width}{Numeric. Transition band half width.}

\item{kernel_size}{Integer.}
}
\value{
Tensor [kernel_size].
}
\description{
Kaiser sinc low-pass filter kernel
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{dit_flux2}
\alias{dit_flux2}
\title{FLUX.2 Transformer (MMDiT)}
\description{
Fresh R port of Flux2Transformer2DModel from the diffusers reference
implementation (Apache-2.0,
src/diffusers/models/transformers/transformer_flux2.py). Defaults are
the klein-4B configuration (5 double + 20 single blocks). Guidance
embeddings (FLUX.2-dev) are not implemented; klein is step-distilled
with \code{guidance_embeds = false}. Timestep conditioning has no
pooled-text component, and the three modulation projections are
shared across all blocks.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{upsampler_ltx23}
\alias{upsampler_ltx23}
\title{LTX-2.3 Spatial Latent Upsampler}
\description{
Fresh R port of the LTX latent upsampler from the diffusers reference
(Apache-2.0, pipelines/ltx2/latent_upsampler.py and
pipeline_ltx2_latent_upsample.py), with the LTX 2.3 configuration:
Conv3d ResBlock stages around a per-frame 2x pixel-shuffle spatial
upsampler (no rational resampler). Operates on unnormalized latents.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{Downsample2D}
\alias{Downsample2D}
\title{Downsample Block}
\usage{
Downsample2D(channels)
}
\description{
Downsample Block
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{encode_bpe}
\alias{encode_bpe}
\title{Encode text to token IDs}
\usage{
encode_bpe(tokenizer, text, add_special_tokens = TRUE, max_length = NULL,
           padding = "none", truncation = FALSE, return_tensors = "list")
}
\arguments{
\item{tokenizer}{A bpe_tokenizer object.}

\item{text}{Character string or vector to encode.}

\item{add_special_tokens}{Logical. Add BOS/EOS tokens.}

\item{max_length}{Integer. Maximum sequence length (NULL for no limit).}

\item{padding}{Character. Padding strategy: "none", "max_length", or "longest".}

\item{truncation}{Logical. Truncate to max_length.}

\item{return_tensors}{Character. Return type: "list" or "pt" (torch tensors).}
}
\value{
List with input_ids and attention_mask.
}
\description{
Encode text to token IDs
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_read_audio}
\alias{ltx23_read_audio}
\title{Read an audio file as 16 kHz stereo PCM}
\usage{
ltx23_read_audio(path, sample_rate = 16000L)
}
\arguments{
\item{path}{Audio file.}

\item{sample_rate}{Integer.}
}
\value{
Matrix [2, samples] in [-1, 1].
}
\description{
Decodes MP3/WAV/etc. via \code{av} to 16-bit PCM at the target rate
and parses the RIFF container in base R.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{qwen_bpe_tokenizer}
\alias{qwen_bpe_tokenizer}
\title{Load a Qwen2 byte-level BPE tokenizer}
\usage{
qwen_bpe_tokenizer(tokenizer_path)
}
\arguments{
\item{tokenizer_path}{Path to a tokenizer.json (or a directory
containing one).}
}
\value{
A \code{qwen_tokenizer} object.
}
\description{
Load a Qwen2 byte-level BPE tokenizer
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{VAEMidBlock}
\alias{VAEMidBlock}
\title{VAE Mid Block}
\usage{
VAEMidBlock(channels, norm_groups = 32)
}
\arguments{
\item{channels}{Number of channels}
}
\description{
VAE Mid Block
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_map_connector_key}
\alias{ltx23_map_connector_key}
\title{Map an official connectors checkpoint key to the R module name}
\usage{
ltx23_map_connector_key(key)
}
\arguments{
\item{key}{Character. Checkpoint key.}
}
\value{
Character. Module parameter name.
}
\description{
Map an official connectors checkpoint key to the R module name
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.ltx23_jit_run_stack}
\alias{.ltx23_jit_run_stack}
\title{Run the block stack through the compiled TorchScript path}
\usage{
.ltx23_jit_run_stack(blocks, hidden_states, audio_hidden_states,
                     encoder_hidden_states, audio_encoder_hidden_states, temb,
                     temb_audio, temb_ca_scale_shift,
                     temb_ca_audio_scale_shift, temb_ca_gate,
                     temb_ca_audio_gate, temb_prompt, temb_prompt_audio,
                     video_rotary_emb, audio_rotary_emb, ca_video_rotary_emb,
                     ca_audio_rotary_emb, encoder_attention_mask = NULL,
                     audio_encoder_attention_mask = NULL,
                     cond_token_index = NULL)
}
\value{
list(hidden_states, audio_hidden_states)
}
\description{
One R-to-libtorch crossing for all blocks: no per-op dispatch, no R
tensor garbage, fused SDPA. Masks must already be additive
\code{[B, 1, 1, S]} (or NULL); rope tensors are the \code{[.., r]}
cos/sin pairs used by the eager path.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_connector_transformer_1d}
\alias{ltx23_connector_transformer_1d}
\title{1D connector transformer}
\usage{
ltx23_connector_transformer_1d(num_attention_heads = 32L,
                               attention_head_dim = 128L, num_layers = 8L,
                               num_learnable_registers = 128L,
                               rope_base_seq_len = 4096L, rope_theta = 10000,
                               rope_double_precision = TRUE, eps = 1e-06,
                               rope_type = "split", gated_attention = TRUE)
}
\arguments{
\item{num_learnable_registers}{Integer or NULL. Register count (the
sequence length must be divisible by it).}

\item{eps}{Numeric. Norm epsilon.}

\item{gated_attention}{Logical. Per-head attention output gates.}

\item{num_attention_heads,attention_head_dim,num_layers}{Transformer shape.}

\item{rope_base_seq_len,rope_theta,rope_double_precision,rope_type}{RoPE config.}
}
\value{
Module whose forward(hidden_states, attention_mask,
  attn_mask_binarize_threshold) returns
  \code{list(hidden_states, attention_mask)}: the transformed
  sequence and the (possibly binarized) mask that accompanies it.
}
\description{
Replaces padded positions with learnable registers (valid tokens are
front-aligned in their original order; the tail is filled with
registers indexed by absolute position, after which the attention mask
is cleared), then runs 1D transformer blocks with rotary embeddings.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.build_fallback_devices}
\alias{.build_fallback_devices}
\title{Build fallback device configuration}
\usage{
.build_fallback_devices(model, strategy)
}
\arguments{
\item{model}{Character. Model type.}

\item{strategy}{Character. Memory strategy.}
}
\value{
Named list of device assignments.
}
\description{
Build fallback device configuration
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_load_pipeline}
\alias{ltx23_load_pipeline}
\title{Load the LTX-2.3 generation components from a single-file checkpoint}
\usage{
ltx23_load_pipeline(checkpoint_path, device = "cuda", dtype = "bfloat16",
                    transformer_device = "cpu",
                    components = c("dit", "connectors", "vae", "audio_vae", "vocoder"),
                    pin = TRUE, attn_chunk = NULL, phase_offload = TRUE,
                    verbose = TRUE)
}
\arguments{
\item{checkpoint_path}{Path to the single-file checkpoint (e.g.
\code{ltx-2.3-22b-distilled-1.1.safetensors}) or to an fp8 artifact
directory produced by \code{\link{ltx23_quantize_fp8}}. With the fp8
artifact, the transformer loads with CPU-resident fp8 weights that
stream to \code{device} during the forward pass.}

\item{device}{Character. Device for the small components (VAEs,
vocoder, connectors) and, with fp8, the transformer residents.}

\item{dtype}{Character. "bfloat16" (checkpoint native) or "float32".}

\item{transformer_device}{Character. Device for the transformer
weights when loading the plain (non-fp8) checkpoint.}

\item{components}{Character vector. Which components to load.}

\item{pin}{Logical. Pin fp8 host memory (fp8 artifact only).}

\item{attn_chunk}{Integer or NULL. Query-chunk size for attention
(see \code{\link{ltx23_set_attn_chunk}}).}

\item{phase_offload}{Logical. Load the small components (connectors,
VAEs, vocoder) to the CPU; the pipeline moves each onto the compute
device only for its phase.}

\item{verbose}{Logical.}
}
\value{
A list with the loaded modules and the checkpoint config,
  class \code{ltx23_pipeline}.
}
\description{
Builds the transformer, connectors, video VAE, audio VAE, and vocoder
with the LTX 2.3 configuration and streams the checkpoint weights into
them. The Gemma3 text encoder ships separately (see
\code{\link{load_gemma3_text_encoder}}).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_tone_map_latents}
\alias{ltx23_tone_map_latents}
\title{Sigmoid tone mapping for latents}
\usage{
ltx23_tone_map_latents(latents, compression)
}
\arguments{
\item{latents}{Tensor.}

\item{compression}{Numeric in [0, 1].}
}
\value{
Tone-mapped latents.
}
\description{
Compresses the latent dynamic range (cf. diffusers
\code{tone_map_latents}). \code{compression} 0 is identity, 1 is the
full effect.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{get_required_components}
\alias{get_required_components}
\title{Get required components for each model type}
\usage{
get_required_components(model_name)
}
\arguments{
\item{model_name}{A character string representing the name of the model.}
}
\value{
A character vector of required components for the specified model.
}
\description{
This function returns a list of required components for each supported model type.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{resident_unload}
\alias{resident_unload}
\title{Drop a resident handle entirely}
\usage{
resident_unload(res)
}
\arguments{
\item{res}{A \code{diffuseR_resident} handle.}
}
\value{
Invisibly the handle, with state "unloaded".
}
\description{
Releases the GPU copy if any, drops the pipeline and the pinned host
storage, and marks the handle unloaded. Terminal: nothing but
\code{\link{resident_status}} works afterwards.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{get_bos_id}
\alias{get_bos_id}
\title{Get BOS token ID}
\usage{
get_bos_id(tokenizer)
}
\description{
Get BOS token ID
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{staging}
\alias{staging}
\alias{staging_ltx23}
\title{Pinned Staging for Phase-Sequential Components}
\description{
Phase offloading moves each large component (transformer,
connectors, VAEs, vocoder, text encoders) between CPU and GPU every
render. From pageable memory those copies run through the driver's
bounce buffer at a fraction of PCIe speed; page-locked (pinned) host
memory transfers by DMA at full rate. Each component's parameters
and buffers are pinned once at load; onload swaps every tensor to
a non-blocking GPU copy of its pinned source, and offload simply
re-points the tensors at the still-valid pinned copies — weights
are immutable during inference, so offload moves no bytes at all.
}
\details{
Costs: the model's host copies become non-swappable for the life
of the pipeline (no extra RAM - set_data repoints the same
tensors), and page-locking adds ~9s to pipeline load. Measured
post byte-LUT (768x512x49, NF4, RTX 5060 Ti): ~7s saved per render
(warm renders 64-66s pageable vs 57-59s pinned; denoise and decode
identical, the delta is pure transfer), so pinning breaks even on
the second render and costs a single-render session ~2s net. On by
default; page-locking failure falls back silently per component,
and \code{options(diffuseR.pin_staging = FALSE)} before the loader
opts out (e.g. under host memory pressure, where unswappable pages
turn thrashing into OOM). The LTX pipeline, the Gemma3 encoder, and
the FLUX-family image loaders (flux1, flux2, zimage) all stage
pinned weights; \code{\link{recommend}} computes the RAM-aware
\code{pin} default per model.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{audio_encode_ltx23}
\alias{audio_encode_ltx23}
\title{Audio Conditioning Frontend for LTX-2.3}
\description{
Turns user audio into the normalized, packed audio latents the joint
denoiser conditions on (lip sync): decode to 16 kHz stereo PCM,
log-mel via a causal STFT (filter 1024, hop 160, 64 slaney-normed
mel bins to 8 kHz — the checkpoint's preprocessing spec), then the
audio VAE encoder in argmax mode. The STFT and mel-filterbank
constructors were verified against the checkpoint's stored vocoder
bases (identical up to bf16 rounding), so the convention matches
training.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_quantize_fp8}
\alias{ltx23_quantize_fp8}
\title{Quantize an LTX-2.3 checkpoint to FP8 shards}
\usage{
ltx23_quantize_fp8(checkpoint_path, output_dir = NULL, shard_bytes = 1.9e+09,
                   force = FALSE, verbose = TRUE)
}
\arguments{
\item{checkpoint_path}{Source .safetensors (46 GB bf16 single file).}

\item{output_dir}{Output directory for shards + manifest; NULL (the
default) resolves under tools::R_user_dir("diffuseR", "data").}

\item{shard_bytes}{Numeric. Target shard size in bytes. The default
1.9e9 keeps every shard under the 2^31-byte (~2.15 GB) ceiling that
stock CRAN safetensors can read. Pass a larger value (e.g. 4e9) only
for local builds you will read back with a fork-patched safetensors.}

\item{force}{Logical. Re-quantize even if a valid manifest exists.}

\item{verbose}{Logical.}
}
\value{
Invisibly, the manifest list.
}
\description{
Streams the single-file bf16 checkpoint tensor by tensor. DiT
attention/FFN linear weights are stored as float8_e4m3fn with a
float32 absmax/448 per-tensor scale (\code{<key>_scale} sibling);
everything else is copied through unchanged. Output shards carry the
original key names plus a manifest for skip-if-exists.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_downsampler3d}
\alias{ltx23_video_downsampler3d}
\title{Pixel-unshuffle 3D downsampler}
\usage{
ltx23_video_downsampler3d(in_channels, out_channels, stride = c(1L, 1L, 1L),
                          spatial_padding_mode = "zeros")
}
\arguments{
\item{stride}{Length-3 integer vector (t, h, w).}

\item{spatial_padding_mode}{Character.}

\item{in_channels,out_channels}{Integers.}
}
\value{
Module whose forward(hidden_states, causal) returns the
  space-to-depth downsampled tensor plus its residual: spatial and
  temporal extents shrink by \code{stride}, channels grow to match.
}
\description{
Conv followed by space/time-to-channel rearrangement, plus a grouped
channel-mean residual of the same rearrangement.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_apply_interleaved_rotary_emb}
\alias{ltx23_apply_interleaved_rotary_emb}
\title{Apply interleaved rotary embeddings}
\usage{
ltx23_apply_interleaved_rotary_emb(x, freqs)
}
\arguments{
\item{x}{Tensor of shape [B, S, C].}

\item{freqs}{List of two tensors (cos, sin), each [B, S, C].}
}
\value{
Tensor with the same shape and dtype as \code{x}.
}
\description{
Rotates adjacent element pairs of the last dimension:
\code{out = x * cos + rotate_half(x) * sin} with pairs interleaved
(elements 1,2 form the first complex pair).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_per_token_rms_norm}
\alias{ltx23_per_token_rms_norm}
\title{Per-token RMS normalization over the channel axis}
\usage{
ltx23_per_token_rms_norm(x, eps = 1e-06)
}
\arguments{
\item{x}{Tensor [B, S, C, L] of stacked per-layer hidden states.}

\item{eps}{Numeric. Stability epsilon.}
}
\value{
Tensor of the same shape.
}
\description{
Per-token RMS normalization over the channel axis
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{save_frames}
\alias{save_frames}
\title{Save Video Frames as Individual Images}
\usage{
save_frames(video, dir, prefix = "frame_", format = "png", verbose = TRUE)
}
\arguments{
\item{video}{Array of video frames [T, H, W, C].}

\item{dir}{Directory to save frames in.}

\item{prefix}{Character. Filename prefix (default "frame_").}

\item{format}{Character. Image format: "png" or "jpg".}

\item{verbose}{Logical.}
}
\value{
Invisibly returns vector of saved file paths.
}
\description{
Save Video Frames as Individual Images
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_memory_profile}
\alias{ltx23_memory_profile}
\title{Get an LTX-2.3 memory profile}
\usage{
ltx23_memory_profile(vram_gb = NULL)
}
\arguments{
\item{vram_gb}{Numeric or NULL (auto-detect free VRAM).}
}
\value{
Named list with device/dtype placement, \code{attn_chunk},
  \code{pin_weights}, and resolution caps.
}
\description{
Selects transformer precision, component placement, and attention
chunking for the available VRAM. Measured on an RTX 5060 Ti (16 GB):
fp8 streaming peaks ~11.6 GB (without phase offloading) at
512x320x49; NF4 keeps the whole 22B transformer resident (~12.5 GB)
and removes the ~21 GB/step PCIe weight streaming. The NF4 profile
renders 1280x704x121 with audio in ~23 min at a 15.7 GB peak
(tiled VAE decode, in-place feed-forward GELU, and the default
\code{diffuseR.attn_budget} of 1.5e8 all required at that size).
}
\details{
\describe{
\item{precision "nf4"}{Weights resident on the GPU; fastest steps;
about 9 percent weight round-trip error.}
\item{precision "fp8"}{Weights CPU-resident, streamed per linear;
near-bf16 quality; each step pays the PCIe transfer.}
}

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_load_upsampler}
\alias{ltx23_load_upsampler}
\title{Load the LTX-2.3 spatial upscaler weights}
\usage{
ltx23_load_upsampler(path, device = "cuda", dtype = "bfloat16", verbose = TRUE)
}
\arguments{
\item{path}{Path to e.g. \code{ltx-2.3-spatial-upscaler-x2-1.1.safetensors}.}

\item{verbose}{Logical.}

\item{device,dtype}{Placement for the loaded model.}
}
\value{
The loaded \code{ltx23_latent_upsampler}.
}
\description{
The checkpoint keys match this module tree directly.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{standardize_devices}
\alias{standardize_devices}
\title{Standardize devices configuration}
\usage{
standardize_devices(devices, required_components)
}
\arguments{
\item{devices}{A character string or a named list specifying the devices for model components.}

\item{required_components}{A character vector of required components for the model.}
}
\value{
A named list of devices for each required component.
}
\description{
This function standardizes the device configuration for model components.
It checks if the devices parameter is a single string or a named list, and fills in missing components with reasonable defaults.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_double_block}
\alias{flux_double_block}
\title{FLUX double-stream (MMDiT) transformer block}
\usage{
flux_double_block(dim, num_attention_heads, attention_head_dim)
}
\arguments{
\item{dim}{Integer. Model dimension.}

\item{num_attention_heads}{Integer. Attention heads.}

\item{attention_head_dim}{Integer. Per-head dimension.}
}
\value{
Module whose forward(hidden_states, encoder_hidden_states,
  temb, image_rotary_emb) returns
  \code{list(encoder_hidden_states, hidden_states)}.
}
\description{
Image and text streams each get adaLN-Zero modulation and a
feed-forward; attention is joint across both streams. Reference:
diffusers FluxTransformerBlock.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_ltx2}
\alias{download_ltx2}
\title{Download the LTX-2.3 checkpoint and build a quantized artifact}
\usage{
download_ltx2(quantize = TRUE, precision = c("nf4", "fp8"), output_dir = NULL,
              text_encoder = TRUE, verbose = TRUE)
}
\arguments{
\item{quantize}{Logical. Build the quantized artifact after downloading.}

\item{precision}{"nf4" (~19 GB, readable by every safetensors) or
"fp8" (~26 GB, needs float8 write support).}

\item{output_dir}{Directory for the artifact. NULL derives it from
\code{precision}.}

\item{text_encoder}{Logical. Also fetch the Gemma3 text encoder and
tokenizer (~25 GB, shared with LTX-2.0; from the Lightricks/LTX-2
repo).}

\item{verbose}{Logical.}
}
\value{
Invisibly, a list with \code{checkpoint} (source path or NULL),
  \code{artifact_dir}, \code{precision}, \code{text_encoder_dir}, and
  \code{fp8_dir} for back-compatibility -- the artifact directory when
  \code{precision} is "fp8", NULL otherwise, since a field named
  \code{fp8_dir} pointing at an nf4 artifact would be a trap.
}
\description{
Skips work that is already done: a valid manifest short-circuits
everything; a cached 46 GB source skips the download. The source file
may be deleted after quantization (it is never removed automatically).
}
\details{
Both quantized tiers are buildable here. \code{\link{recommend}}
returns nf4 for LTX on any card with 14 GB or more (it prefers nf4 at
1280 px over fp8 at 1024 px, since video trades weight precision for
resolution), so nf4 is the tier most users want. fp8 additionally
needs a safetensors that can \emph{write} float8; asking for it
without one warns and builds nf4 instead rather than failing inside
the quantizer.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vae_flux2}
\alias{vae_flux2}
\title{FLUX.2 Latent Layout and VAE Helpers}
\description{
Fresh R port of the FLUX.2 latent packing chain from the diffusers
reference (Apache-2.0, src/diffusers/pipelines/flux2/
pipeline_flux2_klein.py). The 32-channel VAE latent is patchified
2x2 into 128 channels, normalized with the VAE's BatchNorm running
statistics (there is no scalar scaling/shift factor in FLUX.2), and
flattened to channels-last tokens for the transformer.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_split_keys}
\alias{ltx23_split_keys}
\title{Split checkpoint keys by component}
\usage{
ltx23_split_keys(keys)
}
\arguments{
\item{keys}{Character vector of checkpoint tensor names.}
}
\value{
Named list of character vectors: \code{dit},
  \code{connectors}, \code{vae}, \code{audio_vae}, \code{vocoder},
  and \code{other} (anything unrecognized; should be empty).
}
\description{
Splits the flat key space of an LTX single-file checkpoint into its
component groups. Connector tensors live under the
\code{model.diffusion_model.} prefix alongside the transformer, plus a
top-level \code{text_embedding_projection.} group; both are routed to
the \code{connectors} component.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vae_decoder_native_from_safetensors}
\alias{vae_decoder_native_from_safetensors}
\title{Build a native VAE decoder from a diffusers safetensors directory}
\usage{
vae_decoder_native_from_safetensors(path, latent_channels = 4L, verbose = TRUE,
                                    ...)
}
\arguments{
\item{path}{Path to the VAE directory (containing
\code{diffusion_pytorch_model.safetensors}) or the file itself.}

\item{latent_channels}{Latent channel count (4 for SD/SDXL, 16 for FLUX).}

\item{verbose}{Print how many parameters were loaded.}

\item{...}{Overrides for \code{\link{vae_decoder_native}} constructor args.}
}
\value{
The native VAE decoder in eval mode.
}
\description{
The safetensors counterpart to the TorchScript decoder path:
constructs \code{\link{vae_decoder_native}} and loads the decoder half
of a diffusers AutoencoderKL checkpoint (no TorchScript, so it works
on Blackwell). \code{latent_channels} defaults to 4 (SD/SDXL); pass 16
for the FLUX/SD3 VAE. The SD/SDXL and FLUX VAEs share the decoder
shape and differ only in that channel count.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_text_connectors}
\alias{ltx23_text_connectors}
\title{LTX-2.3 text connectors}
\usage{
ltx23_text_connectors(caption_channels = 3840L, text_proj_in_factor = 49L,
                      video_connector_num_attention_heads = 32L,
                      video_connector_attention_head_dim = 128L,
                      video_connector_num_layers = 8L,
                      video_connector_num_learnable_registers = 128L,
                      video_gated_attn = TRUE,
                      audio_connector_num_attention_heads = 32L,
                      audio_connector_attention_head_dim = 64L,
                      audio_connector_num_layers = 8L,
                      audio_connector_num_learnable_registers = 128L,
                      audio_gated_attn = TRUE,
                      connector_rope_base_seq_len = 4096L, rope_theta = 10000,
                      rope_double_precision = TRUE, rope_type = "split",
                      video_hidden_dim = 4096L, audio_hidden_dim = 2048L,
                      proj_bias = TRUE)
}
\arguments{
\item{caption_channels}{Integer. Text encoder hidden size (3840 for
Gemma3-12B).}

\item{text_proj_in_factor}{Integer. Number of stacked hidden states
(num_layers + 1 = 49 for Gemma3-12B).}

\item{video_connector_num_learnable_registers}{Integer or NULL.}

\item{video_gated_attn}{Logical.}

\item{audio_connector_num_learnable_registers}{Integer or NULL.}

\item{audio_gated_attn}{Logical.}

\item{proj_bias}{Logical. Projection bias (TRUE for LTX-2.3).}

\item{video_connector_num_attention_heads,video_connector_attention_head_dim,video_connector_num_layers}{Video connector shape (LTX-2.3: 32 x 128, 8 layers).}

\item{audio_connector_num_attention_heads,audio_connector_attention_head_dim,audio_connector_num_layers}{Audio connector shape (LTX-2.3: 32 x 64, 8 layers).}

\item{connector_rope_base_seq_len,rope_theta,rope_double_precision,rope_type}{RoPE config.}

\item{video_hidden_dim,audio_hidden_dim}{Integers. Projection targets
(DiT inner dims: 4096 / 2048).}
}
\value{
Module whose forward(text_encoder_hidden_states,
  attention_mask) returns \code{list(video_text_embedding,
  audio_text_embedding, attention_mask)}: the caption states adapted
  for the video and audio cross-attention streams, plus the binary
  mask to use with them.
}
\description{
Takes raw stacked per-layer text encoder hidden states and produces
the video and audio text embeddings for the DiT: per-token RMS norm,
per-modality sqrt(dim ratio) rescaling and projection, then a
per-modality 1D connector transformer.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{quick_gelu}
\alias{quick_gelu}
\title{QuickGELU activation}
\usage{
quick_gelu(x)
}
\arguments{
\item{x}{Input tensor}
}
\description{
GELU approximation used by OpenAI CLIP: x * sigmoid(1.702 * x)
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_text_encoder2_weights}
\alias{load_text_encoder2_weights}
\title{Load weights from TorchScript text encoder 2 into native encoder}
\usage{
load_text_encoder2_weights(native_encoder, torchscript_path, verbose = TRUE)
}
\arguments{
\item{native_encoder}{Native text encoder 2 module}

\item{torchscript_path}{Path to TorchScript encoder .pt file}

\item{verbose}{Print loading progress}
}
\value{
The native encoder with loaded weights (invisibly)
}
\description{
Load weights from TorchScript text encoder 2 into native encoder
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{tokenizer_unigram}
\alias{tokenizer_unigram}
\title{SentencePiece Unigram Tokenizer}
\description{
Pure R implementation of HuggingFace tokenizer.json files with a
Unigram model (SentencePiece), as used by T5 - FLUX's second text
encoder. Segmentation is Viterbi best-path over the vocab log
probabilities (Kudo 2018, arXiv:1804.10959). The normalizer and
Metaspace pre-tokenizer settings are read from the file.
}
\details{
Limitation: the Precompiled charsmap normalizer (NFKC-style unicode
mapping) is approximated by control-whitespace substitution only;
ASCII and common latin text tokenizes identically to the reference,
exotic unicode may differ.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_gemma3_nf4}
\alias{load_gemma3_nf4}
\title{Load a Gemma3 text encoder from an NF4 artifact}
\usage{
load_gemma3_nf4(artifact_dir, device = "cuda", dtype = "bfloat16",
                pin = getOption("diffuseR.pin_staging", TRUE), verbose = TRUE)
}
\arguments{
\item{artifact_dir}{Directory produced by
\code{\link{gemma3_quantize_nf4}}.}

\item{device}{"cuda" or "cpu".}

\item{dtype}{Compute dtype ("bfloat16" default).}

\item{pin}{Logical. When loading to the CPU, page-lock the weights
so \code{\link{encode_with_gemma3}} can swap the model to the GPU
at DMA speed per encode and back for free (see
\code{\link{staging}}). Default follows
\code{options(diffuseR.pin_staging)}.}

\item{verbose}{Logical.}
}
\value{
A \code{gemma3_text_model} ready for
  \code{\link{encode_with_gemma3}}.
}
\description{
Builds the model as a skeleton at the compute dtype, swaps the
projection linears for NF4 modules filled from the artifact
(dequantized per forward through the shared byte-LUT), copies the
residents, and hard-errors on any parameter the artifact does not
fill.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{reshard_safetensors}
\alias{reshard_safetensors}
\title{Re-shard a large safetensors file into sub-2 GB shards}
\usage{
reshard_safetensors(input, output_dir, base = "diffusion_pytorch_model",
                    shard_bytes = 1.9e+09, verbose = TRUE)
}
\arguments{
\item{input}{Path to the source \code{.safetensors} file, or a
directory containing \code{<base>.safetensors}.}

\item{output_dir}{Output directory for the shards + index.}

\item{base}{Shard basename (default
\code{"diffusion_pytorch_model"}).}

\item{shard_bytes}{Target shard size; the default 1.9e9 keeps each
shard under the ~2.15 GB ceiling.}

\item{verbose}{Logical.}
}
\value{
Invisibly, the path to the written index.json.
}
\description{
Splits a single \code{.safetensors} file into diffusers-style shards
(\code{<base>-00001-of-000NN.safetensors} plus a
\code{<base>.safetensors.index.json} weight map) so it loads on stock
CRAN safetensors, which overflows a 32-bit offset on any file at or
above 2^31 bytes. Reading the oversize source requires a
fork-patched safetensors (a build-machine step); the shards it writes
are fork-free to read. Used to host large fp16 diffusers weights
(e.g. the 5 GB SDXL UNet) unchanged, without quantization.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{resident_load}
\alias{resident_load}
\title{Load a diffusion pipeline as a resident handle}
\usage{
resident_load(model = c("flux2", "flux1", "zimage", "ltx"), device = "cuda",
              ..., verbose = TRUE)
}
\arguments{
\item{model}{One of "flux1", "flux2", "zimage", "ltx".}

\item{device}{Target CUDA device, e.g. "cuda" or "cuda:1".}

\item{...}{Passed to the family loader (\code{\link{flux_load_pipeline}},
\code{\link{flux2_load_pipeline}}, \code{\link{zimage_load_pipeline}}
or \code{\link{ltx23_load_pipeline}}). \code{ltx} requires
\code{checkpoint_path}.}

\item{verbose}{Print progress messages.}
}
\value{
A \code{diffuseR_resident} handle (an environment). Inspect it
  with \code{\link{resident_status}}; the fields of interest are the
  state, the bound device, the component names, and the pinned host
  byte count.
}
\description{
Loads a pipeline once and keeps its weights page-locked on the host
for the life of the handle. The GPU representation is created by
\code{\link{resident_activate}} and destroyed by
\code{\link{resident_deactivate}}, so a 16 GB card can hand itself
between models without either one re-reading its weights from disk.
}
\details{
The handle is bound to one explicit GPU at load: a bare \code{"cuda"}
resolves to the current device now, and every later transition uses
that index, so the handle cannot drift to whichever GPU happens to be
current at transition time.

One caveat on multi-GPU hosts: the family loader itself runs on the
\emph{current} device, and only the residency handle is bound to
\code{device}. Loading with \code{device = "cuda:1"} from a session
whose current device is 0 therefore stages through GPU 0 before the
first activation lands on GPU 1. Wrap the call in
\code{torch::with_device(device = "cuda:1", ...)} when that matters.

The pipeline is left \emph{inactive} (weights pinned on the host, no
VRAM held). Call \code{\link{resident_activate}} before generating.

}
\examples{
\dontrun{
res <- resident_load("flux2")
resident_activate(res)
img <- resident_generate(res, "a cat in a spacesuit", seed = 7)
resident_deactivate(res) # VRAM freed, weights stay pinned in RAM
resident_activate(res) # fast: DMA copy, no disk
resident_unload(res)
}

}
\seealso{
\code{\link{resident_activate}}, \code{\link{resident_status}}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_open_checkpoint}
\alias{ltx23_open_checkpoint}
\title{Open an LTX-2.3 checkpoint}
\usage{
ltx23_open_checkpoint(path, require_version = "2.3")
}
\arguments{
\item{path}{Path to the checkpoint .safetensors file.}

\item{require_version}{Character. Required \code{model_version} prefix
(default "2.3"). Set to NULL to skip the check.}
}
\value{
An object of class \code{ltx23_checkpoint}: a list with
  \code{handle} (safetensors reader), \code{keys}, \code{version},
  \code{config} (parsed component configs, or NULL), and \code{path}.
}
\description{
Opens a single-file LTX checkpoint lazily (header only), validates the
\code{model_version} metadata, and parses the embedded component
configuration.
}
\examples{
\dontrun{
ckpt <- ltx23_open_checkpoint("ltx-2.3-22b-distilled-1.1.safetensors")
str(ltx23_split_keys(ckpt$keys), max.level = 1)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_load_transformer_fp8}
\alias{ltx23_load_transformer_fp8}
\title{Load the LTX-2.3 transformer with FP8 weights}
\usage{
ltx23_load_transformer_fp8(ckpt, device = "cuda", pin = TRUE, verbose = TRUE,
                           ...)
}
\arguments{
\item{ckpt}{An fp8 \code{ltx23_checkpoint}
(\code{\link{ltx23_open_fp8_checkpoint}}).}

\item{device}{Character. Device for the resident (non-fp8) weights.}

\item{pin}{Logical. Pin the fp8 host memory for faster transfers.}

\item{verbose}{Logical.}

\item{...}{Passed to \code{\link{ltx23_transformer}} (tiny test configs).}
}
\value{
The loaded \code{ltx23_transformer}.
}
\description{
Builds the transformer, swaps the official cast-set linears for
\code{\link{ltx23_fp8_linear}}, loads fp8 weights CPU-side (optionally
pinned) and everything else as bfloat16 on \code{device}. Sets
\code{options(diffuseR.block_gc = TRUE)} so the transformer runs
per-block garbage collection over the dequantized temporaries.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_empirical_mu}
\alias{flux2_empirical_mu}
\title{Empirical timestep shift for FLUX.2}
\usage{
flux2_empirical_mu(image_seq_len, num_steps)
}
\arguments{
\item{image_seq_len}{Integer. Packed image token count.}

\item{num_steps}{Integer. Inference steps.}
}
\value{
Numeric mu for \code{\link{flowmatch_set_timesteps}}.
}
\description{
BFL's piecewise-linear fit of the dynamic-shifting mu as a function of
image sequence length and step count; replaces FLUX.1's
calculate_shift. Reference: compute_empirical_mu (adapted from BFL
sampling.py).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_decoder_safetensors}
\alias{load_decoder_safetensors}
\title{Load HF safetensors VAE weights into the native decoder}
\usage{
load_decoder_safetensors(native_decoder, path, verbose = TRUE)
}
\arguments{
\item{native_decoder}{Native VAE decoder module}

\item{path}{Path to the VAE .safetensors file (or a directory
containing diffusion_pytorch_model.safetensors)}

\item{verbose}{Print loading progress}
}
\value{
The native decoder with loaded weights (invisibly)
}
\description{
Loads the decoder half of a diffusers AutoencoderKL safetensors file
(e.g. FLUX.1-schnell's \code{vae/diffusion_pytorch_model.safetensors}).
Keys under \code{decoder.} map to the native module 1:1; encoder and
quant-conv keys are skipped (the FLUX VAE has no quant convs, and
txt2img needs no encoder).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_flux2}
\alias{download_flux2}
\title{Download and Prepare FLUX.2 Klein 4B Weights}
\description{
Downloads FLUX.2-klein-4B from HuggingFace (Apache-2.0, ungated) and
quantizes the 4B transformer to a local fp8 (~4 GB) or NF4 (~2.3 GB)
artifact.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{sd_pipeline_from_safetensors}
\alias{sd_pipeline_from_safetensors}
\title{Assemble a native SD pipeline from a diffusers safetensors directory}
\usage{
sd_pipeline_from_safetensors(diffusers_dir, model_name = "sd21",
                             devices = NULL, unet_dtype = NULL, verbose = TRUE)
}
\arguments{
\item{diffusers_dir}{Directory with \code{unet/}, \code{vae/},
\code{text_encoder/} subdirectories.}

\item{model_name}{Currently "sd21" (SDXL pending its second encoder).}

\item{devices}{Named list of component devices (\code{unet},
\code{decoder}, \code{text_encoder}); defaults to all-CPU.}

\item{unet_dtype}{A torch dtype for the UNet (default float16 on CUDA,
float32 on CPU).}

\item{verbose}{Logical.}
}
\value{
A list with \code{unet}, \code{decoder}, \code{text_encoder}.
}
\description{
Builds the native UNet, VAE decoder, and CLIP text encoder from a
diffusers directory using the \code{*_from_safetensors} constructors,
places each on its component device, and returns the \code{$unet /
$decoder / $text_encoder} list the \code{txt2img_*} denoise loop
expects.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{convert_sd21_pt_to_diffusers}
\alias{convert_sd21_pt_to_diffusers}
\title{Convert cornball SD 2.1 TorchScript weights to a diffusers artifact}
\usage{
convert_sd21_pt_to_diffusers(pt_dir = NULL, output_dir = NULL,
                             dtype = c("float16", "float32"), verbose = TRUE)
}
\arguments{
\item{pt_dir}{Directory holding \code{unet-cpu.pt},
\code{decoder-cpu.pt}, \code{text_encoder-cpu.pt} (default: the
diffuseR \code{sd21} data location).}

\item{output_dir}{Output diffusers directory (default: the
\code{sd_pipeline_from_safetensors} / \code{download_sd21} location).}

\item{dtype}{\code{"float16"} (default, the hosted tier) or
\code{"float32"}.}

\item{verbose}{Logical.}
}
\value{
Invisibly, \code{output_dir}.
}
\description{
Rebuilds a diffusers-layout directory (\code{unet/}, \code{vae/},
\code{text_encoder/}) from the cornball-ai/sd21-R TorchScript
component \code{.pt} files, so the native safetensors pipeline
(\code{\link{download_sd21}} / \code{\link{sd_pipeline_from_safetensors}})
can load SD 2.1 with no TorchScript.
}
\details{
A TorchScript trace preserves the exact parameter tensors, so the
result is bit-identical to the source at the chosen dtype. This is the
provenance-clean way to build the hosted artifact: the upstream
\code{stabilityai/stable-diffusion-2-1} repo was deprecated, SD 2.1 is
CreativeML OpenRAIL++-M (redistributable), and cornball already hosts
these weights as \code{.pt}. At \code{float16} the components are all
sub-2 GB single files (unet ~1.7 GB, text_encoder ~0.65 GB, vae
~0.16 GB), so they load on stock CRAN safetensors.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_model}
\alias{download_model}
\title{Download TorchScript model files for Stable Diffusion}
\usage{
download_model(model_name = "sd21",
               devices = list(unet = "cpu", decoder = "cpu", text_encoder = "cpu"),
               unet_dtype_str = NULL, overwrite = FALSE, show_progress = TRUE,
               download_models = FALSE)
}
\arguments{
\item{model_name}{Name of the model (e.g., "sd21" for stable-diffusion-2-1)}

\item{devices}{Either a single device string or a named list with elements 'unet', 'decoder', 'text_encoder'; optionally 'encoder'}

\item{unet_dtype_str}{Optional: "float16" or "float32" (only applies if unet uses CUDA)}

\item{overwrite}{If TRUE, force re-download of model files}

\item{show_progress}{Show download progress messages}

\item{download_models}{If TRUE, download the model files from HuggingFace}
}
\value{
A named list of full file paths, keyed by component name.
}
\description{
Downloads the required model files (e.g., UNet, decoder, text encoder)
for a given Stable Diffusion model using \code{hfhub::hub_download()}.
}
\details{
Files are cached by hfhub (typically \code{~/.cache/huggingface/hub/}).
Legacy files in the old \code{R_user_dir()} location are also recognized.

}
\examples{
\dontrun{
paths <- download_model("sd21")
}

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{rope_zimage}
\alias{rope_zimage}
\title{Z-Image Rotary Positional Embeddings and Patchify Helpers}
\description{
Fresh R port of the Z-Image position scheme from the diffusers
reference (Apache-2.0,
src/diffusers/models/transformers/transformer_z_image.py RopeEmbedder,
create_coordinate_grid, _patchify_image, _pad_with_ids, unpatchify).
Z-Image uses 3-axis interleaved RoPE with theta 256; frequencies are
built in float64 but the angles are cast to float32 before cos/sin
(torch.polar on a .float() tensor), which differs measurably from the
FLUX convention at large positions. Every sub-sequence is padded to a
multiple of 32 (SEQ_MULTI_OF); caption positions are a 1-based ramp on
axis 1 built over the padded length, image positions sit on axes 2/3
with axis 1 offset just past the caption.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{t5_encoder}
\alias{t5_encoder}
\title{T5 encoder stack}
\usage{
t5_encoder(vocab_size = 32128L, d_model = 4096L, d_kv = 64L, num_heads = 64L,
           d_ff = 10240L, num_layers = 24L,
           relative_attention_num_buckets = 32L,
           relative_attention_max_distance = 128L, layer_norm_epsilon = 1e-06)
}
\arguments{
\item{layer_norm_epsilon}{Numeric.}

\item{vocab_size,d_model,d_kv,num_heads,d_ff,num_layers}{Integers.}

\item{relative_attention_num_buckets,relative_attention_max_distance}{Integers. Relative position bias shape.}
}
\value{
Module whose forward(input_ids) (1-based ids [B, S]) returns
  the last hidden state [B, S, d_model].
}
\description{
Defaults are the T5-v1.1-XXL configuration used by FLUX.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{nf4_ltx23}
\alias{nf4_ltx23}
\title{NF4 Weight Storage for the LTX-2.3 Transformer}
\description{
4-bit NormalFloat quantization (the QLoRA scheme: per-block absmax
normalization against a 16-level quantile code, two indices packed
per byte). At ~4.5 bits/parameter the 22B transformer fits in about
12.5 GB, small enough to stay resident on a 16 GB GPU: no per-step
PCIe weight streaming, at a small quality cost relative to fp8.
Quantization and dequantization are pure torch ops (bucketize,
index_select) - no custom kernels.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{encode_with_qwen3}
\alias{encode_with_qwen3}
\title{Encode prompts with the Qwen3 encoder for FLUX.2}
\usage{
encode_with_qwen3(prompts, model, tokenizer, max_sequence_length = 512L,
                  out_layers = c(9L, 18L, 27L), device = NULL)
}
\arguments{
\item{prompts}{Character vector.}

\item{model}{A \code{\link{qwen3_encoder}}.}

\item{tokenizer}{A \code{\link{qwen_bpe_tokenizer}}.}

\item{max_sequence_length}{Integer. Fixed token length (klein: 512).}

\item{out_layers}{Integer vector. Hidden-state layers (klein-4B:
9, 18, 27).}

\item{device}{Device for the input ids (defaults to the model's).}
}
\value{
Tensor [length(prompts), max_sequence_length,
  3 * hidden_size].
}
\description{
Tokenizes with the chat template, runs the encoder with the padding
mask, and concatenates the requested mid-stack hidden states per
token, matching Flux2KleinPipeline._get_qwen3_prompt_embeds.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_open_fp8_checkpoint}
\alias{ltx23_open_fp8_checkpoint}
\title{Open an FP8 shard directory as a checkpoint}
\usage{
ltx23_open_fp8_checkpoint(dir)
}
\arguments{
\item{dir}{The fp8 artifact directory (with manifest.json).}
}
\value{
An \code{ltx23_checkpoint}.
}
\description{
Presents the sharded fp8 artifact through the same interface as
\code{\link{ltx23_open_checkpoint}} so the group loaders work
unchanged.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{setup_dtype}
\alias{setup_dtype}
\title{Set up dtype based on device configuration}
\usage{
setup_dtype(devices, unet_dtype_str)
}
\arguments{
\item{devices}{A character string or a named list specifying the devices for model components.}

\item{unet_dtype_str}{A character string specifying the data type for the UNet model.}
}
\value{
A torch dtype object based on the main computation device.
}
\description{
Set up dtype based on device configuration
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_sdxl}
\alias{download_sdxl}
\title{Download the Stable Diffusion XL diffusers weights}
\usage{
download_sdxl(verbose = TRUE)
}
\arguments{
\item{verbose}{Logical.}
}
\value{
Invisibly, the diffusers directory (the parent of
  \code{unet/}, \code{vae/}, \code{text_encoder/},
  \code{text_encoder_2/}).
}
\description{
Fetches the UNet (re-sharded to sub-2 GB shards), VAE, and both CLIP
text encoders from the \code{cornball-ai/sdxl-R} HuggingFace dataset
(fp16 diffusers safetensors, converted from the original
\code{stabilityai/stable-diffusion-xl-base-1.0} OpenRAIL++ weights).
About 7 GB, one-time. The native tokenizer and DDIM scheduler need no
downloads.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_ada_layer_norm_single}
\alias{ltx23_ada_layer_norm_single}
\title{Adaptive layer norm single (adaLN-single)}
\usage{
ltx23_ada_layer_norm_single(embedding_dim, num_mod_params = 6L)
}
\arguments{
\item{embedding_dim}{Integer. Model dimension.}

\item{num_mod_params}{Integer. Number of modulation parameter vectors.}
}
\value{
Module whose forward returns
  \code{list(mod_params [N, num_mod_params * dim], embedded_timestep [N, dim])}.
}
\description{
Embeds a timestep/sigma and projects it to a configurable number of
modulation parameter vectors.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_transformer}
\alias{ltx23_transformer}
\title{LTX-2.3 video transformer model}
\usage{
ltx23_transformer(in_channels = 128L, out_channels = 128L, patch_size = 1L,
                  patch_size_t = 1L, num_attention_heads = 32L,
                  attention_head_dim = 128L, cross_attention_dim = 4096L,
                  vae_scale_factors = c(8L, 32L, 32L), pos_embed_max_pos = 20L,
                  base_height = 2048L, base_width = 2048L, gated_attn = TRUE,
                  cross_attn_mod = TRUE, audio_in_channels = 128L,
                  audio_out_channels = 128L, audio_patch_size = 1L,
                  audio_patch_size_t = 1L, audio_num_attention_heads = 32L,
                  audio_attention_head_dim = 64L,
                  audio_cross_attention_dim = 2048L, audio_scale_factor = 4L,
                  audio_pos_embed_max_pos = 20L, audio_sampling_rate = 16000L,
                  audio_hop_length = 160L, audio_gated_attn = TRUE,
                  audio_cross_attn_mod = TRUE, num_layers = 48L,
                  norm_eps = 1e-06, rope_theta = 10000,
                  rope_double_precision = TRUE, causal_offset = 1L,
                  timestep_scale_multiplier = 1000,
                  cross_attn_timestep_scale_multiplier = 1000,
                  rope_type = "split", perturbed_attn = TRUE)
}
\arguments{
\item{cross_attention_dim}{Integer. Video text embedding dimension.}

\item{vae_scale_factors}{Integer vector. VAE (time, height, width) scales.}

\item{audio_cross_attention_dim}{Integer. Audio text embedding dimension.}

\item{num_layers}{Integer. Transformer block count.}

\item{norm_eps}{Numeric. Norm epsilon.}

\item{rope_type}{"split" (LTX-2.3) or "interleaved".}

\item{in_channels,out_channels}{Integers. Video latent channels.}

\item{patch_size,patch_size_t}{Integers. Video patch sizes.}

\item{num_attention_heads,attention_head_dim}{Video attention shape.}

\item{pos_embed_max_pos,base_height,base_width}{RoPE base grid.}

\item{audio_in_channels,audio_out_channels}{Integers. Audio latent channels.}

\item{audio_patch_size,audio_patch_size_t}{Integers. Audio patch sizes.}

\item{audio_num_attention_heads,audio_attention_head_dim}{Audio attention shape.}

\item{audio_scale_factor,audio_pos_embed_max_pos,audio_sampling_rate,audio_hop_length}{Audio latent grid parameters.}

\item{rope_theta,rope_double_precision,causal_offset}{RoPE parameters.}

\item{timestep_scale_multiplier,cross_attn_timestep_scale_multiplier}{Timestep scaling (inputs arrive already scaled; the ratio modulates
the a2v/v2a gates).}

\item{gated_attn,cross_attn_mod,audio_gated_attn,audio_cross_attn_mod,perturbed_attn}{LTX-2.3 feature flags (all TRUE for the 2.3 checkpoints).}
}
\value{
Module whose forward(hidden_states, ...) returns
  \code{list(sample, audio_sample)}: the predicted velocity for the
  video latent tokens and, when the audio branch is active, for the
  audio latent tokens (\code{audio_sample} is NULL otherwise).
}
\description{
Dual-stream audio/video DiT. Text embeddings arrive already projected
to the video (\code{inner_dim}) and audio (\code{audio_inner_dim})
dimensions by the connector modules.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vae_ltx23_modules}
\alias{vae_ltx23_modules}
\title{LTX-2.3 Video VAE Building Blocks}
\description{
Fresh R port of the LTX-2 causal video autoencoder blocks from the
diffusers reference (Apache-2.0,
src/diffusers/models/autoencoders/autoencoder_kl_ltx2.py). Training
and unused inference branches (noise injection, timestep conditioning,
plain-conv downsampling) are intentionally not ported; the 2.3
checkpoints carry no such weights.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_feed_forward}
\alias{ltx23_feed_forward}
\title{LTX feed-forward layer}
\usage{
ltx23_feed_forward(dim, mult = 4L)
}
\arguments{
\item{dim}{Integer. Input/output dimension.}

\item{mult}{Integer. Hidden dimension multiplier.}
}
\value{
Module whose forward(x) returns the projected states passed
  through a tanh-approximated GELU, a tensor of the same shape as
  \code{x}.
}
\description{
Linear -> GELU (tanh approximation) -> Linear with 4x hidden dim,
matching diffusers \code{FeedForward(activation_fn="gelu-approximate")}
state-dict names (net.0.proj, net.2).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{txt2vid_ltx23}
\alias{txt2vid_ltx23}
\title{LTX-2.3 Text-to-Video Pipeline}
\description{
Fresh R port of the LTX-2 text-to-video flow from the diffusers
reference (Apache-2.0, pipelines/ltx2/pipeline_ltx2.py), specialized
for the distilled LTX 2.3 checkpoints: 8-step official sigma schedule,
no classifier-free guidance, joint audio-video denoising with an Euler
velocity step, and audio decoding through the audio VAE and BWE
vocoder to 48 kHz stereo.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{memory_flux}
\alias{memory_flux}
\title{FLUX Memory Profiles}
\description{
VRAM-based execution profiles for the FLUX.1-schnell pipeline. The
12B transformer runs NF4 (~7 GB) or fp8 (~12 GB), phase-onloaded to
the GPU for denoise; the T5-XXL text encoder phase-onloads to the GPU
(bfloat16, pinned) on 14 GB+ cards and computes on the CPU (float32)
below that, where its ~9.8 GB encode phase does not fit.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_audio_decoder}
\alias{ltx23_audio_decoder}
\title{LTX-2.3 audio VAE decoder}
\usage{
ltx23_audio_decoder(base_channels = 128L, output_channels = 2L,
                    num_res_blocks = 2L, latent_channels = 8L,
                    ch_mult = c(1L, 2L, 4L), causality_axis = "height",
                    mel_bins = 64L)
}
\arguments{
\item{base_channels}{Integer.}

\item{output_channels}{Integer. Audio channels (2 = stereo).}

\item{num_res_blocks}{Integer. Per-level ResNet count (a stage runs
\code{num_res_blocks + 1} blocks).}

\item{latent_channels}{Integer.}

\item{ch_mult}{Integer vector. Channel multipliers per level.}

\item{causality_axis}{Character.}

\item{mel_bins}{Integer. Output mel bins (crop/pad target).}
}
\value{
Module whose forward(x) returns the decoded mel
  spectrogram reconstructed from an audio latent.
}
\description{
Latents [B, 8, L, 16] -> mel spectrogram [B, 2, 4L - 3, 64].
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_single_block}
\alias{flux2_single_block}
\title{FLUX.2 single-stream block (parallel)}
\usage{
flux2_single_block(dim, num_attention_heads, attention_head_dim, mlp_ratio = 3,
                   eps = 1e-06, bias = FALSE)
}
\arguments{
\item{dim}{Integer. Model dimension.}

\item{num_attention_heads}{Integer. Attention heads.}

\item{attention_head_dim}{Integer. Per-head dimension.}

\item{mlp_ratio}{Numeric. MLP multiplier (FLUX.2: 3.0).}

\item{eps}{Numeric. Norm epsilon.}

\item{bias}{Logical.}
}
\value{
Module whose forward(hidden_states, temb_mod,
  image_rotary_emb) returns the joint hidden states.
}
\description{
Parameterless LayerNorm with external modulation, then the fused
parallel attention+MLP. Operates on the pre-concatenated [text; image]
sequence (the reference model concatenates once before the stack).
Reference: Flux2SingleTransformerBlock.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_resnet_block3d}
\alias{ltx23_video_resnet_block3d}
\title{LTX 3D ResNet block}
\usage{
ltx23_video_resnet_block3d(in_channels, out_channels = NULL, eps = 1e-06,
                           spatial_padding_mode = "zeros")
}
\arguments{
\item{eps}{Numeric. Shortcut LayerNorm epsilon.}

\item{spatial_padding_mode}{Character.}

\item{in_channels,out_channels}{Integers.}
}
\value{
Module whose forward(inputs, causal) returns \code{inputs}
  plus the residual branch, a tensor of the same shape.
}
\description{
PerChannelRMSNorm -> SiLU -> causal conv, twice, with a LayerNorm +
1x1 Conv3d shortcut when the channel count changes.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_memory_profile}
\alias{flux_memory_profile}
\title{Resolve a FLUX memory profile}
\usage{
flux_memory_profile(vram_gb = NULL)
}
\arguments{
\item{vram_gb}{Numeric or NULL. Available VRAM; auto-detected when
NULL (via nvidia-smi).}
}
\value{
List with \code{name}, \code{precision} ("nf4"/"fp8"/"bf16"),
  \code{attn_chunk}, \code{text_device}, \code{phase_offload},
  \code{max_pixels}, and (advisory) \code{fork_suggested} and
  \code{note}.
}
\description{
A thin adapter over \code{\link{recommend}} for the FLUX.1 pipeline,
kept for back-compatibility. \code{recommend("flux1")} is the policy;
this reshapes it into the legacy profile fields the loader consumes.
Precision now rises with VRAM (nf4 default, fp8 GPU-resident on 14 GB+
cards when safetensors can read float8, bf16 on 24 GB+); the old
bands, which put fp8 in a narrow low-VRAM slot it can no longer fit,
were backwards.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{greedy_tokenize}
\alias{greedy_tokenize}
\title{Greedy longest match tokenization}
\usage{
greedy_tokenize(text, vocab, byte_fallback = FALSE, unk_token = NULL)
}
\description{
Greedy longest match tokenization
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{memory_ltx23}
\alias{memory_ltx23}
\title{LTX-2.3 Memory Profiles and CUDA GC Tuning}
\description{
Memory management for running the 22B LTX-2.3 transformer on limited
VRAM, built on the patterns proven in the whisper and chatterbox
packages: torch allocator GC tuning before the first CUDA op, fp8
CPU-resident streaming weights, query-chunked attention, and
phase-sequential component placement.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_is_quant_key}
\alias{zimage_is_quant_key}
\title{Test whether a Z-Image key is in the quantization cast set}
\usage{
zimage_is_quant_key(key)
}
\arguments{
\item{key}{Character vector of parameter names (diffusers-style).}
}
\value{
Logical vector.
}
\description{
Test whether a Z-Image key is in the quantization cast set
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_cap_pos_ids}
\alias{zimage_cap_pos_ids}
\title{Build Z-Image caption position ids}
\usage{
zimage_cap_pos_ids(cap_padded_len, device = "cpu")
}
\arguments{
\item{cap_padded_len}{Integer. Caption length after padding to a
multiple of 32.}

\item{device}{Device for the resulting tensor.}
}
\value{
Float tensor of shape [cap_padded_len, 3].
}
\description{
Caption tokens ramp 1..cap_padded_len on the first axis (axes 2 and 3
zero). The reference builds the grid over the already-padded length,
so pad tokens continue the ramp rather than sitting at the origin
(the (0,0,0) pad ids it also emits are truncated away in
_prepare_sequence and never reach RoPE).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unet_modules}
\alias{unet_modules}
\title{UNet Modules for Diffusion Models}
\description{
Native R torch modules for UNet architecture.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_flux2_vae_decoder}
\alias{load_flux2_vae_decoder}
\title{Load the FLUX.2 VAE decoder from safetensors}
\usage{
load_flux2_vae_decoder(path, latent_channels = 32L,
                       block_channels = c(512L, 512L, 256L, 128L),
                       norm_groups = 32L, verbose = TRUE)
}
\arguments{
\item{path}{Path to the VAE .safetensors file (or a directory
containing diffusion_pytorch_model.safetensors).}

\item{verbose}{Logical.}

\item{latent_channels,block_channels,norm_groups}{Constructor
arguments for \code{\link{flux2_vae_decoder}}.}
}
\value{
The loaded \code{flux2_vae_decoder} in eval mode.
}
\description{
Loads the decoder half plus post_quant_conv and the BatchNorm running
statistics; encoder and quant_conv keys are skipped (txt2img needs no
encoder).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ddim_scheduler_step}
\alias{ddim_scheduler_step}
\title{Perform a DDIM scheduler step}
\usage{
ddim_scheduler_step(model_output, timestep, sample, schedule, eta = 0,
                    use_clipped_model_output = FALSE, thresholding = FALSE,
                    generator = NULL, variance_noise = NULL,
                    clip_sample = FALSE, set_alpha_to_one = FALSE,
                    prediction_type = c("epsilon", "sample", "v_prediction"),
                    dtype = torch::torch_float32(), device = "cpu")
}
\arguments{
\item{model_output}{Numeric array. The output from the diffusion model, typically
representing predicted noise or the denoised sample depending on `prediction_type`.}

\item{timestep}{Integer. The current timestep in the diffusion process.}

\item{sample}{Numeric array. The current noisy sample at timestep `t`.}

\item{schedule}{List. The DDIM scheduler object containing the necessary parameters created from ddim_scheduler_create()}

\item{eta}{Numeric. Controls the stochasticity of the process. When eta=0, DDIM is
deterministic. When eta=1, it's equivalent to DDPM. Default: 0}

\item{use_clipped_model_output}{Logical. Whether to clip the model output before
computing the sample update. Can improve stability. Default: FALSE}

\item{thresholding}{Logical. Whether to apply thresholding to the output.
Default: FALSE}

\item{generator}{An optional random number generator for reproducibility.
Default: NULL}

\item{variance_noise}{Optional pre-generated noise for the variance when eta > 0.
If NULL and eta > 0, noise will be generated. Default: NULL}

\item{clip_sample}{Logical. Whether to clip the sample. Default: FALSE}

\item{set_alpha_to_one}{Logical. Whether to override the final alpha value to 1.
Used for numerical stability in the final step. Default: FALSE}

\item{prediction_type}{Character. The type of prediction the model outputs.
Options are:
\describe{
\item{"epsilon"}{The model predicts the noise}
\item{"sample"}{The model predicts the denoised sample directly}
\item{"v_prediction"}{The model predicts the velocity vector (v)}
}
Default: "epsilon"}

\item{dtype}{The data type to use for computations. Default is torch_float32().}

\item{device}{The device to use for computations. Options are "cpu" and "cuda".}
}
\value{
A list containing:
  \describe{
    \item{`prev_sample`}{The less noisy sample at timestep t-1}
    \item{`pred_original_sample`}{The predicted denoised sample}
  }
}
\description{
Performs a single denoising step using the DDIM (Denoising Diffusion Implicit Models)
algorithm. This function takes the output from a diffusion model at a specific timestep
and computes the previous (less noisy) sample in the diffusion process.
}
\details{
The DDIM step function implements the core sampling algorithm of DDIM described in
Song et al. 2020. It computes the previous sample x_t-1 given the current
sample x_t and the model output.

The algorithm differs from DDPM by using a non-Markovian diffusion process that
allows for deterministic sampling and fewer inference steps without sacrificing
quality.

When using `prediction_type="epsilon"` (most common), the model predicts the
noise that was added to create the current noisy sample. For `prediction_type="sample"`,
the model predicts the clean sample directly. The `v_prediction` option implements
the v-parameterization from Salimans & Ho (2022).
}
\references{
Song, J., Meng, C., & Ermon, S. (2020).
"Denoising Diffusion Implicit Models."
\url{https://arxiv.org/abs/2010.02502}

Salimans, T., & Ho, J. (2022).
"Progressive Distillation for Fast Sampling of Diffusion Models."
\url{https://arxiv.org/abs/2202.00512}
}
\examples{
if (torch::torch_is_installed()) {
  scheduler <- ddim_scheduler_create(num_inference_steps = 5)
  sample <- torch::torch_randn(c(1, 4, 8, 8))
  model_output <- torch::torch_randn(c(1, 4, 8, 8))
  result <- ddim_scheduler_step(
    model_output = model_output,
    timestep = scheduler$timesteps[1],
    sample = sample,
    schedule = scheduler,
    eta = 0, # Deterministic sampling
    prediction_type = "epsilon")
  result$shape
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_load_pipeline}
\alias{flux2_load_pipeline}
\title{Load the FLUX.2 klein pipeline}
\usage{
flux2_load_pipeline(model_dir = NULL, device = "cuda",
                    precision = c("auto", "fp8", "nf4", "bf16"),
                    text_device = NULL, attn_chunk = NULL,
                    phase_offload = TRUE, pin = NULL, verbose = TRUE)
}
\arguments{
\item{model_dir}{Quantized artifact directory (default: the
\code{download_flux2_klein} location for \code{precision}), or a
raw diffusers transformer directory.}

\item{device}{Character. Compute device.}

\item{precision}{"auto" (default: reuse an existing artifact, else
fp8 when safetensors supports float8, else nf4), "fp8", or "nf4".}

\item{text_device}{Device for the Qwen3 encoder (default:
\code{device}; it encodes in its own phase and offloads).}

\item{attn_chunk}{Integer or NULL. Attention query-chunk override.}

\item{phase_offload}{Logical. One GPU tenant per phase.}

\item{pin}{Logical or NULL. Page-lock the phase-swapped weights for
DMA-rate transfer (see \code{\link{staging}}). NULL (default)
resolves via \code{options(diffuseR.pin_staging)} then the
host-RAM-aware \code{\link{recommend}} decision.}

\item{verbose}{Logical.}
}
\value{
A \code{flux2_pipeline} list.
}
\description{
Loads the quantized transformer artifact plus the FLUX.2 VAE decoder,
Qwen3 text encoder, and tokenizer from the HuggingFace cache
populated by \code{\link{download_flux2_klein}}. With fp8 precision
the ~4 GB transformer stays GPU-resident.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_decoder3d}
\alias{ltx23_video_decoder3d}
\title{LTX-2.3 video decoder}
\usage{
ltx23_video_decoder3d(in_channels = 128L, out_channels = 3L,
                      block_out_channels = c(256L, 512L, 512L, 1024L),
                      spatio_temporal_scaling = c(TRUE, TRUE, TRUE, TRUE),
                      layers_per_block = c(4L, 6L, 4L, 2L, 2L),
                      upsample_type = NULL, patch_size = 4L, patch_size_t = 1L,
                      resnet_norm_eps = 1e-06, is_causal = FALSE,
                      upsample_residual = c(FALSE, FALSE, FALSE, FALSE),
                      upsample_factor = c(2L, 2L, 1L, 2L),
                      spatial_padding_mode = "zeros")
}
\arguments{
\item{block_out_channels}{Integer vector (config order).}

\item{spatio_temporal_scaling}{Logical vector per up block.}

\item{layers_per_block}{Integer vector (config order; first entry is
the mid block after reversal).}

\item{upsample_type}{Character vector per up block (not reversed).}

\item{resnet_norm_eps}{Numeric.}

\item{is_causal}{Logical. FALSE for LTX (symmetric temporal padding).}

\item{upsample_residual}{Logical vector per up block.}

\item{upsample_factor}{Integer vector per up block.}

\item{spatial_padding_mode}{Character.}

\item{in_channels,out_channels}{Integers. Latent and pixel channels.}

\item{patch_size,patch_size_t}{Integers.}
}
\value{
Module whose forward(hidden_states, causal) returns the
  decoded pixel tensor [B, 3, F, H, W], with the final patch axes
  flattened back into height and width.
}
\description{
Latents [B, 128, F, H, W] -> pixel video [B, 3, 8F - 7, 32H, 32W].
Block channel lists are given encoder-side (as in the config) and
reversed internally; \code{upsample_type} is indexed directly.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{txt2img_flux}
\alias{txt2img_flux}
\title{Generate an image with FLUX.1-schnell}
\usage{
txt2img_flux(prompt, pipeline = NULL, width = 1024L, height = 1024L,
             num_inference_steps = 4L, max_sequence_length = 256L, seed = NULL,
             prompt_embeds = NULL, pooled_prompt_embeds = NULL,
             save_file = TRUE, filename = NULL, verbose = TRUE, ...)
}
\arguments{
\item{prompt}{Character. The prompt.}

\item{pipeline}{A \code{flux_pipeline} from
\code{\link{flux_load_pipeline}}; NULL loads one (passing
\code{...} through).}

\item{num_inference_steps}{Integer. Denoising steps (schnell: 4).}

\item{max_sequence_length}{Integer. T5 token length (schnell: 256).}

\item{seed}{Integer or NULL. Initial latents are drawn on the CPU, so
a seed matches a Python diffusers run with a CPU generator.}

\item{save_file}{Logical. Write a PNG.}

\item{filename}{Output path (default derived from the prompt).}

\item{verbose}{Logical, or one of "silent", "progress", "steps".
TRUE = "steps" (full per-phase chatter), FALSE = "silent".
"progress" prints a one-line generation summary plus a denoise
progress bar (interactive) or periodic step ticks (captured logs).}

\item{...}{Passed to \code{\link{flux_load_pipeline}} when
\code{pipeline} is NULL.}

\item{width,height}{Integers, divisible by 16.}

\item{prompt_embeds,pooled_prompt_embeds}{Optional precomputed text
embeddings (skip the text encoders).}
}
\value{
Invisibly, \code{list(image, metadata)} where \code{image} is
  an [H, W, 3] array in [0, 1].
}
\description{
4-step distilled text-to-image generation (no classifier-free
guidance): T5 + CLIP prompt encoding, flow-matching Euler denoising
over the packed latent sequence, and 16-channel VAE decode. With
phase offloading each component is the sole GPU tenant for its phase.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_transformer}
\alias{zimage_transformer}
\title{Z-Image Transformer}
\usage{
zimage_transformer(in_channels = 16L, dim = 3840L, n_layers = 30L,
                   n_refiner_layers = 2L, n_heads = 30L, norm_eps = 1e-05,
                   cap_feat_dim = 2560L, rope_theta = 256, t_scale = 1000,
                   axes_dims = c(32L, 48L, 48L), patch_size = 2L,
                   f_patch_size = 1L)
}
\arguments{
\item{in_channels}{Integer. Latent channels. Default 16.}

\item{dim}{Integer. Model width. Default 3840.}

\item{n_layers}{Integer. Main trunk depth. Default 30.}

\item{n_refiner_layers}{Integer. Refiner depth. Default 2.}

\item{n_heads}{Integer. Attention heads. Default 30.}

\item{norm_eps}{Numeric. RMSNorm epsilon. Default 1e-5.}

\item{cap_feat_dim}{Integer. Caption embedding width. Default 2560.}

\item{rope_theta}{Numeric. RoPE base frequency. Default 256.}

\item{t_scale}{Numeric. Timestep scale. Default 1000.}

\item{axes_dims}{Integer vector. Per-axis rotary dims. Default
c(32, 48, 48).}

\item{patch_size}{Integer. Spatial patch size. Default 2.}

\item{f_patch_size}{Integer. Temporal patch size. Default 1.}
}
\value{
Module whose forward(x, t, cap_feats, chunk_size) returns the
  predicted velocity for the single latent, a tensor [C, F, H, W]
  matching the shape of \code{x}. Note that the checkpoint negates
  this output and consumes a reversed timestep; see
  \code{\link{txt2img_zimage}}.
}
\description{
Fresh R port of ZImageTransformer2DModel from the diffusers reference
(Apache-2.0, src/diffusers/models/transformers/transformer_z_image.py).
Single-stream DiT: image tokens pass through a modulated noise
refiner, caption tokens through an unmodulated context refiner, then
both are concatenated (image first) and run through the main trunk.
The module tree mirrors the reference state-dict keys 1:1
(all_x_embedder.2-1, noise_refiner.N, context_refiner.N, layers.N,
all_final_layer.2-1, t_embedder, cap_embedder, x_pad_token,
cap_pad_token).
}
\details{
This port is batch-of-1: \code{x} is a single latent [C, F, H, W] and
\code{cap_feats} a single caption [L, cap_feat_dim], so sub-sequences
are uniform and no attention mask is needed. Padding to a multiple of
32 tokens uses the learned pad parameters, appended after embedding
(the reference pads raw features with repeats, embeds pointwise, then
overwrites the pad rows with the same learned tokens).

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unet_safetensors}
\alias{unet_safetensors}
\title{Load HF safetensors weights into the native SD/SDXL UNet}
\description{
The native UNet modules mirror the diffusers
\code{UNet2DConditionModel} state-dict keys 1:1, with the sole
exception that the time- (and, for SDXL, add-) embedding MLPs are
flattened from dotted to underscored names
(\code{time_embedding.linear_1} -> \code{time_embedding_linear_1}).
These loaders read \code{unet/diffusion_pytorch_model.safetensors}
(single file or sharded via its \code{.index.json}) and copy each
weight into the matching native parameter, verifying that every native
parameter is filled and no key or shape is left unmatched.
}
\details{
Reads route through the shared sharded opener, so an oversize (>2 GB)
single-file checkpoint on stock CRAN safetensors surfaces the
actionable "rebuild with smaller shards or install the fork" message
rather than a raw 32-bit overflow.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{tokenizer_qwen}
\alias{tokenizer_qwen}
\title{Qwen2 Byte-Level BPE Tokenizer}
\description{
Pure R implementation of the Qwen2 tokenizer (HuggingFace
tokenizer.json, BPE model with ByteLevel pre-tokenization), as used by
FLUX.2 klein's Qwen3 text encoder. Text is split with the GPT-4-style
regex, each pre-token's UTF-8 bytes are mapped through the GPT-2
byte-to-unicode table, and rank-based BPE merges produce the ids.
Added tokens (\code{<|im_start|>}, \code{<think>}, ...) are split out
literally before byte-level encoding.
}
\details{
Limitation: the NFC normalizer is not applied (base R has no NFC);
input is assumed to already be NFC, which holds for ordinary text.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{save_video_ltx23}
\alias{save_video_ltx23}
\title{Save an LTX video (optionally with audio) to MP4}
\usage{
save_video_ltx23(video, filename, fps = 24, audio = NULL, sample_rate = 48000L,
                 verbose = TRUE)
}
\arguments{
\item{video}{Array [frames, height, width, 3] in [0, 1].}

\item{filename}{Output path (.mp4).}

\item{fps}{Numeric.}

\item{audio}{Optional numeric matrix [channels, samples] in [-1, 1].}

\item{sample_rate}{Integer.}

\item{verbose}{Logical.}
}
\value{
Invisibly, the filename.
}
\description{
Uses the av package (Suggests) to encode frames and mux the audio
track.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_t_embedder}
\alias{zimage_t_embedder}
\title{Z-Image timestep embedder}
\usage{
zimage_t_embedder(out_size, mid_size = 1024L, freq_size = 256L)
}
\arguments{
\item{out_size}{Integer. Output width, min(dim, 256).}

\item{mid_size}{Integer. Hidden width. The full model uses 1024.}

\item{freq_size}{Integer. Sinusoid width. Default 256.}
}
\value{
Module whose forward(t) returns the timestep embedding
  [B, out_size].
}
\description{
256-dim cos-first sinusoid (computed in float32) through a
Linear-SiLU-Linear MLP. The model feeds t * t_scale with the
pipeline's t already in [0, 1].
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_apply_split_rotary_emb}
\alias{ltx23_apply_split_rotary_emb}
\title{Apply split rotary embeddings}
\usage{
ltx23_apply_split_rotary_emb(x, freqs)
}
\arguments{
\item{x}{Tensor of shape [B, H, T, D] (per-head layout), or [B, T, H*D]
which is reshaped per-head when \code{freqs} is 4D.}

\item{freqs}{List of two tensors (cos, sin), each [B, H, T, D/2].}
}
\value{
Tensor with the same shape and dtype as \code{x}.
}
\description{
Rotates element pairs formed by splitting the last dimension in half:
element i pairs with element i + d/2. The cos/sin tensors carry half
the head dimension.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{txt2img_sd21}
\alias{txt2img_sd21}
\title{Generate an image from a text prompt using a diffusion pipeline}
\usage{
txt2img_sd21(prompt, negative_prompt = NULL, img_dim = 768, pipeline = NULL,
             devices = "auto", unet_dtype_str = NULL, download_models = FALSE,
             scheduler = "ddim", timesteps = NULL, initial_latents = NULL,
             num_inference_steps = 50, guidance_scale = 7.5, seed = NULL,
             save_file = TRUE, filename = NULL, metadata_path = NULL,
             use_native_decoder = FALSE, use_native_text_encoder = FALSE,
             use_native_unet = FALSE, diffusers_dir = NULL, ...)
}
\arguments{
\item{prompt}{A character string prompt describing the image to generate.}

\item{negative_prompt}{Optional negative prompt to guide the generation.}

\item{img_dim}{Dimension of the output image (e.g., 512 for 512x512).}

\item{pipeline}{Optional A pre-loaded diffusion pipeline. If `NULL`, it will be loaded based on the model name and devices.}

\item{devices}{A named list of devices for each model component (e.g., `list(unet = "cuda", decoder = "cpu", text_encoder = "cpu")`).}

\item{unet_dtype_str}{Optional A character for dtype of the unet component (typically "float16" for cuda and "float32" for cpu; float32 is available for cuda).}

\item{download_models}{Logical indicating whether to download the model files if they are not found.}

\item{scheduler}{Scheduler to use (e.g., `"ddim"`, `"euler"`).}

\item{timesteps}{Optional A vector of timesteps to use.}

\item{initial_latents}{Optional initial latents for the diffusion process.}

\item{num_inference_steps}{Number of inference steps to run.}

\item{guidance_scale}{Scale for classifier-free guidance (typically 7.5).}

\item{seed}{Optional seed for reproducibility.}

\item{save_file}{Logical indicating whether to save the generated image.}

\item{filename}{Optional filename for saving the image. If `NULL`, a default name is generated.}

\item{metadata_path}{Optional file path to save metadata.}

\item{use_native_decoder}{Logical; if TRUE, uses native R torch decoder instead of TorchScript.
Native decoder has better GPU compatibility (especially Blackwell).}

\item{use_native_text_encoder}{Logical; if TRUE, uses native R torch text encoder instead of TorchScript.
Native text encoder has better GPU compatibility (especially Blackwell).}

\item{use_native_unet}{Logical; if TRUE, uses native R torch UNet instead of TorchScript.
Native UNet has better GPU compatibility (especially Blackwell).}

\item{diffusers_dir}{Optional path to a HuggingFace diffusers
directory (with `unet/`, `vae/`, `text_encoder/`). When set, the
pipeline is built natively from safetensors (no TorchScript), via
[sd_pipeline_from_safetensors()]. See [download_sd21()].}

\item{...}{Additional parameters passed to the diffusion process.}
}
\value{
An image array and metadata
}
\description{
This function generates an image based on a text prompt using the Stable Diffusion model.
It allows for various configurations such as model name, device, scheduler, and more.
}
\examples{
\dontrun{
img <- txt2img("a cat wearing sunglasses in space", device = "cuda")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_nf4_quantize}
\alias{ltx23_nf4_quantize}
\title{Quantize a tensor to NF4}
\usage{
ltx23_nf4_quantize(x)
}
\arguments{
\item{x}{Float tensor (any shape; total elements must be a multiple
of 128, i.e. two 64-element blocks - always true for the LTX
linears).}
}
\value{
List with \code{packed} (uint8, two indices per byte) and
  \code{absmax} (float32, one per 64-element block).
}
\description{
Quantize a tensor to NF4
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_audio_encoder}
\alias{ltx23_audio_encoder}
\title{LTX-2.3 audio VAE encoder}
\usage{
ltx23_audio_encoder(base_channels = 128L, in_channels = 2L,
                    num_res_blocks = 2L, latent_channels = 8L,
                    ch_mult = c(1L, 2L, 4L), causality_axis = "height")
}
\arguments{
\item{in_channels}{Integer. Mel channels (2 = stereo).}

\item{base_channels,num_res_blocks,latent_channels,ch_mult,causality_axis}{See \code{\link{ltx23_audio_decoder}}.}
}
\value{
Module whose forward(x) returns the encoded audio latent, a
  tensor downsampled along time and mel axes with the configured
  latent channel count.
}
\description{
Mel spectrogram [B, 2, T, 64] -> latent distribution moments
[B, 2 * latent_channels, ceil(T/4), 16]. Structure mirrors the
decoder: causal convs, parameterless pixel norms, ResNet stages with
stride-2 downsampling between levels (reference LTX2AudioEncoder).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_double_block}
\alias{flux2_double_block}
\title{FLUX.2 double-stream (MMDiT) block}
\usage{
flux2_double_block(dim, num_attention_heads, attention_head_dim, mlp_ratio = 3,
                   eps = 1e-06, bias = FALSE)
}
\arguments{
\item{dim}{Integer. Model dimension.}

\item{num_attention_heads}{Integer. Attention heads.}

\item{attention_head_dim}{Integer. Per-head dimension.}

\item{mlp_ratio}{Numeric. FF multiplier (FLUX.2: 3.0).}

\item{eps}{Numeric. Norm epsilon.}

\item{bias}{Logical.}
}
\value{
Module whose forward(hidden_states, encoder_hidden_states,
  temb_mod_img, temb_mod_txt, image_rotary_emb) returns
  \code{list(encoder_hidden_states, hidden_states)}.
}
\description{
Image and text streams with externally supplied (shift, scale, gate)
modulation triples, joint attention (txt first), and SwiGLU
feed-forwards. Reference: Flux2TransformerBlock.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_rms_norm}
\alias{gemma3_rms_norm}
\title{Gemma3 RMS Normalization}
\usage{
gemma3_rms_norm(dim, eps = 1e-06)
}
\arguments{
\item{dim}{Integer. Hidden dimension.}

\item{eps}{Numeric. Epsilon for numerical stability.}
}
\description{
RMSNorm with optional addition of 1 to weights (Gemma-style).
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{rope_ltx23}
\alias{rope_ltx23}
\title{LTX-2.3 Rotary Positional Embeddings}
\description{
Fresh R port of the LTX rotary positional embedding scheme from the
diffusers reference implementation (Apache-2.0,
src/diffusers/models/transformers/transformer_ltx2.py). LTX 2.3 uses
the "split" RoPE layout everywhere; "interleaved" is kept for
completeness. Frequencies are computed in float64 per the checkpoint
config (frequencies_precision) and applied in float32.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unet_native_from_torchscript}
\alias{unet_native_from_torchscript}
\title{Create native UNet from TorchScript}
\usage{
unet_native_from_torchscript(torchscript_path, verbose = TRUE)
}
\arguments{
\item{torchscript_path}{Path to TorchScript UNet .pt file}

\item{verbose}{Print loading progress}
}
\value{
A native UNet module with loaded weights
}
\description{
Detects architecture and loads weights from a TorchScript UNet file.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_gemma3_text_encoder}
\alias{load_gemma3_text_encoder}
\title{Load Gemma3 Text Model from safetensors}
\usage{
load_gemma3_text_encoder(model_path, device = "cpu", dtype = "float16",
                         pin = getOption("diffuseR.pin_staging", TRUE),
                         verbose = TRUE)
}
\arguments{
\item{model_path}{Character. Path to directory containing model files.}

\item{device}{Character. Device to load model to.}

\item{dtype}{Character. Data type ("float32", "float16", "bfloat16").}

\item{pin}{Logical. When loading to the CPU, page-lock the weights
so \code{\link{encode_with_gemma3}} can swap the model to the GPU
at DMA speed per encode (~0.3 s on, free off) instead of holding
VRAM or reloading. Default follows
\code{options(diffuseR.pin_staging)}.}

\item{verbose}{Logical. Print loading progress.}
}
\value{
Initialized gemma3_text_model with loaded weights.
}
\description{
Loads pre-trained Gemma3 weights from HuggingFace safetensors files.
An NF4 artifact directory (from \code{\link{gemma3_quantize_nf4}})
dispatches to \code{\link{load_gemma3_nf4}}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_encode_audio}
\alias{ltx23_encode_audio}
\title{Encode audio into normalized, packed conditioning latents}
\usage{
ltx23_encode_audio(audio_vae, wav, audio_num_frames, frontend = NULL)
}
\arguments{
\item{audio_vae}{An \code{ltx23_audio_vae} (with encoder weights).}

\item{wav}{Matrix [2, samples] in [-1, 1] at 16 kHz (see
\code{\link{ltx23_read_audio}}).}

\item{audio_num_frames}{Integer. Target latent length.}

\item{frontend}{Optional prebuilt \code{\link{ltx23_audio_mel_frontend}}.}
}
\value{
Packed normalized latents [1, audio_num_frames, 128] (float32).
}
\description{
Pads or trims the waveform so the latent length equals
\code{audio_num_frames} (mel frames \code{4L - 3}, mirroring the
decoder's \code{target_frames}), computes the log-mel, encodes in
argmax mode, packs, and normalizes with the checkpoint statistics.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.pin_component}
\alias{.pin_component}
\title{Pin a component's tensors for fast phase transfer}
\usage{
.pin_component(module, extra = NULL)
}
\arguments{
\item{module}{An nn_module on the CPU.}

\item{extra}{Optional list of additional plain-field tensors to pin
alongside the module's parameters and buffers (e.g. an fp8
linear's \code{weight_fp8}/\code{weight_scale} fields, which live
outside \code{parameters}/\code{buffers}). \code{set_data} mutates
each tensor in place, so the field reference stays valid.}
}
\value{
A list of \code{list(live, pinned)} tensor pairs, or NULL
  if pinning is unavailable (no CUDA, or page-locking failed).
}
\description{
Pin a component's tensors for fast phase transfer
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ddim_scheduler_create}
\alias{ddim_scheduler_create}
\title{Create a DDIM Scheduler}
\usage{
ddim_scheduler_create(num_train_timesteps = 1000, num_inference_steps = 50,
                      eta = 0,
                      beta_schedule = c("linear", "scaled_linear", "cosine"),
                      beta_start = 0.00085, beta_end = 0.012,
                      rescale_betas_zero_snr = FALSE,
                      dtype = torch::torch_float32(),
                      device = torch::torch_device("cpu"))
}
\arguments{
\item{num_train_timesteps}{Integer. The number of diffusion steps used to
train the model. Default: 1000}

\item{num_inference_steps}{Integer. The number of diffusion steps used for
inference. Fewer steps typically means faster inference at the cost of
sample quality. Default: 50}

\item{eta}{Numeric. Controls the amount of
stochasticity. When eta=0, the sampling process is deterministic.
When eta=1, the sampling process is equivalent to DDPM. Default: 0}

\item{beta_schedule}{Character. The beta schedule to use. Options are:
\describe{
\item{"linear"}{Linear beta schedule from beta_start to beta_end}
\item{"scaled_linear"}{Scaled linear schedule, generally gives better results}
\item{"cosine"}{Cosine schedule that approaches zero smoothly}
}
Default: "linear"}

\item{beta_start}{Numeric. The starting value for the beta schedule. Default: 0.00085}

\item{beta_end}{Numeric. The final value for the beta schedule. Default: 0.012}

\item{rescale_betas_zero_snr}{Logical. If TRUE, rescales the beta values}

\item{dtype}{The data type to use for computations. Default is torch_float32(). Options are torch_float16() and torch_float32().}

\item{device}{The device to use for computations. Options are torch_device("cpu"), torch_device("cuda").}
}
\value{
A DDIM scheduler object that can be used with diffusion models to
  generate samples.
}
\description{
Creates a Denoising Diffusion Implicit Models (DDIM) scheduler for use with
diffusion models. DDIM schedulers provide a deterministic sampling process
that offers faster inference compared to DDPM while maintaining high quality
outputs.
}
\details{
DDIM (Denoising Diffusion Implicit Models) was introduced by Song et al. (2020)
as an extension to DDPM (Denoising Diffusion Probabilistic Models). It offers
a deterministic sampling process and allows for controlling the number of
inference steps independently from the training process.

The scheduler contains the noise schedule and methods for computing
alpha, beta, and other parameters used in the diffusion process.
}
\references{
Song, J., Meng, C., & Ermon, S. (2020).
"Denoising Diffusion Implicit Models."
\url{https://arxiv.org/abs/2010.02502}
}
\examples{
if (torch::torch_is_installed()) {
  scheduler <- ddim_scheduler_create(
    num_train_timesteps = 1000,
    num_inference_steps = 5,
    eta = 0.5,
    beta_schedule = "scaled_linear"
  )
  scheduler$timesteps
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{encode_single}
\alias{encode_single}
\title{Encode a single text string}
\usage{
encode_single(tokenizer, text, add_special_tokens = TRUE)
}
\description{
Encode a single text string
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_gemma3_weights}
\alias{load_gemma3_weights}
\title{Load weights into Gemma3 model}
\usage{
load_gemma3_weights(model, weights, verbose = TRUE)
}
\description{
Load weights into Gemma3 model
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{fp8_ltx23}
\alias{fp8_ltx23}
\title{FP8 Weight Storage for the LTX-2.3 Transformer}
\description{
GPU-poor weight handling: the large attention/FFN linears of the DiT
are stored as float8_e4m3fn with per-tensor scales (the official LTX
quantization policy), kept CPU-resident (optionally pinned), and
dequantized on the compute device inside each forward. Everything
else (norms, embeddings, modulation tables, biases) stays bfloat16.
Requires a safetensors build with F8 dtype support.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{create_sliding_window_mask}
\alias{create_sliding_window_mask}
\title{Create sliding window causal attention mask}
\usage{
create_sliding_window_mask(seq_len, window_size, device = "cpu")
}
\description{
Create sliding window causal attention mask
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_to_gpu}
\alias{load_to_gpu}
\title{Load Module to GPU}
\usage{
load_to_gpu(module, device = "cuda")
}
\arguments{
\item{module}{A torch nn_module.}

\item{device}{Character. Target device (default "cuda").}
}
\value{
The module (modified in place).
}
\description{
Moves a torch module and all its parameters to CUDA.
}
\examples{
if (torch::torch_is_installed()) {
  model <- torch::nn_linear(4, 2)
  # "cuda" needs a GPU; "cpu" is the portable round trip.
  load_to_gpu(model, device = "cpu")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{group_norm_32}
\alias{group_norm_32}
\title{Group Normalization (32 groups)}
\usage{
group_norm_32(channels)
}
\description{
Group Normalization (32 groups)
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{hf_download_pt}
\alias{hf_download_pt}
\title{Download a single TorchScript component via hfhub}
\usage{
hf_download_pt(model_name, filename, download = TRUE)
}
\arguments{
\item{model_name}{Model name (e.g., \code{"sd21"}, \code{"sdxl"}).}

\item{filename}{Filename within the repo (e.g., \code{"unet-cpu.pt"}).}

\item{download}{If \code{TRUE}, download from HuggingFace when not cached.}
}
\value{
The local file path (character string).
}
\description{
Internal helper that wraps \code{hfhub::hub_download()} with the cornball-ai
dataset repo naming convention. Falls back to legacy \code{R_user_dir()} paths
if the file exists there but not yet in the hfhub cache.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{encode_unigram}
\alias{encode_unigram}
\title{Encode text with a Unigram tokenizer}
\usage{
encode_unigram(tokenizer, texts, max_length = 256L, add_eos = TRUE, pad = TRUE)
}
\arguments{
\item{tokenizer}{A \code{\link{unigram_tokenizer}}.}

\item{texts}{Character vector of prompts.}

\item{max_length}{Integer. Fixed sequence length (NULL for no
truncation/padding).}

\item{add_eos}{Logical. Append the EOS token.}

\item{pad}{Logical. Right-pad to \code{max_length}.}
}
\value{
List with \code{input_ids} and \code{attention_mask}, each an
  integer matrix [length(texts), max_length] (or ragged lists when
  \code{max_length} is NULL). Ids are 0-based (HuggingFace
  convention); add 1 for R torch embedding lookups.
}
\description{
Normalizes (strip-right, multi-space collapse, control whitespace to
space), applies the Metaspace pre-tokenizer, segments each pre-token
by Viterbi over the Unigram scores, fuses consecutive unknowns, and
appends EOS. T5 semantics: right padding with \code{<pad>} (id 0),
truncation to \code{max_length - 1} before the EOS.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flowmatch_scale_noise}
\alias{flowmatch_scale_noise}
\title{Scale noise for flow matching forward process}
\usage{
flowmatch_scale_noise(sample, timestep, noise, schedule)
}
\arguments{
\item{sample}{torch tensor. The clean sample.}

\item{timestep}{torch tensor. The current timestep.}

\item{noise}{torch tensor. The noise tensor.}

\item{schedule}{List. The FlowMatch scheduler object.}
}
\value{
torch tensor. The noisy sample at timestep t.
}
\description{
Applies the forward process in flow-matching: interpolates between
the clean sample and noise.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vae_decoder_native}
\alias{vae_decoder_native}
\title{Native VAE Decoder}
\usage{
vae_decoder_native(latent_channels = 4, out_channels = 3,
                   block_channels = c(512, 512, 256, 128), norm_groups = 32)
}
\arguments{
\item{latent_channels}{Number of latent channels (4 for SD/SDXL,
16 for FLUX/SD3)}

\item{out_channels}{Number of output channels (default 3 for RGB)}

\item{block_channels}{Decoder block channels (reversed encoder
block_out_channels; default matches SD/SDXL and FLUX)}

\item{norm_groups}{Group norm groups (default 32; must divide every
entry of \code{block_channels})}
}
\value{
An nn_module representing the VAE decoder
}
\description{
Native R torch implementation of the SDXL VAE decoder.
Replaces TorchScript decoder for better GPU compatibility.
}
\examples{
if (torch::torch_is_installed()) {
  # A small decoder; the SD/SDXL defaults are far too large to build
  # inside an example.
  decoder <- vae_decoder_native(latent_channels = 4,
                                block_channels = 32,
                                norm_groups = 32)
  latents <- torch::torch_randn(c(1, 4, 8, 8))
  image <- torch::with_no_grad(decoder(latents))
  image$shape
}

# Real weights come from a downloaded checkpoint.
\dontrun{
decoder <- vae_decoder_native()
load_decoder_weights(decoder, "path/to/decoder.pt")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_ada_layer_norm_zero_single}
\alias{flux_ada_layer_norm_zero_single}
\title{FLUX adaLN-Zero modulation (single-stream)}
\usage{
flux_ada_layer_norm_zero_single(dim)
}
\arguments{
\item{dim}{Integer. Model dimension.}
}
\value{
Module whose forward(x, emb) returns \code{list(x_norm, gate)}.
}
\description{
Three modulation vectors: shift, scale, gate. Reference: diffusers
AdaLayerNormZeroSingle.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_preprocess_frames}
\alias{ltx23_preprocess_frames}
\title{Preprocess an image (or frame stack) for VAE encoding}
\usage{
ltx23_preprocess_frames(x, height, width)
}
\arguments{
\item{x}{Path to a PNG/JPEG, or an array [H, W, 3] (values in
[0, 1]), or a [F, H, W, 3] array of frames.}

\item{height,width}{Integers. Target size (multiples of 32).}
}
\value{
Float32 tensor [1, 3, F, height, width] in [-1, 1].
}
\description{
Mirrors the diffusers VideoProcessor: bilinear resize so the shorter
relative side matches, center-crop to the exact target, and scale to
[-1, 1].
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_read_tail_frames}
\alias{ltx23_read_tail_frames}
\title{Read the trailing frames of a video file}
\usage{
ltx23_read_tail_frames(path, n = 9L)
}
\arguments{
\item{path}{Video file.}

\item{n}{Integer. Trailing frame count.}
}
\value{
Array [n, H, W, 3] in [0, 1].
}
\description{
Extracts the last \code{n} frames of an MP4 (via \code{av}) for use
as continuation conditioning.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_distilled_sigmas}
\alias{ltx23_distilled_sigmas}
\title{Official distilled sigma schedule}
\usage{
ltx23_distilled_sigmas()
}
\value{
Numeric vector of length 9.
}
\description{
The distilled LTX sigma values (with terminal zero appended), as
published in the Apache-2.0 diffusers reference
(pipelines/ltx2/utils.py).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{detect_text_encoder_architecture}
\alias{detect_text_encoder_architecture}
\title{Detect text encoder architecture from TorchScript file}
\usage{
detect_text_encoder_architecture(torchscript_path)
}
\arguments{
\item{torchscript_path}{Path to TorchScript encoder .pt file}
}
\value{
List with vocab_size, context_length, embed_dim, num_layers, num_heads, mlp_dim
}
\description{
Detect text encoder architecture from TorchScript file
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{dit_flux}
\alias{dit_flux}
\title{FLUX Transformer (MMDiT)}
\description{
Fresh R port of FluxTransformer2DModel from the diffusers reference
implementation (Apache-2.0,
src/diffusers/models/transformers/transformer_flux.py). The module
tree mirrors the diffusers state-dict keys 1:1, so checkpoints load
without remapping. FLUX.1-schnell has no guidance embedder
(guidance_embeds = FALSE); the guidance-distilled dev variant is not
implemented.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{download_zimage}
\alias{download_zimage}
\title{Download and Prepare Z-Image-Turbo Weights}
\description{
Downloads Z-Image-Turbo from HuggingFace (Apache-2.0, ungated) and
quantizes the 6B transformer to a local fp8 (~6.3 GB) or NF4
(~3.6 GB) artifact. The checkpoint ships the transformer in float32
(24.6 GB), so the one-time quantize saves a lot of disk and load
time.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{write_wav}
\alias{write_wav}
\title{Write a 16-bit PCM WAV file}
\usage{
write_wav(audio, path, sample_rate = 48000L)
}
\arguments{
\item{audio}{Numeric matrix [channels, samples] in [-1, 1].}

\item{path}{Output path.}

\item{sample_rate}{Integer.}
}
\value{
Invisibly, the path.
}
\description{
Minimal RIFF writer in base R.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_video_vae}
\alias{ltx23_video_vae}
\title{LTX-2.3 video VAE}
\usage{
ltx23_video_vae(in_channels = 3L, out_channels = 3L, latent_channels = 128L,
                block_out_channels = c(256L, 512L, 1024L, 1024L),
                decoder_block_out_channels = c(256L, 512L, 512L, 1024L),
                layers_per_block = c(4L, 6L, 4L, 2L, 2L),
                decoder_layers_per_block = c(4L, 6L, 4L, 2L, 2L),
                spatio_temporal_scaling = c(TRUE, TRUE, TRUE, TRUE),
                decoder_spatio_temporal_scaling = c(TRUE, TRUE, TRUE, TRUE),
                downsample_type = NULL, upsample_type = NULL,
                upsample_residual = c(FALSE, FALSE, FALSE, FALSE),
                upsample_factor = c(2L, 2L, 1L, 2L), patch_size = 4L,
                patch_size_t = 1L, resnet_norm_eps = 1e-06,
                encoder_causal = TRUE, decoder_causal = FALSE,
                encoder_spatial_padding_mode = "zeros",
                decoder_spatial_padding_mode = "zeros")
}
\arguments{
\item{latent_channels}{Integer.}

\item{resnet_norm_eps}{Numeric.}

\item{in_channels,out_channels}{Integers. Pixel channels.}

\item{block_out_channels,layers_per_block,spatio_temporal_scaling,downsample_type}{Encoder configuration (see \code{\link{ltx23_video_encoder3d}}).}

\item{decoder_block_out_channels,decoder_layers_per_block,decoder_spatio_temporal_scaling,upsample_type,upsample_residual,upsample_factor}{Decoder configuration (see \code{\link{ltx23_video_decoder3d}}).}

\item{patch_size,patch_size_t}{Integers. Pixel patchification.}

\item{encoder_causal,decoder_causal}{Logicals. Temporal padding modes.}

\item{encoder_spatial_padding_mode,decoder_spatial_padding_mode}{Characters.}
}
\value{
Module bundling the video encoder and decoder. Its
  forward(z) is \code{decode(z)}, returning pixels for a latent;
  \code{$encode()} and \code{$decode()} are callable separately.
}
\description{
Encoder + decoder + per-channel latent statistics (loaded from the
checkpoint's \code{per_channel_statistics}). The checkpoint's
\code{scaling_factor} is 1.0, so latent (de)normalization is purely
the per-channel affine map.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_map_vae_key}
\alias{ltx23_map_vae_key}
\title{Map an official VAE checkpoint key to the R module name}
\usage{
ltx23_map_vae_key(key)
}
\arguments{
\item{key}{Character. Checkpoint key (with or without "vae." prefix).}
}
\value{
Character. Module parameter/buffer name.
}
\description{
The official checkpoint stores the encoder/decoder as flat block lists
(down_blocks.0-8 / up_blocks.0-8) where downsamplers/upsamplers and
the mid block are separate entries; diffusers (and this port) nest
them. Index mapping per diffusers convert_ltx2_to_diffusers.py.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_qwen3_text_encoder}
\alias{load_qwen3_text_encoder}
\title{Load a Qwen3 encoder from a transformers directory}
\usage{
load_qwen3_text_encoder(model_path, device = "cpu", dtype = "float32",
                        verbose = TRUE, ...)
}
\arguments{
\item{model_path}{Directory with \code{config.json} and
\code{model*.safetensors} (FLUX.2-klein's \code{text_encoder}).}

\item{device}{Character. Target device.}

\item{dtype}{Character. "bfloat16" (GPU) or "float32" (CPU).}

\item{verbose}{Logical.}

\item{...}{Overrides for \code{\link{qwen3_encoder}} arguments.}
}
\value{
The loaded \code{qwen3_encoder} in eval mode.
}
\description{
Streams the (possibly sharded) safetensors weights into
\code{\link{qwen3_encoder}}. The LM head is tied to the embeddings
and skipped.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vocoder_ltx23}
\alias{vocoder_ltx23}
\title{LTX-2.3 Vocoder with Bandwidth Extension}
\description{
Fresh R port of the LTX-2 BigVGAN-style vocoder from the diffusers
reference (Apache-2.0, pipelines/ltx2/vocoder.py). The 2.3 vocoder
runs a 16 kHz stage (hidden 1536, snakebeta activations with
anti-aliased up/downsampling), re-analyzes its output into a causal
log-mel spectrogram, and feeds a bandwidth-extension vocoder whose
residual is added to a Hann-resampled skip path for 48 kHz output.
The Kaiser sinc / Hann filters and STFT bases are checkpoint buffers.
Runs in float32 (small model; snakebeta is precision-sensitive).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_mlp}
\alias{gemma3_mlp}
\title{Gemma3 MLP}
\usage{
gemma3_mlp(config)
}
\arguments{
\item{config}{List with hidden_size and intermediate_size.}
}
\description{
Feed-forward network with gated linear units and GELU activation.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{BasicTransformerBlock}
\alias{BasicTransformerBlock}
\title{Basic Transformer Block}
\usage{
BasicTransformerBlock(dim, n_heads, d_head, context_dim = NULL)
}
\description{
Basic Transformer Block
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flowmatch_scheduler_create}
\alias{flowmatch_scheduler_create}
\title{Create a FlowMatch Euler Discrete Scheduler}
\usage{
flowmatch_scheduler_create(num_train_timesteps = 1000L, shift = 1,
                           use_dynamic_shifting = FALSE, base_shift = 0.5,
                           max_shift = 1.15, base_seq_len = 256L,
                           max_seq_len = 4096L, invert_sigmas = FALSE,
                           shift_terminal = NULL,
                           time_shift_type = c("exponential", "linear"))
}
\arguments{
\item{num_train_timesteps}{Integer. The number of diffusion steps used to
train the model. Default: 1000}

\item{shift}{Numeric. The shift value for the timestep schedule. Default: 1.0}

\item{use_dynamic_shifting}{Logical. Whether to apply timestep shifting
on-the-fly based on the image/video resolution. Default: FALSE}

\item{base_shift}{Numeric. Value to stabilize generation. Increasing
reduces variation. Default: 0.5}

\item{max_shift}{Numeric. Maximum shift allowed. Increasing encourages
more variation. Default: 1.15}

\item{base_seq_len}{Integer. Base sequence length for dynamic shifting. Default: 256}

\item{max_seq_len}{Integer. Maximum sequence length for dynamic shifting. Default: 4096}

\item{invert_sigmas}{Logical. Whether to invert the sigmas (used by some models
like Mochi). Default: FALSE}

\item{shift_terminal}{Numeric or NULL. End value of shifted schedule. Default: NULL}

\item{time_shift_type}{Character. Type of dynamic shifting: "exponential" or
"linear". Default: "exponential"}
}
\value{
A FlowMatch scheduler object (list) containing:
  \describe{
    \item{sigmas}{The noise schedule}
    \item{timesteps}{The timestep schedule}
    \item{num_train_timesteps}{Training timesteps}
    \item{config}{All configuration parameters}
  }
}
\description{
Creates a FlowMatch scheduler for use with flow-matching diffusion models
like LTX-2. FlowMatch schedulers use Euler integration for sampling,
which is simpler and often faster than DDIM-style schedulers.
}
\details{
FlowMatch (Flow Matching) is a framework for training continuous normalizing
flows by regressing onto target probability paths. The Euler discrete scheduler
implements simple Euler integration for sampling from trained flow models.

The core update rule is:
\code{prev_sample = sample + dt * model_output}
where \code{dt = sigma_next - sigma_current}.
}
\references{
Lipman, Y., Chen, R. T. Q., Ben-Hamu, H., Nickel, M., & Le, M. (2022).
"Flow Matching for Generative Modeling."
\url{https://arxiv.org/abs/2210.02747}
}
\examples{
if (torch::torch_is_installed()) {
  scheduler <- flowmatch_scheduler_create(
    num_train_timesteps = 1000,
    shift = 1.0
  )

  # Set timesteps for inference
  scheduler <- flowmatch_set_timesteps(scheduler, num_inference_steps = 8)
  scheduler$timesteps
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{VAEUpBlock}
\alias{VAEUpBlock}
\title{VAE Up Block}
\usage{
VAEUpBlock(in_channels, out_channels, num_resnets = 3, add_upsample = TRUE,
           norm_groups = 32)
}
\arguments{
\item{in_channels}{Input channels}

\item{out_channels}{Output channels}

\item{num_resnets}{Number of resnet blocks (default 3)}

\item{add_upsample}{Whether to add upsampler}
}
\description{
VAE Up Block
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_audio_downsample}
\alias{ltx23_audio_downsample}
\title{LTX audio downsampler}
\usage{
ltx23_audio_downsample(in_channels, causality_axis = "height")
}
\arguments{
\item{in_channels}{Integer.}

\item{causality_axis}{Character.}
}
\value{
Module whose forward(x) returns the strided convolution of
  \code{x}, halving the downsampled axes.
}
\description{
Causal zero-pad followed by a plain stride-2 conv (reference
LTX2AudioDownsample; note the conv is unwrapped, so its checkpoint
key is \code{downsample.conv.*}).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_open_checkpoint}
\alias{flux_open_checkpoint}
\title{Open a FLUX transformer checkpoint directory}
\usage{
flux_open_checkpoint(transformer_dir)
}
\arguments{
\item{transformer_dir}{Directory containing \code{config.json} and the
\code{diffusion_pytorch_model*.safetensors} file(s).}
}
\value{
An object of class \code{ltx23_checkpoint} (shared checkpoint
  interface): list with \code{handle$get_tensor}, \code{keys},
  \code{config}, and \code{path}.
}
\description{
Opens a diffusers-layout transformer directory lazily (headers only).
Sharded checkpoints are resolved through the index.json weight map;
single-file checkpoints are opened directly. The transformer
\code{config.json} is attached as \code{$config}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_denormalize_latents}
\alias{ltx23_denormalize_latents}
\title{Denormalize latents with the VAE's per-channel statistics}
\usage{
ltx23_denormalize_latents(latents, latents_mean, latents_std)
}
\arguments{
\item{latents}{Tensor [B, C, F, H, W].}

\item{latents_mean,latents_std}{Tensors [C].}
}
\value{
Denormalized latents ready for the decoder.
}
\description{
Denormalize latents with the VAE's per-channel statistics
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{encode_qwen}
\alias{encode_qwen}
\title{Encode prompts with the Qwen tokenizer}
\usage{
encode_qwen(tokenizer, texts, max_length = 512L, chat_template = TRUE,
            enable_thinking = FALSE)
}
\arguments{
\item{tokenizer}{A \code{\link{qwen_bpe_tokenizer}}.}

\item{texts}{Character vector of prompts.}

\item{max_length}{Integer. Fixed sequence length (klein: 512). NULL
for no truncation/padding.}

\item{chat_template}{Logical. Wrap in the Qwen3 chat template.}

\item{enable_thinking}{Logical. Leave the model's thinking enabled
(no empty think block). Default FALSE.}
}
\value{
List with \code{input_ids} and \code{attention_mask} integer
  matrices [length(texts), max_length] (ragged lists when
  \code{max_length} is NULL). Ids are 0-based.
}
\description{
With \code{chat_template = TRUE} each prompt is wrapped as a single
user turn with the generation prompt, matching
\code{apply_chat_template(..., add_generation_prompt = TRUE)}. With
\code{enable_thinking = FALSE} (the FLUX.2 klein pipeline behavior)
the template closes with an empty thinking block; with
\code{enable_thinking = TRUE} (the Z-Image pipeline behavior) it ends
at the assistant turn. Right-pads with \code{<|endoftext|>}.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_t5_text_encoder}
\alias{load_t5_text_encoder}
\title{Load a T5 encoder from a transformers directory}
\usage{
load_t5_text_encoder(model_path, device = "cpu", dtype = "float32",
                     verbose = TRUE, ...)
}
\arguments{
\item{model_path}{Directory with \code{config.json} and
\code{model*.safetensors} (FLUX.1-schnell's \code{text_encoder_2}).}

\item{device}{Character. Target device.}

\item{dtype}{Character. "float32" (CPU default; T5 overflows in
float16) or "bfloat16".}

\item{verbose}{Logical.}

\item{...}{Overrides for \code{\link{t5_encoder}} arguments.}
}
\value{
The loaded \code{t5_encoder} in eval mode.
}
\description{
Streams the (possibly sharded) safetensors weights into
\code{\link{t5_encoder}}, stripping the \code{encoder.} key prefix
and aliasing \code{embed_tokens} to the shared embedding.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_audio_resnet_block}
\alias{ltx23_audio_resnet_block}
\title{LTX audio ResNet block}
\usage{
ltx23_audio_resnet_block(in_channels, out_channels = NULL,
                         causality_axis = "height")
}
\arguments{
\item{causality_axis}{Character.}

\item{in_channels,out_channels}{Integers.}
}
\value{
Module whose forward(x) returns \code{x} plus the residual
  branch, a tensor of the same shape as \code{x}.
}
\description{
PixelNorm -> SiLU -> causal conv, twice, with a 1x1 causal conv
shortcut (\code{nin_shortcut}) on channel change.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_antialias_act1d}
\alias{ltx23_antialias_act1d}
\title{Anti-aliased activation}
\usage{
ltx23_antialias_act1d(channels, ratio = 2L, kernel_size = 12L)
}
\arguments{
\item{channels}{Integer. Channels for the SnakeBeta activation.}

\item{ratio,kernel_size}{Integers. Resampling config.}
}
\value{
Module whose forward(x) returns the activation applied at 2x
  rate (upsample, activate, downsample), a tensor of the same shape
  as \code{x}, with the aliasing the raw activation would introduce
  filtered out.
}
\description{
Upsample 2x, apply the activation, downsample 2x.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{jit_ltx23}
\alias{jit_ltx23}
\title{LTX-2.3 JIT Block Stack}
\description{
TorchScript compilation of the 48-block NF4 transformer step (cf.
the torch skill's JIT-decode pattern proven in whisper and
chatterbox). Eager execution crosses R -> lantern per op (~190 us
each) and leaves every intermediate as an R tensor handle that only
dies at gc(); at high resolution that forces a per-block gc()
costing the vast majority of step time. Compiled, the whole block
stack is one crossing: intermediates are freed eagerly by libtorch,
no R garbage accumulates, no per-block gc is needed, and attention
runs through the fused \code{scaled_dot_product_attention} kernel
instead of a materialized score matrix.
}
\details{
Weights are passed per call as a flat \code{List[Tensor]} (borrowed
by reference, no copies) with a fixed per-block layout; the packer
and the TorchScript indices must stay in lockstep (parity-tested).

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_release_dequant_buffers}
\alias{ltx23_release_dequant_buffers}
\title{Release the NF4 dequantization buffers}
\usage{
ltx23_release_dequant_buffers()
}
\value{
Invisibly, NULL.
}
\description{
Frees the cached per-shape weight buffers (e.g. before decoding at
high resolution).
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_mel_stft}
\alias{ltx23_mel_stft}
\title{Causal log-mel spectrogram with checkpoint-loaded bases}
\usage{
ltx23_mel_stft(filter_length = 512L, hop_length = 80L, window_length = 512L,
               num_mel_channels = 64L)
}
\arguments{
\item{filter_length,hop_length,window_length,num_mel_channels}{Integers.}
}
\value{
Module whose forward(waveform) returns the log-mel
  spectrogram [B, n_mels, frames], clamped at 1e-5 before the log.
}
\description{
Causal log-mel spectrogram with checkpoint-loaded bases
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_audio_mel_frontend}
\alias{ltx23_audio_mel_frontend}
\title{Build the 16 kHz log-mel frontend for audio conditioning}
\usage{
ltx23_audio_mel_frontend(filter_length = 1024L, hop_length = 160L,
                         n_mels = 64L, sample_rate = 16000L, fmin = 0,
                         fmax = 8000)
}
\arguments{
\item{filter_length,hop_length,n_mels,sample_rate,fmin,fmax}{The
checkpoint preprocessing parameters (defaults are LTX-2.3's).}
}
\value{
An \code{ltx23_mel_stft} module.
}
\description{
An \code{\link{ltx23_mel_stft}} whose STFT and mel bases are
constructed (not checkpoint-loaded) with the audio VAE's
preprocessing spec.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{unet_sdxl_native}
\alias{unet_sdxl_native}
\title{Native SDXL UNet}
\usage{
unet_sdxl_native(in_channels = 4L, out_channels = 4L,
                 block_out_channels = c(320L, 640L, 1280L),
                 layers_per_block = 2L,
                 transformer_layers_per_block = c(0L, 2L, 10L),
                 cross_attention_dim = 2048L, attention_head_dim = 64L,
                 addition_embed_dim = 1280L, addition_time_embed_dim = 256L)
}
\arguments{
\item{in_channels}{Input channels (default 4 for latent space)}

\item{out_channels}{Output channels (default 4)}

\item{block_out_channels}{Channel multipliers per block}

\item{layers_per_block}{Number of ResBlocks per down/up block}

\item{transformer_layers_per_block}{Transformer depth per block}

\item{cross_attention_dim}{Context dimension from text encoder}

\item{attention_head_dim}{Dimension per attention head}

\item{addition_embed_dim}{Dimension for additional embeddings}

\item{addition_time_embed_dim}{Dimension for time embedding projection}
}
\value{
An nn_module representing the SDXL UNet
}
\description{
Native R torch implementation of SDXL UNet2DConditionModel.
SDXL has a different architecture from SD21:
- 3 down/up blocks (not 4)
- Variable transformer depth per block
- Additional conditioning via add_embedding
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_load_transformer_nf4}
\alias{ltx23_load_transformer_nf4}
\title{Load the LTX-2.3 transformer with resident NF4 weights}
\usage{
ltx23_load_transformer_nf4(ckpt, device = "cuda", verbose = TRUE, ...)
}
\arguments{
\item{ckpt}{An NF4 \code{ltx23_checkpoint}
(\code{\link{ltx23_open_fp8_checkpoint}} reads any shard artifact).}

\item{device}{Character.}

\item{verbose}{Logical.}

\item{...}{Passed to \code{\link{ltx23_transformer}} (tiny test configs).}
}
\value{
The loaded \code{ltx23_transformer}.
}
\description{
Builds the transformer, swaps the cast-set linears for
\code{\link{ltx23_nf4_linear}}, and loads everything onto
\code{device}: at ~4.5 bits/parameter the whole 22B transformer stays
GPU-resident, avoiding per-step weight transfers.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{repeat_kv}
\alias{repeat_kv}
\title{Repeat KV heads for GQA (Grouped Query Attention)}
\usage{
repeat_kv(hidden_states, n_rep)
}
\arguments{
\item{hidden_states}{Tensor of shape [batch, num_kv_heads, seq_len, head_dim]}

\item{n_rep}{Number of repetitions per KV head}
}
\value{
Tensor of shape [batch, num_kv_heads * n_rep, seq_len, head_dim]
}
\description{
Interleaves KV heads to match query heads: [k0,k0,k1,k1,...] not [k0,k1,...,k0,k1,...]
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_nf4_linear}
\alias{ltx23_nf4_linear}
\title{NF4 linear layer}
\usage{
ltx23_nf4_linear(out_features, in_features, bias = TRUE)
}
\arguments{
\item{bias}{Logical.}

\item{out_features,in_features}{Integers.}
}
\value{
Module whose forward(x) returns the linear projection of
  \code{x}, dequantizing the NF4 weight into a reusable buffer first.
  Same result as an \code{nn_linear} of the same shape, at roughly an
  eighth of the resident weight bytes.
}
\description{
Packed weights and per-block scales are registered as buffers, so
they move with the module (uint8 packs are untouched by dtype
conversions). The forward pass dequantizes on the weight's device.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{sdxl_memory_profile}
\alias{sdxl_memory_profile}
\title{Get SDXL Memory Profile}
\usage{
sdxl_memory_profile(vram_gb = NULL)
}
\arguments{
\item{vram_gb}{Numeric. Available VRAM in GB, or NULL for auto-detection.}
}
\value{
A list with memory profile settings.
}
\description{
Determines optimal memory configuration for SDXL image generation
based on available VRAM.
}
\details{
Memory profiles for SDXL:
\describe{
  \item{full_gpu}{16GB+ - All components on CUDA}
  \item{balanced}{10-12GB - UNet + decoder on CUDA, text encoders on CPU}
  \item{unet_gpu}{6-10GB - Only UNet on CUDA, everything else CPU}
  \item{cpu_only}{<6GB - All on CPU}
}

Each profile also specifies:
- cfg_mode: "batched" or "sequential" (sequential halves peak memory)
- cleanup: "none", "phase", or "step" (when to clear VRAM)
- dtype: "float16" or "float32"
- max_resolution: maximum image dimension
}
\examples{
# A stated VRAM budget is deterministic and needs no GPU.
str(sdxl_memory_profile(vram_gb = 8))

str(sdxl_memory_profile(vram_gb = 24))

# Auto-detect free VRAM on this machine.
str(sdxl_memory_profile())
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_apply_rotary_emb}
\alias{flux_apply_rotary_emb}
\title{Apply FLUX rotary embeddings to a per-head tensor}
\usage{
flux_apply_rotary_emb(x, freqs)
}
\arguments{
\item{x}{Tensor of shape [B, H, S, D] (per-head layout).}

\item{freqs}{List of two tensors (cos, sin), each [S, D], from
\code{flux_pos_embed}.}
}
\value{
Tensor with the same shape and dtype as \code{x}.
}
\description{
Rotates adjacent element pairs of the last dimension:
\code{out = x * cos + rotate_half(x) * sin} with pairs interleaved
(elements 1,2 form the first complex pair). Math in float32, result
cast back to the input dtype. Reference: apply_rotary_emb with
use_real_unbind_dim = -1.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_map_dit_key}
\alias{ltx23_map_dit_key}
\title{Map an official DiT checkpoint key to the R module name}
\usage{
ltx23_map_dit_key(key)
}
\arguments{
\item{key}{Character. Checkpoint key (with or without the
\code{model.diffusion_model.} prefix).}
}
\value{
Character. Module parameter/buffer name.
}
\description{
Applies the official-to-diffusers renames for the LTX-2.3 transformer
(cf. diffusers scripts/convert_ltx2_to_diffusers.py). Our module tree
matches the diffusers names, so this is the full mapping.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.resident_check_fits}
\alias{.resident_check_fits}
\title{Refuse a bulk activation that cannot fit}
\usage{
.resident_check_fits(res, free_gb = NULL)
}
\arguments{
\item{res}{A resident handle.}

\item{free_gb}{Free VRAM in GB. NULL measures it. Pass a value to
make the decision deterministic: with no GPU the measurement is 0,
which means "cannot tell" and never refuses, so a test that wants
the refusal has to state the budget rather than depend on the
machine having a card.}
}
\value{
Invisibly TRUE, or an error naming both figures.
}
\description{
Fails before the transfer rather than part-way through it. A partial
onload that OOMs is recoverable (activation rolls back), but it wastes
the transfer and reports a libtorch allocator error instead of the
actual problem, which is that this model does not fit this card.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vocab_size}
\alias{vocab_size}
\title{Get vocabulary size}
\usage{
vocab_size(tokenizer)
}
\arguments{
\item{tokenizer}{A bpe_tokenizer object.}
}
\value{
Integer vocabulary size.
}
\description{
Get vocabulary size
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{vram_report}
\alias{vram_report}
\title{Report VRAM Usage}
\usage{
vram_report(label = "")
}
\arguments{
\item{label}{Character. Label for the report.}
}
\value{
Invisibly returns a list with used and free VRAM in GB.
}
\description{
Prints current VRAM usage from nvidia-smi.
}
\examples{
if (torch::torch_is_installed()) {
  vram_report("After model load")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{resident_generate}
\alias{resident_generate}
\title{Generate from an active resident pipeline}
\usage{
resident_generate(res, prompt, ...)
}
\arguments{
\item{res}{A \code{diffuseR_resident} handle.}

\item{prompt}{Character. The text prompt.}

\item{...}{Passed to \code{\link{txt2img_flux}},
\code{\link{txt2img_flux2}}, \code{\link{txt2img_zimage}} or
\code{\link{txt2vid_ltx2}}.}
}
\value{
Whatever the family generator returns: an image array for the
  image families, a video array for \code{ltx}.
}
\description{
Dispatches to the family's generator with the resident pipeline
supplied, so no weights are re-read. The handle must be active.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{FeedForward}
\alias{FeedForward}
\title{FeedForward Network}
\usage{
FeedForward(dim, mult = 4L)
}
\description{
FeedForward Network
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{latents_to_video}
\alias{latents_to_video}
\title{Create Video from Latents (Helper)}
\usage{
latents_to_video(latents, vae, file, fps = 24, ...)
}
\arguments{
\item{latents}{Tensor of latents from generation.}

\item{vae}{VAE decoder module.}

\item{file}{Output file path.}

\item{fps}{Frames per second.}

\item{...}{Additional arguments to save_video.}
}
\value{
Invisibly returns the output file path.
}
\description{
Convenience function to decode latents and save video in one step.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{.staged_offload}
\alias{.staged_offload}
\title{Return a pinned component to the CPU}
\usage{
.staged_offload(staging)
}
\description{
Weights are immutable during inference, so the pinned host copies
are still current: offload is a pointer swap, no transfer.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_rotary_pos_embed}
\alias{ltx23_rotary_pos_embed}
\title{LTX-2.3 audio/video rotary position embedder}
\usage{
ltx23_rotary_pos_embed(dim, patch_size = 1L, patch_size_t = 1L,
                       base_num_frames = 20L, base_height = 2048L,
                       base_width = 2048L, sampling_rate = 16000L,
                       hop_length = 160L, scale_factors = c(8L, 32L, 32L),
                       theta = 10000, causal_offset = 1L, modality = "video",
                       double_precision = TRUE, rope_type = "split",
                       num_attention_heads = 32L)
}
\arguments{
\item{dim}{Integer. Rotary dimension (attention head dim x heads for
split type at model level; see reference).}

\item{scale_factors}{Integer vector. VAE (time, height, width) scale
factors.}

\item{theta}{Numeric. RoPE theta.}

\item{causal_offset}{Integer. Temporal offset for the causal VAE
(first frame has stride 1).}

\item{modality}{"video" or "audio".}

\item{double_precision}{Logical. Compute base frequencies in float64.}

\item{rope_type}{"split" (LTX 2.3) or "interleaved".}

\item{num_attention_heads}{Integer. Needed for the split layout.}

\item{patch_size,patch_size_t}{Integers. Spatial/temporal patch sizes.}

\item{base_num_frames,base_height,base_width}{Integers. Base grid the
coordinates are normalized against.}

\item{sampling_rate,hop_length}{Integers. Audio spectrogram params.}
}
\value{
Module whose forward(coords, device) returns
  \code{list(cos_freqs, sin_freqs)}, the two rotary tables to apply
  to queries and keys.
}
\description{
Computes RoPE cos/sin frequency tensors from spatiotemporal patch
coordinates. Video coordinates are 3D (frames scaled to seconds via
fps, height, width in pixel space); audio coordinates are 1D
(seconds). Coordinates are patch boundaries [start, end); the midpoint
is used as the position.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{tokenize_gemma3}
\alias{tokenize_gemma3}
\title{Tokenize text for Gemma3}
\usage{
tokenize_gemma3(tokenizer, text, max_length = 1024L, padding = "max_length",
                return_tensors = "pt")
}
\arguments{
\item{tokenizer}{Gemma3 tokenizer object.}

\item{text}{Character vector of prompts.}

\item{max_length}{Integer. Maximum sequence length.}

\item{padding}{Character. Padding strategy ("left", "right", "max_length", "none").}

\item{return_tensors}{Character. Return type ("list" or "pt" for torch tensors).}
}
\value{
List with input_ids and attention_mask.
}
\description{
Tokenize text for Gemma3
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{get_eos_id}
\alias{get_eos_id}
\title{Get EOS token ID}
\usage{
get_eos_id(tokenizer)
}
\description{
Get EOS token ID
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_unpatchify}
\alias{zimage_unpatchify}
\title{Unpatchify Z-Image tokens back to a latent image}
\usage{
zimage_unpatchify(tokens, size, patch_size = 2L, f_patch_size = 1L,
                  out_channels = 16L)
}
\arguments{
\item{tokens}{Tensor of shape [S, pF * p * p * C] with the image
tokens first.}

\item{size}{Integer vector c(F, H, W) of the target latent size.}

\item{patch_size}{Integer spatial patch size. Default 2.}

\item{f_patch_size}{Integer temporal patch size. Default 1.}

\item{out_channels}{Integer number of latent channels. Default 16.}
}
\value{
Tensor of shape [C, F, H, W].
}
\description{
Takes the first F/pF * H/p * W/p tokens (the image span of the
unified sequence) and reassembles [C, F, H, W], matching unpatchify.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_latent_upsampler}
\alias{ltx23_latent_upsampler}
\title{LTX-2.3 latent upsampler model}
\usage{
ltx23_latent_upsampler(in_channels = 128L, mid_channels = 1024L,
                       num_blocks_per_stage = 4L)
}
\arguments{
\item{in_channels}{Integer. Latent channels.}

\item{mid_channels}{Integer.}

\item{num_blocks_per_stage}{Integer.}
}
\value{
Module whose forward(hidden_states) returns the 2x
  spatially upscaled latent, a tensor with the same batch, channel
  and frame counts and doubled height and width.
}
\description{
Latents [B, 128, F, H, W] -> [B, 128, F, 2H, 2W].
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{offload_to_cpu}
\alias{offload_to_cpu}
\title{Offload Module to CPU}
\usage{
offload_to_cpu(module, gc = TRUE)
}
\arguments{
\item{module}{A torch nn_module.}

\item{gc}{Logical. Run garbage collection after offload.}
}
\value{
The module (modified in place).
}
\description{
Moves a torch module and all its parameters to CPU.
}
\examples{
if (torch::torch_is_installed()) {
  model <- torch::nn_linear(4, 2)
  offload_to_cpu(model)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{dit_zimage_modules}
\alias{dit_zimage_modules}
\title{Z-Image Transformer Block Modules}
\description{
Fresh R port of the Z-Image DiT building blocks from the diffusers
reference (Apache-2.0,
src/diffusers/models/transformers/transformer_z_image.py). Z-Image is
a single-stream DiT: text and image tokens share one sequence and one
set of block weights. Each block uses sandwich RMSNorms (a learned
norm before AND after both the attention and the feed-forward) and a
scale/gate-only modulation — four chunks (scale_msa, gate_msa,
scale_mlp, gate_mlp), no shift, gates tanh-squashed, scales 1 + x.
The attention is plain joint self-attention, so the FLUX attention
module is reused with bias = FALSE and eps = 1e-5.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{save_video}
\alias{save_video}
\title{Save Video to File}
\usage{
save_video(video, file, fps = 24, format = NULL, backend = "auto",
           quality = 85, verbose = TRUE)
}
\arguments{
\item{video}{Array of video frames with shape [T, H, W, C] where C is 3 (RGB).
Values should be in [0, 1] range.}

\item{file}{Character. Output file path. Extension determines format.}

\item{fps}{Numeric. Frames per second (default 24).}

\item{format}{Character. Output format: "mp4", "gif", "webm", or "frames".
If NULL, inferred from file extension.}

\item{backend}{Character. Backend to use: "ffmpeg", "av", or "auto".}

\item{quality}{Integer. Quality level 1-100 (for lossy formats).}

\item{verbose}{Logical. Print progress messages.}
}
\value{
Invisibly returns the output file path.
}
\description{
Saves a video array to a file in various formats.
}
\examples{
video <- array(runif(4 * 16 * 16 * 3), dim = c(4, 16, 16, 3))

# Individual PNG frames need no external encoder.
frame_dir <- file.path(tempdir(), "frames")
save_video(video, frame_dir, format = "frames", verbose = FALSE)
length(list.files(frame_dir, pattern = "[.]png$"))
unlink(frame_dir, recursive = TRUE)

# MP4 and GIF need an ffmpeg binary or the 'av' package. Not shown as
# a live example: both backends hand off to an ffmpeg process that
# inherits this session's stdin, and R CMD check feeds the example
# script to R on stdin, so the encoder eats a byte of the script and
# every later example parses one character short.
# save_video(video, "output.mp4", fps = 24)
# save_video(video, "output.gif", fps = 10)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{load_unet_sdxl_weights}
\alias{load_unet_sdxl_weights}
\title{Load weights from TorchScript SDXL UNet into native SDXL UNet}
\usage{
load_unet_sdxl_weights(native_unet, torchscript_path, verbose = TRUE)
}
\arguments{
\item{native_unet}{Native SDXL UNet module}

\item{torchscript_path}{Path to TorchScript SDXL UNet .pt file}

\item{verbose}{Print loading progress}
}
\value{
The native UNet with loaded weights (invisibly)
}
\description{
Load weights from TorchScript SDXL UNet into native SDXL UNet
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux2_patchify_latents}
\alias{flux2_patchify_latents}
\title{Patchify FLUX.2 latents (2x2 -> channels)}
\usage{
flux2_patchify_latents(latents)
}
\arguments{
\item{latents}{Tensor [B, C, H, W]; H and W must be even.}
}
\value{
Tensor [B, C * 4, H / 2, W / 2].
}
\description{
[B, C, H, W] -> [B, 4C, H/2, W/2], channel order (C, ph, pw).
Reference: Flux2KleinPipeline._patchify_latents.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{gemma3_decoder_layer}
\alias{gemma3_decoder_layer}
\title{Gemma3 Decoder Layer}
\usage{
gemma3_decoder_layer(config, layer_idx = 0L)
}
\arguments{
\item{config}{Model configuration.}

\item{layer_idx}{Integer. Layer index.}
}
\description{
Single transformer block with pre-norm attention and MLP.
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_stage2_distilled_sigmas}
\alias{ltx23_stage2_distilled_sigmas}
\title{Stage-2 distilled sigma schedule (two-stage refinement)}
\usage{
ltx23_stage2_distilled_sigmas()
}
\value{
Numeric vector of length 4.
}
\description{
Stage-2 distilled sigma schedule (two-stage refinement)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{zimage_final_layer}
\alias{zimage_final_layer}
\title{Z-Image final layer}
\usage{
zimage_final_layer(hidden_size, out_channels)
}
\arguments{
\item{hidden_size}{Integer. Model width.}

\item{out_channels}{Integer. Patch output dim
(patch^2 * f_patch * latent channels).}
}
\value{
Module whose forward(x, c) returns the token-to-patch
  projection [B, S, out_channels], ready for unpatchifying into a
  latent.
}
\description{
Parameterless LayerNorm scaled by 1 + adaLN(c) (scale only, no
shift), then the token-to-patch projection.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{get_pad_id}
\alias{get_pad_id}
\title{Get padding token ID}
\usage{
get_pad_id(tokenizer)
}
\description{
Get padding token ID
}
\keyword{internal}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flux_pack_latents}
\alias{flux_pack_latents}
\title{Pack FLUX latents into a patch sequence}
\usage{
flux_pack_latents(latents)
}
\arguments{
\item{latents}{Tensor of shape [B, C, H, W]; H and W must be even.}
}
\value{
Tensor of shape [B, (H/2) * (W/2), C * 4].
}
\description{
Packs a [B, C, H, W] latent into 2x2 patches, giving a sequence
[B, (H/2) * (W/2), C * 4]. Reference: FluxPipeline._pack_latents.
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{ltx23_tail_latents}
\alias{ltx23_tail_latents}
\title{Slice the trailing latent frames of a generation for chaining}
\usage{
ltx23_tail_latents(result, k = 2L, latent_shape = NULL)
}
\arguments{
\item{result}{A \code{\link{txt2vid_ltx2}} result list (uses its
\code{latents} and \code{latent_shape}), or the packed latents
tensor [1, S, 128] itself (then \code{latent_shape} is
required).}

\item{k}{Integer. Trailing latent frames to keep (default 2 = the
standard 9-pixel-frame conditioning prefix).}

\item{latent_shape}{Integer vector c(frames, height, width) of the
latent geometry; only needed when \code{result} is a raw tensor.}
}
\value{
Normalized latents [1, 128, k, H', W'] (float32), ready
  for \code{txt2vid_ltx2(condition_latents = )}.
}
\description{
Cuts the last \code{k} latent frames out of a result's video
latents, in the [1, 128, k, H', W'] layout that
\code{txt2vid_ltx2(condition_latents = )} consumes, so one chunk
can seed the next without leaving latent space: no decode, no
re-encode, no video round-trip.
}
\details{
Semantics caveat: a latent frame sliced from inside a sequence
represents 8 pixel frames, while a fresh VAE encode of a k-frame
tail represents 1 + 8(k - 1) pixel frames with its first latent in
first-frame form. The frozen prefix the next generation sees is
therefore not identical to the pixel-path prefix; compare both on
real content before relying on latent-only joins.

}

% tinyrox says don't edit this manually, but it can't stop you!
\name{filename_from_prompt}
\alias{filename_from_prompt}
\title{Generate a filename from a prompt}
\usage{
filename_from_prompt(prompt, datetime = TRUE)
}
\arguments{
\item{prompt}{A character string representing the prompt.}

\item{datetime}{Logical indicating whether to prepend the current date and time to the filename. Default is TRUE.}
}
\value{
A character string representing the generated filename.
}
\description{
This function generates a filename from a prompt by removing all non-alphanumeric characters and replacing them with underscores. The filename is limited to 50 characters. If `datetime` is set to TRUE, the current date and time are prepended to the filename.
}
\examples{
filename_from_prompt("A beautiful sunset over the mountains")
filename_from_prompt("A beautiful sunset over the mountains", datetime = FALSE)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{flowmatch_calculate_shift}
\alias{flowmatch_calculate_shift}
\title{Calculate shift for dynamic shifting}
\usage{
flowmatch_calculate_shift(seq_len, base_seq_len = 256L, max_seq_len = 4096L,
                          base_shift = 0.5, max_shift = 1.15)
}
\arguments{
\item{seq_len}{Integer. The sequence length (num_patches).}

\item{base_seq_len}{Integer. Base sequence length. Default: 256}

\item{max_seq_len}{Integer. Maximum sequence length. Default: 4096}

\item{base_shift}{Numeric. Base shift value. Default: 0.5}

\item{max_shift}{Numeric. Maximum shift value. Default: 1.15}
}
\value{
Numeric. The computed shift value (mu).
}
\description{
Computes the shift parameter (mu) based on sequence length for
resolution-dependent timestep shifting.
}
